//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// makepsilons.cc

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <math.h>

int dim;
int correct [] = {0,1,2,3,4,5,6,7,8,9,10};

// timestamp outputs a time stamp to stdout

void timestamp (char *comm) {
  time_t now;
  char *s;
  now = time(NULL);
  s = ctime(&now);
  s[24] = '\0';
  fprintf(stderr, "%s on %s\n", comm, s); }

void die (char* err) {
  char str [256];
  timestamp("!");
  strcpy(str, "Error *** ");
  strcat(str, err);
  strcat(str, " *** \n");
  cerr << str;
#ifdef unixsys
  fflush(NULL);
#endif
  abort(); }

int sig(int* lis) {
  int li[11];
  int n, k;
  for (int i = 0; i < dim; i++)
    li[i] = lis[i];
  for (int i = 0; i < dim; i++) {
    n = 0;
    for (int j = 0; j < dim; j++)
      if(li[j] == i) {
	n = n + 1;}
    if (n > 1) {
      return 0;}}
  n = 1;
  for (int i = 0; i < dim; i++) {
    k = -1;
    for (int j = 0; j < dim; j++)
      if(li[j] == i) {
	k = j;}
    if(k < 0) {
      return 0;}
    if((abs(k - i) % 2) == 1)
      {n = -n;}
    for (int j = k; j >= i + 1; j--)
      li[j] = li[j - 1];
    li[i] = i;}
  if (memcmp(li, correct, dim*sizeof(int)) != 0)
    die("sort failed");
  return n; }

// factorial computes n!

long long factorial (int n) {
  long long fact;
  fact = 1;
  for (int i = 2; i <= n; i++)
    fact *= i;
  return fact; }

// here begins the main program

int main (int argcount, char **args) {

  int which, num[11], inds[11];
  
  long long indnum = 0;
  long long ind;
  long long thisind;
  long long minusone = 1;

  for (int i = 0; i < 50; i++)
    minusone *= 2;
 
  if (argcount < 2) {
    fprintf(stderr, "Usage is makepsilons <dimension>\n");
    exit(0);}

  sscanf(args[1], "%d", &dim);

  if ((dim < 2) || (dim > 11)) {
    fprintf(stderr, "dimension must > 1 and <= 11\n");
    exit(0);}
  
  setvbuf(stdout, NULL, _IOLBF, 0);
  
  timestamp("makepsilons begun");

  indnum = factorial(dim);
  for (int i = 0; i < dim; i++)
    inds[i] = i;                            // start at 0,1,2,3,...
  inds[dim - 1] -= 1;
  ind = 0;
  while (ind < indnum) {
    which = dim - 1;                        // increment from last index forward
    while (which >= 0) {
      inds[which] += 1;
      if (inds[which] >= dim) {             // try next index
	inds[which] = 0;
	which -= 1;
	continue;}
      break;}
    memset(num, 0, dim * sizeof(int));
    for (int i = 0; i < dim; i++)          // count number of occurances of each index
      num[inds[i]] += 1;
    for (which = 0; which < dim; which++)
      if (num[which] > 1)                  // if any are > 1 we need to try again
	break;
    if ((which < dim) && (num[which] > 1))
      continue;
    thisind = 0;                           // we have one !
    for (int i = 0; i < dim; i++)
      thisind += (((long long) inds[i]) << (4 * i));
    if (sig(inds) < 0)
      thisind += minusone;
    fwrite(&thisind, sizeof(thisind), 1, stdout);
    // printf("%016llx\n", thisind);
    ind += 1;}

  timestamp("makepsilons complete");
}
