//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// preps4.cc

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <iostream.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <math.h>

#define unixsys 1

long sec;
long micro;
long t0sec;
long t0micro;

int rank, last;

char fn[21] = "/home/ptahusr/eps.xx";

long long eps;

FILE *fd = NULL;

long get_vsize () {
  int c;
  int v = 0;
#ifdef unixsys
  FILE* f = fopen("/proc/self/stat", "r");
  if(f == NULL) {
    perror("opening /proc/self/stat");
    return 0;}
  c = fscanf(f, 
	     "%*d %*s %*c %*d %*d %*d %*d %*d %*u %*u %*u %*u %*u %*d %*d %*d %*d %*d %*d %*u %*u %*d %u", 
	     &v);
  if (c != 1)
    cerr << "fscanf error";
  fclose(f);
#endif
  return (long) v;
 }

void time_stamp () {
#ifdef unixsys
  long siz;
  char str [256];
  timeval now;
  long tcsec = t0sec;
  long tcmicro = t0micro;
  gettimeofday(&now, NULL);
  if (sec == -1) {
    sec = now.tv_sec;
    t0sec = sec;
    micro = now.tv_usec;
    t0micro = micro;
    return;}
  if ((now.tv_usec - tcmicro) < 0) {
    tcmicro -= 1000000;
    tcsec += 1;}
  siz = 1 + get_vsize() / (1024 * 1024);
  sprintf (str, "Total elapsed time is %ld.%06ld; size = %ld Mb.\n", 
	   (now.tv_sec - tcsec), (now.tv_usec - tcmicro), siz);
  cerr << str;
  fflush(NULL);
  t0sec = now.tv_sec;
  t0micro = now.tv_usec;
#endif
 }

void time_stamp (char* comment) {
#ifdef unixsys
  int len;
  long siz;
  timeval now;
  char str [256];
  gettimeofday(&now, NULL);
  if (sec == -1) {
    sec = now.tv_sec;
    t0sec = sec;
    micro = now.tv_usec;
    t0micro = micro;
    return;}
  if ((now.tv_usec - micro) < 0) {
    micro -= 1000000;
    sec += 1;}
  siz = 1 + get_vsize() / (1024 * 1024);
  strcpy(str, comment);
  len = strlen(str);
  sprintf(str + len, "; elapsed time is %ld.%06ld; size = %ld Mb.\n", 
	  (now.tv_sec - sec), (now.tv_usec - micro), siz);
  cerr << str;
  fflush(NULL);
  sec = now.tv_sec;
  micro = now.tv_usec;
#endif
 }

void die (char* err) {
  char str [256];
  time_stamp();
  strcpy(str, "Error *** ");
  strcat(str, err);
  strcat(str, " *** \n");
  cerr << str;
#ifdef unixsys
  fflush(NULL);
#endif
  abort(); }

void die (char* err, int code) {
  char str [256];
  time_stamp();
  strcpy(str, "Error *** ");
  strcat(str, err);
  sprintf(str + strlen(str), ", id = %d (%x) *** \n", code, code);
  cerr << str;
#ifdef unixsys
  fflush(NULL);
#endif
  abort(); }

void epsilon_start (int dim) {
  int slen;
  if ((dim < 2) || (dim > 11))
    die("invalid epsilon dimension");
  rank = dim;
  last = -1;
  slen = sprintf(&(fn[18]), "%d", dim) + 18;
  if ((slen < 19) || (slen > 20))
    die("creating epsilon filename");
  fd = fopen(fn, "r");
  if (fd == NULL) {
    perror("opening epsilon file");
    die ("opening epsilon file", errno);} }

void epsilon_stop () {
  int rc;
  last = -1;
  rc = fclose(fd);
  if (rc != 0) {
    perror("closing epsilon file");
    die ("closing epsilon file", errno);}
  fd = NULL; }

int epsilon_index (int which, int ind) {
  int rc;
  long offset;
  if (which != last) {
    offset = which * sizeof(long long);
    rc = fseek(fd, offset, SEEK_SET);
    if (rc != 0) {
      perror("seeking epsilon file");
      die ("seeking epsilon file", errno);}
    rc = fread(&eps, sizeof(long long), 1, fd);
    if (rc != 1) {
      perror("reading epsilon file");
      die ("reading epsilon file", errno);}
    last = which;}
  return ((eps >> (4 * ind)) & 15); }

int epsilon_sign (int which) {
  int rc;
  long offset;
  if (which != last) {
    offset = which * sizeof(long long);
    rc = fseek(fd, offset, SEEK_SET);
    if (rc != 0) {
      perror("seeking epsilon file");
      die ("seeking epsilon file", errno);}
    rc = fread(&eps, sizeof(long long), 1, fd);
    if (rc != 1) {
      perror("reading epsilon file");
      die ("reading epsilon file", errno);}
    last = which;}
  rc = (eps >> 50);
  if (rc == 0)
    return 1;
  else
    return -1; }

int main () {
  int sgn, inds[4];
  epsilon_start(4);
  for (int i = 0; i < 24; i++) {
    for (int j = 0; j < 4; j++)
      inds[j] = epsilon_index(i,j);
    sgn = epsilon_sign(i);
    printf("%d, %d, %d, %d = %d\n",inds[0],inds[1],inds[2],inds[3],sgn);} 
  epsilon_stop();
  return 0; }

      
