//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// main.cc

#include <ptah.h>
#include <element.h>
#include <hash.h>
#include <ntuple.h>
#include <main.h>
#include <metric.h>
#include <d11sugra.h>

bool trueparm[5] = {true, true, true, true, true}; 
bool falseparm[5] = {false, false, false, false, false}; 
bool antiparm[5] = {true, false, true, false, false}; 
bool invielparm[2] = {false, true};
bool omegaparm[3] = {false, true, true};
bool omegantiparm[3] = {false, true, false};

int fpp[number_of_variables] = {0};

int main () {

  parallel = false;

  function_build(number_of_functions + number_of_ansatze, fnumtrm, fnumfct, fpwr, fco);

  g = new ntuple(2, falseparm, falseparm, trueparm, falseparm);
  g->tensor_build(gcnt, gind, gnumtrm, gnumfct, gpwr, gco);
  g->metric = true;

  g_inverse = new ntuple(2, trueparm, falseparm, trueparm, falseparm);
  g_inverse->tensor_build(gicnt, giind, ginumtrm, ginumfct, gipwr, gico);
  g_inverse->metric = true;

  inverse_sqrt_det_g = new sum;
  inverse_sqrt_det_g->sum_build(1, dnumfct, dpwr, dco);

  christoffel = new connection();

  reimann = new curvature();

  reimann->zero();
  delete reimann;
  traceless_ricci->zero();
  delete traceless_ricci;
  g_curv->zero();
  delete g_curv;
  weyl->zero();
  delete weyl;
  einstein_curv->zero();
  delete einstein_curv;

  time_stamp("Starting computation");

  viel = new ntuple(2, antiparm, falseparm, falseparm, falseparm);      // indices are upper, lower (lorentz, st)
  viel->tensor_build(vielcnt, vielind, vielnumtrm, vielnumfct, vielpwr, vielco);

  viel->display("viel");

  vielinv = new ntuple(2, invielparm, falseparm, falseparm, falseparm); // indices are lower, upper (lorentz, st)
  vielinv->tensor_build(vielinvcnt, vielinvind, vielinvnumtrm, vielinvnumfct, vielinvpwr, vielinvco);

  vielinv->display("vielinv");

  time_stamp("Starting spin connection");

  omega = new ntuple(3, omegaparm, omegantiparm, falseparm, falseparm);  // indices are low, up, up (st, lor, lor)
  omega->tensor_build(omegacnt, omegaind, omeganumtrm, omeganumfct, omegapwr, omegaco);

  omega->display("omega");

  susyparm = new ntuple(1, falseparm, falseparm, falseparm, trueparm);   // index is sp
  susyparm->tensor_build(susyparmcnt, susyparmind, susyparmnumtrm, susyparmnumfct, susyparmpwr, susyparmco);

  susyparm->display("sp");

  time_stamp("Starting gauge field");

  fsuuuu = new ntuple(4, trueparm, antiparm, falseparm, falseparm);
  fsuuuu->tensor_build(fsuuuucnt, fsuuuuind, fsuuuunumtrm, fsuuuunumfct, fsuuuupwr, fsuuuuco);

  fsuuuu->display("fsuuuu");

  fs = new ntuple(4, falseparm, antiparm, falseparm, falseparm);
  fs->tensor_build(fscnt, fsind, fsnumtrm, fsnumfct, fspwr, fsco);

  fs->display("fsllll");

  time_stamp("Tensor initialization completed");

  gravitino = d11gravitino();
  gravitino->display("gravitino");
  gravitino->zero();
  delete gravitino;

  time_stamp("Gravitino variations completed");

  gaugefield = d11gauge();
  gaugefield->display("gauge");
  gaugefield->zero();
  delete gaugefield;

  time_stamp("Gauge field equations completed");

  graviton = d11graviton();
  graviton->display("graviton");
  graviton->zero();
  delete graviton;

  time_stamp("Graviton field equations completed");
  entry_stats();
  return 0; }
