//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// gams2txt.cc

#include <stdlib.h>
#include <string.h>
#include <iostream.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "gammas.h"

#define unixsys

char fni[33] = "/home/ptahusr/gammaxx.abcdefghij";

int last, thisorder;

long long gam;

long sec = -1;
long micro;
long t0sec;
long t0micro;

FILE *fdi = NULL;

long get_wsize () {
  int c;
  int v = 0;
#ifdef unixsys
  FILE* f = fopen("/proc/self/stat", "r");
  if(f == NULL) {
    perror("opening /proc/self/stat");
    return 0;}
  c = fscanf(f, 
	     "%*d %*s %*c %*d %*d %*d %*d %*d %*u %*u %*u %*u %*u %*d %*d %*d %*d %*d %*d %*u %*u %*d %*u %u", 
	     &v);
  if (c != 1)
    cerr << "fscanf error";
  fclose(f);
#endif
  return (long) v;
 }

extern void time_stamp () {
#ifdef unixsys
  long siz;
  char str [256];
  timeval now;
  long tcsec = t0sec;
  long tcmicro = t0micro;
  gettimeofday(&now, NULL);
  if (sec == -1) {
    sec = now.tv_sec;
    t0sec = sec;
    micro = now.tv_usec;
    t0micro = micro;
    return;}
  if ((now.tv_usec - tcmicro) < 0) {
    tcmicro -= 1000000;
    tcsec += 1;}
  siz = 1 + get_wsize() / (1024 * 1024);
  sprintf (str, "Total elapsed time is %ld.%06ld; size = %ld Mb.\n", 
	   (now.tv_sec - tcsec), (now.tv_usec - tcmicro), siz);
  cerr << str;
  fflush(NULL);
  t0sec = now.tv_sec;
  t0micro = now.tv_usec;
#endif
 }

extern void time_stamp (char* comment) {
#ifdef unixsys
  int len;
  long siz;
  timeval now;
  char str [256];
  gettimeofday(&now, NULL);
  if (sec == -1) {
    sec = now.tv_sec;
    t0sec = sec;
    micro = now.tv_usec;
    t0micro = micro;
    return;}
  if ((now.tv_usec - micro) < 0) {
    micro -= 1000000;
    sec += 1;}
  siz = 1 + get_wsize() / (1024 * 1024);
  strcpy(str, comment);
  len = strlen(str);
  sprintf(str + len, "; elapsed time is %ld.%06ld; size = %ld Mb.\n", 
	  (now.tv_sec - sec), (now.tv_usec - micro), siz);
  cerr << str;
  fflush(NULL);
  sec = now.tv_sec;
  micro = now.tv_usec;
#endif
 }

extern void die (char* err) {
  char str [256];
  time_stamp();
  strcpy(str, "Error *** ");
  strcat(str, err);
  strcat(str, " *** \n");
  cerr << str;
#ifdef unixsys
  fflush(NULL);
#endif
  abort(); }

extern void die (char* err, int code) {
  char str [256];
  time_stamp();
  strcpy(str, "Error *** ");
  strcat(str, err);
  sprintf(str + strlen(str), ", id = %d (%x) *** \n", code, code);
  cerr << str;
#ifdef unixsys
  fflush(NULL);
#endif
  abort(); }

void gamma_start (int order) {
  int slen;
  if ((order < 1) || (order > dimension))
    die("invalid gamma order");
  last = -1;
  thisorder = order;
  slen = sprintf(&(fni[19]), "%2d", dimension);
  if (slen != 2)
    die("creating gamma input filename");
  fni[21] = '.';
  fni[22 + order] = 0x00;
  fdi = fopen(fni, "r");
  if (fdi == NULL) {
    perror("opening gamma input file");
    die ("opening gamma input file", errno);} }

void gamma_stop () {
  int rc;
  last = -1;
  rc = fclose(fdi);
  if (rc != 0) {
    perror("closing gamma input file");
    die ("closing gamma input file", errno);}
  fdi = NULL; }

int gamma_index (int which, int ind) {
  int rc;
  long offset;
  if (which != last) {
    offset = which * sizeof(long long);
    rc = fseek(fdi, offset, SEEK_SET);
    if (rc != 0) {
      perror("seeking gamma file");
      die ("seeking gamma file", errno);}
    rc = fread(&gam, sizeof(long long), 1, fdi);
    if (rc != 1) {
      perror("reading gamma file");
      die ("reading gamma file", errno);}
    last = which;}
  if (ind >= thisorder)
    return ((gam >> (8 * (ind - thisorder) + 4 * thisorder)) & 255);
  else
    return ((gam >> (4 * ind)) & 15); }

int gamma_sign (int which) {
  int rc;
  long offset;
  if (which != last) {
    offset = which * sizeof(long long);
    rc = fseek(fdi, offset, SEEK_SET);
    if (rc != 0) {
      perror("seeking gamma file");
      die ("seeking gamma file", errno);}
    rc = fread(&gam, sizeof(long long), 1, fdi);
    if (rc != 1) {
      perror("reading gamma file");
      die ("reading gamma file", errno);}
    last = which;}
  rc = ((gam >> 50) & 1);
  if (rc == 0)
    return 1;
  else
    return -1; }

int gamma_i (int which) {
  int rc;
  long offset;
  if (which != last) {
    offset = which * sizeof(long long);
    rc = fseek(fdi, offset, SEEK_SET);
    if (rc != 0) {
      perror("seeking gamma file");
      die ("seeking gamma file", errno);}
    rc = fread(&gam, sizeof(long long), 1, fdi);
    if (rc != 1) {
      perror("reading gamma file");
      die ("reading gamma file", errno);}
    last = which;}
  rc = (gam >> 51);
  if (rc == 0)
    return 0;
  else
    return 1; }

int main (int argcount, char **args) {

  int order, inno;

  int inds[dimension + 2];

  if (argcount < 2) {
    fprintf(stderr, "Usage is gams2txt <number of antisymmetric indices>\n");
    exit(0);}

  sscanf(args[1], "%d", &order);

  if ((order < 1) || (order >= dimension)) {
    fprintf(stderr, "order must > 0 and < dimension\n");
    exit(0);}

  time_stamp();
  
  setvbuf(stdout, NULL, _IOLBF, 0);

  inno = dimension;
  for (int i = 0; i < (order - 1); i++)
    inno *= (dimension - i -1);
  inno *= spinor_dimension;

  gamma_start(order);
  for (int i = 0; i < inno; i++) {
    for (int j = 0; j < (order + 2); j++)
      if (j < order) {
	inds[j] = gamma_index(i, j);
	if (inds[j] == 0)
	  inds[j] = dimension;}
      else
	inds[j] = gamma_index(i, j) + 1;
    cout << "gammam[[";
    for (int j = 0; j < (order + 1); j++)
      cout << inds[j] << ",";
    cout << inds[order + 1] << "]] = ";
    switch (4 * gamma_i(i) + gamma_sign(i) + 1) {
      case 0:
	cout << "-1;\n";
	break;
      case 2:
	cout << "1;\n";
	break;
      case 4:
	cout << "-I;\n";
	break;
      case 6:
	cout << "I;\n";
	break;
    default:
      die("switch error");}}

  gamma_stop();


  time_stamp("gams2txt completed");
  return 0; }
