//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// makegammas.cc

// currently limited to antisymmetric gamma products up to 8 indices, dim < 16

#include <stdlib.h>
#include <string.h>
#include <iostream.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "gammas.h"

#define unixsys

char fni[33] = "/home/ptahusr/gammaxx.abcdefghij";
char fn1[24] = "/home/ptahusr/gammaxx.a";

int last, thisorder;

long long gam, gam1;

long long gammaa[dimension * spinor_dimension];

long sec = -1;
long micro;
long t0sec;
long t0micro;

FILE *fdi = NULL, *fd1 = NULL;

long get_wsize () {
  int c;
  int v = 0;
#ifdef unixsys
  FILE* f = fopen("/proc/self/stat", "r");
  if(f == NULL) {
    perror("opening /proc/self/stat");
    return 0;}
  c = fscanf(f, 
	     "%*d %*s %*c %*d %*d %*d %*d %*d %*u %*u %*u %*u %*u %*d %*d %*d %*d %*d %*d %*u %*u %*d %*u %u", 
	     &v);
  if (c != 1)
    cerr << "fscanf error";
  fclose(f);
#endif
  return (long) v;
 }

extern void time_stamp () {
#ifdef unixsys
  long siz;
  char str [256];
  timeval now;
  long tcsec = t0sec;
  long tcmicro = t0micro;
  gettimeofday(&now, NULL);
  if (sec == -1) {
    sec = now.tv_sec;
    t0sec = sec;
    micro = now.tv_usec;
    t0micro = micro;
    return;}
  if ((now.tv_usec - tcmicro) < 0) {
    tcmicro -= 1000000;
    tcsec += 1;}
  siz = 1 + get_wsize() / (1024 * 1024);
  sprintf (str, "Total elapsed time is %ld.%06ld; size = %ld Mb.\n", 
	   (now.tv_sec - tcsec), (now.tv_usec - tcmicro), siz);
  cerr << str;
  fflush(NULL);
  t0sec = now.tv_sec;
  t0micro = now.tv_usec;
#endif
 }

extern void time_stamp (char* comment) {
#ifdef unixsys
  int len;
  long siz;
  timeval now;
  char str [256];
  gettimeofday(&now, NULL);
  if (sec == -1) {
    sec = now.tv_sec;
    t0sec = sec;
    micro = now.tv_usec;
    t0micro = micro;
    return;}
  if ((now.tv_usec - micro) < 0) {
    micro -= 1000000;
    sec += 1;}
  siz = 1 + get_wsize() / (1024 * 1024);
  strcpy(str, comment);
  len = strlen(str);
  sprintf(str + len, "; elapsed time is %ld.%06ld; size = %ld Mb.\n", 
	  (now.tv_sec - sec), (now.tv_usec - micro), siz);
  cerr << str;
  fflush(NULL);
  sec = now.tv_sec;
  micro = now.tv_usec;
#endif
 }

extern void die (char* err) {
  char str [256];
  time_stamp();
  strcpy(str, "Error *** ");
  strcat(str, err);
  strcat(str, " *** \n");
  cerr << str;
#ifdef unixsys
  fflush(NULL);
#endif
  abort(); }

extern void die (char* err, int code) {
  char str [256];
  time_stamp();
  strcpy(str, "Error *** ");
  strcat(str, err);
  sprintf(str + strlen(str), ", id = %d (%x) *** \n", code, code);
  cerr << str;
#ifdef unixsys
  fflush(NULL);
#endif
  abort(); }

void gamma_start (int order) {
  int slen, rc;
  if ((order < 1) || (order > dimension))
    die("invalid gamma order");
  thisorder = order;
  last = -1;
  slen = sprintf(&(fni[19]), "%2d", dimension);
  if (slen != 2)
    die("creating gamma input filename");
  fni[21] = '.';
  fni[22 + order] = 0x00;
  fdi = fopen(fni, "r");
  if (fdi == NULL) {
    perror("opening gamma input file");
    die ("opening gamma input file", errno);}
  if (order == 1)
    fd1 = fdi;
  else {
    slen = sprintf(&(fn1[19]), "%2d", dimension);
    if (slen != 2)
      die("creating gamma.a filename");
    fn1[21] = '.';
    fd1 = fopen(fn1, "r");
    if (fd1 == NULL) {
      perror("opening gamma.a file");
      die ("opening gamma.a file", errno);}}
  for (int i = 0; i < dimension * spinor_dimension; i++) {
    rc = fread(&(gammaa[i]), sizeof(long long), 1, fd1);
    if (rc != 1) {
      perror("reading gamma.a file");
      die ("reading gamma.a file", errno);}}
  if (order != 1) {
    rc = fclose(fd1);
    if (rc != 0) {
      perror("closing gamma.a file");
      die ("closing gamma.a file", errno);}}
  fd1 = NULL; }

void gamma_stop () {
  int rc;
  last = -1;
  rc = fclose(fdi);
  if (rc != 0) {
    perror("closing gamma input file");
    die ("closing gamma input file", errno);}
  fdi = NULL; }

int gamma_index (int which, int ind) {
  int rc;
  long offset;
  if (which != last) {
    offset = which * sizeof(long long);
    rc = fseek(fdi, offset, SEEK_SET);
    if (rc != 0) {
      perror("seeking gamma file");
      die ("seeking gamma file", errno);}
    rc = fread(&gam, sizeof(long long), 1, fdi);
    if (rc != 1) {
      perror("reading gamma file");
      die ("reading gamma file", errno);}
    last = which;}
  if (ind >= thisorder)
    return ((gam >> (8 * (ind - thisorder) + 4 * thisorder)) & 255);
  else
    return ((gam >> (4 * ind)) & 15); }

int gamma_sign (int which) {
  int rc;
  long offset;
  if (which != last) {
    offset = which * sizeof(long long);
    rc = fseek(fdi, offset, SEEK_SET);
    if (rc != 0) {
      perror("seeking gamma file");
      die ("seeking gamma file", errno);}
    rc = fread(&gam, sizeof(long long), 1, fdi);
    if (rc != 1) {
      perror("reading gamma file");
      die ("reading gamma file", errno);}
    last = which;}
  rc = ((gam >> 50) & 1);
  if (rc == 0)
    return 1;
  else
    return -1; }

int gamma_i (int which) {
  int rc;
  long offset;
  if (which != last) {
    offset = which * sizeof(long long);
    rc = fseek(fdi, offset, SEEK_SET);
    if (rc != 0) {
      perror("seeking gamma file");
      die ("seeking gamma file", errno);}
    rc = fread(&gam, sizeof(long long), 1, fdi);
    if (rc != 1) {
      perror("reading gamma file");
      die ("reading gamma file", errno);}
    last = which;}
  rc = (gam >> 51);
  if (rc == 0)
    return 0;
  else
    return 1; }

int main (int argcount, char **args) {

  int order, inno, tempi, tempc;
  int inds[dimension + 3];

  bool found;

  long long thisgam;
  long long sign_mask = (((long long) 1) << 50);
  long long i_mask = (((long long) 1) << 51);

  time_stamp();

  if (argcount < 1) {
    fprintf(stderr, "Usage is makegammas <number of antisymmetric indices>\n");
    exit(0);}

  sscanf(args[1], "%d", &order);

  if ((order < 1) || (order >= dimension)) {
    fprintf(stderr, "order must > 0 and < dimension\n");
    exit(0);}
  
  setvbuf(stdout, NULL, _IOLBF, 0);

  if (order == 1) {
    for (int i = 0; i < (dimension * spinor_dimension); i++) {
      thisgam = gammaind[3 * i] +
	(gammaind[3 * i + 1] << 4) +
	(gammaind[3 * i + 2] << 12);
      if (gammapwr[2 * i] != -1)
	thisgam += (((long long) 1) << 51);
      if (gammaco[i] < 0)
	thisgam += (((long long) 1) << 50);
      fwrite(&thisgam, sizeof(thisgam), 1, stdout);}
    time_stamp("makegammas complete");
    exit(0);}

  gamma_start(order - 1);

  inno = dimension;
  for (int i = 0; i < (order - 2); i++)
    inno *= (dimension - i - 1);
  inno *= spinor_dimension;

  for (int i = 0; i < inno; i++) {
    for (int k = 0; k < (order - 1); k++)
      inds[k] = gamma_index(i, k);                     // spacetime indices
    inds[order] = gamma_index(i, order - 1);           // leave room for new spacetime index
    inds[order + 1] = gamma_index(i, order);           //   when inserting spinor indices
    for (int j = 0; j < (dimension * spinor_dimension); j++) {
      tempi = (gammaa[j] >> 4) & 255;                  // gamma_a first spinor index
      if (tempi != inds[order + 1])                    // doesn't contribute to matrix product
	continue;
      inds[order - 1] = gammaa[j] & 15;                // insert new spacetime index
      found = false;                                   // check for dupes
      for (int k = 0; k < (order - 1); k++)
	if (inds[k] == inds[order - 1]) {
	  found = true;
	  break;}
      if (found)
	continue;
      inds[order + 1] = (gammaa[j] >> 12) & 255;       // new second spinor index
      thisgam = 0;
      for (int k = 0; k < order; k++)
	thisgam += (((long long) inds[k]) << (4 * k));
      for (int k = 0; k < 2; k++)
	thisgam += (((long long) inds[k + order]) << ((8 * k) + order * 4));
      tempc = gamma_sign(i);
      if ((gammaa[j] & sign_mask) > 0)
	tempc *= -1;
      tempi = gamma_i(i);
      if (((gammaa[j] & i_mask) > 0) && (tempi == 1)) {
	tempc *= -1;
	tempi = 0;}
      else
	tempi += (gammaa[j] >> 51);
      if (tempc < 0)
	thisgam += sign_mask;
      if (tempi > 0)
	thisgam += i_mask;
      fwrite(&thisgam, sizeof(thisgam), 1, stdout);
      inds[order + 1] = gamma_index(i, order);}}                 // put it back!

  gamma_stop();

  time_stamp("makegammas completed");
  return 0; }
