//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// ptah.h

#include <stdlib.h>
#include <string.h>
#include <iostream.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <math.h>
#include <options.h>
#include <params.h>

#ifdef debug
#define element_magic          0x01010101
#define term_magic             0x02020202
#define invariant_term_magic   0x03030303
#define hash_magic             0x04040404
#define sum_magic              0x05050505
#define invariant_magic        0x06060606
#define component_magic        0x07070707
#define ntuple_magic           0x08080808
#define connection_magic       0x09090909
#define curvature_magic        0x0a0a0a0a
#endif

//------------------------------------------------------------------------
// miscellaneous prototypes
//------------------------------------------------------------------------

void time_stamp();                // time_stamp outputs a time stamp to stdout or stderr
void time_stamp(char* comment);
void die(char*);
void die(char*, int);
void to_stderr(char*);
void to_stdout(char*);
void close_std();

int epsilon_index(int, int);
int epsilon_sign(int);
void epsilon_start(int);
void epsilon_stop();

extern bool parallel;
extern bool parent;

extern int pwr[];

extern long sec;
extern long micro;
extern long t0sec;
extern long t0micro;

void set_powers(int num, int *pairs);

void entry_stats();
void clear_entries();
void head_stats();

#ifdef keepstats
#ifdef debug
extern int keep_head_count[];
#endif
#endif

#ifdef tracentry
extern long trace_ff;
extern long trace_tf;
extern long trace_ei;
extern long trace_e;
extern long trace_ke;
extern long trace_ne;
extern long trace_ie;
extern long trace_re;
extern long trace_td;
extern long trace_t;
extern long trace_kt;
extern long trace_nt;
extern long trace_it;
extern long trace_int;
extern long trace_kint;
extern long trace_nint;
extern long trace_iit;
extern long trace_rinl;
extern long trace_hc;
extern long trace_sh;
extern long trace_iah;
extern long trace_rfl;
extern long trace_sit;
extern long trace_srt;
extern long trace_ft;
extern long trace_f;
extern long trace_sa;
extern long trace_smu;
extern long trace_sumu;
extern long trace_nsmu;
extern long trace_nms;
extern long trace_flr;
extern long trace_ntr;
extern long trace_szq;
extern long trace_fit;
extern long trace_irt;
extern long trace_ia;
extern long trace_rfi;
extern long trace_co;
extern long trace_co2;
extern long trace_kco;
extern long trace_nco;
extern long trace_ic;
extern long trace_rc;
extern long trace_fc;
extern long trace_fi;
extern long trace_csu;
extern long trace_nst;
extern long trace_nss;
extern long trace_na;
extern long trace_n1;
extern long trace_n;
extern long trace_nal;
extern long trace_kn;
extern long trace_ni;
extern long trace_mfil;
extern long trace_msl;
extern long trace_ncn;
extern long trace_nci;
extern long trace_nz;
extern long trace_nc;
extern long trace_nan;
extern long trace_fil;
extern long trace_ncl;
extern long trace_nr;
extern long trace_nl;
extern long trace_ncd;
extern long trace_nsc;
extern long trace_nnc;
extern long trace_nd;
extern long trace_ns;
extern long trace_nsi;
extern long trace_ndi;
extern long trace_gb;
extern long trace_ge;
extern long trace_gin;
extern long trace_gs;
extern long trace_gi;
extern long trace_gam;
extern long trace_eb;
extern long trace_ee;
extern long trace_epi;
extern long trace_es;
extern long trace_cc;
extern long trace_ccu;
extern long trace_s;
extern long trace_ks;
extern long trace_sz;
extern long trace_sm;
extern long trace_sd;
extern long trace_ss;
extern long trace_sn;
extern long trace_i;
extern long trace_ki;
extern long trace_iz;
extern long trace_im;
extern long trace_is;
extern long trace_cd;
extern long trace_ps;
extern long trace_pst;
extern long trace_tin;
extern long trace_ti;
extern long trace_pi;
extern long trace_ip;
extern long trace_sp;
extern long trace_si;
extern long trace_spa;
extern long trace_as;
extern long trace_ci;
extern long trace_spar;
extern long trace_sps;
extern long trace_cspar;
extern long trace_tspar;
extern long trace_com;
extern long trace_npn;
extern long trace_nps;
extern long trace_nfp;
extern long trace_npd;
#endif
