(* sugra - functions for supergravity computations *)
(* PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 

 e-mail address - kenneth.koehler@uc.edu

 This program is free software; you can redistribute it and/or modify it under 
 the terms of the GNU General Public License as published by the Free Software 
 Foundation; either version 2 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 PURPOSE. See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public License along with this 
 program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 Cambridge, MA 02139, USA. 
 
 sugra.nb *)
Off[General::spell, General::spell1, Syntax::com];
Off[Power::infy];
$HistoryLength = 0;
simpler[expr_] := Factor[PowerExpand[expr]];
timestamp[] := Block[{now},
		(now = Date[];
			Print["Time is ", now[[4]], ":", now[[5]], ":", now[[6]],
				" on ", now[[2]], "/", now[[3]], "/", now[[1]], "."];)];
backup[] := Block[{},
		(timestamp[];
			NotebookSave[EvaluationNotebook[]];
        		Run[sync];)];
spacetime[xx_List, gg_List] := Block[{ginv, Christoffel},
		(dim = Length[xx];
			ginv = simpler[Inverse[gg]];
			currentst = {xx, gg, ginv};
			Christoffel = connection[];
			Return[currentst];)];
fastst[xx_List, gg_List] := Block[{ginv, Christoffel},
		(dim = Length[xx];
			ginv = simpler[Inverse[gg]];
			currentst = {xx, gg, ginv};
			Return[currentst];)];
lorentzmetric[nn_] := DiagonalMatrix[Flatten[{Table[1, {ii, nn - 1}], -1}]];
components[ten_List] := Block[{rank, dummy},
		(rank = TensorRank[ten];
			If[rank == 1,
				Do[If[ten[[ii]] == 0, dummy = 0,
						Print["[", ii, "] = ", ten[[ii]]],
						Print["[", ii, "] = ", ten[[ii]]]], {ii, Length[ten]}],];
			If[rank == 2,
				Do[Do[If[ten[[ii, jj]] == 0, dummy = 0,
						Print["[", ii, ",", jj, "] = ", ten[[ii, jj]]],
						Print["[", ii, ",", jj, "] = ", ten[[ii, jj]]]],
					{jj, Length[ten[[ii]]]}], {ii, Length[ten]}],];
			If[rank == 3,
				Do[Do[Do[If[ten[[ii, jj, kk]] == 0, dummy = 0,
						Print["[", ii, ",", jj, ",", kk, "] = ", ten[[ii, jj, kk]]],
						Print["[", ii, ",", jj, ",", kk, "] = ", ten[[ii, jj, kk]]]],
					{kk, Length[ten[[ii, jj]]]}], {jj, Length[ten[[ii]]]}], {ii, 
              Length[ten]}],];
			If[rank == 4,
				Do[Do[Do[Do[If[ten[[ii, jj, kk, ll]] == 0, dummy = 0,
						Print["[", ii, ",", jj, ",", kk, ",", ll, "] = ", 
                      ten[[ii, jj, kk, ll]]],
						Print["[", ii, ",", jj, ",", kk, ",", ll, "] = ", 
                      ten[[ii, jj, kk, ll]]]],
					{ll, Length[ten[[ii, jj, kk]]]}], {kk, Length[ten[[ii, jj]]]}],
						{jj, Length[ten[[ii]]]}], {ii, Length[ten]}],];
			If[rank == 5,
				Do[Do[Do[Do[Do[If[ten[[ii, jj, kk, ll, mm]] == 0, dummy = 0,
						Print["[", ii, ",", jj, ",", kk, ",", ll, ",", mm, "] = ",
											ten[[ii, jj, kk, ll, mm]]],
						Print["[", ii, ",", jj, ",", kk, ",", ll, ",", mm, "] = ",
											ten[[ii, jj, kk, ll, mm]]]], {mm, Length[ten[[ii, jj, kk, ll]]]}],
					{ll, Length[ten[[ii, jj, kk]]]}], {kk, Length[ten[[ii, jj]]]}],
						{jj, Length[ten[[ii]]]}], {ii, Length[ten]}],];)];
tenrank[ten_List] := Block[{dim, rank},
		(dim = Length[currentst[[1]]];
			rank = 0;
			If[Length[ten] == dim, (rank = 1;
			If[(Head[ten[[1]]] === List) && (Length[ten[[1]]] == dim), (rank = 2;
			If[(Head[ten[[1, 1]]] === List) && (Length[ten[[1, 1]]] == dim), (rank = 
                      3;
			If[(Head[ten[[1, 1, 1]]] === List) && (Length[ten[[1, 1, 1]]] == 
                            dim), (rank = 4;
			If[(Head[ten[[1, 1, 1, 1]]] === List) && (Length[ten[[1, 1, 1, 1]]] == 
                                dim), rank = 5,];),];),];),];),];
			Return[rank];)];
raise[ten_List, nn_] := Block[{dim, ginv, rank, tenr},
		(If[Length[currentst] < 3,
				(Print["Evaluate spacetime[x,g] first"];
					Return[];),];
			rank = tenrank[ten];
			If[rank < nn,
				(Print["Error - insufficient rank"];
					Return[];),];
			ginv = currentst[[3]];
			dim = Length[ginv[[1]]];
			If[Length[ten] == dim, ,
				(Print["Error - tensor is incorrect dimension"];
				Return[];)];
			If[rank == 1,
				tenr = Table[simpler[Sum[ginv[[ii, jj]]ten[[jj]],
								{jj, dim}]], {ii, dim}],];
			If[rank == 2,
				tenr = Switch[nn,
						1, Table[Table[simpler[Sum[ginv[[ii, kk]]ten[[kk, jj]],
								{kk, dim}]], {jj, dim}], {ii, dim}],
						2, Table[Table[simpler[Sum[ginv[[jj, kk]]ten[[ii, kk]],
								{kk, dim}]], {jj, dim}], {ii, dim}]],];
			If[rank == 3,
				tenr = Switch[nn,
						1, Table[Table[Table[simpler[Sum[ginv[[ii, ll]]ten[[ll, jj, kk]],
								{ll, dim}]], {kk, dim}], {jj, dim}], {ii, dim}],
						2, Table[Table[Table[simpler[Sum[ginv[[jj, ll]]ten[[ii, ll, kk]],
								{ll, dim}]], {kk, dim}], {jj, dim}], {ii, dim}], 3, 
              Table[Table[Table[simpler[Sum[ginv[[kk, ll]]ten[[ii, jj, ll]],
								{ll, dim}]], {kk, dim}], {jj, dim}], {ii, dim}]],];
			If[rank == 4,
				tenr = Switch[nn,
						1, Table[
                Table[Table[
                    Table[simpler[Sum[ginv[[ii, mm]]ten[[mm, jj, kk, ll]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],
						2, Table[
                Table[Table[
                    Table[simpler[Sum[ginv[[jj, mm]]ten[[ii, mm, kk, ll]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],
						3, Table[
                Table[Table[
                    Table[simpler[Sum[ginv[[kk, mm]]ten[[ii, jj, mm, ll]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],
						4, Table[
                Table[Table[
                    Table[simpler[Sum[ginv[[ll, mm]]ten[[ii, jj, kk, mm]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}]],];
			If[rank == 5,
				tenr = Switch[nn,
						1, Table[Table[Table[Table[Table[simpler[
												Sum[ginv[[ii, qq]]ten[[qq, jj, kk, ll, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						2, Table[Table[Table[Table[Table[simpler[
												Sum[ginv[[jj, qq]]ten[[ii, qq, kk, ll, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						3, Table[Table[Table[Table[Table[simpler[
												Sum[ginv[[kk, qq]]ten[[ii, jj, qq, ll, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						4, Table[Table[Table[Table[Table[simpler[
												Sum[ginv[[ll, qq]]ten[[ii, jj, kk, qq, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						5, Table[Table[Table[Table[Table[simpler[
												Sum[ginv[[mm, qq]]ten[[ii, jj, kk, ll, qq]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}]],];
			Return[tenr];)];
lower[ten_List, nn_] := Block[{dim, gg, rank, tenr},
		(If[Length[currentst] < 3,
				(Print["Evaluate spacetime[x,g] first"];
					Return[];),];
			rank = tenrank[ten];
			If[rank < nn,
				(Print["Error - insufficient rank"];
					Return[];),];
			gg = currentst[[2]];
			dim = Length[gg[[1]]];
			If[Length[ten] == dim, ,
				(Print["Error - tensor is incorrect dimension"];
				Return[];)];
			If[rank == 1,
				tenr = Table[simpler[Sum[gg[[ii, jj]]ten[[jj]],
								{jj, dim}]], {ii, dim}],];
			If[rank == 2,
				tenr = Switch[nn,
						1, Table[Table[simpler[Sum[gg[[ii, kk]]ten[[kk, jj]],
								{kk, dim}]], {jj, dim}], {ii, dim}],
						2, Table[Table[simpler[Sum[gg[[jj, kk]]ten[[ii, kk]],
								{kk, dim}]], {jj, dim}], {ii, dim}]],];
			If[rank == 3,
				tenr = Switch[nn,
						1, Table[Table[Table[simpler[Sum[gg[[ii, ll]]ten[[ll, jj, kk]],
								{ll, dim}]], {kk, dim}], {jj, dim}], {ii, dim}],
						2, Table[Table[Table[simpler[Sum[gg[[jj, ll]]ten[[ii, ll, kk]],
								{ll, dim}]], {kk, dim}], {jj, dim}], {ii, dim}], 3, 
              Table[Table[Table[simpler[Sum[gg[[kk, ll]]ten[[ii, jj, ll]],
								{ll, dim}]], {kk, dim}], {jj, dim}], {ii, dim}]],];
			If[rank == 4,
				tenr = Switch[nn,
						1, Table[
                Table[Table[
                    Table[simpler[Sum[gg[[ii, mm]]ten[[mm, jj, kk, ll]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],
						2, Table[
                Table[Table[
                    Table[simpler[Sum[gg[[jj, mm]]ten[[ii, mm, kk, ll]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],
						3, Table[
                Table[Table[
                    Table[simpler[Sum[gg[[kk, mm]]ten[[ii, jj, mm, ll]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],
						4, Table[
                Table[Table[
                    Table[simpler[Sum[gg[[ll, mm]]ten[[ii, jj, kk, mm]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}]],];
			If[rank == 5,
				tenr = Switch[nn,
						1, Table[Table[Table[Table[Table[simpler[
												Sum[gg[[ii, qq]]ten[[qq, jj, kk, ll, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						2, Table[Table[Table[Table[Table[simpler[
												Sum[gg[[jj, qq]]ten[[ii, qq, kk, ll, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						3, Table[Table[Table[Table[Table[simpler[
												Sum[gg[[kk, qq]]ten[[ii, jj, qq, ll, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						4, Table[Table[Table[Table[Table[simpler[
												Sum[gg[[ll, qq]]ten[[ii, jj, kk, qq, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						5, Table[Table[Table[Table[Table[simpler[
												Sum[gg[[mm, qq]]ten[[ii, jj, kk, ll, qq]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}]],];
			Return[tenr];)];
norm[ve_List] := Block[{dim, gg, nn},
		(If[Length[currentst] < 3,
				(Print["Evaluate spacetime[x,g] first"];
					Return[];),];
			gg = currentst[[2]];
			dim = Length[gg[[1]]];
			If[Length[ve] == dim, ,
				(Print["Error - vector is incorrect dimension"];
				Return[];)];
			nn = simpler[Sum[Sum[gg[[aa, bb]]ve[[aa]]ve[[bb]],
							{bb, dim}], {aa, dim}]];
			Return[nn];)];
twist[ve_List] := Block[{dim, xx, Christoffel, tw},
		(If[Length[currentst] < 4,
				(Print["Evaluate spacetime[x,g] first"];
					Return[];),];
			xx = currentst[[1]];
			dim = Length[xx];
			If[Length[ve] == dim, ,
				(Print["Error - vector is incorrect dimension"];
				Return[];)];
			Christoffel = currentst[[4]];
			tw = Table[simpler[Sum[Sum[Sum[
						Signature[{ii, jj, kk, ll}]ve[[jj]]*(D[ve[[ll]], xx[[kk]]] -
												Sum[Christoffel[[mm, kk, ll]]ve[[mm]], {mm, dim}]),
									{ll, dim}], {kk, dim}], {jj, dim}]], {ii, dim}];
			Return[tw];)];
coordxform[ten_List, xx_List, xp_List] := 
    Block[{dim, rank, tenp, rep, xst, xform},
		(If[Length[currentst] < 3,
				(Print["Evaluate spacetime[x,g] first"];
					Return[];),];
			xst = currentst[[1]];
			rank = tenrank[ten];
			dim = Length[xst];
			If[Length[ten] == dim, ,
				(Print["Error - tensor is incorrect dimension"];
					Return[];)];
			If[Length[xx] == dim, ,
				(Print["Error - x[xp] list is incorrect dimension"];
					Return[];)];
			If[Length[xp] == dim, ,
				(Print["Error - xp list is incorrect dimension"];
					Return[];)];
			rep = ten /. Thread[xst -> xx];
			xform = 
          Table[Table[simpler[D[xx[[jj]], xp[[ii]]]], {jj, dim}], {ii, dim}];
			If[rank == 1,
				tenp = Table[simpler[Sum[xform[[ii, jj]]rep[[jj]],
								{jj, dim}]], {ii, dim}],];
			If[rank == 2,
				tenp = Table[Table[simpler[Sum[Sum[
										xform[[ii, kk]]xform[[jj, ll]]rep[[kk, ll]],
										{ll, dim}], {kk, dim}]], {jj, dim}], {ii, dim}],];
			If[rank == 3,
				tenp = Table[Table[Table[simpler[Sum[Sum[Sum[
												xform[[ii, ll]]xform[[jj, mm]]xform[[kk, nn]]rep[[ll, mm, nn]],
												{nn, dim}], {mm, dim}], {ll, dim}]], {kk, dim}], {jj, dim}], {ii, 
                dim}],];
			If[rank == 4,
				tenp = Table[Table[Table[Table[simpler[Sum[Sum[Sum[Sum[
														
                              xform[[ii, mm]]xform[[jj, nn]]xform[[kk, 
                                    pp]]xform[[ll, qq]]rep[[mm, nn, pp, qq]],
														{qq, dim}], {pp, dim}], {nn, dim}], {mm, dim}]],
									{ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],];
			If[rank == 5,
				tenp = Table[Table[Table[Table[Table[simpler[Sum[Sum[Sum[Sum[Sum[
																xform[[ii, nn]]xform[[jj, pp]]xform[[kk, qq]]*
																	xform[[ll, vv]]xform[[mm, ww]]rep[[nn, pp, qq, vv, ww]],
																{ww, dim}], {vv, dim}], {qq, dim}], {pp, dim}], {nn, dim}]],
										{mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],];
			Return[tenp];)];
coordxform[ten_List, xx_List, xf_List, xp_List] := 
    Block[{dim, rank, tenp, rep, xform},
		(rank = TensorRank[ten];
			dim = Length[xx];
			If[Length[ten] == dim, ,
				(Print["Error - tensor is incorrect dimension"];
					Return[];)];
			If[Length[xf] == dim, ,
				(Print["Error - function xx(xp) list is incorrect dimension"];
					Return[];)];
			If[Length[xp] == dim, ,
				(Print["Error - xprime list is incorrect dimension"];
					Return[];)];
			rep = ten /. Thread[xx -> xf];
			xform = 
          Table[Table[simpler[D[xf[[jj]], xp[[ii]]]], {jj, dim}], {ii, dim}];
			If[rank == 1,
				tenp = Table[simpler[Sum[xform[[ii, jj]]rep[[jj]],
								{jj, dim}]], {ii, dim}],];
			If[rank == 2,
				tenp = Table[Table[simpler[Sum[Sum[
										xform[[ii, kk]]xform[[jj, ll]]rep[[kk, ll]],
										{ll, dim}], {kk, dim}]], {jj, dim}], {ii, dim}],];
			If[rank == 3,
				tenp = Table[Table[Table[simpler[Sum[Sum[Sum[
												xform[[ii, ll]]xform[[jj, mm]]xform[[kk, nn]]rep[[ll, mm, nn]],
												{nn, dim}], {mm, dim}], {ll, dim}]], {kk, dim}], {jj, dim}], {ii, 
                dim}],];
			If[rank == 4,
				tenp = Table[Table[Table[Table[simpler[Sum[Sum[Sum[Sum[
														
                              xform[[ii, mm]]xform[[jj, nn]]xform[[kk, 
                                    pp]]xform[[ll, qq]]rep[[mm, nn, pp, qq]],
														{qq, dim}], {pp, dim}], {nn, dim}], {mm, dim}]],
									{ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],];
			If[rank == 5,
				tenp = Table[Table[Table[Table[Table[simpler[Sum[Sum[Sum[Sum[Sum[
																xform[[ii, nn]]xform[[jj, pp]]xform[[kk, qq]]*
																	xform[[ll, vv]]xform[[mm, ww]]rep[[nn, pp, qq, vv, ww]],
																{ww, dim}], {vv, dim}], {qq, dim}], {pp, dim}], {nn, dim}]],
										{mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],];
			Return[tenp];)];
sphericalmetric[nn_] := Block[{zz, zval, gs, xx,
			z1, z2, z3, z4, z5, z6, z7, z8, z9, z10, z11},
		(zz = Table[Symbol["z" <> ToString[ii]], {ii, nn}];
			xx = Flatten[{r, ph,
              Table[Symbol["t" <> ToString[ii]], {ii, nn - 2}]}];
			zzval = 
          Flatten[{r*Product[Sin[xx[[ii]]], {ii, 2, nn}], 
              Table[r*Cos[xx[[ii]]]*
                  Product[Sin[xx[[jj]]], {jj, ii + 1, nn}], {ii, 2, nn}]}];
			Thread[Evaluate[zz] = zzval];
			Print["Spherical metric check: ", 
          Simplify[Sum[zz[[ii]]^2, {ii, nn}]] == (r^2)];
			gs = Table[Table[Simplify[Sum[
								D[zz[[kk]], xx[[ii]]]D[zz[[kk]], xx[[jj]]],
								{kk, nn}]], {jj, nn}], {ii, nn}];
			Return[gs];)];
connection[] := Block[{dim, gg, ginv, xx, Christoffel},
		(If[Length[currentst] < 3,
				(Print["Evaluate spacetime[x,g] first"];
					Return[];),];
			xx = currentst[[1]];
			gg = currentst[[2]];
			ginv = currentst[[3]];
			dim = Length[xx];
			Christoffel = Table[Table[Table[0, {kk, dim}], {jj, dim}], {ii, dim}];
			Do[Do[Do[Christoffel[[ii, jj, kk]] =
							simpler[Sum[ginv[[ii, ll]](D[gg[[kk, ll]], xx[[jj]]] +
												D[gg[[jj, ll]], xx[[kk]]] - D[gg[[jj, kk]], xx[[ll]]]), {ll, 
                        dim}]/2],
						{kk, jj, dim}], {jj, dim}], {ii, dim}];
			Do[Do[Do[Christoffel[[ii, kk, jj]] = Christoffel[[ii, jj, kk]],
						{kk, jj + 1, dim}], {jj, dim - 1}], {ii, dim}];
			If[Length[currentst] == 3, AppendTo[currentst, Christoffel],
				currentst[[4]] = Christoffel];
			Return[Christoffel];)];
covder[ten_List] := Block[{dim, xx, chris, der, rank},
		(If[Length[currentst] < 4,
				(Print["Evaluate spacetime[x,g] first"];
					Return[];),];
			xx = currentst[[1]];
			chris = currentst[[4]];
			dim = Length[xx];
			rank = tenrank[ten];
			der = Switch[rank,
				1, Table[Table[simpler[D[ten[[ii]], xx[[jj]]] -
							Sum[chris[[nn, ii, jj]]ten[[nn]], {nn, dim}]], {jj, dim}], {ii, dim}],
				2, Table[Table[Table[simpler[
									D[ten[[ii, jj]], xx[[kk]]] -
							Sum[chris[[nn, ii, kk]]ten[[nn, jj]] +
												chris[[nn, jj, kk]]ten[[ii, nn]], {nn, dim}]],
								{kk, dim}], {jj, dim}], {ii, dim}],
				3, Table[Table[Table[Table[simpler[
									D[ten[[ii, jj, kk]], xx[[ll]]] -
							Sum[chris[[nn, ii, ll]]ten[[nn, jj, kk]] +
												chris[[nn, jj, ll]]ten[[ii, nn, kk]] +
												chris[[nn, kk, ll]]ten[[ii, jj, nn]], {nn, dim}]],
									{ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],
				4, Table[Table[Table[Table[Table[simpler[
									D[ten[[ii, jj, kk, ll]], xx[[mm]]] -
							Sum[chris[[nn, ii, mm]]ten[[nn, jj, kk, ll]] +
												chris[[nn, jj, mm]]ten[[ii, nn, kk, ll]] +
												chris[[nn, kk, mm]]ten[[ii, jj, nn, ll]] +
													chris[[nn, ll, mm]]ten[[ii, jj, kk, nn]], {nn, dim}]],
									{mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}]];
			Return[der];)];
curvature[] := Block[{Rcurv, dim, xx, gg, Christoffel},
		(If[Length[currentst] < 4,
				(Print["Evaluate spacetime[x,g] first"];
					Return[];),];
			xx = currentst[[1]];
			gg = currentst[[2]];
			Christoffel = currentst[[4]];
			dim = Length[xx];
			Rcurv = 
          Table[Table[Table[Table[0, {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
              dim}];
			Do[Do[Do[Do[Rcurv[[ii, jj, kk, mm]] =
								simpler[Sum[(D[Christoffel[[ll, ii, kk]], xx[[jj]]] -
												D[Christoffel[[ll, jj, kk]], xx[[ii]]] +
												Sum[Christoffel[[nn, ii, kk]]Christoffel[[ll, nn, jj]] -
														Christoffel[[nn, jj, kk]]Christoffel[[ll, nn, ii]],
													{nn, dim}])gg[[ll, mm]], {ll, dim}]],
							{mm, kk + 1, dim}], {kk, dim - 1}], {jj, ii + 1, dim}], {ii, 
            dim - 1}];
			Do[Do[Do[Do[Rcurv[[jj, ii, kk, ll]] = -Rcurv[[ii, jj, kk, ll]],
							{ll, kk + 1, dim}], {kk, dim - 1}], {jj, ii + 1, dim}], {ii, 
            dim - 1}];
			Do[Do[Do[Do[Rcurv[[ii, jj, ll, kk]] = -Rcurv[[ii, jj, kk, ll]],
							{ll, kk + 1, dim}], {kk, dim - 1}], {jj, ii + 1, dim}], {ii, 
            dim - 1}];
			Do[Do[Do[Do[Rcurv[[jj, ii, ll, kk]] = Rcurv[[ii, jj, kk, ll]],
							{ll, kk + 1, dim}], {kk, dim - 1}], {jj, ii + 1, dim}], {ii, 
            dim - 1}];
			If[Length[currentst] == 4, AppendTo[currentst, Rcurv],
				currentst[[5]] = Rcurv];
			Return[Rcurv];)];
ricci[] := Block[{Ricci, dim, ginv, Rcurv},
		(If[Length[currentst] < 5,
				(Print["Evaluate curvature[] first"];
					Return[];),];
			ginv = currentst[[3]];
			Rcurv = currentst[[5]];
			dim = Length[ginv[[1]]];
			Ricci = Table[Table[0, {jj, dim}], {ii, dim}];
			Do[Do[Ricci[[ii, jj]] = 
              simpler[Sum[Sum[Rcurv[[ii, mm, jj, ll]]ginv[[mm, ll]],
									{ll, dim}], {mm, dim}]], {jj, ii, dim}], {ii, dim}];
			Do[Do[Ricci[[jj, ii]] = Ricci[[ii, jj]], {jj, ii + 1, dim}], {ii, 
            dim - 1}];
			If[Length[currentst] == 5, AppendTo[currentst, Ricci],
				currentst[[6]] = Ricci];
			Return[Ricci];)];
scalarcurv[] := Block[{dim, sc, ginv, Ricci},
		(If[Length[currentst] < 6,
				(Print["Evaluate curvature[] and ricci[] first"];
					Return[];),];
			ginv = currentst[[3]];
			Ricci = currentst[[6]];
			dim = Length[ginv[[1]]];
			sc = simpler[Sum[Sum[
				Ricci[[ii, jj]]ginv[[ii, jj]], {jj, dim}], {ii, dim}]];
			Return[sc];)];
kretschmann[] := 
	Block[{ginv, Rcurv, R1up, R2up, R3up, R4up, kinv, dim},
		(If[Length[currentst] < 5,
				(Print["Evaluate curvature[] first"];
					Return[];),];
			ginv = currentst[[3]];
			Rcurv = currentst[[5]];
			dim = Length[ginv[[1]]];
			R1up = Table[Table[Table[Table[simpler[Sum[
										Rcurv[[ii, jj, kk, nn]]ginv[[nn, ll]], {nn, dim}]],
								{ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}];
			R2up = Table[Table[Table[Table[simpler[Sum[
							R1up[[ii, jj, nn, ll]]ginv[[nn, kk]], {nn, dim}]],
								{ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}];
			R3up = Table[Table[Table[Table[simpler[Sum[
							R2up[[ii, nn, kk, ll]]ginv[[nn, jj]], {nn, dim}]],
								{ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}];
			R4up = Table[Table[Table[Table[simpler[Sum[
							R3up[[nn, jj, kk, ll]]ginv[[nn, ii]], {nn, dim}]],
								{ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}];
			kinv = simpler[Factor[simpler[Sum[Sum[simpler[Sum[Sum[
							4Rcurv[[ii, jj, kk, ll]]R4up[[ii, jj, kk, ll]],
												{ll, kk + 1, dim}], {kk, dim - 1}]],
									{jj, ii + 1, dim}], {ii, dim - 1}]]]];
			Return[kinv];)];
surfaceg[] := Block[{xx, Christoffel, nullkv, dim, dkv, sg},
		(If[Length[currentst] < 4,
				(Print["Evaluate spacetime[x,g] first"];
					Return[];),];
			xx = currentst[[1]];
			Christoffel = currentst[[4]];
			dim = Length[xx];
			nullkv = Table[0, {ii, dim}];
			nullkv[[dim]] = 1;
			dkv = Table[Table[simpler[D[nullkv[[ii]], xx[[jj]]] +
								Sum[Christoffel[[ii, jj, kk]]nullkv[[kk]], {kk, dim}]],
						{jj, dim}], {ii, dim}];
			sg = simpler[Sqrt[Sum[Sum[
								dkv[[ii, jj]]dkv[[jj, ii]]/2, {jj, dim}], {ii, dim}]]];
			Return[sg];)];
expansion[ve_List] := 
	Block[{xx, gg, ginv, Christoffel, dim, bten, check, expan},
		(If[Length[currentst] < 4,
				(Print["Evaluate spacetime[x,g] first"];
					Return[];),];
			xx = currentst[[1]];
			gg = currentst[[2]];
			ginv = currentst[[3]];
			Christoffel = currentst[[4]];
			dim = Length[xx];
			If[Length[ve] == dim, ,
				(Print["Error - vector is incorrect dimension"];
					Return[];)];
			bten = Table[Table[simpler[Sum[gg[[ll, ii]](D[ve[[ii]], xx[[jj]]] +
				Sum[Christoffel[[ii, jj, kk]]ve[[kk]], {kk, dim}]), {ii, dim}]],
			{jj, dim}], {ll, dim}];
			check = Table[simpler[Sum[ve[[jj]]bten[[jj, ii]],
					{jj, dim}]], {ii, dim}];
			Do[If[check[[ii]] == 0, dummy = 0,
					Print["orthogonality check[", ii, "] = ", check[[ii]]],
					Print["orthogonality check[", ii, "] = ", check[[ii]]]],
				{ii, dim}];
			check = Table[simpler[Sum[ve[[jj]]bten[[ii, jj]],
					{jj, dim}]], {ii, dim}];
			Do[If[check[[ii]] == 0, dummy = 0,
					Print["orthogonality check 2[", ii, "] = ", check[[ii]]],
					Print["orthogonality check 2[", ii, "] = ", check[[ii]]]],
				{ii, dim}];
		expan = simpler[Sum[Sum[ginv[[ii, jj]]bten[[ii, jj]],
				{jj, dim}], {ii, dim}]];
			Return[expan];)];
nullexpansion[ve_List] := 
	Block[{xx, gg, ginv, Christoffel, dim, bten, check, expan},
		(If[Length[currentst] < 4,
				(Print["Evaluate spacetime[x,g] first"];
					Return[];),];
			xx = currentst[[1]];
			gg = currentst[[2]];
			ginv = currentst[[3]];
			Christoffel = currentst[[4]];
			dim = Length[xx];
			If[Length[ve] == dim, ,
				(Print["Error - vector is incorrect dimension"];
					Return[];)];
			bten = Table[Table[simpler[(D[ve[[ii]], xx[[jj]]] -
									Sum[Christoffel[[kk, ii, jj]]ve[[kk]], {kk, dim}])],
						{jj, dim}], {ii, dim}];
			check = Table[Table[Table[simpler[
								ve[[ii]]bten[[jj, kk]] - ve[[ii]]bten[[kk, jj]] +
									ve[[kk]]bten[[ii, jj]] - ve[[kk]]bten[[jj, ii]] +
									ve[[jj]]bten[[kk, ii]] - ve[[jj]]bten[[ii, kk]]],
							{kk, dim}], {jj, dim}], {ii, dim}];
			Do[Do[Do[If[check[[ii, jj, kk]] == 0, dummy = 0,
								Print["hypersurface orthogonality check[",
								ii, ",", jj, ",", kk, "] = ", check[[ii, jj, kk]]],
							Print["hypersurface orthogonality check[",
								ii, ",", jj, ",", kk, "] = ", check[[ii, jj, kk]]]],
						{kk, dim}], {jj, dim}], {ii, dim}];
			expan = simpler[Sum[Sum[ginv[[ii, jj]]bten[[ii, jj]],
							{jj, dim}], {ii, dim}]];
			Return[expan];)];
inter[expr_, tl_List, nn_] := If[nn == 1,
		Integrate[expr, {tl[[1]], 0, 2Pi}],
		Integrate[inter[expr, tl, nn - 1], {tl[[nn]], 0, Pi}]];
surfacearea[tl_List] := Block[{dim,nn,area},
	(If[Length[currentst]<3,
		(Print["Evaluate spacetime[x,g] first"];
		Return[];),];
	gg = currentst[[2]];
	dim = Length[gg[[1]]];
	nn = Length[tl];
	area = simpler[inter[Sqrt[-Product[gg[[ii,ii]],{ii,dim}]],tl,nn]];
	Return[area];)];
vielbein[ee_] := Block[{einv, xx},
		(If[Length[currentst] < 3,
				(Print["Evaluate spacetime[x,g] first"];
					Return[];),];
			xx = currentst[[1]];
			If[Length[xx] == Length[ee], ,
				(Print["Error - vielbein is incorrect dimension"];
					Return[];)];
			currentframe = {ee};
			einv = simpler[Transpose[Inverse[ee]]];
			AppendTo[currentframe, einv];
			checkvielbein[];
			Return[currentframe];)];
checkvielbein[] := Block[{ee, einv, gg, ginv, dim, test, eta},
		(If[Length[currentst] < 3,
				(Print["Evaluate spacetime[x,g] first"];
					Return[];),];
			If[Length[currentframe] < 2,
				(Print["Evaluate vielbein[e] first"];
					Return[];),];
			ee = currentframe[[1]];
			einv = currentframe[[2]];
			gg = currentst[[2]];
			ginv = currentst[[3]];
			dim = Length[ee[[1]]];
			eta = lorentzmetric[dim];
			test = Table[Table[simpler[Sum[Sum[ee[[aa, ii]]ee[[bb, jj]]eta[[aa, bb]],
									{bb, dim}], {aa, dim}]], {jj, dim}], {ii, dim}];
			Print["e e eta = g: ", FullSimplify[test == gg]];
			test = Table[Table[simpler[Sum[Sum[
									ginv[[ii, jj]]ee[[aa, ii]]ee[[bb, jj]],
									{jj, dim}], {ii, dim}]], {bb, dim}], {aa, dim}];
			Print["e e ginv = eta: ", FullSimplify[test == eta]];
			test = Table[Table[simpler[simpler[Sum[Sum[
										einv[[aa, ii]]einv[[bb, jj]]eta[[aa, bb]],
										{bb, dim}], {aa, dim}]]], {jj, dim}], {ii, dim}];
			Print["einv einv eta = ginv: ", FullSimplify[test == ginv]];
			test = Table[Table[simpler[Sum[Sum[
									gg[[ii, jj]]einv[[aa, ii]]einv[[bb, jj]],
									{jj, dim}], {ii, dim}]], {bb, dim}], {aa, dim}];
			Print["g einv einv = eta: ", FullSimplify[test == eta]];)];
lorentz[ten_List, nn_] := Block[{einv, dim, rank, tenr},
		(If[Length[currentframe] < 2,
				(Print["Evaluate vielbein[e] first"];
					Return[];),];
			rank = tenrank[ten];
			If[rank < nn,
				(Print["Error - insufficient rank"];
					Return[];),];
			einv = currentframe[[2]];
			dim = Length[einv];
			If[Length[ten] == dim, ,
				(Print["Error - tensor is incorrect dimension"];
					Return[];)];
			If[rank == 1,
				tenr = Table[simpler[Sum[einv[[ii, jj]]ten[[jj]],
								{jj, dim}]], {ii, dim}],];
			If[rank == 2,
				tenr = Switch[nn,
						1, Table[Table[simpler[Sum[einv[[ii, kk]]ten[[kk, jj]],
								{kk, dim}]], {jj, dim}], {ii, dim}],
						2, Table[Table[simpler[Sum[einv[[jj, kk]]ten[[ii, kk]],
								{kk, dim}]], {jj, dim}], {ii, dim}]],];
			If[rank == 3,
				tenr = Switch[nn,
						1, Table[Table[Table[simpler[Sum[einv[[ii, ll]]ten[[ll, jj, kk]],
								{ll, dim}]], {kk, dim}], {jj, dim}], {ii, dim}],
						2, Table[Table[Table[simpler[Sum[einv[[jj, ll]]ten[[ii, ll, kk]],
								{ll, dim}]], {kk, dim}], {jj, dim}], {ii, dim}], 3, 
              Table[Table[Table[simpler[Sum[einv[[kk, ll]]ten[[ii, jj, ll]],
								{ll, dim}]], {kk, dim}], {jj, dim}], {ii, dim}]],];
			If[rank == 4,
				tenr = Switch[nn,
						1, Table[
                Table[Table[
                    Table[simpler[Sum[einv[[ii, mm]]ten[[mm, jj, kk, ll]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],
						2, Table[
                Table[Table[
                    Table[simpler[Sum[einv[[jj, mm]]ten[[ii, mm, kk, ll]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],
						3, Table[
                Table[Table[
                    Table[simpler[Sum[einv[[kk, mm]]ten[[ii, jj, mm, ll]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],
						4, Table[
                Table[Table[
                    Table[simpler[Sum[einv[[ll, mm]]ten[[ii, jj, kk, mm]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}]],];
			If[rank == 5,
				tenr = Switch[nn,
						1, Table[Table[Table[Table[Table[simpler[
												Sum[einv[[ii, qq]]ten[[qq, jj, kk, ll, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						2, Table[Table[Table[Table[Table[simpler[
												Sum[einv[[jj, qq]]ten[[ii, qq, kk, ll, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						3, Table[Table[Table[Table[Table[simpler[
												Sum[einv[[kk, qq]]ten[[ii, jj, qq, ll, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						4, Table[Table[Table[Table[Table[simpler[
												Sum[einv[[ll, qq]]ten[[ii, jj, kk, qq, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						5, Table[Table[Table[Table[Table[simpler[
												Sum[einv[[mm, qq]]ten[[ii, jj, kk, ll, qq]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}]],];
			Return[tenr];)];
curved[ten_List, nn_] := Block[{ee, dim, rank, tenr},
		(If[Length[currentframe] < 2,
				(Print["Evaluate vielbein[e] first"];
					Return[];),];
			rank = tenrank[ten];
			If[rank < nn,
				(Print["Error - insufficient rank"];
					Return[];),];
			ee = currentframe[[1]];
			dim = Length[ee];
			If[Length[ten] == dim, ,
				(Print["Error - tensor is incorrect dimension"];
					Return[];)];
			If[rank == 1,
				tenr = Table[simpler[Sum[ee[[jj, ii]]ten[[jj]],
								{jj, dim}]], {ii, dim}],];
			If[rank == 2,
				tenr = Switch[nn,
						1, Table[Table[simpler[Sum[ee[[kk, ii]]ten[[kk, jj]],
								{kk, dim}]], {jj, dim}], {ii, dim}],
						2, Table[Table[simpler[Sum[ee[[kk, jj]]ten[[ii, kk]],
								{kk, dim}]], {jj, dim}], {ii, dim}]],];
			If[rank == 3,
				tenr = Switch[nn,
						1, Table[Table[Table[simpler[Sum[ee[[ll, ii]]ten[[ll, jj, kk]],
								{ll, dim}]], {kk, dim}], {jj, dim}], {ii, dim}],
						2, Table[Table[Table[simpler[Sum[ee[[ll, jj]]ten[[ii, ll, kk]],
								{ll, dim}]], {kk, dim}], {jj, dim}], {ii, dim}], 3, 
              Table[Table[Table[simpler[Sum[ee[[ll, kk]]ten[[ii, jj, ll]],
								{ll, dim}]], {kk, dim}], {jj, dim}], {ii, dim}]],];
			If[rank == 4,
				tenr = Switch[nn,
						1, Table[
                Table[Table[
                    Table[simpler[Sum[ee[[mm, ii]]ten[[mm, jj, kk, ll]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],
						2, Table[
                Table[Table[
                    Table[simpler[Sum[ee[[mm, jj]]ten[[ii, mm, kk, ll]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],
						3, Table[
                Table[Table[
                    Table[simpler[Sum[ee[[mm, kk]]ten[[ii, jj, mm, ll]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],
						4, Table[
                Table[Table[
                    Table[simpler[Sum[ee[[mm, ll]]ten[[ii, jj, kk, mm]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}]],];
			If[rank == 5,
				tenr = Switch[nn,
						1, Table[Table[Table[Table[Table[simpler[
												Sum[ee[[qq, ii]]ten[[qq, jj, kk, ll, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						2, Table[Table[Table[Table[Table[simpler[
												Sum[ee[[qq, jj]]ten[[ii, qq, kk, ll, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						3, Table[Table[Table[Table[Table[simpler[
												Sum[ee[[qq, kk]]ten[[ii, jj, qq, ll, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						4, Table[Table[Table[Table[Table[simpler[
												Sum[ee[[qq, ll]]ten[[ii, jj, kk, qq, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						5, Table[Table[Table[Table[Table[simpler[
												Sum[ee[[qq, mm]]ten[[ii, jj, kk, ll, qq]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}]],];
			Return[tenr];)];
lorentzraise[ten_List, nn_] := Block[{dim, xx, ginv, rank, tenr},
		(If[Length[currentst] < 3,
				(Print["Evaluate spacetime[x,g] first"];
					Return[];),];
			rank = tenrank[ten];
			If[rank < nn,
				(Print["Error - insufficient rank"];
					Return[];),];
			xx = currentst[[1]];
			dim = Length[xx];
			If[Length[ten] == dim, ,
				(Print["Error - tensor is incorrect dimension"];
				Return[];)];
			ginv = lorentzmetric[dim];
			If[rank == 1,
				tenr = Table[simpler[Sum[ginv[[ii, jj]]ten[[jj]],
								{jj, dim}]], {ii, dim}],];
			If[rank == 2,
				tenr = Switch[nn,
						1, Table[Table[simpler[Sum[ginv[[ii, kk]]ten[[kk, jj]],
								{kk, dim}]], {jj, dim}], {ii, dim}],
						2, Table[Table[simpler[Sum[ginv[[jj, kk]]ten[[ii, kk]],
								{kk, dim}]], {jj, dim}], {ii, dim}]],];
			If[rank == 3,
				tenr = Switch[nn,
						1, Table[Table[Table[simpler[Sum[ginv[[ii, ll]]ten[[ll, jj, kk]],
								{ll, dim}]], {kk, dim}], {jj, dim}], {ii, dim}],
						2, Table[Table[Table[simpler[Sum[ginv[[jj, ll]]ten[[ii, ll, kk]],
								{ll, dim}]], {kk, dim}], {jj, dim}], {ii, dim}], 3, 
              Table[Table[Table[simpler[Sum[ginv[[kk, ll]]ten[[ii, jj, ll]],
								{ll, dim}]], {kk, dim}], {jj, dim}], {ii, dim}]],];
			If[rank == 4,
				tenr = Switch[nn,
						1, Table[
                Table[Table[
                    Table[simpler[Sum[ginv[[ii, mm]]ten[[mm, jj, kk, ll]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],
						2, Table[
                Table[Table[
                    Table[simpler[Sum[ginv[[jj, mm]]ten[[ii, mm, kk, ll]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],
						3, Table[
                Table[Table[
                    Table[simpler[Sum[ginv[[kk, mm]]ten[[ii, jj, mm, ll]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}],
						4, Table[
                Table[Table[
                    Table[simpler[Sum[ginv[[ll, mm]]ten[[ii, jj, kk, mm]],
								{mm, dim}]], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}]],];
			If[rank == 5,
				tenr = Switch[nn,
						1, Table[Table[Table[Table[Table[simpler[
												Sum[ginv[[ii, qq]]ten[[qq, jj, kk, ll, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						2, Table[Table[Table[Table[Table[simpler[
												Sum[ginv[[jj, qq]]ten[[ii, qq, kk, ll, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						3, Table[Table[Table[Table[Table[simpler[
												Sum[ginv[[kk, qq]]ten[[ii, jj, qq, ll, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						4, Table[Table[Table[Table[Table[simpler[
												Sum[ginv[[ll, qq]]ten[[ii, jj, kk, qq, mm]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}],
						5, Table[Table[Table[Table[Table[simpler[
												Sum[ginv[[mm, qq]]ten[[ii, jj, kk, ll, qq]],
								{qq, dim}]], {mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}]],];
			Return[tenr];)];
epsilon[nn_, vind_List] := Block[{eps, epsind, ind, dim},
		(If[Length[currentst] < 3,
				(Print["Evaluate spacetime[x,g] first"];
					Return[];),];
			dim = Length[currentst[[1]]];
			If[Length[vind] == 0, ind = Table[ii, {ii, dim}], ind = vind];
			If[nn == 2,
				(eps = Table[Table[0, {jj, dim}], {ii, dim}];
					epsind = Permutations[ind];
					Do[eps[[epsind[[ii, 1]], epsind[[ii, 2]]]] = 
							Signature[epsind[[ii]]], {ii, Length[epsind]}];),];
			If[nn == 3,
				(eps = Table[Table[Table[0, {kk, dim}], {jj, dim}], {ii, dim}];
					epsind = Permutations[ind];
					Do[eps[[epsind[[ii, 1]], epsind[[ii, 2]], epsind[[ii, 3]]]] = 
							Signature[epsind[[ii]]], {ii, Length[epsind]}];),];
			If[nn == 4,
				(eps = 
              Table[Table[
                  Table[Table[0, {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}];
					epsind = Permutations[ind];
					Do[eps[[epsind[[ii, 1]], epsind[[ii, 2]], epsind[[ii, 3]], 
                    epsind[[ii, 4]]]] = 
							Signature[epsind[[ii]]], {ii, Length[epsind]}];),];
			If[nn == 5,
				(eps = Table[Table[Table[Table[Table[0,
											{mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}];
					epsind = Permutations[ind];
					Do[eps[[epsind[[ii, 1]], epsind[[ii, 2]], epsind[[ii, 3]],
								epsind[[ii, 4]], epsind[[ii, 5]]]] = 
							Signature[epsind[[ii]]], {ii, Length[epsind]}];),];
			Return[eps];)];
spinconn[] := Block[{ee, einv, xx, dim, chris, eta, omega, check},
		(If[Length[currentst] < 4,
				(Print["Evaluate spacetime[x,g] first"];
					Return[];),];
			If[Length[currentframe] < 2,
				(Print["Evaluate vielbein[e] first"];
					Return[];),];
			ee = currentframe[[1]];
			einv = currentframe[[2]];
			xx = currentst[[1]];
			dim = Length[xx];
			chris = currentst[[4]];
			eta = lorentzmetric[dim];
			omega = Table[Table[Table[0, {kk, dim}], {jj, dim}], {ii, dim}];
			Do[Do[Do[(omega[[ii, aa, bb]] = simpler[Sum[Sum[
										-eta[[bb, cc]]einv[[cc, jj]](D[ee[[aa, jj]], xx[[ii]]] -
													Sum[chris[[kk, ii, jj]]ee[[aa, kk]], {kk, dim}]),
										{jj, dim}], {cc, dim}]];
							omega[[ii, bb, aa]] = -omega[[ii, aa, bb]];),
						{bb, aa + 1, dim}], {aa, dim - 1}], {ii, dim}];
			check = Table[Table[Table[simpler[D[ee[[aa, jj]], xx[[ii]]] +
						Sum[Sum[eta[[bb, cc]]omega[[ii, aa, bb]]ee[[cc, jj]],
								{cc, dim}], {bb, dim}] - Sum[chris[[kk, ii, jj]]ee[[aa, kk]],
							{kk, dim}]], {jj, dim}], {ii, dim}], {aa, dim}];
			components[check];
			Return[omega];)];
sigma = {{{0, 1}, {1, 0}}, {{0, -I}, {I, 0}},
		{{1, 0}, {0, -1}}, IdentityMatrix[2]};
anticomm[xx_List, yy_List] := Dot[xx, yy] + Dot[yy, xx];
recurprod[mm_List, nn_] := 
    If[nn == 1, mm[[1]], Dot[recurprod[mm, nn - 1], mm[[nn]]]];
gammamatrix[t_, ss_] :=
	 
    Block[{mm, ind, gmdim, fact, fact2, check, mtest, dummy, gammam},
		(mm = t + ss;
			If[Floor[mm/2] == (mm/2), fact = I^(Mod[mm, 4]/2),
					fact = (-1)^((-3 - mm)/4)];
			gmdim = 2^Floor[mm/2];
			ind = Table[Flatten[{Table[3, {jj, Floor[(ii - 1)/2]}], 2 - Mod[ii, 2],
							Table[4, {jj, Floor[mm/2] - Floor[(ii - 1)/2] - 1}]}],
					{ii, 2Floor[mm/2]}];
			If[Floor[mm/2] == (mm/2), dummy = 0,
				AppendTo[ind, Table[4, {ii, Floor[mm/2]}]]];
			gammam = Table[Table[Table[Product[
					sigma[[ind[[kk, ll]],
						1 + Mod[Floor[(ii - 1)/(2^Floor[mm/2 - ll])], 2],
						1 + Mod[Floor[(jj - 1)/(2^Floor[mm/2 - ll])], 2]]],
					{ll, mm/2}], {jj, gmdim}], {ii, gmdim}], {kk, mm}];
			If[(Floor[mm/2] == (mm/2)), dummy = 0, If[(mm - 1) > 0,
					gammam[[mm]] = fact*recurprod[gammam, mm - 1],]];
			If[t == 1, gammam[[mm]] = gammam[[mm]]*I,];
			check = 
          Table[Table[anticomm[gammam[[ii]], gammam[[jj]]], {jj, mm}], {ii, 
              mm}];
			Print["Anticommutator check: ",
				{Table[check[[ii, ii]] == 2IdentityMatrix[gmdim], {ii, mm - 1}],
				check[[mm, mm]] == ((-1)^t)2IdentityMatrix[gmdim],
					Table[Table[{ii, jj, check[[ii, jj]] == 0IdentityMatrix[gmdim]},
						{jj, ii + 1, mm}], {ii, 1, mm - 1}]}];
			fact2 = fact;
			If[t == 1, fact2 = I^(1 - Mod[Floor[mm/2], 2]),];
			mtest = fact2*recurprod[gammam, mm];
			AppendTo[gammam, mtest];
			Print["Chiral operator check: ",
					Dot[mtest, mtest] == IdentityMatrix[gmdim]];
		Return[gammam];)];
checkgamma[gammam_List, t_, ss_] :=
	 
    Block[{mm, ind, gmdim, fact, fact2, check, mtest, dummy},
		(mm = t + ss;
			gmdim = 2^Floor[mm/2];
			check = 
          Table[Table[anticomm[gammam[[ii]], gammam[[jj]]], {jj, mm}], {ii, 
              mm}];
			Print["Anticommutator check: ",
				{Table[check[[ii, ii]] == 2IdentityMatrix[gmdim], {ii, mm - 1}],
				check[[mm, mm]] == ((-1)^t)2IdentityMatrix[gmdim],
					Table[Table[{ii, jj, check[[ii, jj]] == 0IdentityMatrix[gmdim]},
						{jj, ii + 1, mm}], {ii, 1, mm - 1}]}];
			Print["Chiral operator check: ",
					Dot[gammam[[mm + 1]], gammam[[mm + 1]]] == IdentityMatrix[gmdim]];)];
crossmatrix[mat1_List, mat2_List] := 
	Block[{row1, col1, row2, col2, crow, ccol, cross},
		(row1 = Length[mat1];
			col1 = Length[mat1[[1]]];
			row2 = Length[mat2];
			col2 = Length[mat2[[1]]];
			crow = row1*row2;
			ccol = col1*col2;
			cross = Table[Table[
						mat1[[1 + Mod[Floor[(ii - 1)/row2], row1],
							1 + Mod[Floor[(jj - 1)/col2], col1]]]*
							mat2[[1 + Mod[(ii - 1), row2], 1 + Mod[(jj - 1), col2]]],
						{jj, ccol}], {ii, crow}];
			Return[cross];)];
splitgamma[gamma1_List, gamma2_List] := Block[{p1, p2, gd, sgamma},
		(p1 = Length[gamma1] - 1;
			p2 = Length[gamma2] - 1;
			sgamma = Table[0, {ii, p1 + p2}];
			If[p1 == (2Floor[p1/2]),
				(gd = Length[gamma2[[1]]];
					Do[sgamma[[ii]] = crossmatrix[gamma1[[ii]], IdentityMatrix[gd]], {ii, 
                p1}];
					Do[sgamma[[ii + p1]] = crossmatrix[gamma1[[p1 + 1]], gamma2[[ii]]], {ii, 
                p2}];),
				(gd = Length[gamma1[[1]]];
					Do[sgamma[[ii]] = crossmatrix[gamma1[[ii]], gamma2[[p2 + 1]]], {ii, 
                p1}];
					Do[sgamma[[ii + p1]] = crossmatrix[IdentityMatrix[gd],
								gamma2[[ii]]], {ii, p2}];)];
			AppendTo[sgamma, recurprod[sgamma, p1 + p2]];
			Return[sgamma];)];
antigamma[gammad_, nn_]  := Block[{dim, agamma, ind},
		(dim = Length[gammad] - 1;
			If[nn == 2,
				agamma = Table[Table[Dot[gammad[[jj]], gammad[[kk]]],
								{kk, dim}], {jj, dim}],];
			If[nn == 3,
				(agamma = Table[Table[Table[0, {kk, dim}], {jj, dim}], {ii, dim}];
							Do[Do[Do[(
											ind = Permutations[{ii, kk, ll}];
											
                    agamma[[ii, kk, ll]] = 
                      Dot[gammad[[ii]], gammad[[kk]], gammad[[ll]]];
											ind = Rest[ind];
											Do[agamma[[ind[[mm, 1]], ind[[mm, 2]], ind[[mm, 3]]]] =
													Signature[ind[[mm]]]agamma[[ii, kk, ll]], {mm, Length[ind]}];),
										{ll, kk + 1, dim}], {kk, ii + 1, dim - 1}], {ii, dim - 2}];),];
			If[nn == 4,
				(agamma = Table[Table[Table[Table[0,
										{ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}];
					Do[Do[Do[Do[(
										ind = Permutations[{ii, kk, ll, mm}];
										agamma[[ii, kk, ll, mm]] = Dot[gammad[[ii]], gammad[[kk]],
												gammad[[ll]], gammad[[mm]]];
										ind = Rest[ind];
										
                      Do[agamma[[ind[[pp, 1]], ind[[pp, 2]], ind[[pp, 3]], 
                              ind[[pp, 4]]]] =
												Signature[ind[[pp]]]agamma[[ii, kk, ll, mm]], {pp, 
                          Length[ind]}];),
									{mm, ll + 1, dim}], {ll, kk + 1, dim - 1}],
							{kk, ii + 1, dim - 2}], {ii, dim - 3}];),];
			If[nn == 5,
				(agamma = Table[Table[Table[Table[Table[0,
										{mm, dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, dim}];
					Do[Do[Do[Do[Do[(
										ind = Permutations[{ii, kk, ll, mm, qq}];
										agamma[[ii, kk, ll, mm, qq]] = Dot[gammad[[ii]], gammad[[kk]],
															gammad[[ll]], gammad[[mm]], gammad[[qq]]];
										ind = Rest[ind];
										Do[agamma[[ind[[pp, 1]], ind[[pp, 2]], ind[[pp, 3]],
														ind[[pp, 4]], ind[[pp, 5]]]] =
												Signature[ind[[pp]]]agamma[[ii, kk, ll, mm, qq]], {pp, 
                            Length[ind]}];),
									{qq, mm + 1, dim}], {mm, ll + 1, dim - 1}], {ll, kk + 1, dim - 2}],
							{kk, ii + 1, dim - 3}], {ii, dim - 4}];),];
			Return[agamma];)];
sugra[] := Block[{},
		(Print["Available functions are\n",
			"   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n",
			"   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n",
			"   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n",
			"   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
			"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n",
			"   expansion[vector], nullexpansion[vector], surfacearea[x],\n",
			"   vielbein[e], checkvielbein[], lorentz[tensor,index],\n",
			"   curved[tensor,index], lorentzraise[tensor,index], epsilon[d,indices],\n\
",
			"   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n",
			"   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n",
			"Tensors are limited to rank 5.  All indices are lower (l) except:\n",
			"   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n",
			"   vielbein[[u,l]], inverse vielbein[[l,u]] & spin connection[[l,u,u]].\n \
\n",
			"All indices are spacetime (s) except:\n",
			"   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n",
			"   all gamma matrix indices.\n \n",
			"The last component in any index is the time component.\n \n",
			"All computations are performed with currentst and currentframe."];
        Print["AVOID USING C, D, E, I, N and O as variables !"];)];
sugra[];