//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// contract.cc

#include <ptah.h>
#include <element.h>
#include <hash.h>
#include <ntuple.h>

extern int ntuple::contract(int number_of_ntuples, ntuple** ntuple_list, 
			     ntuple* destination_ntuple, 
			     bool limit_redundant_indices) {
  int index_value, mult;
  long long overall_coefficient;
#ifdef trace_iter
  char str [256];
#endif
  sum new_sum;
  int* free_indices = new int [destination_ntuple->rank];
  if (free_indices == NULL)
    die("allocating free_indices");
  memset(free_indices, 0, destination_ntuple->rank * sizeof(int));
  sum** multiply_list = new sum* [number_of_ntuples + 1];
  if (multiply_list == NULL)
    die("allocating multiply_list");
#ifdef tracentry
  trace_ncn += 1;
#endif
  overall_coefficient = 1;
  mult = make_sum_list(number_of_ntuples, ntuple_list, limit_redundant_indices);
  if (mult == 0) {
    delete [] free_indices;
    delete [] multiply_list;
    destination_ntuple->zero();
    return 1;}
  while (parallelize(destination_ntuple)) {
    multiply_list[0] = sum_list[current_sum_index]->value;
    for (int n = 0; n < number_of_ntuples; n++)
      multiply_list[n + 1] = ntuple_list[n]->sum_list[current_sum_index]->value;
    new_sum.multiply(number_of_ntuples + 1, multiply_list, &overall_coefficient);
#ifdef trace_iter
    sprintf(str, "Processing current_sum_index = %d; indices = ", current_sum_index);
#endif
    for (int i = 0; i < destination_ntuple->rank; i++) {
      index_value = (indices_list[current_sum_index] >> (6 * i)) & 63;
#ifdef debug
      if (index_value == 63)
	die("indices_list has a dummy index");
#endif
#ifdef trace_iter
      sprintf(str + strlen(str), "%d ", index_value);
#endif
      free_indices[i] = index_value;}
#ifdef trace_iter
    to_stderr (str);
#endif
    destination_ntuple->add(free_indices, &new_sum);
    new_sum.zero();}
  destination_ntuple->simplify();
  delete [] free_indices;
  delete [] multiply_list;
  return mult; }

extern int ntuple::contract(int number_of_ntuples, ntuple** ntuple_list, 
			     sum* destination_invariant, 
			     bool limit_redundant_indices) {
  int mult;
  long long overall_coefficient;
#ifdef trace_iter
  char str [256];
#endif
  sum** multiply_list = new sum* [number_of_ntuples + 1];
  if (multiply_list == NULL)
    die("allocating multiply_list");
#ifdef tracentry
  trace_nci += 1;
#endif
  overall_coefficient = 1;
  mult = make_sum_list(number_of_ntuples, ntuple_list, limit_redundant_indices);
  if (mult == 0) {
    delete [] multiply_list;
    destination_invariant->zero();
    return 1;}
  while (parallelize(destination_invariant)) {
#ifdef trace_iter
    sprintf(str, "Processing current_sum_index = %d.", current_sum_index);
    to_stderr (str);
#endif
    multiply_list[0] = sum_list[current_sum_index]->value;
    for (int n = 0; n < number_of_ntuples; n++)
      multiply_list[n + 1] = ntuple_list[n]->sum_list[current_sum_index]->value;
    destination_invariant->multiply(number_of_ntuples + 1, multiply_list, &overall_coefficient);}
  delete [] multiply_list;
  return mult; }

extern ntuple* ntuple::self_contract(int index_1, int index_2) {
#ifdef tracentry
  trace_nsc += 1;
#endif
  int index;
#ifdef debug
  if ((index_1 < 0) || (index_1 >= rank) || (index_2 < 0) || (index_2 >= rank))
    die("self_contract - index out of bounds");
  if (rank <= 2)
    die("attempting to create a scalar with self_contract");
  if ((upper[index_1] && upper[index_2]) || (!upper[index_1] && !upper[index_2]))
    die("attempting to self_contract 2 upper or lower indices");
#endif
  sum* new_sum = new sum;
  if (new_sum == NULL)
    die("allocating new_sum");
  sum** multiply_list = new sum* [2];
  if (multiply_list == NULL)
    die("allocating multiply_list");
  bool* new_upper = new bool [rank - 2];
  if (new_upper == NULL)
    die("allocating new_upper");
  bool* new_antisymmetric = new bool [rank - 2];
  if (new_antisymmetric == NULL)
    die("allocating new_antisymmetric");
  bool* new_symmetric = new bool [rank - 2];
  if (new_symmetric == NULL)
    die("allocating new_symmetric");
  bool* new_spinor = new bool [rank - 2];
  if (new_spinor == NULL)
    die("allocating new_spinor");
  long long one = 1;
  index = 0;
  for(int i = 0; i < rank; i++) {
    if ((i == index_1) || (i == index_2))
      continue;
#ifdef debug
    if ((index < 0) || (index >= rank))
      die("self-contract - index out of bounds");
#endif
    new_upper[index] = upper[i];
    new_antisymmetric[index] = false;
    new_symmetric[index] = false;
    new_spinor[index] = spinor[i];
    index += 1;}
  ntuple* new_ntuple = new ntuple(rank-2, new_upper, new_antisymmetric, new_symmetric, new_spinor);
  if (new_ntuple == NULL)
    die("allocating new_ntuple");
  for (int i = 0; i < rank; i++)
    index_list[i] = -1;
  index_list[index_1] = 1;
  index_list[index_2] = 1;
  make_free_index_list(false);
  for (current_free_index = 0; current_free_index < free_index_length; current_free_index++) {
    index = 0;
    for (int i = 0; i < rank; i++) {
      index_list[i] = free_index(i);
      if (index_list[i] < 0)
	continue;
#ifdef debug
      if ((index < 0) || (index >= rank))
	die("self-contract - index out of bounds (2)");
#endif
      new_ntuple->index_list[index] = index_list[i];
      index += 1;}
    for (int i = 0; i < dimension; i++) {             // contract for one new component
      index_list[index_1] = i;
      index_list[index_2] = i;
      current_component = sub(index_list);
      if (current_component == NULL)
	continue;
      multiply_list[0] = current_component->value;
      new_sum->multiply(1, multiply_list, &one);}
    new_sum->simplify();
    new_ntuple->set(new_ntuple->index_list, new_sum);
    new_sum = new sum;
    if (new_sum == NULL)
      die("allocating new_sum");}
  delete new_sum;
  delete [] new_upper;
  delete [] new_antisymmetric;
  delete [] new_symmetric;
  delete [] new_spinor;
  delete [] multiply_list;
  return new_ntuple; }

extern ntuple* ntuple::ntuple_contract(int number_of_ntuples, ntuple** ntuple_list, int* index_matrix) {
#ifdef tracentry
  trace_nnc += 1;
#endif
  // each index list in index_matrix is max_rank long and has summation ordinals in each position
  int number_of_free_indices = 0;
  int number_of_sum_indices = 0;
  int current_index, mult;
  int max_rank = 0;
  for (int n = 0; n < number_of_ntuples; n++) {
#ifdef debug
    if (ntuple_list[n]->ntuple_magic_no != ntuple_magic)
      die("attempting to contract a non-ntuple");
#endif
    if (ntuple_list[n]->rank > max_rank)
      max_rank = ntuple_list[n]->rank;}
  for (int n = 0; n < number_of_ntuples; n++)
    for (int j = 0; j < ntuple_list[n]->rank; j++) {
      current_index = index_matrix[n * max_rank + j];
      if (current_index < 0)
	number_of_free_indices += 1;
      else
	number_of_sum_indices += 1;}
#ifdef debug
  if (number_of_free_indices <= 0)
    die("attempting to create a scalar with contract");
  if ((number_of_sum_indices % 2) != 0)
    die("improper contraction indices");
#endif
  number_of_sum_indices /= 2;
#ifdef debug
  for (int n = 0; n < number_of_ntuples; n++)
    for (int i = 0; i < ntuple_list[n]->rank; i++) {
      current_index = index_matrix[n * max_rank + i];
      if (current_index > number_of_sum_indices)
	die("contraction index too large");}
#endif
  int* check_upper = new int [number_of_sum_indices];
  if (check_upper == NULL)
    die("allocating check_upper");
  memset(check_upper, 0, number_of_sum_indices * sizeof(int));
  for (int n = 0; n < number_of_ntuples; n++)
    for (int j = 0; j < ntuple_list[n]->rank; j++) {
      current_index = index_matrix[n * max_rank + j];
      if (current_index >= 0) {
#ifdef debug
	if (current_index >= rank)
	  die("contract - list index out of bounds");
#endif
	if (ntuple_list[n]->upper[j])
	  check_upper[current_index] += 1;}}
#ifdef debug
  for (int i = 0; i < number_of_sum_indices; i++) {
    if (check_upper[i] != 1)
      die("attempting to contract same height indices");}
#endif
  int index = 0;
  bool* new_upper = new bool [number_of_free_indices];
  if (new_upper == NULL)
    die("allocating new_upper");
  bool* new_antisymmetric = new bool [number_of_free_indices];
  if (new_antisymmetric == NULL)
    die("allocating new_antisymmetric");
  bool* new_symmetric = new bool [number_of_free_indices];
  if (new_symmetric == NULL)
    die("allocating new_symmetric");
  bool* new_spinor = new bool [number_of_free_indices];
  if (new_spinor == NULL)
    die("allocating new_spinor");
  for (int n = 0; n < number_of_ntuples; n++)
    for (int j = 0; j < ntuple_list[n]->rank; j++) {
      current_index = index_matrix[n * max_rank + j];
      ntuple_list[n]->index_list[j] = current_index;
      if (current_index < 0) {
#ifdef debug
	if ((index < 0) || (index >= number_of_free_indices))
	  die("contract - index out of bounds");
#endif
	new_upper[index] = ntuple_list[n]->upper[j];
	new_antisymmetric[index] = ntuple_list[n]->antisymmetric[j];
	new_symmetric[index] = ntuple_list[n]->symmetric[j];
	if ((j == (ntuple_list[n]->rank - 1)) ||
	    (index_matrix[n * max_rank + j + 1] > -1)) {
	  new_antisymmetric[index] = false;
	  new_symmetric[index] = false;}
	new_spinor[index] = ntuple_list[n]->spinor[j];
	index += 1;}}
  if ((index % 2) == 1) {
    new_antisymmetric[index - 1] = false;
    new_symmetric[index - 1] = false;}
  ntuple* new_ntuple = new ntuple(number_of_free_indices, new_upper, new_antisymmetric, new_symmetric, new_spinor);
  if (new_ntuple == NULL)
    die("allocating new_ntuple");
  mult = ntuple_list[0]->contract((number_of_ntuples - 1), &(ntuple_list[1]), new_ntuple, true);
  if (mult != 1)
    new_ntuple->scalar_multiply(mult);
  new_ntuple->fill();
  delete [] check_upper;
  delete [] new_upper;
  delete [] new_antisymmetric;
  delete [] new_symmetric;
  delete [] new_spinor;
  return new_ntuple; }

extern sum* ntuple::scalar() {
#ifdef tracentry
  trace_ns += 1;
#endif
#ifdef debug
  if (rank != 2)
    die("attempting to create a scalar without a bivector");
  if ((upper[0] && upper[1]) || (!upper[0] && !upper[1]))
    die("attempting to self_contract 2 upper or lower indices");
#endif
  sum* new_sum = new sum(large_hash_size);
  if (new_sum == NULL)
    die("allocating new_sum");
  sum** multiply_list = new sum* [2];
  if (multiply_list == NULL)
    die("allocating multiply_list");
  long long one = 1;
  for (int i = 0; i < dimension; i++) {               // contract
    index_list[0] = i;
    index_list[1] = i;
    current_component = sub(index_list);
    if (current_component == NULL)
      continue;
    multiply_list[0] = current_component->value;
    new_sum->multiply(1, multiply_list, &one);}
  delete [] multiply_list;
  new_sum->simplify_invariant();
  return new_sum; }

extern sum* ntuple::scalar_invariant(int number_of_ntuples, ntuple** ntuple_list, 
				     int* index_matrix, bool simplifyit) {
#ifdef tracentry
  trace_nsi += 1;
#endif
  // each index list in index_matrix is rank long and has summation ordinals in each position
  int number_of_free_indices = 0;
  int number_of_sum_indices = 0;
  int current_index, mult, max_rank = 0;
  for (int n = 0; n < number_of_ntuples; n++) {
#ifdef debug
    if (ntuple_list[n]->ntuple_magic_no != ntuple_magic)
      die("attempting to create a scalar invariant with a non-ntuple");
#endif
    if (ntuple_list[n]->rank > max_rank)
      max_rank = ntuple_list[n]->rank;}
  for (int n = 0; n < number_of_ntuples; n++) {
    for (int j = 0; j < ntuple_list[n]->rank; j++) {
      current_index = index_matrix[n * max_rank + j];
      if (current_index < 0)
	number_of_free_indices += 1;
      else
	number_of_sum_indices += 1;}}
#ifdef debug
  if (number_of_free_indices > 0)
    die("attempting to create a scalar invariant with free indices");
  if ((number_of_sum_indices % 2) != 0)
    die("improper contraction indices");
#endif
  number_of_sum_indices /= 2;
#ifdef debug
  for (int n = 0; n < number_of_ntuples; n++)
    for (int i = 0; i < ntuple_list[n]->rank; i++) {
      current_index = index_matrix[n * max_rank + i];
      if (current_index > number_of_sum_indices)
	die("contraction index too large");}
#endif
  int* check_upper = new int [number_of_sum_indices];
  if (check_upper == NULL)
    die("allocating check_upper");
  memset(check_upper, 0, number_of_sum_indices * sizeof(int));
  for (int n = 0; n < number_of_ntuples; n++)
    for (int j = 0; j < ntuple_list[n]->rank; j++) {
      current_index = index_matrix[n * max_rank + j];
#ifdef debug
      if ((current_index < 0) || (current_index >= rank))
	die("scalar_invariant - index out of bounds");
#endif
	if (ntuple_list[n]->upper[j])
	  check_upper[current_index] += 1;}
#ifdef debug
  for (int i = 0; i < number_of_sum_indices; i++) {
    if (check_upper[i] != 1)
      die("attempting to contract same height indices");}
#endif
  sum* new_sum = new sum(large_hash_size);
  if (new_sum == NULL)
    die("allocating new_sum");
  for (int n = 0; n < number_of_ntuples; n++)
    for (int i = 0; i < ntuple_list[n]->rank; i++)
      ntuple_list[n]->index_list[i] = index_matrix[n * max_rank + i];
  mult = ntuple_list[0]->contract((number_of_ntuples - 1), &(ntuple_list[1]), new_sum, true);
  if (mult != 1)
    new_sum->scalar_multiply(mult);
  if (simplifyit)
    new_sum->simplify_invariant();
  delete [] check_upper;
  return new_sum; }
