//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// d10Bsugra.cc

#include <ptah.h>
#include <element.h>
#include <hash.h>
#include <ntuple.h>
#include <d10Bsugra.h>

extern ntuple* d10Bdilatino () {
  int dpind[4] = {0, -1,
		  -1, 0};
  int fs1ind[9] = {-1, 0, 1,
		    -1, 0, -1,
		    -1, 1, -1};
  int fs2ind[6] = {0, -1, -1,
		   -1, 0, -1};
  int abcind[3] = {0, 1, 2};
  int dind[4] = {-1, 0,
		 0, -1};
  int c1o12[number_of_fractions] = {-2, -1};
  bool fsanti[3] = {true, false, false};
  long long negone = -1;
  int expmphio2_powers[17] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, -4, 0, 0, 0, 0, 0};
  sum expmphio2;
  ntuple *dphilu, *fssllluu, *fsllluuu, *fsgam3, *empfsgam3, *dilino, *dilinop;
  ntuple *nlist[3];
  nlist[0] = dphi;
  nlist[1] = vielinv;
  dphilu = dphi->ntuple_contract(2, nlist, dpind);
  to_stderr ("dphilu completed.");
  dphilu->fast_lorentz_raise(0);
  nlist[0] = fs;
  nlist[1] = vielinv;
  nlist[2] = vielinv->clone();
  fssllluu = fs->ntuple_contract(3, nlist, fs1ind);   // indices are st, lor, lor
  fssllluu->fast_lorentz_raise(1);
  fssllluu->fast_lorentz_raise(2);
  to_stderr ("fssllluu completed.");
  nlist[2]->zero();
  delete nlist[2];
  nlist[0] = fssllluu;
  nlist[1] = vielinv;
  fsllluuu = fssllluu->ntuple_contract(2, nlist, fs2ind);  // indices are up, up, low
  fssllluu->zero();
  delete fssllluu;
  fsllluuu->fast_lorentz_raise(2);
  for (int i = 0; i < 3; i++)
    fsllluuu->antisymmetric[i] = fsanti[i];
  to_stderr ("fsllluuu completed.");
  fsgam3 = fsllluuu->gamma(3, abcind);   // indices are sp, sp
  fsllluuu->zero();
  delete fsllluuu;
  fsgam3->scalar_multiply(c1o12);
  expmphio2.add(&negone, expmphio2_powers);
  empfsgam3 = fsgam3->multiply(&expmphio2);
  fsgam3->zero();
  delete fsgam3;
  to_stderr ("fsgam3 completed.");
  dilino = dphilu->gamma(1, abcind);   // indices are sp, sp
  dphilu->zero();
  delete dphilu;
  dilino->add(empfsgam3);
  empfsgam3->zero();
  delete empfsgam3;
  nlist[0] = dilino;
  nlist[1] = susyparm;
  dilinop = dilino->ntuple_contract(2, nlist, dind);    // index is sp
  to_stderr ("dilatino completed.");
  expmphio2.zero();
  dilino->zero();
  delete dilino;
  return dilinop;
}

extern ntuple* d10Bgravitino () {
  long long one = 1;
  int abind[3] = {-1, 0, 1};
  int abcdind[3] = {1, 2, 3};
  int gind[6] = {-1, -1, 0,
		 0, -1, -1};
  int fs1ind[9] = {-1, 0, 1,
		    -1, 0, -1,
		    -1, 1, -1};
  int fs2ind[6] = {0, -1, -1,
		   -1, 0, -1};
  int fs3ind[6] = {0, -1, -1,
		   0, -1, -1};
  int fs3bind[6] = {-1, -1, 0,
		    0, -1, -1,};
  int c1o4[number_of_fractions] = {-2, 0};
  int c1o96[number_of_fractions] = {-5, -1};
  bool fsanti[3] = {true, false, false};
  int expmphio2_powers[17] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, -4, 0, 0, 0, 0, 0};
  sum expmphio2;
  ntuple *grav, *omegam, *cdsp, *fssllluu, *fsllluuu, *fsgam3, *fsgam4, *fsgam4c, *empfsgam4c, *fsgam4ds;
  ntuple *nlist[3];
  omegam = omega->gamma(2, abind);   // indices are st, sp, sp
  to_stderr ("omegam completed.");
  nlist[0] = omegam;
  nlist[1] = susyparm;
  grav = omegam->ntuple_contract(2, nlist, gind);  // indices are st, sp
  omegam->zero();
  delete omegam;
  grav->scalar_multiply(c1o4);
  cdsp = susyparm->covariant_derivative();       // indices are sp, st
  cdsp->transpose();  // now indices are st, sp
  grav->add(cdsp);
  cdsp->zero();
  delete cdsp;
  nlist[0] = fs;
  nlist[1] = vielinv;
  nlist[2] = vielinv->clone();
  fssllluu = fs->ntuple_contract(3, nlist, fs1ind);  // indices are st, lor, lor
  fssllluu->fast_lorentz_raise(1);
  fssllluu->fast_lorentz_raise(2);
  to_stderr ("fssllluu completed.");
  nlist[2]->zero();
  delete nlist[2];
  nlist[0] = fssllluu;
  nlist[1] = vielinv;
  fsllluuu = fssllluu->ntuple_contract(2, nlist, fs2ind);  // indices are up, up, low
  fsllluuu->fast_lorentz_raise(2);
  for (int i = 0; i < 3; i++)
    fsllluuu->antisymmetric[i] = fsanti[i];
  to_stderr ("fsllluuu completed.");
  fsgam4 = fsllluuu->gamma(4, abcdind);   // indices are lor, sp, sp
  fsllluuu->zero();
  delete fsllluuu;
  nlist[0] = viel;
  nlist[1] = fsgam4;
  fsgam4c = fsgam4->ntuple_contract(2, nlist, fs3ind);   // indices are st, sp, sp
  fsgam4->zero();
  delete fsgam4;
  fsgam3 = fssllluu->gamma(2, abind);  // indices are st, sp, sp
  fssllluu->zero();
  delete fssllluu;
  fsgam3->scalar_multiply(-9);
  fsgam4c->add(fsgam3);
  fsgam3->zero();
  delete fsgam3;
  expmphio2.add(&one, expmphio2_powers);
  empfsgam4c = fsgam4c->multiply(&expmphio2);
  fsgam4c->zero();
  delete fsgam4c;
  empfsgam4c->scalar_multiply(c1o96);
  nlist[0] = empfsgam4c;
  nlist[1] = susyparm;
  fsgam4ds = empfsgam4c->ntuple_contract(2, nlist, fs3bind);  // indices are st, sp
  empfsgam4c->zero();
  delete empfsgam4c;
  grav->add(fsgam4ds);
  fsgam4ds->zero();
  delete fsgam4ds;
  to_stderr ("grav completed.");
  expmphio2.zero();
  return grav; }

extern sum* d10Bdilaton () {
  long long one = 1;
  int inds[3] = {0, 1, 2};
  int fsind[6] = {0, 1, 2,
		  0, 1, 2};
  int c1o12[number_of_fractions] = {-2, -1};
  sum *fssq, fssqep, *dil;
  sum *mlist[2];
  ntuple *cddphi, *cddphiu, *fsuuu;
  ntuple* nlist[2];
  int expmphi_powers[17] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8, -8, 0, 0, 0, 0, 0};
  sum expmphi;
  cddphi = dphi->covariant_derivative();
  cddphiu = cddphi->raise(1, inds);
  cddphi->zero();
  delete cddphi;
  fsuuu = fs->raise(3, inds);
  nlist[0] = fs;
  nlist[1] = fsuuu;
  fssq = fs->scalar_invariant(2, nlist, fsind, true);
  fsuuu->zero();
  delete fsuuu;
  expmphi.add(&one, expmphi_powers);
  mlist[0] = fssq;
  mlist[1] = &expmphi;
  fssqep.multiply(2, mlist, &one);
  fssq->zero();
  delete fssq;
  fssqep.scalar_multiply(c1o12);
  expmphi.zero();
  dil = cddphiu->scalar();
  cddphiu->zero();
  delete cddphiu;
  mlist[0] = &fssqep;
  dil->multiply(1, mlist, &one);
  fssqep.zero();
  return dil;
}

extern ntuple* d10Bgauge () {
  long long one = 1;
  int rind[1] = {3};
  ntuple *epfs, *empepfs, *cdepfs, *cdepfsu, *feq;
  int expmphi_powers[17] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8, -8, 0, 0, 0, 0, 0};
  sum expmphi;
  epfs = fs->clone();
  expmphi.add(&one, expmphi_powers);
  empepfs = epfs->multiply(&expmphi);
  expmphi.zero();
  epfs->zero();
  delete epfs;
  cdepfs = empepfs->covariant_derivative();
  empepfs->zero();
  delete empepfs;
  cdepfsu = cdepfs->raise(1, rind);
  cdepfs->zero();
  delete cdepfs;
  feq = cdepfsu->self_contract(0, 3);
  cdepfsu->zero();
  delete cdepfsu;
  return feq; }

extern ntuple* d10Bgraviton () {
  long long one = 1;
  int rind[1] = {1};
  int fsind[6] = {0, 1, 2,
		  0, 1, 2};
  int fsijind[6] = {-1, 0, 1,
		    -1, 0, 1};
  int inds[3] = {0, 1, 2};
  int c1o2[number_of_fractions] = {-1, 0};
  int c1o4[number_of_fractions] = {-2, 0};
  int c1o12[number_of_fractions] = {-2, -1};
  int expmphi_powers[17] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8, -8, 0, 0, 0, 0, 0};
  sum expmphi, *dphsq;
  bool falseparm[2] = {false, false};
  bool uparm[2] = {false, true};
  sum *fssq;
  sum *mlist[2];
  ntuple *fsuuu, *geq, *fsifsjlu, *empfsifsjlu, *fssqij, *dphij, *dphijlu;
  ntuple *nlist[2];
  geq = ricci->raise(1, rind);
  fsuuu = fs->raise(3, inds);
  nlist[0] = fs;
  nlist[1] = fsuuu;
  fssq = fs->scalar_invariant(2, nlist, fsind, true);
  fssq->scalar_multiply(c1o12);
  fssq->scalar_multiply(-1);
  nlist[0] = fs;
  nlist[1] = fsuuu;
  fsifsjlu = fs->ntuple_contract(2, nlist, fsijind);
  fsuuu->zero();
  delete fsuuu;
  fssqij = new ntuple(2, uparm, falseparm, falseparm, falseparm);
  for (int i = 0; i < dimension; i++) {
    inds[0] = i;
    inds[1] = i;
    mlist[0] = fssq;
    mlist[1] = new sum;
    mlist[1]->multiply(1, mlist, &one);
    fssqij->set(inds, mlist[1]);}
  fsifsjlu->add(fssqij);
  fssqij->zero();
  delete fssqij;
  expmphi.add(&one, expmphi_powers);
  empfsifsjlu = fsifsjlu->multiply(&expmphi);
  expmphi.zero();
  fsifsjlu->zero();
  delete fsifsjlu;
  empfsifsjlu->scalar_multiply(c1o4);
  empfsifsjlu->scalar_multiply(-1);
  geq->add(empfsifsjlu);
  empfsifsjlu->zero();
  delete empfsifsjlu;
  dphij = new ntuple(2, falseparm, falseparm, falseparm, falseparm);
  for (int i = 0; i < dimension; i++)
    for (int j = 0; j < dimension; j++) {
      inds[0] = i;
      inds[1] = j;
      mlist[0] = (dphi->sub(&(inds[0])))->value;
      mlist[1] = (dphi->sub(&(inds[1])))->value;
      dphsq = new sum;
      dphsq->multiply(2, mlist, &one);
      dphij->set(inds, dphsq);}
  dphijlu = dphij->raise(1, rind);
  dphij->zero();
  delete dphij;
  dphijlu->scalar_multiply(c1o2);
  dphijlu->scalar_multiply(-1);
  geq->add(dphijlu);
  dphijlu->zero();
  delete dphijlu;
  return geq; }
