//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// d11sugra.cc

#include <ptah.h>
#include <element.h>
#include <hash.h>
#include <ntuple.h>
#include <d11sugra.h>

extern ntuple* d11gravitino () {
  int abind[3] = {-1, 0, 1};
  int abcdeind[4] = {1, 2, 3, 4};
  int abcind[4] = {-1, 0, 1, 2};
  int gind[6] = {-1, -1, 0,
		 0, -1, -1};
  int fs1ind[16] = {-1, 0, 1, 2,
		    -1, 0, -1, -1,
		    -1, 1, -1, -1,
		    -1, 2, -1, -1};
  int fs2ind[8] = {0, -1, -1, -1, 
		    -1, 0, -1, -1};
  int fs3ind[9] = {0, -1, 1,
		   0, -1, -1,
		   1, -1, -1,};
  int fs3bind[6] = {-1, -1, 0,
		    0, -1, -1,};
  int c1o4[number_of_fractions] = {-2, 0};
  int c1o18[number_of_fractions] = {-1, -2};
  int c1o144[number_of_fractions] = {-4, -2};
  bool fsanti[4] = {true, false, true, false};
  ntuple *grav, *omegam, *cdsp, *fsslllluuu, *fslllluuuu, *fsgam5, *fsgam5ds, *fsgam3, *fsgam3ds;
  ntuple *nlist[4];
  omegam = omega->gamma(2, abind);   // indices are st, sp, sp
  to_stderr ("omegam completed.");
  omegam->display("omegam");
  nlist[0] = omegam;
  nlist[1] = susyparm;
  grav = omegam->ntuple_contract(2, nlist, gind);  // indices are st, sp
  to_stderr ("grav completed.");
  omegam->zero();
  delete omegam;
  grav->scalar_multiply(c1o4);
  cdsp = susyparm->covariant_derivative();       // indices are sp, st
  cdsp->display("cdsp");
  cdsp->transpose();  // now indices are st, sp
  grav->add(cdsp);
  cdsp->zero();
  delete cdsp;
  nlist[0] = fs;
  nlist[1] = vielinv;
  nlist[2] = vielinv->clone();
  nlist[3] = vielinv->clone();
  fsslllluuu = fs->ntuple_contract(4, nlist, fs1ind);
  to_stderr ("fsslllluuu completed.");
  fsslllluuu->fast_lorentz_raise(1);
  fsslllluuu->fast_lorentz_raise(2);
  fsslllluuu->fast_lorentz_raise(3);
  fsslllluuu->antisymmetric[2] = true;
  nlist[0] = fsslllluuu;
  fslllluuuu = fsslllluuu->ntuple_contract(2, nlist, fs2ind);  // introduces -1 (look at free indices!)
  to_stderr ("fslllluuuu completed.");
  fslllluuuu->fast_lorentz_raise(3);
  for (int i = 0; i < 4; i++)
    fslllluuuu->antisymmetric[i] = fsanti[i];
  fslllluuuu->display("fslllluuuu");
  for (int i = 1; i < 4; i++) {
    nlist[i]->zero();
    delete nlist[i];}
  fsgam5 = fslllluuuu->gamma(5, abcdeind);  // indices are lor, sp, sp
  to_stderr ("fsgam5 completed.");
  fsgam5->display("fsgam5");
  nlist[0] = fsgam5;
  nlist[1] = viel;
  nlist[2] = susyparm;
  fsgam5ds = fsgam5->ntuple_contract(3, nlist, fs3ind);    // indices are sp, st
  to_stderr ("fsgam5ds completed.");
  fsgam5->zero();
  delete fsgam5;
  fsgam5ds->transpose();         // indices are now st, sp
  fsgam5ds->scalar_multiply(-1);
  fsgam5ds->scalar_multiply(c1o144);
  grav->add(fsgam5ds);
  fsgam5ds->zero();
  delete fsgam5ds;
  fslllluuuu->zero();
  delete fslllluuuu;
  fsgam3 = fsslllluuu->gamma(3, abcind);  // indices are st, sp, sp
  to_stderr ("fsgam3 completed.");
  nlist[0] = fsgam3;
  nlist[1] = susyparm;
  fsgam3ds = fsgam3->ntuple_contract(2, nlist, fs3bind);   // indices are st, sp
  to_stderr ("fsgam3ds completed.");
  fsgam3->zero();
  delete fsgam3;
  fsgam3ds->scalar_multiply(-1);
  fsgam3ds->scalar_multiply(c1o18);
  grav->add(fsgam3ds);
  fsgam3ds->zero();
  delete fsgam3ds;
  fsslllluuu->zero();
  delete fsslllluuu;
  return grav; }

extern ntuple* d11gauge () {
  int fsind[14] = {-1, -1, -1, 0, 1, 2, 3,
		   0, 1, 2, 3, -1, -1, -1};
  int rind[5] = {0, 1, 2, 3, 4};
  int c1o576[number_of_fractions] = {-6, -2};
  ntuple *cdfs, *cdfsu, *flhs, *fsd, *frhs, *feq;
  ntuple *nlist[2];
  cdfs = fs->covariant_derivative();
  cdfsu = cdfs->raise(1, &(rind[4]));
  cdfs->zero();
  delete cdfs;
  flhs = cdfsu->self_contract(0, 4);
  cdfsu->zero();
  delete cdfsu;
  fsd = fs->dual(4, &(fsind[3]));
  nlist[0] = fsd;
  nlist[1] = fs;
  frhs = fsd->ntuple_contract(2, nlist, fsind);
  fsd->zero();
  delete fsd;
  frhs->scalar_multiply(c1o576);
  feq = flhs->raise(3, rind);
  flhs->zero();
  delete flhs;
  feq->add(frhs);
  frhs->zero();
  delete frhs;
  return feq; }

extern ntuple* d11graviton () {
  int rind[4] = {0, 1, 2, 3};
  int fsind[8] = {-1, 0, 1, 2,
		  -1, 0, 1, 2};
  int sind[8] = {0, 1, 2, 3,
		 0, 1, 2, 3};
  int c1o2[number_of_fractions] = {-1, 0};
  int c1o3[number_of_fractions] = {0, -1};
  int c1o24[number_of_fractions] = {-3, -1};
  sum *fssq;
  ntuple *fsluuu, *geq, *sp, *fsifsj, *fssqt;
  ntuple *nlist[2];
  fsluuu = fsuuuu->lower(1, &(rind[0]));
  geq = ricci->clone();
  ricci->zero();
  delete ricci;
  sp = g->multiply(scacurv);
  scacurv->zero();
  delete scacurv;
  sp->scalar_multiply(-1);
  sp->scalar_multiply(c1o2);
  geq->add(sp);
  sp->zero();
  delete sp;
  nlist[0] = fsluuu;
  nlist[1] = fs;
  fsifsj = fsluuu->ntuple_contract(2, nlist, fsind);
//    fsifsj->display("fsifsj");
  fsluuu->zero();
  delete fsluuu;
  fsifsj->scalar_multiply(-1);
  fsifsj->scalar_multiply(c1o3);
  geq->add(fsifsj);
  fsifsj->zero();
  delete fsifsj;
  nlist[0] = fsuuuu;
  fssq = fsuuuu->scalar_invariant(2, (ntuple **) nlist, (int *) sind, true);
  fsuuuu->zero();
  delete fsuuuu;
  fssq->display("fssq");
  fssqt = g->multiply(fssq);
  fssq->zero();
  delete fssq;
  fssqt->scalar_multiply(c1o24);
  geq->add(fssqt);
  fssqt->zero();
  delete fssqt;
  return geq; }
