//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// dual.cc

#include <ptah.h>
#include <element.h>
#include <hash.h>
#include <ntuple.h>

extern ntuple* ntuple::dual(int number_to_dualize, int* index_list_param) {
#ifdef tracentry
  trace_nd += 1;
#endif
  int index;
  int mult = 1;
  bool skip, indexq = true;
#ifdef trace_iter
  char str [256];
#endif
#ifdef debug
  if ((number_to_dualize < 1) || (number_to_dualize > dimension))
    die("invalid number of indices in dual");
  for (int i = 0; i < number_to_dualize; i++)
    indexq = indexq && !upper[index_list_param[i]];
  if (!indexq)
    die("attempting to dualize on upper indices");
#endif
  index = 0;                                    // count number of antisymmetric index pairs
  for (int i = 0; i < (number_to_dualize - 1); i++)
    if (antisymmetric[index_list_param[i]] && ((index_list_param[i + 1]) == (index_list_param[i] + 1)))
      index += 1;
  for (int i = 0; i < index; i++)
    mult *= 2;
  int new_rank = dimension + rank - 2 * number_to_dualize;
  int* new_index_list = new int [new_rank];
  if (new_index_list == NULL)
    die("allocating new_index_list");
  bool* new_upper = new bool [new_rank];
  if (new_upper == NULL)
    die("allocating new_upper");
  bool* new_antisymmetric = new bool [new_rank];
  if (new_antisymmetric == NULL)
    die("allocating new_antisymmetric");
  bool* new_symmetric = new bool [new_rank];
  if (new_symmetric == NULL)
    die("allocating new_symmetric");
  bool* new_spinor = new bool [new_rank];
  if (new_spinor == NULL)
    die("allocating new_spinor");
  long long overall_coefficient = 0;
  sum* new_sum = new sum;
  if (new_sum == NULL)
    die("allocating new_sum");
  sum** multiply_list = new sum* [2];
  if (multiply_list == NULL)
    die("allocating multiply_list");
  index = 0;                                      // do new free indices first
  for (int i = 0; i < (dimension - number_to_dualize); i++) { 
#ifdef debug
    if ((index < 0) || (index >= new_rank))
      die("dual - index out of bounds");
#endif
    new_upper[index] = true;
    if ((index % 2) == 0)
      new_antisymmetric[index] = true;
    else
      new_antisymmetric[index] = true;
    new_symmetric[index] = false;
    new_spinor[index] = false;
    index += 1;}
  if ((index % 2) == 1)                    // make correction if ended on even index
    new_antisymmetric[index - 1] = false;
  for (int i = 0; i < rank; i++) {         // now do old free indices
    indexq = false;
    for (int j = 0; j < number_to_dualize; j++)
      indexq = indexq || (i == index_list_param[j]);
    if (indexq)                            // skip contracted indices
      continue;
#ifdef debug
    if ((index < 0) || (index >= rank))
      die("dual - index out of bounds (2)");
#endif
    new_upper[index] = upper[i];
    new_antisymmetric[index] = antisymmetric[i];
    new_symmetric[index] = symmetric[i];
    new_spinor[index] = spinor[i];
    index += 1;}
  ntuple* new_ntuple = new ntuple(new_rank, new_upper, new_antisymmetric, new_symmetric, new_spinor);
  if (new_ntuple == NULL)
    die("allocating new_ntuple");
  multiply_list[1] = inverse_sqrt_det_g;
  for (int i = 0; i < rank; i++)
    index_list[i] = -1;
  for (int i = 0; i < number_to_dualize; i++) {
#ifdef debug
    if ((index_list_param[i] < 0) || (index_list_param[i] >= rank))
      die("dual - index parameter out of bounds");
#endif
    index_list[index_list_param[i]] = 1;}
  epsilon_start(dimension);
  make_free_index_list(true);
  while (free_parallelize(new_ntuple)) {
    for (int i = 0; i < rank; i++)
      index_list[i] = free_index(i);
    for (int i = 0; i < epsilon_length; i++) {
      for (int j = 0; j < number_to_dualize; j++) {
#ifdef debug
	if ((index_list_param[j] < 0) || (index_list_param[j] >= rank))
	  die("dual - index parameter out of bounds (2)");
#endif
	index_list[index_list_param[j]] = epsilon_index(i, dimension + j - number_to_dualize);}
      for (int j = 0; j < (dimension - number_to_dualize); j++)     // copy uncontracted epsilon indices
	new_index_list[j] = epsilon_index(i, j);
      index = dimension - number_to_dualize;
      for (int j = 0; j < rank; j++) {               // copy uncontracted ntuple indices
	indexq = false;
	for (int k = 0; k < number_to_dualize; k++)
	  indexq = indexq || (j == index_list_param[k]);
	if (indexq)                                  // skip contracted indices
	  continue;
#ifdef debug
	if ((index < 0) || (index >= rank))
	  die("dual - index out of bounds (2)");
#endif
	new_index_list[index] = index_list[j];
	index += 1;}
      skip = false;
      for (int j = 0; j < rank - 1; j++)
	if (antisymmetric[j] && (index_list[j] > index_list[j + 1])) {
	  skip = true;
	  break;}
      for (int j = 0; j < rank - 1; j++)
	if (antisymmetric[j] && (new_index_list[j] > new_index_list[j + 1])) {
	  skip = true;
	  break;}
      if (!skip) {
	current_component = sub(index_list);
	if (current_component != NULL) {
	  multiply_list[0] = current_component->value;
	  overall_coefficient = epsilon_sign(i) * mult;
	  new_sum->multiply(2, multiply_list, &overall_coefficient);
#ifdef trace_iter
	  sprintf(str, "Processing current_free_index = %d; indices = ", current_free_index);
	  for (int j = 0; j < new_ntuple->rank; j++)
	    sprintf(str + strlen(str), "%d ", new_index_list[j]);
	  to_stderr (str);
#endif
	  new_ntuple->add(new_index_list, new_sum);
	  new_sum->zero();}}}}
  new_ntuple->simplify();
  new_ntuple->fill();
  epsilon_stop();
  delete new_sum;
  delete [] new_upper;
  delete [] new_antisymmetric;
  delete [] new_symmetric;
  delete [] new_spinor;
  delete [] multiply_list;
  delete [] new_index_list;
  return new_ntuple; }

extern sum* ntuple::dual_invariant(int number_of_ntuples, ntuple** ntuple_list, 
				   int* index_matrix, bool simplifyit) {
  // THIS must have the largest rank of any in the ntuple list
#ifdef tracentry
  trace_ndi += 1;
#endif
  // each index list in index_matrix is this->rank long and has epsilon number in each position
  int number_of_sum_indices = 0;
  int current_index;
#ifdef trace_iter
  char str [256];
#endif
  for (int n = 0; n < number_of_ntuples; n++) {
#ifdef debug
    if (ntuple_list[n]->ntuple_magic_no != ntuple_magic)
      die("attempting to create a dual invariant with a non-ntuple");
#endif
    for (int j = 0; j < ntuple_list[n]->rank; j++) {
      number_of_sum_indices += 1;}}
  number_of_epsilons = number_of_sum_indices / dimension;
#ifdef debug
  if ((number_of_sum_indices % dimension) != 0)
    die("improper contraction indices");
  for (int n = 0; n < number_of_ntuples; n++)
    for (int i = 0; i < ntuple_list[n]->rank; i++) {
      current_index = index_matrix[n * rank + i];
      if (current_index > (number_of_epsilons - 1))
	die("epsilon index too large");}
  for (int n = 0; n < number_of_ntuples; n++)
    for (int j = 0; j < ntuple_list[n]->rank; j++) {
	if (ntuple_list[n]->upper[j])
	  die("attempting to dualize upper indices");}
#endif
  int ep, number_to_multiply;
  number_to_multiply = number_of_ntuples + number_of_epsilons;
  which_epsilon = new int [number_of_epsilons];  // which value we're looking at in a given symbol
  if (which_epsilon == NULL)
    die("allocating which_epsilon");
  int* epsilon_indices = new int [number_of_epsilons];  // which index we're looking at in a given symbol
  if (epsilon_indices == NULL)
    die("allocating epsilon_indices");
  bool zero_factor;
  sum* new_sum = new sum(large_hash_size);
  if (new_sum == NULL)
    die("allocating new_sum");
  sum** multiply_list = new sum* [number_of_ntuples + number_of_epsilons];
  if (multiply_list == NULL)
    die("allocating multiply_list");
  long long mult = 1;
  epsilon_start(dimension);
  ep = 0;
  for (int n = 0; n < number_of_ntuples; n++)
    for (int j = 0; j < ntuple_list[n]->rank; j++)
      if (ntuple_list[n]->antisymmetric[j]) {
	ep += 1;
	j += 1;}
  for (int i = 0; i < ep; i++)
    mult *= 2;
  long long overall_coefficient = 0;
  for (int i = number_of_ntuples; i < number_of_ntuples + number_of_epsilons; i++)
    multiply_list[i] = inverse_sqrt_det_g;
  while (parallelize_dual(new_sum)) {
#ifdef trace_iter
    strcpy(str, "Processing epsilons: ");
    for (int w = 0; w < number_of_epsilons; w++)
      sprintf(str + strlen(str), "%d ", which_epsilon[w]);
    to_stderr (str);
#endif
    memset(epsilon_indices, 0, number_of_epsilons *sizeof(int));
    for (int n = 0; n < number_of_ntuples; n++)       // set contraction indices
      for (int j = 0; j < ntuple_list[n]->rank; j++) {
	current_index = index_matrix[n * rank + j];   // which symbol to use for this ntuple index
	ep = which_epsilon[current_index];            // which specific value to use
	ntuple_list[n]->index_list[j] = epsilon_index(ep, epsilon_indices[current_index]);
	epsilon_indices[current_index] += 1;
#ifdef debug
	if (epsilon_indices[current_index] > dimension)
	  die("epsilon_indices overflow");
#endif
      }
    zero_factor = false;
    for (int n = 0; n < number_of_ntuples; n++) {     // now do the math
#ifndef abbreviated
      for (int j = 0; j < (ntuple_list[n]->rank - 1); j++)
	if (ntuple_list[n]->antisymmetric[j]) {
	  if ((ntuple_list[n]->index_list[j] > ntuple_list[n]->index_list[j + 1]) &&
	      (index_matrix[n * rank + j] == index_matrix[n * rank + j + 1])) {
	    zero_factor = true;
	    break;}
	  j += 1;}
      if (zero_factor)
	break;
#endif
      current_component = ntuple_list[n]->sub(ntuple_list[n]->index_list);
      if (current_component == NULL) {
	zero_factor = true;
	break;}
      multiply_list[n] = current_component->value;}
    if (!zero_factor) {
      overall_coefficient = mult;
      for (int j = 0; j < number_of_epsilons; j++) {
	overall_coefficient *= epsilon_sign(which_epsilon[j]);}
      new_sum->multiply(number_to_multiply, multiply_list, &overall_coefficient);}}
  if (simplifyit)
    new_sum->simplify_invariant();
  epsilon_stop();
  delete [] which_epsilon;
  delete [] epsilon_indices;
  delete [] multiply_list;
  return new_sum; }
