//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// epsilon.cc

#include <ptah.h>
#include <stdio.h>
#include <errno.h>

extern int epsilon_length;

#ifdef epsilons_in_memory
long long *epsilon_matrix;
#endif

int rank, last;

char fn[22] = "/home/ptahusr/eps.xx";

long long eps;

FILE *fd = NULL;

extern void epsilon_start (int dim) {
#ifdef tracentry
  trace_eb += 1;
#endif
  int slen, rc;
  if ((dim < 2) || (dim > 11))
    die("invalid epsilon dimension");
  rank = dim;
  last = -1;
  slen = sprintf(&(fn[18]), "%d", dim) + 18;
  if ((slen < 19) || (slen > 20))
    die("creating epsilon filename");
  fd = fopen(fn, "r");
  if (fd == NULL) {
    perror("opening epsilon file");
    die ("opening epsilon file", errno);}
  epsilon_length = 1;
  for (int i = 2; i <= dim; i++)
    epsilon_length *= i;
#ifdef abbreviated
  if ((dim % 2) == 0)
    for (int i = 1; i <= (dim/2); i++)
      epsilon_length /= 2;
#endif
#ifdef epsilons_in_memory
  epsilon_matrix = new long long [epsilon_length];
  if (epsilon_matrix == NULL)
    die("creating epsilon_matrix");
  rc = fread(epsilon_matrix, sizeof(long long), epsilon_length, fd);
  if (rc != epsilon_length) {
    perror("reading epsilon file into matrix");
    die ("reading epsilon file", errno);}
  rc = 0;
  for (eps = 0; eps < epsilon_length; eps++)
    if ((epsilon_matrix[eps] >> 50) == 0)
      rc += 1;
    else
      rc -= 1;
#ifdef abbreviated
  for (int i = 2; i <= (dim/2); i++)
    rc /= i;
  if (rc != 1)
    die ("epsilon matrix failed consistency check");
#else
  if (rc != 0)
    die ("epsilon matrix failed consistency check");
#endif
#endif
 }

extern void epsilon_stop () {
#ifdef tracentry
  trace_ee += 1;
#endif
  int rc;
  last = -1;
  rc = fclose(fd);
  if (rc != 0) {
    perror("closing epsilon file");
    die ("closing epsilon file", errno);}
  fd = NULL;
#ifdef epsilons_in_memory
  delete [] epsilon_matrix;
#endif
  epsilon_length = 0; }

extern int epsilon_index (int which, int ind) {
#ifdef tracentry
  trace_epi += 1;
#endif
#ifndef epsilons_in_memory
  int rc;
  long offset;
  if (which != last) {
    offset = which * sizeof(long long);
    rc = fseek(fd, offset, SEEK_SET);
    if (rc != 0) {
      perror("seeking epsilon file");
      die ("seeking epsilon file", errno);}
    rc = fread(&eps, sizeof(long long), 1, fd);
    if (rc != 1) {
      perror("reading epsilon file");
      die ("reading epsilon file", errno);}
    last = which;}
#endif
#ifdef epsilons_in_memory
  eps = epsilon_matrix[which];
#endif
  return ((eps >> (4 * ind)) & 15); }

extern int epsilon_sign (int which) {
#ifdef tracentry
  trace_es += 1;
#endif
  int rc;
#ifndef epsilons_in_memory
  long offset;
  if (which != last) {
    offset = which * sizeof(long long);
    rc = fseek(fd, offset, SEEK_SET);
    if (rc != 0) {
      perror("seeking epsilon file");
      die ("seeking epsilon file", errno);}
    rc = fread(&eps, sizeof(long long), 1, fd);
    if (rc != 1) {
      perror("reading epsilon file");
      die ("reading epsilon file", errno);}
    last = which;}
#endif
#ifdef epsilons_in_memory
  eps = epsilon_matrix[which];
#endif
  rc = (eps >> 50);
  if (rc == 0)
    return 1;
  else
    return -1; }
