//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// gamma.cc

#include <ptah.h>
#include <stdio.h>
#include <errno.h>
#include <element.h>
#include <hash.h>
#include <ntuple.h>

char fni[33] = "/home/ptahusr/gammaxx.abcdefghij";
char suf[11] = "abcdefghij";

int lastg, thisorder;

long long gam;

FILE *fdi = NULL;

extern void gamma_start (int order) {
#ifdef tracentry
  trace_gb += 1;
#endif
  int slen;
  if ((order < 1) || (order > dimension))
    die("invalid gamma order");
  lastg = -1;
  thisorder = order;
  memcpy(&(fni[22]), suf, 11);
  slen = sprintf(&(fni[19]), "%2d", dimension);
  if (slen != 2)
    die("creating gamma input filename");
  fni[21] = '.';
  fni[22 + order] = 0x00;
  fdi = fopen(fni, "r");
  if (fdi == NULL) {
    perror("opening gamma input file");
    die ("opening gamma input file", errno);} }

extern void gamma_stop () {
#ifdef tracentry
  trace_ge += 1;
#endif
  int rc;
  lastg = -1;
  rc = fclose(fdi);
  if (rc != 0) {
    perror("closing gamma input file");
    die ("closing gamma input file", errno);}
  fdi = NULL; }

extern int gamma_index (int which, int ind) {
#ifdef tracentry
  trace_gin += 1;
#endif
  int rc;
  long offset;
  if (which != lastg) {
    offset = which * sizeof(long long);
    rc = fseek(fdi, offset, SEEK_SET);
    if (rc != 0) {
      perror("seeking gamma file");
      die ("seeking gamma file", errno);}
    rc = fread(&gam, sizeof(long long), 1, fdi);
    if (rc != 1) {
      perror("reading gamma file");
      die ("reading gamma file", errno);}
    lastg = which;}
  if (ind >= thisorder)
    return ((gam >> (8 * (ind - thisorder) + 4 * thisorder)) & 255);
  else
    return ((gam >> (4 * ind)) & 15); }

extern int gamma_sign (int which) {
#ifdef tracentry
  trace_gs += 1;
#endif
  int rc;
  long offset;
  if (which != lastg) {
    offset = which * sizeof(long long);
    rc = fseek(fdi, offset, SEEK_SET);
    if (rc != 0) {
      perror("seeking gamma file");
      die ("seeking gamma file", errno);}
    rc = fread(&gam, sizeof(long long), 1, fdi);
    if (rc != 1) {
      perror("reading gamma file");
      die ("reading gamma file", errno);}
    lastg = which;}
  rc = ((gam >> 50) & 1);
  if (rc == 0)
    return 1;
  else
    return -1; }

extern int gamma_i (int which) {
#ifdef tracentry
  trace_gi += 1;
#endif
  int rc;
  long offset;
  if (which != lastg) {
    offset = which * sizeof(long long);
    rc = fseek(fdi, offset, SEEK_SET);
    if (rc != 0) {
      perror("seeking gamma file");
      die ("seeking gamma file", errno);}
    rc = fread(&gam, sizeof(long long), 1, fdi);
    if (rc != 1) {
      perror("reading gamma file");
      die ("reading gamma file", errno);}
    lastg = which;}
  rc = (gam >> 51);
  if (rc == 0)
    return 0;
  else
    return 1; }

extern ntuple* ntuple::gamma(int order, int* index_list_param) {
  // index_list_param[i] is gamma index to contract with ith ntuple index
#ifdef tracentry
  trace_gam += 1;
#endif
  int index, number_to_gammaize, gamma_length;
  int mult = 1;
  int pwrs[number_of_variables] = {0};
  bool skip, indexq = true;
  long long one = 1;
#ifdef trace_iter
  char str [256];
#endif
#ifdef debug
  if ((order < 1) || (order >= dimension))
    die("invalid order passed to gamma");
#endif
  index = 0;                            // count number of indices to gammaize
  for (int i = 0; i < rank; i++)
    if (index_list_param[i] >= 0) {
#ifdef debug
      if (index_list_param[i] >= order)
	die("index list param out of bounds");
      indexq = indexq && upper[i];
#endif
      index += 1;}
  number_to_gammaize = index;
#ifdef debug
  if (!indexq)
    die("attempting to gammaize on lower indices");
  if (number_to_gammaize > order)
    die("number to gammaize > order");
#endif
  gamma_length = dimension;
  for (int i = 0; i < (order - 1); i++)
    gamma_length *= (dimension - i - 1);
  gamma_length *= spinor_dimension;
  int new_rank = rank + order - 2 * number_to_gammaize + 2;
  index = 0;                             // count number of antisymmetric index pairs
  for (int i = 0; i < (rank - 1); i++)
    if ((index_list_param[i] >= 0) && antisymmetric[i] && 
	((index_list_param[i + 1]) == (index_list_param[i] + 1)))
      index += 1;
  for (int i = 0; i < index; i++)
    mult *= 2;
  int* new_index_list = new int [new_rank];
  if (new_index_list == NULL)
    die("allocating new_index_list");
  bool* new_upper = new bool [new_rank];
  if (new_upper == NULL)
    die("allocating new_upper");
  bool* new_antisymmetric = new bool [new_rank];
  if (new_antisymmetric == NULL)
    die("allocating new_antisymmetric");
  bool* new_symmetric = new bool [new_rank];
  if (new_symmetric == NULL)
    die("allocating new_symmetric");
  bool* new_spinor = new bool [new_rank];
  if (new_spinor == NULL)
    die("allocating new_spinor");
  long long overall_coefficient = 0;
  sum* new_sum = new sum;
  if (new_sum == NULL)
    die("allocating new_sum");
  sum** multiply_list = new sum* [2];
  if (multiply_list == NULL)
    die("allocating multiply_list");
  index = 0;
  for (int i = 0; i < rank; i++) {         // do old free indices first
    if (index_list_param[i] >= 0)
      continue;
#ifdef debug
    if ((index < 0) || (index >= rank))
      die("gamma - index out of bounds");
#endif
    new_upper[index] = upper[i];
    new_antisymmetric[index] = antisymmetric[i];
    new_symmetric[index] = symmetric[i];
    new_spinor[index] = spinor[i];
    index += 1;}
  if ((index % 2) == 1)                    // make correction if ended on even index
    new_antisymmetric[index - 1] = false;
  for (int i = 0; i < (order - number_to_gammaize); i++) {     // now do new free lorentz indices
#ifdef debug
    if ((index < 0) || (index >= new_rank))
      die("gamma - index out of bounds (2)");
#endif
    new_upper[index] = false;
    if ((order-number_to_gammaize > 1) && ((i % 2) == 0))
      new_antisymmetric[index] = true;
    else
      new_antisymmetric[index] = false;
    new_symmetric[index] = false;
    new_spinor[index] = false;
    index += 1;}
  for (int i = 0; i < 2; i++) {               // lastly do new spinor indices;
#ifdef debug
    if ((index < 0) || (index >= new_rank))
      die("gamma - index out of bounds");
#endif
    new_upper[index] = false;
    new_antisymmetric[index] = false;
    new_symmetric[index] = false;
    new_spinor[index] = true;
    index += 1;}
  new_upper[index - 1] = true;              // end spinor index is upper
  ntuple* new_ntuple = new ntuple(new_rank, new_upper, new_antisymmetric, new_symmetric, new_spinor);
  if (new_ntuple == NULL)
    die("allocating new_ntuple");
  for (int i = 0; i < rank; i++)
    index_list[i] = index_list_param[i];
  multiply_list[1] = new sum;
  gamma_start(order);
  make_free_index_list(true);
  if (free_index_length == 0) {
    free_index_list = new long long [1];
    if (free_index_list == NULL)
      die("allocating free_index_list");
    free_index_length = 1;}
  while (free_parallelize(new_ntuple)) {
    for (int i = 0; i < rank; i++)                   // copy free indices
      if (index_list_param[i] < 0)
	index_list[i] = free_index(i);
    for (int i = 0; i < gamma_length; i++) {   // for each nonzero gamma component
      for (int j = 0; j < rank; j++)                // copy gamma contracted indices
	if (index_list_param[j] >= 0)
	  index_list[j] = gamma_index(i, index_list_param[j]);
      index = 0;
      for (int j = 0; j < rank; j++)                // copy uncontracted ntuple indices
	if (index_list_param[j] < 0) {
#ifdef debug
	  if ((index < 0) || (index >= rank))
	    die("gamma - index out of bounds (3)");
#endif
	  new_index_list[index] = index_list[j];
	  index += 1;}
      for (int j = 0; j < (order + 2); j++) {   // copy uncontracted gamma indices
	skip = false;
	for (int k = 0; k < rank; k++)
	  if (j == index_list_param[k]) {
	    skip = true;
	    break;}
	if (skip)
	  continue;
#ifdef debug
	if ((index < 0) || (index >= new_rank))
	  die("gamma - index out of bounds (4)");
#endif
	new_index_list[index] = gamma_index(i, j);
	index += 1;}
      skip = false;
      for (int j = 0; j < rank - 1; j++)
	if (index_list_param[j] >= 0)
	  if (antisymmetric[j] && (index_list[j] > index_list[j + 1])) {
	    skip = true;
	    break;}
      for (int j = 0; j < new_rank - 1; j++)
	if (new_antisymmetric[j] && (new_index_list[j] > new_index_list[j + 1])) {
	  skip = true;
	  break;}
      if (!skip) {
	current_component = sub(index_list);
	if (current_component != NULL) {
	  multiply_list[0] = current_component->value;
	  if (gamma_i(i) == 0) {
	    pwrs[dimension + number_of_functions + number_of_ansatze] = 0;
	    multiply_list[1]->add(&one, pwrs);}
	  else {
	    pwrs[dimension + number_of_functions + number_of_ansatze] = 1;
	    multiply_list[1]->add(&one, pwrs);}
	  overall_coefficient = gamma_sign(i) * mult;
	  new_sum->multiply(2, multiply_list, &overall_coefficient);
	  multiply_list[1]->zero();
	  new_ntuple->add(new_index_list, new_sum);
	  new_sum->zero();}}}
#ifdef trace_iter
    sprintf(str, "Processing current_free_index = %d; indices = ", current_free_index);
    for (int j = 0; j < new_ntuple->rank; j++)
      sprintf(str + strlen(str), "%d ", new_index_list[j]);
    to_stderr (str);
#endif
  }
  new_ntuple->simplify();
  new_ntuple->fill();
  gamma_stop();
  delete new_sum;
  delete multiply_list[1];
  delete [] new_upper;
  delete [] new_antisymmetric;
  delete [] new_symmetric;
  delete [] new_spinor;
  delete [] multiply_list;
  delete [] new_index_list;
  return new_ntuple; }
