//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// geometry.cc

#include <ptah.h>
#include <element.h>
#include <hash.h>
#include <ntuple.h>

extern connection::connection() {
#ifdef tracentry
  trace_cc += 1;
#endif
#ifdef debug
  connection_magic_no = connection_magic;
#endif
  sum half;
  sum* new_sum = new sum;
  sum* temp_sum = new sum;
  sum** multiply_list = new sum* [3];
  long long one = 1;
  long long minus_one = -1;
  memset(&factor_powers_temp, 0, number_of_variables * sizeof(int));
  factor_powers_temp[number_of_variables - number_of_fractions] = -1;
  half.add(&one, factor_powers_temp);
  rank = 3;
  ntuple_init();
  multiply_list[0] = temp_sum;
  multiply_list[2] = &half;
  upper[0] = true;
  symmetric[1] = true;
  for (int i = 0; i < dimension; i++)
    for (int j = 0; j < dimension; j++)
      for (int k = j; k < dimension; k++) {
	for (int l = 0; l < dimension; l++) {
	  index_list[0] = l;
	  index_list[1] = j;
	  current_component = g->sub(index_list);
	  if (current_component != NULL) {
	    temp_sum->derivative(current_component->value, k, &one);}
	  index_list[0] = l;
	  index_list[1] = k;
	  current_component = g->sub(index_list);
	  if (current_component != NULL) {
	    temp_sum->derivative(current_component->value, j, &one);}
	  index_list[0] = j;
	  index_list[1] = k;
	  current_component = g->sub(index_list);
	  if (current_component != NULL) {
	    temp_sum->derivative(current_component->value, l, &minus_one);}
	  index_list[0] = i;
	  index_list[1] = l;
	  current_component = g_inverse->sub(index_list);
	  if ((current_component != NULL) && (!temp_sum->zeroq())) {
	    multiply_list[1] = current_component->value;
	    new_sum->multiply(3, multiply_list, &one);}
	  temp_sum->zero();}
	index_list[0] = i;
	index_list[1] = j;
	index_list[2] = k;
	new_sum->simplify();
	set(index_list, new_sum);
	new_sum = new sum;}
  fill();
  half.zero();
  delete new_sum;
  delete temp_sum;
  delete [] multiply_list; }

extern curvature::curvature() {
#ifdef tracentry
  trace_ccu += 1;
#endif
#ifdef debug
  curvature_magic_no = curvature_magic;
#endif
  int first = 0;
  bool alltrue [4] = {true, false, true, false};
  bool allfalse [4] = {false, false, false, false};
  sum denom, denom2;
  sum* temp_sum = new sum;
  sum* new_sum = new sum;
  sum** multiply_list = new sum* [3];
  ntuple* ri1up;
  long long one = 1;
  long long minus_one = -1;
  long long two = 2;
  rank = 4;
  ntuple_init();
  upper[0] = true;
  for (int i = 0; i < dimension; i++)
    for (int j = 0; j < dimension; j++)
      for (int k = 0; k < dimension - 1; k++)
	for (int l = k + 1; l < dimension; l++) {
	  index_list[0] = i;
	  index_list[1] = l;
	  index_list[2] = j;
	  current_component = christoffel->sub(index_list);
	  if (current_component != NULL)
	    new_sum->derivative(current_component->value, k, &one);
	  index_list[1] = k;
	  current_component = christoffel->sub(index_list);
	  if (current_component != NULL)
	    new_sum->derivative(current_component->value, l, &minus_one);
	  for (int m = 0; m < dimension; m++) {
	    index_list[0] = i;
	    index_list[1] = m;
	    index_list[2] = k;
	    current_component = christoffel->sub(index_list);
	    if (current_component == NULL)
	      continue;
	    multiply_list[0] = current_component->value;
	    index_list[0] = m;
	    index_list[1] = l;
	    index_list[2] = j;
	    current_component = christoffel->sub(index_list);
	    if (current_component == NULL)
	      continue;
	    multiply_list[1] = current_component->value;
	    new_sum->multiply(2, multiply_list, &one);}
	  for (int m = 0; m < dimension; m++) {
	    index_list[0] = i;
	    index_list[1] = m;
	    index_list[2] = l;
	    current_component = christoffel->sub(index_list);
	    if (current_component == NULL)
	      continue;
	    multiply_list[0] = current_component->value;
	    index_list[0] = m;
	    index_list[1] = k;
	    index_list[2] = j;
	    current_component = christoffel->sub(index_list);
	    if (current_component == NULL)
	      continue;
	    multiply_list[1] = current_component->value;
	    new_sum->multiply(2, multiply_list, &minus_one);}
	  index_list[0] = i;
	  index_list[1] = j;
	  index_list[2] = k;
	  index_list[3] = l;
	  new_sum->simplify();
	  set(index_list, new_sum);
	  new_sum = new sum;}
  antisymmetric[2] = true;
  fill();

  riemann = this;
  ricci = riemann->self_contract(0,2);
  riemann_all_low = riemann->lower(1, &first);
  riemann_all_low->antisymmetric[0] = true;
  ri1up = ricci->raise(1, &first);
  scacurv = ri1up->scalar();

//  C[a,b,c,d] = R[a,b,c,d] + 
//    (g[a,d]R[b,c] - g[a,c]R[b,d] + g[b,c]R[a,d] - g[b,d]R[a,c]) (2/(dim-2)) +
//    R (g[a,c]g[b,d] - g[a,d]g[b,c]) (2/((dim-1)(dim-2)))
// or
//  s[a,b] = r[a,b] - g[a,b] r / dim
//  e[a,b,c,d] = - (g[a,d]s[b,c] - g[a,c]s[b,d] + g[b,c]s[a,d] - g[b,d]s[a,c]) (2/(dim-2))
//  g[a,b,c,d] = r (g[a,c]g[b,d] - g[a,d]g[b,c]) (2/((dim-1)dim))
//  c[a,b,c,d] = r[a,b,c,d] - e[a,b,c,d] - g[a,b,c,d]

  traceless_ricci = ricci->clone();
  if (scacurv->number_of_components > 0) {
    memcpy(&factor_powers_temp, &factor_powers_dim, number_of_variables * sizeof(int));
    denom.add(&minus_one, factor_powers_temp);
    for (int a = 0; a < dimension; a++)
      for (int b = 0; b < dimension; b++) {
	index_list[0] = a;
	index_list[1] = b;
	current_component = g->sub(index_list);
	if ((current_component != NULL) && (current_component->value != NULL)) {
	  multiply_list[0] = current_component->value;
	  multiply_list[1] = &denom;
	  multiply_list[2] = scacurv;
	  new_sum->multiply(3, multiply_list, &one);
	  new_sum->simplify();
	  traceless_ricci->add(index_list, new_sum);
	  new_sum->zero();}}}

  einstein_curv = new ntuple(4, allfalse, alltrue, allfalse, allfalse);
  g_curv = new ntuple(4, allfalse, alltrue, allfalse, allfalse);
  weyl = riemann_all_low->clone();
  if (traceless_ricci->number_of_components > 0) {
    memcpy(&factor_powers_temp, &factor_powers_dimm2, number_of_variables * sizeof(int));
    denom.zero();
    denom.add(&two, factor_powers_temp);
    memcpy(&factor_powers_temp, &factor_powers_dimdimm1, number_of_variables * sizeof(int));
    denom2.add(&two, factor_powers_temp);
    for (int a = 0; a < dimension; a++)
      for (int b = 0; b < dimension; b++)
	if (a != b)
	for (int c = 0; c < dimension; c++)
	  for (int d = 0; d < dimension; d++)
	    if (c != d) {
	    index_list[0] = a;
	    index_list[1] = d;
	    current_component = g->sub(index_list);
	    if ((current_component != NULL) && (current_component->value != NULL)) {
	      multiply_list[0] = current_component->value;
	      index_list[0] = b;
	      index_list[1] = c;
	      current_component = traceless_ricci->sub(index_list);
	      if ((current_component != NULL) && (current_component->value != NULL)) {
		multiply_list[1] = current_component->value;
		new_sum->multiply(2, multiply_list, &one);}}    // compute -E for convenience
	    index_list[0] = a;
	    index_list[1] = c;
	    current_component = g->sub(index_list);
	    if ((current_component != NULL) && (current_component->value != NULL)) {
	      multiply_list[0] = current_component->value;
	      index_list[0] = b;
	      index_list[1] = d;
	      current_component = traceless_ricci->sub(index_list);
	      if ((current_component != NULL) && (current_component->value != NULL)) {
		multiply_list[1] = current_component->value;
		new_sum->multiply(2, multiply_list, &minus_one);}}
	    index_list[0] = b;
	    index_list[1] = c;
	    current_component = g->sub(index_list);
	    if ((current_component != NULL) && (current_component->value != NULL)) {
	      multiply_list[0] = current_component->value;
	      index_list[0] = a;
	      index_list[1] = d;
	      current_component = traceless_ricci->sub(index_list);
	      if ((current_component != NULL) && (current_component->value != NULL)) {
		multiply_list[1] = current_component->value;
		new_sum->multiply(2, multiply_list, &one);}}
	    index_list[0] = b;
	    index_list[1] = d;
	    current_component = g->sub(index_list);
	    if ((current_component != NULL) && (current_component->value != NULL)) {
	      multiply_list[0] = current_component->value;
	      index_list[0] = a;
	      index_list[1] = c;
	      current_component = traceless_ricci->sub(index_list);
	      if ((current_component != NULL) && (current_component->value != NULL)) {
		multiply_list[1] = current_component->value;
		new_sum->multiply(2, multiply_list, &minus_one);}}
	    multiply_list[0] = new_sum;
	    multiply_list[1] = &denom;
	    temp_sum->multiply(2, multiply_list, &one);
	    temp_sum->simplify();
	    index_list[0] = a;
	    index_list[1] = b;
	    index_list[2] = c;
	    index_list[3] = d;
	    einstein_curv->set(index_list, temp_sum);
	    temp_sum = new sum;
	    new_sum->zero();
	    if (scacurv->number_of_components > 0) {
	      index_list[0] = a;
	      index_list[1] = c;
	      current_component = g->sub(index_list);
	      if ((current_component != NULL) && (current_component->value != NULL)) {
		multiply_list[0] = current_component->value;
		index_list[0] = b;
		index_list[1] = d;
		current_component = g->sub(index_list);
		if ((current_component != NULL) && (current_component->value != NULL)) {
		  multiply_list[1] = current_component->value;
		  new_sum->multiply(2, multiply_list, &minus_one);}}   // compute -G for convenience
	      index_list[0] = a;
	      index_list[1] = d;
	      current_component = g->sub(index_list);
	      if ((current_component != NULL) && (current_component->value != NULL)) {
		multiply_list[0] = current_component->value;
		index_list[0] = b;
		index_list[1] = c;
		current_component = g->sub(index_list);
		if ((current_component != NULL) && (current_component->value != NULL)) {
		  multiply_list[1] = current_component->value;
		  new_sum->multiply(2, multiply_list, &one);}}
	      multiply_list[0] = new_sum;
	      multiply_list[1] = &denom2;
	      multiply_list[2] = scacurv;
	      temp_sum->multiply(3, multiply_list, &one);
	      temp_sum->simplify();
	      index_list[0] = a;
	      index_list[1] = b;
	      index_list[2] = c;
	      index_list[3] = d;
	      g_curv->set(index_list, temp_sum);
	      temp_sum = new sum;
	      new_sum->zero();}}}

  weyl->add(einstein_curv);
  weyl->add(g_curv);
  weyl->simplify();
  einstein_curv->scalar_multiply(minus_one);
  g_curv->scalar_multiply(minus_one);
  denom.zero();
  denom2.zero();
  delete temp_sum;
  delete new_sum;
  delete [] multiply_list; }
