//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// indexing.cc

#include <ptah.h>
#include <element.h>
#include <hash.h>
#include <ntuple.h>

extern ntuple* ntuple::raise(int numind, int* indices) {
#ifdef tracentry
  trace_nr += 1;
#endif
  int number_to_contract = 0;
  bool found;
#ifdef debug
  if (numind > rank)
    die("raise - invalid parameter");
#endif
  int* indices_to_contract = new int [numind];
  if (indices_to_contract == NULL)
    die("allocating indices_to_contract");
  sum* new_sum = new sum;
  if (new_sum == NULL)
    die("allocating new_sum");
  sum** multiply_list = new sum* [numind];
  if (multiply_list == NULL)
    die("allocating multiply_list");
  ntuple** contract_list = new ntuple* [numind];
  if (contract_list == NULL)
    die("allocating contract_list");
#ifdef debug
  for (int i = 0; i < numind; i++) {
    if ((indices[i] < 0) || (indices[i] >= rank))
      die("raise - index out of bounds");
    if (upper[indices[i]])
      die("attempting to raise an upper index");
    if (spinor[indices[i]])
      die("attempting to raise a spinor index");}
#endif
  ntuple* new_ntuple = new ntuple(rank, upper, antisymmetric, symmetric, spinor);
  if (new_ntuple == NULL)
    die("allocating new_ntuple");
                    // if raising an index which has (anti)symmetry and its pair is not
                    // being raised, set both to false
  for (int i = 0; i < numind; i++) {
    if (antisymmetric[indices[i]]) {
      found = false;
      for (int j = 0; j < numind; j++)
	if (indices[j] == (indices[i] + 1)) {
	  found = true;
	  break;}
      if (!found) {
	new_ntuple->antisymmetric[indices[i]] = false;
	new_ntuple->antisymmetric[indices[i] + 1] = false;}}
    if (symmetric[indices[i]]) {
      found = false;
      for (int j = 0; j < numind; j++)
	if (indices[j] == (indices[i] + 1)) {
	  found = true;
	  break;}
      if (!found) {
	new_ntuple->symmetric[indices[i]] = false;
	new_ntuple->symmetric[indices[i] + 1] = false;}}}
  contract_list[0] = g_inverse;
  for (int index = 0; index < numind; index++) {
    new_ntuple->upper[indices[index]] = true;
    indices_to_contract[number_to_contract] = indices[index];
    number_to_contract += 1;}
  for (int i = 1; i < number_to_contract; i++)
    contract_list[i] = g_inverse->clone();
  for (int i = 0; i < rank; i++)
    index_list[i] = -1;
  for (int i = 0; i < number_to_contract; i++) {
    index_list[indices_to_contract[i]] = i;
    contract_list[i]->index_list[0] = i;
    contract_list[i]->index_list[1] = -1;}
  contract(number_to_contract, contract_list, new_ntuple, true);
  for (int i = 1; i < number_to_contract; i++) {
    contract_list[i]->zero();
    delete contract_list[i];}
  delete new_sum;
  delete [] multiply_list;
  delete [] contract_list;
  delete [] indices_to_contract;
  new_ntuple->fill();
  return new_ntuple; }

extern ntuple* ntuple::lower(int numind, int* indices) {
#ifdef tracentry
  trace_nl += 1;
#endif
#ifdef debug
  if (numind > rank)
    die("lower - invalid parameter");
#endif
  int* indices_to_contract = new int [numind];
  if (indices_to_contract == NULL)
    die("allocating indices_to_contract");
  int number_to_contract = 0;
  bool found;
  sum* new_sum = new sum;
  if (new_sum == NULL)
    die("allocating new_sum");
  sum** multiply_list = new sum* [numind];
  if (multiply_list == NULL)
    die("allocating multiply_list");
  ntuple** contract_list = new ntuple* [numind];
  if (contract_list == NULL)
    die("allocating contract_list");
#ifdef debug
  for (int i = 0; i < numind; i++) {
    if ((indices[i] < 0) || (indices[i] >= rank))
      die("lower - index out of bounds");
  if (!upper[indices[i]])
    die("attempting to lower a lower index");
  if (spinor[indices[i]])
    die("attempting to lower a spinor index");}
#endif
  ntuple* new_ntuple = new ntuple(rank, upper, antisymmetric, symmetric, spinor);
  if (new_ntuple == NULL)
    die("allocating new_ntuple");
                    // if lowering an index which has (anti)symmetry and its pair is not
                    // being lowered, set both to false
  for (int i = 0; i < numind; i++) {
    if (antisymmetric[indices[i]]) {
      found = false;
      for (int j = 0; j < numind; j++)
	if (indices[j] == (indices[i] + 1)) {
	  found = true;
	  break;}
      if (!found) {
	new_ntuple->antisymmetric[indices[i]] = false;
	new_ntuple->antisymmetric[indices[i] + 1] = false;}}
    if (symmetric[indices[i]]) {
      found = false;
      for (int j = 0; j < numind; j++)
	if (indices[j] == (indices[i] + 1)) {
	  found = true;
	  break;}
      if (!found) {
	new_ntuple->symmetric[indices[i]] = false;
	new_ntuple->symmetric[indices[i] + 1] = false;}}}
  for (int index = 0; index < numind; index++) {
    new_ntuple->upper[indices[index]] = false;
    indices_to_contract[number_to_contract] = indices[index];
    contract_list[number_to_contract] = g;
    number_to_contract += 1;}
  for (int i = 1; i < number_to_contract; i++)
    contract_list[i] = g->clone();
  for (int i = 0; i < rank; i++)
    index_list[i] = -1;
  for (int i = 0; i < number_to_contract; i++) {
    index_list[indices_to_contract[i]] = i;
    contract_list[i]->index_list[0] = i;
    contract_list[i]->index_list[1] = -1;}
  contract(number_to_contract, contract_list, new_ntuple, true);
  for (int i = 1; i < number_to_contract; i++) {
    contract_list[i]->zero();
    delete contract_list[i];}
  delete new_sum;
  delete [] multiply_list;
  delete [] contract_list;
  delete [] indices_to_contract;
  new_ntuple->fill();
  return new_ntuple; }
