//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// invariant.cc

#include <ptah.h>
#include <element.h>
#include <hash.h>

extern invariant::invariant() {
#ifdef tracentry
  trace_i += 1;
#endif
#ifdef debug
  hash_magic_no = hash_magic;
  sum_magic_no = sum_magic;
  invariant_magic_no = invariant_magic;
#endif
  delete [] head;
  head = new element* [large_hash_size];
  if (head == NULL)
    die("allocating head");
  size = large_hash_size;
#ifdef keepstats
#ifdef debug
  delete [] headcount;
  headcount = new int [large_hash_size];
  if (headcount == NULL)
    die("allocating headcount");
  memset(headcount, 0, large_hash_size * sizeof(int));
#endif
#endif
  number_of_components = 0;
  last_hash = -1;
  for (int i = 0; i < size; i++)
    head[i] = NULL;
  invariant_head = (invariant_term**) new invariant_term* [number_to_simplify * simplify_size];
  if (invariant_head == NULL)
    die("allocating invariant_head");
  for (int i = 0; i < (number_to_simplify * simplify_size); i++)
      invariant_head[i] = NULL; }

extern invariant::~invariant() {
#ifdef tracentry
  trace_ki += 1;
#endif
#ifdef debug
  invariant_magic_no = -invariant_magic;
  for (int i = 0; i < (number_to_simplify * simplify_size); i++)
    if (invariant_head[i] != NULL)
      die("attempting to delete non-empty invariant");
#endif
  delete [] invariant_head; }

extern void invariant::zero() {
#ifdef tracentry
  trace_iz += 1;
#endif
  invariant_term *current_term;
  for (int i = 0; i < size; i++)
    {current_term = (invariant_term *) first_term(i);
    while (current_term != NULL)
      {remove_from_invariant(current_term);
      set_hash(i);
      remove_term(current_term);
      current_term = (invariant_term *) first_term(i);}}        // start over since we just removed the head
}

extern void invariant::multiply(int number_of_sums, sum** sum_list, long long* overall_coefficient) {
#ifdef tracentry
  trace_im += 1;
#endif
  int* head_index = new int [number_of_sums];
  if (head_index == NULL)
    die("allocating head_index");
  int* first_head_index = new int [number_of_sums];
  if (first_head_index == NULL)
    die("allocating first_head_index");
  term** current_term = new term* [number_of_sums];
  if (current_term == NULL)
    die("allocating current_term");
  term** first_current_term = new term* [number_of_sums];
  if (first_current_term == NULL)
    die("allocating first_current_term");
  long long product_of_coefficients;
  bool nonzero, not_done = true;
  for (int i = 0; i < number_of_sums; i++) {
#ifdef debug
    if (sum_list[i]->sum_magic_no != sum_magic)
      die("attempting to multiply a non-sum");
#endif
    head_index[i] = 0;                                    // find first term in sum
    current_term[i] = sum_list[i]->first_term(0);
    while ((head_index[i] < (sum_list[i]->size - 1)) && (current_term[i] == NULL))
      {head_index[i] += 1;
      current_term[i] = sum_list[i]->first_term(head_index[i]);}
    if (current_term[i] == NULL)                          // one sum is zero, so product is also
      {not_done = false;
      break;}
    first_head_index[i] = head_index[i];                  // save for next round
    first_current_term[i] = current_term[i];}
  while (not_done) {
    memset(&factor_powers_temp_i, 0, number_of_variables * sizeof(int));
    product_of_coefficients = *overall_coefficient;
    nonzero = false;
    for (int i = 0; i < number_of_sums; i++)              // multiply all current terms
      if (current_term[i] != NULL) {
	product_of_coefficients *= current_term[i]->coefficient;
	nonzero = true;
	for (int j = 0; j < number_of_variables; j++)
	  factor_powers_temp_i[j] += current_term[i]->factor_powers[j];}
    if (nonzero)
      add(&product_of_coefficients, factor_powers_temp_i);
    for (int i = 0; i < number_of_sums; i++)              // check all sums
      {current_term[i] = current_term[i]->next_term();    // next term in sum
      while ((head_index[i] < (sum_list[i]->size - 1)) && (current_term[i] == NULL))
	{head_index[i] += 1;
	current_term[i] = sum_list[i]->first_term(head_index[i]);}
      if (current_term[i] != NULL)                        // not through all terms in this sum
	break;
      head_index[i] = first_head_index[i];                // start this sum over
      current_term[i] = first_current_term[i];            // and check the next
      if (i >= (number_of_sums - 1))                      // are we done?
	not_done = false;}}
  delete [] current_term;
  delete [] first_current_term;
  delete [] head_index;
  delete [] first_head_index; }
