//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// main.cc

#include <ptah.h>
#include <element.h>
#include <hash.h>
#include <ntuple.h>
#include <main.h>
#include <metric.h>
#include <d10Bsugra.h>

ntuple* ad2;
ntuple* ad3;

ntuple* rieuuuu;
ntuple* rieuull;
ntuple* rieulul;
ntuple* riesq;
ntuple* riesq2;
ntuple* riesqlluu;
ntuple* riesqulluul;

ntuple* ricuu;
ntuple* ricul;
ntuple* ricsq;

ntuple* cdrie;
ntuple* cdrieuuuul;
ntuple* cdrieuuuuu;
ntuple* cdrieullll;
ntuple* cdrieconuuu;
ntuple* cdrieconlll;

ntuple* cdric;
ntuple* cdricuuu;
ntuple* cdricull;
ntuple* cdricuul;
ntuple* cdriconu;
ntuple* cdriconl;

ntuple* rlist[5];

sum* invar;

sum* slist[3];

component* comp;
component* comp2;

int inds[6];
int indlist[5] = {0,1,2,3,4};

int riesqmat[8] = {0,1,2,3, 0,1,2,3};
int riesqulluulmat[8] = {-1,0,-1,-1, -1,-1,-1,0};
int riecub1mat[8] = {0,1,-1,-1, -1,-1,0,1};
int riecub2mat[8] = {0,-1,1,-1, -1,0,-1,1};
int riecub3mat[8] = {0,-1,1,-1, -1,1,-1,0};
int riecubmat[8] = {0,1,2,3, 0,2,1,3};

int ricsqmat[4] = {0,1, 0,1};
int riccubmat[4] = {0,-1, -1,0};
int riericsqmat[12] = {0,1,2,3, 0,2,-1,-1, 1,3,-1,-1};
int riesqricmat[8] = {0,-1,-1,-1, -1,-1,-1,0};
		      
int cdriesq1mat[10] = {0,1,2,3,4, 0,1,2,3,4};
int cdriesq2mat[6] = {0,1,2, 0,1,2};

int cdricsq2mat[6] = {0,1,2, 0,2,1};
int cdricsq3mat[2] = {0,0};

int rstarrmat[4] = {0,0,0,0};
int rstarrrrmat[8] = {0,0,0,0,0,0,0,0};
int eulermat[20] = {0,0,1,1, 0,0,1,1, 0,0,1,1, 0,0,1,1, 0,0,1,1};

bool trueparm[2] = {true, true}; 
bool antiparm[3] = {true,true,false};
bool falseparm[3] = {false, false, false}; 

int len;

int fpp[number_of_variables] = {0};

long long two = 2;
long long four = 4;

int main (int argcount, char **args) {

  char str[256];

  parallel = false;         // initial stuff does not require parallel

  time_stamp ();
  function_build (number_of_functions + number_of_ansatze, fnumtrm, fnumfct, fpwr, fco);

  g = new ntuple (2, falseparm, falseparm, trueparm, falseparm);
  g->tensor_build (gcnt, gind, gnumtrm, gnumfct, gpwr, gco);
  g->metric = true;

  g_inverse = new ntuple (2, trueparm, falseparm, trueparm, falseparm);
  g_inverse->tensor_build (gicnt, giind, ginumtrm, ginumfct, gipwr, gico);
  g_inverse->metric = true;

  inverse_sqrt_det_g = new sum;
  inverse_sqrt_det_g->sum_build (1, dnumfct, dpwr, dco);

  ad2 = new ntuple (2, falseparm, trueparm, falseparm, falseparm);
  ad2->tensor_build (ad2cnt, ad2ind, ad2numtrm, ad2numfct, ad2pwr, ad2co);

  ad3 = new ntuple (3, falseparm, antiparm, falseparm, falseparm);
  ad3->tensor_build (ad3cnt, ad3ind, ad3numtrm, ad3numfct, ad3pwr, ad3co);

  christoffel = new connection ();

  riemann = new curvature ();

  if (argcount >= 2) {
    for (int i = 0; i < number_of_functions; i++) {

      sprintf (str, "f[[%d]]", i + 1);
      functions[i]->display (str);

      sprintf (str, "Print[%d, %d];", i, function_power[i]);
      to_stdout (str);

      for (int j = 0; j < dimension; j++) {
	sprintf (str, "derivs[[%d,%d]]", i + 1, i * dimension + j + 1);
	derivatives[i * dimension + j]->display (str);}}

    len = sprintf (str, "simp[expr_]:=expr/.{");
    for (int i = 0; i < number_of_functions; i++) {
      len += sprintf (str + len, "%c->f[[%d]]^(1/%d)", variables[dimension + i], i + 1, function_power[i]);
      if (i < (number_of_functions - 1))
	len += sprintf (str + len, ",");}
    sprintf (str + len, "};");
    to_stdout (str);

    g->display ("g");
    g_inverse->display ("ginv");
    inverse_sqrt_det_g->display ("isrdmg");}

  scacurv->display ("sc");

  traceless_ricci->zero ();
  delete traceless_ricci;
  g_curv->zero ();
  delete g_curv;
  weyl->zero ();
  delete weyl;
  einstein_curv->zero ();
  delete einstein_curv;

  time_stamp ("Initialization completed");

  if (argcount >= 2) {
    parallel = false;

    rieuull = riemann->raise (1, &(indlist[1]));
    rieuull->antisymmetric[0] = true;
    rieuuuu = rieuull->raise (2, &(indlist[2]));
    
    rlist[0] = riemann_all_low;
    rlist[1] = rieuuuu;
    invar = riemann_all_low->scalar_invariant (2, (ntuple**) rlist, riesqmat, false);
    invar->display ("Riesq");
    invar->zero ();
    delete invar;

    if (ricci->number_of_components > 0) {

      ricul = ricci->raise (1, &(indlist[0]));
      ricuu = ricul->raise (1, &(indlist[1]));
      ricuu->symmetric[0] = true;
      
      rlist[0] = ricci;
      rlist[1] = ricuu;
      invar = ricci->scalar_invariant (2, (ntuple**) rlist, ricsqmat, false);
      invar->display ("Ricsq");
      invar->zero ();
      delete invar;}
    
    time_stamp ("Quadratic invariants completed");

    rieulul = riemann->raise (1, &(indlist[2]));

    rlist[0] = rieuull;
    rlist[1] = riemann_all_low;
    riesq = rieuull->ntuple_contract (2, (ntuple**) rlist, riecub1mat);
    rlist[0] = rieuuuu;
    rlist[1] = riesq;
    invar = rieuuuu->scalar_invariant (2, (ntuple**) rlist, riesqmat, false);
    invar->display ("Riecub1");
    invar->zero ();
    delete invar;

    rlist[0] = rieulul;
    rlist[1] = riemann_all_low;
    riesq2 = rieulul->ntuple_contract (2, (ntuple**) rlist, riecub2mat);
    rlist[0] = rieuuuu;
    rlist[1] = riesq2;
    invar = rieuuuu->scalar_invariant (2, (ntuple**) rlist, riecubmat, false);
    invar->display ("Riecub2");
    invar->zero ();
    delete invar;
    riesq2->zero ();
    delete riesq2;

    rlist[0] = rieulul;
    rlist[1] = riemann_all_low;
    riesq2 = rieulul->ntuple_contract (2, (ntuple**) rlist, riecub3mat);
    rlist[0] = rieuuuu;
    rlist[1] = riesq2;
    invar = rieuuuu->scalar_invariant (2, (ntuple**) rlist, riecubmat, false);
    invar->display ("Riecub3");
    invar->zero ();
    delete invar;
    rieulul->zero ();
    delete rieulul;
    riesq2->zero ();
    delete riesq2;

    if (ricci->number_of_components > 0) {
      
      rlist[0] = ricul;
      rlist[1] = ricci;
      ricsq = ricul->ntuple_contract (2, (ntuple**) rlist, riccubmat);
      rlist[0] = ricuu;
      rlist[1] = ricsq;
      invar = ricuu->scalar_invariant (2, (ntuple**) rlist, ricsqmat, false);
      invar->display ("Riccub");
      invar->zero ();
      delete invar;
      ricuu->zero ();
      delete ricuu;
      ricsq->zero ();
      delete ricsq;

      rlist[0] = rieuuuu;
      rlist[1] = ricci;
      rlist[2] = ricci->clone ();
      invar = rieuuuu->scalar_invariant (3, (ntuple**) rlist, riericsqmat, false);
      invar->display ("Riericsq");
      invar->zero ();
      delete invar;
      
      rlist[0] = ricul;
      rlist[1] = riemann_all_low;
      ricsq = ricul->ntuple_contract (2, (ntuple**) rlist, riesqricmat);
      rlist[0] = rieuuuu;
      rlist[1] = ricsq;
      invar = rieuuuu->scalar_invariant (2, (ntuple**) rlist, riesqmat, false);
      invar->display ("Riesqric");
      invar->zero ();
      delete invar;
      rieuuuu->zero ();
      delete rieuuuu;
      ricul->zero ();
      delete ricul;
      ricsq->zero ();
      delete ricsq;}
    
    time_stamp ("Cubic invariants completed");
    
    cdrie = riemann_all_low->covariant_derivative ();
    
    cdrieullll = cdrie->raise (1, &(indlist[0]));
    cdrieuuuul = cdrieullll->raise (3, &(indlist[1]));
    cdrieuuuul->antisymmetric[0] = true;
    cdrieuuuuu = cdrieuuuul->raise (1, &(indlist[4]));
    
    rlist[0] = cdrie;
    rlist[1] = cdrieuuuuu;
    invar = cdrie->scalar_invariant (2, (ntuple**) rlist, cdriesq1mat, false);
    invar->display ("Cdriesq1");
    invar->zero ();
    delete invar;
    cdrie->zero ();
    delete cdrie;
    cdrieuuuuu->zero ();
    delete cdrieuuuuu;

    cdrieconuuu = cdrieuuuul->self_contract (0, 4);
    cdrieconlll = cdrieullll->self_contract (0, 4);
    rlist[0] = cdrieconuuu;
    rlist[1] = cdrieconlll;
    invar = cdrieconuuu->scalar_invariant (2, (ntuple**) rlist, cdriesq2mat, false);
    invar->display ("Cdriesq2");
    invar->zero ();
    delete invar;
    cdrieuuuul->zero ();
    delete cdrieuuuul;
    cdrieullll->zero ();
    delete cdrieullll;
    cdrieconuuu->zero ();
    delete cdrieconuuu;
    cdrieconlll->zero ();
    delete cdrieconlll;

    if (ricci->number_of_components > 0) {

      cdric = ricci->covariant_derivative ();

      cdricull = cdric->raise (1, &(indlist[0]));
      cdricuul = cdricull->raise (1, &(indlist[1]));
      cdricuul->symmetric[0] = true;
      cdricuuu = cdricuul->raise (1, &(indlist[2]));

      rlist[0] = cdricuuu;
      rlist[1] = cdric;
      invar = cdricuuu->scalar_invariant (2, (ntuple**) rlist, cdriesq2mat, false);
      invar->display ("Cdricsq1");
      invar->zero ();
      delete invar;

      invar = cdricuuu->scalar_invariant (2, (ntuple**) rlist, cdricsq2mat, false);
      invar->display ("Cdricsq2");
      invar->zero ();
      delete invar;
      cdricuuu->zero ();
      delete cdricuuu;

      cdriconu = cdricuul->self_contract (0, 2);
      cdriconl = cdricull->self_contract (0, 2);
      rlist[0] = cdriconu;
      rlist[1] = cdriconl;
      invar = cdriconu->scalar_invariant (2, (ntuple**) rlist, cdricsq3mat, false);
      invar->display ("Cdricsq3");
      invar->zero ();
      cdricuul->zero ();
      delete cdricuul;
      cdricull->zero ();
      delete cdricull;
      cdriconu->zero ();
      delete cdriconu;
      cdriconl->zero ();
      delete cdriconl;}
    
    time_stamp ("Covariant derivative invariants computed");

    if (dimension == 4) {
      rlist[0] = riesq;
      invar = riesq->dual_invariant (1, (ntuple**) rlist, rstarrmat, false);
      invar->display ("RstarR");
      invar->zero ();}
    else {
      riesqlluu = riesq->raise (2, &(indlist[2]));
      invar = new sum;
      for (int i = 0; i < dimension - 1; i++)
	for (int j = i + 1; j < dimension; j++)
	  for (int l = 0; l < 2; l++) {
	    inds[0] = i;
	    inds[1] = j;
	    switch (l) {
	    case 0: inds[2] = i; inds[3] = j; break;
	    case 1: inds[2] = j; inds[3] = i; break;}
	    comp = ad2->sub (inds);
	    comp2 = ad2->sub (inds + 2);
	    if (((comp != NULL) && (comp->value != NULL) && (! (comp->value)->zeroq ())) &&
		((comp2 != NULL) && (comp2->value != NULL) && (! (comp2->value)->zeroq ()))) {
	      slist[0] = comp->value;
	      slist[1] = comp2->value;
	      comp = riesqlluu->sub (inds);
	      if ((comp != NULL) && (comp->value != NULL) && (! (comp->value)->zeroq())) {
		slist[2] = comp->value;
		invar->multiply (3, slist, &four);}}}
//        invar->simplify_invariant ();
      invar->display ("RstarR");
      invar->zero ();
      riesqlluu->zero ();
      delete riesqlluu;}

    if (dimension == 8) {
      rlist[0] = riesq;
      rlist[1] = riesq->clone ();
      invar = riesq->dual_invariant (2, (ntuple**) rlist, rstarrrrmat, false);
      invar->display_add ("RstarRRR");
      invar->zero ();
      rlist[1]->zero ();
      delete rlist[1];}
      
    riesq->zero ();
    delete riesq;

    rlist[0] = rieuull;
    rlist[1] = rieuull->clone ();
    riesqulluul = rieuull->ntuple_contract (2, (ntuple**) rlist, riesqulluulmat);
    rieuull->zero ();
    delete rieuull;

    invar = new sum;
    for (int i = 0; i < dimension; i++)
      for (int j = 0; j < dimension - 1; j++)
	for (int k = j + 1; k < dimension; k++)
	  for (int n = 0; n < 6; n++) {
	    inds[0] = i;
	    inds[1] = j;
	    inds[2] = k;
	    switch (n) {
	    case 0: inds[3] = i; inds[4] = j; inds[5] = k; break;
	    case 1: inds[3] = j; inds[4] = k; inds[5] = i; break;
	    case 2: inds[3] = k; inds[4] = i; inds[5] = j; break;
	    case 3: inds[3] = i; inds[4] = k; inds[5] = j; break;
	    case 4: inds[3] = j; inds[4] = i; inds[5] = k; break;
	    case 5: inds[3] = k; inds[4] = j; inds[5] = i; break;}
	    comp = ad3->sub (inds);
	    comp2 = ad3->sub (inds + 3);
	    if (((comp != NULL) && (comp->value != NULL) && (! (comp->value)->zeroq ())) &&
		((comp2 != NULL) && (comp2->value != NULL) && (! (comp2->value)->zeroq ()))) {
	      slist[0] = comp->value;
	      slist[1] = comp2->value;
	      inds[0] = i;
	      inds[3] = j;
	      inds[4] = k;
	      switch (n) {
	      case 0: inds[1] = i; inds[2] = j; inds[5] = k; break;
	      case 1: inds[1] = j; inds[2] = k; inds[5] = i; break;
	      case 2: inds[1] = k; inds[2] = i; inds[5] = j; break;
	      case 3: inds[1] = i; inds[2] = k; inds[5] = j; break;
	      case 4: inds[1] = j; inds[2] = i; inds[5] = k; break;
	      case 5: inds[1] = k; inds[2] = j; inds[5] = i; break;}
	      comp = riesqulluul->sub (inds);
	      if ((comp != NULL) && (comp->value != NULL) && (! (comp->value)->zeroq ())) {
		slist[2] = comp->value;
		invar->multiply (3, slist, &two);}}}
//      invar->simplify_invariant ();
    invar->display ("gRstarR");
    invar->zero ();
    riesqulluul->zero ();
    delete riesqulluul;}
  
  if ((dimension % 2) == 0) {

//    parallel = true;         // all should start together

    rlist[0] = riemann_all_low;
    for (int i = 1; i < dimension/2; i++)
      rlist[i] = riemann_all_low->clone ();
    invar = riemann_all_low->dual_invariant (dimension/2, (ntuple**) rlist, eulermat, false);
    invar->display_add ("Euler");
    for (int i = 1; i < dimension/2; i++) {
      rlist[i]->zero ();
      delete rlist[i];}}

  time_stamp ("Dual invariants completed");
  
  entry_stats ();
  return 0; }
