//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// misc.cc

#include <ptah.h>
#include <element.h>
#include <hash.h>
#include <ntuple.h>

//  long get_rss () {
//    int c;
//    rusage ruse;
//    c = getrusage(RUSAGE_SELF, &ruse);
//    if (c < 0)
//      perror("getrusage");
//    return ruse.ru_maxrss; }

long get_vsize () {
  int c;
  int v = 0;
#ifdef unixsys
  FILE* f = fopen("/proc/self/stat", "r");
  if(f == NULL) {
    perror("opening /proc/self/stat");
    return 0;}
  c = fscanf(f, 
	     "%*d %*s %*c %*d %*d %*d %*d %*d %*u %*u %*u %*u %*u %*d %*d %*d %*d %*d %*d %*u %*u %*d %u", 
	     &v);
  if (c != 1)
    cerr << "fscanf error";
  fclose(f);
#endif
  return (long) v;
 }

extern void time_stamp () {
#ifdef unixsys
  long siz;
  char str [256];
  timeval now;
  long tcsec = t0sec;
  long tcmicro = t0micro;
  gettimeofday(&now, NULL);
  if (sec == -1) {
    sec = now.tv_sec;
    t0sec = sec;
    micro = now.tv_usec;
    t0micro = micro;
    return;}
  if ((now.tv_usec - tcmicro) < 0) {
    tcmicro -= 1000000;
    tcsec += 1;}
  siz = 1 + get_vsize() / (1024 * 1024);
  sprintf (str, "Total elapsed time is %ld.%06ld; size = %ld Mb.", 
	   (now.tv_sec - tcsec), (now.tv_usec - tcmicro), siz);
  to_stderr (str);
  t0sec = now.tv_sec;
  t0micro = now.tv_usec;
#endif
 }

extern void time_stamp (char* comment) {
#ifdef unixsys
  int len;
  long siz;
  timeval now;
  char str [256];
  gettimeofday(&now, NULL);
  if (sec == -1) {
    sec = now.tv_sec;
    t0sec = sec;
    micro = now.tv_usec;
    t0micro = micro;
    return;}
  if ((now.tv_usec - micro) < 0) {
    micro -= 1000000;
    sec += 1;}
  siz = 1 + get_vsize() / (1024 * 1024);
  strcpy(str, comment);
  len = strlen(str);
  sprintf(str + len, "; elapsed time is %ld.%06ld; size = %ld Mb.", 
	  (now.tv_sec - sec), (now.tv_usec - micro), siz);
  to_stderr (str);
  sec = now.tv_sec;
  micro = now.tv_usec;
#endif
 }

extern void die (char* err) {
  char str [256];
  time_stamp();
  strcpy(str, "Error *** ");
  strcat(str, err);
  strcat(str, " *** ");
  to_stderr(str);
  abort(); }

extern void die (char* err, int code) {
  char str [256];
  time_stamp();
  strcpy(str, "Error *** ");
  strcat(str, err);
  sprintf(str + strlen(str), ", id = %d (%x) *** ", code, code);
  to_stderr(str);
  abort(); }

extern void set_powers(int num, int *pairs) {
  for (int i = 0; i < number_of_variables; i++)
    pwr[i] = 0;
  for (int i = 0; i < num; i++)
    if (pairs[2 * i] >= 0) {
#ifdef debug
      if (pairs[2 * i] >= number_of_variables)
	die("invalid parameter to set_powers");
#endif
      pwr[pairs[2 * i]] = pairs[2 * i + 1];} }

extern void function_build(int numf, int *numtrm, int *numfct, int *fpwr, long long *co) {
  term* te;
  sum* su;
  int offset = 0;
  int fctoffset = 0;
#ifdef debug
  if (numf > (number_of_functions + number_of_ansatze))
    die("invalid parameter to function_build");
#endif
  for (int i = 0; i < numf; i++) {
    if (i < number_of_functions) {
      for (int j = 0; j < numtrm[i * (dimension + 1)]; j++) {      // term loop
	set_powers(numfct[offset + j], &fpwr[fctoffset * 2]);
	fctoffset += numfct[offset + j];
	if (function_terms[i] == NULL)
	  function_terms[i] = new term(&co[offset + j], pwr);
	else
	  function_terms[i] = (new term(&co[offset + j], pwr))->insert_term(function_terms[i]);}
      offset += numtrm[i * (dimension + 1)];
      functions[i] = new sum();
      te = function_terms[i];
      while (te != NULL) {
	functions[i]->add(&(te->coefficient),te->factor_powers);
	te = te->next_term();}}
    else {
      fctoffset += numfct[offset];
      offset += numtrm[i * (dimension + 1)];}
    for (int j = 0; j < dimension; j++) {
      su = new sum();
      for (int k = 0; k < numtrm[i * (dimension + 1) + j + 1]; k++) {  // term loop
	set_powers(numfct[offset + k], &fpwr[fctoffset * 2]);
	fctoffset += numfct[offset + k];
	su->add(&co[offset + k], pwr);}
      offset += numtrm[i * (dimension + 1) + j + 1];
      derivatives[i * dimension + j] = su;}} }

extern void entry_stats () {
#ifdef tracentry
  long t;
  int len = 0;
  char str[8192];
  if ((t =  trace_e) > 0) len += sprintf (str + len, " %ld entries into element,", t);
  if ((t =  trace_ke) > 0) len += sprintf (str + len, " %ld entries into ~element,", t);
  if ((t =  trace_td) > 0) len += sprintf (str + len, " %ld entries into dummy term,\n", t);
  if ((t =  trace_t) > 0) len += sprintf (str + len, " %ld entries into term,", t);
  if ((t =  trace_kt) > 0) len += sprintf (str + len, " %ld entries into ~term,", t);
  if ((t =  trace_int) > 0) len += sprintf (str + len, " %ld entries into invariant_term,\n", t);
  if ((t =  trace_kint) > 0) len += sprintf (str + len, " %ld entries into ~invariant_term,", t);
  if ((t =  trace_co) > 0) len += sprintf (str + len, " %ld entries into component(),", t);
  if ((t =  trace_co2) > 0) len += sprintf (str + len, " %ld entries into component(sum),\n", t);
  if ((t =  trace_kco) > 0) len += sprintf (str + len, " %ld entries into ~component,", t);
  if ((t =  trace_s) > 0) len += sprintf (str + len, " %ld entries into sum,", t);
  if ((t =  trace_ks) > 0) len += sprintf (str + len, " %ld entries into ~sum,\n", t);
  if ((t =  trace_i) > 0) len += sprintf (str + len, " %ld entries into invariant,", t);
  if ((t =  trace_ki) > 0) len += sprintf (str + len, " %ld entries into ~invariant,", t);
  if ((t =  trace_n1) > 0) len += sprintf (str + len, " %ld entries into ntuple(),\n", t);
  if ((t =  trace_n) > 0) len += sprintf (str + len, " %ld entries into ntuple,", t);
  if ((t =  trace_nal) > 0) len += sprintf (str + len, " %ld entries into ntuple(all),", t);
  if ((t =  trace_kn) > 0) len += sprintf (str + len, " %ld entries into ~ntuple,\n", t);
  if ((t =  trace_cc) > 0) len += sprintf (str + len, " %ld entries into connection,", t);
  if ((t =  trace_ccu) > 0) len += sprintf (str + len, " %ld entries into curvature,", t);
  if ((t =  trace_ei) > 0) len += sprintf (str + len, " %ld entries into element_init,\n", t);
  if ((t =  trace_ni) > 0) len += sprintf (str + len, " %ld entries into ntuple_init,", t);
  if ((t =  trace_ff) > 0) len += sprintf (str + len, " %ld entries into fraction_cd,", t);
  if ((t =  trace_tf) > 0) len += sprintf (str + len, " %ld entries into normalize,\n", t);
  if ((t =  trace_hc) > 0) len += sprintf (str + len, " %ld entries into hash_code,", t);
  if ((t =  trace_sh) > 0) len += sprintf (str + len, " %ld entries into set_hash,", t);
  if ((t =  trace_f) > 0) len += sprintf (str + len, " %ld entries into find,\n", t);
  if ((t =  trace_fi) > 0) len += sprintf (str + len, " %ld entries into free_index,", t);
  if ((t =  trace_csu) > 0) len += sprintf (str + len, " %ld entries into sub,", t);
  if ((t =  trace_mfil) > 0) len += sprintf (str + len, " %ld entries into make_free_index_list,\n", t);
  if ((t =  trace_msl) > 0) len += sprintf (str + len, " %ld entries into make_sum_list,", t);
  if ((t =  trace_ft) > 0) len += sprintf (str + len, " %ld entries into first_term,", t);
  if ((t =  trace_fit) > 0) len += sprintf (str + len, " %ld entries into first_invariant,\n", t);
  if ((t =  trace_fc) > 0) len += sprintf (str + len, " %ld entries into first_component,", t);
  if ((t =  trace_ie) > 0) len += sprintf (str + len, " %ld entries into insert_element,", t);
  if ((t =  trace_it) > 0) len += sprintf (str + len, " %ld entries into element.insert_term,\n", t);
  if ((t =  trace_sit) > 0) len += sprintf (str + len, " %ld entries into sum.insert_term,", t);
  if ((t =  trace_iah) > 0) len += sprintf (str + len, " %ld entries into insert_at_head,", t);
  if ((t =  trace_iit) > 0) len += sprintf (str + len, " %ld entries into insert_invariant_term,\n", t);
  if ((t =  trace_ic) > 0) len += sprintf (str + len, " %ld entries into insert_component,", t);
  if ((t =  trace_ne) > 0) len += sprintf (str + len, " %ld entries into next_element,", t);
  if ((t =  trace_nt) > 0) len += sprintf (str + len, " %ld entries into next_term,\n", t);
  if ((t =  trace_nint) > 0) len += sprintf (str + len, " %ld entries into next_invariant_term,", t);
  if ((t =  trace_nco) > 0) len += sprintf (str + len, " %ld entries into next_component,", t);
  if ((t =  trace_re) > 0) len += sprintf (str + len, " %ld entries into remove_element,\n", t);
  if ((t =  trace_rfl) > 0) len += sprintf (str + len, " %ld entries into remove_from_list,", t);
  if ((t =  trace_srt) > 0) len += sprintf (str + len, " %ld entries into sum.remove_term,", t);
  if ((t =  trace_irt) > 0) len += sprintf (str + len, " %ld entries into invariant.remove_term,\n", t);
  if ((t =  trace_rinl) > 0) len += sprintf (str + len, " %ld entries into remove_invariant_links,", t);
  if ((t =  trace_rfi) > 0) len += sprintf (str + len, " %ld entries into remove_from_invariant,", t);
  if ((t =  trace_rc) > 0) len += sprintf (str + len, " %ld entries into remove_component,\n", t);
  if ((t =  trace_nst) > 0) len += sprintf (str + len, " %ld entries into set(term),", t);
  if ((t =  trace_nss) > 0) len += sprintf (str + len, " %ld entries into set(sum),", t);
  if ((t =  trace_szq) > 0) len += sprintf (str + len, " %ld entries into sum.zeroq,\n", t);
  if ((t =  trace_sa) > 0) len += sprintf (str + len, " %ld entries into sum.add,", t);
  if ((t =  trace_smu) > 0) len += sprintf (str + len, " %ld entries into scalar_multiply,", t);
  if ((t =  trace_sz) > 0) len += sprintf (str + len, " %ld entries into zero,\n", t);
  if ((t =  trace_sumu) > 0) len += sprintf (str + len, " %ld entries into multiply,", t);
  if ((t =  trace_sd) > 0) len += sprintf (str + len, " %ld entries into derivative,", t);
  if ((t =  trace_ss) > 0) len += sprintf (str + len, " %ld entries into sum.simplify,\n", t);
  if ((t =  trace_ia) > 0) len += sprintf (str + len, " %ld entries into invariant.add,", t);
  if ((t =  trace_iz) > 0) len += sprintf (str + len, " %ld entries into invariant zero,", t);
  if ((t =  trace_im) > 0) len += sprintf (str + len, " %ld entries into invariant multiply,\n", t);
  if ((t =  trace_is) > 0) len += sprintf (str + len, " %ld entries into simplify,", t);
  if ((t =  trace_nz) > 0) len += sprintf (str + len, " %ld entries into ntuple zero,", t);
  if ((t =  trace_nc) > 0) len += sprintf (str + len, " %ld entries into ntuple copy,\n", t);
  if ((t =  trace_nan) > 0) len += sprintf (str + len, " %ld entries into ntuple add to ntuple,", t);
  if ((t =  trace_fil) > 0) len += sprintf (str + len, " %ld entries into fill,", t);
  if ((t =  trace_ncl) > 0) len += sprintf (str + len, " %ld entries into clone,\n", t);
  if ((t =  trace_na) > 0) len += sprintf (str + len, " %ld entries into ntuple.add,", t);
  if ((t =  trace_nsmu) > 0) len += sprintf (str + len, " %ld entries into ntuple.scalar_multiply,", t);
  if ((t =  trace_nms) > 0) len += sprintf (str + len, " %ld entries into ntuple.multiply,\n", t);
  if ((t =  trace_flr) > 0) len += sprintf (str + len, " %ld entries into fast_lorentz_raise,", t);
  if ((t =  trace_ntr) > 0) len += sprintf (str + len, " %ld entries into transpose,", t);
  if ((t =  trace_ncd) > 0) len += sprintf (str + len, " %ld entries into covariant_derivative,\n", t);
  if ((t =  trace_ncn) > 0) len += sprintf (str + len, " %ld entries into contract to ntuple,", t);
  if ((t =  trace_nci) > 0) len += sprintf (str + len, " %ld entries into contract to invariant,", t);
  if ((t =  trace_nsc) > 0) len += sprintf (str + len, " %ld entries into self_contract,\n", t);
  if ((t =  trace_nnc) > 0) len += sprintf (str + len, " %ld entries into ntuple_contract,", t);
  if ((t =  trace_sn) > 0) len += sprintf (str + len, " %ld entries into ntuple.simplify,", t);
  if ((t =  trace_ns) > 0) len += sprintf (str + len, " %ld entries into scalar,\n", t);
  if ((t =  trace_nsi) > 0) len += sprintf (str + len, " %ld entries into scalar_invariant,", t);
  if ((t =  trace_nr) > 0) len += sprintf (str + len, " %ld entries into raise,", t);
  if ((t =  trace_nl) > 0) len += sprintf (str + len, " %ld entries into lower,\n", t);
  if ((t =  trace_nd) > 0) len += sprintf (str + len, " %ld entries into dual,", t);
  if ((t =  trace_ndi) > 0) len += sprintf (str + len, " %ld entries into dual_invariant,", t);
  if ((t =  trace_gb) > 0) len += sprintf (str + len, " %ld entries into gamma_start,\n", t);
  if ((t =  trace_ge) > 0) len += sprintf (str + len, " %ld entries into gamma_stop,", t);
  if ((t =  trace_gin) > 0) len += sprintf (str + len, " %ld entries into gamma_index,", t);
  if ((t =  trace_gs) > 0) len += sprintf (str + len, " %ld entries into gamma_sign,\n", t);
  if ((t =  trace_gi) > 0) len += sprintf (str + len, " %ld entries into gamma_i,", t);
  if ((t =  trace_gam) > 0) len += sprintf (str + len, " %ld entries into gamma,", t);
  if ((t =  trace_eb) > 0) len += sprintf (str + len, " %ld entries into epsilon_start,\n", t);
  if ((t =  trace_ee) > 0) len += sprintf (str + len, " %ld entries into epsilon_stop,", t);
  if ((t =  trace_epi) > 0) len += sprintf (str + len, " %ld entries into epsilon_index,", t);
  if ((t =  trace_es) > 0) len += sprintf (str + len, " %ld entries into epsilon_sign,\n", t);
  if ((t = trace_cd) > 0) len += sprintf (str + len, " %ld entries into check_done,", t);
  if ((t = trace_ps) > 0) len += sprintf (str + len, " %ld entries into parallelize_start,", t);
  if ((t = trace_pst) > 0) len += sprintf (str + len, " %ld entries into parallelize_step,\n", t);
  if ((t = trace_tin) > 0) len += sprintf (str + len, " %ld entries into tids index,", t);
  if ((t = trace_ti) > 0) len += sprintf (str + len, " %ld entries into tids_initialize,", t);
  if ((t = trace_pi) > 0) len += sprintf (str + len, " %ld entries into parallel_initialize,\n", t);
  if ((t = trace_ip) > 0) len += sprintf (str + len, " %ld entries into init_parcels,", t);
  if ((t = trace_sp) > 0) len += sprintf (str + len, " %ld entries into sync_process,", t);
  if ((t = trace_si) > 0) len += sprintf (str + len, " %ld entries into sync_indices,\n", t);
  if ((t = trace_spa) > 0) len += sprintf (str + len, " %ld entries into start_parcel,", t);
  if ((t = trace_as) > 0) len += sprintf (str + len, " %ld entries into add_sum,", t);
  if ((t = trace_ci) > 0) len += sprintf (str + len, " %ld entries into check_incoming,\n", t);
  if ((t = trace_spar) > 0) len += sprintf (str + len, " %ld entries into send,", t);
  if ((t = trace_sps) > 0) len += sprintf (str + len, " %ld entries into special send,", t);
  if ((t = trace_cspar) > 0) len += sprintf (str + len, " %ld entries into send component,\n", t);
  if ((t = trace_tspar) > 0) len += sprintf (str + len, " %ld entries into send term,", t);
  if ((t = trace_com) > 0) len += sprintf (str + len, " %ld entries into commit,", t);
  if ((t = trace_npn) > 0) len += sprintf (str + len, " %ld entries into parallelize ntuple,\n", t);
  if ((t = trace_nps) > 0) len += sprintf (str + len, " %ld entries into parallelize sum,", t);
  if ((t = trace_nfp) > 0) len += sprintf (str + len, " %ld entries into free_parallelize,", t);
  if ((t = trace_npd) > 0) len += sprintf (str + len, " %ld entries into parallelize_dual", t);
  to_stderr (str);
#endif
  close_std (); }

extern void clear_entries () {
#ifdef tracentry
  trace_ff = 0;
  trace_tf = 0;
  trace_ei = 0;
  trace_e = 0;
  trace_ke = 0;
  trace_ne = 0;
  trace_ie = 0;
  trace_re = 0;
  trace_td = 0;
  trace_t = 0;
  trace_kt = 0;
  trace_nt = 0;
  trace_it = 0;
  trace_int = 0;
  trace_kint = 0;
  trace_nint = 0;
  trace_iit = 0;
  trace_rinl = 0;
  trace_hc = 0;
  trace_sh = 0;
  trace_iah = 0;
  trace_rfl = 0;
  trace_sit = 0;
  trace_srt = 0;
  trace_ft = 0;
  trace_f = 0;
  trace_sa = 0;
  trace_smu = 0;
  trace_sumu = 0;
  trace_nsmu = 0;
  trace_nms = 0;
  trace_flr = 0;
  trace_ntr = 0;
  trace_szq = 0;
  trace_fit = 0;
  trace_irt = 0;
  trace_ia = 0;
  trace_rfi = 0;
  trace_co = 0;
  trace_co2 = 0;
  trace_kco = 0;
  trace_nco = 0;
  trace_ic = 0;
  trace_rc = 0;
  trace_fc = 0;
  trace_fi = 0;
  trace_csu = 0;
  trace_nst = 0;
  trace_nss = 0;
  trace_na = 0;
  trace_n1 = 0;
  trace_n = 0;
  trace_nal = 0;
  trace_kn = 0;
  trace_ni = 0;
  trace_mfil = 0;
  trace_msl = 0;
  trace_ncn = 0;
  trace_nci = 0;
  trace_nz = 0;
  trace_nc = 0;
  trace_nan = 0;
  trace_fil = 0;
  trace_ncl = 0;
  trace_nr = 0;
  trace_nl = 0;
  trace_ncd = 0;
  trace_nsc = 0;
  trace_nnc = 0;
  trace_nd = 0;
  trace_ns = 0;
  trace_nsi = 0;
  trace_ndi = 0;
  trace_gb = 0;
  trace_ge = 0;
  trace_gin = 0;
  trace_gs = 0;
  trace_gi = 0;
  trace_gam = 0;
  trace_eb = 0;
  trace_ee = 0;
  trace_epi = 0;
  trace_es = 0;
  trace_cc = 0;
  trace_ccu = 0;
  trace_s = 0;
  trace_ks = 0;
  trace_sz = 0;
  trace_sm = 0;
  trace_sd = 0;
  trace_ss = 0;
  trace_sn = 0;
  trace_i = 0;
  trace_ki = 0;
  trace_iz = 0;
  trace_im = 0;
  trace_is = 0;
  trace_cd = 0;
  trace_ps = 0;
  trace_pst = 0;
  trace_tin = 0;
  trace_ti = 0;
  trace_pi = 0;
  trace_ip = 0;
  trace_sp = 0;
  trace_si = 0;
  trace_spa = 0;
  trace_as = 0;
  trace_ci = 0;
  trace_spar = 0;
  trace_sps = 0;
  trace_cspar = 0;
  trace_tspar = 0;
  trace_com = 0;
  trace_npn = 0;
  trace_nps = 0;
  trace_nfp = 0;
  trace_npd = 0;
#endif
 }

#ifdef keepstats
#ifdef debug
void head_stats() {
  int max_count = 0, len = 0;
  char str[2048];
  for (int i = 0; i < large_hash_size; i++)
    if (keep_head_count[i] > max_count)
      max_count = keep_head_count[i];
  int* bins = new int[max_count + 1];
  if (bins == NULL)
    die("allocating bins");
  memset(bins, 0, (max_count + 1) * sizeof(int));
  for (int i = 0; i < large_hash_size; i++)
    bins[keep_head_count[i]] += 1;
  to_stderr("Headcount bin ");
  for (int i = 0; i <= max_count; i++) {
    if (bins[i] > 0) {
      len += sprintf (str + len, "%d = %d, ", i , bins[i]);}}
  to_stderr (str);;
  delete [] bins; }
#endif
#endif

#ifdef unixsys
FILE *fe, *fo;
bool erropen = false, outopen = false;
#endif

extern void to_stderr (char* msg) {
  char str [8200];
  int len = 0, rc = 0;
#ifndef unixsys
  cerr << msg << "\n";
#endif
#ifdef unixsys
  time_t now;
  tm *now_tm;
  now = time (NULL);
  now_tm = localtime (&now);
  len = strftime (str, 30, "%c ", now_tm);
  strcpy (str + len, msg);
  if (!erropen) {
    fe = fopen ("/home/ptahusr/ptah.err", "a");
    if(fe == NULL) {
      perror ("opening /home/ptahusr/ptah.err");
      abort ();}
    erropen = true;}
  len = strlen (str);
  rc = fwrite (str, 1, len, fe);
  if (rc != len) {
    perror ("writing to /home/ptahusr/ptah.err");
    abort ();}
  fwrite ("\n", 1, 1, fe);
  fflush (fe);
#endif
 }

extern void to_stdout (char* msg) {
  int len, rc = 0;
#ifndef unixsys
  cout << msg << "\n";
#endif
#ifdef unixsys
  if (!outopen) {
    fo = fopen ("/home/ptahusr/ptah.out", "a");
    if(fo == NULL) {
      perror ("opening /home/ptahusr/ptah.out");
      die ("opening /home/ptahusr/ptah.out");}
    fwrite ("(*  *)\n", 1, 7, fo);
    outopen = true;}
  len = strlen (msg);
  rc = fwrite (msg, 1, len, fo);
  if (rc != len) {
    perror ("writing to /home/ptahusr/ptah.out");
    die ("writing to /home/ptahusr/ptah.out");}
  fwrite ("\n", 1, 1, fo);
  fflush (fo);
#endif
 }

extern void close_std () {
  fclose (fe);
  erropen = false;
  fclose (fo);
  outopen = false;
  sync (); }
