//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// ntuple.cc

#include <ptah.h>
#include <element.h>
#include <hash.h>
#include <ntuple.h>

extern ntuple::ntuple () {
#ifdef tracentry
  trace_n1 += 1;
#endif
#ifdef debug
  hash_magic_no = hash_magic;
  ntuple_magic_no = ntuple_magic;
#endif
  head = new element* [regular_hash_size];
  if (head == NULL)
    die ("allocating head");
#ifdef keepstats
#ifdef debug
  headcount = new int [regular_hash_size];
  if (headcount == NULL)
    die ("allocating headcount");
  memset (headcount, 0, regular_hash_size * sizeof (int));
#endif
#endif
  metric = false;
  size = regular_hash_size;
  last_hash = -1;
  number_of_components = 0;
  free_index_list = NULL;
  sum_list = NULL;
  indices_list = NULL;
  casize = -1;
  for (int i = 0; i < size; i++)
    head[i] = NULL; }

extern ntuple::ntuple (int rank_param) {
#ifdef tracentry
  trace_n += 1;
#endif
#ifdef debug
  hash_magic_no = hash_magic;
  ntuple_magic_no = ntuple_magic;
#endif
  head = new element* [regular_hash_size];
  if (head == NULL)
    die ("allocating head");
#ifdef keepstats
#ifdef debug
  headcount = new int [regular_hash_size];
  if (headcount == NULL)
    die ("allocating headcount");
  memset (headcount, 0, regular_hash_size * sizeof (int));
#endif
#endif
  metric = false;
  size = regular_hash_size;
  rank = rank_param;
  last_hash = -1;
  number_of_components = 0;
  for (int i = 0; i < size; i++)
    head[i] = NULL; 
  ntuple_init (); }

extern ntuple::ntuple (int rank_param, bool* upper_param, bool* antisymmetric_param, 
		      bool* symmetric_param, bool* spinor_param) {
  casize = 1;
#ifdef tracentry
  trace_nal += 1;
#endif
#ifdef debug
  hash_magic_no = hash_magic;
  ntuple_magic_no = ntuple_magic;
  for (int i = 0; i < rank_param; i++)
    if (antisymmetric_param[i] && symmetric_param[i])
      die ("confusion in symmetrization of indices");
  if ((((rank_param - 1) % 2) == 0) && antisymmetric_param[rank_param -1])
    die ("antisymmetrization must be in pairs");
  if ((((rank_param - 1) % 2) == 0) && symmetric_param[rank_param -1])
    die ("symmetrization must be in pairs");
#endif
  head = new element* [regular_hash_size];
  if (head == NULL)
    die ("allocating head");
#ifdef keepstats
#ifdef debug
  headcount = new int [regular_hash_size];
  if (headcount == NULL)
    die ("allocating headcount");
  memset (headcount, 0, regular_hash_size * sizeof (int));
#endif
#endif
  metric = false;
  size = regular_hash_size;
  rank = rank_param;
  index_list = new int [rank + 1];
  if (index_list == NULL)
    die ("allocating index_list");
  last_hash = -1;
  number_of_components = 0;
  for (int i = 0; i < size; i++)
    head[i] = NULL;
  free_index_list = NULL;
  sum_list = NULL;
  indices_list = NULL;
  upper = new bool [rank];
  if (upper == NULL)
    die ("allocating upper");
  antisymmetric = new bool [rank];
  if (antisymmetric == NULL)
    die ("allocating antisymmetric");
  symmetric = new bool [rank];
  if (symmetric == NULL)
    die ("allocating symmetric");
  spinor = new bool [rank];
  if (spinor == NULL)
    die ("allocating spinor");
  for (int i = 0; i < rank; i++) {
    upper[i] = upper_param[i];
    antisymmetric[i] = antisymmetric_param[i];
    symmetric[i] = symmetric_param[i];
    spinor[i] = spinor_param[i];}
  if (rank > 6)
    casize = -1;
  else {
    for (int i = 0; i < rank; i++)
      if (spinor[i])
	casize *= spinor_dimension;
      else
	casize *= dimension;
    comparray = new component* [casize];
    if (comparray == NULL)
      die ("allocating comparray");
    memset (comparray, NULL, casize * sizeof (component*));} }

extern ntuple::~ntuple () {
#ifdef tracentry
  trace_kn += 1;
#endif
#ifdef debug
  for (int i = 0; i < size; i++)
    if (head[i] != NULL)
      die ("attempting to delete non-empty ntuple");
  ntuple_magic_no = -ntuple_magic;
#endif
  delete [] index_list;
  if (free_index_list != NULL) {
    delete [] free_index_list;
    free_index_list = NULL;}
  if (sum_list != NULL) {
    delete [] sum_list;
    sum_list = NULL;}
  if (indices_list != NULL) {
    delete [] indices_list;
    indices_list = NULL;}
  delete [] upper;
  delete [] antisymmetric;
  delete [] symmetric;
  delete [] spinor;
  if (rank <= 6)
    delete [] comparray;
  casize = -1;
#ifdef keepstats
#ifdef debug
  delete [] headcount;
#endif
#endif
  delete [] head; }

extern void ntuple::ntuple_init () {
  casize = 1;
#ifdef tracentry
  trace_ni += 1;
#endif
  metric = false;
  free_index_list = NULL;
  sum_list = NULL;
  indices_list = NULL;
  index_list = new int [rank + 1];
  if (index_list == NULL)
    die ("allocating index_list");
  upper = new bool [rank];
  if (upper == NULL)
    die ("allocating upper");
  antisymmetric = new bool [rank];
  if (antisymmetric == NULL)
    die ("allocating antisymmetric");
  symmetric = new bool [rank];
  if (symmetric == NULL)
    die ("allocating symmetric");
  spinor = new bool [rank];
  if (spinor == NULL)
    die ("allocating spinor");
  for (int i = 0; i < rank; i++) {
    upper[i] = false;
    antisymmetric[i] = false;
    symmetric[i] = false;
    spinor[i] = false;}
  if (rank > 6)
    casize = -1;
  else {
    for (int i = 0; i < rank; i++)
      casize *= dimension;
    comparray = new component* [casize];
    if (comparray == NULL)
      die ("allocating comparray");
    memset (comparray, NULL, casize * sizeof (component*));} }

extern void ntuple::zero () {
#ifdef tracentry
  trace_nz += 1;
#endif
  component *current_component;
  for (int i = 0; i < regular_hash_size; i++) {
    current_component = first_component (i);
    while (current_component != NULL) {
      (current_component->value)->zero ();
      set_hash (i);
      remove_component (current_component);
      current_component = first_component (i);}}
  if (rank <= 6)
    memset (comparray, NULL, casize * sizeof (component*)); }

extern void ntuple::copy (ntuple* from) {
#ifdef tracentry
  trace_nc += 1;
#endif
  long long one = 1;
  component *current_component;
  for (int i = 0; i < regular_hash_size; i++) {
    current_component = from->first_component (i);
    while (current_component != NULL) {
      if (current_component->value != NULL)
	if (! (current_component->value)->zeroq ()) {
	  sum* new_sum = new sum;
	  if (new_sum == NULL)
	    die ("allocating new_sum");
	  new_sum->multiply (1, &(current_component->value), &one);
	  for (int j = 0; j < rank; j++)
	    index_list[j] = (current_component->indices >> (6 * j)) & 63;
	  set (index_list, new_sum);}
      current_component = current_component->next_component ();}} }

extern void ntuple::add (ntuple* from) {
#ifdef tracentry
  trace_nan += 1;
#endif
#ifdef debug
  if (rank != from->rank)
    die ("rank mismatch in ntuple add");
  bool upchk = true, spchk = true;
  for (int i = 0; i < rank; i++) {
    upchk = upchk && (upper[i] == from->upper[i]);
    spchk = spchk && (spinor[i] == from->spinor[i]);}
  if (! upchk)
    die ("index level mismatch in ntuple add");
  if (! spchk)
    die ("spinor mismatch in ntuple add");
#endif
  component *current_component;
  for (int i = 0; i < regular_hash_size; i++) {
    current_component = from->first_component (i);
    while (current_component != NULL) {
      for (int j = 0; j < rank; j++)
	index_list[j] = (current_component->indices >> (6 * j)) & 63;
      add (index_list, current_component->value);
      current_component = current_component->next_component ();}} }

extern ntuple* ntuple::clone () {
#ifdef tracentry
  trace_ncl += 1;
#endif
  ntuple* new_ntuple = new ntuple (rank, upper, antisymmetric, symmetric, spinor);
  if (new_ntuple == NULL)
    die ("allocating new_ntuple");
  new_ntuple->copy (this);
  new_ntuple->metric = metric;
  return new_ntuple; }

extern void ntuple::fill () {
#ifdef tracentry
  trace_fil += 1;
#endif
  bool needit = false;
  int t;
  long long one = 1;
  long long minus_one = -1;
  component *current_component, *new_component;
  for (int i = 0; i < rank; i++)
    if (antisymmetric[i] || symmetric[i]) {
      needit = true;
      break;}
  if (!needit)
    return;
  for (int r = 0; r < rank - 1; r++) {
    if (!(symmetric[r] || antisymmetric[r]))
      continue;
    for (int i = 0; i < regular_hash_size; i++) {
      current_component = first_component (i);
      while (current_component != NULL) {
	if (current_component->value != NULL)
	  if (! (current_component->value)->zeroq ()) {
	    for (int j = 0; j < rank; j++) 
	      index_list[j] = (current_component->indices >> (6 * j)) & 63;
	    if (antisymmetric[r] && (index_list[r] < index_list[r + 1])) {
	      t = index_list[r];
	      index_list[r] = index_list[r + 1];
	      index_list[r + 1] = t;
	      new_component = sub (index_list);
	      if (new_component == NULL) {
		sum* new_sum = new sum;
		if (new_sum == NULL)
		  die ("allocating new_sum");
		new_sum->multiply (1, &(current_component->value), &minus_one);
		set (index_list, new_sum);}}
	    if (symmetric[r] && (index_list[r] < index_list[r + 1])) {
	      t = index_list[r];
	      index_list[r] = index_list[r + 1];
	      index_list[r + 1] = t;
	      new_component = sub (index_list);
	      if (new_component == NULL) {
		sum* new_sum = new sum;
		if (new_sum == NULL)
		  die ("allocating new_sum");
		new_sum->multiply (1, &(current_component->value), &one);
		set (index_list, new_sum);}}}
	current_component = current_component->next_component ();}}
    if (symmetric[r] || antisymmetric[r])
      r += 1;} }

extern ntuple* ntuple::covariant_derivative () {
#ifdef tracentry
  trace_ncd += 1;
#endif
#ifdef debug
  for (int i = 0; i < rank; i++) {
    if (upper[i])
      die ("attempting to take covder of tensor with upper indices");}
#endif
  int sdim, tdim;
  sum* new_sum = new sum;
  if (new_sum == NULL)
    die ("allocating new_sum");
  sum** multiply_list = new sum* [2];
  if (multiply_list == NULL)
    die ("allocating multiply_list");
  bool* new_upper = new bool [rank + 1];
  if (new_upper == NULL)
    die ("allocating new_upper");
  bool* new_antisymmetric = new bool [rank + 1];
  if (new_antisymmetric == NULL)
    die ("allocating new_antisymmetric");
  bool* new_symmetric = new bool [rank + 1];
  if (new_symmetric == NULL)
    die ("allocating new_symmetric");
  bool* new_spinor = new bool [rank + 1];
  if (new_spinor == NULL)
    die ("allocating new_spinor");
  long long one = 1;
  long long minus_one = -1;
  component* save_component;
  for(int i = 0; i < rank; i++) {
    new_upper[i] = upper[i];
    new_antisymmetric[i] = antisymmetric[i];
    new_symmetric[i] = symmetric[i];
    new_spinor[i] = spinor[i];}
  new_upper[rank] = false;
  new_antisymmetric[rank] = false;
  new_symmetric[rank] = false;
  new_spinor[rank] = false;
  ntuple* new_ntuple = new ntuple (rank + 1, new_upper, new_antisymmetric, new_symmetric, new_spinor);
  if (new_ntuple == NULL)
    die ("allocating new_ntuple");
              // cannot do this with free_index_list !
  memset (index_list, 0, rank * sizeof (int));
  if (spinor[0])
    sdim = spinor_dimension;
  else
    sdim = dimension;
  while (index_list[0] < sdim) {
    save_component = sub (index_list);
    for (int var = 0; var < dimension; var++) {
      if (save_component != NULL)
	new_sum->derivative (save_component->value, var, &one);
      christoffel->index_list[2] = var;
      for (int r = 0; r < rank; r++) {
	if (spinor[r])
	  continue;
	christoffel->index_list[1] = index_list[r];
	for (int n = 0; n < dimension; n++) {
	  index_list[r] = n;
	  current_component = sub (index_list);
	  if (current_component != NULL) {
	    multiply_list[0] = current_component->value;
	    christoffel->index_list[0] = n;
	    current_component = christoffel->sub (christoffel->index_list);
	    if (current_component != NULL) {
	      multiply_list[1] = current_component->value;
	      new_sum->multiply (2, multiply_list, &minus_one);}}}
	index_list[r] = christoffel->index_list[1];}
      index_list[rank] = var;
      new_sum->simplify ();
      new_ntuple->add (index_list, new_sum);
      new_sum->zero ();}
    for (int i = rank - 1; i >= 0; i--) {
      index_list[i] += 1;
      if (spinor[i])
	tdim = spinor_dimension;
      else
	tdim = dimension;
      if ((index_list[i] >= tdim) && (i > 0))
	index_list[i] = 0;
      else
	break;}}
//    new_ntuple->fill ();
  delete new_sum;
  delete [] new_upper;
  delete [] new_antisymmetric;
  delete [] new_symmetric;
  delete [] multiply_list;
  return new_ntuple; }

extern void ntuple::display (char* name) {
  char label [256];
  char label_index [3];
  int mathematica_index;
#ifdef debug
  int our_index_list [12];
#endif
  for (int i = 0; i < regular_hash_size; i++) {
    current_component = first_component (i);
    while (current_component != NULL) {
      strcpy (label, name);
      strcat (label, "[[");
      for (int j = 0; j < rank; j++) {
	mathematica_index = (current_component->indices >> (6 * j)) & 63;
#ifdef debug
	our_index_list[j] = mathematica_index;
#endif
	if (!spinor[j]) {
	  if (mathematica_index == 0)
	    mathematica_index = dimension;}
	else
	  mathematica_index += 1;
	sprintf (label_index, "%d", mathematica_index);
	strcat (label, (char *) &label_index);
	if (j < (rank - 1))
	  strcat (label, ",");}
#ifdef debug
      if (rank <= 6)
	if (! comparray[ca_index (our_index_list)])
	  die ("comparray component identified as zero");
#endif
      strcat (label, "]]");
      (current_component->value)->display (label);
      current_component = current_component->next_component ();}} }

extern void ntuple::display_add (char* name) {
  char label [256];
  char label_index [3];
  int mathematica_index;
#ifdef debug
  int our_index_list [12];
#endif
  for (int i = 0; i < regular_hash_size; i++) {
    current_component = first_component (i);
    while (current_component != NULL) {
      strcpy (label, name);
      strcat (label, "[[");
      for (int j = 0; j < rank; j++) {
	mathematica_index = (current_component->indices >> (6 * j)) & 63;
#ifdef debug
	our_index_list[j] = mathematica_index;
#endif
	if (!spinor[j]) {
	  if (mathematica_index == 0)
	    mathematica_index = dimension;}
	else
	  mathematica_index += 1;
	sprintf (label_index, "%d", mathematica_index);
	strcat (label, (char *) &label_index);
	if (j < (rank - 1))
	  strcat (label, ",");}
#ifdef debug
      if (rank <= 6)
	if (! comparray[ca_index (our_index_list)])
	  die ("comparray component identified as zero");
#endif
      strcat (label, "]]");
      (current_component->value)->display_add (label);
      current_component = current_component->next_component ();}} }

#ifdef debug
extern void ntuple::dump (char* name) {
  char label [256];
  char label_index [3];
  int mathematica_index;
#ifdef debug
  int our_index_list [12];
#endif
  bool check;
  if (!parent)
    return;
  cerr << "Dump of n-tuple " << name << ": \n";
  cerr << "     rank = " << rank << " \n";
  cerr << "     number of components = " << number_of_components << " \n";
  cerr << "     metric = " << metric << " \n";
  check = false;
  for (int i = 0; i < rank; i++)
    check = check || upper[i];
  if (check) {
    cerr << "     indices ";
    for (int i = 0; i < rank; i++)
      if (upper[i])
	cerr << i << " ";
    cerr << "are upper \n";}
  check = false;
  for (int i = 0; i < rank; i++)
    check = check || antisymmetric[i];
  if (check) {
    cerr << "     indices ";
    for (int i = 0; i < rank; i++)
      if (antisymmetric[i])
	cerr << i << " ";
    cerr << "are antisymmetric \n";}
  check = false;
  for (int i = 0; i < rank; i++)
    check = check || symmetric[i];
  if (check) {
    cerr << "     indices ";
    for (int i = 0; i < rank; i++)
      if (symmetric[i])
	cerr << i << " ";
    cerr << "are symmetric \n";}
  cerr << "     index_list = ";
  for (int i = 0; i < rank; i++)
    cerr << index_list[i] << " ";
  cerr << " \n";
  cerr << "     current_free_index = " << current_free_index << " \n"; 
  cerr << "     free_index_length = " << free_index_length << " \n"; 
  if ((free_index_list != NULL) && (free_index_length > 0)) {
    cerr << "     free_index_list = ";
    for (int i = 0; i < free_index_length; i++) {
      cerr << "( ";
      for (int j = 0; j < rank; j++)
	cerr << ((free_index_list[i] >> (6 * j)) & 63) << " ";
      cerr << ") ";}
    cerr << " \n";}
  cerr << "     current_sum_index = " << current_sum_index << " \n"; 
  cerr << "     sum_list_length = " << sum_list_length << " \n"; 
  if ((sum_list != NULL) && (sum_list_length > 0)) {
    cerr << "     sum_list = ";
    for (int i = 0; i < sum_list_length; i++) {
      cerr << "( ";
      current_component = sum_list[i];
      for (int j = 0; j < rank; j++)
	cerr << ((current_component->indices >> (6 * j)) & 63) << " ";
      cerr << ") ";}
    cerr << " \n";}
  if (casize > 0) {
    cerr << "     comparray indices = ";
    for (int i = 0; i < casize; i++)
      if (comparray[i])
	cerr << i << ", ";
    cerr << " \n";}
  for (int i = 0; i < regular_hash_size; i++) {
    current_component = first_component (i);
    if (current_component != NULL)
      cerr << "        hash list " << i << " \n";
    while (current_component != NULL) {
      strcpy (label, name);
      strcat (label, "[[");
      for (int j = 0; j < rank; j++) {
	mathematica_index = (current_component->indices >> (6 * j)) & 63;
#ifdef debug
	our_index_list[j] = mathematica_index;
#endif
	if (!spinor[j]) {
	  if (mathematica_index == 0)
	    mathematica_index = dimension;}
	else
	  mathematica_index += 1;
	sprintf (label_index, "%d", mathematica_index);
	strcat (label, (char *) &label_index);
	if (j < (rank - 1))
	  strcat (label, ",");}
#ifdef debug
      if (rank <= 6)
	if (! comparray[ca_index (our_index_list)])
	  die ("comparray component identified as zero");
#endif
      strcat (label, "]]");
      (current_component->value)->dump (label);
      current_component = current_component->next_component ();}} }
#endif

extern void ntuple::tensor_build (int numc, int *ind, int *numtrm, int *numfct, int *gpwr, long long *co) {
  sum* su;
  int offset = 0;
  int fctoffset = 0;
  for (int i = 0; i < numc; i++) {                        // component loop
    if (numtrm[i] == 1) {
      set_powers (numfct[offset], &gpwr[fctoffset * 2]);
      fctoffset += numfct[offset];
      set (&ind[i * rank], new term (&co[offset], pwr));}
    else {
      su = new sum;
      for (int j = 0; j < numtrm[i]; j++) {               // term loop
	set_powers (numfct[offset + j], &gpwr[fctoffset * 2]);
	fctoffset += numfct[offset + j];
	su->add (&co[offset + j], pwr);}
      set (&ind[i * rank], su);}
    offset += numtrm[i];} }
