//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// parallel.cc

#include <ptah.h>
#include <element.h>
#include <hash.h>
#include <ntuple.h>
#include <signal.h>
#include <errno.h>

bool parallel_start = true;

extern bool ntuple::parallelize (ntuple* dest) {
#ifdef tracentry
  trace_npn += 1;
#endif
  if (sum_list_length <= 0)
    return false;
  if (!parallel) {
    if (parallel_start) {
      current_sum_index = 0;
      parallel_start = false;
      return true;}
    current_sum_index += 1;
    if (current_sum_index >= sum_list_length) {
      parallel_start = true;
      return false;}
    return true;}
  die ("trying to run parallel"); }

extern bool ntuple::parallelize (sum* dest) {
#ifdef tracentry
  trace_nps += 1;
#endif
  if (sum_list_length <= 0)
    return false;
  if (!parallel) {
    if (parallel_start) {
      current_sum_index = 0;
      parallel_start = false;
      return true;}
    current_sum_index += 1;
    if (current_sum_index >= sum_list_length) {
      parallel_start = true;
      return false;}
    return true;}
  die ("trying to run parallel"); }

extern bool ntuple::free_parallelize (ntuple* dest) {
#ifdef tracentry
  trace_nfp += 1;
#endif
  if (free_index_length <= 0)
    return false;
  if (!parallel) {
    if (parallel_start) {
      current_free_index = 0;
      parallel_start = false;
      return true;}
    current_free_index += 1;
    if (current_free_index >= free_index_length) {
      parallel_start = true;
      return false;}
    return true;}
  die ("trying to run parallel"); }

extern bool ntuple::parallelize_dual (sum* dest) {
  int ep;
  bool sum_not_done = true;
#ifdef tracentry
  trace_npd += 1;
#endif
  if (!parallel) {
    if (parallel_start) {
      memset(which_epsilon, 0, number_of_epsilons *sizeof(int));
      parallel_start = false;
      return true;}
    ep = number_of_epsilons - 1;                    // adjust epsilon indices
    while (sum_not_done) {
      which_epsilon[ep] += 1;
      if (which_epsilon[ep] >= epsilon_length) {    // move to next epsilon index
	which_epsilon[ep] = 0;
	ep -= 1;
	if (ep < 0) {
	  parallel_start = true;
	  sum_not_done = false;}}
      else
	break;}
    return sum_not_done;}
  die ("trying to run parallel"); }
