//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// paramize.cc

#include <ptah.h>
#include <element.h>
#include <hash.h>
#include <ntuple.h>

extern bool ntuple::make_free_index_list(bool limit_redundant_indices = false) {
#ifdef tracentry
  trace_mfil += 1;
#endif
  int free_size = 1, index = 0;
  int compare, node;
  int* check_indices;
  int* left;
  int* right;
  int* nodes;
#ifdef debug
//    char str [256];
#endif
  long long* temp_index_list;
  long long index_value;
  bool found, ret = false;
  free_index_length = 0;
  if (free_index_list != NULL)
    delete [] free_index_list;
  for (int i = 0; i < rank; i++)
    if (index_list[i] < 0) {                          // (-1) -> free index
      if (spinor[i])
	free_size *= spinor_dimension;
      else
	free_size *= dimension;}
  if (free_size == 1) {
    free_index_list = NULL;
    return false;}
  temp_index_list = new long long [free_size];
  if (temp_index_list == NULL)
    die("allocating temp_index_list");
  check_indices = new int[rank];
  if (check_indices == NULL)
    die("allocating check_indices");
  for (int i = 0; i < regular_hash_size; i++) {
    current_component = first_component(i);
    while (current_component != NULL) {
      index_value = current_component->indices;
      if (limit_redundant_indices) {
	for (int j = 0; j < rank; j++) 
	  check_indices[j] = (index_value >> (6 * j)) & 63;
	found = false;
	for (int r = 0; r < rank - 1; r++) {
	  if ((index_list[r] >= 0) || (index_list[r + 1] >= 0))
	    continue;   // don't limit redundant indices on summed indices
	  if (antisymmetric[r] && (check_indices[r + 1] <= check_indices[r])) {
	    found = true;
	    break;}
	  if (symmetric[r] && (check_indices[r + 1] < check_indices[r])) {
	    found = true;
	    ret = true;
	    break;}
	  if (antisymmetric[r] || symmetric[r])
	    r += 1;}
	if (found) {
	  current_component = current_component->next_component();
	  continue;}}
      for (int j = 0; j < rank; j++)
	if (index_list[j] >= 0)                       // sum index
	  index_value |= 63 << (6 * j);
      found = false;
      for (int j = index - 1; j >= 0; j--)
	if (index_value == temp_index_list[j]) {
	  found = true;
	  break;}
      if (!found) {
#ifdef debug
	if ((index < 0) || (index >= free_size))
	  die("free_index out of bounds");
#endif
	temp_index_list[index] = index_value;
	index += 1;}
      current_component = current_component->next_component();}}
  free_index_length = index;
  left = new int [free_index_length];            // binary tree sort so all tasks correspond
  if (left == NULL)
    die("allocating left");
  right = new int [free_index_length];
  if (right == NULL)
    die("allocating right");
  nodes = new int [free_index_length];
  if (nodes == NULL)
    die("allocating nodes");
  memset(left, 0, free_index_length * sizeof(int));
  memset(right, 0, free_index_length * sizeof(int));
  index = 0;
  compare = 1;
  while (compare < free_index_length) {
#ifdef debug
    if (temp_index_list[compare] == temp_index_list[index])
      die("sorting free indices");
#endif
    if (temp_index_list[compare] < temp_index_list[index]) {
      if (left[index] == 0) {
	left[index] = compare;
	compare += 1;
	index = 0;
	continue;}
      index = left[index];
      continue;}
    if (right[index] == 0) {
      right[index] = compare;
      compare += 1;
      index = 0;
      continue;}
    index = right[index];}
  free_index_list = new long long [free_index_length];
  if (free_index_list == NULL)
    die("allocating free_index_list");
  index = 0;
  compare = 0;
  node = -1;
  while (index < free_index_length) {
    if (left[compare] > 0) {
      node += 1;
#ifdef debug
      if (node >= free_index_length)
	die("node failure in free index sort");
#endif
      nodes[node] = compare;
      compare = left[compare];
      left[nodes[node]] = 0;
      continue;}
    if (temp_index_list[compare] >= 0) {          // hasn't already been inserted
      free_index_list[index] = temp_index_list[compare];
      temp_index_list[compare] = -1;
      index += 1;
      if (index >= free_index_length)
	break;}
    if (right[compare] > 0) {
      node += 1;
#ifdef debug
      if (node >= free_index_length)
	die("node failure in free index sort");
#endif
      nodes[node] = compare;
      compare = right[compare];
      right[nodes[node]] = 0;
      continue;}
    if (node < 0)       // must be completed unbalanced!
      break;
    compare = nodes[node];
    node -= 1;}
#ifdef debug
  if (index != free_index_length)
    die("free index sort ended prematurely");
//    sprintf(str, "Free_index_length = %d.", free_index_length);
//    to_stderr (str);
//    strcpy(str, "Free indices =");
//    for (int i = 0; i < free_index_length; i++) {
//      sprintf(str + strlen(str), " %Lo", free_index_list[i]);
//      if (((i % 8) == 7)  || (i == (free_index_length - 1))) {
//        to_stderr (str);}}
#endif
  delete [] check_indices;
  delete [] temp_index_list;
  delete [] left;
  delete [] right;
  delete [] nodes;
  return ret; }

extern int ntuple::make_sum_list(int number_of_ntuples, ntuple** ntuple_list, 
					bool limit_redundant_indices = false) {
                     // all indices summed over must appear in "this"'s index_list
  // free indices occur in order throughout ntuple_list EXCEPT for metric indices,
  //     which occur in place of those of the original ntuple being raised / lowered
#ifdef tracentry
  trace_msl += 1;
#endif
  int sum_size, index, max_rank = rank;
  int multiplier = 1;
  int number_of_sums = 0, number_of_free_indices = 0;
#ifdef debug
  int largest_index = 0;
#endif
  int* free_indices = NULL;
  long long free_indices_value;
  bool non_zero, free_not_done = true, sum_not_done, skip;
  for (int i = 0; i < rank; i++)                 // compute number of summation and free indices
    if (index_list[i] >= 0) {                    // sum index
      number_of_sums += 1;
#ifdef debug
      if (index_list[i] > largest_index)
	largest_index = index_list[i];
#endif
      }
    else
      number_of_free_indices += 1;
#ifdef debug
  if ((number_of_ntuples > number_of_sums) || 
      (largest_index != (number_of_sums - 1)))
    die("number of sums is inconsistent");
  for (int n = 0; n < number_of_ntuples; n++)
    for (int i = 0; i < ntuple_list[n]->rank; i++)
      if (ntuple_list[n]->index_list[i] > largest_index)
	die("largest index is inconsistent");
#endif
  sum_list_length = 0;
  make_free_index_list(limit_redundant_indices);
  if ((number_of_free_indices > 0) && (free_index_length == 0)) {
#ifdef debug
    to_stderr ("make_free_index_list found no components");
#endif
    return 0;}
  int* sum_skip = new int[rank];
  if (sum_skip == NULL)
    die("allocating sum_skip");
  if (limit_redundant_indices) {
    memset(sum_skip, 0, rank * sizeof(int));
    for (int i = 0; i < rank - 1; i++) {
      if (symmetric[i]) {
	for (int n = 0; n < number_of_ntuples; n++)
	  for (int j = 0; j < (ntuple_list[n]->rank - 1); j++)
	    if ((index_list[i] >= 0) && (index_list[i + 1] >= 0) &&
		(index_list[i] == ntuple_list[n]->index_list[j]) &&
		(index_list[i + 1] == ntuple_list[n]->index_list[j + 1]) &&
		ntuple_list[n]->symmetric[j]) {
	      sum_skip[i] = 1;
	      j += 1;}
	i += 1;}
      if (antisymmetric[i]) {
	for (int n = 0; n < number_of_ntuples; n++)
	  for (int j = 0; j < (ntuple_list[n]->rank - 1); j++)
	    if ((index_list[i] >= 0) && (index_list[i + 1] >= 0) &&
		(index_list[i] == ntuple_list[n]->index_list[j]) &&
		(index_list[i + 1] == ntuple_list[n]->index_list[j + 1]) &&
		ntuple_list[n]->antisymmetric[j]) {
	      sum_skip[i] = -1;
	      multiplier *= 2;
	      j += 1;}
	i += 1;}}}
  int* sum_indices = new int [number_of_sums];
  if (sum_indices == NULL)
    die("allocating sum_indices");
  int* sum_dimension = new int [number_of_sums];
  if (sum_dimension == NULL)
    die("allocating sum_dimension");
  sum_size = number_of_components;               // compute number of products in sum
  if (sum_list != NULL)
    delete [] sum_list;
  for (int n = 0; n < number_of_ntuples; n++) {
    sum_size *= ntuple_list[n]->number_of_components;
    if (ntuple_list[n]->metric)
      sum_size = (int) (sum_size * sum_fudge_factor / dimension);    // adjust sum_size for contractions
    if (ntuple_list[n]->rank > max_rank)            // get max rank while we're at it
      max_rank = ntuple_list[n]->rank;
    if (ntuple_list[n]->sum_list != NULL)
      delete [] ntuple_list[n]->sum_list;}
  int* temp_index_list = new int [max_rank];
  if (temp_index_list == NULL)
    die("allocating temp_index_list");
  sum_list = new component* [sum_size];
  if (sum_list == NULL)
    die("allocating sum_list");
  current_free_index = 0;
  if (indices_list != NULL)
    delete [] indices_list;
  indices_list = new long long [sum_size];
  if (indices_list == NULL)
    die("allocating indices_list");
  for (int n = 0; n < number_of_ntuples; n++) {  // finish number of free indices etc.
    for (int i = 0; i < ntuple_list[n]->rank; i++)
      if (ntuple_list[n]->index_list[i] < 0)
	number_of_free_indices += 1;
      else {
	if (ntuple_list[n]->spinor[i])
	  sum_dimension[ntuple_list[n]->index_list[i]] = spinor_dimension;
	else
	  sum_dimension[ntuple_list[n]->index_list[i]] = dimension;}
    ntuple_list[n]->sum_list = new component* [sum_size];
    if (ntuple_list[n]->sum_list == NULL)
      die("allocating ntuple_list[n]->sum_list");
    ntuple_list[n]->make_free_index_list(limit_redundant_indices);
    ntuple_list[n]->current_free_index = 0;}
  if (number_of_free_indices > 0) {
    free_indices = new int[number_of_free_indices];
    if (free_indices == NULL)
      die("allocating free_indices");}
  while (free_not_done) {                        // free index loop
    memset(sum_indices, 0, number_of_sums * sizeof(int));
    if (number_of_free_indices > 0) {            // compute free indices
      index = 0;
      for (int i = 0; i < rank; i++)             // first do this indices
	if (index_list[i] < 0) {                 // free index
#ifdef debug
	  if ((index < 0) || (index >= number_of_free_indices))
	    die("index out of bounds in make_sum_list (1)");
#endif
	  free_indices[index] = free_index(i);
	  index += 1;}
	else {
	  sum_not_done = true;                  // do metric indices here
	  for (int n = 0; n < number_of_ntuples; n++) {
	    if (ntuple_list[n]->metric)
	      for (int j = 0; j < 2; j++)
		if (ntuple_list[n]->index_list[j] == index_list[i]) {
#ifdef debug
		  if ((index < 0) || (index >= number_of_free_indices))
		    die("index out of bounds in make_sum_list (2)");
#endif
		  free_indices[index] = ntuple_list[n]->free_index((j + 1) % 2);
		  index += 1;
		  sum_not_done = false;
		  break;}
	    if (!sum_not_done)
	      break;}}
      for (int n = 0; n < number_of_ntuples; n++)    // non-metric free indices
	if (!ntuple_list[n]->metric)
	  for (int j = 0; j < ntuple_list[n]->rank; j++)
	    if (ntuple_list[n]->index_list[j] < 0) {
#ifdef debug
	      if ((index < 0) || (index >= number_of_free_indices))
		die("index out of bounds in make_sum_list (3)");
#endif
	      free_indices[index] = ntuple_list[n]->free_index(j);
	      index += 1;}
      free_indices_value = 0;                    // compact the free indices
      for (int i = 0; i < number_of_free_indices; i++)
	free_indices_value |= ((long long) (free_indices[i] & 63) << (6 * i));}
    else
      free_indices_value = -1;
    sum_not_done = true;
    while (sum_not_done) {                       // summation loop
      for (int i = 0; i < rank; i++)
	if (index_list[i] < 0)
	  temp_index_list[i] = free_index(i);
	else
	  temp_index_list[i] = sum_indices[index_list[i]];
      if (limit_redundant_indices) {
	skip = false;
	for (int i = 0; i < (rank - 1); i++)
	  if ((sum_skip[i] < 0) && (temp_index_list[i + 1] <= temp_index_list[i])) {
	    skip = true;
	    break;}
	if (skip)
	  current_component = NULL;
	else
	  current_component = sub(temp_index_list);}
      else
	  current_component = sub(temp_index_list);
      if (current_component != NULL) {
	non_zero = true;
	for (int n = 0; n < number_of_ntuples; n++) {
	  for (int i = 0; i < ntuple_list[n]->rank; i++)
	    if (ntuple_list[n]->index_list[i] < 0)
	      temp_index_list[i] = ntuple_list[n]->free_index(i);
	    else
	      temp_index_list[i] = sum_indices[ntuple_list[n]->index_list[i]];
	  ntuple_list[n]->current_component = ntuple_list[n]->sub(temp_index_list);
	  if (ntuple_list[n]->current_component == NULL)
	    non_zero = false;}
	if (non_zero) {
#ifdef debug
	  if (sum_list_length >= sum_size)
	    die("sum list length out of bounds");
#endif
	  sum_list[sum_list_length] = current_component;
	  for (int n = 0; n < number_of_ntuples; n++)
	    ntuple_list[n]->sum_list[sum_list_length] = ntuple_list[n]->current_component;
	  indices_list[sum_list_length] = free_indices_value;
	  sum_list_length += 1; }}                   // next product in the sum
      for (int i = number_of_sums - 1; i >= 0; i--) {
	sum_indices[i] += 1;                         // increment summation index
	if (sum_indices[i] >= sum_dimension[i]) {
	  sum_indices[i] = 0;                        // next summation index
	  if (i == 0)                                // unless we're done
	    sum_not_done = false;}
	else
	  break;}}                                   // next free index in the sum
    for (int n = number_of_ntuples - 1; n >= 0; n--) {
      ntuple_list[n]->current_free_index += 1;
      if (ntuple_list[n]->current_free_index >= ntuple_list[n]->free_index_length) {
	ntuple_list[n]->current_free_index = 0;
	if (n == 0) {
	  current_free_index += 1;
	  if (current_free_index >= free_index_length)
	    free_not_done = false;
	  break;}}
      else
	break;}}
  for (int n = 0; n < number_of_ntuples; n++)
    ntuple_list[n]->sum_list_length = sum_list_length;
  delete [] sum_skip;
  delete [] sum_indices;
  delete [] sum_dimension;
  if (number_of_free_indices > 0)
    delete [] free_indices;
  delete [] temp_index_list;
  return multiplier; }
