//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// simplify.cc

#include <ptah.h>
#include <element.h>
#include <hash.h>
#include <ntuple.h>
 
extern void sum::simplify () {
#ifdef tracentry
  trace_ss += 1;
#endif
  int inv_no = 0;
  int number_to_multiply, fn, fnind, save_power;
  bool simplified, gcd_not_one = false;
  long long minus;
  long long one = 1;
  long long new_coeff;
  term *current_term;
  term *current_function_term;
  invariant_term *current_inv_term;
  sum *gcd = NULL;
  sum **this_list = new sum* [1];
  if (number_of_components < 10)
    return;
  sum *this_save = new sum;
  this_list[0] = this;
  this_save->multiply (1, this_list, &one);
  number_to_simplify = 0;
  for (int i = 0; i < number_of_passes; i++)
    number_to_simplify += pass_num_fns[i];
  gcd_not_one = false;
  sum* temp_sum = new sum;
  if (temp_sum == NULL)
    die("allocating temp_sum");
  sum* temp_sum_2 = new sum;
  if (temp_sum_2 == NULL)
    die("allocating temp_sum_2");
  sum** multiply_list = new sum* [number_to_simplify * simplify_size];
  if (multiply_list == NULL)
    die("allocating multiply_list");
  invariant *temp_invariant[2];
  temp_invariant[0] = new invariant;
  if (temp_invariant[0] == NULL)
    die("allocating temp_invariant[0]");
  temp_invariant[1] = new invariant;
  if (temp_invariant[1] == NULL)
    die("allocating temp_invariant[1]");
  memset(&gcd_factor_powers, 0, number_of_variables * sizeof(int));
  for (int i = 0; i < size; i++) {                                     // find gcd
    current_term = first_term(i);
    while (current_term != NULL) {
      for (int j = 0; j < number_of_variables - number_of_fractions; j++)
	if (current_term->factor_powers[j] < gcd_factor_powers[j]) {
	  gcd_not_one = true;
	  gcd_factor_powers[j] = current_term->factor_powers[j];}
      current_term = current_term->next_term();}}
  multiply_list[0] = (sum*) this;
  if (gcd_not_one) {                                                   // multiply through by gcd
    for (int j = 0; j < number_of_variables - number_of_fractions; j++)
      gcd_factor_powers[j] = - gcd_factor_powers[j];
    gcd = new sum;
    if (gcd == NULL)
      die("allocating gcd");
    gcd->add(&one, gcd_factor_powers);
    multiply_list[1] = gcd;
    temp_sum_2->multiply(2, multiply_list, &one);}
  else
    temp_sum_2->multiply(1, multiply_list, &one);
  for (int i = 0; i < temp_sum_2->size; i++) {
    current_term = temp_sum_2->first_term(i);
    while (current_term != NULL) {
      number_to_multiply = 1;
      multiply_list[0] = temp_sum;
      memcpy(&factor_powers_temp_s, current_term->factor_powers, number_of_variables * sizeof(int));
      for (int fnind = 0; fnind < number_of_functions; fnind++) {     // expand all functions
	fn = fnind + dimension;
	if (current_term->factor_powers[fn] > 0) {
	  factor_powers_temp_s[fn] = current_term->factor_powers[fn] % function_power[fnind];
	  for (int j = 0; j < (current_term->factor_powers[fn] / function_power[fnind]); j++) {
	    multiply_list[number_to_multiply] = functions[fnind];
	    number_to_multiply += 1;}}}
      temp_sum->zero();
      temp_sum->add(&(current_term->coefficient), factor_powers_temp_s);
      temp_invariant[0]->multiply(number_to_multiply, multiply_list, &one);
      current_term = current_term->next_term();}}
#ifdef debug
#ifdef keepstats
  memcpy(&keep_head_count, temp_invariant[0]->headcount, large_hash_size * sizeof(int));
#endif
#endif
  for (int fpind = 0; fpind < number_to_simplify; fpind++) {     // for all functions
    fnind = pass_fns[fpind];;
#ifdef debug
    if ((sort_variable[fnind] < 0) || (sort_variable[fnind] >= number_of_variables))
      die("simplify index out of bounds");
#endif
    fn = fnind + dimension;
    simplified = true;
    while (simplified) {                  // repeat until all factors of function are out
      simplified = false;
      for (int p = (simplify_size - 1); p >= function_terms[fnind]->factor_powers[sort_variable[fnind]]; p--) {
	// for highest to lowest powers of sort var
	current_inv_term = temp_invariant[inv_no]->first_invariant(fpind, p);
	while (current_inv_term != NULL) {                  //   while not done with terms with that power
	  memcpy(&factor_powers_temp_s, current_inv_term->factor_powers, number_of_variables * sizeof(int));
	  factor_powers_temp_s[sort_variable[fnind]] -= function_terms[fnind]->factor_powers[sort_variable[fnind]];
	  save_power = factor_powers_temp_s[sort_variable[fnind]];  // divide out leading term
	  new_coeff = current_inv_term->coefficient / function_terms[fnind]->coefficient;
	  temp_invariant[(inv_no + 1) % 2]->add(&new_coeff, factor_powers_temp_s);
	  current_function_term = function_terms[fnind]->next_term(); // start subtracting from 2nd term
	  while (current_function_term != NULL) {
	    for (int k = 0; k < number_of_variables; k++)
	      factor_powers_temp_s[k] += current_function_term->factor_powers[k];
	    minus = - new_coeff * current_function_term->coefficient;
	    temp_invariant[inv_no]->add(&minus, factor_powers_temp_s);
	    memcpy(&factor_powers_temp_s, current_inv_term->factor_powers, number_of_variables * sizeof(int));
	    factor_powers_temp_s[sort_variable[fnind]] = save_power;
	    current_function_term = current_function_term->next_term();}
	  minus = - current_inv_term->coefficient;
	  memcpy(&factor_powers_temp_s, current_inv_term->factor_powers, number_of_variables * sizeof(int));
	  temp_invariant[inv_no]->add(&minus, factor_powers_temp_s);
	  simplified = true;
	  current_inv_term = temp_invariant[inv_no]->first_invariant(fpind, p);}} // start at top of this list
      if (simplified) {
	if (temp_invariant[inv_no]->number_of_components == 0) {
	  gcd_factor_powers[fn] -= function_power[fnind];  //       factor out the function
	  gcd_not_one = true;
	  inv_no = (inv_no + 1) % 2;}                      //       and try again
	else {
	  multiply_list[0] = temp_invariant[(inv_no + 1) % 2]; //   restore dividend and try new function
	  multiply_list[1] = functions[fnind];
	  temp_invariant[inv_no]->multiply(2, multiply_list, &one);
	  temp_invariant[(inv_no + 1) % 2]->zero();
	  simplified = false;}}}}
  zero();                                                              // zero this invariant
  multiply_list[0] = (sum*) temp_invariant[inv_no];
  if (gcd_not_one) {                                       // divide through by gcd
    if (gcd != NULL) {
      gcd->zero();
      for (int j = 0; j < number_of_variables - number_of_fractions; j++)
	gcd_factor_powers[j] = - gcd_factor_powers[j];}
    else {
      gcd = new sum;
      if (gcd == NULL)
	die("allocating gcd");}
    gcd->add(&one, gcd_factor_powers);
    multiply_list[1] = gcd;
    multiply(2, multiply_list, &one);}
  else
    multiply(1, multiply_list, &one);
  delete [] multiply_list;
  if (gcd != NULL) {
    gcd->zero();
    delete gcd;
    gcd = NULL;}
  for (int i = 0; i < 2; i++)
    if (temp_invariant[i] != NULL) {
      temp_invariant[i]->zero();
      delete temp_invariant[i];}
  temp_sum->zero();
  delete temp_sum;
  temp_sum_2->zero();
  delete temp_sum_2;
  if (number_of_components > this_save->number_of_components) {
    this->zero ();
    this_list[0] = this_save;
    this->multiply (1, this_list, &one);}
  this_save->zero ();
  delete this_save; }

extern void sum::simplify_invariant () {
#ifdef tracentry
  trace_is += 1;
#endif
#ifdef keepstats
  int starting_count = number_of_components;
  char str [256];
#endif
  if (number_of_components < 10)
    return;
#ifdef keepstats
  time_stamp("Before simplifying");
  sprintf(str, "Starting with %d terms.", starting_count);
  to_stderr (str);
#endif
  simplify();
#ifdef keepstats
#ifdef debug
  head_stats();
#endif
  time_stamp("Simplification complete");
  sprintf(str, "Simplified to %d terms.", number_of_components);
  to_stderr (str);
#endif
 }

extern void ntuple::simplify() {
#ifdef tracentry
  trace_sn += 1;
#endif
  int inds[rank];
  component *next;
  for (int i = 0; i < regular_hash_size; i++) {
    current_component = first_component(i);
    while (current_component != NULL) {
      if (current_component->value != NULL)
	(current_component->value)->simplify();
      if ((current_component->value)->zeroq()) {
	for (int j = 0; j < rank; j++)
	  inds[j] = (current_component->indices >> (6 * j)) & 63;
	if (rank <= 6)
	  comparray[ca_index(inds)] = NULL;
	next = current_component->next_component();
	set_hash(i);
	remove_component(current_component);
	current_component = next;
	continue;}
      current_component = current_component->next_component();}} }
