//
// PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 Kenneth R. Koehler 
//
// e-mail address - kenneth.koehler@uc.edu
//
// This program is free software; you can redistribute it and/or modify it under the terms of the
// GNU General Public License as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version. 
//
// This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU General Public License for more details. 
//
// You should have received a copy of the GNU General Public License along with this program; if
// not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
//
// sum.cc

#include <ptah.h>
#include <element.h>
#include <hash.h>

extern sum::sum () {
#ifdef tracentry
  trace_s += 1;
#endif
#ifdef debug
  hash_magic_no = hash_magic;
  sum_magic_no = sum_magic;
#endif
  head = new element* [regular_hash_size];
  if (head == NULL)
    die ("allocating head");
#ifdef keepstats
#ifdef debug
  headcount = new int [regular_hash_size];
  if (headcount == NULL)
    die ("allocating headcount");
  memset (headcount, 0, regular_hash_size * sizeof (int));
#endif
#endif
  size = regular_hash_size;
  rank = number_of_variables - number_of_fractions;
  number_of_components = 0;
  last_hash = -1;
  for (int i = 0; i < size; i++)
    head[i] = NULL; }

extern sum::sum (int hash_size) {
#ifdef tracentry
  trace_s += 1;
#endif
#ifdef debug
  hash_magic_no = hash_magic;
  sum_magic_no = sum_magic;
#endif
  head = new element* [hash_size];
  if (head == NULL)
    die ("allocating head");
#ifdef keepstats
#ifdef debug
  headcount = new int [hash_size];
  if (headcount == NULL)
    die ("allocating headcount");
  memset (headcount, 0, hash_size * sizeof (int));
#endif
#endif
  size = hash_size;
  rank = number_of_variables - number_of_fractions;
  number_of_components = 0;
  last_hash = -1;
  for (int i = 0; i < size; i++)
    head[i] = NULL; }

extern sum::~sum () {
#ifdef tracentry
  trace_ks += 1;
#endif
#ifdef debug
  for (int i = 0; i < size; i++)
    if (head[i] != NULL)
      die ("attempting to delete non-empty sum");
  hash_magic_no = -hash_magic;
  sum_magic_no = -sum_magic;
#endif
#ifdef keepstats
#ifdef debug
  delete [] headcount;
#endif
#endif
  delete [] head; }

extern void sum::sum_build (int numt, int *numfct, int *fpwr, long long *co) {
  int fctoffset = 0;
  for (int j = 0; j < numt; j++) {                 // term loop
    set_powers (numfct[j], &fpwr[fctoffset * 2]);
    fctoffset += numfct[j];}
  add (&co[0], pwr); }

extern void sum::zero () {
#ifdef tracentry
  trace_sz += 1;
#endif
  term *current_term;
  for (int i = 0; i < size; i++)
    {current_term = first_term (i);
    while (current_term != NULL)
      {set_hash (i);
      remove_term (current_term);
      current_term = first_term (i);}}        // start over since we just removed the head
}

extern void sum::multiply (int number_of_sums, sum** sum_list, long long* overall_coefficient) {
#ifdef tracentry
  trace_sumu += 1;
#endif
  int* head_index = new int [number_of_sums];
  if (head_index == NULL)
    die ("allocating head_index");
  int* first_head_index = new int [number_of_sums];
  if (first_head_index == NULL)
    die ("allocating first_head_index");
  term** current_term = new term* [number_of_sums];
  if (current_term == NULL)
    die ("allocating current_term");
  term** first_current_term = new term* [number_of_sums];
  if (first_current_term == NULL)
    die ("allocating first_current_term");
  long long product_of_coefficients;
  bool nonzero, not_done = true;
  for (int i = 0; i < number_of_sums; i++) {
#ifdef debug
    if (sum_list[i]->sum_magic_no != sum_magic)
      die ("attempting to multiply a non-sum");
#endif
    head_index[i] = 0;                                    // find first term in sum
    current_term[i] = sum_list[i]->first_term (0);
    while ((head_index[i] < (sum_list[i]->size - 1)) && (current_term[i] == NULL))
      {head_index[i] += 1;
      current_term[i] = sum_list[i]->first_term (head_index[i]);}
    if (current_term[i] == NULL)                          // one sum is zero, so product is also
      {not_done = false;
      break;}
    first_head_index[i] = head_index[i];                  // save for next round
    first_current_term[i] = current_term[i];}
  while (not_done)
    {memset (&factor_powers_temp, 0, number_of_variables * sizeof (int));
    product_of_coefficients = *overall_coefficient;
    nonzero = false;
    for (int i = 0; i < number_of_sums; i++)              // multiply all current terms
      if (current_term[i] != NULL)
	{product_of_coefficients *= current_term[i]->coefficient;
	nonzero = true;
	for (int j = 0; j < number_of_variables; j++)
	  factor_powers_temp[j] += current_term[i]->factor_powers[j];}
    if (nonzero)
      add (&product_of_coefficients, factor_powers_temp);
    for (int i = 0; i < number_of_sums; i++)              // check all sums
      {current_term[i] = current_term[i]->next_term ();    // next term in sum
      while ((head_index[i] < (sum_list[i]->size - 1)) && (current_term[i] == NULL))
	{head_index[i] += 1;
	current_term[i] = sum_list[i]->first_term (head_index[i]);}
      if (current_term[i] != NULL)                        // not through all terms in this sum
	break;
      head_index[i] = first_head_index[i];                // start this sum over
      current_term[i] = first_current_term[i];            // and check the next
      if (i >= (number_of_sums - 1))                      // are we done?
	not_done = false;}}
  delete [] current_term;
  delete [] first_current_term;
  delete [] head_index;
  delete [] first_head_index; }

extern void sum::derivative (sum* function, int variable, long long* overall_coefficient) {
#ifdef tracentry
  trace_sd += 1;
#endif
  term *current_term;
  sum chain_rule;
  sum* chain_sums [2];
  int power;
  long long product;
  long long one = 1;
#ifdef debug
  if (function->sum_magic_no != sum_magic)
    die ("attempting to take the derivative of a non-sum");
#endif
  for (int i = 0; i < function->size; i++)                 // look at every list of terms
    {current_term = function->first_term (i);
    while (current_term != NULL) {                         // look at every term in this list
#ifdef debug
      if (current_term->term_magic_no != term_magic)
	die ("attempting to take the derivative of a non-term");
#endif
      for (int var = 0; var < (number_of_variables - number_of_constants -
			       number_of_fractions); var++)  // look at every variable
	if (current_term->factor_powers[var] != 0)         // does this term depend on this variable?
	  {if (var < dimension)                            // is it a coordinate?
	    {if (var == variable)                          // is it this coordinate?
	      {memcpy (&factor_powers_temp_d, current_term->factor_powers, number_of_variables * sizeof (int));
	      power = factor_powers_temp_d[var];           // dx^n/dx = n x^(n-1)
	      factor_powers_temp_d[var] -= 1;
	      product = current_term->coefficient * *overall_coefficient;
	      product *= power;
	      add (&product, factor_powers_temp_d);}}
	  else                                            // must be a function
	    {if (argument[(var - dimension) * dimension + variable])
		                                          // does this function depend on this variable?
	      {memcpy (&factor_powers_temp_d, current_term->factor_powers, number_of_variables * sizeof (int));
	      power = factor_powers_temp_d[var];          // df^n/dx = n f^(n-1) df/dx
	      factor_powers_temp_d[var] -= 1;
	      product = current_term->coefficient * *overall_coefficient;
	      product *= power;
	      chain_rule.add (&product, factor_powers_temp_d);
	      chain_sums[0] = &chain_rule;
	      chain_sums[1] = derivatives[(var - dimension) * dimension + variable];
	      multiply (2, chain_sums, &one);
	      chain_rule.zero ();}}}
      current_term = current_term->next_term ();}} }

extern void sum::display (char* name) {
  int len;
  char str[255];
  term *current_term;
  sprintf (str, "%s = Plus[", name);
  to_stdout (str);
  for (int i = 0; i < size; i++) {
    current_term = first_term (i);
    while (current_term != NULL) {
      len = sprintf (str, "Times[ %lld,", current_term->coefficient);
      for (int k = 0; k < number_of_variables; k++)
	if (current_term->factor_powers[k] != 0)
	  len += sprintf (str + len, "Power[ %c, %d],", variables[k], current_term->factor_powers[k]);
      sprintf (str + len, "1],");
      to_stdout (str);
      current_term = current_term->next_term ();}}
  sprintf (str, "0];");
  to_stdout (str); }

extern void sum::display_add (char* name) {
  int len;
  char str[255];
  term *current_term;
  sprintf (str, "%s += Plus[", name);
  to_stdout (str);
  for (int i = 0; i < size; i++) {
    current_term = first_term (i);
    while (current_term != NULL) {
      len = sprintf (str, "Times[ %lld,", current_term->coefficient);
      for (int k = 0; k < number_of_variables; k++)
	if (current_term->factor_powers[k] != 0)
	  len += sprintf (str + len, "Power[ %c, %d],", variables[k], current_term->factor_powers[k]);
      sprintf (str + len, "1],");
      to_stdout (str);
      current_term = current_term->next_term ();}}
  sprintf (str, "0];");
  to_stdout (str); }

#ifdef debug
extern void sum::dump (char* name) {
  char s[24];
  term *current_term;
  if (!parent)
    return;
  cerr << "     " << name << " = Sum of";
  for (int k = 0; k < size; k++) {
    current_term = first_term (k);
    while (current_term != NULL) {
      sprintf (s, ", %lld", current_term->coefficient);
      cerr << s << " ";
      for (int l = 0; l < number_of_variables; l++)
	if(current_term->factor_powers[l] != 0) 
	  {cerr << variables[l] << "^" << current_term->factor_powers[l] << " ";}
      current_term = current_term->next_term ();}}
  cerr << " \n";}
#endif
