(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     19036,        481]*)
(*NotebookOutlinePosition[     20026,        513]*)
(*  CellTagsIndexPosition[     19982,        509]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/ccif.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,<oldx,>func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("AVOID USING C, D, E, I, N and O as variables !"\)], "Print"]
}, Open  ]],

Cell[BoxData[
    \(\(ptahdir\  = \ "\<~/projects/ptah/\>";\)\)], "Input"],

Cell[BoxData[{
    \(\(ls\  = \ {{a, b, c, d, e, f, g, h, i, j}, \[IndentingNewLine]{k, l, 
            m, n, o, p, q, r, s, t}, \[IndentingNewLine]{u, v, w, x, y, z, A, 
            B, C, D}, \[IndentingNewLine]{F, G, H, J, K, L, M, N, O, 
            P}, \[IndentingNewLine]{Q, R, S, T, U, V, W, X, Y, 
            Z}};\)\), "\[IndentingNewLine]", 
    \(\(xx = ls[\([1]\)];\)\n\), "\n", 
    \(\(gg\  = \ IdentityMatrix[10];\)\)}], "Input"],

Cell[BoxData[
    \(\(spacetime[xx, gg];\)\)], "Input"],

Cell[BoxData[
    \(\(gp\  = \ gg;\)\)], "Input"],

Cell[BoxData[
    \(\(ginvp\  = \ simpler[Inverse[gp]];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(irmdgp\  = \ 
      Simplify[PowerExpand[1/Sqrt[Simplify[Det[gp]]]]]\)], "Input"],

Cell[BoxData[
    \(1\)], "Output"]
}, Open  ]],

Cell[BoxData[
    \(\(ast1\  = \ 
        Table[Table[ls[\([1, ii]\)] - ls[\([1, jj]\)], {jj, 10}], {ii, 
            10}];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TableForm[ast1]\)], "Input"],

Cell[BoxData[
    TagBox[GridBox[{
          {
            "0", \(a - b\), \(a - c\), \(a - d\), \(a - e\), \(a - f\), \(a - 
              g\), \(a - h\), \(a - i\), \(a - j\)},
          {\(\(-a\) + b\), 
            "0", \(b - c\), \(b - d\), \(b - e\), \(b - f\), \(b - g\), \(b - 
              h\), \(b - i\), \(b - j\)},
          {\(\(-a\) + c\), \(\(-b\) + c\), 
            "0", \(c - d\), \(c - e\), \(c - f\), \(c - g\), \(c - h\), \(c - 
              i\), \(c - j\)},
          {\(\(-a\) + d\), \(\(-b\) + d\), \(\(-c\) + d\), 
            "0", \(d - e\), \(d - f\), \(d - g\), \(d - h\), \(d - i\), \(d - 
              j\)},
          {\(\(-a\) + e\), \(\(-b\) + e\), \(\(-c\) + e\), \(\(-d\) + e\), 
            "0", \(e - f\), \(e - g\), \(e - h\), \(e - i\), \(e - j\)},
          {\(\(-a\) + f\), \(\(-b\) + f\), \(\(-c\) + f\), \(\(-d\) + 
              f\), \(\(-e\) + f\), 
            "0", \(f - g\), \(f - h\), \(f - i\), \(f - j\)},
          {\(\(-a\) + g\), \(\(-b\) + g\), \(\(-c\) + g\), \(\(-d\) + 
              g\), \(\(-e\) + g\), \(\(-f\) + g\), 
            "0", \(g - h\), \(g - i\), \(g - j\)},
          {\(\(-a\) + h\), \(\(-b\) + h\), \(\(-c\) + h\), \(\(-d\) + 
              h\), \(\(-e\) + h\), \(\(-f\) + h\), \(\(-g\) + h\), 
            "0", \(h - i\), \(h - j\)},
          {\(\(-a\) + i\), \(\(-b\) + i\), \(\(-c\) + i\), \(\(-d\) + 
              i\), \(\(-e\) + i\), \(\(-f\) + i\), \(\(-g\) + i\), \(\(-h\) + 
              i\), "0", \(i - j\)},
          {\(\(-a\) + j\), \(\(-b\) + j\), \(\(-c\) + j\), \(\(-d\) + 
              j\), \(\(-e\) + j\), \(\(-f\) + j\), \(\(-g\) + j\), \(\(-h\) + 
              j\), \(\(-i\) + j\), "0"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      (TableForm[ #]&)]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(\(ast2\  = \ 
        Table[Table[ls[\([2, ii]\)] - ls[\([2, jj]\)], {jj, 10}], {ii, 
            10}];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TableForm[ast2]\)], "Input"],

Cell[BoxData[
    TagBox[GridBox[{
          {
            "0", \(k - l\), \(k - m\), \(k - n\), \(k - o\), \(k - p\), \(k - 
              q\), \(k - r\), \(k - s\), \(k - t\)},
          {\(\(-k\) + l\), 
            "0", \(l - m\), \(l - n\), \(l - o\), \(l - p\), \(l - q\), \(l - 
              r\), \(l - s\), \(l - t\)},
          {\(\(-k\) + m\), \(\(-l\) + m\), 
            "0", \(m - n\), \(m - o\), \(m - p\), \(m - q\), \(m - r\), \(m - 
              s\), \(m - t\)},
          {\(\(-k\) + n\), \(\(-l\) + n\), \(\(-m\) + n\), 
            "0", \(n - o\), \(n - p\), \(n - q\), \(n - r\), \(n - s\), \(n - 
              t\)},
          {\(\(-k\) + o\), \(\(-l\) + o\), \(\(-m\) + o\), \(\(-n\) + o\), 
            "0", \(o - p\), \(o - q\), \(o - r\), \(o - s\), \(o - t\)},
          {\(\(-k\) + p\), \(\(-l\) + p\), \(\(-m\) + p\), \(\(-n\) + 
              p\), \(\(-o\) + p\), 
            "0", \(p - q\), \(p - r\), \(p - s\), \(p - t\)},
          {\(\(-k\) + q\), \(\(-l\) + q\), \(\(-m\) + q\), \(\(-n\) + 
              q\), \(\(-o\) + q\), \(\(-p\) + q\), 
            "0", \(q - r\), \(q - s\), \(q - t\)},
          {\(\(-k\) + r\), \(\(-l\) + r\), \(\(-m\) + r\), \(\(-n\) + 
              r\), \(\(-o\) + r\), \(\(-p\) + r\), \(\(-q\) + r\), 
            "0", \(r - s\), \(r - t\)},
          {\(\(-k\) + s\), \(\(-l\) + s\), \(\(-m\) + s\), \(\(-n\) + 
              s\), \(\(-o\) + s\), \(\(-p\) + s\), \(\(-q\) + s\), \(\(-r\) + 
              s\), "0", \(s - t\)},
          {\(\(-k\) + t\), \(\(-l\) + t\), \(\(-m\) + t\), \(\(-n\) + 
              t\), \(\(-o\) + t\), \(\(-p\) + t\), \(\(-q\) + t\), \(\(-r\) + 
              t\), \(\(-s\) + t\), "0"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      (TableForm[ #]&)]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(\(ast3\  = \ 
        Table[Table[ls[\([3, ii]\)] - ls[\([3, jj]\)], {jj, 10}], {ii, 
            10}];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TableForm[ast3]\)], "Input"],

Cell[BoxData[
    TagBox[GridBox[{
          {
            "0", \(u - v\), \(u - w\), \(u - x\), \(u - y\), \(u - 
              z\), \(\(-A\) + u\), \(\(-B\) + u\), \(\(-C\) + u\), \(\(-D\) + 
              u\)},
          {\(\(-u\) + v\), 
            "0", \(v - w\), \(v - x\), \(v - y\), \(v - z\), \(\(-A\) + 
              v\), \(\(-B\) + v\), \(\(-C\) + v\), \(\(-D\) + v\)},
          {\(\(-u\) + w\), \(\(-v\) + w\), 
            "0", \(w - x\), \(w - y\), \(w - z\), \(\(-A\) + w\), \(\(-B\) + 
              w\), \(\(-C\) + w\), \(\(-D\) + w\)},
          {\(\(-u\) + x\), \(\(-v\) + x\), \(\(-w\) + x\), 
            "0", \(x - y\), \(x - z\), \(\(-A\) + x\), \(\(-B\) + 
              x\), \(\(-C\) + x\), \(\(-D\) + x\)},
          {\(\(-u\) + y\), \(\(-v\) + y\), \(\(-w\) + y\), \(\(-x\) + y\), 
            "0", \(y - z\), \(\(-A\) + y\), \(\(-B\) + y\), \(\(-C\) + 
              y\), \(\(-D\) + y\)},
          {\(\(-u\) + z\), \(\(-v\) + z\), \(\(-w\) + z\), \(\(-x\) + 
              z\), \(\(-y\) + z\), 
            "0", \(\(-A\) + z\), \(\(-B\) + z\), \(\(-C\) + z\), \(\(-D\) + 
              z\)},
          {\(A - u\), \(A - v\), \(A - w\), \(A - x\), \(A - y\), \(A - z\), 
            "0", \(A - B\), \(A - C\), \(A - D\)},
          {\(B - u\), \(B - v\), \(B - w\), \(B - x\), \(B - y\), \(B - 
              z\), \(\(-A\) + B\), "0", \(B - C\), \(B - D\)},
          {\(C - u\), \(C - v\), \(C - w\), \(C - x\), \(C - y\), \(C - 
              z\), \(\(-A\) + C\), \(\(-B\) + C\), "0", \(C - D\)},
          {\(D - u\), \(D - v\), \(D - w\), \(D - x\), \(D - y\), \(D - 
              z\), \(\(-A\) + D\), \(\(-B\) + D\), \(\(-C\) + D\), "0"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      (TableForm[ #]&)]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(\(ast4\  = \ 
        Table[Table[ls[\([4, ii]\)] - ls[\([4, jj]\)], {jj, 10}], {ii, 
            10}];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TableForm[ast4]\)], "Input"],

Cell[BoxData[
    TagBox[GridBox[{
          {
            "0", \(F - G\), \(F - H\), \(F - J\), \(F - K\), \(F - L\), \(F - 
              M\), \(F - N\), \(F - O\), \(F - P\)},
          {\(\(-F\) + G\), 
            "0", \(G - H\), \(G - J\), \(G - K\), \(G - L\), \(G - M\), \(G - 
              N\), \(G - O\), \(G - P\)},
          {\(\(-F\) + H\), \(\(-G\) + H\), 
            "0", \(H - J\), \(H - K\), \(H - L\), \(H - M\), \(H - N\), \(H - 
              O\), \(H - P\)},
          {\(\(-F\) + J\), \(\(-G\) + J\), \(\(-H\) + J\), 
            "0", \(J - K\), \(J - L\), \(J - M\), \(J - N\), \(J - O\), \(J - 
              P\)},
          {\(\(-F\) + K\), \(\(-G\) + K\), \(\(-H\) + K\), \(\(-J\) + K\), 
            "0", \(K - L\), \(K - M\), \(K - N\), \(K - O\), \(K - P\)},
          {\(\(-F\) + L\), \(\(-G\) + L\), \(\(-H\) + L\), \(\(-J\) + 
              L\), \(\(-K\) + L\), 
            "0", \(L - M\), \(L - N\), \(L - O\), \(L - P\)},
          {\(\(-F\) + M\), \(\(-G\) + M\), \(\(-H\) + M\), \(\(-J\) + 
              M\), \(\(-K\) + M\), \(\(-L\) + M\), 
            "0", \(M - N\), \(M - O\), \(M - P\)},
          {\(\(-F\) + N\), \(\(-G\) + N\), \(\(-H\) + N\), \(\(-J\) + 
              N\), \(\(-K\) + N\), \(\(-L\) + N\), \(\(-M\) + N\), 
            "0", \(N - O\), \(N - P\)},
          {\(\(-F\) + O\), \(\(-G\) + O\), \(\(-H\) + O\), \(\(-J\) + 
              O\), \(\(-K\) + O\), \(\(-L\) + O\), \(\(-M\) + O\), \(\(-N\) + 
              O\), "0", \(O - P\)},
          {\(\(-F\) + P\), \(\(-G\) + P\), \(\(-H\) + P\), \(\(-J\) + 
              P\), \(\(-K\) + P\), \(\(-L\) + P\), \(\(-M\) + P\), \(\(-N\) + 
              P\), \(\(-O\) + P\), "0"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      (TableForm[ #]&)]], "Output"]
}, Open  ]],

Cell[BoxData[
    \(\(ast5\  = \ 
        Table[Table[ls[\([5, ii]\)] - ls[\([5, jj]\)], {jj, 10}], {ii, 
            10}];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(TableForm[ast5]\)], "Input"],

Cell[BoxData[
    TagBox[GridBox[{
          {
            "0", \(Q - R\), \(Q - S\), \(Q - T\), \(Q - U\), \(Q - V\), \(Q - 
              W\), \(Q - X\), \(Q - Y\), \(Q - Z\)},
          {\(\(-Q\) + R\), 
            "0", \(R - S\), \(R - T\), \(R - U\), \(R - V\), \(R - W\), \(R - 
              X\), \(R - Y\), \(R - Z\)},
          {\(\(-Q\) + S\), \(\(-R\) + S\), 
            "0", \(S - T\), \(S - U\), \(S - V\), \(S - W\), \(S - X\), \(S - 
              Y\), \(S - Z\)},
          {\(\(-Q\) + T\), \(\(-R\) + T\), \(\(-S\) + T\), 
            "0", \(T - U\), \(T - V\), \(T - W\), \(T - X\), \(T - Y\), \(T - 
              Z\)},
          {\(\(-Q\) + U\), \(\(-R\) + U\), \(\(-S\) + U\), \(\(-T\) + U\), 
            "0", \(U - V\), \(U - W\), \(U - X\), \(U - Y\), \(U - Z\)},
          {\(\(-Q\) + V\), \(\(-R\) + V\), \(\(-S\) + V\), \(\(-T\) + 
              V\), \(\(-U\) + V\), 
            "0", \(V - W\), \(V - X\), \(V - Y\), \(V - Z\)},
          {\(\(-Q\) + W\), \(\(-R\) + W\), \(\(-S\) + W\), \(\(-T\) + 
              W\), \(\(-U\) + W\), \(\(-V\) + W\), 
            "0", \(W - X\), \(W - Y\), \(W - Z\)},
          {\(\(-Q\) + X\), \(\(-R\) + X\), \(\(-S\) + X\), \(\(-T\) + 
              X\), \(\(-U\) + X\), \(\(-V\) + X\), \(\(-W\) + X\), 
            "0", \(X - Y\), \(X - Z\)},
          {\(\(-Q\) + Y\), \(\(-R\) + Y\), \(\(-S\) + Y\), \(\(-T\) + 
              Y\), \(\(-U\) + Y\), \(\(-V\) + Y\), \(\(-W\) + Y\), \(\(-X\) + 
              Y\), "0", \(Y - Z\)},
          {\(\(-Q\) + Z\), \(\(-R\) + Z\), \(\(-S\) + Z\), \(\(-T\) + 
              Z\), \(\(-U\) + Z\), \(\(-V\) + Z\), \(\(-W\) + Z\), \(\(-X\) + 
              Z\), \(\(-Y\) + Z\), "0"}
          },
        RowSpacings->1,
        ColumnSpacings->3,
        RowAlignments->Baseline,
        ColumnAlignments->{Left}],
      (TableForm[ #]&)]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(dim = 10;\)\), "\[IndentingNewLine]", 
    \(coords\  = \ xx\), "\[IndentingNewLine]", 
    \(ints\  = \ {2, 3}\), "\n", 
    \(constants\  = \ Flatten[Table[ls[\([ii]\)], {ii, 2, 5}]]\), "\n", 
    \(functions\  = \ {}\), "\[IndentingNewLine]", 
    \(ansatze\  = \ {}\)}], "Input"],

Cell[BoxData[
    \({a, b, c, d, e, f, g, h, i, j}\)], "Output"],

Cell[BoxData[
    \({2, 3}\)], "Output"],

Cell[BoxData[
    \({k, l, m, n, o, p, q, r, s, t, u, v, w, x, y, z, A, B, C, D, F, G, H, 
      J, K, L, M, N, O, P, Q, R, S, T, U, V, W, X, Y, Z}\)], "Output"],

Cell[BoxData[
    \({}\)], "Output"],

Cell[BoxData[
    \({}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(headers[];\)\)], "Input"],

Cell[BoxData[
    \(Part::"partw" \(\(:\)\(\ \)\) 
      "Part \!\(1\) of \!\({}\) does not exist."\)], "Message"],

Cell[BoxData[
    \("functions completed..."\)], "Print"],

Cell[BoxData[
    \("g completed..."\)], "Print"],

Cell[BoxData[
    \("gi completed..."\)], "Print"],

Cell[BoxData[
    \("d completed..."\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(tensor[ast1, "\<ast1\>"];\)\), "\[IndentingNewLine]", 
    \(\(tensor[ast2, "\<ast2\>"];\)\), "\[IndentingNewLine]", 
    \(\(tensor[ast3, "\<ast3\>"];\)\), "\[IndentingNewLine]", 
    \(\(tensor[ast4, "\<ast4\>"];\)\), "\[IndentingNewLine]", 
    \(\(tensor[ast5, "\<ast5\>"];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("ast1"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ast1", " completed..."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("ast2"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ast2", " completed..."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("ast3"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ast3", " completed..."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("ast4"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ast4", " completed..."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("ast5"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ast5", " completed..."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(backup[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]45\[InvisibleSpace]":"\[InvisibleSpace]27\[InvisibleSpace]" \
on "\[InvisibleSpace]1\[InvisibleSpace]"/"\[InvisibleSpace]19\[InvisibleSpace]\
"/"\[InvisibleSpace]2003\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 45, ":", 27, " on ", 1, "/", 19, "/", 2003, "."],
      Editable->False]], "Print"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "ken", \
"projects", "mathematica", "packages"}, "d11brane.nb.ps", CharacterEncoding \
-> "ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 70, 1, 27, "Input"],
Cell[1874, 57, 2984, 44, 375, "Print"],
Cell[4861, 103, 81, 1, 23, "Print"]
}, Open  ]],
Cell[4957, 107, 74, 1, 27, "Input"],
Cell[5034, 110, 443, 7, 139, "Input"],
Cell[5480, 119, 55, 1, 27, "Input"],
Cell[5538, 122, 49, 1, 27, "Input"],
Cell[5590, 125, 70, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[5685, 130, 101, 2, 27, "Input"],
Cell[5789, 134, 35, 1, 27, "Output"]
}, Open  ]],
Cell[5839, 138, 138, 3, 27, "Input"],

Cell[CellGroupData[{
Cell[6002, 145, 48, 1, 27, "Input"],
Cell[6053, 148, 1855, 36, 189, "Output"]
}, Open  ]],
Cell[7923, 187, 138, 3, 27, "Input"],

Cell[CellGroupData[{
Cell[8086, 194, 48, 1, 27, "Input"],
Cell[8137, 197, 1855, 36, 189, "Output"]
}, Open  ]],
Cell[10007, 236, 138, 3, 27, "Input"],

Cell[CellGroupData[{
Cell[10170, 243, 48, 1, 27, "Input"],
Cell[10221, 246, 1842, 35, 189, "Output"]
}, Open  ]],
Cell[12078, 284, 138, 3, 27, "Input"],

Cell[CellGroupData[{
Cell[12241, 291, 48, 1, 27, "Input"],
Cell[12292, 294, 1855, 36, 189, "Output"]
}, Open  ]],
Cell[14162, 333, 138, 3, 27, "Input"],

Cell[CellGroupData[{
Cell[14325, 340, 48, 1, 27, "Input"],
Cell[14376, 343, 1855, 36, 189, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16268, 384, 306, 6, 107, "Input"],
Cell[16577, 392, 64, 1, 27, "Output"],
Cell[16644, 395, 40, 1, 27, "Output"],
Cell[16687, 398, 161, 2, 43, "Output"],
Cell[16851, 402, 36, 1, 27, "Output"],
Cell[16890, 405, 36, 1, 27, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16963, 411, 47, 1, 27, "Input"],
Cell[17013, 414, 114, 2, 23, "Message"],
Cell[17130, 418, 57, 1, 23, "Print"],
Cell[17190, 421, 49, 1, 23, "Print"],
Cell[17242, 424, 50, 1, 23, "Print"],
Cell[17295, 427, 49, 1, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[17381, 433, 316, 5, 91, "Input"],
Cell[17700, 440, 159, 3, 23, "Print"],
Cell[17862, 445, 159, 3, 23, "Print"],
Cell[18024, 450, 159, 3, 23, "Print"],
Cell[18186, 455, 159, 3, 23, "Print"],
Cell[18348, 460, 159, 3, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[18544, 468, 46, 1, 27, "Input"],
Cell[18593, 471, 427, 7, 23, "Print"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

