(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     23339,        752]*)
(*NotebookOutlinePosition[     24326,        784]*)
(*  CellTagsIndexPosition[     24282,        780]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/ccif.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,<oldx,>func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("AVOID USING C, D, E, I, N and O as variables !"\)], "Print"]
}, Closed]],

Cell[BoxData[{
    \(\(f\  = \ Table[0, {i, 50}];\)\), "\[IndentingNewLine]", 
    \(\(derivs\  = \ 
        Table[Table[0, {i, 1000}], {j, 50}];\)\), "\[IndentingNewLine]", 
    \(\(g\  = Table[Table[0, {i, 10}], {j, 10}]\ ;\)\), "\n", 
    \(\(ginv\  = \ 
        Table[Table[0, {i, 10}], {j, 10}];\)\), "\[IndentingNewLine]", 
    \(\(Euler\  = \ 0;\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< /home/ken/projects/ptah/ptah.out\)], "Input"],

Cell[BoxData[
    InterpretationBox[\(0\[InvisibleSpace]4\),
      SequenceForm[ 0, 4],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(1\[InvisibleSpace]2\),
      SequenceForm[ 1, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(2\[InvisibleSpace]2\),
      SequenceForm[ 2, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(3\[InvisibleSpace]2\),
      SequenceForm[ 3, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(4\[InvisibleSpace]2\),
      SequenceForm[ 4, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(5\[InvisibleSpace]2\),
      SequenceForm[ 5, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(6\[InvisibleSpace]2\),
      SequenceForm[ 6, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(7\[InvisibleSpace]2\),
      SequenceForm[ 7, 2],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[f];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(k + r\^6\)\),
      SequenceForm[ "[", 1, "] = ", 
        Plus[ k, 
          Power[ r, 6]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]"] = "\
\[InvisibleSpace]r\),
      SequenceForm[ "[", 2, "] = ", r],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - p\^2\)\),
      SequenceForm[ "[", 3, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ p, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - q\^2\)\),
      SequenceForm[ "[", 4, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ q, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - s\^2\)\),
      SequenceForm[ "[", 5, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ s, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - u\^2\)\),
      SequenceForm[ "[", 6, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ u, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - v\^2\)\),
      SequenceForm[ "[", 7, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ v, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - w\^2\)\),
      SequenceForm[ "[", 8, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ w, 2]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[derivs];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(3\ r\^5\)\/\(2\ b\
\^3\)\),
      SequenceForm[ "[", 1, ",", 2, "] = ", 
        Times[ 
          Rational[ 3, 2], 
          Power[ b, -3], 
          Power[ r, 5]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]12\[InvisibleSpace]"] = "\[InvisibleSpace]1\/\(2\ c\)\),
      SequenceForm[ "[", 2, ",", 12, "] = ", 
        Times[ 
          Rational[ 1, 2], 
          Power[ c, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]23\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(p\/P\)\)\),
      SequenceForm[ "[", 3, ",", 23, "] = ", 
        Times[ -1, p, 
          Power[ P, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]34\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(q\/Q\)\)\),
      SequenceForm[ "[", 4, ",", 34, "] = ", 
        Times[ -1, q, 
          Power[ Q, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]45\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(s\/S\)\)\),
      SequenceForm[ "[", 5, ",", 45, "] = ", 
        Times[ -1, s, 
          Power[ S, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]56\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(u\/U\)\)\),
      SequenceForm[ "[", 6, ",", 56, "] = ", 
        Times[ -1, u, 
          Power[ U, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]67\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(v\/V\)\)\),
      SequenceForm[ "[", 7, ",", 67, "] = ", 
        Times[ -1, v, 
          Power[ V, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]78\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(w\/W\)\)\),
      SequenceForm[ "[", 8, ",", 78, "] = ", 
        Times[ -1, w, 
          Power[ W, -1]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[g];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]b\/c\^3\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ b, 
          Power[ c, -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\ c\)\/P\^2\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ b, c, 
          Power[ P, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\ c\ \
P\^2\)\/Q\^2\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ b, c, 
          Power[ P, 2], 
          Power[ Q, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\ c\ P\^2\ \
Q\^2\)\/S\^2\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ b, c, 
          Power[ P, 2], 
          Power[ Q, 2], 
          Power[ S, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\ c\ P\^2\ Q\^2\ \
S\^2\)\/U\^2\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ b, c, 
          Power[ P, 2], 
          Power[ Q, 2], 
          Power[ S, 2], 
          Power[ U, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\ c\ P\^2\ Q\^2\ \
S\^2\ U\^2\)\/V\^2\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ b, c, 
          Power[ P, 2], 
          Power[ Q, 2], 
          Power[ S, 2], 
          Power[ U, 2], 
          Power[ V, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\ c\ P\^2\ Q\^2\ \
S\^2\ U\^2\ V\^2\)\/W\^2\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ b, c, 
          Power[ P, 2], 
          Power[ Q, 2], 
          Power[ S, 2], 
          Power[ U, 2], 
          Power[ V, 2], 
          Power[ W, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\ c\ P\^2\ Q\^2\ \
S\^2\ U\^2\ V\^2\ W\^2\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ b, c, 
          Power[ P, 2], 
          Power[ Q, 2], 
          Power[ S, 2], 
          Power[ U, 2], 
          Power[ V, 2], 
          Power[ W, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]c\^9\/b\^3\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ 
          Power[ b, -3], 
          Power[ c, 9]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = \
"\[InvisibleSpace]\(-\(c\^9\/b\^3\)\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ -1, 
          Power[ b, -3], 
          Power[ c, 9]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[ginv];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]c\^3\/b\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, -1], 
          Power[ c, 3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]P\^2\/\(b\ c\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, -1], 
          Power[ c, -1], 
          Power[ P, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]Q\^2\/\(b\ c\ \
P\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ b, -1], 
          Power[ c, -1], 
          Power[ P, -2], 
          Power[ Q, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]S\^2\/\(b\ c\ P\^2\ \
Q\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ b, -1], 
          Power[ c, -1], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ S, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]U\^2\/\(b\ c\ P\^2\ \
Q\^2\ S\^2\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ b, -1], 
          Power[ c, -1], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ S, -2], 
          Power[ U, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]V\^2\/\(b\ c\ P\^2\ \
Q\^2\ S\^2\ U\^2\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ b, -1], 
          Power[ c, -1], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ S, -2], 
          Power[ U, -2], 
          Power[ V, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]W\^2\/\(b\ c\ P\^2\ \
Q\^2\ S\^2\ U\^2\ V\^2\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Power[ b, -1], 
          Power[ c, -1], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ S, -2], 
          Power[ U, -2], 
          Power[ V, -2], 
          Power[ W, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]1\/\(b\ c\ P\^2\ \
Q\^2\ S\^2\ U\^2\ V\^2\ W\^2\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ 
          Power[ b, -1], 
          Power[ c, -1], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ S, -2], 
          Power[ U, -2], 
          Power[ V, -2], 
          Power[ W, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]b\^3\/c\^9\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ 
          Power[ b, 3], 
          Power[ c, -9]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = \
"\[InvisibleSpace]\(-\(b\^3\/c\^9\)\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ -1, 
          Power[ b, 3], 
          Power[ c, -9]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(isrdmg\)], "Input"],

Cell[BoxData[
    \(1\/\(b\ c\^11\ P\^5\ Q\^4\ S\^3\ U\^2\ V\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[sc]]\)], "Input"],

Cell[BoxData[
    \(\(-\(\(9\ k\^2\)\/\(2\ \@r\ \((k + r\^6)\)\^\(9/4\)\)\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riesq]]\)], "Input"],

Cell[BoxData[
    \(\(27\ k\^2\ \((211\ k\^2 - 448\ k\ r\^6 + 1792\ r\^12)\)\)\/\(16\ r\ \
\((k + r\^6)\)\^\(9/2\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Ricsq]]\)], "Input"],

Cell[BoxData[
    \(\(2349\ k\^4\)\/\(4\ r\ \((k + r\^6)\)\^\(9/2\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riecub1]]\)], "Input"],

Cell[BoxData[
    \(\(81\ k\^3\ \((\(-3757\)\ k\^3 + 24864\ k\^2\ r\^6 - 84504\ k\ r\^12 + \
89600\ r\^18)\)\)\/\(128\ r\^\(3/2\)\ \((k + r\^6)\)\^\(27/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riecub2]]\)], "Input"],

Cell[BoxData[
    \(\(81\ k\^3\ \((\(-2747\)\ k\^3 + 27552\ k\^2\ r\^6 - 20328\ k\ r\^12 + \
125440\ r\^18)\)\)\/\(512\ r\^\(3/2\)\ \((k + r\^6)\)\^\(27/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riecub3]]\)], "Input"],

Cell[BoxData[
    \(\(81\ k\^3\ \((505\ k\^3 + 1344\ k\^2\ r\^6 + 32088\ k\ r\^12 + 17920\ \
r\^18)\)\)\/\(256\ r\^\(3/2\)\ \((k + r\^6)\)\^\(27/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riccub]]\)], "Input"],

Cell[BoxData[
    \(\(-\(\(40095\ k\^6\)\/\(8\ r\^\(3/2\)\ \((k + \
r\^6)\)\^\(27/4\)\)\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riericsq]]\)], "Input"],

Cell[BoxData[
    \(\(729\ k\^5\ \((\(-25\)\ k + 168\ r\^6)\)\)\/\(8\ r\^\(3/2\)\ \((k + \
r\^6)\)\^\(27/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riesqric]]\)], "Input"],

Cell[BoxData[
    \(\(-\(\(243\ k\^4\ \((389\ k\^2 - 1736\ k\ r\^6 + 
                3164\ r\^12)\)\)\/\(64\ r\^\(3/2\)\ \((k + \
r\^6)\)\^\(27/4\)\)\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Cdriesq1]]\)], "Input"],

Cell[BoxData[
    \(\(27\ k\^2\ \((499\ k\^4 + 8680\ k\^3\ r\^6 + 282576\ k\^2\ r\^12 - \
401408\ k\ r\^18 + 286720\ r\^24)\)\)\/\(32\ r\^\(3/2\)\ \((k + \
r\^6)\)\^\(27/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Cdriesq2]]\)], "Input"],

Cell[BoxData[
    \(\(81\ k\^4\ \((25\ k\^2 - 2632\ k\ r\^6 + 70000\ r\^12)\)\)\/\(32\ \
r\^\(3/2\)\ \((k + r\^6)\)\^\(27/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Cdricsq1]]\)], "Input"],

Cell[BoxData[
    \(\(81\ k\^4\ \((283\ k\^2 + 3752\ k\ r\^6 + 46480\ r\^12)\)\)\/\(32\ r\^\
\(3/2\)\ \((k + r\^6)\)\^\(27/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Cdricsq2]]\)], "Input"],

Cell[BoxData[
    \(\(81\ k\^4\ \((541\ k\^2 + 10136\ k\ r\^6 + 22960\ r\^12)\)\)\/\(64\ \
r\^\(3/2\)\ \((k + r\^6)\)\^\(27/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Cdricsq3]]\)], "Input"],

Cell[BoxData[
    \(\(81\ k\^4\ \((k + 28\ r\^6)\)\^2\)\/\(64\ r\^\(3/2\)\ \((k + r\^6)\)\^\
\(27/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[RstarR]]\)], "Input"],

Cell[BoxData[
    \(\(10973491200\ k\^2\ \((211\ k\^2 - 448\ k\ r\^6 + 1792\ \
r\^12)\)\)\/\(r\ \((k + r\^6)\)\^\(9/2\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[gRstarR]]\)], "Input"],

Cell[BoxData[
    \(\(-\(\(171460800\ k\^2\ \((485\ k\^2 + 448\ k\ r\^6 - 
                1792\ r\^12)\)\)\/\(r\ \((k + r\^6)\)\^\(9/2\)\)\)\)\)], \
"Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Euler]]\)], "Input"],

Cell[BoxData[
    \(\(-\(\(3444525\ k\^5\ \((5\ k + 8\ r\^6)\)\^2\ \((5\ k\^3 + 
                24\ k\^2\ r\^6 - 102\ k\ r\^12 - 
                256\ r\^18)\)\)\/\(8\ r\^\(5/2\)\ \((k + \
r\^6)\)\^\(45/4\)\)\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(backup[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]11\[InvisibleSpace]":"\
\[InvisibleSpace]14\[InvisibleSpace]":"\[InvisibleSpace]23\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]9\[InvisibleSpace]\
"/"\[InvisibleSpace]2003\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 11, ":", 14, ":", 23, " on ", 4, "/", 9, "/", 2003, "."],
      Editable->False]], "Print"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "ken", \
"projects", "ptah", "mathematica"}, "dbraneval.nb.ps", CharacterEncoding -> \
"ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 70, 1, 27, "Input"],
Cell[1874, 57, 2984, 44, 375, "Print"],
Cell[4861, 103, 81, 1, 23, "Print"]
}, Closed]],
Cell[4957, 107, 368, 7, 88, "Input"],

Cell[CellGroupData[{
Cell[5350, 118, 68, 1, 27, "Input"],
Cell[5421, 121, 121, 3, 23, "Print"],
Cell[5545, 126, 121, 3, 23, "Print"],
Cell[5669, 131, 121, 3, 23, "Print"],
Cell[5793, 136, 121, 3, 23, "Print"],
Cell[5917, 141, 121, 3, 23, "Print"],
Cell[6041, 146, 121, 3, 23, "Print"],
Cell[6165, 151, 121, 3, 23, "Print"],
Cell[6289, 156, 121, 3, 23, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[6447, 164, 51, 1, 24, "Input"],
Cell[6501, 167, 231, 6, 25, "Print"],
Cell[6735, 175, 179, 4, 23, "Print"],
Cell[6917, 181, 256, 7, 25, "Print"],
Cell[7176, 190, 256, 7, 25, "Print"],
Cell[7435, 199, 256, 7, 25, "Print"],
Cell[7694, 208, 256, 7, 25, "Print"],
Cell[7953, 217, 256, 7, 25, "Print"],
Cell[8212, 226, 256, 7, 25, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8505, 238, 56, 1, 27, "Input"],
Cell[8564, 241, 343, 9, 41, "Print"],
Cell[8910, 252, 305, 7, 40, "Print"],
Cell[9218, 261, 286, 6, 37, "Print"],
Cell[9507, 269, 286, 6, 38, "Print"],
Cell[9796, 277, 286, 6, 37, "Print"],
Cell[10085, 285, 286, 6, 37, "Print"],
Cell[10374, 293, 286, 6, 37, "Print"],
Cell[10663, 301, 286, 6, 37, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[10986, 312, 51, 1, 24, "Input"],
Cell[11040, 315, 274, 6, 40, "Print"],
Cell[11317, 323, 284, 6, 40, "Print"],
Cell[11604, 331, 317, 8, 42, "Print"],
Cell[11924, 341, 348, 9, 41, "Print"],
Cell[12275, 352, 379, 10, 41, "Print"],
Cell[12657, 364, 410, 11, 41, "Print"],
Cell[13070, 377, 441, 12, 41, "Print"],
Cell[13514, 391, 440, 12, 25, "Print"],
Cell[13957, 405, 299, 7, 41, "Print"],
Cell[14259, 414, 318, 8, 41, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[14614, 427, 54, 1, 24, "Input"],
Cell[14671, 430, 296, 7, 41, "Print"],
Cell[14970, 439, 329, 8, 41, "Print"],
Cell[15302, 449, 363, 10, 41, "Print"],
Cell[15668, 461, 395, 11, 42, "Print"],
Cell[16066, 474, 427, 12, 42, "Print"],
Cell[16496, 488, 459, 13, 42, "Print"],
Cell[16958, 503, 491, 14, 42, "Print"],
Cell[17452, 519, 495, 14, 41, "Print"],
Cell[17950, 535, 299, 7, 41, "Print"],
Cell[18252, 544, 318, 8, 41, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[18607, 557, 39, 1, 24, "Input"],
Cell[18649, 560, 76, 1, 45, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[18762, 566, 51, 1, 27, "Input"],
Cell[18816, 569, 91, 1, 50, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[18944, 575, 54, 1, 27, "Input"],
Cell[19001, 578, 130, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[19168, 585, 54, 1, 27, "Input"],
Cell[19225, 588, 83, 1, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[19345, 594, 56, 1, 27, "Input"],
Cell[19404, 597, 171, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[19612, 604, 56, 1, 27, "Input"],
Cell[19671, 607, 172, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[19880, 614, 56, 1, 27, "Input"],
Cell[19939, 617, 164, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[20140, 624, 55, 1, 27, "Input"],
Cell[20198, 627, 105, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[20340, 634, 57, 1, 27, "Input"],
Cell[20400, 637, 124, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[20561, 644, 57, 1, 27, "Input"],
Cell[20621, 647, 168, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[20826, 655, 57, 1, 27, "Input"],
Cell[20886, 658, 189, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[21112, 666, 57, 1, 27, "Input"],
Cell[21172, 669, 141, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[21350, 676, 57, 1, 27, "Input"],
Cell[21410, 679, 142, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[21589, 686, 57, 1, 27, "Input"],
Cell[21649, 689, 143, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[21829, 696, 57, 1, 27, "Input"],
Cell[21889, 699, 117, 2, 49, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[22043, 706, 55, 1, 27, "Input"],
Cell[22101, 709, 135, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[22273, 716, 56, 1, 27, "Input"],
Cell[22332, 719, 159, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[22528, 727, 54, 1, 27, "Input"],
Cell[22585, 730, 227, 4, 49, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[22849, 739, 46, 1, 27, "Input"],
Cell[22898, 742, 425, 7, 23, "Print"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

