(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     29655,        854]*)
(*NotebookOutlinePosition[     30645,        886]*)
(*  CellTagsIndexPosition[     30601,        882]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/ccif.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,<oldx,>func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("AVOID USING C, D, E, I, N and O as variables !"\)], "Print"]
}, Open  ]],

Cell[BoxData[
    \(\(dim\  = \ 10;\)\)], "Input"],

Cell[BoxData[{
    \(\(ad2\  = \ 
        Table[Table[0, {j, dim}], {i, dim}];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        Do[\((\[IndentingNewLine]inddim\  = \ 
              Join[DeleteCases[Join[Table[l, {l, dim - 1}], {0}], i | j], {i, 
                  j}]; \[IndentingNewLine]ad2[\([i, j]\)]\  = \ 
              Signature[inddim] 
                Factorial[dim - 2];\[IndentingNewLine])\), {j, dim}], {i, 
          dim}];\)\)}], "Input"],

Cell[BoxData[{
    \(\(ad3\  = \ 
        Table[Table[Table[0, {k, dim}], {j, dim}], {i, 
            dim}];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        Do[Do[\((\[IndentingNewLine]inddim\  = \ 
                Join[DeleteCases[Join[Table[l, {l, dim - 1}], {0}], 
                    i | j | k], {i, j, k}]; \[IndentingNewLine]ad3[\([i, j, 
                    k]\)]\  = \ 
                Signature[inddim] 
                  Factorial[dim - 3];\[IndentingNewLine])\), {k, dim}], {j, 
            dim}], {i, dim}];\)\)}], "Input"],

Cell[BoxData[
    \(\(ptahdir\  = \ "\<~/projects/ptah/\>";\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(sphericalmetric[8]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Spherical metric check: "\[InvisibleSpace]True\),
      SequenceForm[ "Spherical metric check: ", True],
      Editable->False]], "Print"],

Cell[BoxData[
    \({{1, 0, 0, 0, 0, 0, 0, 0}, {0, 
        r\^2\ Sin[t1]\^2\ Sin[t2]\^2\ Sin[t3]\^2\ Sin[t4]\^2\ Sin[t5]\^2\ \
Sin[t6]\^2, 0, 0, 0, 0, 0, 0}, {0, 0, 
        r\^2\ Sin[t2]\^2\ Sin[t3]\^2\ Sin[t4]\^2\ Sin[t5]\^2\ Sin[t6]\^2, 0, 
        0, 0, 0, 0}, {0, 0, 0, 
        r\^2\ Sin[t3]\^2\ Sin[t4]\^2\ Sin[t5]\^2\ Sin[t6]\^2, 0, 0, 0, 
        0}, {0, 0, 0, 0, r\^2\ Sin[t4]\^2\ Sin[t5]\^2\ Sin[t6]\^2, 0, 0, 
        0}, {0, 0, 0, 0, 0, r\^2\ Sin[t5]\^2\ Sin[t6]\^2, 0, 0}, {0, 0, 0, 0, 
        0, 0, r\^2\ Sin[t6]\^2, 0}, {0, 0, 0, 0, 0, 0, 0, r\^2}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(gs\  = \ 
      coordxform[
        sphericalmetric[8], {r, ph, t1, t2, t3, t4, t5, 
          t6}, \[IndentingNewLine]{r, y, ArcCos[w], ArcCos[v], ArcCos[u], 
          ArcCos[s], ArcCos[q], ArcCos[p]}, \[IndentingNewLine]{r, y, w, v, 
          u, s, q, p}]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Spherical metric check: "\[InvisibleSpace]True\),
      SequenceForm[ "Spherical metric check: ", True],
      Editable->False]], "Print"],

Cell[BoxData[
    \({{1, 0, 0, 0, 0, 0, 0, 
        0}, {0, \((\(-1\) + p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
              q)\)\ r\^2\ \((\(-1\) + s)\)\ \((1 + s)\)\ \((\(-1\) + 
              u)\)\ \((1 + u)\)\ \((\(-1\) + v)\)\ \((1 + v)\)\ \((\(-1\) + 
              w)\)\ \((1 + w)\), 0, 0, 0, 0, 0, 0}, {0, 
        0, \(\((\(-1\) + p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + q)\)\ \
r\^2\ \((\(-1\) + s)\)\ \((1 + s)\)\ \((\(-1\) + u)\)\ \((1 + u)\)\ \((\(-1\) \
+ v)\)\ \((1 + v)\)\)\/\(\((\(-1\) + w)\)\ \((1 + w)\)\), 0, 0, 0, 0, 0}, {0, 
        0, 0, \(-\(\(\((\(-1\) + p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
                    q)\)\ r\^2\ \((\(-1\) + s)\)\ \((1 + s)\)\ \((\(-1\) + 
                    u)\)\ \((1 + u)\)\)\/\(\((\(-1\) + v)\)\ \((1 + 
                    v)\)\)\)\), 0, 0, 0, 0}, {0, 0, 0, 
        0, \(\((\(-1\) + p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + q)\)\ \
r\^2\ \((\(-1\) + s)\)\ \((1 + s)\)\)\/\(\((\(-1\) + u)\)\ \((1 + u)\)\), 0, 
        0, 0}, {0, 0, 0, 0, 
        0, \(-\(\(\((\(-1\) + p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
                    q)\)\ r\^2\)\/\(\((\(-1\) + s)\)\ \((1 + s)\)\)\)\), 0, 
        0}, {0, 0, 0, 0, 0, 
        0, \(\((\(-1\) + p)\)\ \((1 + p)\)\ r\^2\)\/\(\((\(-1\) + q)\)\ \((1 \
+ q)\)\), 0}, {0, 0, 0, 0, 0, 0, 
        0, \(-\(r\^2\/\(\((\(-1\) + p)\)\ \((1 + p)\)\)\)\)}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(x\  = \ {r, p, q, s, u, v, w, y, z, t};\)\), "\n", 
    \(\(b[r]\  = \ \((k + r^6)\)^\((1/4)\);\)\), "\n", 
    \(\(c[r]\  = \ r^\((1/2)\);\)\n\), "\n", 
    \(\(g\  = \ IdentityMatrix[10];\)\), "\[IndentingNewLine]", 
    \(\(g[\([1, 1]\)]\  *= \ b[r]/c[r]^3;\)\), "\n", 
    \(\(Do[
        g[\([i, i]\)]\  = \ gs[\([10 - i, 10 - i]\)] b[r]/c[r]^3, {i, 2, 
          8}];\)\n\), "\n", 
    \(\(g[\([9, 9]\)]\  = \ simpler[\((c[r]^3/b[r])\)^3];\)\), "\n", 
    \(\(g[\([10, 10]\)]\  = \ \(-g[\([9, 9]\)]\);\)\), "\n", 
    \(\(components[g];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\((k + \
r\^6)\)\^\(1/4\)\/r\^\(3/2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ r, 
            Rational[ -3, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 6]], 
            Rational[ 1, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(\@r\ \((k + \
r\^6)\)\^\(1/4\)\)\/\(\((\(-1\) + p)\)\ \((1 + p)\)\)\)\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ -1, 
          Power[ 
            Plus[ -1, p], -1], 
          Power[ 
            Plus[ 1, p], -1], 
          Power[ r, 
            Rational[ 1, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 6]], 
            Rational[ 1, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\((\(-1\) + p)\)\ \
\((1 + p)\)\ \@r\ \((k + r\^6)\)\^\(1/4\)\)\/\(\((\(-1\) + q)\)\ \((1 + \
q)\)\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Power[ 
            Plus[ -1, q], -1], 
          Power[ 
            Plus[ 1, q], -1], 
          Power[ r, 
            Rational[ 1, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 6]], 
            Rational[ 1, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(\((\(-1\) + 
                    p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
                    q)\)\ \@r\ \((k + r\^6)\)\^\(1/4\)\)\/\(\((\(-1\) + 
                    s)\)\ \((1 + s)\)\)\)\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ -1, 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Plus[ -1, q], 
          Plus[ 1, q], 
          Power[ r, 
            Rational[ 1, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 6]], 
            Rational[ 1, 4]], 
          Power[ 
            Plus[ -1, s], -1], 
          Power[ 
            Plus[ 1, s], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(\((\(-1\) + p)\)\ \
\((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + q)\)\ \@r\ \((k + r\^6)\)\^\(1/4\)\ \((\
\(-1\) + s)\)\ \((1 + s)\)\)\/\(\((\(-1\) + u)\)\ \((1 + u)\)\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Plus[ -1, q], 
          Plus[ 1, q], 
          Power[ r, 
            Rational[ 1, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 6]], 
            Rational[ 1, 4]], 
          Plus[ -1, s], 
          Plus[ 1, s], 
          Power[ 
            Plus[ -1, u], -1], 
          Power[ 
            Plus[ 1, u], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(\((\(-1\) + 
                    p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
                    q)\)\ \@r\ \((k + r\^6)\)\^\(1/4\)\ \((\(-1\) + 
                    s)\)\ \((1 + s)\)\ \((\(-1\) + u)\)\ \((1 + 
                    u)\)\)\/\(\((\(-1\) + v)\)\ \((1 + v)\)\)\)\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ -1, 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Plus[ -1, q], 
          Plus[ 1, q], 
          Power[ r, 
            Rational[ 1, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 6]], 
            Rational[ 1, 4]], 
          Plus[ -1, s], 
          Plus[ 1, s], 
          Plus[ -1, u], 
          Plus[ 1, u], 
          Power[ 
            Plus[ -1, v], -1], 
          Power[ 
            Plus[ 1, v], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(\((\(-1\) + p)\)\ \
\((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + q)\)\ \@r\ \((k + r\^6)\)\^\(1/4\)\ \((\
\(-1\) + s)\)\ \((1 + s)\)\ \((\(-1\) + u)\)\ \((1 + u)\)\ \((\(-1\) + v)\)\ \
\((1 + v)\)\)\/\(\((\(-1\) + w)\)\ \((1 + w)\)\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Plus[ -1, q], 
          Plus[ 1, q], 
          Power[ r, 
            Rational[ 1, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 6]], 
            Rational[ 1, 4]], 
          Plus[ -1, s], 
          Plus[ 1, s], 
          Plus[ -1, u], 
          Plus[ 1, u], 
          Plus[ -1, v], 
          Plus[ 1, v], 
          Power[ 
            Plus[ -1, w], -1], 
          Power[ 
            Plus[ 1, w], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(\((\(-1\) + 
              p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
              q)\)\ \@r\ \((k + r\^6)\)\^\(1/4\)\ \((\(-1\) + s)\)\ \((1 + 
              s)\)\ \((\(-1\) + u)\)\ \((1 + u)\)\ \((\(-1\) + v)\)\ \((1 + 
              v)\)\ \((\(-1\) + w)\)\ \((1 + w)\)\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Plus[ -1, q], 
          Plus[ 1, q], 
          Power[ r, 
            Rational[ 1, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 6]], 
            Rational[ 1, 4]], 
          Plus[ -1, s], 
          Plus[ 1, s], 
          Plus[ -1, u], 
          Plus[ 1, u], 
          Plus[ -1, v], 
          Plus[ 1, v], 
          Plus[ -1, w], 
          Plus[ 1, w]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]r\^\(9/2\)\/\((k + \
r\^6)\)\^\(3/4\)\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ 
          Power[ r, 
            Rational[ 9, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 6]], 
            Rational[ -3, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = \
"\[InvisibleSpace]\(-\(r\^\(9/2\)\/\((k + r\^6)\)\^\(3/4\)\)\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ -1, 
          Power[ r, 
            Rational[ 9, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 6]], 
            Rational[ -3, 4]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell["\<\
We next parameterize the metric so that each component is a \
polynomial:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(gp\  = \ 
        simpler[g /. {\((k + r^6)\) \[Rule] b^4, 
              r \[Rule] 
                c^2, \((\(-1\) + 
                    p)\) \[Rule] \(-P^2\)/\((1 + p)\), \((\(-1\) + 
                    q)\) \[Rule] \(-Q^2\)/\((1 + q)\), \((\(-1\) + 
                    s)\) \[Rule] \(-S^2\)/\((1 + s)\), \((\(-1\) + 
                    u)\) \[Rule] \(-U^2\)/\((1 + u)\), \((\(-1\) + 
                    v)\) \[Rule] \(-V^2\)/\((1 + v)\), \((\(-1\) + 
                    w)\) \[Rule] \(-W^2\)/\((1 + 
                      w)\)}];\)\), "\[IndentingNewLine]", 
    \(\(components[gp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]b\/c\^3\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ b, 
          Power[ c, -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\ c\)\/P\^2\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ b, c, 
          Power[ P, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\ c\ \
P\^2\)\/Q\^2\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ b, c, 
          Power[ P, 2], 
          Power[ Q, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\ c\ P\^2\ \
Q\^2\)\/S\^2\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ b, c, 
          Power[ P, 2], 
          Power[ Q, 2], 
          Power[ S, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\ c\ P\^2\ Q\^2\ \
S\^2\)\/U\^2\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ b, c, 
          Power[ P, 2], 
          Power[ Q, 2], 
          Power[ S, 2], 
          Power[ U, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\ c\ P\^2\ Q\^2\ \
S\^2\ U\^2\)\/V\^2\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ b, c, 
          Power[ P, 2], 
          Power[ Q, 2], 
          Power[ S, 2], 
          Power[ U, 2], 
          Power[ V, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\ c\ P\^2\ Q\^2\ \
S\^2\ U\^2\ V\^2\)\/W\^2\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ b, c, 
          Power[ P, 2], 
          Power[ Q, 2], 
          Power[ S, 2], 
          Power[ U, 2], 
          Power[ V, 2], 
          Power[ W, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\ c\ P\^2\ Q\^2\ \
S\^2\ U\^2\ V\^2\ W\^2\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ b, c, 
          Power[ P, 2], 
          Power[ Q, 2], 
          Power[ S, 2], 
          Power[ U, 2], 
          Power[ V, 2], 
          Power[ W, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]c\^9\/b\^3\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ 
          Power[ b, -3], 
          Power[ c, 9]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = \
"\[InvisibleSpace]\(-\(c\^9\/b\^3\)\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ -1, 
          Power[ b, -3], 
          Power[ c, 9]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(ginvp\  = \ simpler[Inverse[gp]];\)\), "\[IndentingNewLine]", 
    \(\(components[ginvp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]c\^3\/b\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, -1], 
          Power[ c, 3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]P\^2\/\(b\ c\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, -1], 
          Power[ c, -1], 
          Power[ P, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]Q\^2\/\(b\ c\ \
P\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ b, -1], 
          Power[ c, -1], 
          Power[ P, -2], 
          Power[ Q, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]S\^2\/\(b\ c\ P\^2\ \
Q\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ b, -1], 
          Power[ c, -1], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ S, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]U\^2\/\(b\ c\ P\^2\ \
Q\^2\ S\^2\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ b, -1], 
          Power[ c, -1], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ S, -2], 
          Power[ U, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]V\^2\/\(b\ c\ P\^2\ \
Q\^2\ S\^2\ U\^2\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ b, -1], 
          Power[ c, -1], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ S, -2], 
          Power[ U, -2], 
          Power[ V, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]W\^2\/\(b\ c\ P\^2\ \
Q\^2\ S\^2\ U\^2\ V\^2\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Power[ b, -1], 
          Power[ c, -1], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ S, -2], 
          Power[ U, -2], 
          Power[ V, -2], 
          Power[ W, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]1\/\(b\ c\ P\^2\ \
Q\^2\ S\^2\ U\^2\ V\^2\ W\^2\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ 
          Power[ b, -1], 
          Power[ c, -1], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ S, -2], 
          Power[ U, -2], 
          Power[ V, -2], 
          Power[ W, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]b\^3\/c\^9\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ 
          Power[ b, 3], 
          Power[ c, -9]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = \
"\[InvisibleSpace]\(-\(b\^3\/c\^9\)\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ -1, 
          Power[ b, 3], 
          Power[ c, -9]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(irmdgp\  = \ 
      Simplify[PowerExpand[1/Sqrt[\(-Simplify[Det[gp]]\)]]]\)], "Input"],

Cell[BoxData[
    \(1\/\(b\ c\^11\ P\^5\ Q\^4\ S\^3\ U\^2\ V\)\)], "Output"]
}, Open  ]],

Cell["\<\
Don't forget the sub[i] in the derivatives (takes care of change in \
time coord from i=10 to i=0 in c++)\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(btab\  = \ {b^4, b[r]^4, 
        Table[D[b[r]^4, x[\([sub[i]]\)]], {i, 10}]/\((4 
               b^3)\)}\), "\[IndentingNewLine]", 
    \(ctab\  = {c^2, c[r]^2, 
        Table[D[c[r]^2, x[\([sub[i]]\)]], {i, 10}]/\((2  
              c)\)}\), "\[IndentingNewLine]", 
    \(Ptab\  = \ {P^2, 1 - p^2, 
        Table[D[1 - p^2, x[\([sub[i]]\)]], {i, 10}]/\((2  
              P)\)}\), "\[IndentingNewLine]", 
    \(Qtab\  = \ {Q^2, 1 - q^2, 
        Table[D[1 - q^2, x[\([sub[i]]\)]], {i, 10}]/\((2  
              Q)\)}\), "\[IndentingNewLine]", 
    \(Stab\  = \ {S^2, 1 - s^2, 
        Table[D[1 - s^2, x[\([sub[i]]\)]], {i, 10}]/\((2  
              S)\)}\), "\[IndentingNewLine]", 
    \(Utab\  = \ {U^2, 1 - u^2, 
        Table[D[1 - u^2, x[\([sub[i]]\)]], {i, 10}]/\((2  
              U)\)}\), "\[IndentingNewLine]", 
    \(Vtab\  = \ {V^2, 1 - v^2, 
        Table[D[1 - v^2, x[\([sub[i]]\)]], {i, 10}]/\((2  
              V)\)}\), "\[IndentingNewLine]", 
    \(Wtab\  = \ {W^2, 1 - w^2, 
        Table[D[1 - w^2, x[\([sub[i]]\)]], {i, 10}]/\((2  W)\)}\)}], "Input"],

Cell[BoxData[
    \({b\^4, 
      k + r\^6, {0, \(3\ r\^5\)\/\(2\ b\^3\), 0, 0, 0, 0, 0, 0, 0, 
        0}}\)], "Output"],

Cell[BoxData[
    \({c\^2, r, {0, 1\/\(2\ c\), 0, 0, 0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({P\^2, 
      1 - p\^2, {0, 0, \(-\(p\/P\)\), 0, 0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({Q\^2, 
      1 - q\^2, {0, 0, 0, \(-\(q\/Q\)\), 0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({S\^2, 
      1 - s\^2, {0, 0, 0, 0, \(-\(s\/S\)\), 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({U\^2, 
      1 - u\^2, {0, 0, 0, 0, 0, \(-\(u\/U\)\), 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({V\^2, 
      1 - v\^2, {0, 0, 0, 0, 0, 0, \(-\(v\/V\)\), 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({W\^2, 
      1 - w\^2, {0, 0, 0, 0, 0, 0, 0, \(-\(w\/W\)\), 0, 0}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(dim = 10;\)\), "\[IndentingNewLine]", 
    \(coords\  = \ x\), "\[IndentingNewLine]", 
    \(ints\  = \ {2, 3}\), "\n", 
    \(constants\  = \ {I, k}\), "\n", 
    \(functions\  = \ {btab, ctab, Ptab, Qtab, Stab, Utab, Vtab, 
        Wtab}\), "\[IndentingNewLine]", 
    \(ansatze\  = \ {}\)}], "Input"],

Cell[BoxData[
    \({r, p, q, s, u, v, w, y, z, t}\)], "Output"],

Cell[BoxData[
    \({2, 3}\)], "Output"],

Cell[BoxData[
    \({\[ImaginaryI], k}\)], "Output"],

Cell[BoxData[
    \({{b\^4, 
        k + r\^6, {0, \(3\ r\^5\)\/\(2\ b\^3\), 0, 0, 0, 0, 0, 0, 0, 
          0}}, {c\^2, r, {0, 1\/\(2\ c\), 0, 0, 0, 0, 0, 0, 0, 0}}, {P\^2, 
        1 - p\^2, {0, 0, \(-\(p\/P\)\), 0, 0, 0, 0, 0, 0, 0}}, {Q\^2, 
        1 - q\^2, {0, 0, 0, \(-\(q\/Q\)\), 0, 0, 0, 0, 0, 0}}, {S\^2, 
        1 - s\^2, {0, 0, 0, 0, \(-\(s\/S\)\), 0, 0, 0, 0, 0}}, {U\^2, 
        1 - u\^2, {0, 0, 0, 0, 0, \(-\(u\/U\)\), 0, 0, 0, 0}}, {V\^2, 
        1 - v\^2, {0, 0, 0, 0, 0, 0, \(-\(v\/V\)\), 0, 0, 0}}, {W\^2, 
        1 - w\^2, {0, 0, 0, 0, 0, 0, 0, \(-\(w\/W\)\), 0, 0}}}\)], "Output"],

Cell[BoxData[
    \({}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(headers[];\)\)], "Input"],

Cell[BoxData[
    \("functions completed..."\)], "Print"],

Cell[BoxData[
    \("g completed..."\)], "Print"],

Cell[BoxData[
    \("gi completed..."\)], "Print"],

Cell[BoxData[
    \("d completed..."\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(tensor[ad2, "\<ad2\>"];\)\), "\[IndentingNewLine]", 
    \(\(tensor[ad3, "\<ad3\>"];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("ad2"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ad2", " completed..."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("ad3"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ad3", " completed..."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(backup[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]19\[InvisibleSpace]":"\
\[InvisibleSpace]2\[InvisibleSpace]":"\[InvisibleSpace]47\[InvisibleSpace]" \
on "\[InvisibleSpace]3\[InvisibleSpace]"/"\[InvisibleSpace]30\[InvisibleSpace]\
"/"\[InvisibleSpace]2003\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 19, ":", 2, ":", 47, " on ", 3, "/", 30, "/", 2003, "."],
      Editable->False]], "Print"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "ken", \
"projects", "mathematica", "packages"}, "d11brane.nb.ps", CharacterEncoding \
-> "ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 70, 1, 27, "Input"],
Cell[1874, 57, 2984, 44, 375, "Print"],
Cell[4861, 103, 81, 1, 23, "Print"]
}, Open  ]],
Cell[4957, 107, 50, 1, 27, "Input"],
Cell[5010, 110, 451, 9, 91, "Input"],
Cell[5464, 121, 536, 11, 91, "Input"],
Cell[6003, 134, 74, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[6102, 139, 51, 1, 27, "Input"],
Cell[6156, 142, 177, 3, 23, "Print"],
Cell[6336, 147, 580, 9, 117, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6953, 161, 292, 6, 59, "Input"],
Cell[7248, 169, 177, 3, 23, "Print"],
Cell[7428, 174, 1389, 21, 282, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8854, 200, 581, 11, 187, "Input"],
Cell[9438, 213, 429, 12, 42, "Print"],
Cell[9870, 227, 572, 16, 46, "Print"],
Cell[10445, 245, 641, 19, 46, "Print"],
Cell[11089, 266, 793, 22, 45, "Print"],
Cell[11885, 290, 801, 23, 45, "Print"],
Cell[12689, 315, 974, 27, 45, "Print"],
Cell[13666, 344, 963, 28, 45, "Print"],
Cell[14632, 374, 963, 27, 50, "Print"],
Cell[15598, 403, 429, 12, 44, "Print"],
Cell[16030, 417, 446, 12, 44, "Print"]
}, Open  ]],
Cell[16491, 432, 97, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[16613, 439, 630, 12, 91, "Input"],
Cell[17246, 453, 274, 6, 40, "Print"],
Cell[17523, 461, 284, 6, 40, "Print"],
Cell[17810, 469, 317, 8, 42, "Print"],
Cell[18130, 479, 348, 9, 41, "Print"],
Cell[18481, 490, 379, 10, 41, "Print"],
Cell[18863, 502, 410, 11, 41, "Print"],
Cell[19276, 515, 441, 12, 41, "Print"],
Cell[19720, 529, 440, 12, 25, "Print"],
Cell[20163, 543, 299, 7, 41, "Print"],
Cell[20465, 552, 318, 8, 41, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[20820, 565, 128, 2, 43, "Input"],
Cell[20951, 569, 296, 7, 41, "Print"],
Cell[21250, 578, 329, 8, 41, "Print"],
Cell[21582, 588, 363, 10, 41, "Print"],
Cell[21948, 600, 395, 11, 42, "Print"],
Cell[22346, 613, 427, 12, 42, "Print"],
Cell[22776, 627, 459, 13, 42, "Print"],
Cell[23238, 642, 491, 14, 42, "Print"],
Cell[23732, 658, 495, 14, 41, "Print"],
Cell[24230, 674, 299, 7, 41, "Print"],
Cell[24532, 683, 318, 8, 41, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[24887, 696, 106, 2, 27, "Input"],
Cell[24996, 700, 76, 1, 45, "Output"]
}, Open  ]],
Cell[25087, 704, 129, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[25241, 711, 1094, 23, 139, "Input"],
Cell[26338, 736, 121, 3, 45, "Output"],
Cell[26462, 741, 85, 1, 44, "Output"],
Cell[26550, 744, 101, 2, 41, "Output"],
Cell[26654, 748, 101, 2, 42, "Output"],
Cell[26758, 752, 101, 2, 41, "Output"],
Cell[26862, 756, 101, 2, 41, "Output"],
Cell[26966, 760, 101, 2, 41, "Output"],
Cell[27070, 764, 101, 2, 41, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[27208, 771, 326, 7, 107, "Input"],
Cell[27537, 780, 64, 1, 27, "Output"],
Cell[27604, 783, 40, 1, 27, "Output"],
Cell[27647, 786, 52, 1, 27, "Output"],
Cell[27702, 789, 606, 9, 143, "Output"],
Cell[28311, 800, 36, 1, 27, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[28384, 806, 47, 1, 27, "Input"],
Cell[28434, 809, 57, 1, 23, "Print"],
Cell[28494, 812, 49, 1, 23, "Print"],
Cell[28546, 815, 50, 1, 23, "Print"],
Cell[28599, 818, 49, 1, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[28685, 824, 123, 2, 43, "Input"],
Cell[28811, 828, 157, 3, 23, "Print"],
Cell[28971, 833, 157, 3, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[29165, 841, 46, 1, 27, "Input"],
Cell[29214, 844, 425, 7, 23, "Print"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

