(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     19378,        622]*)
(*NotebookOutlinePosition[     20365,        654]*)
(*  CellTagsIndexPosition[     20321,        650]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/ccif.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,<oldx,>func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("AVOID USING C, D, E, I, N and O as variables !"\)], "Print"]
}, Closed]],

Cell[BoxData[{
    \(\(f\  = \ Table[0, {i, 50}];\)\), "\[IndentingNewLine]", 
    \(\(derivs\  = \ 
        Table[Table[0, {i, 1000}], {j, 50}];\)\), "\[IndentingNewLine]", 
    \(\(g\  = Table[Table[0, {i, 10}], {j, 10}]\ ;\)\), "\n", 
    \(\(ginv\  = \ 
        Table[Table[0, {i, 10}], {j, 10}];\)\), "\[IndentingNewLine]", 
    \(\(Euler\  = \ 0;\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< /home/ken/projects/ptah/ptah.out\)], "Input"],

Cell[BoxData[
    InterpretationBox[\(0\[InvisibleSpace]16\),
      SequenceForm[ 0, 16],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(1\[InvisibleSpace]8\),
      SequenceForm[ 1, 8],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(2\[InvisibleSpace]2\),
      SequenceForm[ 2, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(3\[InvisibleSpace]2\),
      SequenceForm[ 3, 2],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[f];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(k + r\^2\)\),
      SequenceForm[ "[", 1, "] = ", 
        Plus[ k, 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]"] = "\
\[InvisibleSpace]r\),
      SequenceForm[ "[", 2, "] = ", r],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - p\^2\)\),
      SequenceForm[ "[", 3, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ p, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - q\^2\)\),
      SequenceForm[ "[", 4, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ q, 2]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[derivs];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]r\/\(8\ b\^15\)\),
      SequenceForm[ "[", 1, ",", 2, "] = ", 
        Times[ 
          Rational[ 1, 8], 
          Power[ b, -15], r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]12\[InvisibleSpace]"] = "\[InvisibleSpace]1\/\(8\ c\^7\)\),
      SequenceForm[ "[", 2, ",", 12, "] = ", 
        Times[ 
          Rational[ 1, 8], 
          Power[ c, -7]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]23\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(p\/P\)\)\),
      SequenceForm[ "[", 3, ",", 23, "] = ", 
        Times[ -1, p, 
          Power[ P, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]34\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(q\/Q\)\)\),
      SequenceForm[ "[", 4, ",", 34, "] = ", 
        Times[ -1, q, 
          Power[ Q, -1]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[g];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]b\^12\/c\^12\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, 12], 
          Power[ c, -12]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^12\ \
c\^4\)\/P\^2\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, 12], 
          Power[ c, 4], 
          Power[ P, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^12\ c\^4\ P\^2\
\)\/Q\^2\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ b, 12], 
          Power[ c, 4], 
          Power[ P, 2], 
          Power[ Q, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^12\ c\^4\ P\^2\
\ Q\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ b, 12], 
          Power[ c, 4], 
          Power[ P, 2], 
          Power[ Q, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]c\^4\/b\^4\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ b, -4], 
          Power[ c, 4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]c\^4\/b\^4\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ b, -4], 
          Power[ c, 4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]c\^4\/b\^4\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Power[ b, -4], 
          Power[ c, 4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]c\^4\/b\^4\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ 
          Power[ b, -4], 
          Power[ c, 4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]c\^4\/b\^4\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ 
          Power[ b, -4], 
          Power[ c, 4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = \
"\[InvisibleSpace]\(-\(c\^4\/b\^4\)\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ -1, 
          Power[ b, -4], 
          Power[ c, 4]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[ginv];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]c\^12\/b\^12\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, -12], 
          Power[ c, 12]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]P\^2\/\(b\^12\ c\^4\
\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, -12], 
          Power[ c, -4], 
          Power[ P, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]Q\^2\/\(b\^12\ c\^4\
\ P\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ b, -12], 
          Power[ c, -4], 
          Power[ P, -2], 
          Power[ Q, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]1\/\(b\^12\ c\^4\ P\
\^2\ Q\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ b, -12], 
          Power[ c, -4], 
          Power[ P, -2], 
          Power[ Q, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]b\^4\/c\^4\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ b, 4], 
          Power[ c, -4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]b\^4\/c\^4\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ b, 4], 
          Power[ c, -4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]b\^4\/c\^4\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Power[ b, 4], 
          Power[ c, -4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]b\^4\/c\^4\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ 
          Power[ b, 4], 
          Power[ c, -4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]b\^4\/c\^4\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ 
          Power[ b, 4], 
          Power[ c, -4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = \
"\[InvisibleSpace]\(-\(b\^4\/c\^4\)\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ -1, 
          Power[ b, 4], 
          Power[ c, -4]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(isrdmg\)], "Input"],

Cell[BoxData[
    \(1\/\(b\^12\ c\^12\ P\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[sc]]\)], "Input"],

Cell[BoxData[
    \(\(3\ k\^2\)\/\(2\ \@r\ \((k + r\^2)\)\^\(11/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riesq]]\)], "Input"],

Cell[BoxData[
    \(\(3\ k\^2\ \((59\ k\^2 + 192\ k\ r\^2 + 384\ r\^4)\)\)\/\(16\ r\ \((k + \
r\^2)\)\^\(11/2\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Ricsq]]\)], "Input"],

Cell[BoxData[
    \(\(33\ k\^4\)\/\(4\ r\ \((k + r\^2)\)\^\(11/2\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riecub1]]\)], "Input"],

Cell[BoxData[
    \(\(3\ k\^3\ \((841\ k\^3 + 4032\ k\^2\ r\^2 + 6408\ k\ r\^4 + 768\ r\^6)\
\)\)\/\(128\ r\^\(3/2\)\ \((k + r\^2)\)\^\(33/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riecub2]]\)], "Input"],

Cell[BoxData[
    \(\(3\ k\^3\ \((895\ k\^3 + 4608\ k\^2\ r\^2 + 15480\ k\ r\^4 + 20736\ \
r\^6)\)\)\/\(512\ r\^\(3/2\)\ \((k + r\^2)\)\^\(33/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riecub3]]\)], "Input"],

Cell[BoxData[
    \(\(9\ k\^3\ \((9\ k\^3 + 96\ k\^2\ r\^2 + 1512\ k\ r\^4 + 3328\ \
r\^6)\)\)\/\(256\ r\^\(3/2\)\ \((k + r\^2)\)\^\(33/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riccub]]\)], "Input"],

Cell[BoxData[
    \(\(75\ k\^6\)\/\(8\ r\^\(3/2\)\ \((k + r\^2)\)\^\(33/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riericsq]]\)], "Input"],

Cell[BoxData[
    \(\(3\ k\^5\ \((37\ k + 72\ r\^2)\)\)\/\(8\ r\^\(3/2\)\ \((k + \
r\^2)\)\^\(33/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riesqric]]\)], "Input"],

Cell[BoxData[
    \(\(3\ k\^4\ \((169\ k\^2 + 552\ k\ r\^2 + 636\ r\^4)\)\)\/\(64\ r\^\(3/2\
\)\ \((k + r\^2)\)\^\(33/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Cdriesq1]]\)], "Input"],

Cell[BoxData[
    \(\(3\ k\^2\ \((59\ k\^4 + 840\ k\^3\ r\^2 + 5904\ k\^2\ r\^4 + 6144\ k\ \
r\^6 + 18432\ r\^8)\)\)\/\(32\ r\^\(3/2\)\ \((k + r\^2)\)\^\(33/4\)\)\)], \
"Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Cdriesq2]]\)], "Input"],

Cell[BoxData[
    \(\(3\ k\^4\ \((11\ k\^2 + 456\ k\ r\^2 + 4752\ r\^4)\)\)\/\(32\ r\^\(3/2\
\)\ \((k + r\^2)\)\^\(33/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Cdricsq1]]\)], "Input"],

Cell[BoxData[
    \(\(9\ k\^4\ \((11\ k\^2 + 200\ k\ r\^2 + 1104\ r\^4)\)\)\/\(32\ r\^\(3/2\
\)\ \((k + r\^2)\)\^\(33/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Cdricsq2]]\)], "Input"],

Cell[BoxData[
    \(\(3\ k\^4\ \((55\ k\^2 + 744\ k\ r\^2 + 1872\ r\^4)\)\)\/\(64\ r\^\(3/2\
\)\ \((k + r\^2)\)\^\(33/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Cdricsq3]]\)], "Input"],

Cell[BoxData[
    \(\(9\ k\^4\ \((k + 12\ r\^2)\)\^2\)\/\(64\ r\^\(3/2\)\ \((k + \
r\^2)\)\^\(33/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[RstarR]]\)], "Input"],

Cell[BoxData[
    \(\(1219276800\ k\^2\ \((59\ k\^2 + 192\ k\ r\^2 + 384\ r\^4)\)\)\/\(r\ \
\((k + r\^2)\)\^\(11/2\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[gRstarR]]\)], "Input"],

Cell[BoxData[
    \(\(19051200\ k\^2\ \((\(-29\)\ k\^2 + 192\ k\ r\^2 + 384\ r\^4)\)\)\/\(r\
\ \((k + r\^2)\)\^\(11/2\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Euler]]\)], "Input"],

Cell[BoxData[
    \(\(2025\ k\^6\ \((8\ k\^3 + 57\ k\^2\ r\^2 + 48\ k\ r\^4 - 64\ r\^6)\)\)\
\/\(4\ \@r\ \((k + r\^2)\)\^\(55/4\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(backup[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]14\[InvisibleSpace]":"\
\[InvisibleSpace]19\[InvisibleSpace]":"\[InvisibleSpace]4\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]9\[InvisibleSpace]\
"/"\[InvisibleSpace]2003\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 14, ":", 19, ":", 4, " on ", 4, "/", 9, "/", 2003, "."],
      Editable->False]], "Print"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "ken", \
"projects", "ptah", "mathematica"}, "dbraneval.nb.ps", CharacterEncoding -> \
"ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 70, 1, 27, "Input"],
Cell[1874, 57, 2984, 44, 375, "Print"],
Cell[4861, 103, 81, 1, 23, "Print"]
}, Closed]],
Cell[4957, 107, 368, 7, 88, "Input"],

Cell[CellGroupData[{
Cell[5350, 118, 68, 1, 27, "Input"],
Cell[5421, 121, 123, 3, 23, "Print"],
Cell[5547, 126, 121, 3, 23, "Print"],
Cell[5671, 131, 121, 3, 23, "Print"],
Cell[5795, 136, 121, 3, 23, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[5953, 144, 51, 1, 24, "Input"],
Cell[6007, 147, 231, 6, 25, "Print"],
Cell[6241, 155, 179, 4, 23, "Print"],
Cell[6423, 161, 256, 7, 25, "Print"],
Cell[6682, 170, 256, 7, 25, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6975, 182, 56, 1, 27, "Input"],
Cell[7034, 185, 311, 7, 37, "Print"],
Cell[7348, 194, 308, 7, 40, "Print"],
Cell[7659, 203, 286, 6, 37, "Print"],
Cell[7948, 211, 286, 6, 38, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[8271, 222, 51, 1, 24, "Input"],
Cell[8325, 225, 303, 7, 41, "Print"],
Cell[8631, 234, 338, 9, 41, "Print"],
Cell[8972, 245, 369, 10, 42, "Print"],
Cell[9344, 257, 368, 10, 25, "Print"],
Cell[9715, 269, 299, 7, 41, "Print"],
Cell[10017, 278, 299, 7, 41, "Print"],
Cell[10319, 287, 299, 7, 41, "Print"],
Cell[10621, 296, 299, 7, 41, "Print"],
Cell[10923, 305, 299, 7, 41, "Print"],
Cell[11225, 314, 318, 8, 41, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[11580, 327, 54, 1, 24, "Input"],
Cell[11637, 330, 303, 7, 41, "Print"],
Cell[11943, 339, 339, 9, 41, "Print"],
Cell[12285, 350, 371, 10, 41, "Print"],
Cell[12659, 362, 375, 10, 41, "Print"],
Cell[13037, 374, 299, 7, 41, "Print"],
Cell[13339, 383, 299, 7, 41, "Print"],
Cell[13641, 392, 299, 7, 41, "Print"],
Cell[13943, 401, 299, 7, 41, "Print"],
Cell[14245, 410, 299, 7, 41, "Print"],
Cell[14547, 419, 318, 8, 41, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[14902, 432, 39, 1, 24, "Input"],
Cell[14944, 435, 56, 1, 44, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15037, 441, 51, 1, 27, "Input"],
Cell[15091, 444, 83, 1, 50, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15211, 450, 54, 1, 27, "Input"],
Cell[15268, 453, 127, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15432, 460, 54, 1, 27, "Input"],
Cell[15489, 463, 82, 1, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15608, 469, 56, 1, 27, "Input"],
Cell[15667, 472, 158, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15862, 479, 56, 1, 27, "Input"],
Cell[15921, 482, 161, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16119, 489, 56, 1, 27, "Input"],
Cell[16178, 492, 155, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16370, 499, 55, 1, 27, "Input"],
Cell[16428, 502, 91, 1, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16556, 508, 57, 1, 27, "Input"],
Cell[16616, 511, 116, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16769, 518, 57, 1, 27, "Input"],
Cell[16829, 521, 137, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[17003, 528, 57, 1, 27, "Input"],
Cell[17063, 531, 178, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[17278, 539, 57, 1, 27, "Input"],
Cell[17338, 542, 137, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[17512, 549, 57, 1, 27, "Input"],
Cell[17572, 552, 137, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[17746, 559, 57, 1, 27, "Input"],
Cell[17806, 562, 137, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[17980, 569, 57, 1, 27, "Input"],
Cell[18040, 572, 116, 2, 49, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[18193, 579, 55, 1, 27, "Input"],
Cell[18251, 582, 132, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[18420, 589, 56, 1, 27, "Input"],
Cell[18479, 592, 135, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[18651, 599, 54, 1, 27, "Input"],
Cell[18708, 602, 145, 2, 50, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[18890, 609, 46, 1, 27, "Input"],
Cell[18939, 612, 423, 7, 23, "Print"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

