(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     23293,        696]*)
(*NotebookOutlinePosition[     24283,        728]*)
(*  CellTagsIndexPosition[     24239,        724]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/ccif.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,<oldx,>func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("AVOID USING C, D, E, I, N and O as variables !"\)], "Print"]
}, Closed]],

Cell[BoxData[
    \(\(dim\  = \ 10;\)\)], "Input"],

Cell[BoxData[{
    \(\(ad2\  = \ 
        Table[Table[0, {j, dim}], {i, dim}];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        Do[\((\[IndentingNewLine]inddim\  = \ 
              Join[DeleteCases[Join[Table[l, {l, dim - 1}], {0}], i | j], {i, 
                  j}]; \[IndentingNewLine]ad2[\([i, j]\)]\  = \ 
              Signature[inddim] 
                Factorial[dim - 2];\[IndentingNewLine])\), {j, dim}], {i, 
          dim}];\)\)}], "Input"],

Cell[BoxData[{
    \(\(ad3\  = \ 
        Table[Table[Table[0, {k, dim}], {j, dim}], {i, 
            dim}];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        Do[Do[\((\[IndentingNewLine]inddim\  = \ 
                Join[DeleteCases[Join[Table[l, {l, dim - 1}], {0}], 
                    i | j | k], {i, j, k}]; \[IndentingNewLine]ad3[\([i, j, 
                    k]\)]\  = \ 
                Signature[inddim] 
                  Factorial[dim - 3];\[IndentingNewLine])\), {k, dim}], {j, 
            dim}], {i, dim}];\)\)}], "Input"],

Cell[BoxData[
    \(\(ptahdir\  = \ "\<~/projects/ptah/\>";\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(sphericalmetric[4]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Spherical metric check: "\[InvisibleSpace]True\),
      SequenceForm[ "Spherical metric check: ", True],
      Editable->False]], "Print"],

Cell[BoxData[
    \({{1, 0, 0, 0}, {0, r\^2\ Sin[t1]\^2\ Sin[t2]\^2, 0, 0}, {0, 0, 
        r\^2\ Sin[t2]\^2, 0}, {0, 0, 0, r\^2}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(gs\  = \ 
      coordxform[
        sphericalmetric[4], {r, ph, t1, t2}, \[IndentingNewLine]{r, y, 
          ArcCos[q], ArcCos[p]}, \[IndentingNewLine]{r, y, q, p}]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Spherical metric check: "\[InvisibleSpace]True\),
      SequenceForm[ "Spherical metric check: ", True],
      Editable->False]], "Print"],

Cell[BoxData[
    \({{1, 0, 0, 
        0}, {0, \((\(-1\) + p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
              q)\)\ r\^2, 0, 0}, {0, 
        0, \(\((\(-1\) + p)\)\ \((1 + p)\)\ r\^2\)\/\(\((\(-1\) + q)\)\ \((1 \
+ q)\)\), 0}, {0, 0, 
        0, \(-\(r\^2\/\(\((\(-1\) + p)\)\ \((1 + p)\)\)\)\)}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(x\  = \ {r, p, q, y, s, u, v, w, z, t};\)\), "\n", 
    \(\(b[r]\  = \ \((k + r^2)\)^\((1/16)\);\)\), "\n", 
    \(\(c[r]\  = \ r^\((1/8)\);\)\n\), "\n", 
    \(\(g\  = \ IdentityMatrix[10];\)\), "\[IndentingNewLine]", 
    \(\(g[\([1, 1]\)]\  *= \ b[r]^12/c[r]^12;\)\), "\n", 
    \(\(Do[
        g[\([i, i]\)]\  = \ gs[\([6 - i, 6 - i]\)] b[r]^12/c[r]^12, {i, 2, 
          4}];\)\n\), "\n", 
    \(\ \(Do[
        g[\([i, i]\)] = \ simpler[\((c[r]/b[r])\)^4], {i, 5, 9}];\)\), "\n", 
    \(\(g[\([10, 10]\)]\  = \ \(-g[\([9, 9]\)]\);\)\), "\n", 
    \(\(components[g];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\((k + \
r\^2)\)\^\(3/4\)\/r\^\(3/2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ r, 
            Rational[ -3, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 2]], 
            Rational[ 3, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(\@r\ \((k + \
r\^2)\)\^\(3/4\)\)\/\(\((\(-1\) + p)\)\ \((1 + p)\)\)\)\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ -1, 
          Power[ 
            Plus[ -1, p], -1], 
          Power[ 
            Plus[ 1, p], -1], 
          Power[ r, 
            Rational[ 1, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 2]], 
            Rational[ 3, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\((\(-1\) + p)\)\ \
\((1 + p)\)\ \@r\ \((k + r\^2)\)\^\(3/4\)\)\/\(\((\(-1\) + q)\)\ \((1 + \
q)\)\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Power[ 
            Plus[ -1, q], -1], 
          Power[ 
            Plus[ 1, q], -1], 
          Power[ r, 
            Rational[ 1, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 2]], 
            Rational[ 3, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\((\(-1\) + 
              p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
              q)\)\ \@r\ \((k + r\^2)\)\^\(3/4\)\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Plus[ -1, q], 
          Plus[ 1, q], 
          Power[ r, 
            Rational[ 1, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 2]], 
            Rational[ 3, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\@r\/\((k + r\^2)\)\
\^\(1/4\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ r, 
            Rational[ 1, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 2]], 
            Rational[ -1, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\@r\/\((k + r\^2)\)\
\^\(1/4\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ r, 
            Rational[ 1, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 2]], 
            Rational[ -1, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\@r\/\((k + r\^2)\)\
\^\(1/4\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Power[ r, 
            Rational[ 1, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 2]], 
            Rational[ -1, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\@r\/\((k + r\^2)\)\
\^\(1/4\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ 
          Power[ r, 
            Rational[ 1, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 2]], 
            Rational[ -1, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]\@r\/\((k + r\^2)\)\
\^\(1/4\)\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ 
          Power[ r, 
            Rational[ 1, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 2]], 
            Rational[ -1, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\@r\/\((k + r\
\^2)\)\^\(1/4\)\)\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ -1, 
          Power[ r, 
            Rational[ 1, 2]], 
          Power[ 
            Plus[ k, 
              Power[ r, 2]], 
            Rational[ -1, 4]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell["\<\
We next parameterize the metric so that each component is a \
polynomial:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(gp\  = \ 
        simpler[g /. {\((k + r^2)\) \[Rule] b^16, 
              r \[Rule] 
                c^8, \((\(-1\) + 
                    p)\) \[Rule] \(-P^2\)/\((1 + p)\), \((\(-1\) + 
                    q)\) \[Rule] \(-Q^2\)/\((1 + 
                      q)\)}];\)\), "\[IndentingNewLine]", 
    \(\(components[gp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]b\^12\/c\^12\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, 12], 
          Power[ c, -12]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^12\ \
c\^4\)\/P\^2\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, 12], 
          Power[ c, 4], 
          Power[ P, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^12\ c\^4\ P\^2\
\)\/Q\^2\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ b, 12], 
          Power[ c, 4], 
          Power[ P, 2], 
          Power[ Q, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^12\ c\^4\ P\^2\
\ Q\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ b, 12], 
          Power[ c, 4], 
          Power[ P, 2], 
          Power[ Q, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]c\^4\/b\^4\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ b, -4], 
          Power[ c, 4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]c\^4\/b\^4\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ b, -4], 
          Power[ c, 4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]c\^4\/b\^4\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Power[ b, -4], 
          Power[ c, 4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]c\^4\/b\^4\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ 
          Power[ b, -4], 
          Power[ c, 4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]c\^4\/b\^4\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ 
          Power[ b, -4], 
          Power[ c, 4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = \
"\[InvisibleSpace]\(-\(c\^4\/b\^4\)\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ -1, 
          Power[ b, -4], 
          Power[ c, 4]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(ginvp\  = \ simpler[Inverse[gp]];\)\), "\[IndentingNewLine]", 
    \(\(components[ginvp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]c\^12\/b\^12\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, -12], 
          Power[ c, 12]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]P\^2\/\(b\^12\ c\^4\
\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, -12], 
          Power[ c, -4], 
          Power[ P, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]Q\^2\/\(b\^12\ c\^4\
\ P\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ b, -12], 
          Power[ c, -4], 
          Power[ P, -2], 
          Power[ Q, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]1\/\(b\^12\ c\^4\ P\
\^2\ Q\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ b, -12], 
          Power[ c, -4], 
          Power[ P, -2], 
          Power[ Q, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]b\^4\/c\^4\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ b, 4], 
          Power[ c, -4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]b\^4\/c\^4\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ b, 4], 
          Power[ c, -4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]b\^4\/c\^4\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Power[ b, 4], 
          Power[ c, -4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]b\^4\/c\^4\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ 
          Power[ b, 4], 
          Power[ c, -4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]b\^4\/c\^4\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ 
          Power[ b, 4], 
          Power[ c, -4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = \
"\[InvisibleSpace]\(-\(b\^4\/c\^4\)\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ -1, 
          Power[ b, 4], 
          Power[ c, -4]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(irmdgp\  = \ 
      Simplify[PowerExpand[1/Sqrt[\(-Simplify[Det[gp]]\)]]]\)], "Input"],

Cell[BoxData[
    \(1\/\(b\^12\ c\^12\ P\)\)], "Output"]
}, Open  ]],

Cell["\<\
Don't forget the sub[i] in the derivatives (takes care of change in \
time coord from i=10 to i=0 in c++)\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(btab\  = \ {b^16, b[r]^16, 
        Table[D[b[r]^16, x[\([sub[i]]\)]], {i, 10}]/\((16 
               b^15)\)}\), "\[IndentingNewLine]", 
    \(ctab\  = {c^8, c[r]^8, 
        Table[D[c[r]^8, x[\([sub[i]]\)]], {i, 10}]/\((8 
               c^7)\)}\), "\[IndentingNewLine]", 
    \(Ptab\  = \ {P^2, 1 - p^2, 
        Table[D[1 - p^2, x[\([sub[i]]\)]], {i, 10}]/\((2  
              P)\)}\), "\[IndentingNewLine]", 
    \(Qtab\  = \ {Q^2, 1 - q^2, 
        Table[D[1 - q^2, x[\([sub[i]]\)]], {i, 10}]/\((2  Q)\)}\)}], "Input"],

Cell[BoxData[
    \({b\^16, 
      k + r\^2, {0, r\/\(8\ b\^15\), 0, 0, 0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({c\^8, r, {0, 1\/\(8\ c\^7\), 0, 0, 0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({P\^2, 
      1 - p\^2, {0, 0, \(-\(p\/P\)\), 0, 0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({Q\^2, 
      1 - q\^2, {0, 0, 0, \(-\(q\/Q\)\), 0, 0, 0, 0, 0, 0}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(dim = 10;\)\), "\[IndentingNewLine]", 
    \(coords\  = \ x\), "\[IndentingNewLine]", 
    \(ints\  = \ {2, 3}\), "\n", 
    \(constants\  = \ {I, k}\), "\n", 
    \(functions\  = \ {btab, ctab, Ptab, Qtab}\), "\[IndentingNewLine]", 
    \(ansatze\  = \ {}\)}], "Input"],

Cell[BoxData[
    \({r, p, q, y, s, u, v, w, z, t}\)], "Output"],

Cell[BoxData[
    \({2, 3}\)], "Output"],

Cell[BoxData[
    \({\[ImaginaryI], k}\)], "Output"],

Cell[BoxData[
    \({{b\^16, 
        k + r\^2, {0, r\/\(8\ b\^15\), 0, 0, 0, 0, 0, 0, 0, 0}}, {c\^8, 
        r, {0, 1\/\(8\ c\^7\), 0, 0, 0, 0, 0, 0, 0, 0}}, {P\^2, 
        1 - p\^2, {0, 0, \(-\(p\/P\)\), 0, 0, 0, 0, 0, 0, 0}}, {Q\^2, 
        1 - q\^2, {0, 0, 0, \(-\(q\/Q\)\), 0, 0, 0, 0, 0, 0}}}\)], "Output"],

Cell[BoxData[
    \({}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(headers[];\)\)], "Input"],

Cell[BoxData[
    \("functions completed..."\)], "Print"],

Cell[BoxData[
    \("g completed..."\)], "Print"],

Cell[BoxData[
    \("gi completed..."\)], "Print"],

Cell[BoxData[
    \("d completed..."\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(tensor[ad2, "\<ad2\>"];\)\), "\[IndentingNewLine]", 
    \(\(tensor[ad3, "\<ad3\>"];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("ad2"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ad2", " completed..."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("ad3"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ad3", " completed..."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(backup[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]23\[InvisibleSpace]":"\
\[InvisibleSpace]2\[InvisibleSpace]":"\[InvisibleSpace]32\[InvisibleSpace]" \
on "\[InvisibleSpace]3\[InvisibleSpace]"/"\[InvisibleSpace]30\[InvisibleSpace]\
"/"\[InvisibleSpace]2003\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 23, ":", 2, ":", 32, " on ", 3, "/", 30, "/", 2003, "."],
      Editable->False]], "Print"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "ken", \
"projects", "mathematica", "packages"}, "d11brane.nb.ps", CharacterEncoding \
-> "ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 70, 1, 27, "Input"],
Cell[1874, 57, 2984, 44, 375, "Print"],
Cell[4861, 103, 81, 1, 23, "Print"]
}, Closed]],
Cell[4957, 107, 50, 1, 24, "Input"],
Cell[5010, 110, 451, 9, 91, "Input"],
Cell[5464, 121, 536, 11, 91, "Input"],
Cell[6003, 134, 74, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[6102, 139, 51, 1, 27, "Input"],
Cell[6156, 142, 177, 3, 23, "Print"],
Cell[6336, 147, 144, 2, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6517, 154, 198, 4, 59, "Input"],
Cell[6718, 160, 177, 3, 23, "Print"],
Cell[6898, 165, 319, 6, 75, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[7254, 176, 610, 12, 187, "Input"],
Cell[7867, 190, 429, 12, 42, "Print"],
Cell[8299, 204, 572, 16, 46, "Print"],
Cell[8874, 222, 641, 19, 46, "Print"],
Cell[9518, 243, 613, 17, 30, "Print"],
Cell[10134, 262, 422, 12, 46, "Print"],
Cell[10559, 276, 422, 12, 46, "Print"],
Cell[10984, 290, 422, 12, 46, "Print"],
Cell[11409, 304, 422, 12, 46, "Print"],
Cell[11834, 318, 422, 12, 46, "Print"],
Cell[12259, 332, 439, 12, 46, "Print"]
}, Open  ]],
Cell[12713, 347, 97, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[12835, 354, 359, 8, 43, "Input"],
Cell[13197, 364, 303, 7, 41, "Print"],
Cell[13503, 373, 338, 9, 41, "Print"],
Cell[13844, 384, 369, 10, 42, "Print"],
Cell[14216, 396, 368, 10, 25, "Print"],
Cell[14587, 408, 299, 7, 41, "Print"],
Cell[14889, 417, 299, 7, 41, "Print"],
Cell[15191, 426, 299, 7, 41, "Print"],
Cell[15493, 435, 299, 7, 41, "Print"],
Cell[15795, 444, 299, 7, 41, "Print"],
Cell[16097, 453, 318, 8, 41, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16452, 466, 128, 2, 43, "Input"],
Cell[16583, 470, 303, 7, 41, "Print"],
Cell[16889, 479, 339, 9, 41, "Print"],
Cell[17231, 490, 371, 10, 41, "Print"],
Cell[17605, 502, 375, 10, 41, "Print"],
Cell[17983, 514, 299, 7, 41, "Print"],
Cell[18285, 523, 299, 7, 41, "Print"],
Cell[18587, 532, 299, 7, 41, "Print"],
Cell[18889, 541, 299, 7, 41, "Print"],
Cell[19191, 550, 299, 7, 41, "Print"],
Cell[19493, 559, 318, 8, 41, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[19848, 572, 106, 2, 27, "Input"],
Cell[19957, 576, 56, 1, 44, "Output"]
}, Open  ]],
Cell[20028, 580, 129, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[20182, 587, 545, 11, 75, "Input"],
Cell[20730, 600, 104, 2, 41, "Output"],
Cell[20837, 604, 88, 1, 44, "Output"],
Cell[20928, 607, 101, 2, 41, "Output"],
Cell[21032, 611, 101, 2, 42, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[21170, 618, 293, 6, 107, "Input"],
Cell[21466, 626, 64, 1, 27, "Output"],
Cell[21533, 629, 40, 1, 27, "Output"],
Cell[21576, 632, 52, 1, 27, "Output"],
Cell[21631, 635, 315, 5, 78, "Output"],
Cell[21949, 642, 36, 1, 27, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[22022, 648, 47, 1, 27, "Input"],
Cell[22072, 651, 57, 1, 23, "Print"],
Cell[22132, 654, 49, 1, 23, "Print"],
Cell[22184, 657, 50, 1, 23, "Print"],
Cell[22237, 660, 49, 1, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[22323, 666, 123, 2, 43, "Input"],
Cell[22449, 670, 157, 3, 23, "Print"],
Cell[22609, 675, 157, 3, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[22803, 683, 46, 1, 27, "Input"],
Cell[22852, 686, 425, 7, 23, "Print"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

