(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     19017,        613]*)
(*NotebookOutlinePosition[     20004,        645]*)
(*  CellTagsIndexPosition[     19960,        641]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/ccif.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,<oldx,>func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("AVOID USING C, D, E, I, N and O as variables !"\)], "Print"]
}, Closed]],

Cell[BoxData[{
    \(\(f\  = \ Table[0, {i, 50}];\)\), "\[IndentingNewLine]", 
    \(\(derivs\  = \ 
        Table[Table[0, {i, 1000}], {j, 50}];\)\), "\[IndentingNewLine]", 
    \(\(g\  = Table[Table[0, {i, 11}], {j, 11}]\ ;\)\), "\n", 
    \(\(ginv\  = \ 
        Table[Table[0, {i, 11}], {j, 11}];\)\), "\[IndentingNewLine]", 
    \(\(Euler\  = \ 0;\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< /home/ken/projects/ptah/ptah.out\)], "Input"],

Cell[BoxData[
    InterpretationBox[\(0\[InvisibleSpace]3\),
      SequenceForm[ 0, 3],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(1\[InvisibleSpace]2\),
      SequenceForm[ 1, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(2\[InvisibleSpace]2\),
      SequenceForm[ 2, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(3\[InvisibleSpace]2\),
      SequenceForm[ 3, 2],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[f];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(k + r\^3\)\),
      SequenceForm[ "[", 1, "] = ", 
        Plus[ k, 
          Power[ r, 3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - p\^2\)\),
      SequenceForm[ "[", 2, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ p, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - q\^2\)\),
      SequenceForm[ "[", 3, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ q, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - s\^2\)\),
      SequenceForm[ "[", 4, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ s, 2]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[derivs];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]r\^2\/b\^2\),
      SequenceForm[ "[", 1, ",", 2, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]14\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(p\/P\)\)\),
      SequenceForm[ "[", 2, ",", 14, "] = ", 
        Times[ -1, p, 
          Power[ P, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]26\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(q\/Q\)\)\),
      SequenceForm[ "[", 3, ",", 26, "] = ", 
        Times[ -1, q, 
          Power[ Q, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]38\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(s\/S\)\)\),
      SequenceForm[ "[", 4, ",", 38, "] = ", 
        Times[ -1, s, 
          Power[ S, -1]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[g];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]b\^2\/r\^2\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, 2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]b\^2\/P\^2\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, 2], 
          Power[ P, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^2\ \
P\^2\)\/Q\^2\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ b, 2], 
          Power[ P, 2], 
          Power[ Q, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^2\ P\^2\ \
Q\^2\)\/S\^2\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ b, 2], 
          Power[ P, 2], 
          Power[ Q, 2], 
          Power[ S, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^2\ P\^2\ Q\^2\ \
S\^2\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ b, 2], 
          Power[ P, 2], 
          Power[ Q, 2], 
          Power[ S, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]r\/b\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ b, -1], r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]r\/b\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Power[ b, -1], r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]r\/b\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ 
          Power[ b, -1], r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]r\/b\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ 
          Power[ b, -1], r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]r\/b\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ 
          Power[ b, -1], r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(r\/b\)\)\),
      SequenceForm[ "[", 11, ",", 11, "] = ", 
        Times[ -1, 
          Power[ b, -1], r]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[ginv];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]r\^2\/b\^2\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]P\^2\/b\^2\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]Q\^2\/\(b\^2\ \
P\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]S\^2\/\(b\^2\ P\^2\ \
Q\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ S, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]1\/\(b\^2\ P\^2\ \
Q\^2\ S\^2\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ S, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]b\/r\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ b, 
          Power[ r, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]b\/r\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ b, 
          Power[ r, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]b\/r\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ b, 
          Power[ r, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]b\/r\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ b, 
          Power[ r, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]b\/r\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ b, 
          Power[ r, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(b\/r\)\)\),
      SequenceForm[ "[", 11, ",", 11, "] = ", 
        Times[ -1, b, 
          Power[ r, -1]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(isrdmg\)], "Input"],

Cell[BoxData[
    \(1\/\(b\^2\ P\^2\ Q\ r\^2\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[sc]]\)], "Input"],

Cell[BoxData[
    \(\(3\ k\^2\)\/\(2\ \((k + r\^3)\)\^\(8/3\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riesq]]\)], "Input"],

Cell[BoxData[
    \(\(9\ k\^2\ \((13\ k\^2 + 32\ k\ r\^3 + 160\ r\^6)\)\)\/\(4\ \((k + \
r\^3)\)\^\(16/3\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Ricsq]]\)], "Input"],

Cell[BoxData[
    \(\(207\ k\^4\)\/\(4\ \((k + r\^3)\)\^\(16/3\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riecub1]]\)], "Input"],

Cell[BoxData[
    \(\(3\ k\^3\ \((121\ k\^3 + 816\ k\^2\ r\^3 + 1152\ k\ r\^6 - 320\ \
r\^9)\)\)\/\(8\ \((k + r\^3)\)\^8\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riecub2]]\)], "Input"],

Cell[BoxData[
    \(\(3\ k\^3\ \((221\ k\^3 + 1776\ k\^2\ r\^3 + 5328\ k\ r\^6 + 14720\ \
r\^9)\)\)\/\(32\ \((k + r\^3)\)\^8\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riecub3]]\)], "Input"],

Cell[BoxData[
    \(\(3\ k\^3\ \((25\ k\^3 + 240\ k\^2\ r\^3 + 1044\ k\ r\^6 + 3760\ \
r\^9)\)\)\/\(8\ \((k + r\^3)\)\^8\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riccub]]\)], "Input"],

Cell[BoxData[
    \(\(675\ k\^6\)\/\(8\ \((k + r\^3)\)\^8\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riericsq]]\)], "Input"],

Cell[BoxData[
    \(\(27\ k\^5\ \((25\ k + 144\ r\^3)\)\)\/\(8\ \((k + r\^3)\)\^8\)\)], \
"Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Riesqric]]\)], "Input"],

Cell[BoxData[
    \(\(27\ k\^4\ \((19\ k\^2 + 96\ k\ r\^3 + 80\ r\^6)\)\)\/\(16\ \((k + \
r\^3)\)\^8\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Cdriesq1]]\)], "Input"],

Cell[BoxData[
    \(\(72\ k\^2\ r\^6\ \((56\ k\^2 - 100\ k\ r\^3 + 175\ r\^6)\)\)\/\((k + r\
\^3)\)\^8\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Cdriesq2]]\)], "Input"],

Cell[BoxData[
    \(\(4770\ k\^4\ r\^6\)\/\((k + r\^3)\)\^8\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Cdricsq1]]\)], "Input"],

Cell[BoxData[
    \(\(3474\ k\^4\ r\^6\)\/\((k + r\^3)\)\^8\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Cdricsq2]]\)], "Input"],

Cell[BoxData[
    \(\(1089\ k\^4\ r\^6\)\/\((k + r\^3)\)\^8\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[Cdricsq3]]\)], "Input"],

Cell[BoxData[
    \(\(36\ k\^4\ r\^6\)\/\((k + r\^3)\)\^8\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[RstarR]]\)], "Input"],

Cell[BoxData[
    \(\(1185137049600\ k\^2\ \((13\ k\^2 + 32\ k\ r\^3 + 160\ r\^6)\)\)\/\((k \
+ r\^3)\)\^\(16/3\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[simp[gRstarR]]\)], "Input"],

Cell[BoxData[
    \(\(-\(\(14631321600\ k\^2\ \((33\ k\^2 - 32\ k\ r\^3 - 
                160\ r\^6)\)\)\/\((k + r\^3)\)\^\(16/3\)\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(backup[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]17\[InvisibleSpace]":"\
\[InvisibleSpace]52\[InvisibleSpace]":"\[InvisibleSpace]28\[InvisibleSpace]" \
on "\[InvisibleSpace]5\[InvisibleSpace]"/"\[InvisibleSpace]6\[InvisibleSpace]\
"/"\[InvisibleSpace]2003\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 17, ":", 52, ":", 28, " on ", 5, "/", 6, "/", 2003, "."],
      Editable->False]], "Print"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "ken", \
"projects", "ptah", "mathematica"}, "dbraneval.nb.ps", CharacterEncoding -> \
"ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 70, 1, 27, "Input"],
Cell[1874, 57, 2984, 44, 375, "Print"],
Cell[4861, 103, 81, 1, 23, "Print"]
}, Closed]],
Cell[4957, 107, 368, 7, 88, "Input"],

Cell[CellGroupData[{
Cell[5350, 118, 68, 1, 27, "Input"],
Cell[5421, 121, 121, 3, 23, "Print"],
Cell[5545, 126, 121, 3, 23, "Print"],
Cell[5669, 131, 121, 3, 23, "Print"],
Cell[5793, 136, 121, 3, 23, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[5951, 144, 51, 1, 24, "Input"],
Cell[6005, 147, 231, 6, 25, "Print"],
Cell[6239, 155, 256, 7, 25, "Print"],
Cell[6498, 164, 256, 7, 25, "Print"],
Cell[6757, 173, 256, 7, 25, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[7050, 185, 56, 1, 27, "Input"],
Cell[7109, 188, 299, 7, 41, "Print"],
Cell[7411, 197, 286, 6, 37, "Print"],
Cell[7700, 205, 286, 6, 38, "Print"],
Cell[7989, 213, 286, 6, 37, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8312, 224, 51, 1, 27, "Input"],
Cell[8366, 227, 299, 7, 41, "Print"],
Cell[8668, 236, 299, 7, 41, "Print"],
Cell[8970, 245, 336, 9, 42, "Print"],
Cell[9309, 256, 367, 10, 41, "Print"],
Cell[9679, 268, 366, 10, 25, "Print"],
Cell[10048, 280, 271, 6, 37, "Print"],
Cell[10322, 288, 271, 6, 37, "Print"],
Cell[10596, 296, 271, 6, 37, "Print"],
Cell[10870, 304, 271, 6, 37, "Print"],
Cell[11144, 312, 275, 6, 37, "Print"],
Cell[11422, 320, 288, 6, 37, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[11747, 331, 54, 1, 27, "Input"],
Cell[11804, 334, 299, 7, 41, "Print"],
Cell[12106, 343, 299, 7, 41, "Print"],
Cell[12408, 352, 337, 9, 41, "Print"],
Cell[12748, 363, 369, 10, 42, "Print"],
Cell[13120, 375, 373, 10, 41, "Print"],
Cell[13496, 387, 271, 6, 40, "Print"],
Cell[13770, 395, 271, 6, 40, "Print"],
Cell[14044, 403, 271, 6, 40, "Print"],
Cell[14318, 411, 271, 6, 40, "Print"],
Cell[14592, 419, 275, 6, 40, "Print"],
Cell[14870, 427, 288, 6, 40, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15195, 438, 39, 1, 27, "Input"],
Cell[15237, 441, 60, 1, 45, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15334, 447, 51, 1, 27, "Input"],
Cell[15388, 450, 77, 1, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15502, 456, 54, 1, 27, "Input"],
Cell[15559, 459, 122, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15718, 466, 54, 1, 27, "Input"],
Cell[15775, 469, 80, 1, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15892, 475, 56, 1, 27, "Input"],
Cell[15951, 478, 136, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16124, 485, 56, 1, 27, "Input"],
Cell[16183, 488, 140, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16360, 495, 56, 1, 27, "Input"],
Cell[16419, 498, 136, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16592, 505, 55, 1, 27, "Input"],
Cell[16650, 508, 73, 1, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16760, 514, 57, 1, 27, "Input"],
Cell[16820, 517, 99, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16956, 524, 57, 1, 27, "Input"],
Cell[17016, 527, 116, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[17169, 534, 57, 1, 27, "Input"],
Cell[17229, 537, 116, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[17382, 544, 57, 1, 27, "Input"],
Cell[17442, 547, 73, 1, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[17552, 553, 57, 1, 27, "Input"],
Cell[17612, 556, 73, 1, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[17722, 562, 57, 1, 27, "Input"],
Cell[17782, 565, 73, 1, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[17892, 571, 57, 1, 27, "Input"],
Cell[17952, 574, 71, 1, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[18060, 580, 55, 1, 27, "Input"],
Cell[18118, 583, 127, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[18282, 590, 56, 1, 27, "Input"],
Cell[18341, 593, 149, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[18527, 600, 46, 1, 27, "Input"],
Cell[18576, 603, 425, 7, 23, "Print"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

