(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     29969,        889]*)
(*NotebookOutlinePosition[     30959,        921]*)
(*  CellTagsIndexPosition[     30915,        917]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/ccif.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,<oldx,>func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("AVOID USING C, D, E, I, N and O as variables !"\)], "Print"]
}, Closed]],

Cell["\<\
From \"BPS Branes in Supergravity\" by K. S. Stelle, hep-th/9803116 \
- the d=11 magnetic 5-brane\
\>", "Text"],

Cell[CellGroupData[{

Cell["Gamma Matrices", "Subsubsection"],

Cell[BoxData[{
    \(\(gammap\  = \ gammamatrix[0, 5];\)\), "\n", 
    \(\(gammaq\  = \ gammamatrix[1, 5];\)\), "\n", 
    \(\(gamma11\  = \ splitgamma[gammap, gammaq];\)\), "\n", 
    \(\(checkgamma[gamma11, 1, 10];\)\)}], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["SUSY Parameters", "Subsubsection"],

Cell[BoxData[{
    \(\(projp\  = \ 
        IdentityMatrix[4] + gammap[\([6]\)];\)\), "\[IndentingNewLine]", 
    \(\(projq\  = \ IdentityMatrix[8] + gammaq[\([7]\)];\)\)}], "Input"],

Cell[BoxData[
    \(\(Clear[pp1, pp2, pp3, pp4, pq1, pq2, pq3, pq4, pq5, pq6, pq7, 
        pq8];\)\)], "Input"],

Cell[BoxData[{
    \(\(parmp\  = \ 
        Table[{Symbol["\<pp\>" <> ToString[i]]}, {i, 4}];\)\), "\n", 
    \(\(parmq = \ 
        Table[{Symbol["\<pq\>" <> ToString[i]]}, {i, 8}];\)\), "\n", 
    \(\(parm11\  = \ crossmatrix[parmp, parmq];\)\)}], "Input"],

Cell[BoxData[
    \(Union[Flatten[Dot[projq, parmq]]]\)], "Input"],

Cell[BoxData[{
    \(\(pq1\  = \ 0;\)\), "\n", 
    \(\(pq4\  = \ 0;\)\), "\[IndentingNewLine]", 
    \(\(pq6\  = \ 0;\)\), "\[IndentingNewLine]", 
    \(\(pq7\  = \ 0;\)\)}], "Input"]
}, Open  ]],

Cell[BoxData[
    \(\(dim\  = \ 11;\)\)], "Input"],

Cell[BoxData[{
    \(\(ad2\  = \ 
        Table[Table[0, {j, dim}], {i, dim}];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        Do[\((\[IndentingNewLine]inddim\  = \ 
              Join[DeleteCases[Join[Table[l, {l, dim - 1}], {0}], i | j], {i, 
                  j}]; \[IndentingNewLine]ad2[\([i, j]\)]\  = \ 
              Signature[inddim] 
                Factorial[dim - 2];\[IndentingNewLine])\), {j, dim}], {i, 
          dim}];\)\)}], "Input"],

Cell[BoxData[{
    \(\(ad3\  = \ 
        Table[Table[Table[0, {k, dim}], {j, dim}], {i, 
            dim}];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        Do[Do[\((\[IndentingNewLine]inddim\  = \ 
                Join[DeleteCases[Join[Table[l, {l, dim - 1}], {0}], 
                    i | j | k], {i, j, k}]; \[IndentingNewLine]ad3[\([i, j, 
                    k]\)]\  = \ 
                Signature[inddim] 
                  Factorial[dim - 3];\[IndentingNewLine])\), {k, dim}], {j, 
            dim}], {i, dim}];\)\)}], "Input"],

Cell[BoxData[
    \(\(ptahdir\  = \ "\<~/projects/ptah/\>";\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(sphericalmetric[5]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Spherical metric check: "\[InvisibleSpace]True\),
      SequenceForm[ "Spherical metric check: ", True],
      Editable->False]], "Print"],

Cell[BoxData[
    \({{1, 0, 0, 0, 0}, {0, r\^2\ Sin[t1]\^2\ Sin[t2]\^2\ Sin[t3]\^2, 0, 0, 
        0}, {0, 0, r\^2\ Sin[t2]\^2\ Sin[t3]\^2, 0, 0}, {0, 0, 0, 
        r\^2\ Sin[t3]\^2, 0}, {0, 0, 0, 0, r\^2}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(gs\  = \ 
      coordxform[
        sphericalmetric[5], {r, ph, t1, t2, t3}, \[IndentingNewLine]{r, y, 
          ArcCos[s], ArcCos[q], ArcCos[p]}, \[IndentingNewLine]{r, y, s, q, 
          p}]\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Spherical metric check: "\[InvisibleSpace]True\),
      SequenceForm[ "Spherical metric check: ", True],
      Editable->False]], "Print"],

Cell[BoxData[
    \({{1, 0, 0, 0, 
        0}, {0, \(-\((\(-1\) + p)\)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
              q)\)\ r\^2\ \((\(-1\) + s)\)\ \((1 + s)\), 0, 0, 0}, {0, 
        0, \(-\(\(\((\(-1\) + p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
                    q)\)\ r\^2\)\/\(\((\(-1\) + s)\)\ \((1 + s)\)\)\)\), 0, 
        0}, {0, 0, 
        0, \(\((\(-1\) + p)\)\ \((1 + p)\)\ r\^2\)\/\(\((\(-1\) + q)\)\ \((1 \
+ q)\)\), 0}, {0, 0, 0, 
        0, \(-\(r\^2\/\(\((\(-1\) + p)\)\ \((1 + p)\)\)\)\)}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(x\  = \ {r, p, q, s, y, n, u, v, w, z, t};\)\), "\n", 
    \(\(b[r]\  = \ \((k + r^3)\)^\((1/3)\);\)\), "\n", 
    \(\(c[r]\  = \ r;\)\n\), "\n", 
    \(\(g\  = \ IdentityMatrix[11];\)\), "\[IndentingNewLine]", 
    \(\(g[\([1, 1]\)]\  *= \ b[r]^2/c[r]^2;\)\), "\n", 
    \(\(Do[
        g[\([i, i]\)]\  = \ gs[\([7 - i, 7 - i]\)] b[r]^2/c[r]^2, {i, 2, 
          5}];\)\n\), "\n", 
    \(\ \(Do[g[\([i, i]\)] = \ simpler[c[r]/b[r]], {i, 6, 10}];\)\), "\n", 
    \(\(g[\([11, 11]\)]\  = \ \(-g[\([10, 10]\)]\);\)\), "\n", 
    \(\(components[g];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\((k + \
r\^3)\)\^\(2/3\)\/r\^2\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ r, -2], 
          Power[ 
            Plus[ k, 
              Power[ r, 3]], 
            Rational[ 2, 3]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\((k + r\^3)\)\
\^\(2/3\)\/\(\((\(-1\) + p)\)\ \((1 + p)\)\)\)\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ -1, 
          Power[ 
            Plus[ -1, p], -1], 
          Power[ 
            Plus[ 1, p], -1], 
          Power[ 
            Plus[ k, 
              Power[ r, 3]], 
            Rational[ 2, 3]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\((\(-1\) + p)\)\ \
\((1 + p)\)\ \((k + r\^3)\)\^\(2/3\)\)\/\(\((\(-1\) + q)\)\ \((1 + q)\)\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Power[ 
            Plus[ -1, q], -1], 
          Power[ 
            Plus[ 1, q], -1], 
          Power[ 
            Plus[ k, 
              Power[ r, 3]], 
            Rational[ 2, 3]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(\((\(-1\) + 
                    p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
                    q)\)\ \((k + r\^3)\)\^\(2/3\)\)\/\(\((\(-1\) + 
                    s)\)\ \((1 + s)\)\)\)\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ -1, 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Plus[ -1, q], 
          Plus[ 1, q], 
          Power[ 
            Plus[ k, 
              Power[ r, 3]], 
            Rational[ 2, 3]], 
          Power[ 
            Plus[ -1, s], -1], 
          Power[ 
            Plus[ 1, s], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\((\(-1\) + 
                p)\)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
              q)\)\ \((k + r\^3)\)\^\(2/3\)\ \((\(-1\) + s)\)\ \((1 + 
              s)\)\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ -1, 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Plus[ -1, q], 
          Plus[ 1, q], 
          Power[ 
            Plus[ k, 
              Power[ r, 3]], 
            Rational[ 2, 3]], 
          Plus[ -1, s], 
          Plus[ 1, s]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]r\/\((k + r\^3)\)\^\
\(1/3\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ r, 
          Power[ 
            Plus[ k, 
              Power[ r, 3]], 
            Rational[ -1, 3]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]r\/\((k + r\^3)\)\^\
\(1/3\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ r, 
          Power[ 
            Plus[ k, 
              Power[ r, 3]], 
            Rational[ -1, 3]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]r\/\((k + r\^3)\)\^\
\(1/3\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ r, 
          Power[ 
            Plus[ k, 
              Power[ r, 3]], 
            Rational[ -1, 3]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]r\/\((k + r\^3)\)\^\
\(1/3\)\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ r, 
          Power[ 
            Plus[ k, 
              Power[ r, 3]], 
            Rational[ -1, 3]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]r\/\((k + \
r\^3)\)\^\(1/3\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ r, 
          Power[ 
            Plus[ k, 
              Power[ r, 3]], 
            Rational[ -1, 3]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(r\/\((k + \
r\^3)\)\^\(1/3\)\)\)\),
      SequenceForm[ "[", 11, ",", 11, "] = ", 
        Times[ -1, r, 
          Power[ 
            Plus[ k, 
              Power[ r, 3]], 
            Rational[ -1, 3]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell["\<\
We next parameterize the metric so that each component is a \
polynomial:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(gp\  = \ 
        simpler[g /. {\((k + r^3)\) \[Rule] 
                b^3, \((\(-1\) + 
                    p)\) \[Rule] \(-P^2\)/\((1 + p)\), \((\(-1\) + 
                    q)\) \[Rule] \(-Q^2\)/\((1 + q)\), \((\(-1\) + 
                    s)\) \[Rule] \(-S^2\)/\((1 + 
                      s)\)}];\)\), "\[IndentingNewLine]", 
    \(\(components[gp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]b\^2\/r\^2\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, 2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]b\^2\/P\^2\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, 2], 
          Power[ P, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^2\ \
P\^2\)\/Q\^2\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ b, 2], 
          Power[ P, 2], 
          Power[ Q, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^2\ P\^2\ \
Q\^2\)\/S\^2\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ b, 2], 
          Power[ P, 2], 
          Power[ Q, 2], 
          Power[ S, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^2\ P\^2\ Q\^2\ \
S\^2\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ b, 2], 
          Power[ P, 2], 
          Power[ Q, 2], 
          Power[ S, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]r\/b\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ b, -1], r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]r\/b\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Power[ b, -1], r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]r\/b\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ 
          Power[ b, -1], r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]r\/b\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ 
          Power[ b, -1], r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]r\/b\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ 
          Power[ b, -1], r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(r\/b\)\)\),
      SequenceForm[ "[", 11, ",", 11, "] = ", 
        Times[ -1, 
          Power[ b, -1], r]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(ginvp\  = \ simpler[Inverse[gp]];\)\), "\[IndentingNewLine]", 
    \(\(components[ginvp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]r\^2\/b\^2\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]P\^2\/b\^2\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]Q\^2\/\(b\^2\ \
P\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]S\^2\/\(b\^2\ P\^2\ \
Q\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ S, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]1\/\(b\^2\ P\^2\ \
Q\^2\ S\^2\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ S, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]b\/r\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ b, 
          Power[ r, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]b\/r\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ b, 
          Power[ r, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]b\/r\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ b, 
          Power[ r, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]b\/r\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ b, 
          Power[ r, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]b\/r\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ b, 
          Power[ r, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(b\/r\)\)\),
      SequenceForm[ "[", 11, ",", 11, "] = ", 
        Times[ -1, b, 
          Power[ r, -1]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(irmdgp\  = \ 
      Simplify[PowerExpand[1/Sqrt[\(-Simplify[Det[gp]]\)]]]\)], "Input"],

Cell[BoxData[
    \(1\/\(b\^2\ P\^2\ Q\ r\^2\)\)], "Output"]
}, Open  ]],

Cell[BoxData[{
    \(\(e\  = \ 
        Table[Table[0, {j, dim}], {i, dim}];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        e[\([i, i]\)] = simpler[Sqrt[g[\([i, i]\)]]], {i, 
          dim - 1}];\)\), "\[IndentingNewLine]", 
    \(\(e[\([dim, dim]\)]\  = \ 
        simpler[Sqrt[\(-g[\([dim, dim]\)]\)]];\)\), "\[IndentingNewLine]", 
    \(\(vielbein[e];\)\), "\[IndentingNewLine]", 
    \(\(omega\  = \ spinconn[];\)\)}], "Input"],

Cell[BoxData[
    \(\(einv\  = \ currentframe[\([2]\)];\)\)], "Input"],

Cell[CellGroupData[{

Cell["SUSY Parameter Ansatz", "Subsubsection"],

Cell[BoxData[{
    \(\(pp1\  = \ 1;\)\), "\n", 
    \(\(pp2\  = \ 1;\)\), "\[IndentingNewLine]", 
    \(\(pp3\  = \ 1;\)\), "\n", 
    \(\(pp4\  = \ 1;\)\), "\[IndentingNewLine]", 
    \(\(pq2\  = \ s;\)\), "\[IndentingNewLine]", 
    \(\(pq3\  = \ s;\)\), "\[IndentingNewLine]", 
    \(\(pq5\  = \ s;\)\), "\[IndentingNewLine]", 
    \(\(pq8\  = \ s;\)\)}], "Input"]
}, Open  ]],

Cell[BoxData[{
    \(\(ep\  = \ 
        simpler[e /. {\((k + r^3)\) \[Rule] b^3, 
              r \[Rule] 
                c^3, \((\(-1\) + 
                    p)\) \[Rule] \(-P^2\)/\((1 + p)\), \((\(-1\) + 
                    q)\) \[Rule] \(-Q^2\)/\((1 + q)\), \((\(-1\) + 
                    s)\) \[Rule] \(-S^2\)/\((1 + 
                      s)\)}];\)\), "\[IndentingNewLine]", 
    \(\(epinv\  = \ 
        simpler[einv /. {\((k + r^3)\) \[Rule] b^3, 
              r \[Rule] 
                c^3, \((\(-1\) + 
                    p)\) \[Rule] \(-P^2\)/\((1 + p)\), \((\(-1\) + 
                    q)\) \[Rule] \(-Q^2\)/\((1 + q)\), \((\(-1\) + 
                    s)\) \[Rule] \(-S^2\)/\((1 + s)\)}];\)\)}], "Input"],

Cell[BoxData[
    \(\(components[ep];\)\)], "Input"],

Cell[BoxData[
    \(\(components[epinv];\)\)], "Input"],

Cell[BoxData[
    \(\(omegap\  = \ 
        simpler[omega /. {\((k + r^3)\) \[Rule] b^3, 
              r \[Rule] 
                c^3, \((\(-1\) + 
                    p)\) \[Rule] \(-P^2\)/\((1 + p)\), \((\(-1\) + 
                    q)\) \[Rule] \(-Q^2\)/\((1 + q)\), \((\(-1\) + 
                    s)\) \[Rule] \(-S^2\)/\((1 + s)\)}];\)\)], "Input"],

Cell[BoxData[
    \(\(components[omegap];\)\)], "Input"],

Cell[BoxData[
    \(\(components[parm11];\)\)], "Input"],

Cell[CellGroupData[{

Cell["Gauge Field Strength Ansatz", "Subsubsection"],

Cell["We might also express f as d[a Abcd]/6...", "Text"],

Cell[BoxData[
    \(\(f[r_]\  := \ 3  k/\((2  r^5)\);\)\)], "Input"],

Cell[BoxData[
    \(\(eps5\  = \ epsilon[5, {1, 2, 3, 4, 5}];\)\)], "Input"],

Cell[BoxData[
    \(\(fsp\  = \ 
        Table[Table[Table[Table[0, {i, 11}], {j, 11}], {kk, 11}], {l, 
            11}];\)\)], "Input"],

Cell[BoxData[
    \(\(fslp\  = \ 
        Table[Table[Table[Table[0, {i, 11}], {j, 11}], {kk, 11}], {l, 
            11}];\)\)], "Input"],

Cell[BoxData[
    \(\(Do[
        Do[Do[Do[
              fsp[\([i, j, kk, l]\)]\  = \ 
                Sum[Sum[\[IndentingNewLine]eps5[\([i, j, kk, l, 
                          m]\)] \((f/\((b[r]/c[r])\)^10)\) gp[\([m, n]\)] 
                      x[\([n]\)], \[IndentingNewLine]{n, 5}], {m, 5}], {l, 
                5}], {kk, 5}], {j, 5}], {i, 5}];\)\)], "Input"],

Cell[BoxData[
    \(\(Do[
        Do[Do[Do[
              fslp[\([i, j, kk, l]\)]\  = \ 
                Sum[Sum[
                    Sum[Sum[\[IndentingNewLine]fsp[\([m, n, pp, qq]\)] 
                          gp[\([i, m]\)] gp[\([j, n]\)] gp[\([kk, pp]\)] 
                          gp[\([l, qq]\)], \[IndentingNewLine]{qq, 5}], {pp, 
                        5}], {n, 5}], {m, 5}], {l, 5}], {kk, 5}], {j, 
            5}], {i, 5}];\)\)], "Input"],

Cell[BoxData[
    \(fsp[\([1, 2, 3, 4]\)]\)], "Input"],

Cell[BoxData[
    \(fslp[\([1, 2, 3, 4]\)]\)], "Input"]
}, Open  ]],

Cell["\<\
Don't forget the sub[i] in the derivatives (takes care of change in \
time coord from i=10 to i=0 in c++)\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(btab\  = \ {b^3, b[r]^3, 
        Table[D[b[r]^3, x[\([sub[i]]\)]], {i, dim}]/\((3 
               b^2)\)}\), "\[IndentingNewLine]", 
    \(Ptab\  = \ {P^2, 1 - p^2, 
        Table[D[1 - p^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              P)\)}\), "\[IndentingNewLine]", 
    \(Qtab\  = \ {Q^2, 1 - q^2, 
        Table[D[1 - q^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              Q)\)}\), "\[IndentingNewLine]", 
    \(Stab\  = \ {S^2, 1 - s^2, 
        Table[D[1 - s^2, x[\([sub[i]]\)]], {i, dim}]/\((2  S)\)}\)}], "Input"],

Cell[BoxData[
    \({b\^3, 
      k + r\^3, {0, r\^2\/b\^2, 0, 0, 0, 0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({P\^2, 
      1 - p\^2, {0, 0, \(-\(p\/P\)\), 0, 0, 0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({Q\^2, 
      1 - q\^2, {0, 0, 0, \(-\(q\/Q\)\), 0, 0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({S\^2, 
      1 - s\^2, {0, 0, 0, 0, \(-\(s\/S\)\), 0, 0, 0, 0, 0, 0}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(dim = 11;\)\), "\[IndentingNewLine]", 
    \(coords\  = \ x\), "\[IndentingNewLine]", 
    \(ints\  = \ {2, 3}\), "\n", 
    \(constants\  = \ {I, k}\), "\n", 
    \(functions\  = \ {btab, Ptab, Qtab, Stab}\), "\[IndentingNewLine]", 
    \(ansatze\  = \ {}\)}], "Input"],

Cell[BoxData[
    \({r, p, q, s, y, n, u, v, w, z, t}\)], "Output"],

Cell[BoxData[
    \({2, 3}\)], "Output"],

Cell[BoxData[
    \({\[ImaginaryI], k}\)], "Output"],

Cell[BoxData[
    \({{b\^3, k + r\^3, {0, r\^2\/b\^2, 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {P\^2, 
        1 - p\^2, {0, 0, \(-\(p\/P\)\), 0, 0, 0, 0, 0, 0, 0, 0}}, {Q\^2, 
        1 - q\^2, {0, 0, 0, \(-\(q\/Q\)\), 0, 0, 0, 0, 0, 0, 0}}, {S\^2, 
        1 - s\^2, {0, 0, 0, 0, \(-\(s\/S\)\), 0, 0, 0, 0, 0, 0}}}\)], "Output"],

Cell[BoxData[
    \({}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(headers[];\)\)], "Input"],

Cell[BoxData[
    \("functions completed..."\)], "Print"],

Cell[BoxData[
    \("g completed..."\)], "Print"],

Cell[BoxData[
    \("gi completed..."\)], "Print"],

Cell[BoxData[
    \("d completed..."\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(tensor[ad2, "\<ad2\>"];\)\), "\[IndentingNewLine]", 
    \(\(tensor[ad3, "\<ad3\>"];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("ad2"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ad2", " completed..."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("ad3"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ad3", " completed..."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[BoxData[{
    \(\(tensor[ep, "\<viel\>"];\)\), "\[IndentingNewLine]", 
    \(\(tensor[epinv, "\<vielinv\>"];\)\), "\[IndentingNewLine]", 
    \(\(tensor[omegap, "\<omega\>"];\)\), "\[IndentingNewLine]", 
    \(\(tensor[Flatten[parm11], "\<susyparm\>"];\)\), "\[IndentingNewLine]", 
    \(\(tensor[fsp, "\<fsuuuu\>"];\)\)}], "Input"],

Cell[BoxData[
    \(\(tensor[fslp, "\<fs\>"];\)\)], "Input"],

Cell["\<\
Run the following cell AND MAKEGAMMAS whenever the brane number \
(and hence the gamma split) changes\
\>", "Text"],

Cell[BoxData[
    \(\(gammaout[gamma11, "\<gamma\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(backup[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]17\[InvisibleSpace]":"\
\[InvisibleSpace]46\[InvisibleSpace]":"\[InvisibleSpace]7\[InvisibleSpace]" \
on "\[InvisibleSpace]5\[InvisibleSpace]"/"\[InvisibleSpace]6\[InvisibleSpace]\
"/"\[InvisibleSpace]2003\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 17, ":", 46, ":", 7, " on ", 5, "/", 6, "/", 2003, "."],
      Editable->False]], "Print"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "ken", \
"projects", "mathematica", "packages"}, "d11brane.nb.ps", CharacterEncoding \
-> "ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 70, 1, 27, "Input"],
Cell[1874, 57, 2984, 44, 375, "Print"],
Cell[4861, 103, 81, 1, 23, "Print"]
}, Closed]],
Cell[4957, 107, 121, 3, 29, "Text"],

Cell[CellGroupData[{
Cell[5103, 114, 39, 0, 42, "Subsubsection"],
Cell[5145, 116, 232, 4, 75, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[5414, 125, 40, 0, 42, "Subsubsection"],
Cell[5457, 127, 182, 3, 43, "Input"],
Cell[5642, 132, 112, 2, 27, "Input"],
Cell[5757, 136, 258, 5, 59, "Input"],
Cell[6018, 143, 66, 1, 27, "Input"],
Cell[6087, 146, 184, 4, 75, "Input"]
}, Open  ]],
Cell[6286, 153, 50, 1, 27, "Input"],
Cell[6339, 156, 451, 9, 91, "Input"],
Cell[6793, 167, 536, 11, 91, "Input"],
Cell[7332, 180, 74, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[7431, 185, 51, 1, 27, "Input"],
Cell[7485, 188, 177, 3, 23, "Print"],
Cell[7665, 193, 221, 3, 47, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[7923, 201, 227, 5, 59, "Input"],
Cell[8153, 208, 177, 3, 23, "Print"],
Cell[8333, 213, 536, 9, 116, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8906, 227, 584, 11, 187, "Input"],
Cell[9493, 240, 396, 11, 42, "Print"],
Cell[9892, 253, 511, 14, 45, "Print"],
Cell[10406, 269, 582, 16, 45, "Print"],
Cell[10991, 287, 736, 20, 44, "Print"],
Cell[11730, 309, 662, 18, 26, "Print"],
Cell[12395, 329, 371, 10, 40, "Print"],
Cell[12769, 341, 371, 10, 40, "Print"],
Cell[13143, 353, 371, 10, 40, "Print"],
Cell[13517, 365, 371, 10, 40, "Print"],
Cell[13891, 377, 375, 10, 40, "Print"],
Cell[14269, 389, 388, 10, 40, "Print"]
}, Open  ]],
Cell[14672, 402, 97, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[14794, 409, 396, 8, 75, "Input"],
Cell[15193, 419, 299, 7, 41, "Print"],
Cell[15495, 428, 299, 7, 41, "Print"],
Cell[15797, 437, 336, 9, 42, "Print"],
Cell[16136, 448, 367, 10, 41, "Print"],
Cell[16506, 460, 366, 10, 25, "Print"],
Cell[16875, 472, 271, 6, 37, "Print"],
Cell[17149, 480, 271, 6, 37, "Print"],
Cell[17423, 488, 271, 6, 37, "Print"],
Cell[17697, 496, 271, 6, 37, "Print"],
Cell[17971, 504, 275, 6, 37, "Print"],
Cell[18249, 512, 288, 6, 37, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[18574, 523, 128, 2, 43, "Input"],
Cell[18705, 527, 299, 7, 41, "Print"],
Cell[19007, 536, 299, 7, 41, "Print"],
Cell[19309, 545, 337, 9, 41, "Print"],
Cell[19649, 556, 369, 10, 42, "Print"],
Cell[20021, 568, 373, 10, 41, "Print"],
Cell[20397, 580, 271, 6, 40, "Print"],
Cell[20671, 588, 271, 6, 40, "Print"],
Cell[20945, 596, 271, 6, 40, "Print"],
Cell[21219, 604, 271, 6, 40, "Print"],
Cell[21493, 612, 275, 6, 40, "Print"],
Cell[21771, 620, 288, 6, 40, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[22096, 631, 106, 2, 27, "Input"],
Cell[22205, 635, 60, 1, 45, "Output"]
}, Open  ]],
Cell[22280, 639, 433, 9, 91, "Input"],
Cell[22716, 650, 70, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[22811, 655, 46, 0, 42, "Subsubsection"],
Cell[22860, 657, 367, 8, 139, "Input"]
}, Open  ]],
Cell[23242, 668, 728, 15, 107, "Input"],
Cell[23973, 685, 52, 1, 27, "Input"],
Cell[24028, 688, 55, 1, 27, "Input"],
Cell[24086, 691, 356, 7, 59, "Input"],
Cell[24445, 700, 56, 1, 27, "Input"],
Cell[24504, 703, 56, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[24585, 708, 52, 0, 42, "Subsubsection"],
Cell[24640, 710, 57, 0, 32, "Text"],
Cell[24700, 712, 68, 1, 27, "Input"],
Cell[24771, 715, 76, 1, 27, "Input"],
Cell[24850, 718, 136, 3, 27, "Input"],
Cell[24989, 723, 137, 3, 27, "Input"],
Cell[25129, 728, 367, 7, 59, "Input"],
Cell[25499, 737, 449, 9, 59, "Input"],
Cell[25951, 748, 54, 1, 27, "Input"],
Cell[26008, 751, 55, 1, 27, "Input"]
}, Open  ]],
Cell[26078, 755, 129, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[26232, 762, 546, 11, 75, "Input"],
Cell[26781, 775, 101, 2, 45, "Output"],
Cell[26885, 779, 104, 2, 41, "Output"],
Cell[26992, 783, 104, 2, 42, "Output"],
Cell[27099, 787, 104, 2, 41, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[27240, 794, 293, 6, 107, "Input"],
Cell[27536, 802, 67, 1, 27, "Output"],
Cell[27606, 805, 40, 1, 27, "Output"],
Cell[27649, 808, 52, 1, 27, "Output"],
Cell[27704, 811, 318, 4, 79, "Output"],
Cell[28025, 817, 36, 1, 27, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[28098, 823, 47, 1, 27, "Input"],
Cell[28148, 826, 57, 1, 23, "Print"],
Cell[28208, 829, 49, 1, 23, "Print"],
Cell[28260, 832, 50, 1, 23, "Print"],
Cell[28313, 835, 49, 1, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[28399, 841, 123, 2, 43, "Input"],
Cell[28525, 845, 157, 3, 23, "Print"],
Cell[28685, 850, 157, 3, 23, "Print"]
}, Open  ]],
Cell[28857, 856, 337, 5, 91, "Input"],
Cell[29197, 863, 60, 1, 27, "Input"],
Cell[29260, 866, 125, 3, 32, "Text"],
Cell[29388, 871, 68, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[29481, 876, 46, 1, 27, "Input"],
Cell[29530, 879, 423, 7, 23, "Print"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

