(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     30239,        935]*)
(*NotebookOutlinePosition[     31226,        967]*)
(*  CellTagsIndexPosition[     31182,        963]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/ccif.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,<oldx,>func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("AVOID USING C, D, E, I, N and O as variables !"\)], "Print"]
}, Closed]],

Cell[BoxData[{
    \(\(f\  = \ Table[0, {i, 50}];\)\), "\[IndentingNewLine]", 
    \(\(derivs\  = \ 
        Table[Table[0, {i, 1000}], {j, 50}];\)\), "\[IndentingNewLine]", 
    \(\(g\  = Table[Table[0, {i, 11}], {j, 11}]\ ;\)\), "\n", 
    \(\(ginv\  = \ 
        Table[Table[0, {i, 11}], {j, 11}];\)\), "\[IndentingNewLine]", 
    \(\(Euler\  = \ 0;\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< /home/ken/projects/ptah/ptah.out\)], "Input"],

Cell[BoxData[
    InterpretationBox[\(0\[InvisibleSpace]1\),
      SequenceForm[ 0, 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(1\[InvisibleSpace]2\),
      SequenceForm[ 1, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(2\[InvisibleSpace]2\),
      SequenceForm[ 2, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(3\[InvisibleSpace]2\),
      SequenceForm[ 3, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(4\[InvisibleSpace]2\),
      SequenceForm[ 4, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(5\[InvisibleSpace]2\),
      SequenceForm[ 5, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(6\[InvisibleSpace]2\),
      SequenceForm[ 6, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(7\[InvisibleSpace]2\),
      SequenceForm[ 7, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(8\[InvisibleSpace]2\),
      SequenceForm[ 8, 2],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[f];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(a\^2\ b\^2 + r\^2\)\),
      SequenceForm[ "[", 1, "] = ", 
        Plus[ 
          Times[ 
            Power[ a, 2], 
            Power[ b, 2]], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(\(-m\) + a\^2\ r\^6 + r\^8\)\),
      SequenceForm[ "[", 2, "] = ", 
        Plus[ 
          Times[ -1, m], 
          Times[ 
            Power[ a, 2], 
            Power[ r, 6]], 
          Power[ r, 8]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - b\^2\)\),
      SequenceForm[ "[", 3, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ b, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - p\^2\)\),
      SequenceForm[ "[", 4, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ p, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - q\^2\)\),
      SequenceForm[ "[", 5, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ q, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - s\^2\)\),
      SequenceForm[ "[", 6, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ s, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - u\^2\)\),
      SequenceForm[ "[", 7, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ u, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - v\^2\)\),
      SequenceForm[ "[", 8, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ v, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - w\^2\)\),
      SequenceForm[ "[", 9, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ w, 2]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[derivs];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ r\)\),
      SequenceForm[ "[", 1, ",", 9, "] = ", 
        Times[ 2, r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ a\^2\ b\)\),
      SequenceForm[ "[", 1, ",", 10, "] = ", 
        Times[ 2, 
          Power[ a, 2], b]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]20\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(3\ a\^2\ \
r\^5\)\/G + \(4\ r\^7\)\/G\)\),
      SequenceForm[ "[", 2, ",", 20, "] = ", 
        Plus[ 
          Times[ 3, 
            Power[ a, 2], 
            Power[ G, -1], 
            Power[ r, 5]], 
          Times[ 4, 
            Power[ G, -1], 
            Power[ r, 7]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]32\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(b\/B\)\)\),
      SequenceForm[ "[", 3, ",", 32, "] = ", 
        Times[ -1, b, 
          Power[ B, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]35\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(p\/P\)\)\),
      SequenceForm[ "[", 4, ",", 35, "] = ", 
        Times[ -1, p, 
          Power[ P, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]47\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(q\/Q\)\)\),
      SequenceForm[ "[", 5, ",", 47, "] = ", 
        Times[ -1, q, 
          Power[ Q, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]59\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(s\/S\)\)\),
      SequenceForm[ "[", 6, ",", 59, "] = ", 
        Times[ -1, s, 
          Power[ S, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]71\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(u\/U\)\)\),
      SequenceForm[ "[", 7, ",", 71, "] = ", 
        Times[ -1, u, 
          Power[ U, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]83\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(v\/V\)\)\),
      SequenceForm[ "[", 8, ",", 83, "] = ", 
        Times[ -1, v, 
          Power[ V, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]95\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(w\/W\)\)\),
      SequenceForm[ "[", 9, ",", 95, "] = ", 
        Times[ -1, w, 
          Power[ W, -1]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[g];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\/P\^2 - \
\(B\^2\ r\^2\)\/P\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, -2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, -2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ r\^2\)\/Q\
\^2 - \(B\^2\ P\^2\ r\^2\)\/Q\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, -2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, -2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ Q\^2\ \
r\^2\)\/S\^2 - \(B\^2\ P\^2\ Q\^2\ r\^2\)\/S\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, -2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, -2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ Q\^2\ \
r\^2\ S\^2\)\/U\^2 - \(B\^2\ P\^2\ Q\^2\ r\^2\ S\^2\)\/U\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, -2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, -2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ Q\^2\ \
r\^2\ S\^2\ U\^2\)\/V\^2 - \(B\^2\ P\^2\ Q\^2\ r\^2\ S\^2\ U\^2\)\/V\^2\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, 2], 
            Power[ V, -2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, 2], 
            Power[ V, -2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ Q\^2\ \
r\^2\ S\^2\ U\^2\ V\^2\)\/W\^2 - \(B\^2\ P\^2\ Q\^2\ r\^2\ S\^2\ U\^2\ V\^2\)\
\/W\^2\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, 2], 
            Power[ V, 2], 
            Power[ W, -2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, 2], 
            Power[ V, 2], 
            Power[ W, -2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(P\^2\ Q\^2\ r\^2\ \
S\^2\ U\^2\ V\^2\ W\^2 - B\^2\ P\^2\ Q\^2\ r\^2\ S\^2\ U\^2\ V\^2\ W\^2\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, 2], 
            Power[ V, 2], 
            Power[ W, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, 2], 
            Power[ V, 2], 
            Power[ W, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(F\ r\^6\)\/G\^2\),
      
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ F, 
          Power[ G, -2], 
          Power[ r, 6]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]F\/B\^2\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ 
          Power[ B, -2], F]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(a\^4\ \
B\^2\)\/F - \(a\^4\ B\^4\)\/F + \(a\^2\ B\^4\ m\)\/\(F\ r\^6\) + \(2\ a\^2\ B\
\^2\ r\^2\)\/F - \(a\^2\ B\^4\ r\^2\)\/F + \(B\^2\ r\^4\)\/F\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Plus[ 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1]], 
          Times[ -1, 
            Power[ a, 4], 
            Power[ B, 4], 
            Power[ F, -1]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], m, 
            Power[ r, -6]], 
          Times[ 2, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ \
m\)\/\(F\ r\^6\)\),
      SequenceForm[ "[", 10, ",", 11, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, 
          Power[ r, -6]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ \
m\)\/\(F\ r\^6\)\),
      SequenceForm[ "[", 11, ",", 10, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, 
          Power[ r, -6]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(a\^2\/F\)\) \
+ \(a\^2\ B\^2\)\/F + m\/\(F\ r\^6\) - r\^2\/F\)\),
      SequenceForm[ "[", 11, ",", 11, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 2], 
            Power[ F, -1]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1]], 
          Times[ 
            Power[ F, -1], m, 
            Power[ r, -6]], 
          Times[ -1, 
            Power[ F, -1], 
            Power[ r, 2]]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[ginv];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]P\^2\/\(b\^2\ \
r\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, 2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]Q\^2\/\(b\^2\ P\^2\ \
r\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, 2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]S\^2\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]U\^2\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\ S\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, -2], 
          Power[ U, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]V\^2\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\ S\^2\ U\^2\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, -2], 
          Power[ U, -2], 
          Power[ V, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]W\^2\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\ S\^2\ U\^2\ V\^2\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, -2], 
          Power[ U, -2], 
          Power[ V, -2], 
          Power[ W, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]1\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\ S\^2\ U\^2\ V\^2\ W\^2\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, -2], 
          Power[ U, -2], 
          Power[ V, -2], 
          Power[ W, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]G\^2\/\(F\ r\^6\)\),
      
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ 
          Power[ F, -1], 
          Power[ G, 2], 
          Power[ r, -6]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]B\^2\/F\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ 
          Power[ B, 2], 
          Power[ F, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(m\/\(B\^2\ \
F\ G\^2\)\)\) - \(a\^2\ r\^6\)\/\(F\ G\^2\) + \(a\^2\ r\^6\)\/\(B\^2\ F\ G\^2\
\) + r\^8\/\(B\^2\ F\ G\^2\)\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], m], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 6]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 6]], 
          Times[ 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 8]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\)\/\(F\ \
G\^2\)\),
      SequenceForm[ "[", 10, ",", 11, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\)\/\(F\ \
G\^2\)\),
      SequenceForm[ "[", 11, ",", 10, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(\(a\^2\ \
B\^2\ m\)\/\(F\ G\^2\)\)\) - \(a\^4\ r\^6\)\/\(F\ G\^2\) + \(a\^4\ B\^2\ r\^6\
\)\/\(F\ G\^2\) - \(2\ a\^2\ r\^8\)\/\(F\ G\^2\) + \(a\^2\ B\^2\ r\^8\)\/\(F\ \
G\^2\) - r\^10\/\(F\ G\^2\)\)\),
      SequenceForm[ "[", 11, ",", 11, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], m], 
          Times[ -1, 
            Power[ a, 4], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 6]], 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 6]], 
          Times[ -2, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 8]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 8]], 
          Times[ -1, 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 10]]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(isrdmg\)], "Input"],

Cell[BoxData[
    \(1\/\(b\^7\ F\ P\^5\ Q\^4\ r\^7\ S\^3\ U\^2\ V\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[sc]]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riesq]]\)], "Input"],

Cell[BoxData[
    \(\(24\ m\^2\ \((98\ a\^8\ b\^8 + 465\ a\^6\ b\^6\ r\^2 + 858\ a\^4\ b\^4\
\ r\^4 + 745\ a\^2\ b\^2\ r\^6 + 270\ r\^8)\)\)\/\(r\^16\ \((a\^2\ b\^2 + \
r\^2)\)\^6\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riecub1]]\)], "Input"],

Cell[BoxData[
    \(\(\(1\/\(r\^24\ \((a\^2\ b\^2 + r\^2)\)\^9\)\)\((24\ m\^3\ \((2968\ \
a\^12\ b\^12 + 21777\ a\^10\ b\^10\ r\^2 + 68160\ a\^8\ b\^8\ r\^4 + 
            116886\ a\^6\ b\^6\ r\^6 + 116228\ a\^4\ b\^4\ r\^8 + 
            63745\ a\^2\ b\^2\ r\^10 + 15180\ r\^12)\))\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riecub2]]\)], "Input"],

Cell[BoxData[
    \(\(\(1\/\(r\^24\ \((a\^2\ b\^2 + r\^2)\)\^9\)\)\((6\ m\^3\ \((1736\ \
a\^12\ b\^12 + 11865\ a\^10\ b\^10\ r\^2 + 34008\ a\^8\ b\^8\ r\^4 + 
            52710\ a\^6\ b\^6\ r\^6 + 47572\ a\^4\ b\^4\ r\^8 + 
            24905\ a\^2\ b\^2\ r\^10 + 6420\ r\^12)\))\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riecub3]]\)], "Input"],

Cell[BoxData[
    \(\(-\(\(1\/\(r\^24\ \((a\^2\ b\^2 + r\^2)\)\^9\)\)\((48\ m\^3\ \((154\ a\
\^12\ b\^12 + 1239\ a\^10\ b\^10\ r\^2 + 4269\ a\^8\ b\^8\ r\^4 + 
              8022\ a\^6\ b\^6\ r\^6 + 8582\ a\^4\ b\^4\ r\^8 + 
              4855\ a\^2\ b\^2\ r\^10 + 1095\ r\^12)\))\)\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Cdriesq1]]\)], "Input"],

Cell[BoxData[
    \(\(\(1\/\(r\^24\ \((a\^2\ b\^2 + r\^2)\)\^9\)\)\((192\ m\^2\ \((\(-980\)\
\ a\^12\ b\^12\ m - 6832\ a\^10\ b\^10\ m\ r\^2 - 
            20213\ a\^8\ b\^8\ m\ r\^4 + 980\ a\^14\ b\^12\ r\^6 - 
            32704\ a\^6\ b\^6\ m\ r\^6 + 6930\ a\^12\ b\^10\ r\^8 + 
            882\ a\^12\ b\^12\ r\^8 - 30863\ a\^4\ b\^4\ m\ r\^8 + 
            20727\ a\^10\ b\^8\ r\^10 + 6318\ a\^10\ b\^10\ r\^10 - 
            16450\ a\^2\ b\^2\ m\ r\^10 + 33747\ a\^8\ b\^6\ r\^12 + 
            19170\ a\^8\ b\^8\ r\^12 - 4050\ m\ r\^12 + 
            31815\ a\^6\ b\^4\ r\^14 + 31752\ a\^6\ b\^6\ r\^14 + 
            16695\ a\^4\ b\^2\ r\^16 + 30618\ a\^4\ b\^4\ r\^16 + 
            3850\ a\^2\ r\^18 + 16650\ a\^2\ b\^2\ r\^18 + 
            4050\ r\^20)\))\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[RstarR]]\)], "Input"],

Cell[BoxData[
    \(\(12641461862400\ m\^2\ \((98\ a\^8\ b\^8 + 465\ a\^6\ b\^6\ r\^2 + 858\
\ a\^4\ b\^4\ r\^4 + 745\ a\^2\ b\^2\ r\^6 + 270\ r\^8)\)\)\/\(r\^16\ \((a\^2\
\ b\^2 + r\^2)\)\^6\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[gRstarR]]\)], "Input"],

Cell[BoxData[
    \(\(156067430400\ m\^2\ \((98\ a\^8\ b\^8 + 465\ a\^6\ b\^6\ r\^2 + 858\ \
a\^4\ b\^4\ r\^4 + 745\ a\^2\ b\^2\ r\^6 + 270\ r\^8)\)\)\/\(r\^16\ \((a\^2\ \
b\^2 + r\^2)\)\^6\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(backup[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]3\[InvisibleSpace]":"\
\[InvisibleSpace]16\[InvisibleSpace]":"\[InvisibleSpace]5\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]9\[InvisibleSpace]\
"/"\[InvisibleSpace]2003\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 3, ":", 16, ":", 5, " on ", 4, "/", 9, "/", 2003, "."],
      Editable->False]], "Print"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "ken", \
"projects", "ptah", "mathematica"}, "dbraneval.nb.ps", CharacterEncoding -> \
"ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 70, 1, 27, "Input"],
Cell[1874, 57, 2984, 44, 375, "Print"],
Cell[4861, 103, 81, 1, 23, "Print"]
}, Closed]],
Cell[4957, 107, 368, 7, 88, "Input"],

Cell[CellGroupData[{
Cell[5350, 118, 68, 1, 27, "Input"],
Cell[5421, 121, 121, 3, 23, "Print"],
Cell[5545, 126, 121, 3, 23, "Print"],
Cell[5669, 131, 121, 3, 23, "Print"],
Cell[5793, 136, 121, 3, 23, "Print"],
Cell[5917, 141, 121, 3, 23, "Print"],
Cell[6041, 146, 121, 3, 23, "Print"],
Cell[6165, 151, 121, 3, 23, "Print"],
Cell[6289, 156, 121, 3, 23, "Print"],
Cell[6413, 161, 121, 3, 23, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[6571, 169, 51, 1, 24, "Input"],
Cell[6625, 172, 310, 9, 25, "Print"],
Cell[6938, 183, 345, 10, 25, "Print"],
Cell[7286, 195, 256, 7, 25, "Print"],
Cell[7545, 204, 256, 7, 25, "Print"],
Cell[7804, 213, 256, 7, 25, "Print"],
Cell[8063, 222, 256, 7, 25, "Print"],
Cell[8322, 231, 256, 7, 25, "Print"],
Cell[8581, 240, 256, 7, 25, "Print"],
Cell[8840, 249, 256, 7, 25, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[9133, 261, 56, 1, 27, "Input"],
Cell[9192, 264, 252, 5, 23, "Print"],
Cell[9447, 271, 285, 6, 25, "Print"],
Cell[9735, 279, 463, 13, 41, "Print"],
Cell[10201, 294, 286, 6, 40, "Print"],
Cell[10490, 302, 286, 6, 37, "Print"],
Cell[10779, 310, 286, 6, 38, "Print"],
Cell[11068, 318, 286, 6, 37, "Print"],
Cell[11357, 326, 286, 6, 37, "Print"],
Cell[11646, 334, 286, 6, 37, "Print"],
Cell[11935, 342, 286, 6, 37, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[12258, 353, 51, 1, 24, "Input"],
Cell[12312, 356, 455, 13, 41, "Print"],
Cell[12770, 371, 525, 15, 42, "Print"],
Cell[13298, 388, 591, 17, 41, "Print"],
Cell[13892, 407, 657, 19, 41, "Print"],
Cell[14552, 428, 723, 21, 41, "Print"],
Cell[15278, 451, 791, 24, 41, "Print"],
Cell[16072, 477, 779, 23, 25, "Print"],
Cell[16854, 502, 316, 8, 41, "Print"],
Cell[17173, 512, 274, 6, 39, "Print"],
Cell[17450, 520, 1096, 34, 41, "Print"],
Cell[18549, 556, 354, 9, 41, "Print"],
Cell[18906, 567, 354, 9, 41, "Print"],
Cell[19263, 578, 645, 19, 41, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[19945, 602, 54, 1, 24, "Input"],
Cell[20002, 605, 337, 9, 41, "Print"],
Cell[20342, 616, 369, 10, 41, "Print"],
Cell[20714, 628, 401, 11, 42, "Print"],
Cell[21118, 641, 433, 12, 42, "Print"],
Cell[21554, 655, 465, 13, 42, "Print"],
Cell[22022, 670, 497, 14, 42, "Print"],
Cell[22522, 686, 501, 14, 41, "Print"],
Cell[23026, 702, 339, 9, 41, "Print"],
Cell[23368, 713, 296, 7, 41, "Print"],
Cell[23667, 722, 900, 27, 41, "Print"],
Cell[24570, 751, 323, 8, 37, "Print"],
Cell[24896, 761, 323, 8, 37, "Print"],
Cell[25222, 771, 1267, 39, 41, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[26526, 815, 39, 1, 24, "Input"],
Cell[26568, 818, 81, 1, 45, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[26686, 824, 49, 1, 27, "Input"],
Cell[26738, 827, 35, 1, 27, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[26810, 833, 52, 1, 27, "Input"],
Cell[26865, 836, 195, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[27097, 844, 54, 1, 27, "Input"],
Cell[27154, 847, 300, 4, 85, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[27491, 856, 54, 1, 27, "Input"],
Cell[27548, 859, 296, 4, 67, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[27881, 868, 54, 1, 27, "Input"],
Cell[27938, 871, 300, 4, 67, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[28275, 880, 55, 1, 27, "Input"],
Cell[28333, 883, 784, 12, 121, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[29154, 900, 53, 1, 27, "Input"],
Cell[29210, 903, 207, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[29454, 911, 54, 1, 27, "Input"],
Cell[29511, 914, 205, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[29753, 922, 46, 1, 27, "Input"],
Cell[29802, 925, 421, 7, 23, "Print"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

