(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     46445,       1389]*)
(*NotebookOutlinePosition[     47435,       1421]*)
(*  CellTagsIndexPosition[     47391,       1417]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/ccif.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,<oldx,>func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("AVOID USING C, D, E, I, N and O as variables !"\)], "Print"]
}, Open  ]],

Cell[BoxData[
    \(\(dim\  = \ 11;\)\)], "Input"],

Cell[BoxData[{
    \(\(ad2\  = \ 
        Table[Table[0, {j, dim}], {i, dim}];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        Do[\((\[IndentingNewLine]inddim\  = \ 
              Join[DeleteCases[Join[Table[l, {l, dim - 1}], {0}], i | j], {i, 
                  j}]; \[IndentingNewLine]ad2[\([i, j]\)]\  = \ 
              Signature[inddim] 
                Factorial[dim - 2];\[IndentingNewLine])\), {j, dim}], {i, 
          dim}];\)\)}], "Input"],

Cell[BoxData[{
    \(\(ad3\  = \ 
        Table[Table[Table[0, {k, dim}], {j, dim}], {i, 
            dim}];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        Do[Do[\((\[IndentingNewLine]inddim\  = \ 
                Join[DeleteCases[Join[Table[l, {l, dim - 1}], {0}], 
                    i | j | k], {i, j, k}]; \[IndentingNewLine]ad3[\([i, j, 
                    k]\)]\  = \ 
                Signature[inddim] 
                  Factorial[dim - 3];\[IndentingNewLine])\), {k, dim}], {j, 
            dim}], {i, dim}];\)\)}], "Input"],

Cell[BoxData[
    \(\(ptahdir\  = \ "\<~/projects/ptah/\>";\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(xp\  = \ 
        Flatten[{Table[Symbol["\<t\>" <> ToString[i]], {i, dim - 5}], 
            If[dim \[GreaterEqual] 5, p2, {}], r, th, ph, 
            t}];\)\), "\[IndentingNewLine]", 
    \(\(rho\  = \ Sqrt[r^2 + a^2  Cos[th]^2];\)\), "\[IndentingNewLine]", 
    \(\(delta\  = \ 
        m/\((r^\((dim - 5)\) rho^2)\);\)\), "\[IndentingNewLine]", 
    \(\(psi\  = \ 
        r^\((dim - 5)\) 
          rho^2/\((r^\((dim - 5)\) \((r^2 + a^2)\) - 
                m)\);\)\), "\[IndentingNewLine]", 
    \(\(gp\  = \ 0*IdentityMatrix[dim];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        gp[\([i, i]\)]\  = \ 
          r^2  Cos[th]^2\ Product[Sin[xp[\([j - 1]\)]]^2, {j, 2, i}], {i, 1, 
          dim - 4}];\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim - 3, dim - 3]\)]\  = \ psi;\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim - 2, dim - 2]\)]\  = \ rho^2;\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim - 1, dim - 1]\)]\  = \ \((r^2 + a^2)\) Sin[th]^2 + 
          delta*a^2  Sin[th]^4;\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim - 1, dim]\)]\  = \ 
        delta*a*Sin[th]^2;\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim, dim - 1]\)]\  = \ 
        gp[\([dim - 1, dim]\)];\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim, dim]\)]\  = \ \(-1\) + delta;\)\), "\[IndentingNewLine]", 
    \(\(gp\  = \ simpler[gp];\)\), "\[IndentingNewLine]", 
    \(\(Print["\<dim = \>", dim, "\<, x = \>", 
        xp, "\<, metric components are:\>"];\)\), "\[IndentingNewLine]", 
    \(\(components[gp];\)\), "\[IndentingNewLine]", 
    \(\(If[dim \[Equal] 4, 
        x = {r, b, y, t}, \[IndentingNewLine]x\  = \ 
          Flatten[{Take[{p, q, s, u, v, w}, dim - 5], 
              If[dim \[GreaterEqual] 5, z, {}], r, b, y, 
              t}]];\)\), "\[IndentingNewLine]", 
    \(\(g\  = \ 
        simpler[coordxform[gp, xp, 
            If[dim \[Equal] 4, {r, ArcCos[b], y, 
                t}, \[IndentingNewLine]Flatten[{Thread[
                    ArcCos[Take[{p, q, s, u, v, w}, dim - 5]]], 
                  If[dim \[GreaterEqual] 5, z, {}], r, ArcCos[b], y, t}]], 
            x]];\)\), "\[IndentingNewLine]", 
    \(\(Print["\<dim = \>", dim, "\<, x = \>", 
        x, "\<, metric components are:\>"];\)\), "\[IndentingNewLine]", 
    \(\(components[g];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("dim = "\[InvisibleSpace]11\[InvisibleSpace]", x = "\
\[InvisibleSpace]{t1, t2, t3, t4, t5, t6, p2, r, th, ph, 
          t}\[InvisibleSpace]", metric components are:"\),
      SequenceForm[ 
      "dim = ", 11, ", x = ", {t1, t2, t3, t4, t5, t6, p2, r, th, ph, t}, 
        ", metric components are:"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\ \
Cos[th]\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Cos[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\ Cos[th]\^2\ \
Sin[t1]\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Cos[ th], 2], 
          Power[ 
            Sin[ t1], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\ Cos[th]\^2\ \
Sin[t1]\^2\ Sin[t2]\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Cos[ th], 2], 
          Power[ 
            Sin[ t1], 2], 
          Power[ 
            Sin[ t2], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\ Cos[th]\^2\ \
Sin[t1]\^2\ Sin[t2]\^2\ Sin[t3]\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Cos[ th], 2], 
          Power[ 
            Sin[ t1], 2], 
          Power[ 
            Sin[ t2], 2], 
          Power[ 
            Sin[ t3], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\ Cos[th]\^2\ \
Sin[t1]\^2\ Sin[t2]\^2\ Sin[t3]\^2\ Sin[t4]\^2\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Cos[ th], 2], 
          Power[ 
            Sin[ t1], 2], 
          Power[ 
            Sin[ t2], 2], 
          Power[ 
            Sin[ t3], 2], 
          Power[ 
            Sin[ t4], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\ Cos[th]\^2\ \
Sin[t1]\^2\ Sin[t2]\^2\ Sin[t3]\^2\ Sin[t4]\^2\ Sin[t5]\^2\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Cos[ th], 2], 
          Power[ 
            Sin[ t1], 2], 
          Power[ 
            Sin[ t2], 2], 
          Power[ 
            Sin[ t3], 2], 
          Power[ 
            Sin[ t4], 2], 
          Power[ 
            Sin[ t5], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\ Cos[th]\^2\ \
Sin[t1]\^2\ Sin[t2]\^2\ Sin[t3]\^2\ Sin[t4]\^2\ Sin[t5]\^2\ Sin[t6]\^2\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Cos[ th], 2], 
          Power[ 
            Sin[ t1], 2], 
          Power[ 
            Sin[ t2], 2], 
          Power[ 
            Sin[ t3], 2], 
          Power[ 
            Sin[ t4], 2], 
          Power[ 
            Sin[ t5], 2], 
          Power[ 
            Sin[ t6], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^6\ \((r\^2 + a\
\^2\ Cos[th]\^2)\)\)\/\(\(-m\) + a\^2\ r\^6 + r\^8\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ 
          Power[ r, 6], 
          Power[ 
            Plus[ 
              Times[ -1, m], 
              Times[ 
                Power[ a, 2], 
                Power[ r, 6]], 
              Power[ r, 8]], -1], 
          Plus[ 
            Power[ r, 2], 
            Times[ 
              Power[ a, 2], 
              Power[ 
                Cos[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2 + 
          a\^2\ Cos[th]\^2\)\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Plus[ 
          Power[ r, 2], 
          Times[ 
            Power[ a, 2], 
            Power[ 
              Cos[ th], 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]\(Sin[th]\^2\ \((a\
\^2\ r\^8 + r\^10 + a\^4\ r\^6\ Cos[th]\^2 + a\^2\ r\^8\ Cos[th]\^2 + a\^2\ m\
\ Sin[th]\^2)\)\)\/\(r\^6\ \((r\^2 + a\^2\ Cos[th]\^2)\)\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ 
          Power[ r, -6], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Power[ r, 8]], 
            Power[ r, 10], 
            Times[ 
              Power[ a, 4], 
              Power[ r, 6], 
              Power[ 
                Cos[ th], 2]], 
            Times[ 
              Power[ a, 2], 
              Power[ r, 8], 
              Power[ 
                Cos[ th], 2]], 
            Times[ 
              Power[ a, 2], m, 
              Power[ 
                Sin[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\ Sin[th]\^2\
\)\/\(r\^6\ \((r\^2 + a\^2\ Cos[th]\^2)\)\)\),
      SequenceForm[ "[", 10, ",", 11, "] = ", 
        Times[ a, m, 
          Power[ r, -6], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\ Sin[th]\^2\
\)\/\(r\^6\ \((r\^2 + a\^2\ Cos[th]\^2)\)\)\),
      SequenceForm[ "[", 11, ",", 10, "] = ", 
        Times[ a, m, 
          Power[ r, -6], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(m - r\^8 - a\^2\ \
r\^6\ Cos[th]\^2\)\/\(r\^6\ \((r\^2 + a\^2\ Cos[th]\^2)\)\)\),
      SequenceForm[ "[", 11, ",", 11, "] = ", 
        Times[ 
          Power[ r, -6], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Plus[ m, 
            Times[ -1, 
              Power[ r, 8]], 
            Times[ -1, 
              Power[ a, 2], 
              Power[ r, 6], 
              Power[ 
                Cos[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("dim = "\[InvisibleSpace]11\[InvisibleSpace]", x = "\
\[InvisibleSpace]{p, q, s, u, v, w, z, r, b, y, 
          t}\[InvisibleSpace]", metric components are:"\),
      SequenceForm[ 
      "dim = ", 11, ", x = ", {p, q, s, u, v, w, z, r, b, y, t}, 
        ", metric components are:"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(b\^2\ r\^2\)\
\/\(\((\(-1\) + p)\)\ \((1 + p)\)\)\)\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ -1, 
          Power[ b, 2], 
          Power[ 
            Plus[ -1, p], -1], 
          Power[ 
            Plus[ 1, p], -1], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^2\ \((\(-1\) + \
p)\)\ \((1 + p)\)\ r\^2\)\/\(\((\(-1\) + q)\)\ \((1 + q)\)\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, 2], 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Power[ 
            Plus[ -1, q], -1], 
          Power[ 
            Plus[ 1, q], -1], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(b\^2\ \
\((\(-1\) + p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
                    q)\)\ r\^2\)\/\(\((\(-1\) + s)\)\ \((1 + s)\)\)\)\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ -1, 
          Power[ b, 2], 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Plus[ -1, q], 
          Plus[ 1, q], 
          Power[ r, 2], 
          Power[ 
            Plus[ -1, s], -1], 
          Power[ 
            Plus[ 1, s], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^2\ \((\(-1\) + \
p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + q)\)\ r\^2\ \((\(-1\) + s)\)\ \
\((1 + s)\)\)\/\(\((\(-1\) + u)\)\ \((1 + u)\)\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ b, 2], 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Plus[ -1, q], 
          Plus[ 1, q], 
          Power[ r, 2], 
          Plus[ -1, s], 
          Plus[ 1, s], 
          Power[ 
            Plus[ -1, u], -1], 
          Power[ 
            Plus[ 1, u], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(b\^2\ \
\((\(-1\) + p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
                    q)\)\ r\^2\ \((\(-1\) + s)\)\ \((1 + s)\)\ \((\(-1\) + 
                    u)\)\ \((1 + u)\)\)\/\(\((\(-1\) + v)\)\ \((1 + 
                    v)\)\)\)\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ -1, 
          Power[ b, 2], 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Plus[ -1, q], 
          Plus[ 1, q], 
          Power[ r, 2], 
          Plus[ -1, s], 
          Plus[ 1, s], 
          Plus[ -1, u], 
          Plus[ 1, u], 
          Power[ 
            Plus[ -1, v], -1], 
          Power[ 
            Plus[ 1, v], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^2\ \((\(-1\) + \
p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + q)\)\ r\^2\ \((\(-1\) + s)\)\ \
\((1 + s)\)\ \((\(-1\) + u)\)\ \((1 + u)\)\ \((\(-1\) + v)\)\ \((1 + v)\)\)\/\
\(\((\(-1\) + w)\)\ \((1 + w)\)\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ b, 2], 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Plus[ -1, q], 
          Plus[ 1, q], 
          Power[ r, 2], 
          Plus[ -1, s], 
          Plus[ 1, s], 
          Plus[ -1, u], 
          Plus[ 1, u], 
          Plus[ -1, v], 
          Plus[ 1, v], 
          Power[ 
            Plus[ -1, w], -1], 
          Power[ 
            Plus[ 1, w], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^2\ \((\(-1\) + 
              p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
              q)\)\ r\^2\ \((\(-1\) + s)\)\ \((1 + s)\)\ \((\(-1\) + 
              u)\)\ \((1 + u)\)\ \((\(-1\) + v)\)\ \((1 + v)\)\ \((\(-1\) + 
              w)\)\ \((1 + w)\)\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Power[ b, 2], 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Plus[ -1, q], 
          Plus[ 1, q], 
          Power[ r, 2], 
          Plus[ -1, s], 
          Plus[ 1, s], 
          Plus[ -1, u], 
          Plus[ 1, u], 
          Plus[ -1, v], 
          Plus[ 1, v], 
          Plus[ -1, w], 
          Plus[ 1, w]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^6\ \((a\^2\ \
b\^2 + r\^2)\)\)\/\(\(-m\) + a\^2\ r\^6 + r\^8\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ 
          Power[ r, 6], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Power[ b, 2]], 
            Power[ r, 2]], 
          Power[ 
            Plus[ 
              Times[ -1, m], 
              Times[ 
                Power[ a, 2], 
                Power[ r, 6]], 
              Power[ r, 8]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(a\^2\ b\^2 + 
                r\^2\)\/\(\((\(-1\) + b)\)\ \((1 + b)\)\)\)\)\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ -1, 
          Power[ 
            Plus[ -1, b], -1], 
          Power[ 
            Plus[ 1, b], -1], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Power[ b, 2]], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(\((\(-1\) + 
                    b)\)\ \((1 + b)\)\ \((a\^2\ m - a\^2\ b\^2\ m + 
                    a\^4\ b\^2\ r\^6 + a\^2\ r\^8 + a\^2\ b\^2\ r\^8 + 
                    r\^10)\)\)\/\(r\^6\ \((a\^2\ b\^2 + r\^2)\)\)\)\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Times[ -1, 
          Plus[ -1, b], 
          Plus[ 1, b], 
          Power[ r, -6], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ a, 2], 
                Power[ b, 2]], 
              Power[ r, 2]], -1], 
          Plus[ 
            Times[ 
              Power[ a, 2], m], 
            Times[ -1, 
              Power[ a, 2], 
              Power[ b, 2], m], 
            Times[ 
              Power[ a, 4], 
              Power[ b, 2], 
              Power[ r, 6]], 
            Times[ 
              Power[ a, 2], 
              Power[ r, 8]], 
            Times[ 
              Power[ a, 2], 
              Power[ b, 2], 
              Power[ r, 8]], 
            Power[ r, 10]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(a\ \
\((\(-1\) + b)\)\ \((1 + b)\)\ m\)\/\(r\^6\ \((a\^2\ b\^2 + r\^2)\)\)\)\)\),
      SequenceForm[ "[", 10, ",", 11, "] = ", 
        Times[ -1, a, 
          Plus[ -1, b], 
          Plus[ 1, b], m, 
          Power[ r, -6], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ a, 2], 
                Power[ b, 2]], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(a\ \
\((\(-1\) + b)\)\ \((1 + b)\)\ m\)\/\(r\^6\ \((a\^2\ b\^2 + r\^2)\)\)\)\)\),
      SequenceForm[ "[", 11, ",", 10, "] = ", 
        Times[ -1, a, 
          Plus[ -1, b], 
          Plus[ 1, b], m, 
          Power[ r, -6], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ a, 2], 
                Power[ b, 2]], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(\(-m\) + 
                a\^2\ b\^2\ r\^6 + 
                r\^8\)\/\(r\^6\ \((a\^2\ b\^2 + r\^2)\)\)\)\)\),
      SequenceForm[ "[", 11, ",", 11, "] = ", 
        Times[ -1, 
          Power[ r, -6], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ a, 2], 
                Power[ b, 2]], 
              Power[ r, 2]], -1], 
          Plus[ 
            Times[ -1, m], 
            Times[ 
              Power[ a, 2], 
              Power[ b, 2], 
              Power[ r, 6]], 
            Power[ r, 8]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell["\<\
We next parameterize the metric so that each component is a \
polynomial:\
\>", "Text"],

Cell[BoxData[
    \(\(rules\  = \ {a\^2\ b\^2 + r\^2 \[Rule] 
            F, \(a\^2\) r^6 - m\  + r\^8 \[Rule] 
            G^2, \[IndentingNewLine]\((\(-1\) + 
                b)\) \[Rule] \(-B^2\)/\((1 + b)\), 
          b^2 \[Rule] 
            1 - B^2, \[IndentingNewLine]\((\(-1\) + 
                p)\) \[Rule] \(-P^2\)/\((1 + p)\), \((\(-1\) + 
                q)\) \[Rule] \(-Q^2\)/\((1 + q)\), \((\(-1\) + 
                s)\) \[Rule] \(-S^2\)/\((1 + s)\), \((\(-1\) + 
                u)\) \[Rule] \(-U^2\)/\((1 + u)\), \((\(-1\) + 
                v)\) \[Rule] \(-V^2\)/\((1 + v)\), \((\(-1\) + 
                w)\) \[Rule] \(-W^2\)/\((1 + w)\)};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(gp\  = \ Expand[g /. rules];\)\), "\[IndentingNewLine]", 
    \(\(components[gp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\/P\^2 - \
\(B\^2\ r\^2\)\/P\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, -2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, -2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ r\^2\)\/Q\
\^2 - \(B\^2\ P\^2\ r\^2\)\/Q\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, -2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, -2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ Q\^2\ \
r\^2\)\/S\^2 - \(B\^2\ P\^2\ Q\^2\ r\^2\)\/S\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, -2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, -2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ Q\^2\ \
r\^2\ S\^2\)\/U\^2 - \(B\^2\ P\^2\ Q\^2\ r\^2\ S\^2\)\/U\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, -2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, -2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ Q\^2\ \
r\^2\ S\^2\ U\^2\)\/V\^2 - \(B\^2\ P\^2\ Q\^2\ r\^2\ S\^2\ U\^2\)\/V\^2\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, 2], 
            Power[ V, -2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, 2], 
            Power[ V, -2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ Q\^2\ \
r\^2\ S\^2\ U\^2\ V\^2\)\/W\^2 - \(B\^2\ P\^2\ Q\^2\ r\^2\ S\^2\ U\^2\ V\^2\)\
\/W\^2\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, 2], 
            Power[ V, 2], 
            Power[ W, -2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, 2], 
            Power[ V, 2], 
            Power[ W, -2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(P\^2\ Q\^2\ r\^2\ \
S\^2\ U\^2\ V\^2\ W\^2 - B\^2\ P\^2\ Q\^2\ r\^2\ S\^2\ U\^2\ V\^2\ W\^2\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, 2], 
            Power[ V, 2], 
            Power[ W, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, 2], 
            Power[ V, 2], 
            Power[ W, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(F\ r\^6\)\/G\^2\),
      
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ F, 
          Power[ G, -2], 
          Power[ r, 6]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]F\/B\^2\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ 
          Power[ B, -2], F]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(a\^4\ \
B\^2\)\/F - \(a\^4\ B\^4\)\/F + \(a\^2\ B\^4\ m\)\/\(F\ r\^6\) + \(2\ a\^2\ B\
\^2\ r\^2\)\/F - \(a\^2\ B\^4\ r\^2\)\/F + \(B\^2\ r\^4\)\/F\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Plus[ 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1]], 
          Times[ -1, 
            Power[ a, 4], 
            Power[ B, 4], 
            Power[ F, -1]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], m, 
            Power[ r, -6]], 
          Times[ 2, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ \
m\)\/\(F\ r\^6\)\),
      SequenceForm[ "[", 10, ",", 11, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, 
          Power[ r, -6]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ \
m\)\/\(F\ r\^6\)\),
      SequenceForm[ "[", 11, ",", 10, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, 
          Power[ r, -6]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(a\^2\/F\)\) \
+ \(a\^2\ B\^2\)\/F + m\/\(F\ r\^6\) - r\^2\/F\)\),
      SequenceForm[ "[", 11, ",", 11, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 2], 
            Power[ F, -1]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1]], 
          Times[ 
            Power[ F, -1], m, 
            Power[ r, -6]], 
          Times[ -1, 
            Power[ F, -1], 
            Power[ r, 2]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(ginvp\  = \ 
        Expand[Simplify[
            Factor[Inverse[g]] /. rules]];\)\), "\[IndentingNewLine]", 
    \(\(components[ginvp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]P\^2\/\(b\^2\ \
r\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, 2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]Q\^2\/\(b\^2\ P\^2\ \
r\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, 2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]S\^2\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]U\^2\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\ S\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, -2], 
          Power[ U, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]V\^2\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\ S\^2\ U\^2\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, -2], 
          Power[ U, -2], 
          Power[ V, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]W\^2\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\ S\^2\ U\^2\ V\^2\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, -2], 
          Power[ U, -2], 
          Power[ V, -2], 
          Power[ W, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]1\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\ S\^2\ U\^2\ V\^2\ W\^2\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, -2], 
          Power[ U, -2], 
          Power[ V, -2], 
          Power[ W, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]G\^2\/\(F\ r\^6\)\),
      
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ 
          Power[ F, -1], 
          Power[ G, 2], 
          Power[ r, -6]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]B\^2\/F\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Times[ 
          Power[ B, 2], 
          Power[ F, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(m\/\(B\^2\ \
F\ G\^2\)\)\) - \(a\^2\ r\^6\)\/\(F\ G\^2\) + \(a\^2\ r\^6\)\/\(B\^2\ F\ G\^2\
\) + r\^8\/\(B\^2\ F\ G\^2\)\)\),
      SequenceForm[ "[", 10, ",", 10, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], m], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 6]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 6]], 
          Times[ 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 8]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]10\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\)\/\(F\ \
G\^2\)\),
      SequenceForm[ "[", 10, ",", 11, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]10\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\)\/\(F\ \
G\^2\)\),
      SequenceForm[ "[", 11, ",", 10, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]11\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(\(a\^2\ \
B\^2\ m\)\/\(F\ G\^2\)\)\) - \(a\^4\ r\^6\)\/\(F\ G\^2\) + \(a\^4\ B\^2\ r\^6\
\)\/\(F\ G\^2\) - \(2\ a\^2\ r\^8\)\/\(F\ G\^2\) + \(a\^2\ B\^2\ r\^8\)\/\(F\ \
G\^2\) - r\^10\/\(F\ G\^2\)\)\),
      SequenceForm[ "[", 11, ",", 11, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], m], 
          Times[ -1, 
            Power[ a, 4], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 6]], 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 6]], 
          Times[ -2, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 8]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 8]], 
          Times[ -1, 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 10]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(irmdgp\  = \ 
      simpler[1/PowerExpand[Sqrt[simpler[\(-Det[g]\)]] /. rules]]\)], "Input"],

Cell[BoxData[
    \(1\/\(b\^7\ F\ P\^5\ Q\^4\ r\^7\ S\^3\ U\^2\ V\)\)], "Output"]
}, Open  ]],

Cell["\<\
Don't forget the sub[i] in the derivatives (takes care of change in \
time coord from i=10 to i=0 in c++)\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(Btab\  = \ {B^2, 1 - b^2, 
        Table[D[1 - b^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              B)\)}\), "\[IndentingNewLine]", 
    \(Ftab\  = \ {F, a\^2\ b\^2 + r\^2, 
        Table[D[a\^2\ b\^2 + r\^2, x[\([sub[i]]\)]], {i, 
            dim}]}\), "\[IndentingNewLine]", 
    \(Gtab\  = \ {G^2, \(a\^2\) r^6 - m\  + r\^8, 
        Table[D[\(a\^2\) r^6 - m\  + r\^8, x[\([sub[i]]\)]], {i, dim}]/\((2  
              G)\)}\), "\[IndentingNewLine]", 
    \(Ptab\  = \ {P^2, 1 - p^2, 
        Table[D[1 - p^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              P)\)}\), "\[IndentingNewLine]", 
    \(Qtab\  = \ {Q^2, 1 - q^2, 
        Table[D[1 - q^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              Q)\)}\), "\[IndentingNewLine]", 
    \(Stab\  = \ {S^2, 1 - s^2, 
        Table[D[1 - s^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              S)\)}\), "\[IndentingNewLine]", 
    \(Utab\  = \ {U^2, 1 - u^2, 
        Table[D[1 - u^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              U)\)}\), "\[IndentingNewLine]", 
    \(Vtab\  = \ {V^2, 1 - v^2, 
        Table[D[1 - v^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              V)\)}\), "\[IndentingNewLine]", 
    \(Wtab\  = \ {W^2, 1 - w^2, 
        Table[D[1 - w^2, x[\([sub[i]]\)]], {i, dim}]/\((2  W)\)}\)}], "Input"],

Cell[BoxData[
    \({B\^2, 
      1 - b\^2, {0, 0, 0, 0, 0, 0, 0, 0, 0, \(-\(b\/B\)\), 0}}\)], "Output"],

Cell[BoxData[
    \({F, 
      a\^2\ b\^2 + r\^2, {0, 0, 0, 0, 0, 0, 0, 0, 2\ r, 2\ a\^2\ b, 
        0}}\)], "Output"],

Cell[BoxData[
    \({G\^2, \(-m\) + a\^2\ r\^6 + r\^8, {0, 0, 0, 0, 0, 0, 0, 
        0, \(6\ a\^2\ r\^5 + 8\ r\^7\)\/\(2\ G\), 0, 0}}\)], "Output"],

Cell[BoxData[
    \({P\^2, 
      1 - p\^2, {0, \(-\(p\/P\)\), 0, 0, 0, 0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({Q\^2, 
      1 - q\^2, {0, 0, \(-\(q\/Q\)\), 0, 0, 0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({S\^2, 
      1 - s\^2, {0, 0, 0, \(-\(s\/S\)\), 0, 0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({U\^2, 
      1 - u\^2, {0, 0, 0, 0, \(-\(u\/U\)\), 0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({V\^2, 
      1 - v\^2, {0, 0, 0, 0, 0, \(-\(v\/V\)\), 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({W\^2, 
      1 - w\^2, {0, 0, 0, 0, 0, 0, \(-\(w\/W\)\), 0, 0, 0, 0}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(coords\  = \ x\), "\[IndentingNewLine]", 
    \(ints\  = \ {2, 3}\), "\n", 
    \(constants\  = \ {I, m, a}\), "\n", 
    \(functions\  = \ 
      Take[{Ftab, Gtab, Btab, Ptab, Qtab, Stab, Utab, Vtab, Wtab}, 
        If[dim \[Equal] 4, 3, dim - 2]]\), "\[IndentingNewLine]", 
    \(ansatze\  = \ {}\)}], "Input"],

Cell[BoxData[
    \({p, q, s, u, v, w, z, r, b, y, t}\)], "Output"],

Cell[BoxData[
    \({2, 3}\)], "Output"],

Cell[BoxData[
    \({\[ImaginaryI], m, a}\)], "Output"],

Cell[BoxData[
    \({{F, 
        a\^2\ b\^2 + r\^2, {0, 0, 0, 0, 0, 0, 0, 0, 2\ r, 2\ a\^2\ b, 
          0}}, {G\^2, \(-m\) + a\^2\ r\^6 + r\^8, {0, 0, 0, 0, 0, 0, 0, 
          0, \(6\ a\^2\ r\^5 + 8\ r\^7\)\/\(2\ G\), 0, 0}}, {B\^2, 
        1 - b\^2, {0, 0, 0, 0, 0, 0, 0, 0, 0, \(-\(b\/B\)\), 0}}, {P\^2, 
        1 - p\^2, {0, \(-\(p\/P\)\), 0, 0, 0, 0, 0, 0, 0, 0, 0}}, {Q\^2, 
        1 - q\^2, {0, 0, \(-\(q\/Q\)\), 0, 0, 0, 0, 0, 0, 0, 0}}, {S\^2, 
        1 - s\^2, {0, 0, 0, \(-\(s\/S\)\), 0, 0, 0, 0, 0, 0, 0}}, {U\^2, 
        1 - u\^2, {0, 0, 0, 0, \(-\(u\/U\)\), 0, 0, 0, 0, 0, 0}}, {V\^2, 
        1 - v\^2, {0, 0, 0, 0, 0, \(-\(v\/V\)\), 0, 0, 0, 0, 0}}, {W\^2, 
        1 - w\^2, {0, 0, 0, 0, 0, 0, \(-\(w\/W\)\), 0, 0, 0, 0}}}\)], "Output"],

Cell[BoxData[
    \({}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(headers[];\)\)], "Input"],

Cell[BoxData[
    \("functions completed..."\)], "Print"],

Cell[BoxData[
    \("g completed..."\)], "Print"],

Cell[BoxData[
    \("gi completed..."\)], "Print"],

Cell[BoxData[
    \("d completed..."\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(tensor[ad2, "\<ad2\>"];\)\), "\[IndentingNewLine]", 
    \(\(tensor[ad3, "\<ad3\>"];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("ad2"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ad2", " completed..."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("ad3"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ad3", " completed..."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(backup[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]19\[InvisibleSpace]":"\[InvisibleSpace]37\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]5\[InvisibleSpace]\
"/"\[InvisibleSpace]2003\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 19, ":", 37, " on ", 4, "/", 5, "/", 2003, "."],
      Editable->False]], "Print"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "ken", \
"projects", "mathematica", "packages"}, "d11brane.nb.ps", CharacterEncoding \
-> "ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 70, 1, 27, "Input"],
Cell[1874, 57, 2984, 44, 375, "Print"],
Cell[4861, 103, 81, 1, 23, "Print"]
}, Open  ]],
Cell[4957, 107, 50, 1, 27, "Input"],
Cell[5010, 110, 451, 9, 91, "Input"],
Cell[5464, 121, 536, 11, 91, "Input"],
Cell[6003, 134, 74, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[6102, 139, 2298, 44, 379, "Input"],
Cell[8403, 185, 375, 7, 23, "Print"],
Cell[8781, 194, 330, 9, 25, "Print"],
Cell[9114, 205, 387, 11, 25, "Print"],
Cell[9504, 218, 444, 13, 25, "Print"],
Cell[9951, 233, 501, 15, 25, "Print"],
Cell[10455, 250, 558, 17, 25, "Print"],
Cell[11016, 269, 615, 19, 25, "Print"],
Cell[11634, 290, 672, 21, 25, "Print"],
Cell[12309, 313, 671, 20, 41, "Print"],
Cell[12983, 335, 395, 11, 25, "Print"],
Cell[13381, 348, 1148, 36, 43, "Print"],
Cell[14532, 386, 565, 16, 43, "Print"],
Cell[15100, 404, 565, 16, 43, "Print"],
Cell[15668, 422, 742, 22, 43, "Print"],
Cell[16413, 446, 357, 7, 23, "Print"],
Cell[16773, 455, 451, 12, 44, "Print"],
Cell[17227, 469, 518, 14, 44, "Print"],
Cell[17748, 485, 632, 17, 43, "Print"],
Cell[18383, 504, 680, 19, 43, "Print"],
Cell[19066, 525, 834, 23, 43, "Print"],
Cell[19903, 550, 842, 24, 43, "Print"],
Cell[20748, 576, 842, 23, 41, "Print"],
Cell[21593, 601, 641, 19, 41, "Print"],
Cell[22237, 622, 547, 15, 43, "Print"],
Cell[22787, 639, 1188, 35, 43, "Print"],
Cell[23978, 676, 569, 15, 42, "Print"],
Cell[24550, 693, 569, 15, 42, "Print"],
Cell[25122, 710, 724, 21, 43, "Print"]
}, Open  ]],
Cell[25861, 734, 97, 3, 32, "Text"],
Cell[25961, 739, 675, 12, 83, "Input"],

Cell[CellGroupData[{
Cell[26661, 755, 120, 2, 43, "Input"],
Cell[26784, 759, 455, 13, 41, "Print"],
Cell[27242, 774, 525, 15, 42, "Print"],
Cell[27770, 791, 591, 17, 41, "Print"],
Cell[28364, 810, 657, 19, 41, "Print"],
Cell[29024, 831, 723, 21, 41, "Print"],
Cell[29750, 854, 791, 24, 41, "Print"],
Cell[30544, 880, 779, 23, 25, "Print"],
Cell[31326, 905, 316, 8, 41, "Print"],
Cell[31645, 915, 274, 6, 39, "Print"],
Cell[31922, 923, 1096, 34, 41, "Print"],
Cell[33021, 959, 354, 9, 41, "Print"],
Cell[33378, 970, 354, 9, 41, "Print"],
Cell[33735, 981, 645, 19, 41, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[34417, 1005, 175, 4, 43, "Input"],
Cell[34595, 1011, 337, 9, 41, "Print"],
Cell[34935, 1022, 369, 10, 41, "Print"],
Cell[35307, 1034, 401, 11, 42, "Print"],
Cell[35711, 1047, 433, 12, 42, "Print"],
Cell[36147, 1061, 465, 13, 42, "Print"],
Cell[36615, 1076, 497, 14, 42, "Print"],
Cell[37115, 1092, 501, 14, 41, "Print"],
Cell[37619, 1108, 339, 9, 41, "Print"],
Cell[37961, 1119, 296, 7, 41, "Print"],
Cell[38260, 1128, 900, 27, 41, "Print"],
Cell[39163, 1157, 323, 8, 37, "Print"],
Cell[39489, 1167, 323, 8, 37, "Print"],
Cell[39815, 1177, 1267, 39, 41, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[41119, 1221, 112, 2, 27, "Input"],
Cell[41234, 1225, 81, 1, 45, "Output"]
}, Open  ]],
Cell[41330, 1229, 129, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[41484, 1236, 1287, 26, 164, "Input"],
Cell[42774, 1264, 104, 2, 44, "Output"],
Cell[42881, 1268, 119, 3, 29, "Output"],
Cell[43003, 1273, 148, 2, 45, "Output"],
Cell[43154, 1277, 104, 2, 41, "Output"],
Cell[43261, 1281, 104, 2, 42, "Output"],
Cell[43368, 1285, 104, 2, 41, "Output"],
Cell[43475, 1289, 104, 2, 41, "Output"],
Cell[43582, 1293, 104, 2, 41, "Output"],
Cell[43689, 1297, 104, 2, 41, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[43830, 1304, 333, 7, 107, "Input"],
Cell[44166, 1313, 67, 1, 27, "Output"],
Cell[44236, 1316, 40, 1, 27, "Output"],
Cell[44279, 1319, 55, 1, 27, "Output"],
Cell[44337, 1322, 761, 11, 205, "Output"],
Cell[45101, 1335, 36, 1, 27, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[45174, 1341, 47, 1, 27, "Input"],
Cell[45224, 1344, 57, 1, 23, "Print"],
Cell[45284, 1347, 49, 1, 23, "Print"],
Cell[45336, 1350, 50, 1, 23, "Print"],
Cell[45389, 1353, 49, 1, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[45475, 1359, 123, 2, 43, "Input"],
Cell[45601, 1363, 157, 3, 23, "Print"],
Cell[45761, 1368, 157, 3, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[45955, 1376, 46, 1, 27, "Input"],
Cell[46004, 1379, 425, 7, 23, "Print"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

