(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     17305,        533]*)
(*NotebookOutlinePosition[     18292,        565]*)
(*  CellTagsIndexPosition[     18248,        561]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/ccif.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,<oldx,>func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("AVOID USING C, D, E, I, N and O as variables !"\)], "Print"]
}, Closed]],

Cell[BoxData[{
    \(\(f\  = \ Table[0, {i, 50}];\)\), "\[IndentingNewLine]", 
    \(\(derivs\  = \ 
        Table[Table[0, {i, 1000}], {j, 50}];\)\), "\[IndentingNewLine]", 
    \(\(g\  = Table[Table[0, {i, 11}], {j, 11}]\ ;\)\), "\n", 
    \(\(ginv\  = \ 
        Table[Table[0, {i, 11}], {j, 11}];\)\), "\[IndentingNewLine]", 
    \(\(Euler\  = \ 0;\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< /home/ken/projects/ptah/ptah.out\)], "Input"],

Cell[BoxData[
    InterpretationBox[\(0\[InvisibleSpace]1\),
      SequenceForm[ 0, 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(1\[InvisibleSpace]2\),
      SequenceForm[ 1, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(2\[InvisibleSpace]2\),
      SequenceForm[ 2, 2],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[f];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(a\^2\ b\^2 + r\^2\)\),
      SequenceForm[ "[", 1, "] = ", 
        Plus[ 
          Times[ 
            Power[ a, 2], 
            Power[ b, 2]], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(a\^2 - m\ r + r\^2\)\),
      SequenceForm[ "[", 2, "] = ", 
        Plus[ 
          Power[ a, 2], 
          Times[ -1, m, r], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - b\^2\)\),
      SequenceForm[ "[", 3, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ b, 2]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[derivs];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ r\)\),
      SequenceForm[ "[", 1, ",", 2, "] = ", 
        Times[ 2, r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ a\^2\ b\)\),
      SequenceForm[ "[", 1, ",", 3, "] = ", 
        Times[ 2, 
          Power[ a, 2], b]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(m\/\(2\ \
G\)\)\) + r\/G\)\),
      SequenceForm[ "[", 2, ",", 6, "] = ", 
        Plus[ 
          Times[ 
            Rational[ -1, 2], 
            Power[ G, -1], m], 
          Times[ 
            Power[ G, -1], r]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]11\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(b\/B\)\)\),
      SequenceForm[ "[", 3, ",", 11, "] = ", 
        Times[ -1, b, 
          Power[ B, -1]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[g];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]F\/G\^2\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ F, 
          Power[ G, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]F\/B\^2\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ B, -2], F]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(a\^4\ B\^2\)\/F \
- \(a\^4\ B\^4\)\/F + \(a\^2\ B\^4\ m\ r\)\/F + \(2\ a\^2\ B\^2\ r\^2\)\/F - \
\(a\^2\ B\^4\ r\^2\)\/F + \(B\^2\ r\^4\)\/F\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Plus[ 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1]], 
          Times[ -1, 
            Power[ a, 4], 
            Power[ B, 4], 
            Power[ F, -1]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], m, r], 
          Times[ 2, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ m\ \
r\)\/F\),
      SequenceForm[ "[", 3, ",", 4, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ m\ \
r\)\/F\),
      SequenceForm[ "[", 4, ",", 3, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-1\) + \(m\ r\)\
\/F\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Plus[ -1, 
          Times[ 
            Power[ F, -1], m, r]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[ginv];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]G\^2\/F\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ F, -1], 
          Power[ G, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]B\^2\/F\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ B, 2], 
          Power[ F, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(a\^2\/\(F\ G\
\^2\)\)\) + a\^2\/\(B\^2\ F\ G\^2\) - \(m\ r\)\/\(B\^2\ F\ G\^2\) + 
          r\^2\/\(B\^2\ F\ G\^2\)\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2]], 
          Times[ -1, 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], m, r], 
          Times[ 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\ r\)\/\(F\ G\
\^2\)\),
      SequenceForm[ "[", 3, ",", 4, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m, r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\ r\)\/\(F\ G\
\^2\)\),
      SequenceForm[ "[", 4, ",", 3, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m, r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(a\^4\/\(F\ G\
\^2\)\)\) + \(a\^4\ B\^2\)\/\(F\ G\^2\) - \(a\^2\ B\^2\ m\ r\)\/\(F\ G\^2\) - \
\(2\ a\^2\ r\^2\)\/\(F\ G\^2\) + \(a\^2\ B\^2\ r\^2\)\/\(F\ G\^2\) - 
          r\^4\/\(F\ G\^2\)\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 4], 
            Power[ F, -1], 
            Power[ G, -2]], 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2]], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], m, r], 
          Times[ -2, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 2]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 4]]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(isrdmg\)], "Input"],

Cell[BoxData[
    \(1\/F\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[sc]]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riesq]]\)], "Input"],

Cell[BoxData[
    \(\(-\(\(12\ m\^2\ \((a\ b - r)\)\ \((a\ b + r)\)\ \((a\^2\ b\^2 - 
                4\ a\ b\ r + r\^2)\)\ \((a\^2\ b\^2 + 4\ a\ b\ r + 
                r\^2)\)\)\/\((a\^2\ b\^2 + r\^2)\)\^6\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riecub1]]\)], "Input"],

Cell[BoxData[
    \(\(12\ m\^3\ r\ \((3\ a\^2\ b\^2 - r\^2)\)\ \((3\ a\^6\ b\^6 - 27\ a\^4\ \
b\^4\ r\^2 + 33\ a\^2\ b\^2\ r\^4 - r\^6)\)\)\/\((a\^2\ b\^2 + r\^2)\)\^9\)], \
"Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riecub2]]\)], "Input"],

Cell[BoxData[
    \(\(6\ m\^3\ r\ \((3\ a\^2\ b\^2 - r\^2)\)\ \((3\ a\^6\ b\^6 - 27\ a\^4\ \
b\^4\ r\^2 + 33\ a\^2\ b\^2\ r\^4 - r\^6)\)\)\/\((a\^2\ b\^2 + r\^2)\)\^9\)], \
"Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riecub3]]\)], "Input"],

Cell[BoxData[
    \(\(3\ m\^3\ r\ \((3\ a\^2\ b\^2 - r\^2)\)\ \((3\ a\^6\ b\^6 - 27\ a\^4\ \
b\^4\ r\^2 + 33\ a\^2\ b\^2\ r\^4 - r\^6)\)\)\/\((a\^2\ b\^2 + r\^2)\)\^9\)], \
"Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Cdriesq1]]\)], "Input"],

Cell[BoxData[
    \(\(\(1\/\((a\^2\ b\^2 + r\^2)\)\^9\)\((180\ m\^2\ \((a\^2\ b\^2 - m\ r + 
            r\^2)\)\ \((a\^4\ b\^4 + 4\ a\^3\ b\^3\ r - 6\ a\^2\ b\^2\ r\^2 - 
            4\ a\ b\ r\^3 + r\^4)\)\ \((a\^4\ b\^4 - 4\ a\^3\ b\^3\ r - 
            6\ a\^2\ b\^2\ r\^2 + 4\ a\ b\ r\^3 + r\^4)\))\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[RstarR]]\)], "Input"],

Cell[BoxData[
    \(\(48\ a\ b\ m\^2\ r\ \((a\^2\ b\^2 - 3\ r\^2)\)\ \((3\ a\^2\ b\^2 - \
r\^2)\)\)\/\((a\^2\ b\^2 + r\^2)\)\^6\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[gRstarR]]\)], "Input"],

Cell[BoxData[
    \(\(-\(\(48\ m\^2\ \((a\ b - r)\)\ \((a\ b + r)\)\ \((a\^2\ b\^2 - 
                4\ a\ b\ r + r\^2)\)\ \((a\^2\ b\^2 + 4\ a\ b\ r + 
                r\^2)\)\)\/\((a\^2\ b\^2 + r\^2)\)\^6\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Euler]]\)], "Input"],

Cell[BoxData[
    \(\(48\ m\^2\ \((a\ b - r)\)\ \((a\ b + r)\)\ \((a\^2\ b\^2 - 4\ a\ b\ r \
+ r\^2)\)\ \((a\^2\ b\^2 + 4\ a\ b\ r + r\^2)\)\)\/\((a\^2\ b\^2 + r\^2)\)\^6\
\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(backup[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]15\[InvisibleSpace]":"\
\[InvisibleSpace]55\[InvisibleSpace]":"\[InvisibleSpace]15\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]8\[InvisibleSpace]\
"/"\[InvisibleSpace]2003\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 15, ":", 55, ":", 15, " on ", 4, "/", 8, "/", 2003, "."],
      Editable->False]], "Print"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "ken", \
"projects", "ptah", "mathematica"}, "dbraneval.nb.ps", CharacterEncoding -> \
"ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 70, 1, 27, "Input"],
Cell[1874, 57, 2984, 44, 375, "Print"],
Cell[4861, 103, 81, 1, 23, "Print"]
}, Closed]],
Cell[4957, 107, 368, 7, 88, "Input"],

Cell[CellGroupData[{
Cell[5350, 118, 68, 1, 27, "Input"],
Cell[5421, 121, 121, 3, 23, "Print"],
Cell[5545, 126, 121, 3, 23, "Print"],
Cell[5669, 131, 121, 3, 23, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[5827, 139, 51, 1, 24, "Input"],
Cell[5881, 142, 310, 9, 25, "Print"],
Cell[6194, 153, 292, 8, 25, "Print"],
Cell[6489, 163, 256, 7, 25, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6782, 175, 56, 1, 27, "Input"],
Cell[6841, 178, 252, 5, 23, "Print"],
Cell[7096, 185, 283, 6, 25, "Print"],
Cell[7382, 193, 401, 11, 37, "Print"],
Cell[7786, 206, 286, 6, 40, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[8109, 217, 51, 1, 24, "Input"],
Cell[8163, 220, 274, 6, 39, "Print"],
Cell[8440, 228, 274, 6, 39, "Print"],
Cell[8717, 236, 1060, 33, 41, "Print"],
Cell[9780, 271, 320, 8, 41, "Print"],
Cell[10103, 281, 320, 8, 41, "Print"],
Cell[10426, 291, 320, 8, 37, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[10783, 304, 54, 1, 24, "Input"],
Cell[10840, 307, 296, 7, 41, "Print"],
Cell[11139, 316, 296, 7, 41, "Print"],
Cell[11438, 325, 841, 25, 41, "Print"],
Cell[12282, 352, 325, 8, 37, "Print"],
Cell[12610, 362, 325, 8, 37, "Print"],
Cell[12938, 372, 1206, 37, 41, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[14181, 414, 39, 1, 24, "Input"],
Cell[14223, 417, 38, 1, 44, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[14298, 423, 49, 1, 27, "Input"],
Cell[14350, 426, 35, 1, 27, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[14422, 432, 52, 1, 27, "Input"],
Cell[14477, 435, 225, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[14739, 443, 54, 1, 27, "Input"],
Cell[14796, 446, 183, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15016, 454, 54, 1, 27, "Input"],
Cell[15073, 457, 182, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15292, 465, 54, 1, 27, "Input"],
Cell[15349, 468, 182, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15568, 476, 55, 1, 27, "Input"],
Cell[15626, 479, 321, 4, 67, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[15984, 488, 53, 1, 27, "Input"],
Cell[16040, 491, 141, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16218, 498, 54, 1, 27, "Input"],
Cell[16275, 501, 225, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16537, 509, 52, 1, 27, "Input"],
Cell[16592, 512, 186, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[16815, 520, 46, 1, 27, "Input"],
Cell[16864, 523, 425, 7, 23, "Print"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

