(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     28183,        820]*)
(*NotebookOutlinePosition[     29173,        852]*)
(*  CellTagsIndexPosition[     29129,        848]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/ccif.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,<oldx,>func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("AVOID USING C, D, E, I, N and O as variables !"\)], "Print"]
}, Open  ]],

Cell[BoxData[
    \(\(dim\  = \ 4;\)\)], "Input"],

Cell[BoxData[{
    \(\(ad2\  = \ 
        Table[Table[0, {j, dim}], {i, dim}];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        Do[\((\[IndentingNewLine]inddim\  = \ 
              Join[DeleteCases[Join[Table[l, {l, dim - 1}], {0}], i | j], {i, 
                  j}]; \[IndentingNewLine]ad2[\([i, j]\)]\  = \ 
              Signature[inddim] 
                Factorial[dim - 2];\[IndentingNewLine])\), {j, dim}], {i, 
          dim}];\)\)}], "Input"],

Cell[BoxData[{
    \(\(ad3\  = \ 
        Table[Table[Table[0, {k, dim}], {j, dim}], {i, 
            dim}];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        Do[Do[\((\[IndentingNewLine]inddim\  = \ 
                Join[DeleteCases[Join[Table[l, {l, dim - 1}], {0}], 
                    i | j | k], {i, j, k}]; \[IndentingNewLine]ad3[\([i, j, 
                    k]\)]\  = \ 
                Signature[inddim] 
                  Factorial[dim - 3];\[IndentingNewLine])\), {k, dim}], {j, 
            dim}], {i, dim}];\)\)}], "Input"],

Cell[BoxData[
    \(\(ptahdir\  = \ "\<~/projects/ptah/\>";\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(xp\  = \ 
        Flatten[{Table[Symbol["\<t\>" <> ToString[i]], {i, dim - 5}], 
            If[dim \[GreaterEqual] 5, p2, {}], r, th, ph, 
            t}];\)\), "\[IndentingNewLine]", 
    \(\(rho\  = \ Sqrt[r^2 + a^2  Cos[th]^2];\)\), "\[IndentingNewLine]", 
    \(\(delta\  = \ 
        m/\((r^\((dim - 5)\) rho^2)\);\)\), "\[IndentingNewLine]", 
    \(\(psi\  = \ 
        r^\((dim - 5)\) 
          rho^2/\((r^\((dim - 5)\) \((r^2 + a^2)\) - 
                m)\);\)\), "\[IndentingNewLine]", 
    \(\(gp\  = \ 0*IdentityMatrix[dim];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        gp[\([i, i]\)]\  = \ 
          r^2  Cos[th]^2\ Product[Sin[xp[\([j - 1]\)]]^2, {j, 2, i}], {i, 1, 
          dim - 4}];\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim - 3, dim - 3]\)]\  = \ psi;\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim - 2, dim - 2]\)]\  = \ rho^2;\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim - 1, dim - 1]\)]\  = \ \((r^2 + a^2)\) Sin[th]^2 + 
          delta*a^2  Sin[th]^4;\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim - 1, dim]\)]\  = \ 
        delta*a*Sin[th]^2;\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim, dim - 1]\)]\  = \ 
        gp[\([dim - 1, dim]\)];\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim, dim]\)]\  = \ \(-1\) + delta;\)\), "\[IndentingNewLine]", 
    \(\(gp\  = \ simpler[gp];\)\), "\[IndentingNewLine]", 
    \(\(Print["\<dim = \>", dim, "\<, x = \>", 
        xp, "\<, metric components are:\>"];\)\), "\[IndentingNewLine]", 
    \(\(components[gp];\)\), "\[IndentingNewLine]", 
    \(\(If[dim \[Equal] 4, 
        x = {r, b, y, t}, \[IndentingNewLine]x\  = \ 
          Flatten[{Take[{p, q, s, u, v, w}, dim - 5], 
              If[dim \[GreaterEqual] 5, z, {}], r, b, y, 
              t}]];\)\), "\[IndentingNewLine]", 
    \(\(g\  = \ 
        simpler[coordxform[gp, xp, 
            If[dim \[Equal] 4, {r, ArcCos[b], y, 
                t}, \[IndentingNewLine]Flatten[{Thread[
                    ArcCos[Take[{p, q, s, u, v, w}, dim - 5]]], 
                  If[dim \[GreaterEqual] 5, z, {}], r, ArcCos[b], y, t}]], 
            x]];\)\), "\[IndentingNewLine]", 
    \(\(Print["\<dim = \>", dim, "\<, x = \>", 
        x, "\<, metric components are:\>"];\)\), "\[IndentingNewLine]", 
    \(\(components[g];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("dim = "\[InvisibleSpace]4\[InvisibleSpace]", x = "\
\[InvisibleSpace]{r, th, ph, t}\[InvisibleSpace]", metric components are:"\),
      SequenceForm[ 
      "dim = ", 4, ", x = ", {r, th, ph, t}, ", metric components are:"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2 + a\^2\ \
Cos[th]\^2\)\/\(a\^2 - m\ r + r\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ -1, m, r], 
              Power[ r, 2]], -1], 
          Plus[ 
            Power[ r, 2], 
            Times[ 
              Power[ a, 2], 
              Power[ 
                Cos[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2 + 
          a\^2\ Cos[th]\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Plus[ 
          Power[ r, 2], 
          Times[ 
            Power[ a, 2], 
            Power[ 
              Cos[ th], 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(Sin[th]\^2\ \
\((a\^2\ r\^2 + r\^4 + a\^4\ Cos[th]\^2 + a\^2\ r\^2\ Cos[th]\^2 + a\^2\ m\ r\
\ Sin[th]\^2)\)\)\/\(r\^2 + a\^2\ Cos[th]\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Power[ r, 2]], 
            Power[ r, 4], 
            Times[ 
              Power[ a, 4], 
              Power[ 
                Cos[ th], 2]], 
            Times[ 
              Power[ a, 2], 
              Power[ r, 2], 
              Power[ 
                Cos[ th], 2]], 
            Times[ 
              Power[ a, 2], m, r, 
              Power[ 
                Sin[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\ r\ \
Sin[th]\^2\)\/\(r\^2 + a\^2\ Cos[th]\^2\)\),
      SequenceForm[ "[", 3, ",", 4, "] = ", 
        Times[ a, m, r, 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\ r\ \
Sin[th]\^2\)\/\(r\^2 + a\^2\ Cos[th]\^2\)\),
      SequenceForm[ "[", 4, ",", 3, "] = ", 
        Times[ a, m, r, 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(m\ r - r\^2 - \
a\^2\ Cos[th]\^2\)\/\(r\^2 + a\^2\ Cos[th]\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Plus[ 
            Times[ m, r], 
            Times[ -1, 
              Power[ r, 2]], 
            Times[ -1, 
              Power[ a, 2], 
              Power[ 
                Cos[ th], 2]]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ th], 2]]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("dim = "\[InvisibleSpace]4\[InvisibleSpace]", x = "\
\[InvisibleSpace]{r, b, y, t}\[InvisibleSpace]", metric components are:"\),
      SequenceForm[ 
      "dim = ", 4, ", x = ", {r, b, y, t}, ", metric components are:"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\^2\ b\^2 + r\^2\
\)\/\(a\^2 - m\ r + r\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Power[ b, 2]], 
            Power[ r, 2]], 
          Power[ 
            Plus[ 
              Power[ a, 2], 
              Times[ -1, m, r], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(a\^2\ b\^2 + 
                r\^2\)\/\(\((\(-1\) + b)\)\ \((1 + b)\)\)\)\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ -1, 
          Power[ 
            Plus[ -1, b], -1], 
          Power[ 
            Plus[ 1, b], -1], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Power[ b, 2]], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(\((\(-1\) + 
                    b)\)\ \((1 + b)\)\ \((a\^4\ b\^2 + a\^2\ m\ r - 
                    a\^2\ b\^2\ m\ r + a\^2\ r\^2 + a\^2\ b\^2\ r\^2 + 
                    r\^4)\)\)\/\(a\^2\ b\^2 + r\^2\)\)\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ -1, 
          Plus[ -1, b], 
          Plus[ 1, b], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ a, 2], 
                Power[ b, 2]], 
              Power[ r, 2]], -1], 
          Plus[ 
            Times[ 
              Power[ a, 4], 
              Power[ b, 2]], 
            Times[ 
              Power[ a, 2], m, r], 
            Times[ -1, 
              Power[ a, 2], 
              Power[ b, 2], m, r], 
            Times[ 
              Power[ a, 2], 
              Power[ r, 2]], 
            Times[ 
              Power[ a, 2], 
              Power[ b, 2], 
              Power[ r, 2]], 
            Power[ r, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(a\ \((\(-1\) \
+ b)\)\ \((1 + b)\)\ m\ r\)\/\(a\^2\ b\^2 + r\^2\)\)\)\),
      SequenceForm[ "[", 3, ",", 4, "] = ", 
        Times[ -1, a, 
          Plus[ -1, b], 
          Plus[ 1, b], m, r, 
          Power[ 
            Plus[ 
              Times[ 
                Power[ a, 2], 
                Power[ b, 2]], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(a\ \((\(-1\) \
+ b)\)\ \((1 + b)\)\ m\ r\)\/\(a\^2\ b\^2 + r\^2\)\)\)\),
      SequenceForm[ "[", 4, ",", 3, "] = ", 
        Times[ -1, a, 
          Plus[ -1, b], 
          Plus[ 1, b], m, r, 
          Power[ 
            Plus[ 
              Times[ 
                Power[ a, 2], 
                Power[ b, 2]], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(a\^2\ b\^2 - 
                m\ r + r\^2\)\/\(a\^2\ b\^2 + r\^2\)\)\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ -1, 
          Power[ 
            Plus[ 
              Times[ 
                Power[ a, 2], 
                Power[ b, 2]], 
              Power[ r, 2]], -1], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Power[ b, 2]], 
            Times[ -1, m, r], 
            Power[ r, 2]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell["\<\
We next parameterize the metric so that each component is a \
polynomial:\
\>", "Text"],

Cell[BoxData[
    \(\(rules\  = \ {a\^2\ b\^2 + r\^2 \[Rule] F, 
          a\^2 - m\ r + r\^2 \[Rule] G^2, 
          a\^2 + r\ \((\(-m\) + r)\) \[Rule] 
            G^2, \[IndentingNewLine]\((\(-1\) + 
                b)\) \[Rule] \(-B^2\)/\((1 + b)\), 
          b^2 \[Rule] 
            1 - B^2, \[IndentingNewLine]\((\(-1\) + 
                p)\) \[Rule] \(-P^2\)/\((1 + p)\), \((\(-1\) + 
                q)\) \[Rule] \(-Q^2\)/\((1 + q)\), \((\(-1\) + 
                s)\) \[Rule] \(-S^2\)/\((1 + s)\), \((\(-1\) + 
                u)\) \[Rule] \(-U^2\)/\((1 + u)\), \((\(-1\) + 
                v)\) \[Rule] \(-V^2\)/\((1 + v)\), \((\(-1\) + 
                w)\) \[Rule] \(-W^2\)/\((1 + w)\)};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(gp\  = \ Expand[g /. rules];\)\), "\[IndentingNewLine]", 
    \(\(components[gp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]F\/G\^2\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ F, 
          Power[ G, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]F\/B\^2\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ B, -2], F]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(a\^4\ B\^2\)\/F \
- \(a\^4\ B\^4\)\/F + \(a\^2\ B\^4\ m\ r\)\/F + \(2\ a\^2\ B\^2\ r\^2\)\/F - \
\(a\^2\ B\^4\ r\^2\)\/F + \(B\^2\ r\^4\)\/F\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Plus[ 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1]], 
          Times[ -1, 
            Power[ a, 4], 
            Power[ B, 4], 
            Power[ F, -1]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], m, r], 
          Times[ 2, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ m\ \
r\)\/F\),
      SequenceForm[ "[", 3, ",", 4, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ m\ \
r\)\/F\),
      SequenceForm[ "[", 4, ",", 3, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-1\) + \(m\ r\)\
\/F\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Plus[ -1, 
          Times[ 
            Power[ F, -1], m, r]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(ginvp\  = \ 
        Expand[Simplify[Inverse[g]] /. rules];\)\), "\[IndentingNewLine]", 
    \(\(components[ginvp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]G\^2\/F\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ F, -1], 
          Power[ G, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]B\^2\/F\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ B, 2], 
          Power[ F, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(a\^2\/\(F\ G\
\^2\)\)\) + a\^2\/\(B\^2\ F\ G\^2\) - \(m\ r\)\/\(B\^2\ F\ G\^2\) + 
          r\^2\/\(B\^2\ F\ G\^2\)\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2]], 
          Times[ -1, 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], m, r], 
          Times[ 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\ r\)\/\(F\ G\
\^2\)\),
      SequenceForm[ "[", 3, ",", 4, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m, r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\ r\)\/\(F\ G\
\^2\)\),
      SequenceForm[ "[", 4, ",", 3, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m, r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(a\^4\/\(F\ G\
\^2\)\)\) + \(a\^4\ B\^2\)\/\(F\ G\^2\) - \(a\^2\ B\^2\ m\ r\)\/\(F\ G\^2\) - \
\(2\ a\^2\ r\^2\)\/\(F\ G\^2\) + \(a\^2\ B\^2\ r\^2\)\/\(F\ G\^2\) - 
          r\^4\/\(F\ G\^2\)\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 4], 
            Power[ F, -1], 
            Power[ G, -2]], 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2]], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], m, r], 
          Times[ -2, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 2]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 4]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(irmdgp\  = \ 
      simpler[1/PowerExpand[Sqrt[simpler[\(-Det[g]\)]] /. rules]]\)], "Input"],

Cell[BoxData[
    \(1\/F\)], "Output"]
}, Open  ]],

Cell["\<\
Don't forget the sub[i] in the derivatives (takes care of change in \
time coord from i=10 to i=0 in c++)\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(Btab\  = \ {B^2, 1 - b^2, 
        Table[D[1 - b^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              B)\)}\), "\[IndentingNewLine]", 
    \(Ftab\  = \ {F, a\^2\ b\^2 + r\^2, 
        Table[D[a\^2\ b\^2 + r\^2, x[\([sub[i]]\)]], {i, 
            dim}]}\), "\[IndentingNewLine]", 
    \(Gtab\  = \ {G^2, a\^2 - m\ r + r\^2, 
        Table[D[a\^2 - m\ r + r\^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              G)\)}\), "\[IndentingNewLine]", 
    \(Ptab\  = \ {P^2, 1 - p^2, 
        Table[D[1 - p^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              P)\)}\), "\[IndentingNewLine]", 
    \(Qtab\  = \ {Q^2, 1 - q^2, 
        Table[D[1 - q^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              Q)\)}\), "\[IndentingNewLine]", 
    \(Stab\  = \ {S^2, 1 - s^2, 
        Table[D[1 - s^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              S)\)}\), "\[IndentingNewLine]", 
    \(Utab\  = \ {U^2, 1 - u^2, 
        Table[D[1 - u^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              U)\)}\), "\[IndentingNewLine]", 
    \(Vtab\  = \ {V^2, 1 - v^2, 
        Table[D[1 - v^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              V)\)}\), "\[IndentingNewLine]", 
    \(Wtab\  = \ {W^2, 1 - w^2, 
        Table[D[1 - w^2, x[\([sub[i]]\)]], {i, dim}]/\((2  W)\)}\)}], "Input"],

Cell[BoxData[
    \({B\^2, 1 - b\^2, {0, 0, \(-\(b\/B\)\), 0}}\)], "Output"],

Cell[BoxData[
    \({F, a\^2\ b\^2 + r\^2, {0, 2\ r, 2\ a\^2\ b, 0}}\)], "Output"],

Cell[BoxData[
    \({G\^2, 
      a\^2 - m\ r + r\^2, {0, \(\(-m\) + 2\ r\)\/\(2\ G\), 0, 0}}\)], "Output"],

Cell[BoxData[
    \({P\^2, 1 - p\^2, {0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({Q\^2, 1 - q\^2, {0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({S\^2, 1 - s\^2, {0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({U\^2, 1 - u\^2, {0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({V\^2, 1 - v\^2, {0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({W\^2, 1 - w\^2, {0, 0, 0, 0}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(coords\  = \ x\), "\[IndentingNewLine]", 
    \(ints\  = \ {2, 3}\), "\n", 
    \(constants\  = \ {I, m, a}\), "\n", 
    \(functions\  = \ 
      Take[{Ftab, Gtab, Btab, Ptab, Qtab, Stab, Utab, Vtab, Wtab}, 
        If[dim \[Equal] 4, 3, dim - 2]]\), "\[IndentingNewLine]", 
    \(ansatze\  = \ {}\)}], "Input"],

Cell[BoxData[
    \({r, b, y, t}\)], "Output"],

Cell[BoxData[
    \({2, 3}\)], "Output"],

Cell[BoxData[
    \({\[ImaginaryI], m, a}\)], "Output"],

Cell[BoxData[
    \({{F, a\^2\ b\^2 + r\^2, {0, 2\ r, 2\ a\^2\ b, 0}}, {G\^2, 
        a\^2 - m\ r + r\^2, {0, \(\(-m\) + 2\ r\)\/\(2\ G\), 0, 0}}, {B\^2, 
        1 - b\^2, {0, 0, \(-\(b\/B\)\), 0}}}\)], "Output"],

Cell[BoxData[
    \({}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(headers[];\)\)], "Input"],

Cell[BoxData[
    \("functions completed..."\)], "Print"],

Cell[BoxData[
    \("g completed..."\)], "Print"],

Cell[BoxData[
    \("gi completed..."\)], "Print"],

Cell[BoxData[
    \("d completed..."\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(tensor[ad2, "\<ad2\>"];\)\), "\[IndentingNewLine]", 
    \(\(tensor[ad3, "\<ad3\>"];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("ad2"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ad2", " completed..."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("ad3"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ad3", " completed..."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(backup[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]3\[InvisibleSpace]":"\
\[InvisibleSpace]33\[InvisibleSpace]":"\[InvisibleSpace]4\[InvisibleSpace]" \
on "\[InvisibleSpace]3\[InvisibleSpace]"/"\[InvisibleSpace]31\[InvisibleSpace]\
"/"\[InvisibleSpace]2003\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 3, ":", 33, ":", 4, " on ", 3, "/", 31, "/", 2003, "."],
      Editable->False]], "Print"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "ken", \
"projects", "mathematica", "packages"}, "d11brane.nb.ps", CharacterEncoding \
-> "ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 70, 1, 27, "Input"],
Cell[1874, 57, 2984, 44, 375, "Print"],
Cell[4861, 103, 81, 1, 23, "Print"]
}, Open  ]],
Cell[4957, 107, 49, 1, 27, "Input"],
Cell[5009, 110, 451, 9, 91, "Input"],
Cell[5463, 121, 536, 11, 91, "Input"],
Cell[6002, 134, 74, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[6101, 139, 2298, 44, 379, "Input"],
Cell[8402, 185, 297, 5, 23, "Print"],
Cell[8702, 192, 573, 17, 41, "Print"],
Cell[9278, 211, 395, 11, 25, "Print"],
Cell[9676, 224, 1075, 34, 43, "Print"],
Cell[10754, 260, 529, 15, 43, "Print"],
Cell[11286, 277, 529, 15, 43, "Print"],
Cell[11818, 294, 692, 21, 43, "Print"],
Cell[12513, 317, 293, 5, 23, "Print"],
Cell[12809, 324, 543, 16, 41, "Print"],
Cell[13355, 342, 547, 15, 43, "Print"],
Cell[13905, 359, 1121, 33, 41, "Print"],
Cell[15029, 394, 533, 14, 40, "Print"],
Cell[15565, 410, 533, 14, 40, "Print"],
Cell[16101, 426, 631, 18, 41, "Print"]
}, Open  ]],
Cell[16747, 447, 97, 3, 32, "Text"],
Cell[16847, 452, 717, 13, 83, "Input"],

Cell[CellGroupData[{
Cell[17589, 469, 120, 2, 43, "Input"],
Cell[17712, 473, 274, 6, 39, "Print"],
Cell[17989, 481, 274, 6, 39, "Print"],
Cell[18266, 489, 1060, 33, 41, "Print"],
Cell[19329, 524, 320, 8, 41, "Print"],
Cell[19652, 534, 320, 8, 41, "Print"],
Cell[19975, 544, 320, 8, 37, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[20332, 557, 154, 3, 43, "Input"],
Cell[20489, 562, 296, 7, 41, "Print"],
Cell[20788, 571, 296, 7, 41, "Print"],
Cell[21087, 580, 841, 25, 41, "Print"],
Cell[21931, 607, 325, 8, 37, "Print"],
Cell[22259, 617, 325, 8, 37, "Print"],
Cell[22587, 627, 1206, 37, 41, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[23830, 669, 112, 2, 27, "Input"],
Cell[23945, 673, 38, 1, 44, "Output"]
}, Open  ]],
Cell[23998, 677, 129, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[24152, 684, 1273, 26, 164, "Input"],
Cell[25428, 712, 76, 1, 44, "Output"],
Cell[25507, 715, 82, 1, 29, "Output"],
Cell[25592, 718, 107, 2, 44, "Output"],
Cell[25702, 722, 64, 1, 29, "Output"],
Cell[25769, 725, 64, 1, 29, "Output"],
Cell[25836, 728, 64, 1, 29, "Output"],
Cell[25903, 731, 64, 1, 29, "Output"],
Cell[25970, 734, 64, 1, 29, "Output"],
Cell[26037, 737, 64, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[26138, 743, 333, 7, 107, "Input"],
Cell[26474, 752, 46, 1, 27, "Output"],
Cell[26523, 755, 40, 1, 27, "Output"],
Cell[26566, 758, 55, 1, 27, "Output"],
Cell[26624, 761, 214, 3, 69, "Output"],
Cell[26841, 766, 36, 1, 27, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[26914, 772, 47, 1, 27, "Input"],
Cell[26964, 775, 57, 1, 23, "Print"],
Cell[27024, 778, 49, 1, 23, "Print"],
Cell[27076, 781, 50, 1, 23, "Print"],
Cell[27129, 784, 49, 1, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[27215, 790, 123, 2, 43, "Input"],
Cell[27341, 794, 157, 3, 23, "Print"],
Cell[27501, 799, 157, 3, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[27695, 807, 46, 1, 27, "Input"],
Cell[27744, 810, 423, 7, 23, "Print"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

