(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     32422,        953]*)
(*NotebookOutlinePosition[     33412,        985]*)
(*  CellTagsIndexPosition[     33368,        981]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/ccif.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,<oldx,>func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("AVOID USING C, D, E, I, N and O as variables !"\)], "Print"]
}, Open  ]],

Cell[BoxData[
    \(\(dim\  = \ 6;\)\)], "Input"],

Cell[BoxData[{
    \(\(ad2\  = \ 
        Table[Table[0, {j, dim}], {i, dim}];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        Do[\((\[IndentingNewLine]inddim\  = \ 
              Join[DeleteCases[Join[Table[l, {l, dim - 1}], {0}], i | j], {i, 
                  j}]; \[IndentingNewLine]ad2[\([i, j]\)]\  = \ 
              Signature[inddim] 
                Factorial[dim - 2];\[IndentingNewLine])\), {j, dim}], {i, 
          dim}];\)\)}], "Input"],

Cell[BoxData[{
    \(\(ad3\  = \ 
        Table[Table[Table[0, {k, dim}], {j, dim}], {i, 
            dim}];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        Do[Do[\((\[IndentingNewLine]inddim\  = \ 
                Join[DeleteCases[Join[Table[l, {l, dim - 1}], {0}], 
                    i | j | k], {i, j, k}]; \[IndentingNewLine]ad3[\([i, j, 
                    k]\)]\  = \ 
                Signature[inddim] 
                  Factorial[dim - 3];\[IndentingNewLine])\), {k, dim}], {j, 
            dim}], {i, dim}];\)\)}], "Input"],

Cell[BoxData[
    \(\(ptahdir\  = \ "\<~/projects/ptah/\>";\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(xp\  = \ 
        Flatten[{Table[Symbol["\<t\>" <> ToString[i]], {i, dim - 5}], 
            If[dim \[GreaterEqual] 5, p2, {}], r, th, ph, 
            t}];\)\), "\[IndentingNewLine]", 
    \(\(rho\  = \ Sqrt[r^2 + a^2  Cos[th]^2];\)\), "\[IndentingNewLine]", 
    \(\(delta\  = \ 
        m/\((r^\((dim - 5)\) rho^2)\);\)\), "\[IndentingNewLine]", 
    \(\(psi\  = \ 
        r^\((dim - 5)\) 
          rho^2/\((r^\((dim - 5)\) \((r^2 + a^2)\) - 
                m)\);\)\), "\[IndentingNewLine]", 
    \(\(gp\  = \ 0*IdentityMatrix[dim];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        gp[\([i, i]\)]\  = \ 
          r^2  Cos[th]^2\ Product[Sin[xp[\([j - 1]\)]]^2, {j, 2, i}], {i, 1, 
          dim - 4}];\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim - 3, dim - 3]\)]\  = \ psi;\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim - 2, dim - 2]\)]\  = \ rho^2;\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim - 1, dim - 1]\)]\  = \ \((r^2 + a^2)\) Sin[th]^2 + 
          delta*a^2  Sin[th]^4;\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim - 1, dim]\)]\  = \ 
        delta*a*Sin[th]^2;\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim, dim - 1]\)]\  = \ 
        gp[\([dim - 1, dim]\)];\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim, dim]\)]\  = \ \(-1\) + delta;\)\), "\[IndentingNewLine]", 
    \(\(gp\  = \ simpler[gp];\)\), "\[IndentingNewLine]", 
    \(\(Print["\<dim = \>", dim, "\<, x = \>", 
        xp, "\<, metric components are:\>"];\)\), "\[IndentingNewLine]", 
    \(\(components[gp];\)\), "\[IndentingNewLine]", 
    \(\(If[dim \[Equal] 4, 
        x = {r, b, y, t}, \[IndentingNewLine]x\  = \ 
          Flatten[{Take[{p, q, s, u, v, w}, dim - 5], 
              If[dim \[GreaterEqual] 5, z, {}], r, b, y, 
              t}]];\)\), "\[IndentingNewLine]", 
    \(\(g\  = \ 
        simpler[coordxform[gp, xp, 
            If[dim \[Equal] 4, {r, ArcCos[b], y, 
                t}, \[IndentingNewLine]Flatten[{Thread[
                    ArcCos[Take[{p, q, s, u, v, w}, dim - 5]]], 
                  If[dim \[GreaterEqual] 5, z, {}], r, ArcCos[b], y, t}]], 
            x]];\)\), "\[IndentingNewLine]", 
    \(\(Print["\<dim = \>", dim, "\<, x = \>", 
        x, "\<, metric components are:\>"];\)\), "\[IndentingNewLine]", 
    \(\(components[g];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("dim = "\[InvisibleSpace]6\[InvisibleSpace]", x = "\
\[InvisibleSpace]{t1, p2, r, th, ph, 
          t}\[InvisibleSpace]", metric components are:"\),
      SequenceForm[ 
      "dim = ", 6, ", x = ", {t1, p2, r, th, ph, t}, 
        ", metric components are:"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\ \
Cos[th]\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Cos[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\ Cos[th]\^2\ \
Sin[t1]\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Cos[ th], 2], 
          Power[ 
            Sin[ t1], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\ \((r\^2 + a\^2\
\ Cos[th]\^2)\)\)\/\(\(-m\) + a\^2\ r + r\^3\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ r, 
          Power[ 
            Plus[ 
              Times[ -1, m], 
              Times[ 
                Power[ a, 2], r], 
              Power[ r, 3]], -1], 
          Plus[ 
            Power[ r, 2], 
            Times[ 
              Power[ a, 2], 
              Power[ 
                Cos[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2 + 
          a\^2\ Cos[th]\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Plus[ 
          Power[ r, 2], 
          Times[ 
            Power[ a, 2], 
            Power[ 
              Cos[ th], 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(Sin[th]\^2\ \
\((a\^2\ r\^3 + r\^5 + a\^4\ r\ Cos[th]\^2 + a\^2\ r\^3\ Cos[th]\^2 + a\^2\ m\
\ Sin[th]\^2)\)\)\/\(r\ \((r\^2 + a\^2\ Cos[th]\^2)\)\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ r, -1], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Power[ r, 3]], 
            Power[ r, 5], 
            Times[ 
              Power[ a, 4], r, 
              Power[ 
                Cos[ th], 2]], 
            Times[ 
              Power[ a, 2], 
              Power[ r, 3], 
              Power[ 
                Cos[ th], 2]], 
            Times[ 
              Power[ a, 2], m, 
              Power[ 
                Sin[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\ \
Sin[th]\^2\)\/\(r\ \((r\^2 + a\^2\ Cos[th]\^2)\)\)\),
      SequenceForm[ "[", 5, ",", 6, "] = ", 
        Times[ a, m, 
          Power[ r, -1], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\ \
Sin[th]\^2\)\/\(r\ \((r\^2 + a\^2\ Cos[th]\^2)\)\)\),
      SequenceForm[ "[", 6, ",", 5, "] = ", 
        Times[ a, m, 
          Power[ r, -1], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(m - r\^3 - a\^2\ \
r\ Cos[th]\^2\)\/\(r\ \((r\^2 + a\^2\ Cos[th]\^2)\)\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ r, -1], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Plus[ m, 
            Times[ -1, 
              Power[ r, 3]], 
            Times[ -1, 
              Power[ a, 2], r, 
              Power[ 
                Cos[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("dim = "\[InvisibleSpace]6\[InvisibleSpace]", x = "\
\[InvisibleSpace]{p, z, r, b, y, 
          t}\[InvisibleSpace]", metric components are:"\),
      SequenceForm[ 
      "dim = ", 6, ", x = ", {p, z, r, b, y, t}, ", metric components are:"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(b\^2\ r\^2\)\
\/\(\((\(-1\) + p)\)\ \((1 + p)\)\)\)\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ -1, 
          Power[ b, 2], 
          Power[ 
            Plus[ -1, p], -1], 
          Power[ 
            Plus[ 1, p], -1], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-b\^2\)\ \
\((\(-1\) + p)\)\ \((1 + p)\)\ r\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ -1, 
          Power[ b, 2], 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\ \((a\^2\ b\^2 \
+ r\^2)\)\)\/\(\(-m\) + a\^2\ r + r\^3\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ r, 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Power[ b, 2]], 
            Power[ r, 2]], 
          Power[ 
            Plus[ 
              Times[ -1, m], 
              Times[ 
                Power[ a, 2], r], 
              Power[ r, 3]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(a\^2\ b\^2 + 
                r\^2\)\/\(\((\(-1\) + b)\)\ \((1 + b)\)\)\)\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ -1, 
          Power[ 
            Plus[ -1, b], -1], 
          Power[ 
            Plus[ 1, b], -1], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Power[ b, 2]], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(\((\(-1\) + 
                    b)\)\ \((1 + b)\)\ \((a\^2\ m - a\^2\ b\^2\ m + 
                    a\^4\ b\^2\ r + a\^2\ r\^3 + a\^2\ b\^2\ r\^3 + 
                    r\^5)\)\)\/\(r\ \((a\^2\ b\^2 + r\^2)\)\)\)\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ -1, 
          Plus[ -1, b], 
          Plus[ 1, b], 
          Power[ r, -1], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ a, 2], 
                Power[ b, 2]], 
              Power[ r, 2]], -1], 
          Plus[ 
            Times[ 
              Power[ a, 2], m], 
            Times[ -1, 
              Power[ a, 2], 
              Power[ b, 2], m], 
            Times[ 
              Power[ a, 4], 
              Power[ b, 2], r], 
            Times[ 
              Power[ a, 2], 
              Power[ r, 3]], 
            Times[ 
              Power[ a, 2], 
              Power[ b, 2], 
              Power[ r, 3]], 
            Power[ r, 5]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(a\ \((\(-1\) \
+ b)\)\ \((1 + b)\)\ m\)\/\(r\ \((a\^2\ b\^2 + r\^2)\)\)\)\)\),
      SequenceForm[ "[", 5, ",", 6, "] = ", 
        Times[ -1, a, 
          Plus[ -1, b], 
          Plus[ 1, b], m, 
          Power[ r, -1], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ a, 2], 
                Power[ b, 2]], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(a\ \((\(-1\) \
+ b)\)\ \((1 + b)\)\ m\)\/\(r\ \((a\^2\ b\^2 + r\^2)\)\)\)\)\),
      SequenceForm[ "[", 6, ",", 5, "] = ", 
        Times[ -1, a, 
          Plus[ -1, b], 
          Plus[ 1, b], m, 
          Power[ r, -1], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ a, 2], 
                Power[ b, 2]], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(\(-m\) + 
                a\^2\ b\^2\ r + r\^3\)\/\(r\ \((a\^2\ b\^2 + r\^2)\)\)\)\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ -1, 
          Power[ r, -1], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ a, 2], 
                Power[ b, 2]], 
              Power[ r, 2]], -1], 
          Plus[ 
            Times[ -1, m], 
            Times[ 
              Power[ a, 2], 
              Power[ b, 2], r], 
            Power[ r, 3]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell["\<\
We next parameterize the metric so that each component is a \
polynomial:\
\>", "Text"],

Cell[BoxData[
    \(\(rules\  = \ {a\^2\ b\^2 + r\^2 \[Rule] 
            F, \(a\^2\) r - m\  + r\^3 \[Rule] 
            G^2, \[IndentingNewLine]\((\(-1\) + 
                b)\) \[Rule] \(-B^2\)/\((1 + b)\), 
          b^2 \[Rule] 
            1 - B^2, \[IndentingNewLine]\((\(-1\) + 
                p)\) \[Rule] \(-P^2\)/\((1 + p)\), 
          p^2 \[Rule] 
            1 - P^2, \((\(-1\) + 
                q)\) \[Rule] \(-Q^2\)/\((1 + q)\), \((\(-1\) + 
                s)\) \[Rule] \(-S^2\)/\((1 + s)\), \((\(-1\) + 
                u)\) \[Rule] \(-U^2\)/\((1 + u)\), \((\(-1\) + 
                v)\) \[Rule] \(-V^2\)/\((1 + v)\), \((\(-1\) + 
                w)\) \[Rule] \(-W^2\)/\((1 + w)\)};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(gp\  = \ Expand[Factor[g] //. rules];\)\), "\[IndentingNewLine]", 
    \(\(components[gp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\/P\^2 - \
\(B\^2\ r\^2\)\/P\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, -2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, -2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(P\^2\ r\^2 - 
          B\^2\ P\^2\ r\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(F\ r\)\/G\^2\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ F, 
          Power[ G, -2], r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]F\/B\^2\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ B, -2], F]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(a\^4\ B\^2\)\/F \
- \(a\^4\ B\^4\)\/F + \(a\^2\ B\^4\ m\)\/\(F\ r\) + \(2\ a\^2\ B\^2\ \
r\^2\)\/F - \(a\^2\ B\^4\ r\^2\)\/F + \(B\^2\ r\^4\)\/F\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Plus[ 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1]], 
          Times[ -1, 
            Power[ a, 4], 
            Power[ B, 4], 
            Power[ F, -1]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], m, 
            Power[ r, -1]], 
          Times[ 2, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ m\)\/\(F\
\ r\)\),
      SequenceForm[ "[", 5, ",", 6, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, 
          Power[ r, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ m\)\/\(F\
\ r\)\),
      SequenceForm[ "[", 6, ",", 5, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, 
          Power[ r, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(a\^2\/F\)\) \
+ \(a\^2\ B\^2\)\/F + m\/\(F\ r\) - r\^2\/F\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 2], 
            Power[ F, -1]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1]], 
          Times[ 
            Power[ F, -1], m, 
            Power[ r, -1]], 
          Times[ -1, 
            Power[ F, -1], 
            Power[ r, 2]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(ginvp\  = \ 
        Expand[Simplify[
            Factor[Inverse[g]] //. rules]];\)\), "\[IndentingNewLine]", 
    \(\(components[ginvp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]P\^2\/\(b\^2\ \
r\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, 2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]1\/\(b\^2\ P\^2\ \
r\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]G\^2\/\(F\ r\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ F, -1], 
          Power[ G, 2], 
          Power[ r, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]B\^2\/F\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ B, 2], 
          Power[ F, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(m\/\(B\^2\ F\
\ G\^2\)\)\) - \(a\^2\ r\)\/\(F\ G\^2\) + \(a\^2\ r\)\/\(B\^2\ F\ G\^2\) + 
          r\^3\/\(B\^2\ F\ G\^2\)\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], m], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2], r], 
          Times[ 
            Power[ a, 2], 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], r], 
          Times[ 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 3]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\)\/\(F\ G\^2\
\)\),
      SequenceForm[ "[", 5, ",", 6, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\)\/\(F\ G\^2\
\)\),
      SequenceForm[ "[", 6, ",", 5, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(\(a\^2\ B\^2\
\ m\)\/\(F\ G\^2\)\)\) - \(a\^4\ r\)\/\(F\ G\^2\) + \(a\^4\ B\^2\ r\)\/\(F\ G\
\^2\) - \(2\ a\^2\ r\^3\)\/\(F\ G\^2\) + \(a\^2\ B\^2\ r\^3\)\/\(F\ G\^2\) - 
          r\^5\/\(F\ G\^2\)\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], m], 
          Times[ -1, 
            Power[ a, 4], 
            Power[ F, -1], 
            Power[ G, -2], r], 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], r], 
          Times[ -2, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 3]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 3]], 
          Times[ -1, 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 5]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(irmdgp\  = \ 
      simpler[1/PowerExpand[Sqrt[simpler[\(-Det[g]\)]] /. rules]]\)], "Input"],

Cell[BoxData[
    \(1\/\(b\^2\ F\ r\^2\)\)], "Output"]
}, Open  ]],

Cell["\<\
Don't forget the sub[i] in the derivatives (takes care of change in \
time coord from i=10 to i=0 in c++)\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(Btab\  = \ {B^2, 1 - b^2, 
        Table[D[1 - b^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              B)\)}\), "\[IndentingNewLine]", 
    \(Ftab\  = \ {F, a\^2\ b\^2 + r\^2, 
        Table[D[a\^2\ b\^2 + r\^2, x[\([sub[i]]\)]], {i, 
            dim}]}\), "\[IndentingNewLine]", 
    \(Gtab\  = \ {G^2, \(a\^2\) r - m\  + r\^3, 
        Table[D[\(a\^2\) r - m\  + r\^3, x[\([sub[i]]\)]], {i, dim}]/\((2  
              G)\)}\), "\[IndentingNewLine]", 
    \(Ptab\  = \ {P^2, 1 - p^2, 
        Table[D[1 - p^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              P)\)}\), "\[IndentingNewLine]", 
    \(Qtab\  = \ {Q^2, 1 - q^2, 
        Table[D[1 - q^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              Q)\)}\), "\[IndentingNewLine]", 
    \(Stab\  = \ {S^2, 1 - s^2, 
        Table[D[1 - s^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              S)\)}\), "\[IndentingNewLine]", 
    \(Utab\  = \ {U^2, 1 - u^2, 
        Table[D[1 - u^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              U)\)}\), "\[IndentingNewLine]", 
    \(Vtab\  = \ {V^2, 1 - v^2, 
        Table[D[1 - v^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              V)\)}\), "\[IndentingNewLine]", 
    \(Wtab\  = \ {W^2, 1 - w^2, 
        Table[D[1 - w^2, x[\([sub[i]]\)]], {i, dim}]/\((2  W)\)}\)}], "Input"],

Cell[BoxData[
    \({B\^2, 1 - b\^2, {0, 0, 0, 0, \(-\(b\/B\)\), 0}}\)], "Output"],

Cell[BoxData[
    \({F, a\^2\ b\^2 + r\^2, {0, 0, 0, 2\ r, 2\ a\^2\ b, 0}}\)], "Output"],

Cell[BoxData[
    \({G\^2, \(-m\) + a\^2\ r + r\^3, {0, 0, 0, \(a\^2 + 3\ r\^2\)\/\(2\ G\), 
        0, 0}}\)], "Output"],

Cell[BoxData[
    \({P\^2, 1 - p\^2, {0, \(-\(p\/P\)\), 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({Q\^2, 1 - q\^2, {0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({S\^2, 1 - s\^2, {0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({U\^2, 1 - u\^2, {0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({V\^2, 1 - v\^2, {0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({W\^2, 1 - w\^2, {0, 0, 0, 0, 0, 0}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(coords\  = \ x\), "\[IndentingNewLine]", 
    \(ints\  = \ {2, 3}\), "\n", 
    \(constants\  = \ {I, m, a}\), "\n", 
    \(functions\  = \ 
      Take[{Ftab, Gtab, Btab, Ptab, Qtab, Stab, Utab, Vtab, Wtab}, 
        If[dim \[Equal] 4, 3, dim - 2]]\), "\[IndentingNewLine]", 
    \(ansatze\  = \ {}\)}], "Input"],

Cell[BoxData[
    \({p, z, r, b, y, t}\)], "Output"],

Cell[BoxData[
    \({2, 3}\)], "Output"],

Cell[BoxData[
    \({\[ImaginaryI], m, a}\)], "Output"],

Cell[BoxData[
    \({{F, 
        a\^2\ b\^2 + r\^2, {0, 0, 0, 2\ r, 2\ a\^2\ b, 0}}, {G\^2, \(-m\) + 
          a\^2\ r + r\^3, {0, 0, 0, \(a\^2 + 3\ r\^2\)\/\(2\ G\), 0, 
          0}}, {B\^2, 1 - b\^2, {0, 0, 0, 0, \(-\(b\/B\)\), 0}}, {P\^2, 
        1 - p\^2, {0, \(-\(p\/P\)\), 0, 0, 0, 0}}}\)], "Output"],

Cell[BoxData[
    \({}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(headers[];\)\)], "Input"],

Cell[BoxData[
    \("functions completed..."\)], "Print"],

Cell[BoxData[
    \("g completed..."\)], "Print"],

Cell[BoxData[
    \("gi completed..."\)], "Print"],

Cell[BoxData[
    \("d completed..."\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(tensor[ad2, "\<ad2\>"];\)\), "\[IndentingNewLine]", 
    \(\(tensor[ad3, "\<ad3\>"];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("ad2"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ad2", " completed..."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("ad3"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ad3", " completed..."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(backup[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]22\[InvisibleSpace]":"\
\[InvisibleSpace]28\[InvisibleSpace]":"\[InvisibleSpace]2\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]2\[InvisibleSpace]\
"/"\[InvisibleSpace]2003\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 22, ":", 28, ":", 2, " on ", 4, "/", 2, "/", 2003, "."],
      Editable->False]], "Print"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "ken", \
"projects", "mathematica", "packages"}, "d11brane.nb.ps", CharacterEncoding \
-> "ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 70, 1, 27, "Input"],
Cell[1874, 57, 2984, 44, 375, "Print"],
Cell[4861, 103, 81, 1, 23, "Print"]
}, Open  ]],
Cell[4957, 107, 49, 1, 27, "Input"],
Cell[5009, 110, 451, 9, 91, "Input"],
Cell[5463, 121, 536, 11, 91, "Input"],
Cell[6002, 134, 74, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[6101, 139, 2298, 44, 379, "Input"],
Cell[8402, 185, 333, 7, 23, "Print"],
Cell[8738, 194, 330, 9, 25, "Print"],
Cell[9071, 205, 387, 11, 25, "Print"],
Cell[9461, 218, 615, 18, 41, "Print"],
Cell[10079, 238, 395, 11, 25, "Print"],
Cell[10477, 251, 1110, 35, 43, "Print"],
Cell[11590, 288, 558, 16, 43, "Print"],
Cell[12151, 306, 558, 16, 43, "Print"],
Cell[12712, 324, 706, 21, 43, "Print"],
Cell[13421, 347, 316, 6, 23, "Print"],
Cell[13740, 355, 451, 12, 44, "Print"],
Cell[14194, 369, 393, 10, 25, "Print"],
Cell[14590, 381, 585, 17, 41, "Print"],
Cell[15178, 400, 547, 15, 43, "Print"],
Cell[15728, 417, 1150, 34, 43, "Print"],
Cell[16881, 453, 562, 15, 42, "Print"],
Cell[17446, 470, 562, 15, 42, "Print"],
Cell[18011, 487, 671, 19, 43, "Print"]
}, Open  ]],
Cell[18697, 509, 97, 3, 32, "Text"],
Cell[18797, 514, 718, 14, 83, "Input"],

Cell[CellGroupData[{
Cell[19540, 532, 129, 2, 43, "Input"],
Cell[19672, 536, 455, 13, 41, "Print"],
Cell[20130, 551, 458, 13, 25, "Print"],
Cell[20591, 566, 284, 6, 39, "Print"],
Cell[20878, 574, 274, 6, 39, "Print"],
Cell[21155, 582, 1089, 34, 41, "Print"],
Cell[22247, 618, 347, 9, 41, "Print"],
Cell[22597, 629, 347, 9, 41, "Print"],
Cell[22947, 640, 638, 19, 41, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[23622, 664, 176, 4, 43, "Input"],
Cell[23801, 670, 337, 9, 41, "Print"],
Cell[24141, 681, 341, 9, 40, "Print"],
Cell[24485, 692, 329, 8, 41, "Print"],
Cell[24817, 702, 296, 7, 41, "Print"],
Cell[25116, 711, 851, 25, 41, "Print"],
Cell[25970, 738, 319, 8, 37, "Print"],
Cell[26292, 748, 319, 8, 37, "Print"],
Cell[26614, 758, 1216, 37, 41, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[27867, 800, 112, 2, 27, "Input"],
Cell[27982, 804, 54, 1, 44, "Output"]
}, Open  ]],
Cell[28051, 808, 129, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[28205, 815, 1283, 26, 164, "Input"],
Cell[29491, 843, 82, 1, 44, "Output"],
Cell[29576, 846, 88, 1, 29, "Output"],
Cell[29667, 849, 121, 2, 45, "Output"],
Cell[29791, 853, 82, 1, 41, "Output"],
Cell[29876, 856, 70, 1, 29, "Output"],
Cell[29949, 859, 70, 1, 29, "Output"],
Cell[30022, 862, 70, 1, 29, "Output"],
Cell[30095, 865, 70, 1, 29, "Output"],
Cell[30168, 868, 70, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[30275, 874, 333, 7, 107, "Input"],
Cell[30611, 883, 52, 1, 27, "Output"],
Cell[30666, 886, 40, 1, 27, "Output"],
Cell[30709, 889, 55, 1, 27, "Output"],
Cell[30767, 892, 310, 5, 81, "Output"],
Cell[31080, 899, 36, 1, 27, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[31153, 905, 47, 1, 27, "Input"],
Cell[31203, 908, 57, 1, 23, "Print"],
Cell[31263, 911, 49, 1, 23, "Print"],
Cell[31315, 914, 50, 1, 23, "Print"],
Cell[31368, 917, 49, 1, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[31454, 923, 123, 2, 43, "Input"],
Cell[31580, 927, 157, 3, 23, "Print"],
Cell[31740, 932, 157, 3, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[31934, 940, 46, 1, 27, "Input"],
Cell[31983, 943, 423, 7, 23, "Print"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

