(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     22230,        683]*)
(*NotebookOutlinePosition[     23217,        715]*)
(*  CellTagsIndexPosition[     23173,        711]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/ccif.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,<oldx,>func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("AVOID USING C, D, E, I, N and O as variables !"\)], "Print"]
}, Closed]],

Cell[BoxData[{
    \(\(f\  = \ Table[0, {i, 50}];\)\), "\[IndentingNewLine]", 
    \(\(derivs\  = \ 
        Table[Table[0, {i, 1000}], {j, 50}];\)\), "\[IndentingNewLine]", 
    \(\(g\  = Table[Table[0, {i, 11}], {j, 11}]\ ;\)\), "\n", 
    \(\(ginv\  = \ 
        Table[Table[0, {i, 11}], {j, 11}];\)\), "\[IndentingNewLine]", 
    \(\(Euler\  = \ 0;\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< /home/ken/projects/ptah/ptah.out\)], "Input"],

Cell[BoxData[
    InterpretationBox[\(0\[InvisibleSpace]1\),
      SequenceForm[ 0, 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(1\[InvisibleSpace]2\),
      SequenceForm[ 1, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(2\[InvisibleSpace]2\),
      SequenceForm[ 2, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(3\[InvisibleSpace]2\),
      SequenceForm[ 3, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(4\[InvisibleSpace]2\),
      SequenceForm[ 4, 2],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[f];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(a\^2\ b\^2 + r\^2\)\),
      SequenceForm[ "[", 1, "] = ", 
        Plus[ 
          Times[ 
            Power[ a, 2], 
            Power[ b, 2]], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(\(-m\) + a\^2\ r\^2 + r\^4\)\),
      SequenceForm[ "[", 2, "] = ", 
        Plus[ 
          Times[ -1, m], 
          Times[ 
            Power[ a, 2], 
            Power[ r, 2]], 
          Power[ r, 4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - b\^2\)\),
      SequenceForm[ "[", 3, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ b, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - p\^2\)\),
      SequenceForm[ "[", 4, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ p, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - q\^2\)\),
      SequenceForm[ "[", 5, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ q, 2]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[derivs];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ r\)\),
      SequenceForm[ "[", 1, ",", 5, "] = ", 
        Times[ 2, r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ a\^2\ b\)\),
      SequenceForm[ "[", 1, ",", 6, "] = ", 
        Times[ 2, 
          Power[ a, 2], b]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]12\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(a\^2\ r\)\/G + \
\(2\ r\^3\)\/G\)\),
      SequenceForm[ "[", 2, ",", 12, "] = ", 
        Plus[ 
          Times[ 
            Power[ a, 2], 
            Power[ G, -1], r], 
          Times[ 2, 
            Power[ G, -1], 
            Power[ r, 3]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]20\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(b\/B\)\)\),
      SequenceForm[ "[", 3, ",", 20, "] = ", 
        Times[ -1, b, 
          Power[ B, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]23\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(p\/P\)\)\),
      SequenceForm[ "[", 4, ",", 23, "] = ", 
        Times[ -1, p, 
          Power[ P, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]31\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(q\/Q\)\)\),
      SequenceForm[ "[", 5, ",", 31, "] = ", 
        Times[ -1, q, 
          Power[ Q, -1]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[g];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\/P\^2 - \
\(B\^2\ r\^2\)\/P\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, -2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, -2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ r\^2\)\/Q\
\^2 - \(B\^2\ P\^2\ r\^2\)\/Q\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, -2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, -2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(P\^2\ Q\^2\ r\^2 \
- B\^2\ P\^2\ Q\^2\ r\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(F\ r\^2\)\/G\^2\),
      
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ F, 
          Power[ G, -2], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]F\/B\^2\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ B, -2], F]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(a\^4\ B\^2\)\/F \
- \(a\^4\ B\^4\)\/F + \(a\^2\ B\^4\ m\)\/\(F\ r\^2\) + \(2\ a\^2\ B\^2\ \
r\^2\)\/F - \(a\^2\ B\^4\ r\^2\)\/F + \(B\^2\ r\^4\)\/F\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Plus[ 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1]], 
          Times[ -1, 
            Power[ a, 4], 
            Power[ B, 4], 
            Power[ F, -1]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], m, 
            Power[ r, -2]], 
          Times[ 2, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ m\)\/\(F\
\ r\^2\)\),
      SequenceForm[ "[", 6, ",", 7, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ m\)\/\(F\
\ r\^2\)\),
      SequenceForm[ "[", 7, ",", 6, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(a\^2\/F\)\) \
+ \(a\^2\ B\^2\)\/F + m\/\(F\ r\^2\) - r\^2\/F\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 2], 
            Power[ F, -1]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1]], 
          Times[ 
            Power[ F, -1], m, 
            Power[ r, -2]], 
          Times[ -1, 
            Power[ F, -1], 
            Power[ r, 2]]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[ginv];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]P\^2\/\(b\^2\ \
r\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, 2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]Q\^2\/\(b\^2\ P\^2\ \
r\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, 2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]1\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]G\^2\/\(F\ r\^2\)\),
      
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ F, -1], 
          Power[ G, 2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]B\^2\/F\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ B, 2], 
          Power[ F, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(m\/\(B\^2\ F\
\ G\^2\)\)\) - \(a\^2\ r\^2\)\/\(F\ G\^2\) + \(a\^2\ r\^2\)\/\(B\^2\ F\ \
G\^2\) + r\^4\/\(B\^2\ F\ G\^2\)\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], m], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 2]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 2]], 
          Times[ 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\)\/\(F\ G\^2\
\)\),
      SequenceForm[ "[", 6, ",", 7, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\)\/\(F\ G\^2\
\)\),
      SequenceForm[ "[", 7, ",", 6, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(\(a\^2\ B\^2\
\ m\)\/\(F\ G\^2\)\)\) - \(a\^4\ r\^2\)\/\(F\ G\^2\) + \(a\^4\ B\^2\ r\^2\)\/\
\(F\ G\^2\) - \(2\ a\^2\ r\^4\)\/\(F\ G\^2\) + \(a\^2\ B\^2\ r\^4\)\/\(F\ \
G\^2\) - r\^6\/\(F\ G\^2\)\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], m], 
          Times[ -1, 
            Power[ a, 4], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 2]], 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 2]], 
          Times[ -2, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 4]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 4]], 
          Times[ -1, 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 6]]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(isrdmg\)], "Input"],

Cell[BoxData[
    \(1\/\(b\^3\ F\ P\ r\^3\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[sc]]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riesq]]\)], "Input"],

Cell[BoxData[
    \(\(24\ m\^2\ \((3\ a\^8\ b\^8 + 15\ a\^6\ b\^6\ r\^2 + 30\ a\^4\ b\^4\ r\
\^4 + 27\ a\^2\ b\^2\ r\^6 + 25\ r\^8)\)\)\/\(r\^8\ \((a\^2\ b\^2 + \
r\^2)\)\^6\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riecub1]]\)], "Input"],

Cell[BoxData[
    \(\(24\ m\^3\ \((8\ a\^12\ b\^12 + 57\ a\^10\ b\^10\ r\^2 + 186\ a\^8\ \
b\^8\ r\^4 + 382\ a\^6\ b\^6\ r\^6 + 576\ a\^4\ b\^4\ r\^8 + 513\ a\^2\ b\^2\ \
r\^10 + 310\ r\^12)\)\)\/\(r\^12\ \((a\^2\ b\^2 + r\^2)\)\^9\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riecub2]]\)], "Input"],

Cell[BoxData[
    \(\(6\ m\^3\ \((16\ a\^12\ b\^12 + 129\ a\^10\ b\^10\ r\^2 + 438\ a\^8\ b\
\^8\ r\^4 + 782\ a\^6\ b\^6\ r\^6 + 720\ a\^4\ b\^4\ r\^8 + 297\ a\^2\ b\^2\ \
r\^10 + 290\ r\^12)\)\)\/\(r\^12\ \((a\^2\ b\^2 + r\^2)\)\^9\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riecub3]]\)], "Input"],

Cell[BoxData[
    \(\(24\ m\^3\ \((2\ a\^12\ b\^12 + 18\ a\^10\ b\^10\ r\^2 + 63\ a\^8\ \
b\^8\ r\^4 + 100\ a\^6\ b\^6\ r\^6 + 36\ a\^4\ b\^4\ r\^8 - 54\ a\^2\ b\^2\ r\
\^10 - 5\ r\^12)\)\)\/\(r\^12\ \((a\^2\ b\^2 + r\^2)\)\^9\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Cdriesq1]]\)], "Input"],

Cell[BoxData[
    \(\(\(1\/\(r\^12\ \((a\^2\ b\^2 + r\^2)\)\^9\)\)\((576\ m\^2\ \((\(-3\)\ \
a\^12\ b\^12\ m + 3\ a\^14\ b\^12\ r\^2 - 22\ a\^10\ b\^10\ m\ r\^2 + 
            23\ a\^12\ b\^10\ r\^4 + 2\ a\^12\ b\^12\ r\^4 - 
            69\ a\^8\ b\^8\ m\ r\^4 + 75\ a\^10\ b\^8\ r\^6 + 
            16\ a\^10\ b\^10\ r\^6 - 120\ a\^6\ b\^6\ m\ r\^6 + 
            135\ a\^8\ b\^6\ r\^8 + 54\ a\^8\ b\^8\ r\^8 - 
            124\ a\^4\ b\^4\ m\ r\^8 + 140\ a\^6\ b\^4\ r\^10 + 
            104\ a\^6\ b\^6\ r\^10 - 64\ a\^2\ b\^2\ m\ r\^10 + 
            102\ a\^4\ b\^2\ r\^12 + 86\ a\^4\ b\^4\ r\^12 - 50\ m\ r\^12 + 
            42\ a\^2\ r\^14 + 72\ a\^2\ b\^2\ r\^14 + 
            50\ r\^16)\))\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[RstarR]]\)], "Input"],

Cell[BoxData[
    \(\(1382400\ m\^2\ \((3\ a\^8\ b\^8 + 15\ a\^6\ b\^6\ r\^2 + 30\ a\^4\ \
b\^4\ r\^4 + 27\ a\^2\ b\^2\ r\^6 + 25\ r\^8)\)\)\/\(r\^8\ \((a\^2\ b\^2 + \
r\^2)\)\^6\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[gRstarR]]\)], "Input"],

Cell[BoxData[
    \(\(55296\ m\^2\ \((3\ a\^8\ b\^8 + 15\ a\^6\ b\^6\ r\^2 + 30\ a\^4\ b\^4\
\ r\^4 + 27\ a\^2\ b\^2\ r\^6 + 25\ r\^8)\)\)\/\(r\^8\ \((a\^2\ b\^2 + \
r\^2)\)\^6\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(backup[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]20\[InvisibleSpace]":"\
\[InvisibleSpace]2\[InvisibleSpace]":"\[InvisibleSpace]34\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]8\[InvisibleSpace]\
"/"\[InvisibleSpace]2003\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 20, ":", 2, ":", 34, " on ", 4, "/", 8, "/", 2003, "."],
      Editable->False]], "Print"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "ken", \
"projects", "ptah", "mathematica"}, "dbraneval.nb.ps", CharacterEncoding -> \
"ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 70, 1, 27, "Input"],
Cell[1874, 57, 2984, 44, 375, "Print"],
Cell[4861, 103, 81, 1, 23, "Print"]
}, Closed]],
Cell[4957, 107, 368, 7, 88, "Input"],

Cell[CellGroupData[{
Cell[5350, 118, 68, 1, 27, "Input"],
Cell[5421, 121, 121, 3, 23, "Print"],
Cell[5545, 126, 121, 3, 23, "Print"],
Cell[5669, 131, 121, 3, 23, "Print"],
Cell[5793, 136, 121, 3, 23, "Print"],
Cell[5917, 141, 121, 3, 23, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[6075, 149, 51, 1, 24, "Input"],
Cell[6129, 152, 310, 9, 25, "Print"],
Cell[6442, 163, 345, 10, 25, "Print"],
Cell[6790, 175, 256, 7, 25, "Print"],
Cell[7049, 184, 256, 7, 25, "Print"],
Cell[7308, 193, 256, 7, 25, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[7601, 205, 56, 1, 27, "Input"],
Cell[7660, 208, 252, 5, 23, "Print"],
Cell[7915, 215, 283, 6, 25, "Print"],
Cell[8201, 223, 430, 12, 41, "Print"],
Cell[8634, 237, 286, 6, 40, "Print"],
Cell[8923, 245, 286, 6, 37, "Print"],
Cell[9212, 253, 286, 6, 38, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[9535, 264, 51, 1, 24, "Input"],
Cell[9589, 267, 455, 13, 41, "Print"],
Cell[10047, 282, 525, 15, 42, "Print"],
Cell[10575, 299, 515, 15, 25, "Print"],
Cell[11093, 316, 316, 8, 41, "Print"],
Cell[11412, 326, 274, 6, 39, "Print"],
Cell[11689, 334, 1092, 34, 41, "Print"],
Cell[12784, 370, 350, 9, 41, "Print"],
Cell[13137, 381, 350, 9, 41, "Print"],
Cell[13490, 392, 641, 19, 41, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[14168, 416, 54, 1, 24, "Input"],
Cell[14225, 419, 337, 9, 41, "Print"],
Cell[14565, 430, 369, 10, 41, "Print"],
Cell[14937, 442, 373, 10, 41, "Print"],
Cell[15313, 454, 339, 9, 41, "Print"],
Cell[15655, 465, 296, 7, 41, "Print"],
Cell[15954, 474, 896, 27, 41, "Print"],
Cell[16853, 503, 319, 8, 37, "Print"],
Cell[17175, 513, 319, 8, 37, "Print"],
Cell[17497, 523, 1261, 39, 41, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[18795, 567, 39, 1, 24, "Input"],
Cell[18837, 570, 57, 1, 44, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[18931, 576, 49, 1, 27, "Input"],
Cell[18983, 579, 35, 1, 27, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[19055, 585, 52, 1, 27, "Input"],
Cell[19110, 588, 189, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[19336, 596, 54, 1, 27, "Input"],
Cell[19393, 599, 247, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[19677, 607, 54, 1, 27, "Input"],
Cell[19734, 610, 248, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[20019, 618, 54, 1, 27, "Input"],
Cell[20076, 621, 242, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[20355, 629, 55, 1, 27, "Input"],
Cell[20413, 632, 719, 11, 121, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[21169, 648, 53, 1, 27, "Input"],
Cell[21225, 651, 194, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[21456, 659, 54, 1, 27, "Input"],
Cell[21513, 662, 192, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[21742, 670, 46, 1, 27, "Input"],
Cell[21791, 673, 423, 7, 23, "Print"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

