(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     24674,        765]*)
(*NotebookOutlinePosition[     25661,        797]*)
(*  CellTagsIndexPosition[     25617,        793]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/ccif.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,<oldx,>func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("AVOID USING C, D, E, I, N and O as variables !"\)], "Print"]
}, Closed]],

Cell[BoxData[{
    \(\(f\  = \ Table[0, {i, 50}];\)\), "\[IndentingNewLine]", 
    \(\(derivs\  = \ 
        Table[Table[0, {i, 1000}], {j, 50}];\)\), "\[IndentingNewLine]", 
    \(\(g\  = Table[Table[0, {i, 11}], {j, 11}]\ ;\)\), "\n", 
    \(\(ginv\  = \ 
        Table[Table[0, {i, 11}], {j, 11}];\)\), "\[IndentingNewLine]", 
    \(\(RstarRRR\  = \ 0;\)\), "\[IndentingNewLine]", 
    \(\(Euler\  = \ 0;\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< /home/ken/projects/ptah/ptah.out\)], "Input"],

Cell[BoxData[
    InterpretationBox[\(0\[InvisibleSpace]1\),
      SequenceForm[ 0, 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(1\[InvisibleSpace]2\),
      SequenceForm[ 1, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(2\[InvisibleSpace]2\),
      SequenceForm[ 2, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(3\[InvisibleSpace]2\),
      SequenceForm[ 3, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(4\[InvisibleSpace]2\),
      SequenceForm[ 4, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(5\[InvisibleSpace]2\),
      SequenceForm[ 5, 2],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[f];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(a\^2\ b\^2 + r\^2\)\),
      SequenceForm[ "[", 1, "] = ", 
        Plus[ 
          Times[ 
            Power[ a, 2], 
            Power[ b, 2]], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(\(-m\) + a\^2\ r\^3 + r\^5\)\),
      SequenceForm[ "[", 2, "] = ", 
        Plus[ 
          Times[ -1, m], 
          Times[ 
            Power[ a, 2], 
            Power[ r, 3]], 
          Power[ r, 5]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - b\^2\)\),
      SequenceForm[ "[", 3, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ b, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - p\^2\)\),
      SequenceForm[ "[", 4, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ p, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - q\^2\)\),
      SequenceForm[ "[", 5, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ q, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - s\^2\)\),
      SequenceForm[ "[", 6, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ s, 2]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[derivs];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ r\)\),
      SequenceForm[ "[", 1, ",", 6, "] = ", 
        Times[ 2, r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ a\^2\ b\)\),
      SequenceForm[ "[", 1, ",", 7, "] = ", 
        Times[ 2, 
          Power[ a, 2], b]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]14\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(3\ a\^2\ \
r\^2\)\/\(2\ G\) + \(5\ r\^4\)\/\(2\ G\)\)\),
      SequenceForm[ "[", 2, ",", 14, "] = ", 
        Plus[ 
          Times[ 
            Rational[ 3, 2], 
            Power[ a, 2], 
            Power[ G, -1], 
            Power[ r, 2]], 
          Times[ 
            Rational[ 5, 2], 
            Power[ G, -1], 
            Power[ r, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]23\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(b\/B\)\)\),
      SequenceForm[ "[", 3, ",", 23, "] = ", 
        Times[ -1, b, 
          Power[ B, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]26\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(p\/P\)\)\),
      SequenceForm[ "[", 4, ",", 26, "] = ", 
        Times[ -1, p, 
          Power[ P, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]35\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(q\/Q\)\)\),
      SequenceForm[ "[", 5, ",", 35, "] = ", 
        Times[ -1, q, 
          Power[ Q, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]44\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(s\/S\)\)\),
      SequenceForm[ "[", 6, ",", 44, "] = ", 
        Times[ -1, s, 
          Power[ S, -1]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[g];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\/P\^2 - \
\(B\^2\ r\^2\)\/P\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, -2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, -2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ r\^2\)\/Q\
\^2 - \(B\^2\ P\^2\ r\^2\)\/Q\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, -2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, -2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ Q\^2\ \
r\^2\)\/S\^2 - \(B\^2\ P\^2\ Q\^2\ r\^2\)\/S\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, -2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, -2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(P\^2\ Q\^2\ r\^2\ \
S\^2 - B\^2\ P\^2\ Q\^2\ r\^2\ S\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(F\ r\^3\)\/G\^2\),
      
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ F, 
          Power[ G, -2], 
          Power[ r, 3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]F\/B\^2\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ B, -2], F]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(a\^4\ B\^2\)\/F \
- \(a\^4\ B\^4\)\/F + \(a\^2\ B\^4\ m\)\/\(F\ r\^3\) + \(2\ a\^2\ B\^2\ \
r\^2\)\/F - \(a\^2\ B\^4\ r\^2\)\/F + \(B\^2\ r\^4\)\/F\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Plus[ 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1]], 
          Times[ -1, 
            Power[ a, 4], 
            Power[ B, 4], 
            Power[ F, -1]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], m, 
            Power[ r, -3]], 
          Times[ 2, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ m\)\/\(F\
\ r\^3\)\),
      SequenceForm[ "[", 7, ",", 8, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, 
          Power[ r, -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ m\)\/\(F\
\ r\^3\)\),
      SequenceForm[ "[", 8, ",", 7, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, 
          Power[ r, -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(a\^2\/F\)\) \
+ \(a\^2\ B\^2\)\/F + m\/\(F\ r\^3\) - r\^2\/F\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 2], 
            Power[ F, -1]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1]], 
          Times[ 
            Power[ F, -1], m, 
            Power[ r, -3]], 
          Times[ -1, 
            Power[ F, -1], 
            Power[ r, 2]]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[ginv];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]P\^2\/\(b\^2\ \
r\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, 2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]Q\^2\/\(b\^2\ P\^2\ \
r\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, 2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]S\^2\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]1\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\ S\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]G\^2\/\(F\ r\^3\)\),
      
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ F, -1], 
          Power[ G, 2], 
          Power[ r, -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]B\^2\/F\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ B, 2], 
          Power[ F, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(m\/\(B\^2\ F\
\ G\^2\)\)\) - \(a\^2\ r\^3\)\/\(F\ G\^2\) + \(a\^2\ r\^3\)\/\(B\^2\ F\ \
G\^2\) + r\^5\/\(B\^2\ F\ G\^2\)\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], m], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 3]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 3]], 
          Times[ 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 5]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\)\/\(F\ G\^2\
\)\),
      SequenceForm[ "[", 7, ",", 8, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\)\/\(F\ G\^2\
\)\),
      SequenceForm[ "[", 8, ",", 7, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(\(a\^2\ B\^2\
\ m\)\/\(F\ G\^2\)\)\) - \(a\^4\ r\^3\)\/\(F\ G\^2\) + \(a\^4\ B\^2\ r\^3\)\/\
\(F\ G\^2\) - \(2\ a\^2\ r\^5\)\/\(F\ G\^2\) + \(a\^2\ B\^2\ r\^5\)\/\(F\ \
G\^2\) - r\^7\/\(F\ G\^2\)\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], m], 
          Times[ -1, 
            Power[ a, 4], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 3]], 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 3]], 
          Times[ -2, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 5]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 5]], 
          Times[ -1, 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 7]]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(isrdmg\)], "Input"],

Cell[BoxData[
    \(1\/\(b\^4\ F\ P\^2\ Q\ r\^4\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[sc]]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riesq]]\)], "Input"],

Cell[BoxData[
    \(\(12\ m\^2\ \((20\ a\^8\ b\^8 + 99\ a\^6\ b\^6\ r\^2 + 195\ a\^4\ b\^4\ \
r\^4 + 189\ a\^2\ b\^2\ r\^6 + 105\ r\^8)\)\)\/\(r\^10\ \((a\^2\ b\^2 + r\^2)\
\)\^6\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riecub1]]\)], "Input"],

Cell[BoxData[
    \(\(\(1\/\(r\^15\ \((a\^2\ b\^2 + r\^2)\)\^9\)\)\((12\ m\^3\ \((130\ \
a\^12\ b\^12 + 996\ a\^10\ b\^10\ r\^2 + 3387\ a\^8\ b\^8\ r\^4 + 
            6676\ a\^6\ b\^6\ r\^6 + 8212\ a\^4\ b\^4\ r\^8 + 
            5936\ a\^2\ b\^2\ r\^10 + 2135\ r\^12)\))\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riecub2]]\)], "Input"],

Cell[BoxData[
    \(\(6\ m\^3\ \((85\ a\^12\ b\^12 + 636\ a\^10\ b\^10\ r\^2 + 1992\ a\^8\ \
b\^8\ r\^4 + 3316\ a\^6\ b\^6\ r\^6 + 3097\ a\^4\ b\^4\ r\^8 + 1736\ a\^2\ \
b\^2\ r\^10 + 770\ r\^12)\)\)\/\(r\^15\ \((a\^2\ b\^2 + r\^2)\)\^9\)\)], \
"Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riecub3]]\)], "Input"],

Cell[BoxData[
    \(\(3\ m\^3\ \((40\ a\^12\ b\^12 + 276\ a\^10\ b\^10\ r\^2 + 597\ a\^8\ b\
\^8\ r\^4 - 44\ a\^6\ b\^6\ r\^6 - 2018\ a\^4\ b\^4\ r\^8 - 2464\ a\^2\ b\^2\ \
r\^10 - 595\ r\^12)\)\)\/\(r\^15\ \((a\^2\ b\^2 + r\^2)\)\^9\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Cdriesq1]]\)], "Input"],

Cell[BoxData[
    \(\(\(1\/\(r\^15\ \((a\^2\ b\^2 + r\^2)\)\^9\)\)\((12\ m\^2\ \((\(-700\)\ \
a\^12\ b\^12\ m - 5060\ a\^10\ b\^10\ m\ r\^2 + 700\ a\^14\ b\^12\ r\^3 - 
            15631\ a\^8\ b\^8\ m\ r\^4 + 5220\ a\^12\ b\^10\ r\^5 + 
            540\ a\^12\ b\^12\ r\^5 - 26724\ a\^6\ b\^6\ m\ r\^6 + 
            16560\ a\^10\ b\^8\ r\^7 + 4131\ a\^10\ b\^10\ r\^7 - 
            27222\ a\^4\ b\^4\ m\ r\^8 + 28936\ a\^8\ b\^6\ r\^9 + 
            13419\ a\^8\ b\^8\ r\^9 - 16464\ a\^2\ b\^2\ m\ r\^10 + 
            29916\ a\^6\ b\^4\ r\^11 + 24030\ a\^6\ b\^6\ r\^11 - 
            6615\ m\ r\^12 + 18900\ a\^4\ b\^2\ r\^13 + 
            24786\ a\^4\ b\^4\ r\^13 + 5880\ a\^2\ r\^15 + 
            17199\ a\^2\ b\^2\ r\^15 + 6615\ r\^17)\))\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[RstarRRR]]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[RstarR]]\)], "Input"],

Cell[BoxData[
    \(\(24883200\ m\^2\ \((20\ a\^8\ b\^8 + 99\ a\^6\ b\^6\ r\^2 + 195\ a\^4\ \
b\^4\ r\^4 + 189\ a\^2\ b\^2\ r\^6 + 105\ r\^8)\)\)\/\(r\^10\ \((a\^2\ b\^2 + \
r\^2)\)\^6\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[gRstarR]]\)], "Input"],

Cell[BoxData[
    \(\(691200\ m\^2\ \((20\ a\^8\ b\^8 + 99\ a\^6\ b\^6\ r\^2 + 195\ a\^4\ b\
\^4\ r\^4 + 189\ a\^2\ b\^2\ r\^6 + 105\ r\^8)\)\)\/\(r\^10\ \((a\^2\ b\^2 + \
r\^2)\)\^6\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Euler]]\)], "Input"],

Cell[BoxData[
    \(\(9216\ m\^4\ \((117\ a\^6\ b\^6 + 269\ a\^4\ b\^4\ r\^2 - 133\ a\^2\ b\
\^2\ r\^4 - 525\ r\^6)\)\)\/\(r\^18\ \((a\^2\ b\^2 + r\^2)\)\^8\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(backup[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]20\[InvisibleSpace]":"\
\[InvisibleSpace]6\[InvisibleSpace]":"\[InvisibleSpace]58\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]8\[InvisibleSpace]\
"/"\[InvisibleSpace]2003\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 20, ":", 6, ":", 58, " on ", 4, "/", 8, "/", 2003, "."],
      Editable->False]], "Print"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "ken", \
"projects", "ptah", "mathematica"}, "dbraneval.nb.ps", CharacterEncoding -> \
"ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 70, 1, 27, "Input"],
Cell[1874, 57, 2984, 44, 375, "Print"],
Cell[4861, 103, 81, 1, 23, "Print"]
}, Closed]],
Cell[4957, 107, 423, 8, 104, "Input"],

Cell[CellGroupData[{
Cell[5405, 119, 68, 1, 27, "Input"],
Cell[5476, 122, 121, 3, 23, "Print"],
Cell[5600, 127, 121, 3, 23, "Print"],
Cell[5724, 132, 121, 3, 23, "Print"],
Cell[5848, 137, 121, 3, 23, "Print"],
Cell[5972, 142, 121, 3, 23, "Print"],
Cell[6096, 147, 121, 3, 23, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[6254, 155, 51, 1, 24, "Input"],
Cell[6308, 158, 310, 9, 25, "Print"],
Cell[6621, 169, 345, 10, 25, "Print"],
Cell[6969, 181, 256, 7, 25, "Print"],
Cell[7228, 190, 256, 7, 25, "Print"],
Cell[7487, 199, 256, 7, 25, "Print"],
Cell[7746, 208, 256, 7, 25, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8039, 220, 56, 1, 27, "Input"],
Cell[8098, 223, 252, 5, 23, "Print"],
Cell[8353, 230, 283, 6, 25, "Print"],
Cell[8639, 238, 531, 15, 41, "Print"],
Cell[9173, 255, 286, 6, 40, "Print"],
Cell[9462, 263, 286, 6, 37, "Print"],
Cell[9751, 271, 286, 6, 38, "Print"],
Cell[10040, 279, 286, 6, 37, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[10363, 290, 51, 1, 24, "Input"],
Cell[10417, 293, 455, 13, 41, "Print"],
Cell[10875, 308, 525, 15, 42, "Print"],
Cell[11403, 325, 591, 17, 41, "Print"],
Cell[11997, 344, 581, 17, 25, "Print"],
Cell[12581, 363, 316, 8, 41, "Print"],
Cell[12900, 373, 274, 6, 39, "Print"],
Cell[13177, 381, 1092, 34, 41, "Print"],
Cell[14272, 417, 350, 9, 41, "Print"],
Cell[14625, 428, 350, 9, 41, "Print"],
Cell[14978, 439, 641, 19, 41, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[15656, 463, 54, 1, 24, "Input"],
Cell[15713, 466, 337, 9, 41, "Print"],
Cell[16053, 477, 369, 10, 41, "Print"],
Cell[16425, 489, 401, 11, 42, "Print"],
Cell[16829, 502, 405, 11, 41, "Print"],
Cell[17237, 515, 339, 9, 41, "Print"],
Cell[17579, 526, 296, 7, 41, "Print"],
Cell[17878, 535, 896, 27, 41, "Print"],
Cell[18777, 564, 319, 8, 37, "Print"],
Cell[19099, 574, 319, 8, 37, "Print"],
Cell[19421, 584, 1261, 39, 41, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[20719, 628, 39, 1, 24, "Input"],
Cell[20761, 631, 63, 1, 45, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[20861, 637, 49, 1, 27, "Input"],
Cell[20913, 640, 35, 1, 27, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[20985, 646, 52, 1, 27, "Input"],
Cell[21040, 649, 194, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[21271, 657, 54, 1, 27, "Input"],
Cell[21328, 660, 290, 4, 67, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[21655, 669, 54, 1, 27, "Input"],
Cell[21712, 672, 254, 4, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[22003, 681, 54, 1, 27, "Input"],
Cell[22060, 684, 249, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[22346, 692, 55, 1, 27, "Input"],
Cell[22404, 695, 766, 11, 121, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[23207, 711, 55, 1, 27, "Input"],
Cell[23265, 714, 35, 1, 27, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[23337, 720, 53, 1, 27, "Input"],
Cell[23393, 723, 200, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[23630, 731, 54, 1, 27, "Input"],
Cell[23687, 734, 198, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[23922, 742, 52, 1, 27, "Input"],
Cell[23977, 745, 172, 2, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[24186, 752, 46, 1, 27, "Input"],
Cell[24235, 755, 423, 7, 23, "Print"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

