(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     37475,       1111]*)
(*NotebookOutlinePosition[     38465,       1143]*)
(*  CellTagsIndexPosition[     38421,       1139]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/ccif.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,<oldx,>func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("AVOID USING C, D, E, I, N and O as variables !"\)], "Print"]
}, Open  ]],

Cell[BoxData[
    \(\(dim\  = \ 8;\)\)], "Input"],

Cell[BoxData[{
    \(\(ad2\  = \ 
        Table[Table[0, {j, dim}], {i, dim}];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        Do[\((\[IndentingNewLine]inddim\  = \ 
              Join[DeleteCases[Join[Table[l, {l, dim - 1}], {0}], i | j], {i, 
                  j}]; \[IndentingNewLine]ad2[\([i, j]\)]\  = \ 
              Signature[inddim] 
                Factorial[dim - 2];\[IndentingNewLine])\), {j, dim}], {i, 
          dim}];\)\)}], "Input"],

Cell[BoxData[{
    \(\(ad3\  = \ 
        Table[Table[Table[0, {k, dim}], {j, dim}], {i, 
            dim}];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        Do[Do[\((\[IndentingNewLine]inddim\  = \ 
                Join[DeleteCases[Join[Table[l, {l, dim - 1}], {0}], 
                    i | j | k], {i, j, k}]; \[IndentingNewLine]ad3[\([i, j, 
                    k]\)]\  = \ 
                Signature[inddim] 
                  Factorial[dim - 3];\[IndentingNewLine])\), {k, dim}], {j, 
            dim}], {i, dim}];\)\)}], "Input"],

Cell[BoxData[
    \(\(ptahdir\  = \ "\<~/projects/ptah/\>";\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(xp\  = \ 
        Flatten[{Table[Symbol["\<t\>" <> ToString[i]], {i, dim - 5}], 
            If[dim \[GreaterEqual] 5, p2, {}], r, th, ph, 
            t}];\)\), "\[IndentingNewLine]", 
    \(\(rho\  = \ Sqrt[r^2 + a^2  Cos[th]^2];\)\), "\[IndentingNewLine]", 
    \(\(delta\  = \ 
        m/\((r^\((dim - 5)\) rho^2)\);\)\), "\[IndentingNewLine]", 
    \(\(psi\  = \ 
        r^\((dim - 5)\) 
          rho^2/\((r^\((dim - 5)\) \((r^2 + a^2)\) - 
                m)\);\)\), "\[IndentingNewLine]", 
    \(\(gp\  = \ 0*IdentityMatrix[dim];\)\), "\[IndentingNewLine]", 
    \(\(Do[
        gp[\([i, i]\)]\  = \ 
          r^2  Cos[th]^2\ Product[Sin[xp[\([j - 1]\)]]^2, {j, 2, i}], {i, 1, 
          dim - 4}];\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim - 3, dim - 3]\)]\  = \ psi;\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim - 2, dim - 2]\)]\  = \ rho^2;\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim - 1, dim - 1]\)]\  = \ \((r^2 + a^2)\) Sin[th]^2 + 
          delta*a^2  Sin[th]^4;\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim - 1, dim]\)]\  = \ 
        delta*a*Sin[th]^2;\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim, dim - 1]\)]\  = \ 
        gp[\([dim - 1, dim]\)];\)\), "\[IndentingNewLine]", 
    \(\(gp[\([dim, dim]\)]\  = \ \(-1\) + delta;\)\), "\[IndentingNewLine]", 
    \(\(gp\  = \ simpler[gp];\)\), "\[IndentingNewLine]", 
    \(\(Print["\<dim = \>", dim, "\<, x = \>", 
        xp, "\<, metric components are:\>"];\)\), "\[IndentingNewLine]", 
    \(\(components[gp];\)\), "\[IndentingNewLine]", 
    \(\(If[dim \[Equal] 4, 
        x = {r, b, y, t}, \[IndentingNewLine]x\  = \ 
          Flatten[{Take[{p, q, s, u, v, w}, dim - 5], 
              If[dim \[GreaterEqual] 5, z, {}], r, b, y, 
              t}]];\)\), "\[IndentingNewLine]", 
    \(\(g\  = \ 
        simpler[coordxform[gp, xp, 
            If[dim \[Equal] 4, {r, ArcCos[b], y, 
                t}, \[IndentingNewLine]Flatten[{Thread[
                    ArcCos[Take[{p, q, s, u, v, w}, dim - 5]]], 
                  If[dim \[GreaterEqual] 5, z, {}], r, ArcCos[b], y, t}]], 
            x]];\)\), "\[IndentingNewLine]", 
    \(\(Print["\<dim = \>", dim, "\<, x = \>", 
        x, "\<, metric components are:\>"];\)\), "\[IndentingNewLine]", 
    \(\(components[g];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("dim = "\[InvisibleSpace]8\[InvisibleSpace]", x = "\
\[InvisibleSpace]{t1, t2, t3, p2, r, th, ph, 
          t}\[InvisibleSpace]", metric components are:"\),
      SequenceForm[ 
      "dim = ", 8, ", x = ", {t1, t2, t3, p2, r, th, ph, t}, 
        ", metric components are:"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\ \
Cos[th]\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Cos[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\ Cos[th]\^2\ \
Sin[t1]\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Cos[ th], 2], 
          Power[ 
            Sin[ t1], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\ Cos[th]\^2\ \
Sin[t1]\^2\ Sin[t2]\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Cos[ th], 2], 
          Power[ 
            Sin[ t1], 2], 
          Power[ 
            Sin[ t2], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\ Cos[th]\^2\ \
Sin[t1]\^2\ Sin[t2]\^2\ Sin[t3]\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Cos[ th], 2], 
          Power[ 
            Sin[ t1], 2], 
          Power[ 
            Sin[ t2], 2], 
          Power[ 
            Sin[ t3], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^3\ \((r\^2 + a\
\^2\ Cos[th]\^2)\)\)\/\(\(-m\) + a\^2\ r\^3 + r\^5\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ r, 3], 
          Power[ 
            Plus[ 
              Times[ -1, m], 
              Times[ 
                Power[ a, 2], 
                Power[ r, 3]], 
              Power[ r, 5]], -1], 
          Plus[ 
            Power[ r, 2], 
            Times[ 
              Power[ a, 2], 
              Power[ 
                Cos[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2 + 
          a\^2\ Cos[th]\^2\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Plus[ 
          Power[ r, 2], 
          Times[ 
            Power[ a, 2], 
            Power[ 
              Cos[ th], 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(Sin[th]\^2\ \
\((a\^2\ r\^5 + r\^7 + a\^4\ r\^3\ Cos[th]\^2 + a\^2\ r\^5\ Cos[th]\^2 + a\^2\
\ m\ Sin[th]\^2)\)\)\/\(r\^3\ \((r\^2 + a\^2\ Cos[th]\^2)\)\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Power[ r, -3], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Power[ r, 5]], 
            Power[ r, 7], 
            Times[ 
              Power[ a, 4], 
              Power[ r, 3], 
              Power[ 
                Cos[ th], 2]], 
            Times[ 
              Power[ a, 2], 
              Power[ r, 5], 
              Power[ 
                Cos[ th], 2]], 
            Times[ 
              Power[ a, 2], m, 
              Power[ 
                Sin[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\ \
Sin[th]\^2\)\/\(r\^3\ \((r\^2 + a\^2\ Cos[th]\^2)\)\)\),
      SequenceForm[ "[", 7, ",", 8, "] = ", 
        Times[ a, m, 
          Power[ r, -3], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\ \
Sin[th]\^2\)\/\(r\^3\ \((r\^2 + a\^2\ Cos[th]\^2)\)\)\),
      SequenceForm[ "[", 8, ",", 7, "] = ", 
        Times[ a, m, 
          Power[ r, -3], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(m - r\^5 - a\^2\ \
r\^3\ Cos[th]\^2\)\/\(r\^3\ \((r\^2 + a\^2\ Cos[th]\^2)\)\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ 
          Power[ r, -3], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ a, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Plus[ m, 
            Times[ -1, 
              Power[ r, 5]], 
            Times[ -1, 
              Power[ a, 2], 
              Power[ r, 3], 
              Power[ 
                Cos[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("dim = "\[InvisibleSpace]8\[InvisibleSpace]", x = "\
\[InvisibleSpace]{p, q, s, z, r, b, y, 
          t}\[InvisibleSpace]", metric components are:"\),
      SequenceForm[ 
      "dim = ", 8, ", x = ", {p, q, s, z, r, b, y, t}, 
        ", metric components are:"],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(b\^2\ r\^2\)\
\/\(\((\(-1\) + p)\)\ \((1 + p)\)\)\)\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ -1, 
          Power[ b, 2], 
          Power[ 
            Plus[ -1, p], -1], 
          Power[ 
            Plus[ 1, p], -1], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(b\^2\ \((\(-1\) + \
p)\)\ \((1 + p)\)\ r\^2\)\/\(\((\(-1\) + q)\)\ \((1 + q)\)\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, 2], 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Power[ 
            Plus[ -1, q], -1], 
          Power[ 
            Plus[ 1, q], -1], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(b\^2\ \
\((\(-1\) + p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
                    q)\)\ r\^2\)\/\(\((\(-1\) + s)\)\ \((1 + s)\)\)\)\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ -1, 
          Power[ b, 2], 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Plus[ -1, q], 
          Plus[ 1, q], 
          Power[ r, 2], 
          Power[ 
            Plus[ -1, s], -1], 
          Power[ 
            Plus[ 1, s], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-b\^2\)\ \
\((\(-1\) + p)\)\ \((1 + p)\)\ \((\(-1\) + q)\)\ \((1 + 
              q)\)\ r\^2\ \((\(-1\) + s)\)\ \((1 + s)\)\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ -1, 
          Power[ b, 2], 
          Plus[ -1, p], 
          Plus[ 1, p], 
          Plus[ -1, q], 
          Plus[ 1, q], 
          Power[ r, 2], 
          Plus[ -1, s], 
          Plus[ 1, s]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^3\ \((a\^2\ \
b\^2 + r\^2)\)\)\/\(\(-m\) + a\^2\ r\^3 + r\^5\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ r, 3], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Power[ b, 2]], 
            Power[ r, 2]], 
          Power[ 
            Plus[ 
              Times[ -1, m], 
              Times[ 
                Power[ a, 2], 
                Power[ r, 3]], 
              Power[ r, 5]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(a\^2\ b\^2 + 
                r\^2\)\/\(\((\(-1\) + b)\)\ \((1 + b)\)\)\)\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ -1, 
          Power[ 
            Plus[ -1, b], -1], 
          Power[ 
            Plus[ 1, b], -1], 
          Plus[ 
            Times[ 
              Power[ a, 2], 
              Power[ b, 2]], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(\((\(-1\) + 
                    b)\)\ \((1 + b)\)\ \((a\^2\ m - a\^2\ b\^2\ m + 
                    a\^4\ b\^2\ r\^3 + a\^2\ r\^5 + a\^2\ b\^2\ r\^5 + 
                    r\^7)\)\)\/\(r\^3\ \((a\^2\ b\^2 + r\^2)\)\)\)\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ -1, 
          Plus[ -1, b], 
          Plus[ 1, b], 
          Power[ r, -3], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ a, 2], 
                Power[ b, 2]], 
              Power[ r, 2]], -1], 
          Plus[ 
            Times[ 
              Power[ a, 2], m], 
            Times[ -1, 
              Power[ a, 2], 
              Power[ b, 2], m], 
            Times[ 
              Power[ a, 4], 
              Power[ b, 2], 
              Power[ r, 3]], 
            Times[ 
              Power[ a, 2], 
              Power[ r, 5]], 
            Times[ 
              Power[ a, 2], 
              Power[ b, 2], 
              Power[ r, 5]], 
            Power[ r, 7]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(a\ \((\(-1\) \
+ b)\)\ \((1 + b)\)\ m\)\/\(r\^3\ \((a\^2\ b\^2 + r\^2)\)\)\)\)\),
      SequenceForm[ "[", 7, ",", 8, "] = ", 
        Times[ -1, a, 
          Plus[ -1, b], 
          Plus[ 1, b], m, 
          Power[ r, -3], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ a, 2], 
                Power[ b, 2]], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(a\ \((\(-1\) \
+ b)\)\ \((1 + b)\)\ m\)\/\(r\^3\ \((a\^2\ b\^2 + r\^2)\)\)\)\)\),
      SequenceForm[ "[", 8, ",", 7, "] = ", 
        Times[ -1, a, 
          Plus[ -1, b], 
          Plus[ 1, b], m, 
          Power[ r, -3], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ a, 2], 
                Power[ b, 2]], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(\(-m\) + 
                a\^2\ b\^2\ r\^3 + 
                r\^5\)\/\(r\^3\ \((a\^2\ b\^2 + r\^2)\)\)\)\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Times[ -1, 
          Power[ r, -3], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ a, 2], 
                Power[ b, 2]], 
              Power[ r, 2]], -1], 
          Plus[ 
            Times[ -1, m], 
            Times[ 
              Power[ a, 2], 
              Power[ b, 2], 
              Power[ r, 3]], 
            Power[ r, 5]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell["\<\
We next parameterize the metric so that each component is a \
polynomial:\
\>", "Text"],

Cell[BoxData[
    \(\(rules\  = \ {a\^2\ b\^2 + r\^2 \[Rule] 
            F, \(a\^2\) r^3 - m\  + r\^5 \[Rule] 
            G^2, \[IndentingNewLine]\((\(-1\) + 
                b)\) \[Rule] \(-B^2\)/\((1 + b)\), 
          b^2 \[Rule] 
            1 - B^2, \[IndentingNewLine]\((\(-1\) + 
                p)\) \[Rule] \(-P^2\)/\((1 + p)\), \((\(-1\) + 
                q)\) \[Rule] \(-Q^2\)/\((1 + q)\), \((\(-1\) + 
                s)\) \[Rule] \(-S^2\)/\((1 + s)\), \((\(-1\) + 
                u)\) \[Rule] \(-U^2\)/\((1 + u)\), \((\(-1\) + 
                v)\) \[Rule] \(-V^2\)/\((1 + v)\), \((\(-1\) + 
                w)\) \[Rule] \(-W^2\)/\((1 + w)\)};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(gp\  = \ Expand[g /. rules];\)\), "\[IndentingNewLine]", 
    \(\(components[gp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\/P\^2 - \
\(B\^2\ r\^2\)\/P\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, -2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, -2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ r\^2\)\/Q\
\^2 - \(B\^2\ P\^2\ r\^2\)\/Q\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, -2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, -2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ Q\^2\ \
r\^2\)\/S\^2 - \(B\^2\ P\^2\ Q\^2\ r\^2\)\/S\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, -2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, -2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(P\^2\ Q\^2\ r\^2\ \
S\^2 - B\^2\ P\^2\ Q\^2\ r\^2\ S\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(F\ r\^3\)\/G\^2\),
      
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ F, 
          Power[ G, -2], 
          Power[ r, 3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]F\/B\^2\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ B, -2], F]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(a\^4\ B\^2\)\/F \
- \(a\^4\ B\^4\)\/F + \(a\^2\ B\^4\ m\)\/\(F\ r\^3\) + \(2\ a\^2\ B\^2\ \
r\^2\)\/F - \(a\^2\ B\^4\ r\^2\)\/F + \(B\^2\ r\^4\)\/F\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Plus[ 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1]], 
          Times[ -1, 
            Power[ a, 4], 
            Power[ B, 4], 
            Power[ F, -1]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], m, 
            Power[ r, -3]], 
          Times[ 2, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ m\)\/\(F\
\ r\^3\)\),
      SequenceForm[ "[", 7, ",", 8, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, 
          Power[ r, -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ m\)\/\(F\
\ r\^3\)\),
      SequenceForm[ "[", 8, ",", 7, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, 
          Power[ r, -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(a\^2\/F\)\) \
+ \(a\^2\ B\^2\)\/F + m\/\(F\ r\^3\) - r\^2\/F\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 2], 
            Power[ F, -1]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1]], 
          Times[ 
            Power[ F, -1], m, 
            Power[ r, -3]], 
          Times[ -1, 
            Power[ F, -1], 
            Power[ r, 2]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(ginvp\  = \ 
        Expand[Simplify[
            Factor[Inverse[g]] /. rules]];\)\), "\[IndentingNewLine]", 
    \(\(components[ginvp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]P\^2\/\(b\^2\ \
r\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, 2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]Q\^2\/\(b\^2\ P\^2\ \
r\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, 2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]S\^2\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]1\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\ S\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]G\^2\/\(F\ r\^3\)\),
      
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ F, -1], 
          Power[ G, 2], 
          Power[ r, -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]B\^2\/F\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ B, 2], 
          Power[ F, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(m\/\(B\^2\ F\
\ G\^2\)\)\) - \(a\^2\ r\^3\)\/\(F\ G\^2\) + \(a\^2\ r\^3\)\/\(B\^2\ F\ \
G\^2\) + r\^5\/\(B\^2\ F\ G\^2\)\)\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], m], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 3]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 3]], 
          Times[ 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 5]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\)\/\(F\ G\^2\
\)\),
      SequenceForm[ "[", 7, ",", 8, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\)\/\(F\ G\^2\
\)\),
      SequenceForm[ "[", 8, ",", 7, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(\(a\^2\ B\^2\
\ m\)\/\(F\ G\^2\)\)\) - \(a\^4\ r\^3\)\/\(F\ G\^2\) + \(a\^4\ B\^2\ r\^3\)\/\
\(F\ G\^2\) - \(2\ a\^2\ r\^5\)\/\(F\ G\^2\) + \(a\^2\ B\^2\ r\^5\)\/\(F\ \
G\^2\) - r\^7\/\(F\ G\^2\)\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], m], 
          Times[ -1, 
            Power[ a, 4], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 3]], 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 3]], 
          Times[ -2, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 5]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 5]], 
          Times[ -1, 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 7]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(irmdgp\  = \ 
      simpler[1/PowerExpand[Sqrt[simpler[\(-Det[g]\)]] /. rules]]\)], "Input"],

Cell[BoxData[
    \(1\/\(b\^4\ F\ P\^2\ Q\ r\^4\)\)], "Output"]
}, Open  ]],

Cell["\<\
Don't forget the sub[i] in the derivatives (takes care of change in \
time coord from i=10 to i=0 in c++)\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(Btab\  = \ {B^2, 1 - b^2, 
        Table[D[1 - b^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              B)\)}\), "\[IndentingNewLine]", 
    \(Ftab\  = \ {F, a\^2\ b\^2 + r\^2, 
        Table[D[a\^2\ b\^2 + r\^2, x[\([sub[i]]\)]], {i, 
            dim}]}\), "\[IndentingNewLine]", 
    \(Gtab\  = \ {G^2, \(a\^2\) r^3 - m + r\^5, 
        Table[D[\(a\^2\) r^3 - m + r\^5, x[\([sub[i]]\)]], {i, dim}]/\((2  
              G)\)}\), "\[IndentingNewLine]", 
    \(Ptab\  = \ {P^2, 1 - p^2, 
        Table[D[1 - p^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              P)\)}\), "\[IndentingNewLine]", 
    \(Qtab\  = \ {Q^2, 1 - q^2, 
        Table[D[1 - q^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              Q)\)}\), "\[IndentingNewLine]", 
    \(Stab\  = \ {S^2, 1 - s^2, 
        Table[D[1 - s^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              S)\)}\), "\[IndentingNewLine]", 
    \(Utab\  = \ {U^2, 1 - u^2, 
        Table[D[1 - u^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              U)\)}\), "\[IndentingNewLine]", 
    \(Vtab\  = \ {V^2, 1 - v^2, 
        Table[D[1 - v^2, x[\([sub[i]]\)]], {i, dim}]/\((2  
              V)\)}\), "\[IndentingNewLine]", 
    \(Wtab\  = \ {W^2, 1 - w^2, 
        Table[D[1 - w^2, x[\([sub[i]]\)]], {i, dim}]/\((2  W)\)}\)}], "Input"],

Cell[BoxData[
    \({B\^2, 1 - b\^2, {0, 0, 0, 0, 0, 0, \(-\(b\/B\)\), 0}}\)], "Output"],

Cell[BoxData[
    \({F, 
      a\^2\ b\^2 + r\^2, {0, 0, 0, 0, 0, 2\ r, 2\ a\^2\ b, 0}}\)], "Output"],

Cell[BoxData[
    \({G\^2, \(-m\) + a\^2\ r\^3 + r\^5, {0, 0, 0, 0, 
        0, \(3\ a\^2\ r\^2 + 5\ r\^4\)\/\(2\ G\), 0, 0}}\)], "Output"],

Cell[BoxData[
    \({P\^2, 1 - p\^2, {0, \(-\(p\/P\)\), 0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({Q\^2, 1 - q\^2, {0, 0, \(-\(q\/Q\)\), 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({S\^2, 1 - s\^2, {0, 0, 0, \(-\(s\/S\)\), 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({U\^2, 1 - u\^2, {0, 0, 0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({V\^2, 1 - v\^2, {0, 0, 0, 0, 0, 0, 0, 0}}\)], "Output"],

Cell[BoxData[
    \({W\^2, 1 - w\^2, {0, 0, 0, 0, 0, 0, 0, 0}}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(coords\  = \ x\), "\[IndentingNewLine]", 
    \(ints\  = \ {2, 3}\), "\n", 
    \(constants\  = \ {I, m, a}\), "\n", 
    \(functions\  = \ 
      Take[{Ftab, Gtab, Btab, Ptab, Qtab, Stab, Utab, Vtab, Wtab}, 
        If[dim \[Equal] 4, 3, dim - 2]]\), "\[IndentingNewLine]", 
    \(ansatze\  = \ {}\)}], "Input"],

Cell[BoxData[
    \({p, q, s, z, r, b, y, t}\)], "Output"],

Cell[BoxData[
    \({2, 3}\)], "Output"],

Cell[BoxData[
    \({\[ImaginaryI], m, a}\)], "Output"],

Cell[BoxData[
    \({{F, 
        a\^2\ b\^2 + r\^2, {0, 0, 0, 0, 0, 2\ r, 2\ a\^2\ b, 
          0}}, {G\^2, \(-m\) + a\^2\ r\^3 + r\^5, {0, 0, 0, 0, 
          0, \(3\ a\^2\ r\^2 + 5\ r\^4\)\/\(2\ G\), 0, 0}}, {B\^2, 
        1 - b\^2, {0, 0, 0, 0, 0, 0, \(-\(b\/B\)\), 0}}, {P\^2, 
        1 - p\^2, {0, \(-\(p\/P\)\), 0, 0, 0, 0, 0, 0}}, {Q\^2, 
        1 - q\^2, {0, 0, \(-\(q\/Q\)\), 0, 0, 0, 0, 0}}, {S\^2, 
        1 - s\^2, {0, 0, 0, \(-\(s\/S\)\), 0, 0, 0, 0}}}\)], "Output"],

Cell[BoxData[
    \({}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(headers[];\)\)], "Input"],

Cell[BoxData[
    \("functions completed..."\)], "Print"],

Cell[BoxData[
    \("g completed..."\)], "Print"],

Cell[BoxData[
    \("gi completed..."\)], "Print"],

Cell[BoxData[
    \("d completed..."\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(tensor[ad2, "\<ad2\>"];\)\), "\[IndentingNewLine]", 
    \(\(tensor[ad3, "\<ad3\>"];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("ad2"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ad2", " completed..."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("ad3"\[InvisibleSpace]" completed..."\),
      SequenceForm[ "ad3", " completed..."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(backup[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]22\[InvisibleSpace]":"\
\[InvisibleSpace]47\[InvisibleSpace]":"\[InvisibleSpace]18\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]2\[InvisibleSpace]\
"/"\[InvisibleSpace]2003\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 22, ":", 47, ":", 18, " on ", 4, "/", 2, "/", 2003, "."],
      Editable->False]], "Print"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "ken", \
"projects", "mathematica", "packages"}, "d11brane.nb.ps", CharacterEncoding \
-> "ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 70, 1, 27, "Input"],
Cell[1874, 57, 2984, 44, 375, "Print"],
Cell[4861, 103, 81, 1, 23, "Print"]
}, Open  ]],
Cell[4957, 107, 49, 1, 27, "Input"],
Cell[5009, 110, 451, 9, 91, "Input"],
Cell[5463, 121, 536, 11, 91, "Input"],
Cell[6002, 134, 74, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[6101, 139, 2298, 44, 379, "Input"],
Cell[8402, 185, 349, 7, 23, "Print"],
Cell[8754, 194, 330, 9, 25, "Print"],
Cell[9087, 205, 387, 11, 25, "Print"],
Cell[9477, 218, 444, 13, 25, "Print"],
Cell[9924, 233, 501, 15, 25, "Print"],
Cell[10428, 250, 671, 20, 41, "Print"],
Cell[11102, 272, 395, 11, 25, "Print"],
Cell[11500, 285, 1142, 36, 43, "Print"],
Cell[12645, 323, 561, 16, 43, "Print"],
Cell[13209, 341, 561, 16, 43, "Print"],
Cell[13773, 359, 738, 22, 43, "Print"],
Cell[14514, 383, 337, 7, 23, "Print"],
Cell[14854, 392, 451, 12, 44, "Print"],
Cell[15308, 406, 518, 14, 44, "Print"],
Cell[15829, 422, 632, 17, 43, "Print"],
Cell[16464, 441, 568, 15, 25, "Print"],
Cell[17035, 458, 641, 19, 41, "Print"],
Cell[17679, 479, 547, 15, 43, "Print"],
Cell[18229, 496, 1182, 35, 43, "Print"],
Cell[19414, 533, 565, 15, 42, "Print"],
Cell[19982, 550, 565, 15, 42, "Print"],
Cell[20550, 567, 720, 21, 43, "Print"]
}, Open  ]],
Cell[21285, 591, 97, 3, 32, "Text"],
Cell[21385, 596, 675, 12, 83, "Input"],

Cell[CellGroupData[{
Cell[22085, 612, 120, 2, 43, "Input"],
Cell[22208, 616, 455, 13, 41, "Print"],
Cell[22666, 631, 525, 15, 42, "Print"],
Cell[23194, 648, 591, 17, 41, "Print"],
Cell[23788, 667, 581, 17, 25, "Print"],
Cell[24372, 686, 316, 8, 41, "Print"],
Cell[24691, 696, 274, 6, 39, "Print"],
Cell[24968, 704, 1092, 34, 41, "Print"],
Cell[26063, 740, 350, 9, 41, "Print"],
Cell[26416, 751, 350, 9, 41, "Print"],
Cell[26769, 762, 641, 19, 41, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[27447, 786, 175, 4, 43, "Input"],
Cell[27625, 792, 337, 9, 41, "Print"],
Cell[27965, 803, 369, 10, 41, "Print"],
Cell[28337, 815, 401, 11, 42, "Print"],
Cell[28741, 828, 405, 11, 41, "Print"],
Cell[29149, 841, 339, 9, 41, "Print"],
Cell[29491, 852, 296, 7, 41, "Print"],
Cell[29790, 861, 896, 27, 41, "Print"],
Cell[30689, 890, 319, 8, 37, "Print"],
Cell[31011, 900, 319, 8, 37, "Print"],
Cell[31333, 910, 1261, 39, 41, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[32631, 954, 112, 2, 27, "Input"],
Cell[32746, 958, 63, 1, 45, "Output"]
}, Open  ]],
Cell[32824, 962, 129, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[32978, 969, 1283, 26, 164, "Input"],
Cell[34264, 997, 88, 1, 44, "Output"],
Cell[34355, 1000, 101, 2, 29, "Output"],
Cell[34459, 1004, 139, 2, 45, "Output"],
Cell[34601, 1008, 88, 1, 41, "Output"],
Cell[34692, 1011, 88, 1, 42, "Output"],
Cell[34783, 1014, 88, 1, 41, "Output"],
Cell[34874, 1017, 76, 1, 29, "Output"],
Cell[34953, 1020, 76, 1, 29, "Output"],
Cell[35032, 1023, 76, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[35145, 1029, 333, 7, 107, "Input"],
Cell[35481, 1038, 58, 1, 27, "Output"],
Cell[35542, 1041, 40, 1, 27, "Output"],
Cell[35585, 1044, 55, 1, 27, "Output"],
Cell[35643, 1047, 485, 8, 141, "Output"],
Cell[36131, 1057, 36, 1, 27, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[36204, 1063, 47, 1, 27, "Input"],
Cell[36254, 1066, 57, 1, 23, "Print"],
Cell[36314, 1069, 49, 1, 23, "Print"],
Cell[36366, 1072, 50, 1, 23, "Print"],
Cell[36419, 1075, 49, 1, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[36505, 1081, 123, 2, 43, "Input"],
Cell[36631, 1085, 157, 3, 23, "Print"],
Cell[36791, 1090, 157, 3, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[36985, 1098, 46, 1, 27, "Input"],
Cell[37034, 1101, 425, 7, 23, "Print"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

