(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     26107,        805]*)
(*NotebookOutlinePosition[     27094,        837]*)
(*  CellTagsIndexPosition[     27050,        833]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/ccif.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,<oldx,>func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,<oldx,>func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"],

Cell[BoxData[
    \("AVOID USING C, D, E, I, N and O as variables !"\)], "Print"]
}, Closed]],

Cell[BoxData[{
    \(\(f\  = \ Table[0, {i, 50}];\)\), "\[IndentingNewLine]", 
    \(\(derivs\  = \ 
        Table[Table[0, {i, 1000}], {j, 50}];\)\), "\[IndentingNewLine]", 
    \(\(g\  = Table[Table[0, {i, 11}], {j, 11}]\ ;\)\), "\n", 
    \(\(ginv\  = \ 
        Table[Table[0, {i, 11}], {j, 11}];\)\), "\[IndentingNewLine]", 
    \(\(Euler\  = \ 0;\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< /home/ken/projects/ptah/ptah.out\)], "Input"],

Cell[BoxData[
    InterpretationBox[\(0\[InvisibleSpace]1\),
      SequenceForm[ 0, 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(1\[InvisibleSpace]2\),
      SequenceForm[ 1, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(2\[InvisibleSpace]2\),
      SequenceForm[ 2, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(3\[InvisibleSpace]2\),
      SequenceForm[ 3, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(4\[InvisibleSpace]2\),
      SequenceForm[ 4, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(5\[InvisibleSpace]2\),
      SequenceForm[ 5, 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\(6\[InvisibleSpace]2\),
      SequenceForm[ 6, 2],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[f];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(a\^2\ b\^2 + r\^2\)\),
      SequenceForm[ "[", 1, "] = ", 
        Plus[ 
          Times[ 
            Power[ a, 2], 
            Power[ b, 2]], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(\(-m\) + a\^2\ r\^4 + r\^6\)\),
      SequenceForm[ "[", 2, "] = ", 
        Plus[ 
          Times[ -1, m], 
          Times[ 
            Power[ a, 2], 
            Power[ r, 4]], 
          Power[ r, 6]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - b\^2\)\),
      SequenceForm[ "[", 3, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ b, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - p\^2\)\),
      SequenceForm[ "[", 4, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ p, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - q\^2\)\),
      SequenceForm[ "[", 5, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ q, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - s\^2\)\),
      SequenceForm[ "[", 6, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ s, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]"] = "\
\[InvisibleSpace]\(1 - u\^2\)\),
      SequenceForm[ "[", 7, "] = ", 
        Plus[ 1, 
          Times[ -1, 
            Power[ u, 2]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[derivs];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ r\)\),
      SequenceForm[ "[", 1, ",", 7, "] = ", 
        Times[ 2, r]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ a\^2\ b\)\),
      SequenceForm[ "[", 1, ",", 8, "] = ", 
        Times[ 2, 
          Power[ a, 2], b]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]16\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(2\ a\^2\ \
r\^3\)\/G + \(3\ r\^5\)\/G\)\),
      SequenceForm[ "[", 2, ",", 16, "] = ", 
        Plus[ 
          Times[ 2, 
            Power[ a, 2], 
            Power[ G, -1], 
            Power[ r, 3]], 
          Times[ 3, 
            Power[ G, -1], 
            Power[ r, 5]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]26\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(b\/B\)\)\),
      SequenceForm[ "[", 3, ",", 26, "] = ", 
        Times[ -1, b, 
          Power[ B, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]29\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(p\/P\)\)\),
      SequenceForm[ "[", 4, ",", 29, "] = ", 
        Times[ -1, p, 
          Power[ P, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]39\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(q\/Q\)\)\),
      SequenceForm[ "[", 5, ",", 39, "] = ", 
        Times[ -1, q, 
          Power[ Q, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]49\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(s\/S\)\)\),
      SequenceForm[ "[", 6, ",", 49, "] = ", 
        Times[ -1, s, 
          Power[ S, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]59\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(u\/U\)\)\),
      SequenceForm[ "[", 7, ",", 59, "] = ", 
        Times[ -1, u, 
          Power[ U, -1]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[g];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\/P\^2 - \
\(B\^2\ r\^2\)\/P\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, -2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, -2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ r\^2\)\/Q\
\^2 - \(B\^2\ P\^2\ r\^2\)\/Q\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, -2], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, -2], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ Q\^2\ \
r\^2\)\/S\^2 - \(B\^2\ P\^2\ Q\^2\ r\^2\)\/S\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, -2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, -2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(P\^2\ Q\^2\ \
r\^2\ S\^2\)\/U\^2 - \(B\^2\ P\^2\ Q\^2\ r\^2\ S\^2\)\/U\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, -2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, -2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(P\^2\ Q\^2\ r\^2\ \
S\^2\ U\^2 - B\^2\ P\^2\ Q\^2\ r\^2\ S\^2\ U\^2\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Plus[ 
          Times[ 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, 2]], 
          Times[ -1, 
            Power[ B, 2], 
            Power[ P, 2], 
            Power[ Q, 2], 
            Power[ r, 2], 
            Power[ S, 2], 
            Power[ U, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(F\ r\^4\)\/G\^2\),
      
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ F, 
          Power[ G, -2], 
          Power[ r, 4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]F\/B\^2\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Power[ B, -2], F]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(a\^4\ B\^2\)\/F \
- \(a\^4\ B\^4\)\/F + \(a\^2\ B\^4\ m\)\/\(F\ r\^4\) + \(2\ a\^2\ B\^2\ \
r\^2\)\/F - \(a\^2\ B\^4\ r\^2\)\/F + \(B\^2\ r\^4\)\/F\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Plus[ 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1]], 
          Times[ -1, 
            Power[ a, 4], 
            Power[ B, 4], 
            Power[ F, -1]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], m, 
            Power[ r, -4]], 
          Times[ 2, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 4], 
            Power[ F, -1], 
            Power[ r, 2]], 
          Times[ 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ r, 4]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ m\)\/\(F\
\ r\^4\)\),
      SequenceForm[ "[", 8, ",", 9, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, 
          Power[ r, -4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ B\^2\ m\)\/\(F\
\ r\^4\)\),
      SequenceForm[ "[", 9, ",", 8, "] = ", 
        Times[ a, 
          Power[ B, 2], 
          Power[ F, -1], m, 
          Power[ r, -4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(a\^2\/F\)\) \
+ \(a\^2\ B\^2\)\/F + m\/\(F\ r\^4\) - r\^2\/F\)\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 2], 
            Power[ F, -1]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1]], 
          Times[ 
            Power[ F, -1], m, 
            Power[ r, -4]], 
          Times[ -1, 
            Power[ F, -1], 
            Power[ r, 2]]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[ginv];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]P\^2\/\(b\^2\ \
r\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, 2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]Q\^2\/\(b\^2\ P\^2\ \
r\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, 2], 
          Power[ r, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]S\^2\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]U\^2\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\ S\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, -2], 
          Power[ U, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]1\/\(b\^2\ P\^2\ \
Q\^2\ r\^2\ S\^2\ U\^2\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ 
          Power[ b, -2], 
          Power[ P, -2], 
          Power[ Q, -2], 
          Power[ r, -2], 
          Power[ S, -2], 
          Power[ U, -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]G\^2\/\(F\ r\^4\)\),
      
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Times[ 
          Power[ F, -1], 
          Power[ G, 2], 
          Power[ r, -4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]7\[InvisibleSpace]","\
\[InvisibleSpace]7\[InvisibleSpace]"] = "\[InvisibleSpace]B\^2\/F\),
      SequenceForm[ "[", 7, ",", 7, "] = ", 
        Times[ 
          Power[ B, 2], 
          Power[ F, -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(m\/\(B\^2\ F\
\ G\^2\)\)\) - \(a\^2\ r\^4\)\/\(F\ G\^2\) + \(a\^2\ r\^4\)\/\(B\^2\ F\ \
G\^2\) + r\^6\/\(B\^2\ F\ G\^2\)\)\),
      SequenceForm[ "[", 8, ",", 8, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], m], 
          Times[ -1, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 4]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 4]], 
          Times[ 
            Power[ B, -2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 6]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]8\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\)\/\(F\ G\^2\
\)\),
      SequenceForm[ "[", 8, ",", 9, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]8\[InvisibleSpace]"] = "\[InvisibleSpace]\(a\ m\)\/\(F\ G\^2\
\)\),
      SequenceForm[ "[", 9, ",", 8, "] = ", 
        Times[ a, 
          Power[ F, -1], 
          Power[ G, -2], m]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]9\[InvisibleSpace]","\
\[InvisibleSpace]9\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\(\(a\^2\ B\^2\
\ m\)\/\(F\ G\^2\)\)\) - \(a\^4\ r\^4\)\/\(F\ G\^2\) + \(a\^4\ B\^2\ r\^4\)\/\
\(F\ G\^2\) - \(2\ a\^2\ r\^6\)\/\(F\ G\^2\) + \(a\^2\ B\^2\ r\^6\)\/\(F\ \
G\^2\) - r\^8\/\(F\ G\^2\)\)\),
      SequenceForm[ "[", 9, ",", 9, "] = ", 
        Plus[ 
          Times[ -1, 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], m], 
          Times[ -1, 
            Power[ a, 4], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 4]], 
          Times[ 
            Power[ a, 4], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 4]], 
          Times[ -2, 
            Power[ a, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 6]], 
          Times[ 
            Power[ a, 2], 
            Power[ B, 2], 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 6]], 
          Times[ -1, 
            Power[ F, -1], 
            Power[ G, -2], 
            Power[ r, 8]]]],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(isrdmg\)], "Input"],

Cell[BoxData[
    \(1\/\(b\^5\ F\ P\^3\ Q\^2\ r\^5\ S\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[sc]]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riesq]]\)], "Input"],

Cell[BoxData[
    \(\(24\ m\^2\ \((25\ a\^8\ b\^8 + 122\ a\^6\ b\^6\ r\^2 + 235\ a\^4\ b\^4\
\ r\^4 + 220\ a\^2\ b\^2\ r\^6 + 98\ r\^8)\)\)\/\(r\^12\ \((a\^2\ b\^2 + \
r\^2)\)\^6\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riecub1]]\)], "Input"],

Cell[BoxData[
    \(\(\(1\/\(r\^18\ \((a\^2\ b\^2 + r\^2)\)\^9\)\)\((48\ m\^3\ \((155\ \
a\^12\ b\^12 + 1180\ a\^10\ b\^10\ r\^2 + 3906\ a\^8\ b\^8\ r\^4 + 
            7264\ a\^6\ b\^6\ r\^6 + 8079\ a\^4\ b\^4\ r\^8 + 
            5116\ a\^2\ b\^2\ r\^10 + 1484\ r\^12)\))\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riecub2]]\)], "Input"],

Cell[BoxData[
    \(\(\(1\/\(r\^18\ \((a\^2\ b\^2 + r\^2)\)\^9\)\)\((12\ m\^3\ \((145\ \
a\^12\ b\^12 + 1040\ a\^10\ b\^10\ r\^2 + 3122\ a\^8\ b\^8\ r\^4 + 
            5028\ a\^6\ b\^6\ r\^6 + 4673\ a\^4\ b\^4\ r\^8 + 
            2612\ a\^2\ b\^2\ r\^10 + 868\ r\^12)\))\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Riecub3]]\)], "Input"],

Cell[BoxData[
    \(\(-\(\(24\ m\^3\ \((5\ a\^12\ b\^12 + 70\ a\^10\ b\^10\ r\^2 + 
                392\ a\^8\ b\^8\ r\^4 + 1118\ a\^6\ b\^6\ r\^6 + 
                1703\ a\^4\ b\^4\ r\^8 + 1252\ a\^2\ b\^2\ r\^10 + 
                308\ r\^12)\)\)\/\(r\^18\ \((a\^2\ b\^2 + \
r\^2)\)\^9\)\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[Cdriesq1]]\)], "Input"],

Cell[BoxData[
    \(\(\(1\/\(r\^18\ \((a\^2\ b\^2 + r\^2)\)\^9\)\)\((960\ m\^2\ \((\(-30\)\ \
a\^12\ b\^12\ m - 214\ a\^10\ b\^10\ m\ r\^2 + 30\ a\^14\ b\^12\ r\^4 - 
            651\ a\^8\ b\^8\ m\ r\^4 + 219\ a\^12\ b\^10\ r\^6 + 
            25\ a\^12\ b\^12\ r\^6 - 1092\ a\^6\ b\^6\ m\ r\^6 + 
            679\ a\^10\ b\^8\ r\^8 + 186\ a\^10\ b\^10\ r\^8 - 
            1085\ a\^4\ b\^4\ m\ r\^8 + 1155\ a\^8\ b\^6\ r\^10 + 
            588\ a\^8\ b\^8\ r\^10 - 632\ a\^2\ b\^2\ m\ r\^10 + 
            1155\ a\^6\ b\^4\ r\^12 + 1022\ a\^6\ b\^6\ r\^12 - 
            196\ m\ r\^12 + 670\ a\^4\ b\^2\ r\^14 + 
            1047\ a\^4\ b\^4\ r\^14 + 180\ a\^2\ r\^16 + 
            648\ a\^2\ b\^2\ r\^16 + 196\ r\^18)\))\)\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[RstarR]]\)], "Input"],

Cell[BoxData[
    \(\(2438553600\ m\^2\ \((25\ a\^8\ b\^8 + 122\ a\^6\ b\^6\ r\^2 + 235\ \
a\^4\ b\^4\ r\^4 + 220\ a\^2\ b\^2\ r\^6 + 98\ r\^8)\)\)\/\(r\^12\ \((a\^2\ b\
\^2 + r\^2)\)\^6\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[simp[gRstarR]]\)], "Input"],

Cell[BoxData[
    \(\(49766400\ m\^2\ \((25\ a\^8\ b\^8 + 122\ a\^6\ b\^6\ r\^2 + 235\ a\^4\
\ b\^4\ r\^4 + 220\ a\^2\ b\^2\ r\^6 + 98\ r\^8)\)\)\/\(r\^12\ \((a\^2\ b\^2 \
+ r\^2)\)\^6\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(backup[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]20\[InvisibleSpace]":"\
\[InvisibleSpace]15\[InvisibleSpace]":"\[InvisibleSpace]55\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]8\[InvisibleSpace]\
"/"\[InvisibleSpace]2003\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 20, ":", 15, ":", 55, " on ", 4, "/", 8, "/", 2003, "."],
      Editable->False]], "Print"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "ken", \
"projects", "ptah", "mathematica"}, "dbraneval.nb.ps", CharacterEncoding -> \
"ISO8859-1"],
"Magnification"->1},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 70, 1, 27, "Input"],
Cell[1874, 57, 2984, 44, 375, "Print"],
Cell[4861, 103, 81, 1, 23, "Print"]
}, Closed]],
Cell[4957, 107, 368, 7, 88, "Input"],

Cell[CellGroupData[{
Cell[5350, 118, 68, 1, 27, "Input"],
Cell[5421, 121, 121, 3, 23, "Print"],
Cell[5545, 126, 121, 3, 23, "Print"],
Cell[5669, 131, 121, 3, 23, "Print"],
Cell[5793, 136, 121, 3, 23, "Print"],
Cell[5917, 141, 121, 3, 23, "Print"],
Cell[6041, 146, 121, 3, 23, "Print"],
Cell[6165, 151, 121, 3, 23, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[6323, 159, 51, 1, 24, "Input"],
Cell[6377, 162, 310, 9, 25, "Print"],
Cell[6690, 173, 345, 10, 25, "Print"],
Cell[7038, 185, 256, 7, 25, "Print"],
Cell[7297, 194, 256, 7, 25, "Print"],
Cell[7556, 203, 256, 7, 25, "Print"],
Cell[7815, 212, 256, 7, 25, "Print"],
Cell[8074, 221, 256, 7, 25, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8367, 233, 56, 1, 27, "Input"],
Cell[8426, 236, 252, 5, 23, "Print"],
Cell[8681, 243, 283, 6, 25, "Print"],
Cell[8967, 251, 463, 13, 41, "Print"],
Cell[9433, 266, 286, 6, 40, "Print"],
Cell[9722, 274, 286, 6, 37, "Print"],
Cell[10011, 282, 286, 6, 38, "Print"],
Cell[10300, 290, 286, 6, 37, "Print"],
Cell[10589, 298, 286, 6, 37, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[10912, 309, 51, 1, 24, "Input"],
Cell[10966, 312, 455, 13, 41, "Print"],
Cell[11424, 327, 525, 15, 42, "Print"],
Cell[11952, 344, 591, 17, 41, "Print"],
Cell[12546, 363, 657, 19, 41, "Print"],
Cell[13206, 384, 647, 19, 25, "Print"],
Cell[13856, 405, 316, 8, 41, "Print"],
Cell[14175, 415, 274, 6, 39, "Print"],
Cell[14452, 423, 1092, 34, 41, "Print"],
Cell[15547, 459, 350, 9, 41, "Print"],
Cell[15900, 470, 350, 9, 41, "Print"],
Cell[16253, 481, 641, 19, 41, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[16931, 505, 54, 1, 24, "Input"],
Cell[16988, 508, 337, 9, 41, "Print"],
Cell[17328, 519, 369, 10, 41, "Print"],
Cell[17700, 531, 401, 11, 42, "Print"],
Cell[18104, 544, 433, 12, 42, "Print"],
Cell[18540, 558, 437, 12, 41, "Print"],
Cell[18980, 572, 339, 9, 41, "Print"],
Cell[19322, 583, 296, 7, 41, "Print"],
Cell[19621, 592, 896, 27, 41, "Print"],
Cell[20520, 621, 319, 8, 37, "Print"],
Cell[20842, 631, 319, 8, 37, "Print"],
Cell[21164, 641, 1261, 39, 41, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[22462, 685, 39, 1, 24, "Input"],
Cell[22504, 688, 69, 1, 45, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[22610, 694, 49, 1, 27, "Input"],
Cell[22662, 697, 35, 1, 27, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[22734, 703, 52, 1, 27, "Input"],
Cell[22789, 706, 194, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[23020, 714, 54, 1, 27, "Input"],
Cell[23077, 717, 291, 4, 67, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[23405, 726, 54, 1, 27, "Input"],
Cell[23462, 729, 290, 4, 67, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[23789, 738, 54, 1, 27, "Input"],
Cell[23846, 741, 308, 5, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[24191, 751, 55, 1, 27, "Input"],
Cell[24249, 754, 742, 11, 121, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[25028, 770, 53, 1, 27, "Input"],
Cell[25084, 773, 202, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[25323, 781, 54, 1, 27, "Input"],
Cell[25380, 784, 200, 3, 48, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[25617, 792, 46, 1, 27, "Input"],
Cell[25666, 795, 425, 7, 23, "Print"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

