(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     41648,        787]*)
(*NotebookOutlinePosition[     42328,        812]*)
(*  CellTagsIndexPosition[     42284,        808]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["\<\
(* PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 \
Kenneth R. Koehler 

 e-mail address - kenneth.koehler@uc.edu

 This program is free software; you can redistribute it and/or modify it \
under 
 the terms of the GNU General Public License as published by the Free \
Software 
 Foundation; either version 2 of the License, or (at your option) any later \
version. 

 This program is distributed in the hope that it will be useful, but WITHOUT \
ANY WARRANTY;
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A \
PARTICULAR
 PURPOSE. See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public License along with \
this 
 program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 Cambridge, MA 02139, USA. 
 
 ccif.nb\
\>", "Text"],

Cell["\<\
time coordinate must be last
gp, ginvp, and irmdgp (inverse root minus det g) must be specified \
(parameterized versions) *)\
\>", "Text"],

Cell[BoxData[
    \(\(Off[General::spell, General::spell1, Syntax::com];\)\)], "Input"],

Cell[BoxData[
    \(<< ~/projects/ptah/mathematica/sugra.m\)], "Input"],

Cell[BoxData[{
    \(\(findenom[xx_Rational]\  := \ 
        Block[{nn, 
            ff}, \((nn\  = \ Denominator[xx]; \n\t\t\tff\  = \ 
              FactorInteger[
                nn]; \n\t\t\tDo[\(AppendTo[ints, ff[\([ii, 1]\)]];\), {ii, 
                Length[ff]}]; \n\t\t\tints\  = \ 
              Union[ints];)\)];\)\), "\n", 
    \(\(findenom[xx_Times]\  := \ 
        Do[findenom[xx[\([ii]\)]], {ii, Length[xx]}];\)\), "\n", 
    \(\(findenom[xx_Plus]\  := \ 
        Do[findenom[xx[\([ii]\)]], {ii, Length[xx]}];\)\), "\n", 
    \(\(findenom[xx_List]\  := \ 
        Block[{nn}, \((nn\  = \ Flatten[xx]; \n\t\t\t\tDo[
              findenom[nn[\([ii]\)]], {ii, Length[nn]}])\)];\)\n\), "\n", 
    \(\(sub[xx_]\  := \ If[xx == 1, Length[coords], xx - 1];\)\n\), "\n", 
    \(\(countfactors[xx_Times]\  := \ 
        Sum[If[Head[xx[\([ii]\)]] =!= Integer, \n\t\t\tcountfactors[
              xx[\([ii]\)]], 0], {ii, Length[xx]}];\)\), "\n", 
    \(\(countfactors[xx_Rational]\  := \ 
        Length[FactorInteger[Denominator[xx]]];\)\), "\n", 
    \(\(countfactors[xx_]\  := \ 1;\)\n\), "\n", 
    \(\(coeff[xx_]\  := \ 
        Block[{}, \((If[Head[xx] === Complex, 
              Return[xx/I], ]; \[IndentingNewLine]If[Head[xx] === Rational, 
              Return[Sign[xx] Abs[Numerator[xx]]], ]; \[IndentingNewLine]If[
              NumericQ[xx], Return[xx], ]; \[IndentingNewLine]If[
              Head[xx] =!= Times, Return[1], ]; \[IndentingNewLine]If[
              Head[xx[\([1]\)]] === Rational, 
              Return[Sign[xx[\([1]\)]] 
                  Abs[Numerator[xx[\([1]\)]]]], ]; \[IndentingNewLine]If[
              NumericQ[xx[\([1]\)]], Return[xx[\([1]\)]], Return[1], 
              Return[1]];)\)];\)\n\), "\n", 
    \(\(whichvar[xx_Rational, bb_]\  := \ 
        Block[{nn}, \((\n\t\t\tnn\  = \ 
              FactorInteger[
                Denominator[xx]]; \n\t\t\tDo[\((WriteString[
                  mout, \(Position[var, nn[\([ll, 1]\)]]\)[\([1, 1]\)] - 
                    1, "\<,\>", \(-nn[\([ll, 2]\)]\)]; \n\t\t\t\t\tIf[
                  bb || \((ll < Length[nn])\), 
                  WriteString[mout, "\<,\>"], ];)\), {ll, 
                Length[nn]}];)\)];\)\), "\n", 
    \(\(whichvar[xx_Power, 
          bb_]\  := \ \((WriteString[
            mout, \(Position[var, xx[\([1]\)]]\)[\([1, 1]\)] - 1, "\<,\>", 
            xx[\([2]\)]]; \n\t\tIf[bb, 
            WriteString[mout, "\<,\>"], ];)\);\)\), "\n", 
    \(\(whichvar[xx_Symbol, 
          bb_]\  := \ \((WriteString[
            mout, \(Position[var, xx]\)[\([1, 1]\)] - 1, "\<,\>", 
            1]; \n\t\tIf[bb, WriteString[mout, "\<,\>"], ];)\);\)\), "\n", 
    \(\(whichvar[xx_Complex, 
          bb_]\  := \ \((WriteString[
            mout, \(Position[var, I]\)[\([1, 1]\)] - 1, "\<,\>", 1]; \n\t\tIf[
            bb, WriteString[mout, "\<,\>"], ];)\);\)\), "\n", 
    \(\(whichvar[xx_, 
          bb_]\  := \ \((WriteString[mout, \(-1\), "\<,\>", 0]; \n\t\tIf[bb, 
            WriteString[mout, "\<,\>"], ];)\);\)\n\), "\n", 
    \(\(dopows[xx_Times, bb_]\  := \ 
        Do[If[Head[xx[\([kk]\)]] =!= Integer, \n\t\t\twhichvar[xx[\([kk]\)], 
              bb || \((kk < Length[xx])\)], ], {kk, Length[xx]}];\)\), "\n", 
    \(\(dopows[xx_, bb_]\  := \ whichvar[xx, bb];\)\n\), "\n", 
    \(\(exprout[xx_, ss_]\  := \ 
        Block[{terms, numterms, factors, 
            nn}, \((\[IndentingNewLine]nn = 
              Length[xx]; \[IndentingNewLine]WriteString[mout, "\<int \>", 
              ss, "\<numtrm[\>", 
              nn, "\<] = {\>"]; \[IndentingNewLine]terms\  = \ 
              0; \[IndentingNewLine]Do[\((numterms = \ 
                  If[Head[xx[\([ii]\)]] === Plus, Length[xx[\([ii]\)]], 
                    1]; \[IndentingNewLine]WriteString[mout, 
                  numterms]; \[IndentingNewLine]terms\  = \ 
                  terms + numterms; \[IndentingNewLine]If[ii < nn, 
                  WriteString[mout, "\<,\>"], ];)\), {ii, 
                nn}]; \[IndentingNewLine]WriteString[
              mout, "\<};\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<int \>", ss, "\<numfct[\>", 
              terms, "\<] = {\>"]; \[IndentingNewLine]factors\  = \ 
              If[sortvar[\([1]\)] === 0, 0, 
                fun + ans]; \[IndentingNewLine] (*\ \(Print["\<sortvar= \>", 
                  sortvar];\)\ *) \[IndentingNewLine]Do[\((\(If[
                    Head[xx[\([ii]\)]] === 
                      Plus, \[IndentingNewLine]\((Do[
                        If[jj != 
                            sortvar[\([ii]\)], \((WriteString[mout, 
                              countfactors[
                                xx[\([ii, 
                                    jj]\)]]]; \[IndentingNewLine]factors = 
                              factors + 
                                countfactors[
                                  xx[\([ii, 
                                      jj]\)]]; \[IndentingNewLine]If[\((ii < 
                                    nn)\) || \((jj < 
                                    Length[
                                      xx[\([ii]\)]])\) || \
\((sortvar[\([ii]\)] > 0)\), \[IndentingNewLine]WriteString[
                                mout, "\<,\>"], ];)\), ], {jj, 
                          Length[xx[\([ii]\)]]}]; \[IndentingNewLine]If[
                        sortvar[\([ii]\)] =!= 
                          0, \((WriteString[mout, 1]; \[IndentingNewLine]If[
                            ii < nn, 
                            WriteString[
                              mout, "\<,\>"], ];)\), ];)\), \
\[IndentingNewLine]\((WriteString[mout, 
                        countfactors[
                          xx[\([ii]\)]]]; \[IndentingNewLine]factors = 
                        factors + 
                          countfactors[xx[\([ii]\)]]; \[IndentingNewLine]If[
                        ii < nn, 
                        WriteString[mout, "\<,\>"], ];)\)];\))\), {ii, 
                nn}]; \[IndentingNewLine]WriteString[
              mout, "\<};\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<int \>", ss, "\<pwr[\>", 
              2  factors, "\<] = {\>"]; \[IndentingNewLine]Do[\((If[
                  Head[xx[\([ii]\)]] === 
                    Plus, \[IndentingNewLine]\((Do[
                      If[jj != 
                          sortvar[\([ii]\)], \[IndentingNewLine]\((dopows[
                            xx[\([ii, 
                                jj]\)], \((ii < nn)\) || \((jj < 
                                  Length[
                                    xx[\([ii]\)]])\) || \((sortvar[\([ii]\)] \
> 0)\)]; \[IndentingNewLine]WriteString[
                            mout, "\< \>"]; \[IndentingNewLine]If[\((Length[
                                    xx[\([ii]\)]] > 4)\) && \((Mod[jj, 4] == 
                                  0)\) && \((jj != 
                                  Length[
                                    xx[\([ii]\)]])\), \
\[IndentingNewLine]WriteString[
                              mout, "\<\n                   \>"], ];)\), ], \
{jj, Length[xx[\([ii]\)]]}]; \[IndentingNewLine]If[sortvar[\([ii]\)] > 0, 
                      dopows[xx[\([ii, sortvar[\([ii]\)]]\)], 
                        ii < nn], ];)\), \[IndentingNewLine]dopows[
                    xx[\([ii]\)], \((ii < nn)\)]]; \[IndentingNewLine]\(If[
                    ii < nn, 
                    WriteString[mout, "\<\n               \>"], ];\))\), {ii, 
                nn}]; \[IndentingNewLine]WriteString[
              mout, "\<};\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<long long \>", ss, "\<co[\>", 
              terms, "\<] = {\>"]; \[IndentingNewLine]Do[\((If[
                  Head[xx[\([ii]\)]] === 
                    Plus, \[IndentingNewLine]\((Do[
                      If[jj != 
                          sortvar[\([ii]\)], \
\((\[IndentingNewLine]WriteString[mout, 
                            coeff[xx[\([ii, 
                                  jj]\)]]]; \[IndentingNewLine]If[\((ii < 
                                  nn)\) || \((jj < 
                                  Length[
                                    xx[\([ii]\)]])\) || \((sortvar[\([ii]\)] \
> 0)\), \[IndentingNewLine]WriteString[mout, "\<,\>"], ];)\), ], {jj, 
                        Length[xx[\([ii]\)]]}]; \[IndentingNewLine]If[
                      sortvar[\([ii]\)] =!= 
                        0, \((WriteString[mout, 
                          coeff[xx[\([ii, 
                                sortvar[\([ii]\)]]\)]]]; \
\[IndentingNewLine]If[ii < nn, 
                          WriteString[
                            mout, "\<,\>"], ];)\), ];)\), \[IndentingNewLine]\
\((WriteString[mout, coeff[
                        xx[\([ii]\)]]]; \[IndentingNewLine]If[\((ii < nn)\), 
                      WriteString[
                        mout, "\<,\>"], ];)\)]; \[IndentingNewLine]\(If[
                    ii < nn, 
                    WriteString[mout, "\<\n              \>"], ];\))\), {ii, 
                nn}]; \[IndentingNewLine]\(WriteString[
                mout, "\<};\n\>"];\))\)];\)\), "\[IndentingNewLine]", 
    \(\(indout[xx_List, ss_, nn_List]\  := \ 
        Block[{tt, tn, 
            rank}, \((tt\  = Flatten[xx]; \[IndentingNewLine]rank\  = \ 
              TensorRank[xx]; \[IndentingNewLine]tn\  = \ 
              0; \[IndentingNewLine]Do[
              If[tt[\([ii]\)] =!= 0, tn = tn + 1, ], {ii, 
                Length[tt]}]; \[IndentingNewLine]WriteString[
              mout, "\<#define \>", ss, "\<cnt \>", 
              tn, "\<\n\>"]; \[IndentingNewLine]If[
              rank \[Equal] 
                1, \((WriteString[mout, "\<int \>", ss, "\<ind[\>", 
                  tn, "\<] = {\>"]; \[IndentingNewLine]Do[
                  If[xx[\([ii]\)] =!= 
                      0, \((WriteString[
                        mout, \[IndentingNewLine]If[nn[\([1]\)] < 0, ii - 1, 
                          If[ii == nn[\([1]\)], 0, 
                            ii]]]; \[IndentingNewLine]If[\((ii < 
                            Abs[nn[\([1]\)]])\), 
                        WriteString[mout, "\<,\>"], ];)\), ], {ii, 
                    Abs[nn[\([1]\)]]}];)\), ]; \[IndentingNewLine]If[
              rank \[Equal] 
                2, \((WriteString[mout, "\<int \>", ss, "\<ind[\>", 
                  2  tn, "\<] = {\>"]; \[IndentingNewLine]Do[
                  Do[If[xx[\([ii, jj]\)] =!= 
                        0, \((WriteString[
                          mout, \[IndentingNewLine]If[nn[\([1]\)] < 0, 
                            ii - 1, 
                            If[ii == nn[\([1]\)], 0, 
                              ii]], "\<,\>", \[IndentingNewLine]If[
                            nn[\([2]\)] < 0, jj - 1, 
                            If[jj == nn[\([2]\)], 0, 
                              jj]]]; \[IndentingNewLine]If[\((ii < 
                                Abs[nn[\([1]\)]])\) || \((jj < 
                                Abs[nn[\([2]\)]])\), 
                          WriteString[
                            mout, "\<,\>"], ];)\), ], \[IndentingNewLine]{jj, 
                      Abs[nn[\([2]\)]]}], {ii, 
                    Abs[nn[\([1]\)]]}];)\), ]; \[IndentingNewLine]If[
              rank \[Equal] 
                3, \((WriteString[mout, "\<int \>", ss, "\<ind[\>", 
                  3  tn, "\<] = {\>"]; \[IndentingNewLine]Do[
                  Do[Do[
                      If[xx[\([ii, jj, kk]\)] =!= 
                          0, \[IndentingNewLine]\((WriteString[
                            mout, \[IndentingNewLine]If[nn[\([1]\)] < 0, 
                              ii - 1, 
                              If[ii == nn[\([1]\)], 0, 
                                ii]], "\<,\>", \[IndentingNewLine]If[
                              nn[\([2]\)] < 0, jj - 1, 
                              If[jj == nn[\([2]\)], 0, 
                                jj]], "\<,\>", \[IndentingNewLine]If[
                              nn[\([3]\)] < 0, kk - 1, 
                              If[kk == nn[\([3]\)], 0, 
                                kk]]]; \[IndentingNewLine]If[\((ii < 
                                  Abs[nn[\([1]\)]])\) || \((jj < 
                                  Abs[nn[\([2]\)]])\) || \((kk < 
                                  Abs[nn[\([3]\)]])\), 
                            WriteString[
                              mout, "\<,\>"], ];)\), ], \
\[IndentingNewLine]{kk, Abs[nn[\([3]\)]]}], {jj, Abs[nn[\([2]\)]]}], {ii, 
                    Abs[nn[\([1]\)]]}];)\), ]; \[IndentingNewLine]If[
              rank \[Equal] 
                4, \((WriteString[mout, "\<int \>", ss, "\<ind[\>", 
                  4  tn, "\<] = {\>"]; \[IndentingNewLine]Do[
                  Do[Do[Do[
                        If[xx[\([ii, jj, kk, ll]\)] =!= 
                            0, \[IndentingNewLine]\((WriteString[
                              mout, \[IndentingNewLine]If[nn[\([1]\)] < 0, 
                                ii - 1, 
                                If[ii == nn[\([1]\)], 0, 
                                  ii]], "\<,\>", \[IndentingNewLine]If[
                                nn[\([2]\)] < 0, jj - 1, 
                                If[jj == nn[\([2]\)], 0, 
                                  jj]], "\<,\>", \[IndentingNewLine]If[
                                nn[\([3]\)] < 0, kk - 1, 
                                If[kk == nn[\([3]\)], 0, 
                                  kk]], "\<,\>", \[IndentingNewLine]If[
                                nn[\([4]\)] < 0, ll - 1, 
                                If[ll == nn[\([4]\)], 0, 
                                  ll]]]; \[IndentingNewLine]If[\((ii < 
                                    Abs[nn[\([1]\)]])\) || \((jj < 
                                    Abs[nn[\([2]\)]])\) || \((kk < 
                                    Abs[nn[\([3]\)]])\) \
|| \[IndentingNewLine]\((ll < Abs[nn[\([4]\)]])\), 
                              WriteString[
                                mout, "\<,\>"], ];)\), ], \
\[IndentingNewLine]{ll, Abs[nn[\([4]\)]]}], {kk, 
                        Abs[nn[\([3]\)]]}], \[IndentingNewLine]{jj, 
                      Abs[nn[\([2]\)]]}], {ii, 
                    Abs[nn[\([1]\)]]}];)\), ]; \[IndentingNewLine]WriteString[
              mout, "\<};\n\>"];)\)];\)\)}], "Input"],

Cell[BoxData[
    \(\(simpleconcerns[]\  := \ 
        Block[{dep, symdep, passtruc, flatps, remps, 
            sortsymb}, \((\[IndentingNewLine]dep = 
              Table[functions[\([ii, 3]\)], {ii, fun}]; \[IndentingNewLine]If[
              constants[\([1]\)] === I, constart = 2, 
              constart = 1]; \[IndentingNewLine]Do[
              Do[AppendTo[dep[\([ii]\)], 
                  D[functions[\([ii, 2]\)], constants[\([jj]\)]]], {jj, 
                  constart, con}], {ii, fun}]; \[IndentingNewLine]symdep = 
              Table[{}, {ii, 
                  con + dim - constart + 1}]; \[IndentingNewLine]Do[
              Do[If[dep[\([jj, ii]\)] =!= 0, 
                  AppendTo[symdep[\([ii]\)], jj], ], {jj, Length[dep]}], {ii, 
                Length[symdep]}]; 
            Do[Do[Do[
                  If[D[gp[\([jj, kk]\)], coords[\([sub[ii]]\)]] =!= 0, 
                    symdep[\([ii]\)] = 
                      Union[symdep[\([ii]\)], {0}], ], \
\[IndentingNewLine]{kk, dim}], {jj, dim}], {ii, dim}]; 
            Do[Do[Do[
                  If[D[gp[\([jj, kk]\)], constants[\([ii]\)]] =!= 0, 
                    AppendTo[symdep[\([ii + dim - constart + 1]\)], 
                      0], ], \[IndentingNewLine]{kk, dim}], {jj, dim}], {ii, 
                constart, con}]; 
            symdep\  = \ 
              Table[Union[symdep[\([ii]\)]], {ii, 
                  Length[symdep]}];  (*\ \(symdep\  = \ 
                  Select[Union[symdep], 
                    Length[#1] > 1 &];\)\ *) \[IndentingNewLine]symdep\  = \ 
              Select[Union[symdep], Length[#1] > 0 &]; 
            Do[If[MemberQ[symdep[\([ii]\)], 0], 
                symdep[\([ii]\)] = Rest[symdep[\([ii]\)]], ], {ii, 
                Length[symdep]}]; \[IndentingNewLine]passtruc = {}; \
\[IndentingNewLine]pp = 0; \[IndentingNewLine]While[
              Length[symdep] > 
                0, \((\[IndentingNewLine]AppendTo[passtruc, 
                  symdep[\([1]\)]]; \[IndentingNewLine]symdep = 
                  Rest[symdep]; \[IndentingNewLine]pp = 
                  pp + 1; \[IndentingNewLine]ss = 1; \[IndentingNewLine]While[
                  ss \[LessEqual] Length[symdep], \[IndentingNewLine]If[
                    Length[Intersection[passtruc[\([pp]\)], 
                          symdep[\([ss]\)]]] > 
                      0, \[IndentingNewLine]\((passtruc[\([pp]\)] = 
                        Union[passtruc[\([pp]\)], 
                          symdep[\([ss]\)]]; \[IndentingNewLine]symdep = 
                        Drop[symdep, {ss}]; \[IndentingNewLine]ss = 1;)\), 
                    ss = ss + 1]];)\)]; \[IndentingNewLine]WriteString[
              mout, "\<int number_of_passes = \>", 
              Length[passtruc], "\<;\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<int number_to_simplify;\n\>"]; \
\[IndentingNewLine]WriteString[mout, "\<int pass_num_fns[\>", 
              Length[passtruc], "\<] = \>"]; \[IndentingNewLine]WriteString[
              mout, Table[
                Length[passtruc[\([i]\)]], {i, 
                  Length[passtruc]}]]; \[IndentingNewLine]WriteString[
              mout, "\<;\n\>"]; \[IndentingNewLine]flatps\  = \ 
              Flatten[passtruc] - 1; \[IndentingNewLine]WriteString[
              mout, "\<int pass_fns[\>", Length[flatps], "\<] = \>", 
              flatps, "\<;\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<int pass_fn_index;\n\>"]; \[IndentingNewLine]sortvar = \
{}; \[IndentingNewLine]sortsymb\  = \ {}; \[IndentingNewLine]WriteString[
              mout, "\<int sort_variable[\>", 
              fun, "\<] = {\>"]; \[IndentingNewLine]Do[\((funlen\  = \ 
                  Length[functions[\([ii, 2]\)]]; \[IndentingNewLine]it\  = \ 
                  0; \[IndentingNewLine]this = 
                  functions[\([ii, 2]\)]; \[IndentingNewLine]If[
                  AtomQ[this] && \((\(! NumberQ[this]\))\), \((it\  = \ this; 
                    AppendTo[sortsymb, it]; 
                    itn = 1;)\), ]; \[IndentingNewLine]Do[\((If[
                      AtomQ[this[\([jj]\)]] && \((\(! NumberQ[
                              this[\([jj]\)]]\))\), \((it\  = \ 
                          this[\([jj]\)]; AppendTo[sortsymb, it]; itn = jj; \ 
                        Break[];)\), ]; \[IndentingNewLine]If[\((Head[
                            this[\([jj]\)]] === 
                          Power)\), \[IndentingNewLine]If[\((\(! MemberQ[
                              sortsymb, this[\([jj, 1]\)]]\))\), \((it\  = \ 
                            this[\([jj, 1]\)]; AppendTo[sortsymb, it]; 
                          itn = jj; \ 
                          Break[];)\), ], ]; \[IndentingNewLine]If[\((Head[
                            this[\([jj]\)]] === 
                          Times)\), \[IndentingNewLine]If[\((Head[
                                this[\([jj, 2]\)]] === 
                              Power)\) && \((Head[this[\([jj, 1]\)]] === 
                              Integer)\), \[IndentingNewLine]If[\((\(! \
MemberQ[sortsymb, this[\([jj, 2, 1]\)]]\))\), \((it\  = \ 
                              this[\([jj, 2, 1]\)]; AppendTo[sortsymb, it]; 
                            itn = jj; \ 
                            Break[];)\), ], \[IndentingNewLine]If[\((Head[
                                  this[\([jj, 2]\)]] === 
                                Symbol)\) && \((Head[this[\([jj, 1]\)]] === 
                                Integer)\), \[IndentingNewLine]If[\((\(! \
MemberQ[sortsymb, this[\([jj, 2]\)]]\))\), \((it\  = \ this[\([jj, 2]\)]; 
                              AppendTo[sortsymb, it]; itn = jj; \ 
                              Break[];)\), ], ]], ];)\), \
\[IndentingNewLine]{jj, 
                    funlen}]; \[IndentingNewLine]If[\(! NumberQ[
                      it]\), \((WriteString[
                      mout, \(-1\) + \(Position[var, it]\)[\([1, 
                            1]\)]]; \[IndentingNewLine]If[
                      Head[this] === Times, 
                      itn = 1, ]; \[IndentingNewLine]AppendTo[sortvar, 
                      itn]; \[IndentingNewLine]Do[
                      AppendTo[sortvar, 0], {i, 
                        dim}];)\), ]; \[IndentingNewLine]\(If[ii < fun, 
                    WriteString[mout, "\<,\>"], ];\))\), {ii, 
                fun}]; \[IndentingNewLine]WriteString[
              mout, "\<};\n\>"];\[IndentingNewLine])\)];\)\)], "Input"],

Cell[BoxData[{
    \(\(ints\  = \ {};\)\), "\[IndentingNewLine]", 
    \(\(constants\  = \ {};\)\), "\[IndentingNewLine]", 
    \(\(functions\  = \ {};\)\), "\[IndentingNewLine]", 
    \(\(ansatze\  = \ {};\)\)}], "Input"],

Cell[BoxData[
    \(\(headers[]\  := \ 
        Block[{argtest, 
            argexp}, \((dim\  = \ 
              Length[coords]; \[IndentingNewLine]con\  = \ 
              Length[constants]; \[IndentingNewLine]fun\  = \ 
              Length[functions]; \[IndentingNewLine]ans\  = \ 
              Length[ansatze]; \[IndentingNewLine]functions\  = \ 
              Expand[functions]; \[IndentingNewLine]ptab\  = \ {}; \
\[IndentingNewLine]Do[\((dummy = 
                  FactorInteger[
                    Exponent[
                      functions[\([ii, 
                          1]\)], \[IndentingNewLine]\(\((#1 /. {aa_^
                                    bb_ \[Rule] aa, xx_ \[Rule] xx})\) &\)[
                        functions[\([ii, 
                            1]\)]]]]; \[IndentingNewLine]AppendTo[ptab, 
                  Table[dummy[\([jj, 1]\)], {jj, 
                      Length[dummy]}]];)\), \[IndentingNewLine]{ii, 
                Length[functions]}]; 
            ints = Rest[
                Union[Flatten[
                    FactorInteger[\[IndentingNewLine]Union[
                        Flatten[{ints, ptab, 2, dim, dim - 1, 
                            dim - 2}]]]]]]; \[IndentingNewLine]Do[
              findenom[functions[\([ii, 2]\)]], {ii, 
                fun}]; \[IndentingNewLine]Do[
              findenom[ansatze[\([ii, 2]\)]], {ii, 
                ans}]; \[IndentingNewLine]findenom[
              gp]; \[IndentingNewLine]findenom[
              ginvp]; \[IndentingNewLine]findenom[
              irmdgp]; \[IndentingNewLine]num\  = \ 
              dim\  + \ fun\  + \ ans\  + \ Length[constants]\  + \ 
                Length[ints]; \[IndentingNewLine]pout\  = \ 
              OpenWrite[
                ptahdir <> "\<include/params.h\>"]; \
\[IndentingNewLine]WriteString[pout, "\<#define dimension \>", 
              dim, "\<\n\>"]; \[IndentingNewLine]Close[
              pout]; \[IndentingNewLine]pout\  = \ 
              OpenAppend[
                ptahdir <> "\<include/params.h\>"]; \
\[IndentingNewLine]WriteString[pout, "\<#define spinor_dimension \>", 
              2^Floor[dim/2], "\<\n\>"]; \[IndentingNewLine]WriteString[
              pout, "\<#define number_of_variables \>", \
\[IndentingNewLine]dim + fun + ans + con + 
                Length[ints], "\<\n\>"]; \[IndentingNewLine]WriteString[
              pout, "\<#define number_of_functions \>", 
              fun, "\<\n\>"]; \[IndentingNewLine]WriteString[
              pout, "\<#define number_of_ansatze \>", 
              ans, "\<\n\>"]; \[IndentingNewLine]WriteString[
              pout, "\<#define number_of_constants \>", 
              con, "\<\n\>"]; \[IndentingNewLine]WriteString[
              pout, "\<#define number_of_fractions \>", 
              Length[ints], "\<\n\>"]; \[IndentingNewLine]Close[
              pout]; \[IndentingNewLine]var\  = \ {}; \[IndentingNewLine]Do[
              AppendTo[var, coords[\([sub[ii]]\)]], {ii, 
                dim}]; \[IndentingNewLine]Do[
              If[AtomQ[functions[\([ii, 1]\)]], 
                AppendTo[var, 
                  functions[\([ii, 1]\)]], \[IndentingNewLine]AppendTo[var, 
                  functions[\([ii, 1, 1]\)]]], {ii, 
                fun}]; \[IndentingNewLine]Do[
              AppendTo[var, ansatze[\([ii, 1]\)]], {ii, 
                ans}]; \[IndentingNewLine]var\  = \ 
              Join[var, constants]; \[IndentingNewLine]var\  = \ 
              Join[var, ints]; \[IndentingNewLine]mout\  = \ 
              OpenWrite[
                ptahdir <> "\<include/metric.h\>"]; \[IndentingNewLine]Close[
              mout]; \[IndentingNewLine]mout\  = \ 
              OpenAppend[
                ptahdir <> "\<include/metric.h\>"]; \
\[IndentingNewLine]WriteString[mout, "\<char variables[\>", 
              num, "\<] = {\>"]; \[IndentingNewLine]Do[\((WriteString[
                  mout, "\<'\>", 
                  var[\([ii]\)], "\<'\>"]; \[IndentingNewLine]If[ii < num, 
                  WriteString[mout, "\<,\>"], ];)\), {ii, 
                num}]; \[IndentingNewLine]WriteString[
              mout, "\<};\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<int integers[\>", 
              Length[ints], "\<] = {\>"]; \
\[IndentingNewLine]Do[\((WriteString[mout, 
                  ints[\([ii]\)]]; \[IndentingNewLine]If[ii < Length[ints], 
                  WriteString[mout, "\<,\>"], ];)\), {ii, 
                Length[ints]}]; \[IndentingNewLine]WriteString[
              mout, "\<};\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<bool argument[\>", 
              dim*\((fun + 
                    ans)\), "\<] = {\>"]; \[IndentingNewLine]Do[\((Do[\((If[
                      functions[\([ii, 3, jj]\)] =!= 
                        0, \[IndentingNewLine]WriteString[mout, "\<true\>"], 
                      WriteString[
                        mout, "\<false\>"]]; \[IndentingNewLine]\(If[\((ii == 
                              fun)\) && \((jj == dim)\), , 
                        WriteString[mout, "\<,\>"]];\))\), {jj, 
                    dim}]; \[IndentingNewLine]\(If[ii < fun, 
                    WriteString[
                      mout, "\<\n                     \>"], ];\))\), \
\[IndentingNewLine]{ii, fun}]; \[IndentingNewLine]If[ans > 0, 
              WriteString[
                mout, "\<,\n                     \>"], ]; \
\[IndentingNewLine]Do[\((Do[\((If[
                      ansatze[\([ii, 3, jj]\)] =!= 
                        0, \[IndentingNewLine]WriteString[mout, "\<true\>"], 
                      WriteString[
                        mout, "\<false\>"]]; \[IndentingNewLine]\(If[\((ii \
\[Equal] ans)\) && \((jj == dim)\), , WriteString[mout, "\<,\>"]];\))\), {jj, 
                    dim}]; \[IndentingNewLine]\(If[ii < ans, 
                    WriteString[
                      mout, "\<\n                     \>"], ];\))\), \
\[IndentingNewLine]{ii, ans}]; \[IndentingNewLine]WriteString[
              mout, "\<};\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<int function_power[\>", 
              fun, "\<] = {\>"]; \[IndentingNewLine]Do[\((WriteString[mout, 
                  Exponent[functions[\([ii, 1]\)], 
                    var[\([ii + dim]\)]]]; \[IndentingNewLine]If[ii < fun, 
                  WriteString[mout, "\<,\>"], ];)\), {ii, 
                fun}]; \[IndentingNewLine]WriteString[
              mout, "\<};\n\>"]; \[IndentingNewLine]simpleconcerns[]; \
\[IndentingNewLine]WriteString[mout, "\<term* function_terms[\>", 
              fun, "\<] = {\>"]; \[IndentingNewLine]Do[\((WriteString[
                  mout, "\<NULL\>"]; \[IndentingNewLine]If[ii < fun, 
                  WriteString[mout, "\<,\>"], ];)\), {ii, 
                fun}]; \[IndentingNewLine]WriteString[
              mout, "\<};\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<sum* functions[\>", 
              fun, "\<];\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<sum* derivatives[\>", \((fun + ans)\)*
                dim, "\<];\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<int factor_powers_temp[\>", 
              num, "\<];\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<int factor_powers_temp_d[\>", 
              num, "\<];\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<int factor_powers_temp_s[\>", 
              num, "\<];\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<int factor_powers_temp_i[\>", num, "\<];\n\>"]; 
            WriteString[mout, "\<int factor_powers_dim[\>", 
              num, "\<] = {\>"]; \[IndentingNewLine]Do[
              WriteString[mout, "\<0,\>"], {ii, 
                num - Length[ints]}]; \[IndentingNewLine]Do[\((WriteString[
                  mout, If[
                    Length[ff = 
                          Cases[FactorInteger[dim], {ints[\([ii]\)], _}]] > 
                      0, \(-ff[\([1, 2]\)]\), 0]]; \[IndentingNewLine]If[
                  ii =!= Length[ints], 
                  WriteString[mout, "\<,\>"], ];)\), {ii, 
                Length[ints]}]; \[IndentingNewLine]WriteString[
              mout, "\<};\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<int factor_powers_dimm2[\>", 
              num, "\<] = {\>"]; \[IndentingNewLine]Do[
              WriteString[mout, "\<0,\>"], {ii, 
                num - Length[ints]}]; \[IndentingNewLine]Do[\((WriteString[
                  mout, If[
                    Length[ff = 
                          Cases[FactorInteger[
                              dim - 2], {ints[\([ii]\)], _}]] > 
                      0, \(-ff[\([1, 2]\)]\), 0]]; \[IndentingNewLine]If[
                  ii =!= Length[ints], 
                  WriteString[mout, "\<,\>"], ];)\), {ii, 
                Length[ints]}]; \[IndentingNewLine]WriteString[
              mout, "\<};\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<int factor_powers_dimdimm1[\>", 
              num, "\<] = {\>"]; \[IndentingNewLine]Do[
              WriteString[mout, "\<0,\>"], {ii, 
                num - Length[ints]}]; \[IndentingNewLine]Do[\((WriteString[
                  mout, If[
                    Length[ff = 
                          Cases[FactorInteger[
                              dim*\((dim - 1)\)], {ints[\([ii]\)], _}]] > 
                      0, \(-ff[\([1, 2]\)]\), 0]]; \[IndentingNewLine]If[
                  ii =!= Length[ints], 
                  WriteString[mout, "\<,\>"], ];)\), {ii, 
                Length[ints]}]; \[IndentingNewLine]WriteString[
              mout, "\<};\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<int gcd_factor_powers[\>", 
              num, "\<];\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<int pwr[\>", 
              num, "\<];\n\>"]; \[IndentingNewLine]fntab\  = \ 
              Flatten[{Table[
                    Rest[functions[\([ii]\)]], {ii, 
                      fun}], \[IndentingNewLine]Table[
                    Rest[ansatze[\([ii]\)]], {ii, 
                      ans}]}]; \[IndentingNewLine]Do[
              AppendTo[sortvar, 0], {ii, 
                ans*\((dim\  + \ 1)\)}]; \[IndentingNewLine]exprout[
              fntab, "\<f\>"]; \[IndentingNewLine]sortvar = 
              Table[0, {ii, 
                  dim*dim}]; \[IndentingNewLine]Print["\<functions \
completed...\>"]; \[IndentingNewLine]indout[
              gp, "\<g\>", {dim, 
                dim}]; \[IndentingNewLine]tg\  = \ {}; \[IndentingNewLine]Do[
              Do[If[gp[\([ii, jj]\)] =!= 0, 
                  AppendTo[tg, Expand[gp[\([ii, jj]\)]]], ], {jj, dim}], {ii, 
                dim}]; \[IndentingNewLine]exprout[
              tg, "\<g\>"]; \[IndentingNewLine]Print["\<g completed...\>"]; \
\[IndentingNewLine]indout[
              ginvp, "\<gi\>", {dim, 
                dim}]; \[IndentingNewLine]tg\  = \ {}; \[IndentingNewLine]Do[
              Do[If[ginvp[\([ii, jj]\)] =!= 0, 
                  AppendTo[tg, Expand[ginvp[\([ii, jj]\)]]], ], {jj, 
                  dim}], \[IndentingNewLine]{ii, 
                dim}]; \[IndentingNewLine]exprout[
              tg, "\<gi\>"]; \[IndentingNewLine]Print["\<gi completed...\>"]; \
\[IndentingNewLine]tg\  = \ {Expand[irmdgp]}; \[IndentingNewLine]exprout[
              tg, "\<d\>"]; \[IndentingNewLine]Close[
              mout]; \[IndentingNewLine]Print["\<d \
completed...\>"];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(gammaout[xx_List, nam_]\  := \ 
        Block[{tg, rank, 
            dtab}, \((\[IndentingNewLine]rank\  = \ 
              TensorRank[xx]; \[IndentingNewLine]dim\  = \ 
              Length[coords]; \[IndentingNewLine]con\  = \ 
              Length[constants]; \[IndentingNewLine]fun\  = \ 
              Length[functions]; \[IndentingNewLine]ans\  = \ 
              Length[ansatze]; \[IndentingNewLine]var\  = \ {}; \
\[IndentingNewLine]\(Do[
                  AppendTo[var, coords[\([sub[ii]]\)]], {ii, 
                    dim}];\)\[IndentingNewLine]\(Do[
                  If[AtomQ[functions[\([ii, 1]\)]], 
                    AppendTo[var, 
                      functions[\([ii, 1]\)]], \[IndentingNewLine]AppendTo[
                      var, functions[\([ii, 1, 1]\)]]], {ii, 
                    fun}];\)\[IndentingNewLine]\(Do[
                  AppendTo[var, ansatze[\([ii, 1]\)]], {ii, 
                    ans}];\)\[IndentingNewLine]\(var\  = \ 
                  Join[var, constants];\)\[IndentingNewLine]\(var\  = \ 
                Join[var, ints]\); \[IndentingNewLine]mout\  = \ 
              OpenWrite[
                ptahdir <> "\<gammas/gammas.h\>"]; \
\[IndentingNewLine]WriteString[mout, "\<#define dimension \>", 
              dim, "\<\n\>"]; \[IndentingNewLine]WriteString[
              mout, "\<#define spinor_dimension \>", 
              2^Floor[dim/2], "\<\n\>"]; \[IndentingNewLine]sortvar = 
              Table[0, {ii, dim*dim}]; \[IndentingNewLine]If[rank \[Equal] 1, 
              dtab = {Length[xx]}, \[IndentingNewLine]If[rank \[Equal] 2, 
                dtab = {Length[xx], 
                    Length[xx[\([1]\)]]}, \[IndentingNewLine]If[
                  rank \[Equal] 3, 
                  dtab = {Length[xx], Length[xx[\([1]\)]], 
                      Length[xx[\([1, 1]\)]]}, ]]]; \[IndentingNewLine]Do[
              If[dtab[\([i]\)] \[Equal] \((dim + 1)\), 
                dtab[\([i]\)] = dim, ], {i, 
                Length[dtab]}]; \[IndentingNewLine]Do[
              If[dtab[\([i]\)] \[NotEqual] dim, 
                dtab[\([i]\)] = \(-dtab[\([i]\)]\), ], {i, 
                Length[dtab]}]; \[IndentingNewLine]indout[xx, nam, 
              dtab]; \[IndentingNewLine]Do[
              dtab[\([i]\)]\  = \ Abs[dtab[\([i]\)]], {i, 
                Length[dtab]}]; \[IndentingNewLine]tg\  = \ {}; \
\[IndentingNewLine]If[rank \[Equal] 1, 
              Do[If[xx[\([ii]\)] =!= 0, \[IndentingNewLine]AppendTo[tg, 
                    Expand[xx[\([ii]\)]]], ], {ii, 
                  dtab[\([1]\)]}], ]; \[IndentingNewLine]If[rank \[Equal] 2, 
              Do[Do[
                  If[xx[\([ii, jj]\)] =!= 0, \[IndentingNewLine]AppendTo[tg, 
                      Expand[xx[\([ii, jj]\)]]], ], {jj, 
                    dtab[\([2]\)]}], {ii, 
                  dtab[\([1]\)]}], ]; \[IndentingNewLine]If[rank \[Equal] 3, 
              Do[Do[Do[
                    If[xx[\([ii, jj, kk]\)] =!= 
                        0, \[IndentingNewLine]AppendTo[tg, 
                        Expand[xx[\([ii, jj, kk]\)]]], ], {kk, 
                      dtab[\([3]\)]}], {jj, dtab[\([2]\)]}], {ii, 
                  dtab[\([1]\)]}], ]; \[IndentingNewLine]exprout[tg, 
              nam]; \[IndentingNewLine]Close[mout]; \[IndentingNewLine]Print[
              nam, "\< completed...\>"];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(tensor[xx_List, nam_]\  := \ 
        Block[{tg, rank, 
            dtab}, \((\[IndentingNewLine]rank\  = \ 
              TensorRank[xx]; \[IndentingNewLine]mout\  = \ 
              OpenAppend[
                ptahdir <> "\<include/metric.h\>"]; \
\[IndentingNewLine]sortvar = Table[0, {ii, dim*dim}]; \[IndentingNewLine]If[
              rank \[Equal] 1, 
              dtab = {Length[xx]}, \[IndentingNewLine]If[rank \[Equal] 2, 
                dtab = {Length[xx], 
                    Length[xx[\([1]\)]]}, \[IndentingNewLine]If[
                  rank \[Equal] 3, 
                  dtab = {Length[xx], Length[xx[\([1]\)]], 
                      Length[xx[\([1, 1]\)]]}, \[IndentingNewLine]If[
                    rank \[Equal] 4, 
                    dtab = {Length[xx], 
                        Length[xx[\([1]\)]], \[IndentingNewLine]Length[
                          xx[\([1, 1]\)]], 
                        Length[
                          xx[\([1, 1, 1]\)]]}, ]]]]; \[IndentingNewLine]Do[
              If[dtab[\([i]\)] \[NotEqual] dim, 
                dtab[\([i]\)] = \(-dtab[\([i]\)]\), ], {i, 
                Length[dtab]}]; \[IndentingNewLine]indout[xx, nam, 
              dtab]; \[IndentingNewLine]Do[
              dtab[\([i]\)]\  = \ Abs[dtab[\([i]\)]], {i, 
                Length[dtab]}]; \[IndentingNewLine]tg\  = \ {}; \
\[IndentingNewLine]If[rank \[Equal] 1, 
              Do[If[xx[\([ii]\)] =!= 0, \[IndentingNewLine]AppendTo[tg, 
                    Expand[xx[\([ii]\)]]], ], {ii, 
                  dtab[\([1]\)]}], ]; \[IndentingNewLine]If[rank \[Equal] 2, 
              Do[Do[If[
                    xx[\([ii, jj]\)] =!= 0, \[IndentingNewLine]AppendTo[tg, 
                      Expand[xx[\([ii, jj]\)]]], ], {jj, 
                    dtab[\([2]\)]}], {ii, 
                  dtab[\([1]\)]}], ]; \[IndentingNewLine]If[rank \[Equal] 3, 
              Do[Do[Do[
                    If[xx[\([ii, jj, kk]\)] =!= 
                        0, \[IndentingNewLine]AppendTo[tg, 
                        Expand[xx[\([ii, jj, kk]\)]]], ], {kk, 
                      dtab[\([3]\)]}], {jj, dtab[\([2]\)]}], {ii, 
                  dtab[\([1]\)]}], ]; \[IndentingNewLine]If[rank \[Equal] 4, 
              Do[Do[Do[
                    Do[If[xx[\([ii, jj, kk, ll]\)] =!= 
                          0, \[IndentingNewLine]AppendTo[tg, 
                          Expand[
                            xx[\([ii, jj, kk, 
                                ll]\)]]], ], \[IndentingNewLine]{ll, 
                        dtab[\([4]\)]}], {kk, dtab[\([3]\)]}], {jj, 
                    dtab[\([2]\)]}], {ii, 
                  dtab[\([1]\)]}], ]; \[IndentingNewLine]exprout[tg, 
              nam]; \[IndentingNewLine]Close[mout]; \[IndentingNewLine]Print[
              nam, "\< completed...\>"];)\)];\)\)], "Input"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 0}, {Automatic, 0}},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1739, 51, 849, 25, 320, "Text"],
Cell[2591, 78, 149, 4, 50, "Text"],
Cell[2743, 84, 87, 1, 27, "Input"],
Cell[2833, 87, 71, 1, 27, "Input"],
Cell[2907, 90, 14361, 257, 2011, "Input"],
Cell[17271, 349, 6457, 107, 1099, "Input"],
Cell[23731, 458, 222, 4, 75, "Input"],
Cell[23956, 464, 11444, 205, 1899, "Input"],
Cell[35403, 671, 3375, 59, 571, "Input"],
Cell[38781, 732, 2851, 52, 427, "Input"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

