(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     91565,       1803]*)
(*NotebookOutlinePosition[     92591,       1838]*)
(*  CellTagsIndexPosition[     92547,       1834]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \( (*\ sugra\  - \ 
        functions\ for\ supergravity\ computations\ *) \)], "Input"],

Cell["\<\
(* PTAH - Parallel Tensor Analysis Hybrid   Copyright (C) 2002 \
Kenneth R. Koehler 

 e-mail address - kenneth.koehler@uc.edu

 This program is free software; you can redistribute it and/or modify it \
under 
 the terms of the GNU General Public License as published by the Free \
Software 
 Foundation; either version 2 of the License, or (at your option) any later \
version. 

 This program is distributed in the hope that it will be useful, but WITHOUT \
ANY WARRANTY;
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A \
PARTICULAR
 PURPOSE. See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public License along with \
this 
 program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
 Cambridge, MA 02139, USA. 
 
 sugra.nb *)\
\>", "Text"],

Cell[BoxData[{
    \(\(Off[General::spell, General::spell1, Syntax::com];\)\), "\n", 
    \(\(Off[Power::infy];\)\), "\n", 
    \(\($HistoryLength\  = \ 0;\)\)}], "Input"],

Cell[BoxData[
    \(\(simpler[expr_]\  := \ Factor[PowerExpand[expr]];\)\)], "Input"],

Cell[BoxData[
    \(\(timestamp[]\  := \ 
        Block[{now}, \n\t\t\((now\  = \ Date[]; \n\t\t\tPrint["\<Time is \>", 
              now[\([4]\)], "\<:\>", now[\([5]\)], "\<:\>", 
              now[\([6]\)], \n\t\t\t\t"\< on \>", now[\([2]\)], "\</\>", 
              now[\([3]\)], "\</\>", 
              now[\([1]\)], "\<.\>"];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(backup[]\  := \ 
        Block[{}, \n\t\t\((timestamp[]; \n\t\t\tNotebookSave[
              EvaluationNotebook[]]; \[IndentingNewLine]\t\tRun[
              sync];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(spacetime[xx_List, gg_List]\  := \ 
        Block[{ginv, 
            Christoffel}, \n\t\t\((dim\  = \ Length[xx]; \n\t\t\tginv\  = \ 
              simpler[Inverse[gg]]; \n\t\t\tcurrentst\  = \ {xx, gg, 
                ginv}; \n\t\t\tChristoffel\  = \ connection[]; \n\t\t\tReturn[
              currentst];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(fastst[xx_List, gg_List]\  := \ 
        Block[{ginv, 
            Christoffel}, \n\t\t\((dim\  = \ Length[xx]; \n\t\t\tginv\  = \ 
              simpler[Inverse[gg]]; \n\t\t\tcurrentst\  = \ {xx, gg, 
                ginv}; \n\t\t\tReturn[currentst];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(lorentzmetric[nn_]\  := \ 
        DiagonalMatrix[
          Flatten[{Table[1, {ii, nn - 1}], \(-1\)}]];\)\)], "Input"],

Cell[BoxData[
    \(\(components[ten_List]\  := \ 
        Block[{rank, 
            dummy}, \n\t\t\((rank\  = \ TensorRank[ten]; \n\t\t\tIf[
              rank == 1, \n\t\t\t\tDo[
                If[ten[\([ii]\)] == 0, 
                  dummy = 0, \n\t\t\t\t\t\tPrint["\<[\>", ii, "\<] = \>", 
                    ten[\([ii]\)]], \n\t\t\t\t\t\tPrint["\<[\>", 
                    ii, "\<] = \>", ten[\([ii]\)]]], {ii, 
                  Length[ten]}], ]; \n\t\t\tIf[
              rank == 2, \n\t\t\t\tDo[
                Do[If[ten[\([ii, jj]\)] == 0, 
                    dummy = 0, \n\t\t\t\t\t\tPrint["\<[\>", ii, "\<,\>", 
                      jj, "\<] = \>", 
                      ten[\([ii, jj]\)]], \n\t\t\t\t\t\tPrint["\<[\>", 
                      ii, "\<,\>", jj, "\<] = \>", 
                      ten[\([ii, jj]\)]]], \n\t\t\t\t\t{jj, 
                    Length[ten[\([ii]\)]]}], {ii, 
                  Length[ten]}], ]; \n\t\t\tIf[
              rank == 3, \n\t\t\t\tDo[
                Do[Do[If[ten[\([ii, jj, kk]\)] == 0, 
                      dummy = 0, \n\t\t\t\t\t\tPrint["\<[\>", ii, "\<,\>", 
                        jj, "\<,\>", kk, "\<] = \>", 
                        ten[\([ii, jj, kk]\)]], \n\t\t\t\t\t\tPrint["\<[\>", 
                        ii, "\<,\>", jj, "\<,\>", kk, "\<] = \>", 
                        ten[\([ii, jj, kk]\)]]], \n\t\t\t\t\t{kk, 
                      Length[ten[\([ii, jj]\)]]}], {jj, 
                    Length[ten[\([ii]\)]]}], {ii, 
                  Length[ten]}], ]; \n\t\t\tIf[
              rank == 4, \n\t\t\t\tDo[
                Do[Do[Do[
                      If[ten[\([ii, jj, kk, ll]\)] == 0, 
                        dummy = 0, \n\t\t\t\t\t\tPrint["\<[\>", ii, "\<,\>", 
                          jj, "\<,\>", kk, "\<,\>", ll, "\<] = \>", 
                          ten[\([ii, jj, kk, 
                              ll]\)]], \n\t\t\t\t\t\tPrint["\<[\>", 
                          ii, "\<,\>", jj, "\<,\>", kk, "\<,\>", 
                          ll, "\<] = \>", 
                          ten[\([ii, jj, kk, ll]\)]]], \n\t\t\t\t\t{ll, 
                        Length[ten[\([ii, jj, kk]\)]]}], {kk, 
                      Length[ten[\([ii, jj]\)]]}], \n\t\t\t\t\t\t{jj, 
                    Length[ten[\([ii]\)]]}], {ii, 
                  Length[ten]}], ]; \n\t\t\tIf[
              rank == 5, \n\t\t\t\tDo[
                Do[Do[Do[
                      Do[If[ten[\([ii, jj, kk, ll, mm]\)] == 0, 
                          dummy = 0, \n\t\t\t\t\t\tPrint["\<[\>", 
                            ii, "\<,\>", jj, "\<,\>", kk, "\<,\>", 
                            ll, "\<,\>", 
                            mm, "\<] = \>", \n\t\t\t\t\t\t\t\t\t\t\t\
ten[\([ii, jj, kk, ll, mm]\)]], \n\t\t\t\t\t\tPrint["\<[\>", ii, "\<,\>", 
                            jj, "\<,\>", kk, "\<,\>", ll, "\<,\>", 
                            mm, "\<] = \>", \n\t\t\t\t\t\t\t\t\t\t\t\
ten[\([ii, jj, kk, ll, mm]\)]]], {mm, 
                          Length[
                            ten[\([ii, jj, kk, ll]\)]]}], \n\t\t\t\t\t{ll, 
                        Length[ten[\([ii, jj, kk]\)]]}], {kk, 
                      Length[ten[\([ii, jj]\)]]}], \n\t\t\t\t\t\t{jj, 
                    Length[ten[\([ii]\)]]}], {ii, 
                  Length[ten]}], ];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(tenrank[ten_List]\  := \ 
        Block[{dim, 
            rank}, \n\t\t\((dim\  = \ 
              Length[currentst[\([1]\)]]; \n\t\t\trank\  = \ 0; \n\t\t\tIf[
              Length[ten] == 
                dim, \((rank\  = \ 
                  1; \n\t\t\tIf[\((Head[ten[\([1]\)]] === 
                        List)\) && \((Length[ten[\([1]\)]] == 
                        dim)\), \((rank\  = \ 
                      2; \n\t\t\tIf[\((Head[ten[\([1, 1]\)]] === 
                            List)\) && \((Length[ten[\([1, 1]\)]] == 
                            dim)\), \((rank\  = \ 
                          3; \n\t\t\tIf[\((Head[ten[\([1, 1, 1]\)]] === 
                                List)\) && \((Length[ten[\([1, 1, 1]\)]] == 
                                dim)\), \((rank = 
                              4; \n\t\t\tIf[\((Head[ten[\([1, 1, 1, 1]\)]] === 
                                    List)\) && \((Length[
                                      ten[\([1, 1, 1, 1]\)]] == dim)\), 
                              rank = 
                                5, ];)\), ];)\), ];)\), ];)\), ]; \n\t\t\t\
Return[rank];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(raise[ten_List, nn_]\  := \ 
        Block[{dim, ginv, rank, 
            tenr}, \n\t\t\((If[
              Length[currentst] < 
                3, \n\t\t\t\t\((Print["\<Evaluate spacetime[x,g] first\>"]; \n\
\t\t\t\t\tReturn[];)\), ]; \n\t\t\trank\  = \ tenrank[ten]; \n\t\t\tIf[
              rank < nn, \n\t\t\t\t\((Print["\<Error - insufficient rank\>"]; \
\n\t\t\t\t\tReturn[];)\), ]; \n\t\t\tginv\  = \ 
              currentst[\([3]\)]; \n\t\t\tdim\  = \ 
              Length[ginv[\([1]\)]]; \n\t\t\tIf[
              Length[ten] == 
                dim, , \n\t\t\t\t\((Print["\<Error - tensor is incorrect \
dimension\>"]; \n\t\t\t\tReturn[];)\)]; \n\t\t\tIf[
              rank == 1, \n\t\t\t\ttenr\  = \ 
                Table[simpler[
                    Sum[ginv[\([ii, jj]\)] 
                        ten[\([jj]\)], \n\t\t\t\t\t\t\t\t{jj, dim}]], {ii, 
                    dim}], ]; \n\t\t\tIf[
              rank == 2, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      simpler[
                        Sum[ginv[\([ii, kk]\)] 
                            ten[\([kk, jj]\)], \n\t\t\t\t\t\t\t\t{kk, 
                            dim}]], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      simpler[
                        Sum[ginv[\([jj, kk]\)] 
                            ten[\([ii, kk]\)], \n\t\t\t\t\t\t\t\t{kk, 
                            dim}]], {jj, dim}], {ii, dim}]], ]; \n\t\t\tIf[
              rank == 3, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      Table[simpler[
                          Sum[ginv[\([ii, ll]\)] 
                              ten[\([ll, jj, kk]\)], \n\t\t\t\t\t\t\t\t{ll, 
                              dim}]], {kk, dim}], {jj, dim}], {ii, 
                      dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      Table[simpler[
                          Sum[ginv[\([jj, ll]\)] 
                              ten[\([ii, ll, kk]\)], \n\t\t\t\t\t\t\t\t{ll, 
                              dim}]], {kk, dim}], {jj, dim}], {ii, dim}], 3, 
                  Table[Table[
                      Table[simpler[
                          Sum[ginv[\([kk, ll]\)] 
                              ten[\([ii, jj, ll]\)], \n\t\t\t\t\t\t\t\t{ll, 
                              dim}]], {kk, dim}], {jj, dim}], {ii, 
                      dim}]], ]; \n\t\t\tIf[
              rank == 4, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[ginv[\([ii, mm]\)] 
                                ten[\([mm, jj, kk, 
                                    ll]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[ginv[\([jj, mm]\)] 
                                ten[\([ii, mm, kk, 
                                    ll]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}], \n\t\t\t\t\t\t3, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[ginv[\([kk, mm]\)] 
                                ten[\([ii, jj, mm, 
                                    ll]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}], \n\t\t\t\t\t\t4, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[ginv[\([ll, mm]\)] 
                                ten[\([ii, jj, kk, 
                                    mm]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}]], ]; \n\t\t\tIf[
              rank == 5, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                ginv[\([ii, qq]\)] 
                                  ten[\([qq, jj, kk, ll, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                ginv[\([jj, qq]\)] 
                                  ten[\([ii, qq, kk, ll, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t3, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                ginv[\([kk, qq]\)] 
                                  ten[\([ii, jj, qq, ll, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t4, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                ginv[\([ll, qq]\)] 
                                  ten[\([ii, jj, kk, qq, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t5, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                ginv[\([mm, qq]\)] 
                                  ten[\([ii, jj, kk, ll, 
                                      qq]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}]], ]; \n\t\t\tReturn[
              tenr];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(lower[ten_List, nn_]\  := \ 
        Block[{dim, gg, rank, 
            tenr}, \n\t\t\((If[
              Length[currentst] < 
                3, \n\t\t\t\t\((Print["\<Evaluate spacetime[x,g] first\>"]; \n\
\t\t\t\t\tReturn[];)\), ]; \n\t\t\trank\  = \ tenrank[ten]; \n\t\t\tIf[
              rank < nn, \n\t\t\t\t\((Print["\<Error - insufficient rank\>"]; \
\n\t\t\t\t\tReturn[];)\), ]; \n\t\t\tgg\  = \ 
              currentst[\([2]\)]; \n\t\t\tdim\  = \ 
              Length[gg[\([1]\)]]; \n\t\t\tIf[
              Length[ten] == 
                dim, , \n\t\t\t\t\((Print["\<Error - tensor is incorrect \
dimension\>"]; \n\t\t\t\tReturn[];)\)]; \n\t\t\tIf[
              rank == 1, \n\t\t\t\ttenr\  = \ 
                Table[simpler[
                    Sum[gg[\([ii, jj]\)] 
                        ten[\([jj]\)], \n\t\t\t\t\t\t\t\t{jj, dim}]], {ii, 
                    dim}], ]; \n\t\t\tIf[
              rank == 2, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      simpler[
                        Sum[gg[\([ii, kk]\)] 
                            ten[\([kk, jj]\)], \n\t\t\t\t\t\t\t\t{kk, 
                            dim}]], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      simpler[
                        Sum[gg[\([jj, kk]\)] 
                            ten[\([ii, kk]\)], \n\t\t\t\t\t\t\t\t{kk, 
                            dim}]], {jj, dim}], {ii, dim}]], ]; \n\t\t\tIf[
              rank == 3, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      Table[simpler[
                          Sum[gg[\([ii, ll]\)] 
                              ten[\([ll, jj, kk]\)], \n\t\t\t\t\t\t\t\t{ll, 
                              dim}]], {kk, dim}], {jj, dim}], {ii, 
                      dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      Table[simpler[
                          Sum[gg[\([jj, ll]\)] 
                              ten[\([ii, ll, kk]\)], \n\t\t\t\t\t\t\t\t{ll, 
                              dim}]], {kk, dim}], {jj, dim}], {ii, dim}], 3, 
                  Table[Table[
                      Table[simpler[
                          Sum[gg[\([kk, ll]\)] 
                              ten[\([ii, jj, ll]\)], \n\t\t\t\t\t\t\t\t{ll, 
                              dim}]], {kk, dim}], {jj, dim}], {ii, 
                      dim}]], ]; \n\t\t\tIf[
              rank == 4, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[gg[\([ii, mm]\)] 
                                ten[\([mm, jj, kk, 
                                    ll]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[gg[\([jj, mm]\)] 
                                ten[\([ii, mm, kk, 
                                    ll]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}], \n\t\t\t\t\t\t3, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[gg[\([kk, mm]\)] 
                                ten[\([ii, jj, mm, 
                                    ll]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}], \n\t\t\t\t\t\t4, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[gg[\([ll, mm]\)] 
                                ten[\([ii, jj, kk, 
                                    mm]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}]], ]; \n\t\t\tIf[
              rank == 5, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                gg[\([ii, qq]\)] 
                                  ten[\([qq, jj, kk, ll, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                gg[\([jj, qq]\)] 
                                  ten[\([ii, qq, kk, ll, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t3, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                gg[\([kk, qq]\)] 
                                  ten[\([ii, jj, qq, ll, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t4, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                gg[\([ll, qq]\)] 
                                  ten[\([ii, jj, kk, qq, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t5, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                gg[\([mm, qq]\)] 
                                  ten[\([ii, jj, kk, ll, 
                                      qq]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}]], ]; \n\t\t\tReturn[
              tenr];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(norm[ve_List]\  := \ 
        Block[{dim, gg, 
            nn}, \n\t\t\((If[
              Length[currentst] < 
                3, \n\t\t\t\t\((Print["\<Evaluate spacetime[x,g] first\>"]; \n\
\t\t\t\t\tReturn[];)\), ]; \n\t\t\tgg\  = \ 
              currentst[\([2]\)]; \n\t\t\tdim\  = \ 
              Length[gg[\([1]\)]]; \n\t\t\tIf[
              Length[ve] == 
                dim, , \n\t\t\t\t\((Print["\<Error - vector is incorrect \
dimension\>"]; \n\t\t\t\tReturn[];)\)]; \n\t\t\tnn\  = \ 
              simpler[Sum[
                  Sum[gg[\([aa, bb]\)] ve[\([aa]\)] 
                      ve[\([bb]\)], \n\t\t\t\t\t\t\t{bb, dim}], {aa, 
                    dim}]]; \n\t\t\tReturn[nn];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(twist[ve_List]\  := \ 
        Block[{dim, xx, Christoffel, 
            tw}, \n\t\t\((If[
              Length[currentst] < 
                4, \n\t\t\t\t\((Print["\<Evaluate spacetime[x,g] first\>"]; \n\
\t\t\t\t\tReturn[];)\), ]; \n\t\t\txx\  = \ 
              currentst[\([1]\)]; \n\t\t\tdim\  = \ Length[xx]; \n\t\t\tIf[
              Length[ve] == 
                dim, , \n\t\t\t\t\((Print["\<Error - vector is incorrect \
dimension\>"]; \n\t\t\t\tReturn[];)\)]; \n\t\t\tChristoffel\  = \ 
              currentst[\([4]\)]; \n\t\t\ttw\  = \ 
              Table[simpler[
                  Sum[Sum[
                      Sum[\n\t\t\t\t\t\tSignature[{ii, jj, kk, ll}] 
                          ve[\([jj]\)]*\((D[ve[\([ll]\)], 
                                xx[\([kk]\)]] - \n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                Christoffel[\([mm, kk, ll]\)] 
                                  ve[\([mm]\)], {mm, 
                                  dim}])\), \n\t\t\t\t\t\t\t\t\t{ll, 
                          dim}], {kk, dim}], {jj, dim}]], {ii, 
                  dim}]; \n\t\t\tReturn[tw];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(coordxform[ten_List, xx_List, xp_List]\  := \ 
        Block[{dim, rank, tenp, rep, xst, 
            xform}, \n\t\t\((If[
              Length[currentst] < 
                3, \n\t\t\t\t\((Print["\<Evaluate spacetime[x,g] first\>"]; \n\
\t\t\t\t\tReturn[];)\), ]; \n\t\t\txst\  = \ 
              currentst[\([1]\)]; \n\t\t\trank\  = \ 
              tenrank[ten]; \n\t\t\tdim\  = \ Length[xst]; \n\t\t\tIf[
              Length[ten] == 
                dim, , \n\t\t\t\t\((Print["\<Error - tensor is incorrect \
dimension\>"]; \n\t\t\t\t\tReturn[];)\)]; \n\t\t\tIf[
              Length[xx] == 
                dim, , \n\t\t\t\t\((Print["\<Error - x[xp] list is incorrect \
dimension\>"]; \n\t\t\t\t\tReturn[];)\)]; \n\t\t\tIf[
              Length[xp] == 
                dim, , \n\t\t\t\t\((Print["\<Error - xp list is incorrect \
dimension\>"]; \n\t\t\t\t\tReturn[];)\)]; \n\t\t\trep\  = \ 
              ten /. Thread[xst -> xx]; \n\t\t\txform\  = \ 
              Table[Table[
                  simpler[D[xx[\([jj]\)], xp[\([ii]\)]]], {jj, dim}], {ii, 
                  dim}]; \n\t\t\tIf[
              rank == 1, \n\t\t\t\ttenp = 
                Table[simpler[
                    Sum[xform[\([ii, jj]\)] 
                        rep[\([jj]\)], \n\t\t\t\t\t\t\t\t{jj, dim}]], {ii, 
                    dim}], ]; \n\t\t\tIf[
              rank == 2, \n\t\t\t\ttenp = 
                Table[Table[
                    simpler[
                      Sum[Sum[\n\t\t\t\t\t\t\t\t\t\txform[\([ii, kk]\)] 
                            xform[\([jj, ll]\)] 
                            rep[\([kk, ll]\)], \n\t\t\t\t\t\t\t\t\t\t{ll, 
                            dim}], {kk, dim}]], {jj, dim}], {ii, 
                    dim}], ]; \n\t\t\tIf[
              rank == 3, \n\t\t\t\ttenp = 
                Table[Table[
                    Table[simpler[
                        Sum[Sum[
                            Sum[\n\t\t\t\t\t\t\t\t\t\t\t\txform[\([ii, ll]\)] 
                                xform[\([jj, mm]\)] xform[\([kk, nn]\)] 
                                rep[\([ll, mm, 
                                    nn]\)], \n\t\t\t\t\t\t\t\t\t\t\t\t{nn, 
                                dim}], {mm, dim}], {ll, dim}]], {kk, 
                        dim}], {jj, dim}], {ii, dim}], ]; \n\t\t\tIf[
              rank == 4, \n\t\t\t\ttenp = 
                Table[Table[
                    Table[Table[
                        simpler[
                          Sum[Sum[
                              Sum[Sum[\n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\
xform[\([ii, mm]\)] xform[\([jj, nn]\)] xform[\([kk, pp]\)] 
                                    xform[\([ll, qq]\)] 
                                    rep[\([mm, nn, pp, 
                                        qq]\)], \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\
{qq, dim}], {pp, dim}], {nn, dim}], {mm, dim}]], \n\t\t\t\t\t\t\t\t\t{ll, 
                          dim}], {kk, dim}], {jj, dim}], {ii, 
                    dim}], ]; \n\t\t\tIf[
              rank == 5, \n\t\t\t\ttenp = 
                Table[Table[
                    Table[Table[
                        Table[simpler[
                            Sum[Sum[
                                Sum[Sum[
                                    Sum[\n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\
xform[\([ii, nn]\)] xform[\([jj, pp]\)] 
                                        xform[\([kk, 
                                        qq]\)]*\n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\
\t\t\txform[\([ll, vv]\)] xform[\([mm, ww]\)] 
                                        rep[\([nn, pp, qq, vv, 
                                        ww]\)], \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\
\t\t{ww, dim}], {vv, dim}], {qq, dim}], {pp, dim}], {nn, 
                                dim}]], \n\t\t\t\t\t\t\t\t\t\t{mm, 
                            dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                    dim}], ]; \n\t\t\tReturn[tenp];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(coordxform[ten_List, xx_List, xf_List, xp_List]\  := \ 
        Block[{dim, rank, tenp, rep, 
            xform}, \n\t\t\((rank\  = \ TensorRank[ten]; \n\t\t\tdim\  = \ 
              Length[xx]; \n\t\t\tIf[
              Length[ten] == 
                dim, , \n\t\t\t\t\((Print["\<Error - tensor is incorrect \
dimension\>"]; \n\t\t\t\t\tReturn[];)\)]; \n\t\t\tIf[
              Length[xf] == 
                dim, , \n\t\t\t\t\((Print["\<Error - function xx(xp) list is \
incorrect dimension\>"]; \n\t\t\t\t\tReturn[];)\)]; \n\t\t\tIf[
              Length[xp] == 
                dim, , \n\t\t\t\t\((Print["\<Error - xprime list is incorrect \
dimension\>"]; \n\t\t\t\t\tReturn[];)\)]; \n\t\t\trep\  = \ 
              ten /. Thread[xx -> xf]; \n\t\t\txform\  = \ 
              Table[Table[
                  simpler[D[xf[\([jj]\)], xp[\([ii]\)]]], {jj, dim}], {ii, 
                  dim}]; \n\t\t\tIf[
              rank == 1, \n\t\t\t\ttenp = 
                Table[simpler[
                    Sum[xform[\([ii, jj]\)] 
                        rep[\([jj]\)], \n\t\t\t\t\t\t\t\t{jj, dim}]], {ii, 
                    dim}], ]; \n\t\t\tIf[
              rank == 2, \n\t\t\t\ttenp = 
                Table[Table[
                    simpler[
                      Sum[Sum[\n\t\t\t\t\t\t\t\t\t\txform[\([ii, kk]\)] 
                            xform[\([jj, ll]\)] 
                            rep[\([kk, ll]\)], \n\t\t\t\t\t\t\t\t\t\t{ll, 
                            dim}], {kk, dim}]], {jj, dim}], {ii, 
                    dim}], ]; \n\t\t\tIf[
              rank == 3, \n\t\t\t\ttenp = 
                Table[Table[
                    Table[simpler[
                        Sum[Sum[
                            Sum[\n\t\t\t\t\t\t\t\t\t\t\t\txform[\([ii, ll]\)] 
                                xform[\([jj, mm]\)] xform[\([kk, nn]\)] 
                                rep[\([ll, mm, 
                                    nn]\)], \n\t\t\t\t\t\t\t\t\t\t\t\t{nn, 
                                dim}], {mm, dim}], {ll, dim}]], {kk, 
                        dim}], {jj, dim}], {ii, dim}], ]; \n\t\t\tIf[
              rank == 4, \n\t\t\t\ttenp = 
                Table[Table[
                    Table[Table[
                        simpler[
                          Sum[Sum[
                              Sum[Sum[\n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\
xform[\([ii, mm]\)] xform[\([jj, nn]\)] xform[\([kk, pp]\)] 
                                    xform[\([ll, qq]\)] 
                                    rep[\([mm, nn, pp, 
                                        qq]\)], \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\
{qq, dim}], {pp, dim}], {nn, dim}], {mm, dim}]], \n\t\t\t\t\t\t\t\t\t{ll, 
                          dim}], {kk, dim}], {jj, dim}], {ii, 
                    dim}], ]; \n\t\t\tIf[
              rank == 5, \n\t\t\t\ttenp = 
                Table[Table[
                    Table[Table[
                        Table[simpler[
                            Sum[Sum[
                                Sum[Sum[
                                    Sum[\n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\
xform[\([ii, nn]\)] xform[\([jj, pp]\)] xform[\([kk, 
                                        qq]\)]*\n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\
\t\t\txform[\([ll, vv]\)] xform[\([mm, ww]\)] 
                                        rep[\([nn, pp, qq, vv, 
                                        ww]\)], \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\
\t\t{ww, dim}], {vv, dim}], {qq, dim}], {pp, dim}], {nn, 
                                dim}]], \n\t\t\t\t\t\t\t\t\t\t{mm, 
                            dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                    dim}], ]; \n\t\t\tReturn[tenp];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(sphericalmetric[nn_]\  := \ 
        Block[{zz, zval, gs, xx, \n\t\t\tz1, z2, z3, z4, z5, z6, z7, z8, z9, 
            z10, z11}, \n\t\t\((zz\  = \ 
              Table[Symbol["\<z\>" <> ToString[ii]], {ii, 
                  nn}]; \n\t\t\txx\  = \ 
              Flatten[{r, 
                  ph, \[IndentingNewLine]Table[
                    Symbol["\<t\>" <> ToString[ii]], {ii, 
                      nn - 2}]}]; \n\t\t\tzzval = 
              Flatten[{r*Product[Sin[xx[\([ii]\)]], {ii, 2, nn}], 
                  Table[r*Cos[xx[\([ii]\)]]*
                      Product[Sin[xx[\([jj]\)]], {jj, ii + 1, nn}], {ii, 2, 
                      nn}]}]; \n\t\t\tThread[
              Evaluate[zz] = 
                zzval]; \n\t\t\tPrint["\<Spherical metric check: \>", 
              Simplify[
                  Sum[zz[\([ii]\)]^2, {ii, 
                      nn}]] == \((r^2)\)]; \n\t\t\tgs\  = \ 
              Table[Table[
                  Simplify[
                    Sum[\n\t\t\t\t\t\t\t\tD[zz[\([kk]\)], xx[\([ii]\)]] 
                        D[zz[\([kk]\)], xx[\([jj]\)]], \n\t\t\t\t\t\t\t\t{kk, 
                        nn}]], {jj, nn}], {ii, nn}]; \n\t\t\tReturn[
              gs];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(connection[]\  := \ 
        Block[{dim, gg, ginv, xx, 
            Christoffel}, \n\t\t\((If[
              Length[currentst] < 
                3, \n\t\t\t\t\((Print["\<Evaluate spacetime[x,g] first\>"]; \n\
\t\t\t\t\tReturn[];)\), ]; \n\t\t\txx\  = \ 
              currentst[\([1]\)]; \n\t\t\tgg\  = \ 
              currentst[\([2]\)]; \n\t\t\tginv\  = \ 
              currentst[\([3]\)]; \n\t\t\tdim\  = \ 
              Length[xx]; \n\t\t\tChristoffel\  = \ 
              Table[Table[Table[0, {kk, dim}], {jj, dim}], {ii, 
                  dim}]; \n\t\t\tDo[
              Do[Do[Christoffel[\([ii, jj, kk]\)] = \n\t\t\t\t\t\t\tsimpler[
                      Sum[ginv[\([ii, 
                                ll]\)] \((D[gg[\([kk, ll]\)], 
                                  xx[\([jj]\)]]\  + \n\t\t\t\t\t\t\t\t\t\t\t\t\
D[gg[\([jj, ll]\)], xx[\([kk]\)]]\  - \ 
                                D[gg[\([jj, kk]\)], xx[\([ll]\)]])\), {ll, 
                            dim}]/2], \n\t\t\t\t\t\t{kk, jj, dim}], {jj, 
                  dim}], {ii, dim}]; \n\t\t\tDo[
              Do[Do[Christoffel[\([ii, kk, jj]\)] = 
                    Christoffel[\([ii, jj, kk]\)], \n\t\t\t\t\t\t{kk, jj + 1, 
                    dim}], {jj, dim - 1}], {ii, dim}]; \n\t\t\tIf[
              Length[currentst] == 3, 
              AppendTo[currentst, 
                Christoffel], \n\t\t\t\tcurrentst[\([4]\)]\  = \ 
                Christoffel]; \n\t\t\tReturn[Christoffel];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(covder[ten_List]\  := \ 
        Block[{dim, xx, chris, der, 
            rank}, \n\t\t\((If[
              Length[currentst] < 
                4, \n\t\t\t\t\((Print["\<Evaluate spacetime[x,g] first\>"]; \n\
\t\t\t\t\tReturn[];)\), ]; \n\t\t\txx\  = \ 
              currentst[\([1]\)]; \n\t\t\tchris\  = \ 
              currentst[\([4]\)]; \n\t\t\tdim\  = \ 
              Length[xx]; \n\t\t\trank\  = \ tenrank[ten]; \n\t\t\tder\  = \ 
              Switch[rank, \n\t\t\t\t1, 
                Table[Table[
                    simpler[
                      D[ten[\([ii]\)], xx[\([jj]\)]] - \n\t\t\t\t\t\t\tSum[
                          chris[\([nn, ii, jj]\)] ten[\([nn]\)], {nn, 
                            dim}]], {jj, dim}], {ii, dim}], \n\t\t\t\t2, 
                Table[Table[
                    Table[simpler[\n\t\t\t\t\t\t\t\t\tD[ten[\([ii, jj]\)], 
                            xx[\([kk]\)]] - \n\t\t\t\t\t\t\tSum[
                            chris[\([nn, ii, kk]\)] 
                                ten[\([nn, 
                                    jj]\)] + \n\t\t\t\t\t\t\t\t\t\t\t\tchris[\
\([nn, jj, kk]\)] ten[\([ii, nn]\)], {nn, dim}]], \n\t\t\t\t\t\t\t\t{kk, 
                        dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t3, 
                Table[Table[
                    Table[Table[
                        simpler[\n\t\t\t\t\t\t\t\t\tD[ten[\([ii, jj, kk]\)], 
                              xx[\([ll]\)]] - \n\t\t\t\t\t\t\tSum[
                              chris[\([nn, ii, ll]\)] 
                                  ten[\([nn, jj, 
                                      kk]\)] + \n\t\t\t\t\t\t\t\t\t\t\t\t\
chris[\([nn, jj, ll]\)] ten[\([ii, nn, 
                                      kk]\)] + \n\t\t\t\t\t\t\t\t\t\t\t\t\
chris[\([nn, kk, ll]\)] ten[\([ii, jj, nn]\)], {nn, 
                                dim}]], \n\t\t\t\t\t\t\t\t\t{ll, dim}], {kk, 
                        dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t4, 
                Table[Table[
                    Table[Table[
                        Table[simpler[\n\t\t\t\t\t\t\t\t\tD[
                                ten[\([ii, jj, kk, ll]\)], 
                                xx[\([mm]\)]] - \n\t\t\t\t\t\t\tSum[
                                chris[\([nn, ii, mm]\)] 
                                    ten[\([nn, jj, kk, 
                                        ll]\)] + \n\t\t\t\t\t\t\t\t\t\t\t\t\
chris[\([nn, jj, mm]\)] ten[\([ii, nn, kk, 
                                        ll]\)] + \n\t\t\t\t\t\t\t\t\t\t\t\t\
chris[\([nn, kk, mm]\)] ten[\([ii, jj, nn, 
                                        ll]\)] + \n\t\t\t\t\t\t\t\t\t\t\t\t\t\
chris[\([nn, ll, mm]\)] ten[\([ii, jj, kk, nn]\)], {nn, 
                                  dim}]], \n\t\t\t\t\t\t\t\t\t{mm, 
                            dim}], {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                    dim}]]; \n\t\t\tReturn[der];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(curvature[]\  := \ 
        Block[{Rcurv, dim, xx, gg, 
            Christoffel}, \n\t\t\((If[
              Length[currentst] < 
                4, \n\t\t\t\t\((Print["\<Evaluate spacetime[x,g] first\>"]; \n\
\t\t\t\t\tReturn[];)\), ]; \n\t\t\txx\  = \ 
              currentst[\([1]\)]; \n\t\t\tgg\  = \ 
              currentst[\([2]\)]; \n\t\t\tChristoffel\  = \ 
              currentst[\([4]\)]; \n\t\t\tdim\  = \ 
              Length[xx]; \n\t\t\tRcurv\  = \ 
              Table[Table[
                  Table[Table[0, {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                  dim}]; \n\t\t\tDo[
              Do[Do[Do[
                    Rcurv[\([ii, jj, kk, mm]\)] = \n\t\t\t\t\t\t\t\tsimpler[
                        Sum[\((D[Christoffel[\([ll, ii, kk]\)], 
                                  xx[\([jj]\)]]\  - \n\t\t\t\t\t\t\t\t\t\t\t\t\
D[Christoffel[\([ll, jj, kk]\)], 
                                  xx[\([ii]\)]]\  + \n\t\t\t\t\t\t\t\t\t\t\t\t\
Sum[Christoffel[\([nn, ii, kk]\)] Christoffel[\([ll, nn, 
                                        jj]\)]\  - \n\t\t\t\t\t\t\t\t\t\t\t\t\
\t\tChristoffel[\([nn, jj, kk]\)] Christoffel[\([ll, nn, 
                                        ii]\)], \n\t\t\t\t\t\t\t\t\t\t\t\t\t\
{nn, dim}])\) gg[\([ll, mm]\)], {ll, dim}]], \n\t\t\t\t\t\t\t{mm, kk + 1, 
                      dim}], {kk, dim - 1}], {jj, ii + 1, dim}], {ii, 
                dim - 1}]; \n\t\t\tDo[
              Do[Do[Do[
                    Rcurv[\([jj, ii, kk, 
                          ll]\)] = \(-Rcurv[\([ii, jj, kk, 
                            ll]\)]\), \n\t\t\t\t\t\t\t{ll, kk + 1, 
                      dim}], {kk, dim - 1}], {jj, ii + 1, dim}], {ii, 
                dim - 1}]; \n\t\t\tDo[
              Do[Do[Do[
                    Rcurv[\([ii, jj, ll, 
                          kk]\)] = \(-Rcurv[\([ii, jj, kk, 
                            ll]\)]\), \n\t\t\t\t\t\t\t{ll, kk + 1, 
                      dim}], {kk, dim - 1}], {jj, ii + 1, dim}], {ii, 
                dim - 1}]; \n\t\t\tDo[
              Do[Do[Do[
                    Rcurv[\([jj, ii, ll, kk]\)] = 
                      Rcurv[\([ii, jj, kk, ll]\)], \n\t\t\t\t\t\t\t{ll, 
                      kk + 1, dim}], {kk, dim - 1}], {jj, ii + 1, dim}], {ii, 
                dim - 1}]; \n\t\t\tIf[Length[currentst] == 4, 
              AppendTo[currentst, Rcurv], \n\t\t\t\tcurrentst[\([5]\)]\  = \ 
                Rcurv]; \n\t\t\tReturn[Rcurv];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(ricci[]\  := \ 
        Block[{Ricci, dim, ginv, 
            Rcurv}, \n\t\t\((If[
              Length[currentst] < 
                5, \n\t\t\t\t\((Print["\<Evaluate curvature[] first\>"]; \n\t\
\t\t\t\tReturn[];)\), ]; \n\t\t\tginv\  = \ 
              currentst[\([3]\)]; \n\t\t\tRcurv\  = \ 
              currentst[\([5]\)]; \n\t\t\tdim\  = \ 
              Length[ginv[\([1]\)]]; \n\t\t\tRicci\  = \ 
              Table[Table[0, {jj, dim}], {ii, dim}]; \n\t\t\tDo[
              Do[Ricci[\([ii, jj]\)]\  = \ 
                  simpler[
                    Sum[Sum[
                        Rcurv[\([ii, mm, jj, ll]\)] 
                          ginv[\([mm, ll]\)], \n\t\t\t\t\t\t\t\t\t{ll, 
                          dim}], {mm, dim}]], {jj, ii, dim}], {ii, 
                dim}]; \n\t\t\tDo[
              Do[Ricci[\([jj, ii]\)]\  = \ Ricci[\([ii, jj]\)], {jj, ii + 1, 
                  dim}], {ii, dim - 1}]; \n\t\t\tIf[Length[currentst] == 5, 
              AppendTo[currentst, Ricci], \n\t\t\t\tcurrentst[\([6]\)]\  = \ 
                Ricci]; \n\t\t\tReturn[Ricci];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(scalarcurv[]\  := \ 
        Block[{dim, sc, ginv, 
            Ricci}, \n\t\t\((If[
              Length[currentst] < 
                6, \n\t\t\t\t\((Print["\<Evaluate curvature[] and ricci[] \
first\>"]; \n\t\t\t\t\tReturn[];)\), ]; \n\t\t\tginv\  = \ 
              currentst[\([3]\)]; \n\t\t\tRicci\  = \ 
              currentst[\([6]\)]; \n\t\t\tdim\  = \ 
              Length[ginv[\([1]\)]]; \n\t\t\tsc\  = 
              simpler[Sum[
                  Sum[\n\t\t\t\tRicci[\([ii, jj]\)] ginv[\([ii, jj]\)], {jj, 
                      dim}], {ii, dim}]]; \n\t\t\tReturn[
              sc];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(kretschmann[]\  := \ \n\t
        Block[{ginv, Rcurv, R1up, R2up, R3up, R4up, kinv, 
            dim}, \n\t\t\((If[
              Length[currentst] < 
                5, \n\t\t\t\t\((Print["\<Evaluate curvature[] first\>"]; \n\t\
\t\t\t\tReturn[];)\), ]; \n\t\t\tginv\  = \ 
              currentst[\([3]\)]; \n\t\t\tRcurv\  = \ 
              currentst[\([5]\)]; \n\t\t\tdim\  = \ 
              Length[ginv[\([1]\)]]; \n\t\t\tR1up\  = \ 
              Table[Table[
                  Table[Table[
                      simpler[
                        Sum[\n\t\t\t\t\t\t\t\t\t\tRcurv[\([ii, jj, kk, nn]\)] 
                            ginv[\([nn, ll]\)], {nn, 
                            dim}]], \n\t\t\t\t\t\t\t\t{ll, dim}], {kk, 
                      dim}], {jj, dim}], {ii, dim}]; \n\t\t\tR2up\  = \ 
              Table[Table[
                  Table[Table[
                      simpler[
                        Sum[\n\t\t\t\t\t\t\tR1up[\([ii, jj, nn, ll]\)] 
                            ginv[\([nn, kk]\)], {nn, 
                            dim}]], \n\t\t\t\t\t\t\t\t{ll, dim}], {kk, 
                      dim}], {jj, dim}], {ii, dim}]; \n\t\t\tR3up\  = \ 
              Table[Table[
                  Table[Table[
                      simpler[
                        Sum[\n\t\t\t\t\t\t\tR2up[\([ii, nn, kk, ll]\)] 
                            ginv[\([nn, jj]\)], {nn, 
                            dim}]], \n\t\t\t\t\t\t\t\t{ll, dim}], {kk, 
                      dim}], {jj, dim}], {ii, dim}]; \n\t\t\tR4up\  = \ 
              Table[Table[
                  Table[Table[
                      simpler[
                        Sum[\n\t\t\t\t\t\t\tR3up[\([nn, jj, kk, ll]\)] 
                            ginv[\([nn, ii]\)], {nn, 
                            dim}]], \n\t\t\t\t\t\t\t\t{ll, dim}], {kk, 
                      dim}], {jj, dim}], {ii, dim}]; \n\t\t\tkinv\  = \ 
              simpler[Factor[
                  simpler[
                    Sum[Sum[
                        simpler[
                          Sum[Sum[\n\t\t\t\t\t\t\t4 
                                 Rcurv[\([ii, jj, kk, ll]\)] 
                                R4up[\([ii, jj, kk, 
                                    ll]\)], \n\t\t\t\t\t\t\t\t\t\t\t\t{ll, 
                                kk + 1, dim}], {kk, 
                              dim - 1}]], \n\t\t\t\t\t\t\t\t\t{jj, ii + 1, 
                          dim}], {ii, dim - 1}]]]]; \n\t\t\tReturn[
              kinv];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(surfaceg[]\  := \ 
        Block[{xx, Christoffel, nullkv, dim, dkv, 
            sg}, \n\t\t\((If[
              Length[currentst] < 
                4, \n\t\t\t\t\((Print["\<Evaluate spacetime[x,g] first\>"]; \n\
\t\t\t\t\tReturn[];)\), ]; \n\t\t\txx\  = \ 
              currentst[\([1]\)]; \n\t\t\tChristoffel\  = \ 
              currentst[\([4]\)]; \n\t\t\tdim\  = \ 
              Length[xx]; \n\t\t\tnullkv\  = \ 
              Table[0, {ii, dim}]; \n\t\t\tnullkv[\([dim]\)]\  = \ 
              1; \n\t\t\tdkv\  = \ 
              Table[Table[
                  simpler[
                    D[nullkv[\([ii]\)], xx[\([jj]\)]] + \n\t\t\t\t\t\t\t\tSum[
                        Christoffel[\([ii, jj, kk]\)] nullkv[\([kk]\)], {kk, 
                          dim}]], \n\t\t\t\t\t\t{jj, dim}], {ii, 
                  dim}]; \n\t\t\tsg\  = \ 
              simpler[Sqrt[
                  Sum[Sum[\n\t\t\t\t\t\t\t\tdkv[\([ii, jj]\)] 
                        dkv[\([jj, ii]\)]/2, {jj, dim}], {ii, 
                      dim}]]]; \n\t\t\tReturn[sg];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(expansion[ve_List]\  := \ \n\t
        Block[{xx, gg, ginv, Christoffel, dim, bten, check, 
            expan}, \n\t\t\((If[
              Length[currentst] < 
                4, \n\t\t\t\t\((Print["\<Evaluate spacetime[x,g] first\>"]; \n\
\t\t\t\t\tReturn[];)\), ]; \n\t\t\txx\  = \ 
              currentst[\([1]\)]; \n\t\t\tgg\  = \ 
              currentst[\([2]\)]; \n\t\t\tginv\  = \ 
              currentst[\([3]\)]; \n\t\t\tChristoffel\  = \ 
              currentst[\([4]\)]; \n\t\t\tdim\  = \ Length[xx]; \n\t\t\tIf[
              Length[ve] == 
                dim, , \n\t\t\t\t\((Print["\<Error - vector is incorrect \
dimension\>"]; \n\t\t\t\t\tReturn[];)\)]; \n\t\t\tbten\  = \ 
              Table[Table[
                  simpler[
                    Sum[gg[\([ll, 
                            ii]\)] \((D[ve[\([ii]\)], 
                              xx[\([jj]\)]] + \n\t\t\t\tSum[
                              Christoffel[\([ii, jj, kk]\)] 
                                ve[\([kk]\)], {kk, dim}])\), {ii, 
                        dim}]], \n\t\t\t{jj, dim}], {ll, 
                  dim}]; \n\t\t\tcheck\  = \ 
              Table[simpler[
                  Sum[ve[\([jj]\)] bten[\([jj, ii]\)], \n\t\t\t\t\t{jj, 
                      dim}]], {ii, dim}]; \n\t\t\tDo[
              If[check[\([ii]\)] == 0, 
                dummy = 0, \n\t\t\t\t\tPrint["\<orthogonality check[\>", 
                  ii, "\<] = \>", 
                  check[\([ii]\)]], \n\t\t\t\t\tPrint["\<orthogonality check[\
\>", ii, "\<] = \>", check[\([ii]\)]]], \n\t\t\t\t{ii, 
                dim}]; \n\t\t\tcheck\  = \ 
              Table[simpler[
                  Sum[ve[\([jj]\)] bten[\([ii, jj]\)], \n\t\t\t\t\t{jj, 
                      dim}]], {ii, dim}]; \n\t\t\tDo[
              If[check[\([ii]\)] == 0, 
                dummy = 0, \n\t\t\t\t\tPrint["\<orthogonality check 2[\>", 
                  ii, "\<] = \>", 
                  check[\([ii]\)]], \n\t\t\t\t\tPrint["\<orthogonality check \
2[\>", ii, "\<] = \>", check[\([ii]\)]]], \n\t\t\t\t{ii, 
                dim}]; \n\t\texpan\  = \ 
              simpler[Sum[
                  Sum[ginv[\([ii, jj]\)] bten[\([ii, jj]\)], \n\t\t\t\t{jj, 
                      dim}], {ii, dim}]]; \n\t\t\tReturn[
              expan];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(nullexpansion[ve_List]\  := \ \n\t
        Block[{xx, gg, ginv, Christoffel, dim, bten, check, 
            expan}, \n\t\t\((If[
              Length[currentst] < 
                4, \n\t\t\t\t\((Print["\<Evaluate spacetime[x,g] first\>"]; \n\
\t\t\t\t\tReturn[];)\), ]; \n\t\t\txx\  = \ 
              currentst[\([1]\)]; \n\t\t\tgg\  = \ 
              currentst[\([2]\)]; \n\t\t\tginv\  = \ 
              currentst[\([3]\)]; \n\t\t\tChristoffel\  = \ 
              currentst[\([4]\)]; \n\t\t\tdim\  = \ Length[xx]; \n\t\t\tIf[
              Length[ve] == 
                dim, , \n\t\t\t\t\((Print["\<Error - vector is incorrect \
dimension\>"]; \n\t\t\t\t\tReturn[];)\)]; \n\t\t\tbten\  = \ 
              Table[Table[
                  simpler[\((D[ve[\([ii]\)], 
                          xx[\([jj]\)]] - \n\t\t\t\t\t\t\t\t\tSum[
                          Christoffel[\([kk, ii, jj]\)] ve[\([kk]\)], {kk, 
                            dim}])\)], \n\t\t\t\t\t\t{jj, dim}], {ii, 
                  dim}]; \n\t\t\tcheck\  = \ 
              Table[Table[
                  Table[simpler[\n\t\t\t\t\t\t\t\tve[\([ii]\)] 
                          bten[\([jj, kk]\)] - 
                        ve[\([ii]\)] 
                          bten[\([kk, 
                              jj]\)] + \n\t\t\t\t\t\t\t\t\tve[\([kk]\)] 
                          bten[\([ii, jj]\)] - 
                        ve[\([kk]\)] 
                          bten[\([jj, 
                              ii]\)] + \n\t\t\t\t\t\t\t\t\tve[\([jj]\)] 
                          bten[\([kk, ii]\)] - 
                        ve[\([jj]\)] 
                          bten[\([ii, kk]\)]], \n\t\t\t\t\t\t\t{kk, 
                      dim}], {jj, dim}], {ii, dim}]; \n\t\t\tDo[
              Do[Do[If[check[\([ii, jj, kk]\)] == 0, 
                    dummy = 
                      0, \n\t\t\t\t\t\t\t\tPrint["\<hypersurface \
orthogonality check[\>", \n\t\t\t\t\t\t\t\tii, "\<,\>", jj, "\<,\>", 
                      kk, "\<] = \>", 
                      check[\([ii, jj, 
                          kk]\)]], \n\t\t\t\t\t\t\tPrint["\<hypersurface \
orthogonality check[\>", \n\t\t\t\t\t\t\t\tii, "\<,\>", jj, "\<,\>", 
                      kk, "\<] = \>", 
                      check[\([ii, jj, kk]\)]]], \n\t\t\t\t\t\t{kk, 
                    dim}], {jj, dim}], {ii, dim}]; \n\t\t\texpan\  = \ 
              simpler[Sum[
                  Sum[ginv[\([ii, jj]\)] 
                      bten[\([ii, jj]\)], \n\t\t\t\t\t\t\t{jj, dim}], {ii, 
                    dim}]]; \n\t\t\tReturn[expan];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(inter[expr_, tl_List, nn_]\  := \ 
        If[nn == 1, \n\t\tIntegrate[
            expr, {tl[\([1]\)], 0, 2  Pi}], \n\t\tIntegrate[
            inter[expr, tl, nn - 1], {tl[\([nn]\)], 0, Pi}]];\)\)], "Input"],

Cell["\<\
surfacearea[tl_List] := Block[{dim,nn,area},
\t(If[Length[currentst]<3,
\t\t(Print[\"Evaluate spacetime[x,g] first\"];
\t\tReturn[];),];
\tgg = currentst[[2]];
\tdim = Length[gg[[1]]];
\tnn = Length[tl];
\tarea = simpler[inter[Sqrt[-Product[gg[[ii,ii]],{ii,dim}]],tl,nn]];
\tReturn[area];)];\
\>", "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(vielbein[ee_]\  := \ 
        Block[{einv, 
            xx}, \n\t\t\((If[
              Length[currentst] < 
                3, \n\t\t\t\t\((Print["\<Evaluate spacetime[x,g] first\>"]; \n\
\t\t\t\t\tReturn[];)\), ]; \n\t\t\txx\  = \ currentst[\([1]\)]; \n\t\t\tIf[
              Length[xx] == 
                Length[ee], , \n\t\t\t\t\((Print["\<Error - vielbein is \
incorrect dimension\>"]; \n\t\t\t\t\tReturn[];)\)]; \n\t\t\tcurrentframe\  = \
\ {ee}; \n\t\t\teinv\  = \ simpler[Transpose[Inverse[ee]]]; \n\t\t\tAppendTo[
              currentframe, einv]; \n\t\t\tcheckvielbein[]; \n\t\t\tReturn[
              currentframe];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(checkvielbein[]\  := \ 
        Block[{ee, einv, gg, ginv, dim, test, 
            eta}, \n\t\t\((If[
              Length[currentst] < 
                3, \n\t\t\t\t\((Print["\<Evaluate spacetime[x,g] first\>"]; \n\
\t\t\t\t\tReturn[];)\), ]; \n\t\t\tIf[
              Length[currentframe] < 
                2, \n\t\t\t\t\((Print["\<Evaluate vielbein[e] first\>"]; \n\t\
\t\t\t\tReturn[];)\), ]; \n\t\t\tee\  = \ 
              currentframe[\([1]\)]; \n\t\t\teinv\  = \ 
              currentframe[\([2]\)]; \n\t\t\tgg\  = \ 
              currentst[\([2]\)]; \n\t\t\tginv\  = \ 
              currentst[\([3]\)]; \n\t\t\tdim\  = \ 
              Length[ee[\([1]\)]]; \n\t\t\teta\  = \ 
              lorentzmetric[dim]; \n\t\t\ttest\  = \ 
              Table[Table[
                  simpler[
                    Sum[Sum[
                        ee[\([aa, ii]\)] ee[\([bb, jj]\)] 
                          eta[\([aa, bb]\)], \n\t\t\t\t\t\t\t\t\t{bb, 
                          dim}], {aa, dim}]], {jj, dim}], {ii, 
                  dim}]; \n\t\t\tPrint["\<e e eta = g: \>", 
              FullSimplify[test == gg]]; \n\t\t\ttest\  = \ 
              Table[Table[
                  simpler[
                    Sum[Sum[\n\t\t\t\t\t\t\t\t\tginv[\([ii, jj]\)] 
                          ee[\([aa, ii]\)] 
                          ee[\([bb, jj]\)], \n\t\t\t\t\t\t\t\t\t{jj, 
                          dim}], {ii, dim}]], {bb, dim}], {aa, 
                  dim}]; \n\t\t\tPrint["\<e e ginv = eta: \>", 
              FullSimplify[test == eta]]; \n\t\t\ttest\  = \ 
              Table[Table[
                  simpler[
                    simpler[
                      Sum[Sum[\n\t\t\t\t\t\t\t\t\t\teinv[\([aa, ii]\)] 
                            einv[\([bb, jj]\)] 
                            eta[\([aa, bb]\)], \n\t\t\t\t\t\t\t\t\t\t{bb, 
                            dim}], {aa, dim}]]], {jj, dim}], {ii, 
                  dim}]; \n\t\t\tPrint["\<einv einv eta = ginv: \>", 
              FullSimplify[test == ginv]]; \n\t\t\ttest\  = \ 
              Table[Table[
                  simpler[
                    Sum[Sum[\n\t\t\t\t\t\t\t\t\tgg[\([ii, jj]\)] 
                          einv[\([aa, ii]\)] 
                          einv[\([bb, jj]\)], \n\t\t\t\t\t\t\t\t\t{jj, 
                          dim}], {ii, dim}]], {bb, dim}], {aa, 
                  dim}]; \n\t\t\tPrint["\<g einv einv = eta: \>", 
              FullSimplify[test == eta]];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(lorentz[ten_List, nn_]\  := \ 
        Block[{einv, dim, rank, 
            tenr}, \n\t\t\((If[
              Length[currentframe] < 
                2, \n\t\t\t\t\((Print["\<Evaluate vielbein[e] first\>"]; \n\t\
\t\t\t\tReturn[];)\), ]; \n\t\t\trank\  = \ tenrank[ten]; \n\t\t\tIf[
              rank < nn, \n\t\t\t\t\((Print["\<Error - insufficient rank\>"]; \
\n\t\t\t\t\tReturn[];)\), ]; \n\t\t\teinv\  = \ 
              currentframe[\([2]\)]; \n\t\t\tdim\  = \ 
              Length[einv]; \n\t\t\tIf[
              Length[ten] == 
                dim, , \n\t\t\t\t\((Print["\<Error - tensor is incorrect \
dimension\>"]; \n\t\t\t\t\tReturn[];)\)]; \n\t\t\tIf[
              rank == 1, \n\t\t\t\ttenr\  = \ 
                Table[simpler[
                    Sum[einv[\([ii, jj]\)] 
                        ten[\([jj]\)], \n\t\t\t\t\t\t\t\t{jj, dim}]], {ii, 
                    dim}], ]; \n\t\t\tIf[
              rank == 2, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      simpler[
                        Sum[einv[\([ii, kk]\)] 
                            ten[\([kk, jj]\)], \n\t\t\t\t\t\t\t\t{kk, 
                            dim}]], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      simpler[
                        Sum[einv[\([jj, kk]\)] 
                            ten[\([ii, kk]\)], \n\t\t\t\t\t\t\t\t{kk, 
                            dim}]], {jj, dim}], {ii, dim}]], ]; \n\t\t\tIf[
              rank == 3, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      Table[simpler[
                          Sum[einv[\([ii, ll]\)] 
                              ten[\([ll, jj, kk]\)], \n\t\t\t\t\t\t\t\t{ll, 
                              dim}]], {kk, dim}], {jj, dim}], {ii, 
                      dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      Table[simpler[
                          Sum[einv[\([jj, ll]\)] 
                              ten[\([ii, ll, kk]\)], \n\t\t\t\t\t\t\t\t{ll, 
                              dim}]], {kk, dim}], {jj, dim}], {ii, dim}], 3, 
                  Table[Table[
                      Table[simpler[
                          Sum[einv[\([kk, ll]\)] 
                              ten[\([ii, jj, ll]\)], \n\t\t\t\t\t\t\t\t{ll, 
                              dim}]], {kk, dim}], {jj, dim}], {ii, 
                      dim}]], ]; \n\t\t\tIf[
              rank == 4, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[einv[\([ii, mm]\)] 
                                ten[\([mm, jj, kk, 
                                    ll]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[einv[\([jj, mm]\)] 
                                ten[\([ii, mm, kk, 
                                    ll]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}], \n\t\t\t\t\t\t3, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[einv[\([kk, mm]\)] 
                                ten[\([ii, jj, mm, 
                                    ll]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}], \n\t\t\t\t\t\t4, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[einv[\([ll, mm]\)] 
                                ten[\([ii, jj, kk, 
                                    mm]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}]], ]; \n\t\t\tIf[
              rank == 5, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                einv[\([ii, qq]\)] 
                                  ten[\([qq, jj, kk, ll, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                einv[\([jj, qq]\)] 
                                  ten[\([ii, qq, kk, ll, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t3, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                einv[\([kk, qq]\)] 
                                  ten[\([ii, jj, qq, ll, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t4, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                einv[\([ll, qq]\)] 
                                  ten[\([ii, jj, kk, qq, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t5, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                einv[\([mm, qq]\)] 
                                  ten[\([ii, jj, kk, ll, 
                                      qq]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}]], ]; \n\t\t\tReturn[
              tenr];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(curved[ten_List, nn_]\  := \ 
        Block[{ee, dim, rank, 
            tenr}, \n\t\t\((If[
              Length[currentframe] < 
                2, \n\t\t\t\t\((Print["\<Evaluate vielbein[e] first\>"]; \n\t\
\t\t\t\tReturn[];)\), ]; \n\t\t\trank\  = \ tenrank[ten]; \n\t\t\tIf[
              rank < nn, \n\t\t\t\t\((Print["\<Error - insufficient rank\>"]; \
\n\t\t\t\t\tReturn[];)\), ]; \n\t\t\tee\  = \ 
              currentframe[\([1]\)]; \n\t\t\tdim\  = \ Length[ee]; \n\t\t\tIf[
              Length[ten] == 
                dim, , \n\t\t\t\t\((Print["\<Error - tensor is incorrect \
dimension\>"]; \n\t\t\t\t\tReturn[];)\)]; \n\t\t\tIf[
              rank == 1, \n\t\t\t\ttenr\  = \ 
                Table[simpler[
                    Sum[ee[\([jj, ii]\)] 
                        ten[\([jj]\)], \n\t\t\t\t\t\t\t\t{jj, dim}]], {ii, 
                    dim}], ]; \n\t\t\tIf[
              rank == 2, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      simpler[
                        Sum[ee[\([kk, ii]\)] 
                            ten[\([kk, jj]\)], \n\t\t\t\t\t\t\t\t{kk, 
                            dim}]], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      simpler[
                        Sum[ee[\([kk, jj]\)] 
                            ten[\([ii, kk]\)], \n\t\t\t\t\t\t\t\t{kk, 
                            dim}]], {jj, dim}], {ii, dim}]], ]; \n\t\t\tIf[
              rank == 3, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      Table[simpler[
                          Sum[ee[\([ll, ii]\)] 
                              ten[\([ll, jj, kk]\)], \n\t\t\t\t\t\t\t\t{ll, 
                              dim}]], {kk, dim}], {jj, dim}], {ii, 
                      dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      Table[simpler[
                          Sum[ee[\([ll, jj]\)] 
                              ten[\([ii, ll, kk]\)], \n\t\t\t\t\t\t\t\t{ll, 
                              dim}]], {kk, dim}], {jj, dim}], {ii, dim}], 3, 
                  Table[Table[
                      Table[simpler[
                          Sum[ee[\([ll, kk]\)] 
                              ten[\([ii, jj, ll]\)], \n\t\t\t\t\t\t\t\t{ll, 
                              dim}]], {kk, dim}], {jj, dim}], {ii, 
                      dim}]], ]; \n\t\t\tIf[
              rank == 4, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[ee[\([mm, ii]\)] 
                                ten[\([mm, jj, kk, 
                                    ll]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[ee[\([mm, jj]\)] 
                                ten[\([ii, mm, kk, 
                                    ll]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}], \n\t\t\t\t\t\t3, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[ee[\([mm, kk]\)] 
                                ten[\([ii, jj, mm, 
                                    ll]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}], \n\t\t\t\t\t\t4, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[ee[\([mm, ll]\)] 
                                ten[\([ii, jj, kk, 
                                    mm]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}]], ]; \n\t\t\tIf[
              rank == 5, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                ee[\([qq, ii]\)] 
                                  ten[\([qq, jj, kk, ll, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                ee[\([qq, jj]\)] 
                                  ten[\([ii, qq, kk, ll, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t3, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                ee[\([qq, kk]\)] 
                                  ten[\([ii, jj, qq, ll, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t4, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                ee[\([qq, ll]\)] 
                                  ten[\([ii, jj, kk, qq, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t5, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                ee[\([qq, mm]\)] 
                                  ten[\([ii, jj, kk, ll, 
                                      qq]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}]], ]; \n\t\t\tReturn[
              tenr];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(lorentzraise[ten_List, nn_]\  := \ 
        Block[{dim, xx, ginv, rank, 
            tenr}, \n\t\t\((If[
              Length[currentst] < 
                3, \n\t\t\t\t\((Print["\<Evaluate spacetime[x,g] first\>"]; \n\
\t\t\t\t\tReturn[];)\), ]; \n\t\t\trank\  = \ tenrank[ten]; \n\t\t\tIf[
              rank < nn, \n\t\t\t\t\((Print["\<Error - insufficient rank\>"]; \
\n\t\t\t\t\tReturn[];)\), ]; \n\t\t\txx\  = \ 
              currentst[\([1]\)]; \n\t\t\tdim\  = \ Length[xx]; \n\t\t\tIf[
              Length[ten] == 
                dim, , \n\t\t\t\t\((Print["\<Error - tensor is incorrect \
dimension\>"]; \n\t\t\t\tReturn[];)\)]; \n\t\t\tginv\  = \ 
              lorentzmetric[dim]; \n\t\t\tIf[
              rank == 1, \n\t\t\t\ttenr\  = \ 
                Table[simpler[
                    Sum[ginv[\([ii, jj]\)] 
                        ten[\([jj]\)], \n\t\t\t\t\t\t\t\t{jj, dim}]], {ii, 
                    dim}], ]; \n\t\t\tIf[
              rank == 2, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      simpler[
                        Sum[ginv[\([ii, kk]\)] 
                            ten[\([kk, jj]\)], \n\t\t\t\t\t\t\t\t{kk, 
                            dim}]], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      simpler[
                        Sum[ginv[\([jj, kk]\)] 
                            ten[\([ii, kk]\)], \n\t\t\t\t\t\t\t\t{kk, 
                            dim}]], {jj, dim}], {ii, dim}]], ]; \n\t\t\tIf[
              rank == 3, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      Table[simpler[
                          Sum[ginv[\([ii, ll]\)] 
                              ten[\([ll, jj, kk]\)], \n\t\t\t\t\t\t\t\t{ll, 
                              dim}]], {kk, dim}], {jj, dim}], {ii, 
                      dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      Table[simpler[
                          Sum[ginv[\([jj, ll]\)] 
                              ten[\([ii, ll, kk]\)], \n\t\t\t\t\t\t\t\t{ll, 
                              dim}]], {kk, dim}], {jj, dim}], {ii, dim}], 3, 
                  Table[Table[
                      Table[simpler[
                          Sum[ginv[\([kk, ll]\)] 
                              ten[\([ii, jj, ll]\)], \n\t\t\t\t\t\t\t\t{ll, 
                              dim}]], {kk, dim}], {jj, dim}], {ii, 
                      dim}]], ]; \n\t\t\tIf[
              rank == 4, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[ginv[\([ii, mm]\)] 
                                ten[\([mm, jj, kk, 
                                    ll]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[ginv[\([jj, mm]\)] 
                                ten[\([ii, mm, kk, 
                                    ll]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}], \n\t\t\t\t\t\t3, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[ginv[\([kk, mm]\)] 
                                ten[\([ii, jj, mm, 
                                    ll]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}], \n\t\t\t\t\t\t4, 
                  Table[Table[
                      Table[Table[
                          simpler[
                            Sum[ginv[\([ll, mm]\)] 
                                ten[\([ii, jj, kk, 
                                    mm]\)], \n\t\t\t\t\t\t\t\t{mm, 
                                dim}]], {ll, dim}], {kk, dim}], {jj, 
                        dim}], {ii, dim}]], ]; \n\t\t\tIf[
              rank == 5, \n\t\t\t\ttenr\  = \ 
                Switch[nn, \n\t\t\t\t\t\t1, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                ginv[\([ii, qq]\)] 
                                  ten[\([qq, jj, kk, ll, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t2, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                ginv[\([jj, qq]\)] 
                                  ten[\([ii, qq, kk, ll, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t3, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                ginv[\([kk, qq]\)] 
                                  ten[\([ii, jj, qq, ll, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t4, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                ginv[\([ll, qq]\)] 
                                  ten[\([ii, jj, kk, qq, 
                                      mm]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}], \n\t\t\t\t\t\t5, 
                  Table[Table[
                      Table[Table[
                          Table[simpler[\n\t\t\t\t\t\t\t\t\t\t\t\tSum[
                                ginv[\([mm, qq]\)] 
                                  ten[\([ii, jj, kk, ll, 
                                      qq]\)], \n\t\t\t\t\t\t\t\t{qq, 
                                  dim}]], {mm, dim}], {ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}]], ]; \n\t\t\tReturn[
              tenr];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(epsilon[nn_, vind_List]\  := \ 
        Block[{eps, epsind, ind, 
            dim}, \n\t\t\((If[
              Length[currentst] < 
                3, \n\t\t\t\t\((Print["\<Evaluate spacetime[x,g] first\>"]; \n\
\t\t\t\t\tReturn[];)\), ]; \n\t\t\tdim\  = \ 
              Length[currentst[\([1]\)]]; \n\t\t\tIf[Length[vind] == 0, 
              ind\  = \ Table[ii, {ii, dim}], ind\  = \ vind]; \n\t\t\tIf[
              nn == 2, \n\t\t\t\t\((eps\  = \ 
                  Table[Table[0, {jj, dim}], {ii, 
                      dim}]; \n\t\t\t\t\tepsind\  = \ 
                  Permutations[ind]; \n\t\t\t\t\tDo[
                  eps[\([epsind[\([ii, 1]\)], 
                        epsind[\([ii, 
                            2]\)]]\)]\  = \ \n\t\t\t\t\t\t\tSignature[
                      epsind[\([ii]\)]], {ii, 
                    Length[epsind]}];)\), ]; \n\t\t\tIf[
              nn == 3, \n\t\t\t\t\((eps\  = \ 
                  Table[Table[Table[0, {kk, dim}], {jj, dim}], {ii, 
                      dim}]; \n\t\t\t\t\tepsind\  = \ 
                  Permutations[ind]; \n\t\t\t\t\tDo[
                  eps[\([epsind[\([ii, 1]\)], epsind[\([ii, 2]\)], 
                        epsind[\([ii, 
                            3]\)]]\)]\  = \ \n\t\t\t\t\t\t\tSignature[
                      epsind[\([ii]\)]], {ii, 
                    Length[epsind]}];)\), ]; \n\t\t\tIf[
              nn == 4, \n\t\t\t\t\((eps\  = \ 
                  Table[Table[
                      Table[Table[0, {ll, dim}], {kk, dim}], {jj, dim}], {ii, 
                      dim}]; \n\t\t\t\t\tepsind\  = \ 
                  Permutations[ind]; \n\t\t\t\t\tDo[
                  eps[\([epsind[\([ii, 1]\)], epsind[\([ii, 2]\)], 
                        epsind[\([ii, 3]\)], 
                        epsind[\([ii, 
                            4]\)]]\)]\  = \ \n\t\t\t\t\t\t\tSignature[
                      epsind[\([ii]\)]], {ii, 
                    Length[epsind]}];)\), ]; \n\t\t\tIf[
              nn == 5, \n\t\t\t\t\((eps\  = \ 
                  Table[Table[
                      Table[Table[
                          Table[0, \n\t\t\t\t\t\t\t\t\t\t\t{mm, dim}], {ll, 
                            dim}], {kk, dim}], {jj, dim}], {ii, 
                      dim}]; \n\t\t\t\t\tepsind\  = \ 
                  Permutations[ind]; \n\t\t\t\t\tDo[
                  eps[\([epsind[\([ii, 1]\)], epsind[\([ii, 2]\)], 
                        epsind[\([ii, 3]\)], \n\t\t\t\t\t\t\t\tepsind[\([ii, 
                            4]\)], 
                        epsind[\([ii, 
                            5]\)]]\)]\  = \ \n\t\t\t\t\t\t\tSignature[
                      epsind[\([ii]\)]], {ii, 
                    Length[epsind]}];)\), ]; \n\t\t\tReturn[
              eps];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(spinconn[]\  := \ 
        Block[{ee, einv, xx, dim, chris, eta, omega, 
            check}, \n\t\t\((If[
              Length[currentst] < 
                4, \n\t\t\t\t\((Print["\<Evaluate spacetime[x,g] first\>"]; \n\
\t\t\t\t\tReturn[];)\), ]; \n\t\t\tIf[
              Length[currentframe] < 
                2, \n\t\t\t\t\((Print["\<Evaluate vielbein[e] first\>"]; \n\t\
\t\t\t\tReturn[];)\), ]; \n\t\t\tee\  = \ 
              currentframe[\([1]\)]; \n\t\t\teinv\  = \ 
              currentframe[\([2]\)]; \n\t\t\txx\  = \ 
              currentst[\([1]\)]; \n\t\t\tdim\  = \ 
              Length[xx]; \n\t\t\tchris\  = \ 
              currentst[\([4]\)]; \n\t\t\teta\  = \ 
              lorentzmetric[dim]; \n\t\t\tomega\  = \ 
              Table[Table[Table[0, {kk, dim}], {jj, dim}], {ii, 
                  dim}]; \n\t\t\tDo[
              Do[Do[\((omega[\([ii, aa, bb]\)]\  = \ 
                      simpler[
                        Sum[Sum[\n\t\t\t\t\t\t\t\t\t\t\(-eta[\([bb, cc]\)]\) 
                              einv[\([cc, 
                                  jj]\)] \((D[ee[\([aa, jj]\)], 
                                    xx[\([ii]\)]] - \n\t\t\t\t\t\t\t\t\t\t\t\t\
\tSum[chris[\([kk, ii, jj]\)] ee[\([aa, kk]\)], {kk, 
                                      dim}])\), \n\t\t\t\t\t\t\t\t\t\t{jj, 
                              dim}], {cc, 
                            dim}]]; \n\t\t\t\t\t\t\tomega[\([ii, bb, 
                          aa]\)]\  = \ \(-omega[\([ii, aa, 
                            bb]\)]\);)\), \n\t\t\t\t\t\t{bb, aa + 1, 
                    dim}], {aa, dim - 1}], {ii, dim}]; \n\t\t\tcheck\  = \ 
              Table[Table[
                  Table[simpler[
                      D[ee[\([aa, jj]\)], xx[\([ii]\)]] + \n\t\t\t\t\t\tSum[
                          Sum[eta[\([bb, cc]\)] omega[\([ii, aa, bb]\)] 
                              ee[\([cc, jj]\)], \n\t\t\t\t\t\t\t\t{cc, 
                              dim}], {bb, dim}] - 
                        Sum[chris[\([kk, ii, jj]\)] 
                            ee[\([aa, kk]\)], \n\t\t\t\t\t\t\t{kk, 
                            dim}]], {jj, dim}], {ii, dim}], {aa, 
                  dim}]; \n\t\t\tcomponents[check]; \n\t\t\tReturn[
              omega];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(sigma\  = \ {{{0, 1}, {1, 0}}, {{0, \(-I\)}, {I, 0}}, \n\t\t{{1, 
              0}, {0, \(-1\)}}, IdentityMatrix[2]};\)\)], "Input"],

Cell[BoxData[{
    \(\(anticomm[xx_List, yy_List]\  := \ 
        Dot[xx, yy] + Dot[yy, xx];\)\), "\n", 
    \(\(recurprod[mm_List, nn_]\  := \ 
        If[nn == 1, mm[\([1]\)], 
          Dot[recurprod[mm, nn - 1], mm[\([nn]\)]]];\)\)}], "Input"],

Cell[BoxData[
    \(\(gammamatrix[t_, 
          ss_]\  := \n\t\ Block[{mm, ind, gmdim, fact, fact2, check, mtest, 
            dummy, gammam}, \n\t\t\((mm\  = \ t + ss; \n\t\t\tIf[
              Floor[mm/2] == \((mm/2)\), 
              fact = I^\((Mod[mm, 4]/
                      2)\), \n\t\t\t\t\tfact\  = \ \((\(-1\))\)^\((\((\(-3\) \
- mm)\)/4)\)]; \n\t\t\tgmdim\  = \ 2^Floor[mm/2]; \n\t\t\tind\  = \ 
              Table[Flatten[{Table[3, {jj, Floor[\((ii - 1)\)/2]}], 
                    2 - Mod[ii, 2], \n\t\t\t\t\t\t\tTable[
                      4, {jj, 
                        Floor[mm/2] - Floor[\((ii - 1)\)/2] - 
                          1}]}], \n\t\t\t\t\t{ii, 
                  2  Floor[mm/2]}]; \n\t\t\tIf[Floor[mm/2] == \((mm/2)\), 
              dummy = 0, \n\t\t\t\tAppendTo[ind, 
                Table[4, {ii, Floor[mm/2]}]]]; \n\t\t\tgammam\  = \ 
              Table[Table[
                  Table[Product[\n\t\t\t\t\tsigma[\([ind[\([kk, 
                              ll]\)], \n\t\t\t\t\t\t1 + 
                            Mod[Floor[\((ii - 1)\)/\((2^Floor[mm/2 - ll])\)], 
                              2], \n\t\t\t\t\t\t1 + 
                            Mod[Floor[\((jj - 1)\)/\((2^Floor[mm/2 - ll])\)], 
                              2]]\)], \n\t\t\t\t\t{ll, mm/2}], {jj, 
                      gmdim}], {ii, gmdim}], {kk, 
                  mm}]; \n\t\t\tIf[\((Floor[mm/2] == \((mm/2)\))\), 
              dummy = 0, 
              If[\((mm - 1)\) > 0, \n\t\t\t\t\tgammam[\([mm]\)]\  = \ 
                  fact*recurprod[gammam, mm - 1], ]]; \n\t\t\tIf[t == 1, 
              gammam[\([mm]\)]\  = \ 
                gammam[\([mm]\)]*I, ]; \n\t\t\tcheck\  = \ 
              Table[Table[
                  anticomm[gammam[\([ii]\)], gammam[\([jj]\)]], {jj, 
                    mm}], {ii, 
                  mm}]; \n\t\t\tPrint["\<Anticommutator check: \>", \n\t\t\t\t\
{Table[check[\([ii, ii]\)] == 2  IdentityMatrix[gmdim], {ii, 
                    mm - 1}], \n\t\t\t\tcheck[\([mm, 
                      mm]\)] == \((\((\(-1\))\)^t)\) 2 
                     IdentityMatrix[gmdim], \n\t\t\t\t\tTable[
                  Table[{ii, jj, 
                      check[\([ii, jj]\)] == 
                        0  IdentityMatrix[gmdim]}, \n\t\t\t\t\t\t{jj, ii + 1, 
                      mm}], {ii, 1, mm - 1}]}]; \n\t\t\tfact2\  = \ 
              fact; \n\t\t\tIf[t == 1, 
              fact2 = I^\((1 - Mod[Floor[mm/2], 2])\), ]; \n\t\t\tmtest\  = 
              fact2*recurprod[gammam, mm]; \n\t\t\tAppendTo[gammam, 
              mtest]; \n\t\t\tPrint["\<Chiral operator check: \>", \n\t\t\t\t\
\tDot[mtest, mtest] == IdentityMatrix[gmdim]]; \n\t\tReturn[
              gammam];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(checkgamma[gammam_List, t_, 
          ss_]\  := \n\t\ Block[{mm, ind, gmdim, fact, fact2, check, mtest, 
            dummy}, \n\t\t\((mm\  = \ t + ss; \n\t\t\tgmdim\  = \ 
              2^Floor[mm/2]; \n\t\t\tcheck\  = \ 
              Table[Table[
                  anticomm[gammam[\([ii]\)], gammam[\([jj]\)]], {jj, 
                    mm}], {ii, 
                  mm}]; \n\t\t\tPrint["\<Anticommutator check: \>", \n\t\t\t\t\
{Table[check[\([ii, ii]\)] == 2  IdentityMatrix[gmdim], {ii, 
                    mm - 1}], \n\t\t\t\tcheck[\([mm, 
                      mm]\)] == \((\((\(-1\))\)^t)\) 2 
                     IdentityMatrix[gmdim], \n\t\t\t\t\tTable[
                  Table[{ii, jj, 
                      check[\([ii, jj]\)] == 
                        0  IdentityMatrix[gmdim]}, \n\t\t\t\t\t\t{jj, ii + 1, 
                      mm}], {ii, 1, 
                    mm - 1}]}]; \n\t\t\tPrint["\<Chiral operator check: \>", \
\n\t\t\t\t\tDot[gammam[\([mm + 1]\)], gammam[\([mm + 1]\)]] == 
                IdentityMatrix[gmdim]];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(crossmatrix[mat1_List, mat2_List]\  := \ \n\t
        Block[{row1, col1, row2, col2, crow, ccol, 
            cross}, \n\t\t\((row1\  = \ Length[mat1]; \n\t\t\tcol1\  = \ 
              Length[mat1[\([1]\)]]; \n\t\t\trow2\  = \ 
              Length[mat2]; \n\t\t\tcol2\  = \ 
              Length[mat2[\([1]\)]]; \n\t\t\tcrow\  = \ 
              row1*row2; \n\t\t\tccol\  = \ col1*col2; \n\t\t\tcross\  = \ 
              Table[Table[\n\t\t\t\t\t\tmat1[\([1 + 
                          Mod[Floor[\((ii - 1)\)/row2], 
                            row1], \n\t\t\t\t\t\t\t1 + 
                          Mod[Floor[\((jj - 1)\)/col2], 
                            col1]]\)]*\n\t\t\t\t\t\t\tmat2[\([1 + 
                          Mod[\((ii - 1)\), row2], 
                        1 + Mod[\((jj - 1)\), col2]]\)], \n\t\t\t\t\t\t{jj, 
                    ccol}], {ii, crow}]; \n\t\t\tReturn[
              cross];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(splitgamma[gamma1_List, gamma2_List]\  := \ 
        Block[{p1, p2, gd, 
            sgamma}, \n\t\t\((p1\  = \ Length[gamma1] - 1; \n\t\t\tp2\  = \ 
              Length[gamma2] - 1; \n\t\t\tsgamma\  = \ 
              Table[0, {ii, p1 + p2}]; \n\t\t\tIf[
              p1 == \((2  Floor[p1/2])\), \n\t\t\t\t\((gd\  = \ 
                  Length[gamma2[\([1]\)]]; \n\t\t\t\t\tDo[
                  sgamma[\([ii]\)] = 
                    crossmatrix[gamma1[\([ii]\)], IdentityMatrix[gd]], {ii, 
                    p1}]; \n\t\t\t\t\tDo[
                  sgamma[\([ii + p1]\)] = 
                    crossmatrix[gamma1[\([p1 + 1]\)], gamma2[\([ii]\)]], {ii, 
                    p2}];)\), \n\t\t\t\t\((gd\  = \ 
                  Length[gamma1[\([1]\)]]; \n\t\t\t\t\tDo[
                  sgamma[\([ii]\)]\  = \ 
                    crossmatrix[gamma1[\([ii]\)], gamma2[\([p2 + 1]\)]], {ii, 
                    p1}]; \n\t\t\t\t\tDo[
                  sgamma[\([ii + p1]\)] = 
                    crossmatrix[
                      IdentityMatrix[
                        gd], \n\t\t\t\t\t\t\t\tgamma2[\([ii]\)]], {ii, 
                    p2}];)\)]; \n\t\t\tAppendTo[sgamma, 
              recurprod[sgamma, p1 + p2]]; \n\t\t\tReturn[
              sgamma];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(antigamma[gammad_, nn_]\ \  := 
        Block[{dim, agamma, 
            ind}, \n\t\t\((dim\  = \ Length[gammad] - 1; \n\t\t\tIf[
              nn == 2, \n\t\t\t\tagamma\  = \ 
                Table[Table[
                    Dot[gammad[\([jj]\)], 
                      gammad[\([kk]\)]], \n\t\t\t\t\t\t\t\t{kk, dim}], {jj, 
                    dim}], ]; \n\t\t\tIf[
              nn == 3, \n\t\t\t\t\((agamma\  = \ 
                  Table[Table[Table[0, {kk, dim}], {jj, dim}], {ii, 
                      dim}]; \n\t\t\t\t\t\t\tDo[
                  Do[Do[\((\n\t\t\t\t\t\t\t\t\t\t\tind\  = \ 
                          Permutations[{ii, kk, 
                              ll}]; \n\t\t\t\t\t\t\t\t\t\t\tagamma[\([ii, kk, 
                              ll]\)]\  = 
                          Dot[gammad[\([ii]\)], gammad[\([kk]\)], 
                            gammad[\([ll]\)]]; \n\t\t\t\t\t\t\t\t\t\t\tind\  \
= \ Rest[ind]; \n\t\t\t\t\t\t\t\t\t\t\tDo[
                          agamma[\([ind[\([mm, 1]\)], ind[\([mm, 2]\)], 
                                ind[\([mm, 
                                    3]\)]]\)]\  = \n\t\t\t\t\t\t\t\t\t\t\t\t\t\
Signature[ind[\([mm]\)]] agamma[\([ii, kk, ll]\)], {mm, 
                            Length[ind]}];)\), \n\t\t\t\t\t\t\t\t\t\t{ll, 
                        kk + 1, dim}], {kk, ii + 1, dim - 1}], {ii, 
                    dim - 2}];)\), ]; \n\t\t\tIf[
              nn == 4, \n\t\t\t\t\((agamma\  = \ 
                  Table[Table[
                      Table[Table[0, \n\t\t\t\t\t\t\t\t\t\t{ll, dim}], {kk, 
                          dim}], {jj, dim}], {ii, dim}]; \n\t\t\t\t\tDo[
                  Do[Do[Do[\((\n\t\t\t\t\t\t\t\t\t\tind\  = \ 
                            Permutations[{ii, kk, ll, 
                                mm}]; \n\t\t\t\t\t\t\t\t\t\tagamma[\([ii, kk, 
                                ll, mm]\)]\  = 
                            Dot[gammad[\([ii]\)], 
                              gammad[\([kk]\)], \n\t\t\t\t\t\t\t\t\t\t\t\t\
gammad[\([ll]\)], gammad[\([mm]\)]]; \n\t\t\t\t\t\t\t\t\t\tind\  = \ 
                            Rest[ind]; \n\t\t\t\t\t\t\t\t\t\tDo[
                            agamma[\([ind[\([pp, 1]\)], ind[\([pp, 2]\)], 
                                  ind[\([pp, 3]\)], 
                                  ind[\([pp, 
                                      4]\)]]\)]\  = \n\t\t\t\t\t\t\t\t\t\t\t\t\
Signature[ind[\([pp]\)]] agamma[\([ii, kk, ll, mm]\)], {pp, 
                              Length[ind]}];)\), \n\t\t\t\t\t\t\t\t\t{mm, 
                          ll + 1, dim}], {ll, kk + 1, 
                        dim - 1}], \n\t\t\t\t\t\t\t{kk, ii + 1, 
                      dim - 2}], {ii, dim - 3}];)\), ]; \n\t\t\tIf[
              nn == 5, \n\t\t\t\t\((agamma\  = \ 
                  Table[Table[
                      Table[Table[
                          Table[0, \n\t\t\t\t\t\t\t\t\t\t{mm, dim}], {ll, 
                            dim}], {kk, dim}], {jj, dim}], {ii, 
                      dim}]; \n\t\t\t\t\tDo[
                  Do[Do[Do[
                        Do[\((\n\t\t\t\t\t\t\t\t\t\tind\  = \ 
                              Permutations[{ii, kk, ll, mm, 
                                  qq}]; \n\t\t\t\t\t\t\t\t\t\tagamma[\([ii, 
                                  kk, ll, mm, qq]\)]\  = 
                              Dot[gammad[\([ii]\)], 
                                gammad[\([kk]\)], \n\t\t\t\t\t\t\t\t\t\t\t\t\t\
\t\tgammad[\([ll]\)], gammad[\([mm]\)], 
                                gammad[\([qq]\)]]; \n\t\t\t\t\t\t\t\t\t\tind\ \
 = \ Rest[ind]; \n\t\t\t\t\t\t\t\t\t\tDo[
                              agamma[\([ind[\([pp, 1]\)], ind[\([pp, 2]\)], 
                                    ind[\([pp, 
                                        3]\)], \n\t\t\t\t\t\t\t\t\t\t\t\t\t\t\
ind[\([pp, 4]\)], ind[\([pp, 
                                        5]\)]]\)]\  = \n\t\t\t\t\t\t\t\t\t\t\t\
\tSignature[ind[\([pp]\)]] agamma[\([ii, kk, ll, mm, qq]\)], {pp, 
                                Length[ind]}];)\), \n\t\t\t\t\t\t\t\t\t{qq, 
                            mm + 1, dim}], {mm, ll + 1, dim - 1}], {ll, 
                        kk + 1, dim - 2}], \n\t\t\t\t\t\t\t{kk, ii + 1, 
                      dim - 3}], {ii, dim - 4}];)\), ]; \n\t\t\tReturn[
              agamma];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(sugra[]\  := \ 
        Block[{}, \n\t\t\((Print["\<Available functions are\n\>", \n\t\t\t"\< \
  simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n\>", \n\t\t\t\
"\<   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\>", \
\n\t\t\t"\<   norm[vector], twist[vector], \
coordxform[tensor,<oldx,>func,newx],\n\>", \n\t\t\t"\<   sphericalmetric[d], \
connection[], covder[tensor], curvature[],\n\>", \n\t\t\t"\<   ricci[], \
scalarcurv[], kretschmann[], surfaceg[],\n\>", \n\t\t\t"\<   \
expansion[vector], nullexpansion[vector], surfacearea[x],\n\>", \n\t\t\t"\<   \
vielbein[e], checkvielbein[], lorentz[tensor,index],\n\>", \n\t\t\t"\<   \
curved[tensor,index], lorentzraise[tensor,index], epsilon[d,indices],\n\>", \n\
\t\t\t"\<   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n\>", \n\t\t\
\t"\<   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n\>", \n\
\t\t\t"\<Tensors are limited to rank 5.  All indices are lower (l) except:\n\
\>", \n\t\t\t"\<   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n\>", \n\t\t\t"\<   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n\>", \n\t\t\t"\<All indices \
are spacetime (s) except:\n\>", \n\t\t\t"\<   (inverse) vielbein[[lorentz \
(l),s]], spin connection[[s,l,l]] and\n\>", \n\t\t\t"\<   all gamma matrix \
indices.\n \n\>", \n\t\t\t"\<The last component in any index is the time \
component.\n \n\>", \n\t\t\t"\<All computations are performed with currentst \
and currentframe.\>"]; \[IndentingNewLine]Print["\<AVOID USING C, D, E, I, N \
and O as variables !\>"];)\)];\)\)], "Input"],

Cell[BoxData[
    \(\(sugra[];\)\)], "Input"]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
WindowToolbars->{},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{7, 9999},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"Magnification"->1},
PrivateNotebookOptions->{"ColorPalette"->{RGBColor, 128}},
ShowCellLabel->False,
ShowCellTags->False,
RenderingOptions->{"ObjectDithering"->True,
"RasterDithering"->False},
CharacterEncoding->"MacintoshAutomaticEncoding"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 106, 2, 27, "Input"],
Cell[1826, 53, 853, 25, 320, "Text"],
Cell[2682, 80, 171, 3, 59, "Input"],
Cell[2856, 85, 85, 1, 27, "Input"],
Cell[2944, 88, 351, 6, 75, "Input"],
Cell[3298, 96, 206, 4, 75, "Input"],
Cell[3507, 102, 351, 6, 107, "Input"],
Cell[3861, 110, 293, 5, 91, "Input"],
Cell[4157, 117, 141, 3, 27, "Input"],
Cell[4301, 122, 3322, 59, 555, "Input"],
Cell[7626, 183, 1165, 21, 171, "Input"],
Cell[8794, 206, 6682, 126, 1195, "Input"],
Cell[15479, 334, 6646, 126, 1195, "Input"],
Cell[22128, 462, 738, 15, 203, "Input"],
Cell[22869, 479, 1150, 21, 267, "Input"],
Cell[24022, 502, 3932, 74, 795, "Input"],
Cell[27957, 578, 3703, 69, 731, "Input"],
Cell[31663, 649, 1236, 24, 251, "Input"],
Cell[32902, 675, 1507, 27, 315, "Input"],
Cell[34412, 704, 2929, 51, 571, "Input"],
Cell[37344, 757, 2489, 45, 475, "Input"],
Cell[39836, 804, 1122, 21, 251, "Input"],
Cell[40961, 827, 640, 13, 171, "Input"],
Cell[41604, 842, 2509, 49, 427, "Input"],
Cell[44116, 893, 1092, 21, 283, "Input"],
Cell[45211, 916, 2326, 44, 507, "Input"],
Cell[47540, 962, 2599, 48, 507, "Input"],
Cell[50142, 1012, 231, 4, 59, "Input"],
Cell[50376, 1018, 342, 11, 147, "Input"],
Cell[50721, 1031, 671, 12, 219, "Input"],
Cell[51395, 1045, 2498, 48, 459, "Input"],
Cell[53896, 1095, 6680, 126, 1195, "Input"],
Cell[60579, 1223, 6628, 125, 1195, "Input"],
Cell[67210, 1350, 6719, 126, 1211, "Input"],
Cell[73932, 1478, 2796, 52, 523, "Input"],
Cell[76731, 1532, 2293, 41, 459, "Input"],
Cell[79027, 1575, 154, 2, 43, "Input"],
Cell[79184, 1579, 247, 5, 43, "Input"],
Cell[79434, 1586, 2735, 47, 539, "Input"],
Cell[82172, 1635, 1087, 19, 219, "Input"],
Cell[83262, 1656, 949, 16, 235, "Input"],
Cell[84214, 1674, 1301, 24, 283, "Input"],
Cell[85518, 1700, 4345, 73, 875, "Input"],
Cell[89866, 1775, 1647, 23, 363, "Input"],
Cell[91516, 1800, 45, 1, 27, "Input"]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

