(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     44515,       1228]*)
(*NotebookOutlinePosition[     45177,       1252]*)
(*  CellTagsIndexPosition[     45133,       1248]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell[BoxData[
    \(<< projects/mathematica/sugra.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(xp\  = \ {r, th, ph, t};\)\), "\n", 
    \(\(gp\  = \ IdentityMatrix[4];\)\), "\n", 
    \(\(sigma\  = \ r^2 + \((A*Cos[th])\)^2;\)\), "\n", 
    \(\(delta\  = \ r^2 - 2  M*r + e^2 + A^2;\)\), "\n", 
    \(\(gp[\([1, 1]\)]\  = \ sigma/delta;\)\), "\n", 
    \(\(gp[\([2, 2]\)]\  = \ sigma;\)\), "\n", 
    \(\(gp[\([3, 
            3]\)]\  = \ \((\((r^2 + A^2)\)^2 - 
              delta*A^2 \((Sin[th])\)^2)\) \((Sin[th])\)^2/sigma;\)\), "\n", 
    \(\(gp[\([3, 4]\)]\  = \ \(-A\) \((Sin[th])\)^2 \((r^2 + A^2 - delta)\)/
            sigma;\)\), "\n", 
    \(\(gp[\([4, 3]\)]\  = \ \(-A\) \((Sin[th])\)^2 \((r^2 + A^2 - delta)\)/
            sigma;\)\), "\n", 
    \(\(gp[\([4, 4]\)]\  = \ \(-\((delta - A^2 \((Sin[th])\)^2)\)\)/
          sigma;\)\), "\n", 
    \(\(components[gp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2 + A\^2\ \
Cos[th]\^2\)\/\(A\^2 + e\^2 - 2\ M\ r + r\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ A, 2], 
              Power[ e, 2], 
              Times[ -2, M, r], 
              Power[ r, 2]], -1], 
          Plus[ 
            Power[ r, 2], 
            Times[ 
              Power[ A, 2], 
              Power[ 
                Cos[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2 + 
          A\^2\ Cos[th]\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Plus[ 
          Power[ r, 2], 
          Times[ 
            Power[ A, 2], 
            Power[ 
              Cos[ th], 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(Sin[th]\^2\ \
\((\((A\^2 + r\^2)\)\^2 - A\^2\ \((A\^2 + e\^2 - 2\ M\ r + r\^2)\)\ \
Sin[th]\^2)\)\)\/\(r\^2 + A\^2\ Cos[th]\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ A, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2], 
          Plus[ 
            Power[ 
              Plus[ 
                Power[ A, 2], 
                Power[ r, 2]], 2], 
            Times[ -1, 
              Power[ A, 2], 
              Plus[ 
                Power[ A, 2], 
                Power[ e, 2], 
                Times[ -2, M, r], 
                Power[ r, 2]], 
              Power[ 
                Sin[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(A\ \
\((\(-e\^2\) + 2\ M\ r)\)\ Sin[th]\^2\)\/\(r\^2 + A\^2\ Cos[th]\^2\)\)\)\),
      SequenceForm[ "[", 3, ",", 4, "] = ", 
        Times[ -1, A, 
          Plus[ 
            Times[ -1, 
              Power[ e, 2]], 
            Times[ 2, M, r]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ A, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(A\ \
\((\(-e\^2\) + 2\ M\ r)\)\ Sin[th]\^2\)\/\(r\^2 + A\^2\ Cos[th]\^2\)\)\)\),
      SequenceForm[ "[", 4, ",", 3, "] = ", 
        Times[ -1, A, 
          Plus[ 
            Times[ -1, 
              Power[ e, 2]], 
            Times[ 2, M, r]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ A, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-A\^2\) - e\^2 \
+ 2\ M\ r - r\^2 + A\^2\ Sin[th]\^2\)\/\(r\^2 + A\^2\ Cos[th]\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ A, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Plus[ 
            Times[ -1, 
              Power[ A, 2]], 
            Times[ -1, 
              Power[ e, 2]], 
            Times[ 2, M, r], 
            Times[ -1, 
              Power[ r, 2]], 
            Times[ 
              Power[ A, 2], 
              Power[ 
                Sin[ th], 2]]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[BoxData[{
    \(\(A = 0;\)\), "\[IndentingNewLine]", 
    \(\(e = 0;\)\)}], "Input"],

Cell[BoxData[
    \(\(spacetime[xp, gp];\)\)], "Input"],

Cell[BoxData[
    \(\(curvature[];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[currentst[\([5]\)]];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]","\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]M\/\(2\ M - r\)\),
      SequenceForm[ "[", 1, ",", 2, ",", 1, ",", 2, "] = ", 
        Times[ M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]","\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(M\/\(2\ M - 
                r\)\)\)\),
      SequenceForm[ "[", 1, ",", 2, ",", 2, ",", 1, "] = ", 
        Times[ -1, M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]","\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(M\ Sin[th]\^2\)\/\
\(2\ M - r\)\),
      SequenceForm[ "[", 1, ",", 3, ",", 1, ",", 3, "] = ", 
        Times[ M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]","\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(M\ \
Sin[th]\^2\)\/\(2\ M - r\)\)\)\),
      SequenceForm[ "[", 1, ",", 3, ",", 3, ",", 1, "] = ", 
        Times[ -1, M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]","\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(2\ M\)\/r\^3\
\)\)\),
      SequenceForm[ "[", 1, ",", 4, ",", 1, ",", 4, "] = ", 
        Times[ -2, M, 
          Power[ r, -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]","\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ M\)\/r\^3\),
      SequenceForm[ "[", 1, ",", 4, ",", 4, ",", 1, "] = ", 
        Times[ 2, M, 
          Power[ r, -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]","\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(M\/\(2\ M - 
                r\)\)\)\),
      SequenceForm[ "[", 2, ",", 1, ",", 1, ",", 2, "] = ", 
        Times[ -1, M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]","\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]M\/\(2\ M - r\)\),
      SequenceForm[ "[", 2, ",", 1, ",", 2, ",", 1, "] = ", 
        Times[ M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]","\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ M\ r\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 2, ",", 3, ",", 2, ",", 3, "] = ", 
        Times[ 2, M, r, 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]","\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-2\)\ M\ r\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 2, ",", 3, ",", 3, ",", 2, "] = ", 
        Times[ -2, M, r, 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]","\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(M\ \((\(-2\)\ M + \
r)\)\)\/r\^2\),
      SequenceForm[ "[", 2, ",", 4, ",", 2, ",", 4, "] = ", 
        Times[ M, 
          Power[ r, -2], 
          Plus[ 
            Times[ -2, M], r]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]","\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(M\ \((\(-2\)\
\ M + r)\)\)\/r\^2\)\)\),
      SequenceForm[ "[", 2, ",", 4, ",", 4, ",", 2, "] = ", 
        Times[ -1, M, 
          Power[ r, -2], 
          Plus[ 
            Times[ -2, M], r]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]","\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(M\ \
Sin[th]\^2\)\/\(2\ M - r\)\)\)\),
      SequenceForm[ "[", 3, ",", 1, ",", 1, ",", 3, "] = ", 
        Times[ -1, M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]","\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(M\ Sin[th]\^2\)\/\
\(2\ M - r\)\),
      SequenceForm[ "[", 3, ",", 1, ",", 3, ",", 1, "] = ", 
        Times[ M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]","\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-2\)\ M\ r\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 3, ",", 2, ",", 2, ",", 3, "] = ", 
        Times[ -2, M, r, 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]","\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ M\ r\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 3, ",", 2, ",", 3, ",", 2, "] = ", 
        Times[ 2, M, r, 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]","\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(M\ \((2\ M - 
                    r)\)\ Sin[th]\^2\)\/r\^2\)\)\),
      SequenceForm[ "[", 3, ",", 4, ",", 3, ",", 4, "] = ", 
        Times[ -1, M, 
          Plus[ 
            Times[ 2, M], 
            Times[ -1, r]], 
          Power[ r, -2], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]","\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(M\ \((2\ M - r)\)\
\ Sin[th]\^2\)\/r\^2\),
      SequenceForm[ "[", 3, ",", 4, ",", 4, ",", 3, "] = ", 
        Times[ M, 
          Plus[ 
            Times[ 2, M], 
            Times[ -1, r]], 
          Power[ r, -2], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]","\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ M\)\/r\^3\),
      SequenceForm[ "[", 4, ",", 1, ",", 1, ",", 4, "] = ", 
        Times[ 2, M, 
          Power[ r, -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]","\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(2\ M\)\/r\^3\
\)\)\),
      SequenceForm[ "[", 4, ",", 1, ",", 4, ",", 1, "] = ", 
        Times[ -2, M, 
          Power[ r, -3]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]","\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(M\ \((\(-2\)\
\ M + r)\)\)\/r\^2\)\)\),
      SequenceForm[ "[", 4, ",", 2, ",", 2, ",", 4, "] = ", 
        Times[ -1, M, 
          Power[ r, -2], 
          Plus[ 
            Times[ -2, M], r]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]","\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(M\ \((\(-2\)\ M + \
r)\)\)\/r\^2\),
      SequenceForm[ "[", 4, ",", 2, ",", 4, ",", 2, "] = ", 
        Times[ M, 
          Power[ r, -2], 
          Plus[ 
            Times[ -2, M], r]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]","\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(M\ \((2\ M - r)\)\
\ Sin[th]\^2\)\/r\^2\),
      SequenceForm[ "[", 4, ",", 3, ",", 3, ",", 4, "] = ", 
        Times[ M, 
          Plus[ 
            Times[ 2, M], 
            Times[ -1, r]], 
          Power[ r, -2], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]","\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(M\ \((2\ M - 
                    r)\)\ Sin[th]\^2\)\/r\^2\)\)\),
      SequenceForm[ "[", 4, ",", 3, ",", 4, ",", 3, "] = ", 
        Times[ -1, M, 
          Plus[ 
            Times[ 2, M], 
            Times[ -1, r]], 
          Power[ r, -2], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(xpp = {r, th, php, t};\)\), "\[IndentingNewLine]", 
    \(\(gpp\  = \ 
        coordxform[gp, {r, th, php - 2  t, t}, 
          xpp];\)\), "\[IndentingNewLine]", 
    \(\(components[gpp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]r\/\(\(-2\)\ M + \
r\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ r, 
          Power[ 
            Plus[ 
              Times[ -2, M], r], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]r\^2\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Power[ r, 2]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-2\)\ r\^2\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 3, ",", 4, "] = ", 
        Times[ -2, 
          Power[ r, 2], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-2\)\ r\^2\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 4, ",", 3, "] = ", 
        Times[ -2, 
          Power[ r, 2], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-1\) + \(2\ M\)\
\/r + 4\ r\^2\ Sin[th]\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Plus[ -1, 
          Times[ 2, M, 
            Power[ r, -1]], 
          Times[ 4, 
            Power[ r, 2], 
            Power[ 
              Sin[ th], 2]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[BoxData[{
    \(\(spacetime[xpp, gpp];\)\), "\[IndentingNewLine]", 
    \(\(curvature[];\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(components[currentst[\([5]\)]];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]","\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]M\/\(2\ M - r\)\),
      SequenceForm[ "[", 1, ",", 2, ",", 1, ",", 2, "] = ", 
        Times[ M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]","\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(M\/\(2\ M - 
                r\)\)\)\),
      SequenceForm[ "[", 1, ",", 2, ",", 2, ",", 1, "] = ", 
        Times[ -1, M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]","\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(M\ Sin[th]\^2\)\/\
\(2\ M - r\)\),
      SequenceForm[ "[", 1, ",", 3, ",", 1, ",", 3, "] = ", 
        Times[ M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]","\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ M\ \
Sin[th]\^2\)\/\(\(-2\)\ M + r\)\),
      SequenceForm[ "[", 1, ",", 3, ",", 1, ",", 4, "] = ", 
        Times[ 2, M, 
          Power[ 
            Plus[ 
              Times[ -2, M], r], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]","\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(M\ \
Sin[th]\^2\)\/\(2\ M - r\)\)\)\),
      SequenceForm[ "[", 1, ",", 3, ",", 3, ",", 1, "] = ", 
        Times[ -1, M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]","\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(2\ M\ \
Sin[th]\^2\)\/\(\(-2\)\ M + r\)\)\)\),
      SequenceForm[ "[", 1, ",", 3, ",", 4, ",", 1, "] = ", 
        Times[ -2, M, 
          Power[ 
            Plus[ 
              Times[ -2, M], r], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]","\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ M\ \
Sin[th]\^2\)\/\(\(-2\)\ M + r\)\),
      SequenceForm[ "[", 1, ",", 4, ",", 1, ",", 3, "] = ", 
        Times[ 2, M, 
          Power[ 
            Plus[ 
              Times[ -2, M], r], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]","\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ M\ \((\(-2\)\ \
M + r + 2\ r\^3\ Sin[th]\^2)\)\)\/\(\((2\ M - r)\)\ r\^3\)\),
      SequenceForm[ "[", 1, ",", 4, ",", 1, ",", 4, "] = ", 
        Times[ 2, M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1], 
          Power[ r, -3], 
          Plus[ 
            Times[ -2, M], r, 
            Times[ 2, 
              Power[ r, 3], 
              Power[ 
                Sin[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]","\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(2\ M\ \
Sin[th]\^2\)\/\(\(-2\)\ M + r\)\)\)\),
      SequenceForm[ "[", 1, ",", 4, ",", 3, ",", 1, "] = ", 
        Times[ -2, M, 
          Power[ 
            Plus[ 
              Times[ -2, M], r], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]","\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(2\ M\ \
\((\(-2\)\ M + r + 2\ r\^3\ Sin[th]\^2)\)\)\/\(\((2\ M - r)\)\ r\^3\)\)\)\),
      SequenceForm[ "[", 1, ",", 4, ",", 4, ",", 1, "] = ", 
        Times[ -2, M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1], 
          Power[ r, -3], 
          Plus[ 
            Times[ -2, M], r, 
            Times[ 2, 
              Power[ r, 3], 
              Power[ 
                Sin[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]","\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(M\/\(2\ M - 
                r\)\)\)\),
      SequenceForm[ "[", 2, ",", 1, ",", 1, ",", 2, "] = ", 
        Times[ -1, M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]","\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]M\/\(2\ M - r\)\),
      SequenceForm[ "[", 2, ",", 1, ",", 2, ",", 1, "] = ", 
        Times[ M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]","\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ M\ r\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 2, ",", 3, ",", 2, ",", 3, "] = ", 
        Times[ 2, M, r, 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]","\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-4\)\ M\ r\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 2, ",", 3, ",", 2, ",", 4, "] = ", 
        Times[ -4, M, r, 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]","\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-2\)\ M\ r\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 2, ",", 3, ",", 3, ",", 2, "] = ", 
        Times[ -2, M, r, 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]","\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(4\ M\ r\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 2, ",", 3, ",", 4, ",", 2, "] = ", 
        Times[ 4, M, r, 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]","\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-4\)\ M\ r\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 2, ",", 4, ",", 2, ",", 3, "] = ", 
        Times[ -4, M, r, 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]","\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(M\ \((\(-2\)\ M + \
r + 4\ r\^3 - 4\ r\^3\ Cos[2\ th])\)\)\/r\^2\),
      SequenceForm[ "[", 2, ",", 4, ",", 2, ",", 4, "] = ", 
        Times[ M, 
          Power[ r, -2], 
          Plus[ 
            Times[ -2, M], r, 
            Times[ 4, 
              Power[ r, 3]], 
            Times[ -4, 
              Power[ r, 3], 
              Cos[ 
                Times[ 2, th]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]","\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(4\ M\ r\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 2, ",", 4, ",", 3, ",", 2, "] = ", 
        Times[ 4, M, r, 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]","\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(M\ \((\(-2\)\
\ M + r + 4\ r\^3 - 4\ r\^3\ Cos[2\ th])\)\)\/r\^2\)\)\),
      SequenceForm[ "[", 2, ",", 4, ",", 4, ",", 2, "] = ", 
        Times[ -1, M, 
          Power[ r, -2], 
          Plus[ 
            Times[ -2, M], r, 
            Times[ 4, 
              Power[ r, 3]], 
            Times[ -4, 
              Power[ r, 3], 
              Cos[ 
                Times[ 2, th]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]","\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(M\ \
Sin[th]\^2\)\/\(2\ M - r\)\)\)\),
      SequenceForm[ "[", 3, ",", 1, ",", 1, ",", 3, "] = ", 
        Times[ -1, M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]","\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(2\ M\ \
Sin[th]\^2\)\/\(\(-2\)\ M + r\)\)\)\),
      SequenceForm[ "[", 3, ",", 1, ",", 1, ",", 4, "] = ", 
        Times[ -2, M, 
          Power[ 
            Plus[ 
              Times[ -2, M], r], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]","\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(M\ Sin[th]\^2\)\/\
\(2\ M - r\)\),
      SequenceForm[ "[", 3, ",", 1, ",", 3, ",", 1, "] = ", 
        Times[ M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]","\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ M\ \
Sin[th]\^2\)\/\(\(-2\)\ M + r\)\),
      SequenceForm[ "[", 3, ",", 1, ",", 4, ",", 1, "] = ", 
        Times[ 2, M, 
          Power[ 
            Plus[ 
              Times[ -2, M], r], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]","\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-2\)\ M\ r\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 3, ",", 2, ",", 2, ",", 3, "] = ", 
        Times[ -2, M, r, 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]","\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(4\ M\ r\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 3, ",", 2, ",", 2, ",", 4, "] = ", 
        Times[ 4, M, r, 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]","\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ M\ r\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 3, ",", 2, ",", 3, ",", 2, "] = ", 
        Times[ 2, M, r, 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]","\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-4\)\ M\ r\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 3, ",", 2, ",", 4, ",", 2, "] = ", 
        Times[ -4, M, r, 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]","\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(M\ \((\(-2\)\ M + \
r)\)\ Sin[th]\^2\)\/r\^2\),
      SequenceForm[ "[", 3, ",", 4, ",", 3, ",", 4, "] = ", 
        Times[ M, 
          Power[ r, -2], 
          Plus[ 
            Times[ -2, M], r], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]","\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(M\ \((\(-2\)\
\ M + r)\)\ Sin[th]\^2\)\/r\^2\)\)\),
      SequenceForm[ "[", 3, ",", 4, ",", 4, ",", 3, "] = ", 
        Times[ -1, M, 
          Power[ r, -2], 
          Plus[ 
            Times[ -2, M], r], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]","\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(2\ M\ \
Sin[th]\^2\)\/\(\(-2\)\ M + r\)\)\)\),
      SequenceForm[ "[", 4, ",", 1, ",", 1, ",", 3, "] = ", 
        Times[ -2, M, 
          Power[ 
            Plus[ 
              Times[ -2, M], r], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]","\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(2\ M\ \
\((\(-2\)\ M + r + 2\ r\^3\ Sin[th]\^2)\)\)\/\(\((2\ M - r)\)\ r\^3\)\)\)\),
      SequenceForm[ "[", 4, ",", 1, ",", 1, ",", 4, "] = ", 
        Times[ -2, M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1], 
          Power[ r, -3], 
          Plus[ 
            Times[ -2, M], r, 
            Times[ 2, 
              Power[ r, 3], 
              Power[ 
                Sin[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]","\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ M\ \
Sin[th]\^2\)\/\(\(-2\)\ M + r\)\),
      SequenceForm[ "[", 4, ",", 1, ",", 3, ",", 1, "] = ", 
        Times[ 2, M, 
          Power[ 
            Plus[ 
              Times[ -2, M], r], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]","\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ M\ \((\(-2\)\ \
M + r + 2\ r\^3\ Sin[th]\^2)\)\)\/\(\((2\ M - r)\)\ r\^3\)\),
      SequenceForm[ "[", 4, ",", 1, ",", 4, ",", 1, "] = ", 
        Times[ 2, M, 
          Power[ 
            Plus[ 
              Times[ 2, M], 
              Times[ -1, r]], -1], 
          Power[ r, -3], 
          Plus[ 
            Times[ -2, M], r, 
            Times[ 2, 
              Power[ r, 3], 
              Power[ 
                Sin[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]","\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(4\ M\ r\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 4, ",", 2, ",", 2, ",", 3, "] = ", 
        Times[ 4, M, r, 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]","\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(M\ \((\(-2\)\
\ M + r + 4\ r\^3 - 4\ r\^3\ Cos[2\ th])\)\)\/r\^2\)\)\),
      SequenceForm[ "[", 4, ",", 2, ",", 2, ",", 4, "] = ", 
        Times[ -1, M, 
          Power[ r, -2], 
          Plus[ 
            Times[ -2, M], r, 
            Times[ 4, 
              Power[ r, 3]], 
            Times[ -4, 
              Power[ r, 3], 
              Cos[ 
                Times[ 2, th]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]","\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-4\)\ M\ r\ \
Sin[th]\^2\)\),
      SequenceForm[ "[", 4, ",", 2, ",", 3, ",", 2, "] = ", 
        Times[ -4, M, r, 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]","\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(M\ \((\(-2\)\ M + \
r + 4\ r\^3 - 4\ r\^3\ Cos[2\ th])\)\)\/r\^2\),
      SequenceForm[ "[", 4, ",", 2, ",", 4, ",", 2, "] = ", 
        Times[ M, 
          Power[ r, -2], 
          Plus[ 
            Times[ -2, M], r, 
            Times[ 4, 
              Power[ r, 3]], 
            Times[ -4, 
              Power[ r, 3], 
              Cos[ 
                Times[ 2, th]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]","\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(M\ \((\(-2\)\
\ M + r)\)\ Sin[th]\^2\)\/r\^2\)\)\),
      SequenceForm[ "[", 4, ",", 3, ",", 3, ",", 4, "] = ", 
        Times[ -1, M, 
          Power[ r, -2], 
          Plus[ 
            Times[ -2, M], r], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]","\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(M\ \((\(-2\)\ M + \
r)\)\ Sin[th]\^2\)\/r\^2\),
      SequenceForm[ "[", 4, ",", 3, ",", 4, ",", 3, "] = ", 
        Times[ M, 
          Power[ r, -2], 
          Plus[ 
            Times[ -2, M], r], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[TrigExpand[4\ r\^3 - 4\ r\^3\ Cos[2\ th]]]\)], "Input"],

Cell[BoxData[
    \(\(-4\)\ r\^3\ \((\(-1\) + Cos[th]\^2 - Sin[th]\^2)\)\)], "Output"]
}, Open  ]]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
WindowSize->{734, 504},
WindowMargins->{{Automatic, 32}, {-354, Automatic}},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1739, 51, 64, 1, 27, "Input"],
Cell[1806, 54, 2970, 44, 375, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[4813, 103, 822, 16, 184, "Input"],
Cell[5638, 121, 612, 18, 42, "Print"],
Cell[6253, 141, 395, 11, 25, "Print"],
Cell[6651, 154, 979, 30, 47, "Print"],
Cell[7633, 186, 659, 19, 45, "Print"],
Cell[8295, 207, 659, 19, 45, "Print"],
Cell[8957, 228, 821, 25, 45, "Print"]
}, Open  ]],
Cell[9793, 256, 89, 2, 43, "Input"],
Cell[9885, 260, 55, 1, 27, "Input"],
Cell[9943, 263, 49, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[10017, 268, 68, 1, 27, "Input"],
Cell[10088, 271, 452, 10, 39, "Print"],
Cell[10543, 283, 482, 11, 39, "Print"],
Cell[11028, 296, 515, 13, 42, "Print"],
Cell[11546, 311, 528, 13, 42, "Print"],
Cell[12077, 326, 390, 8, 40, "Print"],
Cell[12470, 336, 378, 7, 40, "Print"],
Cell[12851, 345, 482, 11, 39, "Print"],
Cell[13336, 358, 452, 10, 39, "Print"],
Cell[13791, 370, 411, 9, 25, "Print"],
Cell[14205, 381, 417, 9, 25, "Print"],
Cell[14625, 392, 444, 10, 40, "Print"],
Cell[15072, 404, 457, 10, 40, "Print"],
Cell[15532, 416, 528, 13, 42, "Print"],
Cell[16063, 431, 515, 13, 42, "Print"],
Cell[16581, 446, 417, 9, 25, "Print"],
Cell[17001, 457, 411, 9, 25, "Print"],
Cell[17415, 468, 552, 13, 42, "Print"],
Cell[17970, 483, 520, 13, 42, "Print"],
Cell[18493, 498, 378, 7, 40, "Print"],
Cell[18874, 507, 390, 8, 40, "Print"],
Cell[19267, 517, 457, 10, 40, "Print"],
Cell[19727, 529, 444, 10, 40, "Print"],
Cell[20174, 541, 520, 13, 42, "Print"],
Cell[20697, 556, 552, 13, 42, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[21286, 574, 227, 5, 59, "Input"],
Cell[21516, 581, 339, 9, 37, "Print"],
Cell[21858, 592, 248, 5, 25, "Print"],
Cell[22109, 599, 330, 9, 25, "Print"],
Cell[22442, 610, 342, 9, 25, "Print"],
Cell[22787, 621, 342, 9, 25, "Print"],
Cell[23132, 632, 440, 12, 40, "Print"]
}, Open  ]],
Cell[23587, 647, 109, 2, 43, "Input"],

Cell[CellGroupData[{
Cell[23721, 653, 68, 1, 27, "Input"],
Cell[23792, 656, 452, 10, 39, "Print"],
Cell[24247, 668, 482, 11, 39, "Print"],
Cell[24732, 681, 515, 13, 42, "Print"],
Cell[25250, 696, 500, 12, 42, "Print"],
Cell[25753, 710, 528, 13, 42, "Print"],
Cell[26284, 725, 510, 12, 42, "Print"],
Cell[26797, 739, 500, 12, 42, "Print"],
Cell[27300, 753, 700, 18, 44, "Print"],
Cell[28003, 773, 510, 12, 42, "Print"],
Cell[28516, 787, 710, 18, 44, "Print"],
Cell[29229, 807, 482, 11, 39, "Print"],
Cell[29714, 820, 452, 10, 39, "Print"],
Cell[30169, 832, 411, 9, 25, "Print"],
Cell[30583, 843, 417, 9, 25, "Print"],
Cell[31003, 854, 417, 9, 25, "Print"],
Cell[31423, 865, 411, 9, 25, "Print"],
Cell[31837, 876, 417, 9, 25, "Print"],
Cell[32257, 887, 636, 16, 42, "Print"],
Cell[32896, 905, 411, 9, 25, "Print"],
Cell[33310, 916, 649, 16, 42, "Print"],
Cell[33962, 934, 528, 13, 42, "Print"],
Cell[34493, 949, 510, 12, 42, "Print"],
Cell[35006, 963, 515, 13, 42, "Print"],
Cell[35524, 978, 500, 12, 42, "Print"],
Cell[36027, 992, 417, 9, 25, "Print"],
Cell[36447, 1003, 411, 9, 25, "Print"],
Cell[36861, 1014, 411, 9, 25, "Print"],
Cell[37275, 1025, 417, 9, 25, "Print"],
Cell[37695, 1036, 501, 12, 42, "Print"],
Cell[38199, 1050, 514, 12, 42, "Print"],
Cell[38716, 1064, 510, 12, 42, "Print"],
Cell[39229, 1078, 710, 18, 44, "Print"],
Cell[39942, 1098, 500, 12, 42, "Print"],
Cell[40445, 1112, 700, 18, 44, "Print"],
Cell[41148, 1132, 411, 9, 25, "Print"],
Cell[41562, 1143, 649, 16, 42, "Print"],
Cell[42214, 1161, 417, 9, 25, "Print"],
Cell[42634, 1172, 636, 16, 42, "Print"],
Cell[43273, 1190, 514, 12, 42, "Print"],
Cell[43790, 1204, 501, 12, 42, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[44328, 1221, 82, 1, 31, "Input"],
Cell[44413, 1224, 86, 1, 29, "Output"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

