(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     43474,       1133]*)
(*NotebookOutlinePosition[     44513,       1168]*)
(*  CellTagsIndexPosition[     44469,       1164]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(\(Clear["\<Global`*\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< projects/mathematica/sugra.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(xp\  = \ {r, th, ph, t};\)\), "\n", 
    \(\(gp\  = \ IdentityMatrix[4];\)\), "\n", 
    \(\(sigma\  = \ r^2 + \((A*Cos[th])\)^2;\)\), "\n", 
    \(\(delta\  = \ r^2 - 2  M*r + e^2 + A^2;\)\), "\n", 
    \(\(gp[\([1, 1]\)]\  = \ sigma/delta;\)\), "\n", 
    \(\(gp[\([2, 2]\)]\  = \ sigma;\)\), "\n", 
    \(\(gp[\([3, 
            3]\)]\  = \ \((\((r^2 + A^2)\)^2 - 
              delta*A^2 \((Sin[th])\)^2)\) \((Sin[th])\)^2/sigma;\)\), "\n", 
    \(\(gp[\([3, 4]\)]\  = \ \(-A\) \((Sin[th])\)^2 \((r^2 + A^2 - delta)\)/
            sigma;\)\), "\n", 
    \(\(gp[\([4, 3]\)]\  = \ \(-A\) \((Sin[th])\)^2 \((r^2 + A^2 - delta)\)/
            sigma;\)\), "\n", 
    \(\(gp[\([4, 4]\)]\  = \ \(-\((delta - A^2 \((Sin[th])\)^2)\)\)/
          sigma;\)\), "\n", 
    \(\(components[gp];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2 + A\^2\ \
Cos[th]\^2\)\/\(A\^2 + e\^2 - 2\ M\ r + r\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ A, 2], 
              Power[ e, 2], 
              Times[ -2, M, r], 
              Power[ r, 2]], -1], 
          Plus[ 
            Power[ r, 2], 
            Times[ 
              Power[ A, 2], 
              Power[ 
                Cos[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(r\^2 + 
          A\^2\ Cos[th]\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Plus[ 
          Power[ r, 2], 
          Times[ 
            Power[ A, 2], 
            Power[ 
              Cos[ th], 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(Sin[th]\^2\ \
\((\((A\^2 + r\^2)\)\^2 - A\^2\ \((A\^2 + e\^2 - 2\ M\ r + r\^2)\)\ \
Sin[th]\^2)\)\)\/\(r\^2 + A\^2\ Cos[th]\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ A, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2], 
          Plus[ 
            Power[ 
              Plus[ 
                Power[ A, 2], 
                Power[ r, 2]], 2], 
            Times[ -1, 
              Power[ A, 2], 
              Plus[ 
                Power[ A, 2], 
                Power[ e, 2], 
                Times[ -2, M, r], 
                Power[ r, 2]], 
              Power[ 
                Sin[ th], 2]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(A\ \
\((\(-e\^2\) + 2\ M\ r)\)\ Sin[th]\^2\)\/\(r\^2 + A\^2\ Cos[th]\^2\)\)\)\),
      SequenceForm[ "[", 3, ",", 4, "] = ", 
        Times[ -1, A, 
          Plus[ 
            Times[ -1, 
              Power[ e, 2]], 
            Times[ 2, M, r]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ A, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(A\ \
\((\(-e\^2\) + 2\ M\ r)\)\ Sin[th]\^2\)\/\(r\^2 + A\^2\ Cos[th]\^2\)\)\)\),
      SequenceForm[ "[", 4, ",", 3, "] = ", 
        Times[ -1, A, 
          Plus[ 
            Times[ -1, 
              Power[ e, 2]], 
            Times[ 2, M, r]], 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ A, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Power[ 
            Sin[ th], 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-A\^2\) - e\^2 \
+ 2\ M\ r - r\^2 + A\^2\ Sin[th]\^2\)\/\(r\^2 + A\^2\ Cos[th]\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Power[ 
            Plus[ 
              Power[ r, 2], 
              Times[ 
                Power[ A, 2], 
                Power[ 
                  Cos[ th], 2]]], -1], 
          Plus[ 
            Times[ -1, 
              Power[ A, 2]], 
            Times[ -1, 
              Power[ e, 2]], 
            Times[ 2, M, r], 
            Times[ -1, 
              Power[ r, 2]], 
            Times[ 
              Power[ A, 2], 
              Power[ 
                Sin[ th], 2]]]]],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(spacetime[xp, gp];\)\), "\[IndentingNewLine]", 
    \(\(eps4\  = \ \(-epsilon[4, {}]\);\)\), "\n", 
    \(\(x\  = \ {r, c, ph, t};\)\), "\n", 
    \(\(g\  = \ coordxform[gp, {r, ArcCos[c], ph, t}, x];\)\), "\n", 
    \(\(components[g];\)\), "\n", 
    \(\(spacetime[x, g];\)\), "\n", 
    \(\(timestamp[];\)\), "\n", 
    \(\(curvature[];\)\), "\n", 
    \(\(timestamp[];\)\), "\n", 
    \(Factor[kretschmann[]]\), "\[IndentingNewLine]", 
    \(\(timestamp[];\)\), "\[IndentingNewLine]", 
    \(\(irmdg\  = \ 
        Simplify[
          PowerExpand[
            1/Sqrt[\(-Simplify[
                    Det[currentst[\([2]\)]]]\)]]];\)\), \
"\[IndentingNewLine]", 
    \(\(rdual\  = \ 
        Table[Table[Table[Table[0, {l, 4}], {k, 4}], {j, 4}], {i, 
            4}];\)\), "\n", 
    \(\(Do[
        Do[Do[Do[
              rdual[\([i, j, k, l]\)] = 
                2  irmdg\ *\ 
                  Sum[Sum[
                      eps4[\([i, j, m, 
                            n]\)]\ *\n\t\t\t\t\t\t\t\t\tcurrentst[\([5, m, n, 
                            k, l]\)], {n, m + 1, 4}], {m, 4}], \n\t\t\t\t{l, 
                k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(\(Do[
        Do[Do[Do[
              rdual[\([j, i, k, 
                    l]\)]\  = \ \(-rdual[\([i, j, k, l]\)]\), \n\t\t\t\t{l, 
                k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(\(Do[
        Do[Do[Do[
              rdual[\([i, j, l, 
                    k]\)]\  = \ \(-rdual[\([i, j, k, l]\)]\), \n\t\t\t\t{l, 
                k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(\(Do[
        Do[Do[Do[
              rdual[\([j, i, l, k]\)]\  = \ 
                rdual[\([i, j, k, l]\)], \n\t\t\t\t{l, k + 1, 4}], {k, 
              4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(\(r2up\  = \ raise[raise[currentst[\([5]\)], 1], 2];\)\), "\n", 
    \(rstarr\  = \ 
      Simplify[Sum[
          Sum[Sum[Sum[
                4  rdual[\([i, j, k, l]\)] 
                  r2up[\([k, l, i, j]\)], \n\t\t\t\t\t\t{l, k + 1, 4}], {k, 
                4}], {j, i + 1, 4}], {i, 4}]]\), "\n", 
    \(\(timestamp[];\)\), "\n", 
    \(\(cdr\  = \ covder[currentst[\([5]\)]];\)\), "\n", 
    \(\(cdr5u\  = \ 
        raise[raise[raise[raise[raise[cdr, 1], 2], 3], 4], 5];\)\), "\n", 
    \(\(cdrsqun\  = \ 
        Sum[
          Sum[Sum[Sum[
                Sum[4  cdr[\([i, j, k, l, m]\)] 
                    cdr5u[\([i, j, k, l, m]\)], \n\t\t\t\t\t\t{m, 4}], {l, 
                  k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(Factor[cdrsqun]\), "\n", 
    \(\(timestamp[];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(A\^2\ c\^2 + r\^2\
\)\/\(A\^2 + e\^2 - 2\ M\ r + r\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Plus[ 
            Times[ 
              Power[ A, 2], 
              Power[ c, 2]], 
            Power[ r, 2]], 
          Power[ 
            Plus[ 
              Power[ A, 2], 
              Power[ e, 2], 
              Times[ -2, M, r], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(A\^2\ c\^2 + r\^2\
\)\/\(1 - c\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ 
            Plus[ 1, 
              Times[ -1, 
                Power[ c, 2]]], -1], 
          Plus[ 
            Times[ 
              Power[ A, 2], 
              Power[ c, 2]], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\((1 - c\^2)\)\ \
\((\((A\^2 + r\^2)\)\^2 + A\^2\ \((\(-1\) + c\^2)\)\ \((A\^2 + e\^2 + r\ \
\((\(-2\)\ M + r)\))\))\)\)\/\(A\^2\ c\^2 + r\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Plus[ 1, 
            Times[ -1, 
              Power[ c, 2]]], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ A, 2], 
                Power[ c, 2]], 
              Power[ r, 2]], -1], 
          Plus[ 
            Power[ 
              Plus[ 
                Power[ A, 2], 
                Power[ r, 2]], 2], 
            Times[ 
              Power[ A, 2], 
              Plus[ -1, 
                Power[ c, 2]], 
              Plus[ 
                Power[ A, 2], 
                Power[ e, 2], 
                Times[ r, 
                  Plus[ 
                    Times[ -2, M], r]]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(A\ \((\(-1\) \
+ c\^2)\)\ \((e\^2 - 2\ M\ r)\)\)\/\(A\^2\ c\^2 + r\^2\)\)\)\),
      SequenceForm[ "[", 3, ",", 4, "] = ", 
        Times[ -1, A, 
          Plus[ -1, 
            Power[ c, 2]], 
          Plus[ 
            Power[ e, 2], 
            Times[ -2, M, r]], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ A, 2], 
                Power[ c, 2]], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(A\ \((\(-1\) \
+ c\^2)\)\ \((e\^2 - 2\ M\ r)\)\)\/\(A\^2\ c\^2 + r\^2\)\)\)\),
      SequenceForm[ "[", 4, ",", 3, "] = ", 
        Times[ -1, A, 
          Plus[ -1, 
            Power[ c, 2]], 
          Plus[ 
            Power[ e, 2], 
            Times[ -2, M, r]], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ A, 2], 
                Power[ c, 2]], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(A\^2\ c\^2 + 
                e\^2 + r\ \((\(-2\)\ M + r)\)\)\/\(A\^2\ c\^2 + r\^2\)\)\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ -1, 
          Power[ 
            Plus[ 
              Times[ 
                Power[ A, 2], 
                Power[ c, 2]], 
              Power[ r, 2]], -1], 
          Plus[ 
            Times[ 
              Power[ A, 2], 
              Power[ c, 2]], 
            Power[ e, 2], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], r]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]45\[InvisibleSpace]":"\[InvisibleSpace]28\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 45, ":", 28, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]45\[InvisibleSpace]":"\[InvisibleSpace]36\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 45, ":", 36, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(-\(\(1\/\((A\^2\ c\^2 + r\^2)\)\^6\)\((8\ \((\(-7\)\ A\^4\ c\^4\ e\^4 \
+ 6\ A\^6\ c\^6\ M\^2 + 60\ A\^4\ c\^4\ e\^2\ M\ r + 
              34\ A\^2\ c\^2\ e\^4\ r\^2 - 90\ A\^4\ c\^4\ M\^2\ r\^2 - 
              120\ A\^2\ c\^2\ e\^2\ M\ r\^3 - 7\ e\^4\ r\^4 + 
              90\ A\^2\ c\^2\ M\^2\ r\^4 + 12\ e\^2\ M\ r\^5 - 
              6\ M\^2\ r\^6)\))\)\)\)\)], "Output"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]46\[InvisibleSpace]":"\[InvisibleSpace]13\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 46, ":", 13, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(192\ A\ c\ \((A\^4\ c\^4\ M\ \((e\^2 - 3\ M\ r)\) + r\^3\ \((\(-2\)\ \
e\^4 + 5\ e\^2\ M\ r - 3\ M\^2\ r\^2)\) + 2\ A\^2\ c\^2\ r\ \((e\^4 - 5\ e\^2\
\ M\ r + 5\ M\^2\ r\^2)\))\)\)\/\((A\^2\ c\^2 + r\^2)\)\^6\)], "Output"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]46\[InvisibleSpace]":"\[InvisibleSpace]16\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 46, ":", 16, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(\(1\/\((A\^2\ c\^2 + r\^2)\)\^9\)\((16\ \((44\ A\^8\ c\^6\ e\^4 - 
            76\ A\^8\ c\^8\ e\^4 - 32\ A\^6\ c\^6\ e\^6 + 
            45\ A\^10\ c\^10\ M\^2 + 45\ A\^8\ c\^8\ e\^2\ M\^2 - 
            180\ A\^8\ c\^6\ e\^2\ M\ r + 720\ A\^8\ c\^8\ e\^2\ M\ r + 
            604\ A\^6\ c\^6\ e\^4\ M\ r - 90\ A\^8\ c\^8\ M\^3\ r - 
            156\ A\^6\ c\^4\ e\^4\ r\^2 + 712\ A\^6\ c\^6\ e\^4\ r\^2 + 
            588\ A\^4\ c\^4\ e\^6\ r\^2 - 1215\ A\^8\ c\^8\ M\^2\ r\^2 - 
            2340\ A\^6\ c\^6\ e\^2\ M\^2\ r\^2 + 
            180\ A\^6\ c\^4\ e\^2\ M\ r\^3 - 
            2700\ A\^6\ c\^6\ e\^2\ M\ r\^3 - 
            4236\ A\^4\ c\^4\ e\^4\ M\ r\^3 + 2520\ A\^6\ c\^6\ M\^3\ r\^3 - 
            156\ A\^4\ c\^2\ e\^4\ r\^4 - 744\ A\^2\ c\^2\ e\^6\ r\^4 + 
            1890\ A\^6\ c\^6\ M\^2\ r\^4 + 
            9270\ A\^4\ c\^4\ e\^2\ M\^2\ r\^4 + 
            324\ A\^4\ c\^2\ e\^2\ M\ r\^5 - 
            1332\ A\^4\ c\^4\ e\^2\ M\ r\^5 + 
            3540\ A\^2\ c\^2\ e\^4\ M\ r\^5 - 6300\ A\^4\ c\^4\ M\^3\ r\^5 + 
            44\ A\^2\ e\^4\ r\^6 - 712\ A\^2\ c\^2\ e\^4\ r\^6 + 
            76\ e\^6\ r\^6 + 1890\ A\^4\ c\^4\ M\^2\ r\^6 - 
            5364\ A\^2\ c\^2\ e\^2\ M\^2\ r\^6 - 36\ A\^2\ e\^2\ M\ r\^7 + 
            1980\ A\^2\ c\^2\ e\^2\ M\ r\^7 - 260\ e\^4\ M\ r\^7 + 
            2520\ A\^2\ c\^2\ M\^3\ r\^7 + 76\ e\^4\ r\^8 - 
            1215\ A\^2\ c\^2\ M\^2\ r\^8 + 261\ e\^2\ M\^2\ r\^8 - 
            108\ e\^2\ M\ r\^9 - 90\ M\^3\ r\^9 + 
            45\ M\^2\ r\^10)\))\)\)\)], "Output"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]49\[InvisibleSpace]":"\[InvisibleSpace]28\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 49, ":", 28, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(spacetime[xp, gp];\)\), "\n", 
    \(\(x\  = \ {r, c, ph, t};\)\), "\n", 
    \(\(g\  = \ 
        coordxform[gp /. {e \[Rule] 0}, {r, ArcCos[c], ph, t}, 
          x];\)\), "\n", 
    \(\(components[g];\)\), "\n", 
    \(\(spacetime[x, g];\)\), "\n", 
    \(\(timestamp[];\)\), "\n", 
    \(\(curvature[];\)\), "\n", 
    \(\(timestamp[];\)\), "\n", 
    \(Factor[kretschmann[]]\), "\[IndentingNewLine]", 
    \(\(timestamp[];\)\), "\[IndentingNewLine]", 
    \(\(irmdg\  = \ 
        Simplify[
          PowerExpand[
            1/Sqrt[\(-Simplify[
                    Det[currentst[\([2]\)]]]\)]]];\)\), \
"\[IndentingNewLine]", 
    \(\(rdual\  = \ 
        Table[Table[Table[Table[0, {l, 4}], {k, 4}], {j, 4}], {i, 
            4}];\)\), "\n", 
    \(\(Do[
        Do[Do[Do[
              rdual[\([i, j, k, l]\)] = 
                2  irmdg\ *\ 
                  Sum[Sum[
                      eps4[\([i, j, m, 
                            n]\)]\ *\n\t\t\t\t\t\t\t\t\tcurrentst[\([5, m, n, 
                            k, l]\)], {n, m + 1, 4}], {m, 4}], \n\t\t\t\t{l, 
                k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(\(Do[
        Do[Do[Do[
              rdual[\([j, i, k, 
                    l]\)]\  = \ \(-rdual[\([i, j, k, l]\)]\), \n\t\t\t\t{l, 
                k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(\(Do[
        Do[Do[Do[
              rdual[\([i, j, l, 
                    k]\)]\  = \ \(-rdual[\([i, j, k, l]\)]\), \n\t\t\t\t{l, 
                k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(\(Do[
        Do[Do[Do[
              rdual[\([j, i, l, k]\)]\  = \ 
                rdual[\([i, j, k, l]\)], \n\t\t\t\t{l, k + 1, 4}], {k, 
              4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(\(r2up\  = \ raise[raise[currentst[\([5]\)], 1], 2];\)\), "\n", 
    \(rstarr\  = \ 
      Simplify[Sum[
          Sum[Sum[Sum[
                4  rdual[\([i, j, k, l]\)] 
                  r2up[\([k, l, i, j]\)], \n\t\t\t\t\t\t{l, k + 1, 4}], {k, 
                4}], {j, i + 1, 4}], {i, 4}]]\), "\n", 
    \(\(timestamp[];\)\), "\n", 
    \(\(cdr\  = \ covder[currentst[\([5]\)]];\)\), "\n", 
    \(\(cdr5u\  = \ 
        raise[raise[raise[raise[raise[cdr, 1], 2], 3], 4], 5];\)\), "\n", 
    \(\(cdrsqun\  = \ 
        Sum[Sum[Sum[
              Sum[Sum[4  cdr[\([i, j, k, l, m]\)] 
                    cdr5u[\([i, j, k, l, m]\)], \n\t\t\t\t\t\t{m, 4}], {l, 
                  k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(Factor[cdrsqun]\), "\n", 
    \(\(timestamp[];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]\(A\^2\ c\^2 + r\^2\
\)\/\(A\^2 - 2\ M\ r + r\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Plus[ 
            Times[ 
              Power[ A, 2], 
              Power[ c, 2]], 
            Power[ r, 2]], 
          Power[ 
            Plus[ 
              Power[ A, 2], 
              Times[ -2, M, r], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]\(A\^2\ c\^2 + r\^2\
\)\/\(1 - c\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ 
            Plus[ 1, 
              Times[ -1, 
                Power[ c, 2]]], -1], 
          Plus[ 
            Times[ 
              Power[ A, 2], 
              Power[ c, 2]], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(\((\(-1\) + 
                    c\^2)\)\ \((A\^4\ c\^2 + r\^4 + 
                    A\^2\ r\ \((2\ M - 2\ c\^2\ M + r + 
                          c\^2\ r)\))\)\)\/\(A\^2\ c\^2 + r\^2\)\)\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ -1, 
          Plus[ -1, 
            Power[ c, 2]], 
          Power[ 
            Plus[ 
              Times[ 
                Power[ A, 2], 
                Power[ c, 2]], 
              Power[ r, 2]], -1], 
          Plus[ 
            Times[ 
              Power[ A, 4], 
              Power[ c, 2]], 
            Power[ r, 4], 
            Times[ 
              Power[ A, 2], r, 
              Plus[ 
                Times[ 2, M], 
                Times[ -2, 
                  Power[ c, 2], M], r, 
                Times[ 
                  Power[ c, 2], r]]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ A\ \((\(-1\) + \
c\^2)\)\ M\ r\)\/\(A\^2\ c\^2 + r\^2\)\),
      SequenceForm[ "[", 3, ",", 4, "] = ", 
        Times[ 2, A, 
          Plus[ -1, 
            Power[ c, 2]], M, r, 
          Power[ 
            Plus[ 
              Times[ 
                Power[ A, 2], 
                Power[ c, 2]], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(2\ A\ \((\(-1\) + \
c\^2)\)\ M\ r\)\/\(A\^2\ c\^2 + r\^2\)\),
      SequenceForm[ "[", 4, ",", 3, "] = ", 
        Times[ 2, A, 
          Plus[ -1, 
            Power[ c, 2]], M, r, 
          Power[ 
            Plus[ 
              Times[ 
                Power[ A, 2], 
                Power[ c, 2]], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(A\^2\ c\^2 + 
                r\ \((\(-2\)\ M + r)\)\)\/\(A\^2\ c\^2 + r\^2\)\)\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ -1, 
          Power[ 
            Plus[ 
              Times[ 
                Power[ A, 2], 
                Power[ c, 2]], 
              Power[ r, 2]], -1], 
          Plus[ 
            Times[ 
              Power[ A, 2], 
              Power[ c, 2]], 
            Times[ r, 
              Plus[ 
                Times[ -2, M], r]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]49\[InvisibleSpace]":"\[InvisibleSpace]34\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 49, ":", 34, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]49\[InvisibleSpace]":"\[InvisibleSpace]38\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 49, ":", 38, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(-\(\(48\ M\^2\ \((A\ c - r)\)\ \((A\ c + r)\)\ \((A\^2\ c\^2 - 
                4\ A\ c\ r + r\^2)\)\ \((A\^2\ c\^2 + 4\ A\ c\ r + 
                r\^2)\)\)\/\((A\^2\ c\^2 + r\^2)\)\^6\)\)\)], "Output"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]49\[InvisibleSpace]":"\[InvisibleSpace]55\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 49, ":", 55, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(-\(\(192\ A\ c\ M\^2\ r\ \((3\ A\^4\ c\^4 - 10\ A\^2\ c\^2\ r\^2 + 
                3\ r\^4)\)\)\/\((A\^2\ c\^2 + r\^2)\)\^6\)\)\)], "Output"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]49\[InvisibleSpace]":"\[InvisibleSpace]56\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 49, ":", 56, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(\(1\/\((A\^2\ c\^2 + r\^2)\)\^9\)\((720\ M\^2\ \((A\^2\ c\^2 - 
            2\ M\ r + r\^2)\)\ \((A\^4\ c\^4 + 4\ A\^3\ c\^3\ r - 
            6\ A\^2\ c\^2\ r\^2 - 4\ A\ c\ r\^3 + r\^4)\)\ \((A\^4\ c\^4 - 
            4\ A\^3\ c\^3\ r - 6\ A\^2\ c\^2\ r\^2 + 4\ A\ c\ r\^3 + 
            r\^4)\))\)\)\)], "Output"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]51\[InvisibleSpace]":"\[InvisibleSpace]1\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 51, ":", 1, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(spacetime[xp, gp];\)\), "\n", 
    \(\(x\  = \ {r, c, ph, t};\)\), "\n", 
    \(\(g\  = \ 
        coordxform[gp /. {A \[Rule] 0}, {r, ArcCos[c], ph, t}, 
          x];\)\), "\n", 
    \(\(components[g];\)\), "\n", 
    \(\(spacetime[x, g];\)\), "\n", 
    \(\(timestamp[];\)\), "\n", 
    \(\(curvature[];\)\), "\n", 
    \(\(timestamp[];\)\), "\n", 
    \(Factor[kretschmann[]]\), "\[IndentingNewLine]", 
    \(\(timestamp[];\)\), "\[IndentingNewLine]", 
    \(\(irmdg\  = \ 
        Simplify[
          PowerExpand[
            1/Sqrt[\(-Simplify[
                    Det[currentst[\([2]\)]]]\)]]];\)\), \
"\[IndentingNewLine]", 
    \(\(rdual\  = \ 
        Table[Table[Table[Table[0, {l, 4}], {k, 4}], {j, 4}], {i, 
            4}];\)\), "\n", 
    \(\(Do[
        Do[Do[Do[
              rdual[\([i, j, k, l]\)] = 
                2  irmdg\ *\ 
                  Sum[Sum[
                      eps4[\([i, j, m, 
                            n]\)]\ *\n\t\t\t\t\t\t\t\t\tcurrentst[\([5, m, n, 
                            k, l]\)], {n, m + 1, 4}], {m, 4}], \n\t\t\t\t{l, 
                k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(\(Do[
        Do[Do[Do[
              rdual[\([j, i, k, 
                    l]\)]\  = \ \(-rdual[\([i, j, k, l]\)]\), \n\t\t\t\t{l, 
                k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(\(Do[
        Do[Do[Do[
              rdual[\([i, j, l, 
                    k]\)]\  = \ \(-rdual[\([i, j, k, l]\)]\), \n\t\t\t\t{l, 
                k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(\(Do[
        Do[Do[Do[
              rdual[\([j, i, l, k]\)]\  = \ 
                rdual[\([i, j, k, l]\)], \n\t\t\t\t{l, k + 1, 4}], {k, 
              4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(\(r2up\  = \ raise[raise[currentst[\([5]\)], 1], 2];\)\), "\n", 
    \(rstarr\  = \ 
      Simplify[Sum[
          Sum[Sum[Sum[
                4  rdual[\([i, j, k, l]\)] 
                  r2up[\([k, l, i, j]\)], \n\t\t\t\t\t\t{l, k + 1, 4}], {k, 
                4}], {j, i + 1, 4}], {i, 4}]]\), "\n", 
    \(\(timestamp[];\)\), "\n", 
    \(\(cdr\  = \ covder[currentst[\([5]\)]];\)\), "\n", 
    \(\(cdr5u\  = \ 
        raise[raise[raise[raise[raise[cdr, 1], 2], 3], 4], 5];\)\), "\n", 
    \(\(cdrsqun\  = \ 
        Sum[Sum[Sum[
              Sum[Sum[4  cdr[\([i, j, k, l, m]\)] 
                    cdr5u[\([i, j, k, l, m]\)], \n\t\t\t\t\t\t{m, 4}], {l, 
                  k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(Factor[cdrsqun]\), "\n", 
    \(\(timestamp[];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]r\^2\/\(e\^2 - 2\ M\
\ r + r\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Plus[ 
              Power[ e, 2], 
              Times[ -2, M, r], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]r\^2\/\(1 - \
c\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ 
            Plus[ 1, 
              Times[ -1, 
                Power[ c, 2]]], -1], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\((\(-1\) + 
                c\^2)\)\)\ r\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ -1, 
          Plus[ -1, 
            Power[ c, 2]], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(-\(\(e\^2 - 
                2\ M\ r + r\^2\)\/r\^2\)\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ -1, 
          Power[ r, -2], 
          Plus[ 
            Power[ e, 2], 
            Times[ -2, M, r], 
            Power[ r, 2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]51\[InvisibleSpace]":"\[InvisibleSpace]2\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 51, ":", 2, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]51\[InvisibleSpace]":"\[InvisibleSpace]2\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 51, ":", 2, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(8\ \((7\ e\^4 - 12\ e\^2\ M\ r + 6\ M\^2\ r\^2)\)\)\/r\^8\)], "Output"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]51\[InvisibleSpace]":"\[InvisibleSpace]4\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 51, ":", 4, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    \(0\)], "Output"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]51\[InvisibleSpace]":"\[InvisibleSpace]4\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 51, ":", 4, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(-\(\(16\ \((\(-e\^2\) + 2\ M\ r - r\^2)\)\ \((76\ e\^4 - 
                108\ e\^2\ M\ r + 45\ M\^2\ r\^2)\)\)\/r\^12\)\)\)], "Output"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]51\[InvisibleSpace]":"\[InvisibleSpace]13\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 51, ":", 13, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(spacetime[xp, gp];\)\), "\n", 
    \(\(x\  = \ {r, c, ph, t};\)\), "\n", 
    \(\(g\  = \ 
        coordxform[gp /. {A \[Rule] 0, e \[Rule] 0}, {r, ArcCos[c], ph, t}, 
          x];\)\), "\n", 
    \(\(components[g];\)\), "\n", 
    \(\(spacetime[x, g];\)\), "\n", 
    \(\(timestamp[];\)\), "\n", 
    \(\(curvature[];\)\), "\n", 
    \(\(timestamp[];\)\), "\n", 
    \(Factor[kretschmann[]]\), "\[IndentingNewLine]", 
    \(\(timestamp[];\)\), "\[IndentingNewLine]", 
    \(\(irmdg\  = \ 
        Simplify[
          PowerExpand[
            1/Sqrt[\(-Simplify[
                    Det[currentst[\([2]\)]]]\)]]];\)\), \
"\[IndentingNewLine]", 
    \(\(rdual\  = \ 
        Table[Table[Table[Table[0, {l, 4}], {k, 4}], {j, 4}], {i, 
            4}];\)\), "\n", 
    \(\(Do[
        Do[Do[Do[
              rdual[\([i, j, k, l]\)] = 
                2  irmdg\ *\ 
                  Sum[Sum[
                      eps4[\([i, j, m, 
                            n]\)]\ *\n\t\t\t\t\t\t\t\t\tcurrentst[\([5, m, n, 
                            k, l]\)], {n, m + 1, 4}], {m, 4}], \n\t\t\t\t{l, 
                k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(\(Do[
        Do[Do[Do[
              rdual[\([j, i, k, 
                    l]\)]\  = \ \(-rdual[\([i, j, k, l]\)]\), \n\t\t\t\t{l, 
                k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(\(Do[
        Do[Do[Do[
              rdual[\([i, j, l, 
                    k]\)]\  = \ \(-rdual[\([i, j, k, l]\)]\), \n\t\t\t\t{l, 
                k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(\(Do[
        Do[Do[Do[
              rdual[\([j, i, l, k]\)]\  = \ 
                rdual[\([i, j, k, l]\)], \n\t\t\t\t{l, k + 1, 4}], {k, 
              4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(\(r2up\  = \ raise[raise[currentst[\([5]\)], 1], 2];\)\), "\n", 
    \(rstarr\  = \ 
      Simplify[Sum[
          Sum[Sum[Sum[
                4  rdual[\([i, j, k, l]\)] 
                  r2up[\([k, l, i, j]\)], \n\t\t\t\t\t\t{l, k + 1, 4}], {k, 
                4}], {j, i + 1, 4}], {i, 4}]]\), "\n", 
    \(\(timestamp[];\)\), "\n", 
    \(\(cdr\  = \ covder[currentst[\([5]\)]];\)\), "\n", 
    \(\(cdr5u\  = \ 
        raise[raise[raise[raise[raise[cdr, 1], 2], 3], 4], 5];\)\), "\n", 
    \(\(cdrsqun\  = \ 
        Sum[Sum[Sum[
              Sum[Sum[4  cdr[\([i, j, k, l, m]\)] 
                    cdr5u[\([i, j, k, l, m]\)], \n\t\t\t\t\t\t{m, 4}], {l, 
                  k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 4}];\)\), "\n", 
    \(Factor[cdrsqun]\), "\n", 
    \(\(timestamp[];\)\)}], "Input"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]r\/\(\(-2\)\ M + \
r\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ r, 
          Power[ 
            Plus[ 
              Times[ -2, M], r], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]r\^2\/\(1 - \
c\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ 
            Plus[ 1, 
              Times[ -1, 
                Power[ c, 2]]], -1], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\((\(-1\) + 
                c\^2)\)\)\ r\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ -1, 
          Plus[ -1, 
            Power[ c, 2]], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-1\) + \(2\ M\)\
\/r\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Plus[ -1, 
          Times[ 2, M, 
            Power[ r, -1]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]51\[InvisibleSpace]":"\[InvisibleSpace]14\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 51, ":", 14, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]51\[InvisibleSpace]":"\[InvisibleSpace]14\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 51, ":", 14, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(48\ M\^2\)\/r\^6\)], "Output"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]51\[InvisibleSpace]":"\[InvisibleSpace]15\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 51, ":", 15, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    \(0\)], "Output"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]51\[InvisibleSpace]":"\[InvisibleSpace]16\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 51, ":", 16, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(-\(\(720\ M\^2\ \((2\ M - r)\)\)\/r\^9\)\)\)], "Output"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]16\[InvisibleSpace]":"\
\[InvisibleSpace]51\[InvisibleSpace]":"\[InvisibleSpace]22\[InvisibleSpace]" \
on "\[InvisibleSpace]4\[InvisibleSpace]"/"\[InvisibleSpace]17\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 16, ":", 51, ":", 22, " on ", 4, "/", 17, "/", 2001, "."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[BoxData[
    \(\(NotebookSave[EvaluationNotebook[]];\)\)], "Input"]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
WindowToolbars->{},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"Magnification"->1},
PrivateNotebookOptions->{"ColorPalette"->{RGBColor, 128}},
ShowCellLabel->False,
ShowCellTags->False,
RenderingOptions->{"ObjectDithering"->True,
"RasterDithering"->False},
CharacterEncoding->"MacintoshAutomaticEncoding"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1717, 49, 59, 1, 27, "Input"],

Cell[CellGroupData[{
Cell[1801, 54, 64, 1, 27, "Input"],
Cell[1868, 57, 2970, 44, 375, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[4875, 106, 822, 16, 184, "Input"],
Cell[5700, 124, 612, 18, 42, "Print"],
Cell[6315, 144, 395, 11, 25, "Print"],
Cell[6713, 157, 979, 30, 47, "Print"],
Cell[7695, 189, 659, 19, 45, "Print"],
Cell[8357, 210, 659, 19, 45, "Print"],
Cell[9019, 231, 821, 25, 45, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[9877, 261, 2661, 63, 523, "Input"],
Cell[12541, 326, 582, 17, 42, "Print"],
Cell[13126, 345, 503, 15, 42, "Print"],
Cell[13632, 362, 1025, 31, 44, "Print"],
Cell[14660, 395, 609, 17, 42, "Print"],
Cell[15272, 414, 609, 17, 42, "Print"],
Cell[15884, 433, 699, 20, 42, "Print"],
Cell[16586, 455, 427, 7, 23, "Print"],
Cell[17016, 464, 427, 7, 23, "Print"],
Cell[17446, 473, 400, 6, 85, "Output"],
Cell[17849, 481, 427, 7, 23, "Print"],
Cell[18279, 490, 244, 3, 49, "Output"],
Cell[18526, 495, 427, 7, 23, "Print"],
Cell[18956, 504, 1563, 25, 193, "Output"],
Cell[20522, 531, 427, 7, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[20986, 543, 2616, 63, 507, "Input"],
Cell[23605, 608, 546, 16, 42, "Print"],
Cell[24154, 626, 503, 15, 42, "Print"],
Cell[24660, 643, 1005, 29, 42, "Print"],
Cell[25668, 674, 516, 14, 42, "Print"],
Cell[26187, 690, 516, 14, 42, "Print"],
Cell[26706, 706, 665, 19, 42, "Print"],
Cell[27374, 727, 427, 7, 23, "Print"],
Cell[27804, 736, 427, 7, 23, "Print"],
Cell[28234, 745, 225, 3, 49, "Output"],
Cell[28462, 750, 427, 7, 23, "Print"],
Cell[28892, 759, 164, 2, 49, "Output"],
Cell[29059, 763, 427, 7, 23, "Print"],
Cell[29489, 772, 337, 5, 67, "Output"],
Cell[29829, 779, 425, 7, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[30291, 791, 2616, 63, 507, "Input"],
Cell[32910, 856, 430, 12, 42, "Print"],
Cell[33343, 870, 387, 11, 42, "Print"],
Cell[33733, 883, 367, 9, 25, "Print"],
Cell[34103, 894, 428, 11, 42, "Print"],
Cell[34534, 907, 425, 7, 23, "Print"],
Cell[34962, 916, 425, 7, 23, "Print"],
Cell[35390, 925, 93, 1, 46, "Output"],
Cell[35486, 928, 425, 7, 23, "Print"],
Cell[35914, 937, 35, 1, 27, "Output"],
Cell[35952, 940, 425, 7, 23, "Print"],
Cell[36380, 949, 158, 2, 46, "Output"],
Cell[36541, 953, 427, 7, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[37005, 965, 2629, 63, 507, "Input"],
Cell[39637, 1030, 339, 9, 37, "Print"],
Cell[39979, 1041, 387, 11, 42, "Print"],
Cell[40369, 1054, 367, 9, 25, "Print"],
Cell[40739, 1065, 320, 8, 40, "Print"],
Cell[41062, 1075, 427, 7, 23, "Print"],
Cell[41492, 1084, 427, 7, 23, "Print"],
Cell[41922, 1093, 52, 1, 46, "Output"],
Cell[41977, 1096, 427, 7, 23, "Print"],
Cell[42407, 1105, 35, 1, 27, "Output"],
Cell[42445, 1108, 427, 7, 23, "Print"],
Cell[42875, 1117, 78, 1, 46, "Output"],
Cell[42956, 1120, 427, 7, 23, "Print"]
}, Open  ]],
Cell[43398, 1130, 72, 1, 27, "Input"]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

