(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     68040,       1956]*)
(*NotebookOutlinePosition[     68720,       1981]*)
(*  CellTagsIndexPosition[     68676,       1977]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell[BoxData[
    \(<< projects/mathematica/sugra.m\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Available functions are\n"\[InvisibleSpace]"   \
simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n"\
\[InvisibleSpace]"   lorentzmetric[d], components[tensor], \
raise/lower[tensor,index],\n"\[InvisibleSpace]"   norm[vector], \
twist[vector], coordxform[tensor,func,newx],\n"\[InvisibleSpace]"   \
sphericalmetric[d], connection[], covder[tensor], curvature[],\n"\
\[InvisibleSpace]"   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n"\
\[InvisibleSpace]"   expansion[vector], nullexpansion[vector], \
surfacearea[x],\n"\[InvisibleSpace]"   vielbein[e], checkvielbein[], \
lorentz[tensor,index],\n"\[InvisibleSpace]"   curved[tensor,index], \
lorentzraise[tensor,index], epsilon[d,indices],\n"\[InvisibleSpace]"   \
spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n"\[InvisibleSpace]"   \
crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n"\
\[InvisibleSpace]"Tensors are limited to rank 5.  All indices are lower (l) \
except:\n"\[InvisibleSpace]"   x[[upper (u)]], inverse metric[[u,u]], \
Christoffel[[u,l,l]],\n"\[InvisibleSpace]"   vielbein[[u,l]], inverse \
vielbein[[l,u]] & spin connection[[l,u,u]].\n \n"\[InvisibleSpace]"All \
indices are spacetime (s) except:\n"\[InvisibleSpace]"   (inverse) \
vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\n"\[InvisibleSpace]"  \
 all gamma matrix indices.\n \n"\[InvisibleSpace]"The last component in any \
index is the time component.\n \n"\[InvisibleSpace]"All computations are \
performed with currentst and currentframe."\),
      SequenceForm[ 
      "Available functions are\n", 
        "   simpler[expr], timestamp[], backup[], spacetime/fastst[x,g],\n", 
        "   lorentzmetric[d], components[tensor], raise/lower[tensor,index],\n\
", "   norm[vector], twist[vector], coordxform[tensor,func,newx],\n", 
        "   sphericalmetric[d], connection[], covder[tensor], curvature[],\n",
         "   ricci[], scalarcurv[], kretschmann[], surfaceg[],\n", 
        "   expansion[vector], nullexpansion[vector], surfacearea[x],\n", 
        "   vielbein[e], checkvielbein[], lorentz[tensor,index],\n", 
        "   curved[tensor,index], lorentzraise[tensor,index], \
epsilon[d,indices],\n", 
        "   spinconn[], gammamatrix[t,s], checkgamma[gamma,t,s],\n", 
        "   crossmatrix[m1,m2], splitgamma[g1,g2], antigamma[gamma,n].\n \n", 
        "Tensors are limited to rank 5.  All indices are lower (l) except:\n",
         "   x[[upper (u)]], inverse metric[[u,u]], Christoffel[[u,l,l]],\n", 
        "   vielbein[[u,l]], inverse vielbein[[l,u]] & spin \
connection[[l,u,u]].\n \n", "All indices are spacetime (s) except:\n", 
        "   (inverse) vielbein[[lorentz (l),s]], spin connection[[s,l,l]] and\
\n", "   all gamma matrix indices.\n \n", 
        "The last component in any index is the time component.\n \n", 
        "All computations are performed with currentst and currentframe."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Do[\((Print["\<dimension = \>", dim]; \[IndentingNewLine]gs\  = \ 
            sphericalmetric[
              dim - 1]; \[IndentingNewLine]xp\  = \ {r}; \
\[IndentingNewLine]Do[
            AppendTo[xp, Symbol["\<t\>" <> ToString[i]]], {i, 
              dim - 2}]; \[IndentingNewLine]AppendTo[xp, 
            t]; \[IndentingNewLine]gp\  = \ 
            IdentityMatrix[
              dim]; \[IndentingNewLine]gp[\([1, 
                1]\)]\  = \ \(-1\)/\((\(-1\) + 
                  2  M/\((r^\((dim - 3)\))\))\); \[IndentingNewLine]Do[
            gp[\([dim + 1 - i, dim + 1 - i]\)]\  = \ gs[\([i, i]\)], {i, 2, 
              dim - 1}]; \[IndentingNewLine]gp[\([dim, dim]\)]\  = \ \(-1\) + 
              2  M/\((r^\((dim - 3)\))\); \[IndentingNewLine]fastst[xp, 
            gp]; \[IndentingNewLine]x\  = \ 
            Flatten[{r, Take[{b, f, h, q, s, u, v, w}, dim - 3], t1, 
                t}]; \[IndentingNewLine]Print["\<metric solution in \
coordinates: \>", x]; \[IndentingNewLine]g\  = \ 
            coordxform[
              gp, \[IndentingNewLine]Flatten[{r, 
                  Take[{ArcCos[b], ArcCos[f], ArcCos[h], 
                      ArcCos[q], \[IndentingNewLine]ArcCos[s], ArcCos[u], 
                      ArcCos[v], ArcCos[w]}, dim - 3], t1, t}], 
              x]; \[IndentingNewLine]components[
            Simplify[
              g]]; \[IndentingNewLine]timestamp[]; \
\[IndentingNewLine]spacetime[x, 
            g]; \[IndentingNewLine]curvature[]; \
\[IndentingNewLine]timestamp[]; \[IndentingNewLine]Print["\<R^2 = \>", 
            Factor[kretschmann[]]]; \[IndentingNewLine]timestamp[]; \
\[IndentingNewLine]cdr\  = \ 
            covder[currentst[\([5]\)]]; \[IndentingNewLine]cdr5u\  = \ 
            raise[raise[raise[raise[raise[cdr, 1], 2], 3], 4], 
              5]; \[IndentingNewLine]cdrsqun\  = \ 
            Sum[Sum[Sum[
                  Sum[Sum[
                      4  cdr[\([i, j, k, l, m]\)] 
                        cdr5u[\([i, j, k, l, m]\)], \n\t\t\t\t\t\t{m, 
                        4}], {l, k + 1, 4}], {k, 4}], {j, i + 1, 4}], {i, 
                4}]; \[IndentingNewLine]Print["\<(DR)^2 = \>", 
            Factor[cdrsqun]]; \[IndentingNewLine]timestamp[];\
\[IndentingNewLine])\), {dim, 4, 6}];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("dimension = "\[InvisibleSpace]4\),
      SequenceForm[ "dimension = ", 4],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Spherical metric check: "\[InvisibleSpace]True\),
      SequenceForm[ "Spherical metric check: ", True],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("metric solution in coordinates: \
"\[InvisibleSpace]{r, b, t1, t}\),
      SequenceForm[ "metric solution in coordinates: ", {r, b, t1, t}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]r\/\(\(-2\)\ M + \
r\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ r, 
          Power[ 
            Plus[ 
              Times[ -2, M], r], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]r\^2\/\(1 - \
b\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ 
            Plus[ 1, 
              Times[ -1, 
                Power[ b, 2]]], -1], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\((\(-1\) + 
                b\^2)\)\)\ r\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ -1, 
          Plus[ -1, 
            Power[ b, 2]], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-1\) + \(2\ M\)\
\/r\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Plus[ -1, 
          Times[ 2, M, 
            Power[ r, -1]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]18\[InvisibleSpace]":"\
\[InvisibleSpace]30\[InvisibleSpace]":"\[InvisibleSpace]54\[InvisibleSpace]" \
on "\[InvisibleSpace]3\[InvisibleSpace]"/"\[InvisibleSpace]15\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 18, ":", 30, ":", 54, " on ", 3, "/", 15, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]18\[InvisibleSpace]":"\
\[InvisibleSpace]30\[InvisibleSpace]":"\[InvisibleSpace]55\[InvisibleSpace]" \
on "\[InvisibleSpace]3\[InvisibleSpace]"/"\[InvisibleSpace]15\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 18, ":", 30, ":", 55, " on ", 3, "/", 15, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("R^2 = "\[InvisibleSpace]\(48\ M\^2\)\/r\^6\),
      SequenceForm[ "R^2 = ", 
        Times[ 48, 
          Power[ M, 2], 
          Power[ r, -6]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]18\[InvisibleSpace]":"\
\[InvisibleSpace]30\[InvisibleSpace]":"\[InvisibleSpace]56\[InvisibleSpace]" \
on "\[InvisibleSpace]3\[InvisibleSpace]"/"\[InvisibleSpace]15\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 18, ":", 30, ":", 56, " on ", 3, "/", 15, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("(DR)^2 = "\[InvisibleSpace]\(-\(\(720\ M\^2\ \((2\ M \
- r)\)\)\/r\^9\)\)\),
      SequenceForm[ "(DR)^2 = ", 
        Times[ -720, 
          Power[ M, 2], 
          Plus[ 
            Times[ 2, M], 
            Times[ -1, r]], 
          Power[ r, -9]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]18\[InvisibleSpace]":"\
\[InvisibleSpace]31\[InvisibleSpace]":"\[InvisibleSpace]3\[InvisibleSpace]" \
on "\[InvisibleSpace]3\[InvisibleSpace]"/"\[InvisibleSpace]15\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 18, ":", 31, ":", 3, " on ", 3, "/", 15, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("dimension = "\[InvisibleSpace]5\),
      SequenceForm[ "dimension = ", 5],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Spherical metric check: "\[InvisibleSpace]True\),
      SequenceForm[ "Spherical metric check: ", True],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("metric solution in coordinates: \
"\[InvisibleSpace]{r, b, f, t1, t}\),
      SequenceForm[ "metric solution in coordinates: ", {r, b, f, t1, t}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]r\^2\/\(\(-2\)\ M + \
r\^2\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ r, 2], 
          Power[ 
            Plus[ 
              Times[ -2, M], 
              Power[ r, 2]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]r\^2\/\(1 - \
b\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ 
            Plus[ 1, 
              Times[ -1, 
                Power[ b, 2]]], -1], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]r\^2\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Power[ r, 2]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\((\(-1\) + 
              b\^2)\)\ \((\(-1\) + f\^2)\)\ r\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ 
          Plus[ -1, 
            Power[ b, 2]], 
          Plus[ -1, 
            Power[ f, 2]], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-1\) + \(2\ M\)\
\/r\^2\)\),
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Plus[ -1, 
          Times[ 2, M, 
            Power[ r, -2]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]18\[InvisibleSpace]":"\
\[InvisibleSpace]31\[InvisibleSpace]":"\[InvisibleSpace]3\[InvisibleSpace]" \
on "\[InvisibleSpace]3\[InvisibleSpace]"/"\[InvisibleSpace]15\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 18, ":", 31, ":", 3, " on ", 3, "/", 15, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]18\[InvisibleSpace]":"\
\[InvisibleSpace]31\[InvisibleSpace]":"\[InvisibleSpace]4\[InvisibleSpace]" \
on "\[InvisibleSpace]3\[InvisibleSpace]"/"\[InvisibleSpace]15\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 18, ":", 31, ":", 4, " on ", 3, "/", 15, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("R^2 = "\[InvisibleSpace]\((4\ \((\(-72\)\ M\^2 + 
                  72\ b\^2\ M\^2 + 288\ f\^2\ M\^2 - 288\ b\^2\ f\^2\ M\^2 - 
                  432\ f\^4\ M\^2 + 432\ b\^2\ f\^4\ M\^2 + 288\ f\^6\ M\^2 - 
                  288\ b\^2\ f\^6\ M\^2 - 72\ f\^8\ M\^2 + 
                  72\ b\^2\ f\^8\ M\^2 + 4\ M\ r\^2 - 4\ b\^2\ M\ r\^2 - 
                  24\ f\^2\ M\ r\^2 + 24\ b\^2\ f\^2\ M\ r\^2 + 
                  44\ f\^4\ M\ r\^2 - 44\ b\^2\ f\^4\ M\ r\^2 - 
                  32\ f\^6\ M\ r\^2 + 32\ b\^2\ f\^6\ M\ r\^2 + 
                  8\ f\^8\ M\ r\^2 - 8\ b\^2\ f\^8\ M\ r\^2 - r\^4 + 
                  b\^2\ r\^4 + 4\ f\^2\ r\^4 - 6\ b\^2\ f\^2\ r\^4 - 
                  10\ f\^4\ r\^4 + 14\ b\^2\ f\^4\ r\^4 + 8\ f\^6\ r\^4 - 
                  10\ b\^2\ f\^6\ r\^4 - 2\ f\^8\ r\^4 + 
                  2\ b\^2\ f\^8\ r\^4)\))\)/\((\((\(-1\) + b)\)\ \((1 + 
                  b)\)\ \((\(-1\) + f)\)\^4\ \((1 + f)\)\^4\ r\^8)\)\),
      SequenceForm[ "R^2 = ", 
        Times[ 4, 
          Power[ 
            Plus[ -1, b], -1], 
          Power[ 
            Plus[ 1, b], -1], 
          Power[ 
            Plus[ -1, f], -4], 
          Power[ 
            Plus[ 1, f], -4], 
          Power[ r, -8], 
          Plus[ 
            Times[ -72, 
              Power[ M, 2]], 
            Times[ 72, 
              Power[ b, 2], 
              Power[ M, 2]], 
            Times[ 288, 
              Power[ f, 2], 
              Power[ M, 2]], 
            Times[ -288, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ M, 2]], 
            Times[ -432, 
              Power[ f, 4], 
              Power[ M, 2]], 
            Times[ 432, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ M, 2]], 
            Times[ 288, 
              Power[ f, 6], 
              Power[ M, 2]], 
            Times[ -288, 
              Power[ b, 2], 
              Power[ f, 6], 
              Power[ M, 2]], 
            Times[ -72, 
              Power[ f, 8], 
              Power[ M, 2]], 
            Times[ 72, 
              Power[ b, 2], 
              Power[ f, 8], 
              Power[ M, 2]], 
            Times[ 4, M, 
              Power[ r, 2]], 
            Times[ -4, 
              Power[ b, 2], M, 
              Power[ r, 2]], 
            Times[ -24, 
              Power[ f, 2], M, 
              Power[ r, 2]], 
            Times[ 24, 
              Power[ b, 2], 
              Power[ f, 2], M, 
              Power[ r, 2]], 
            Times[ 44, 
              Power[ f, 4], M, 
              Power[ r, 2]], 
            Times[ -44, 
              Power[ b, 2], 
              Power[ f, 4], M, 
              Power[ r, 2]], 
            Times[ -32, 
              Power[ f, 6], M, 
              Power[ r, 2]], 
            Times[ 32, 
              Power[ b, 2], 
              Power[ f, 6], M, 
              Power[ r, 2]], 
            Times[ 8, 
              Power[ f, 8], M, 
              Power[ r, 2]], 
            Times[ -8, 
              Power[ b, 2], 
              Power[ f, 8], M, 
              Power[ r, 2]], 
            Times[ -1, 
              Power[ r, 4]], 
            Times[ 
              Power[ b, 2], 
              Power[ r, 4]], 
            Times[ 4, 
              Power[ f, 2], 
              Power[ r, 4]], 
            Times[ -6, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ r, 4]], 
            Times[ -10, 
              Power[ f, 4], 
              Power[ r, 4]], 
            Times[ 14, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ r, 4]], 
            Times[ 8, 
              Power[ f, 6], 
              Power[ r, 4]], 
            Times[ -10, 
              Power[ b, 2], 
              Power[ f, 6], 
              Power[ r, 4]], 
            Times[ -2, 
              Power[ f, 8], 
              Power[ r, 4]], 
            Times[ 2, 
              Power[ b, 2], 
              Power[ f, 8], 
              Power[ r, 4]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]18\[InvisibleSpace]":"\
\[InvisibleSpace]31\[InvisibleSpace]":"\[InvisibleSpace]10\[InvisibleSpace]" \
on "\[InvisibleSpace]3\[InvisibleSpace]"/"\[InvisibleSpace]15\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 18, ":", 31, ":", 10, " on ", 3, "/", 15, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("(DR)^2 = "\[InvisibleSpace]\(-\(\((16\ \((288\ M\^3 \
- 576\ b\^2\ M\^3 + 288\ b\^4\ M\^3 - 1728\ f\^2\ M\^3 + 
                    3456\ b\^2\ f\^2\ M\^3 - 1728\ b\^4\ f\^2\ M\^3 + 
                    4320\ f\^4\ M\^3 - 8640\ b\^2\ f\^4\ M\^3 + 
                    4320\ b\^4\ f\^4\ M\^3 - 5760\ f\^6\ M\^3 + 
                    11520\ b\^2\ f\^6\ M\^3 - 5760\ b\^4\ f\^6\ M\^3 + 
                    4320\ f\^8\ M\^3 - 8640\ b\^2\ f\^8\ M\^3 + 
                    4320\ b\^4\ f\^8\ M\^3 - 1728\ f\^10\ M\^3 + 
                    3456\ b\^2\ f\^10\ M\^3 - 1728\ b\^4\ f\^10\ M\^3 + 
                    288\ f\^12\ M\^3 - 576\ b\^2\ f\^12\ M\^3 + 
                    288\ b\^4\ f\^12\ M\^3 - 176\ M\^2\ r\^2 + 
                    352\ b\^2\ M\^2\ r\^2 - 176\ b\^4\ M\^2\ r\^2 + 
                    1120\ f\^2\ M\^2\ r\^2 - 2240\ b\^2\ f\^2\ M\^2\ r\^2 + 
                    1120\ b\^4\ f\^2\ M\^2\ r\^2 - 2928\ f\^4\ M\^2\ r\^2 + 
                    5856\ b\^2\ f\^4\ M\^2\ r\^2 - 
                    2928\ b\^4\ f\^4\ M\^2\ r\^2 + 4032\ f\^6\ M\^2\ r\^2 - 
                    8064\ b\^2\ f\^6\ M\^2\ r\^2 + 
                    4032\ b\^4\ f\^6\ M\^2\ r\^2 - 3088\ f\^8\ M\^2\ r\^2 + 
                    6176\ b\^2\ f\^8\ M\^2\ r\^2 - 
                    3088\ b\^4\ f\^8\ M\^2\ r\^2 + 1248\ f\^10\ M\^2\ r\^2 - 
                    2496\ b\^2\ f\^10\ M\^2\ r\^2 + 
                    1248\ b\^4\ f\^10\ M\^2\ r\^2 - 208\ f\^12\ M\^2\ r\^2 + 
                    416\ b\^2\ f\^12\ M\^2\ r\^2 - 
                    208\ b\^4\ f\^12\ M\^2\ r\^2 + 20\ M\ r\^4 - 
                    40\ b\^2\ M\ r\^4 + 20\ b\^4\ M\ r\^4 - 
                    152\ f\^2\ M\ r\^4 + 312\ b\^2\ f\^2\ M\ r\^4 - 
                    160\ b\^4\ f\^2\ M\ r\^4 + 460\ f\^4\ M\ r\^4 - 
                    952\ b\^2\ f\^4\ M\ r\^4 + 492\ b\^4\ f\^4\ M\ r\^4 - 
                    704\ f\^6\ M\ r\^4 + 1456\ b\^2\ f\^6\ M\ r\^4 - 
                    752\ b\^4\ f\^6\ M\ r\^4 + 576\ f\^8\ M\ r\^4 - 
                    1184\ b\^2\ f\^8\ M\ r\^4 + 608\ b\^4\ f\^8\ M\ r\^4 - 
                    240\ f\^10\ M\ r\^4 + 488\ b\^2\ f\^10\ M\ r\^4 - 
                    248\ b\^4\ f\^10\ M\ r\^4 + 40\ f\^12\ M\ r\^4 - 
                    80\ b\^2\ f\^12\ M\ r\^4 + 40\ b\^4\ f\^12\ M\ r\^4 - 
                    2\ r\^6 + 3\ b\^2\ r\^6 - b\^4\ r\^6 + 7\ f\^2\ r\^6 - 
                    20\ b\^2\ f\^2\ r\^6 + 12\ b\^4\ f\^2\ r\^6 - 
                    34\ f\^4\ r\^6 + 94\ b\^2\ f\^4\ r\^6 - 
                    56\ b\^4\ f\^4\ r\^6 + 58\ f\^6\ r\^6 - 
                    152\ b\^2\ f\^6\ r\^6 + 88\ b\^4\ f\^6\ r\^6 - 
                    52\ f\^8\ r\^6 + 127\ b\^2\ f\^8\ r\^6 - 
                    71\ b\^4\ f\^8\ r\^6 + 23\ f\^10\ r\^6 - 
                    52\ b\^2\ f\^10\ r\^6 + 28\ b\^4\ f\^10\ r\^6 - 
                    4\ f\^12\ r\^6 + 8\ b\^2\ f\^12\ r\^6 - 
                    4\ b\^4\ f\^12\ r\^6)\))\)/\((\((\(-1\) + b)\)\^2\ \((1 + \
b)\)\^2\ \((\(-1\) + f)\)\^6\ \((1 + f)\)\^6\ r\^12)\)\)\)\),
      SequenceForm[ "(DR)^2 = ", 
        Times[ -16, 
          Power[ 
            Plus[ -1, b], -2], 
          Power[ 
            Plus[ 1, b], -2], 
          Power[ 
            Plus[ -1, f], -6], 
          Power[ 
            Plus[ 1, f], -6], 
          Power[ r, -12], 
          Plus[ 
            Times[ 288, 
              Power[ M, 3]], 
            Times[ -576, 
              Power[ b, 2], 
              Power[ M, 3]], 
            Times[ 288, 
              Power[ b, 4], 
              Power[ M, 3]], 
            Times[ -1728, 
              Power[ f, 2], 
              Power[ M, 3]], 
            Times[ 3456, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ M, 3]], 
            Times[ -1728, 
              Power[ b, 4], 
              Power[ f, 2], 
              Power[ M, 3]], 
            Times[ 4320, 
              Power[ f, 4], 
              Power[ M, 3]], 
            Times[ -8640, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ M, 3]], 
            Times[ 4320, 
              Power[ b, 4], 
              Power[ f, 4], 
              Power[ M, 3]], 
            Times[ -5760, 
              Power[ f, 6], 
              Power[ M, 3]], 
            Times[ 11520, 
              Power[ b, 2], 
              Power[ f, 6], 
              Power[ M, 3]], 
            Times[ -5760, 
              Power[ b, 4], 
              Power[ f, 6], 
              Power[ M, 3]], 
            Times[ 4320, 
              Power[ f, 8], 
              Power[ M, 3]], 
            Times[ -8640, 
              Power[ b, 2], 
              Power[ f, 8], 
              Power[ M, 3]], 
            Times[ 4320, 
              Power[ b, 4], 
              Power[ f, 8], 
              Power[ M, 3]], 
            Times[ -1728, 
              Power[ f, 10], 
              Power[ M, 3]], 
            Times[ 3456, 
              Power[ b, 2], 
              Power[ f, 10], 
              Power[ M, 3]], 
            Times[ -1728, 
              Power[ b, 4], 
              Power[ f, 10], 
              Power[ M, 3]], 
            Times[ 288, 
              Power[ f, 12], 
              Power[ M, 3]], 
            Times[ -576, 
              Power[ b, 2], 
              Power[ f, 12], 
              Power[ M, 3]], 
            Times[ 288, 
              Power[ b, 4], 
              Power[ f, 12], 
              Power[ M, 3]], 
            Times[ -176, 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ 352, 
              Power[ b, 2], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ -176, 
              Power[ b, 4], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ 1120, 
              Power[ f, 2], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ -2240, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ 1120, 
              Power[ b, 4], 
              Power[ f, 2], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ -2928, 
              Power[ f, 4], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ 5856, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ -2928, 
              Power[ b, 4], 
              Power[ f, 4], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ 4032, 
              Power[ f, 6], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ -8064, 
              Power[ b, 2], 
              Power[ f, 6], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ 4032, 
              Power[ b, 4], 
              Power[ f, 6], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ -3088, 
              Power[ f, 8], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ 6176, 
              Power[ b, 2], 
              Power[ f, 8], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ -3088, 
              Power[ b, 4], 
              Power[ f, 8], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ 1248, 
              Power[ f, 10], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ -2496, 
              Power[ b, 2], 
              Power[ f, 10], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ 1248, 
              Power[ b, 4], 
              Power[ f, 10], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ -208, 
              Power[ f, 12], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ 416, 
              Power[ b, 2], 
              Power[ f, 12], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ -208, 
              Power[ b, 4], 
              Power[ f, 12], 
              Power[ M, 2], 
              Power[ r, 2]], 
            Times[ 20, M, 
              Power[ r, 4]], 
            Times[ -40, 
              Power[ b, 2], M, 
              Power[ r, 4]], 
            Times[ 20, 
              Power[ b, 4], M, 
              Power[ r, 4]], 
            Times[ -152, 
              Power[ f, 2], M, 
              Power[ r, 4]], 
            Times[ 312, 
              Power[ b, 2], 
              Power[ f, 2], M, 
              Power[ r, 4]], 
            Times[ -160, 
              Power[ b, 4], 
              Power[ f, 2], M, 
              Power[ r, 4]], 
            Times[ 460, 
              Power[ f, 4], M, 
              Power[ r, 4]], 
            Times[ -952, 
              Power[ b, 2], 
              Power[ f, 4], M, 
              Power[ r, 4]], 
            Times[ 492, 
              Power[ b, 4], 
              Power[ f, 4], M, 
              Power[ r, 4]], 
            Times[ -704, 
              Power[ f, 6], M, 
              Power[ r, 4]], 
            Times[ 1456, 
              Power[ b, 2], 
              Power[ f, 6], M, 
              Power[ r, 4]], 
            Times[ -752, 
              Power[ b, 4], 
              Power[ f, 6], M, 
              Power[ r, 4]], 
            Times[ 576, 
              Power[ f, 8], M, 
              Power[ r, 4]], 
            Times[ -1184, 
              Power[ b, 2], 
              Power[ f, 8], M, 
              Power[ r, 4]], 
            Times[ 608, 
              Power[ b, 4], 
              Power[ f, 8], M, 
              Power[ r, 4]], 
            Times[ -240, 
              Power[ f, 10], M, 
              Power[ r, 4]], 
            Times[ 488, 
              Power[ b, 2], 
              Power[ f, 10], M, 
              Power[ r, 4]], 
            Times[ -248, 
              Power[ b, 4], 
              Power[ f, 10], M, 
              Power[ r, 4]], 
            Times[ 40, 
              Power[ f, 12], M, 
              Power[ r, 4]], 
            Times[ -80, 
              Power[ b, 2], 
              Power[ f, 12], M, 
              Power[ r, 4]], 
            Times[ 40, 
              Power[ b, 4], 
              Power[ f, 12], M, 
              Power[ r, 4]], 
            Times[ -2, 
              Power[ r, 6]], 
            Times[ 3, 
              Power[ b, 2], 
              Power[ r, 6]], 
            Times[ -1, 
              Power[ b, 4], 
              Power[ r, 6]], 
            Times[ 7, 
              Power[ f, 2], 
              Power[ r, 6]], 
            Times[ -20, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ r, 6]], 
            Times[ 12, 
              Power[ b, 4], 
              Power[ f, 2], 
              Power[ r, 6]], 
            Times[ -34, 
              Power[ f, 4], 
              Power[ r, 6]], 
            Times[ 94, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ r, 6]], 
            Times[ -56, 
              Power[ b, 4], 
              Power[ f, 4], 
              Power[ r, 6]], 
            Times[ 58, 
              Power[ f, 6], 
              Power[ r, 6]], 
            Times[ -152, 
              Power[ b, 2], 
              Power[ f, 6], 
              Power[ r, 6]], 
            Times[ 88, 
              Power[ b, 4], 
              Power[ f, 6], 
              Power[ r, 6]], 
            Times[ -52, 
              Power[ f, 8], 
              Power[ r, 6]], 
            Times[ 127, 
              Power[ b, 2], 
              Power[ f, 8], 
              Power[ r, 6]], 
            Times[ -71, 
              Power[ b, 4], 
              Power[ f, 8], 
              Power[ r, 6]], 
            Times[ 23, 
              Power[ f, 10], 
              Power[ r, 6]], 
            Times[ -52, 
              Power[ b, 2], 
              Power[ f, 10], 
              Power[ r, 6]], 
            Times[ 28, 
              Power[ b, 4], 
              Power[ f, 10], 
              Power[ r, 6]], 
            Times[ -4, 
              Power[ f, 12], 
              Power[ r, 6]], 
            Times[ 8, 
              Power[ b, 2], 
              Power[ f, 12], 
              Power[ r, 6]], 
            Times[ -4, 
              Power[ b, 4], 
              Power[ f, 12], 
              Power[ r, 6]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]18\[InvisibleSpace]":"\
\[InvisibleSpace]31\[InvisibleSpace]":"\[InvisibleSpace]40\[InvisibleSpace]" \
on "\[InvisibleSpace]3\[InvisibleSpace]"/"\[InvisibleSpace]15\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 18, ":", 31, ":", 40, " on ", 3, "/", 15, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("dimension = "\[InvisibleSpace]6\),
      SequenceForm[ "dimension = ", 6],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Spherical metric check: "\[InvisibleSpace]True\),
      SequenceForm[ "Spherical metric check: ", True],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("metric solution in coordinates: \
"\[InvisibleSpace]{r, b, f, h, t1, t}\),
      SequenceForm[ "metric solution in coordinates: ", {r, b, f, h, t1, t}],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]1\[InvisibleSpace]","\
\[InvisibleSpace]1\[InvisibleSpace]"] = "\[InvisibleSpace]r\^3\/\(\(-2\)\ M + \
r\^3\)\),
      SequenceForm[ "[", 1, ",", 1, "] = ", 
        Times[ 
          Power[ r, 3], 
          Power[ 
            Plus[ 
              Times[ -2, M], 
              Power[ r, 3]], -1]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]2\[InvisibleSpace]","\
\[InvisibleSpace]2\[InvisibleSpace]"] = "\[InvisibleSpace]r\^2\/\(1 - \
b\^2\)\),
      SequenceForm[ "[", 2, ",", 2, "] = ", 
        Times[ 
          Power[ 
            Plus[ 1, 
              Times[ -1, 
                Power[ b, 2]]], -1], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]3\[InvisibleSpace]","\
\[InvisibleSpace]3\[InvisibleSpace]"] = "\[InvisibleSpace]\(\((\(-1\) + h\^2)\
\)\ r\^2\)\/\(\(-1\) + f\^2\)\),
      SequenceForm[ "[", 3, ",", 3, "] = ", 
        Times[ 
          Power[ 
            Plus[ -1, 
              Power[ f, 2]], -1], 
          Plus[ -1, 
            Power[ h, 2]], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]4\[InvisibleSpace]","\
\[InvisibleSpace]4\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\((\(-1\) + 
                f\^2)\)\)\ r\^2\)\),
      SequenceForm[ "[", 4, ",", 4, "] = ", 
        Times[ -1, 
          Plus[ -1, 
            Power[ f, 2]], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]5\[InvisibleSpace]","\
\[InvisibleSpace]5\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-\((\(-1\) + 
                b\^2)\)\)\ \((\(-1\) + f\^2)\)\ \((\(-1\) + h\^2)\)\ r\^2\)\),
      
      SequenceForm[ "[", 5, ",", 5, "] = ", 
        Times[ -1, 
          Plus[ -1, 
            Power[ b, 2]], 
          Plus[ -1, 
            Power[ f, 2]], 
          Plus[ -1, 
            Power[ h, 2]], 
          Power[ r, 2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("["\[InvisibleSpace]6\[InvisibleSpace]","\
\[InvisibleSpace]6\[InvisibleSpace]"] = "\[InvisibleSpace]\(\(-1\) + \(2\ M\)\
\/r\^3\)\),
      SequenceForm[ "[", 6, ",", 6, "] = ", 
        Plus[ -1, 
          Times[ 2, M, 
            Power[ r, -3]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]18\[InvisibleSpace]":"\
\[InvisibleSpace]31\[InvisibleSpace]":"\[InvisibleSpace]41\[InvisibleSpace]" \
on "\[InvisibleSpace]3\[InvisibleSpace]"/"\[InvisibleSpace]15\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 18, ":", 31, ":", 41, " on ", 3, "/", 15, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]18\[InvisibleSpace]":"\
\[InvisibleSpace]31\[InvisibleSpace]":"\[InvisibleSpace]44\[InvisibleSpace]" \
on "\[InvisibleSpace]3\[InvisibleSpace]"/"\[InvisibleSpace]15\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 18, ":", 31, ":", 44, " on ", 3, "/", 15, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("R^2 = "\[InvisibleSpace]\((4\ \((\(-240\)\ M\^2 + 
                  240\ b\^2\ M\^2 + 480\ f\^2\ M\^2 - 480\ b\^2\ f\^2\ M\^2 - 
                  240\ f\^4\ M\^2 + 240\ b\^2\ f\^4\ M\^2 + 960\ h\^2\ M\^2 - 
                  960\ b\^2\ h\^2\ M\^2 - 1920\ f\^2\ h\^2\ M\^2 + 
                  1920\ b\^2\ f\^2\ h\^2\ M\^2 + 960\ f\^4\ h\^2\ M\^2 - 
                  960\ b\^2\ f\^4\ h\^2\ M\^2 - 1440\ h\^4\ M\^2 + 
                  1440\ b\^2\ h\^4\ M\^2 + 2880\ f\^2\ h\^4\ M\^2 - 
                  2880\ b\^2\ f\^2\ h\^4\ M\^2 - 1440\ f\^4\ h\^4\ M\^2 + 
                  1440\ b\^2\ f\^4\ h\^4\ M\^2 + 960\ h\^6\ M\^2 - 
                  960\ b\^2\ h\^6\ M\^2 - 1920\ f\^2\ h\^6\ M\^2 + 
                  1920\ b\^2\ f\^2\ h\^6\ M\^2 + 960\ f\^4\ h\^6\ M\^2 - 
                  960\ b\^2\ f\^4\ h\^6\ M\^2 - 240\ h\^8\ M\^2 + 
                  240\ b\^2\ h\^8\ M\^2 + 480\ f\^2\ h\^8\ M\^2 - 
                  480\ b\^2\ f\^2\ h\^8\ M\^2 - 240\ f\^4\ h\^8\ M\^2 + 
                  240\ b\^2\ f\^4\ h\^8\ M\^2 + 4\ M\ r\^3 - 
                  4\ b\^2\ M\ r\^3 - 12\ f\^2\ M\ r\^3 + 
                  12\ b\^2\ f\^2\ M\ r\^3 + 8\ f\^4\ M\ r\^3 - 
                  8\ b\^2\ f\^4\ M\ r\^3 - 36\ h\^2\ M\ r\^3 + 
                  36\ b\^2\ h\^2\ M\ r\^3 + 80\ f\^2\ h\^2\ M\ r\^3 - 
                  80\ b\^2\ f\^2\ h\^2\ M\ r\^3 - 44\ f\^4\ h\^2\ M\ r\^3 + 
                  44\ b\^2\ f\^4\ h\^2\ M\ r\^3 + 80\ h\^4\ M\ r\^3 - 
                  80\ b\^2\ h\^4\ M\ r\^3 - 164\ f\^2\ h\^4\ M\ r\^3 + 
                  164\ b\^2\ f\^2\ h\^4\ M\ r\^3 + 84\ f\^4\ h\^4\ M\ r\^3 - 
                  84\ b\^2\ f\^4\ h\^4\ M\ r\^3 - 68\ h\^6\ M\ r\^3 + 
                  68\ b\^2\ h\^6\ M\ r\^3 + 136\ f\^2\ h\^6\ M\ r\^3 - 
                  136\ b\^2\ f\^2\ h\^6\ M\ r\^3 - 68\ f\^4\ h\^6\ M\ r\^3 + 
                  68\ b\^2\ f\^4\ h\^6\ M\ r\^3 + 20\ h\^8\ M\ r\^3 - 
                  20\ b\^2\ h\^8\ M\ r\^3 - 40\ f\^2\ h\^8\ M\ r\^3 + 
                  40\ b\^2\ f\^2\ h\^8\ M\ r\^3 + 20\ f\^4\ h\^8\ M\ r\^3 - 
                  20\ b\^2\ f\^4\ h\^8\ M\ r\^3 - 3\ r\^6 + 3\ b\^2\ r\^6 + 
                  4\ f\^2\ r\^6 - 6\ b\^2\ f\^2\ r\^6 - 2\ f\^4\ r\^6 + 
                  4\ b\^2\ f\^4\ r\^6 + 6\ h\^2\ r\^6 - 8\ b\^2\ h\^2\ r\^6 - 
                  16\ f\^2\ h\^2\ r\^6 + 24\ b\^2\ f\^2\ h\^2\ r\^6 + 
                  8\ f\^4\ h\^2\ r\^6 - 14\ b\^2\ f\^4\ h\^2\ r\^6 - 
                  15\ h\^4\ r\^6 + 19\ b\^2\ h\^4\ r\^6 + 
                  30\ f\^2\ h\^4\ r\^6 - 40\ b\^2\ f\^2\ h\^4\ r\^6 - 
                  15\ f\^4\ h\^4\ r\^6 + 21\ b\^2\ f\^4\ h\^4\ r\^6 + 
                  14\ h\^6\ r\^6 - 16\ b\^2\ h\^6\ r\^6 - 
                  28\ f\^2\ h\^6\ r\^6 + 32\ b\^2\ f\^2\ h\^6\ r\^6 + 
                  14\ f\^4\ h\^6\ r\^6 - 16\ b\^2\ f\^4\ h\^6\ r\^6 - 
                  5\ h\^8\ r\^6 + 5\ b\^2\ h\^8\ r\^6 + 
                  10\ f\^2\ h\^8\ r\^6 - 10\ b\^2\ f\^2\ h\^8\ r\^6 - 
                  5\ f\^4\ h\^8\ r\^6 + 
                  5\ b\^2\ f\^4\ h\^8\ r\^6)\))\)/\((\((\(-1\) + b)\)\ \((1 + 
                  b)\)\ \((\(-1\) + f)\)\^2\ \((1 + f)\)\^2\ \((\(-1\) + h)\)\
\^4\ \((1 + h)\)\^4\ r\^10)\)\),
      SequenceForm[ "R^2 = ", 
        Times[ 4, 
          Power[ 
            Plus[ -1, b], -1], 
          Power[ 
            Plus[ 1, b], -1], 
          Power[ 
            Plus[ -1, f], -2], 
          Power[ 
            Plus[ 1, f], -2], 
          Power[ 
            Plus[ -1, h], -4], 
          Power[ 
            Plus[ 1, h], -4], 
          Power[ r, -10], 
          Plus[ 
            Times[ -240, 
              Power[ M, 2]], 
            Times[ 240, 
              Power[ b, 2], 
              Power[ M, 2]], 
            Times[ 480, 
              Power[ f, 2], 
              Power[ M, 2]], 
            Times[ -480, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ M, 2]], 
            Times[ -240, 
              Power[ f, 4], 
              Power[ M, 2]], 
            Times[ 240, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ M, 2]], 
            Times[ 960, 
              Power[ h, 2], 
              Power[ M, 2]], 
            Times[ -960, 
              Power[ b, 2], 
              Power[ h, 2], 
              Power[ M, 2]], 
            Times[ -1920, 
              Power[ f, 2], 
              Power[ h, 2], 
              Power[ M, 2]], 
            Times[ 1920, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ h, 2], 
              Power[ M, 2]], 
            Times[ 960, 
              Power[ f, 4], 
              Power[ h, 2], 
              Power[ M, 2]], 
            Times[ -960, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ h, 2], 
              Power[ M, 2]], 
            Times[ -1440, 
              Power[ h, 4], 
              Power[ M, 2]], 
            Times[ 1440, 
              Power[ b, 2], 
              Power[ h, 4], 
              Power[ M, 2]], 
            Times[ 2880, 
              Power[ f, 2], 
              Power[ h, 4], 
              Power[ M, 2]], 
            Times[ -2880, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ h, 4], 
              Power[ M, 2]], 
            Times[ -1440, 
              Power[ f, 4], 
              Power[ h, 4], 
              Power[ M, 2]], 
            Times[ 1440, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ h, 4], 
              Power[ M, 2]], 
            Times[ 960, 
              Power[ h, 6], 
              Power[ M, 2]], 
            Times[ -960, 
              Power[ b, 2], 
              Power[ h, 6], 
              Power[ M, 2]], 
            Times[ -1920, 
              Power[ f, 2], 
              Power[ h, 6], 
              Power[ M, 2]], 
            Times[ 1920, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ h, 6], 
              Power[ M, 2]], 
            Times[ 960, 
              Power[ f, 4], 
              Power[ h, 6], 
              Power[ M, 2]], 
            Times[ -960, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ h, 6], 
              Power[ M, 2]], 
            Times[ -240, 
              Power[ h, 8], 
              Power[ M, 2]], 
            Times[ 240, 
              Power[ b, 2], 
              Power[ h, 8], 
              Power[ M, 2]], 
            Times[ 480, 
              Power[ f, 2], 
              Power[ h, 8], 
              Power[ M, 2]], 
            Times[ -480, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ h, 8], 
              Power[ M, 2]], 
            Times[ -240, 
              Power[ f, 4], 
              Power[ h, 8], 
              Power[ M, 2]], 
            Times[ 240, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ h, 8], 
              Power[ M, 2]], 
            Times[ 4, M, 
              Power[ r, 3]], 
            Times[ -4, 
              Power[ b, 2], M, 
              Power[ r, 3]], 
            Times[ -12, 
              Power[ f, 2], M, 
              Power[ r, 3]], 
            Times[ 12, 
              Power[ b, 2], 
              Power[ f, 2], M, 
              Power[ r, 3]], 
            Times[ 8, 
              Power[ f, 4], M, 
              Power[ r, 3]], 
            Times[ -8, 
              Power[ b, 2], 
              Power[ f, 4], M, 
              Power[ r, 3]], 
            Times[ -36, 
              Power[ h, 2], M, 
              Power[ r, 3]], 
            Times[ 36, 
              Power[ b, 2], 
              Power[ h, 2], M, 
              Power[ r, 3]], 
            Times[ 80, 
              Power[ f, 2], 
              Power[ h, 2], M, 
              Power[ r, 3]], 
            Times[ -80, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ h, 2], M, 
              Power[ r, 3]], 
            Times[ -44, 
              Power[ f, 4], 
              Power[ h, 2], M, 
              Power[ r, 3]], 
            Times[ 44, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ h, 2], M, 
              Power[ r, 3]], 
            Times[ 80, 
              Power[ h, 4], M, 
              Power[ r, 3]], 
            Times[ -80, 
              Power[ b, 2], 
              Power[ h, 4], M, 
              Power[ r, 3]], 
            Times[ -164, 
              Power[ f, 2], 
              Power[ h, 4], M, 
              Power[ r, 3]], 
            Times[ 164, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ h, 4], M, 
              Power[ r, 3]], 
            Times[ 84, 
              Power[ f, 4], 
              Power[ h, 4], M, 
              Power[ r, 3]], 
            Times[ -84, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ h, 4], M, 
              Power[ r, 3]], 
            Times[ -68, 
              Power[ h, 6], M, 
              Power[ r, 3]], 
            Times[ 68, 
              Power[ b, 2], 
              Power[ h, 6], M, 
              Power[ r, 3]], 
            Times[ 136, 
              Power[ f, 2], 
              Power[ h, 6], M, 
              Power[ r, 3]], 
            Times[ -136, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ h, 6], M, 
              Power[ r, 3]], 
            Times[ -68, 
              Power[ f, 4], 
              Power[ h, 6], M, 
              Power[ r, 3]], 
            Times[ 68, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ h, 6], M, 
              Power[ r, 3]], 
            Times[ 20, 
              Power[ h, 8], M, 
              Power[ r, 3]], 
            Times[ -20, 
              Power[ b, 2], 
              Power[ h, 8], M, 
              Power[ r, 3]], 
            Times[ -40, 
              Power[ f, 2], 
              Power[ h, 8], M, 
              Power[ r, 3]], 
            Times[ 40, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ h, 8], M, 
              Power[ r, 3]], 
            Times[ 20, 
              Power[ f, 4], 
              Power[ h, 8], M, 
              Power[ r, 3]], 
            Times[ -20, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ h, 8], M, 
              Power[ r, 3]], 
            Times[ -3, 
              Power[ r, 6]], 
            Times[ 3, 
              Power[ b, 2], 
              Power[ r, 6]], 
            Times[ 4, 
              Power[ f, 2], 
              Power[ r, 6]], 
            Times[ -6, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ r, 6]], 
            Times[ -2, 
              Power[ f, 4], 
              Power[ r, 6]], 
            Times[ 4, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ r, 6]], 
            Times[ 6, 
              Power[ h, 2], 
              Power[ r, 6]], 
            Times[ -8, 
              Power[ b, 2], 
              Power[ h, 2], 
              Power[ r, 6]], 
            Times[ -16, 
              Power[ f, 2], 
              Power[ h, 2], 
              Power[ r, 6]], 
            Times[ 24, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ h, 2], 
              Power[ r, 6]], 
            Times[ 8, 
              Power[ f, 4], 
              Power[ h, 2], 
              Power[ r, 6]], 
            Times[ -14, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ h, 2], 
              Power[ r, 6]], 
            Times[ -15, 
              Power[ h, 4], 
              Power[ r, 6]], 
            Times[ 19, 
              Power[ b, 2], 
              Power[ h, 4], 
              Power[ r, 6]], 
            Times[ 30, 
              Power[ f, 2], 
              Power[ h, 4], 
              Power[ r, 6]], 
            Times[ -40, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ h, 4], 
              Power[ r, 6]], 
            Times[ -15, 
              Power[ f, 4], 
              Power[ h, 4], 
              Power[ r, 6]], 
            Times[ 21, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ h, 4], 
              Power[ r, 6]], 
            Times[ 14, 
              Power[ h, 6], 
              Power[ r, 6]], 
            Times[ -16, 
              Power[ b, 2], 
              Power[ h, 6], 
              Power[ r, 6]], 
            Times[ -28, 
              Power[ f, 2], 
              Power[ h, 6], 
              Power[ r, 6]], 
            Times[ 32, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ h, 6], 
              Power[ r, 6]], 
            Times[ 14, 
              Power[ f, 4], 
              Power[ h, 6], 
              Power[ r, 6]], 
            Times[ -16, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ h, 6], 
              Power[ r, 6]], 
            Times[ -5, 
              Power[ h, 8], 
              Power[ r, 6]], 
            Times[ 5, 
              Power[ b, 2], 
              Power[ h, 8], 
              Power[ r, 6]], 
            Times[ 10, 
              Power[ f, 2], 
              Power[ h, 8], 
              Power[ r, 6]], 
            Times[ -10, 
              Power[ b, 2], 
              Power[ f, 2], 
              Power[ h, 8], 
              Power[ r, 6]], 
            Times[ -5, 
              Power[ f, 4], 
              Power[ h, 8], 
              Power[ r, 6]], 
            Times[ 5, 
              Power[ b, 2], 
              Power[ f, 4], 
              Power[ h, 8], 
              Power[ r, 6]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]18\[InvisibleSpace]":"\
\[InvisibleSpace]34\[InvisibleSpace]":"\[InvisibleSpace]16\[InvisibleSpace]" \
on "\[InvisibleSpace]3\[InvisibleSpace]"/"\[InvisibleSpace]15\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 18, ":", 34, ":", 16, " on ", 3, "/", 15, "/", 2001, "."],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("(DR)^2 = "\[InvisibleSpace]\((4\ \((2550\ M\^3 - 
                  7650\ f\^2\ M\^3 + 7650\ f\^4\ M\^3 - 2550\ f\^6\ M\^3 - 
                  15300\ h\^2\ M\^3 + 45900\ f\^2\ h\^2\ M\^3 - 
                  45900\ f\^4\ h\^2\ M\^3 + 15300\ f\^6\ h\^2\ M\^3 + 
                  38250\ h\^4\ M\^3 - 114750\ f\^2\ h\^4\ M\^3 + 
                  114750\ f\^4\ h\^4\ M\^3 - 38250\ f\^6\ h\^4\ M\^3 - 
                  51000\ h\^6\ M\^3 + 153000\ f\^2\ h\^6\ M\^3 - 
                  153000\ f\^4\ h\^6\ M\^3 + 51000\ f\^6\ h\^6\ M\^3 + 
                  38250\ h\^8\ M\^3 - 114750\ f\^2\ h\^8\ M\^3 + 
                  114750\ f\^4\ h\^8\ M\^3 - 38250\ f\^6\ h\^8\ M\^3 - 
                  15300\ h\^10\ M\^3 + 45900\ f\^2\ h\^10\ M\^3 - 
                  45900\ f\^4\ h\^10\ M\^3 + 15300\ f\^6\ h\^10\ M\^3 + 
                  2550\ h\^12\ M\^3 - 7650\ f\^2\ h\^12\ M\^3 + 
                  7650\ f\^4\ h\^12\ M\^3 - 2550\ f\^6\ h\^12\ M\^3 - 
                  1435\ M\^2\ r\^3 + 4465\ f\^2\ M\^2\ r\^3 - 
                  4625\ f\^4\ M\^2\ r\^3 + 1595\ f\^6\ M\^2\ r\^3 + 
                  9090\ h\^2\ M\^2\ r\^3 - 27910\ f\^2\ h\^2\ M\^2\ r\^3 + 
                  28550\ f\^4\ h\^2\ M\^2\ r\^3 - 
                  9730\ f\^6\ h\^2\ M\^2\ r\^3 - 23765\ h\^4\ M\^2\ r\^3 + 
                  72255\ f\^2\ h\^4\ M\^2\ r\^3 - 
                  73215\ f\^4\ h\^4\ M\^2\ r\^3 + 
                  24725\ f\^6\ h\^4\ M\^2\ r\^3 + 32860\ h\^6\ M\^2\ r\^3 - 
                  99220\ f\^2\ h\^6\ M\^2\ r\^3 + 
                  99860\ f\^4\ h\^6\ M\^2\ r\^3 - 
                  33500\ f\^6\ h\^6\ M\^2\ r\^3 - 25365\ h\^8\ M\^2\ r\^3 + 
                  76255\ f\^2\ h\^8\ M\^2\ r\^3 - 
                  76415\ f\^4\ h\^8\ M\^2\ r\^3 + 
                  25525\ f\^6\ h\^8\ M\^2\ r\^3 + 10370\ h\^10\ M\^2\ r\^3 - 
                  31110\ f\^2\ h\^10\ M\^2\ r\^3 + 
                  31110\ f\^4\ h\^10\ M\^2\ r\^3 - 
                  10370\ f\^6\ h\^10\ M\^2\ r\^3 - 1755\ h\^12\ M\^2\ r\^3 + 
                  5265\ f\^2\ h\^12\ M\^2\ r\^3 - 
                  5265\ f\^4\ h\^12\ M\^2\ r\^3 + 
                  1755\ f\^6\ h\^12\ M\^2\ r\^3 + 128\ M\ r\^6 - 
                  432\ f\^2\ M\ r\^6 + 496\ f\^4\ M\ r\^6 - 
                  192\ f\^6\ M\ r\^6 - 912\ h\^2\ M\ r\^6 + 
                  3024\ f\^2\ h\^2\ M\ r\^6 - 3344\ f\^4\ h\^2\ M\ r\^6 + 
                  1232\ f\^6\ h\^2\ M\ r\^6 + 2736\ h\^4\ M\ r\^6 - 
                  8752\ f\^2\ h\^4\ M\ r\^6 + 9312\ f\^4\ h\^4\ M\ r\^6 - 
                  3296\ f\^6\ h\^4\ M\ r\^6 - 4288\ h\^6\ M\ r\^6 + 
                  13280\ f\^2\ h\^6\ M\ r\^6 - 13696\ f\^4\ h\^6\ M\ r\^6 + 
                  4704\ f\^6\ h\^6\ M\ r\^6 + 3664\ h\^8\ M\ r\^6 - 
                  11104\ f\^2\ h\^8\ M\ r\^6 + 11216\ f\^4\ h\^8\ M\ r\^6 - 
                  3776\ f\^6\ h\^8\ M\ r\^6 - 1616\ h\^10\ M\ r\^6 + 
                  4848\ f\^2\ h\^10\ M\ r\^6 - 4848\ f\^4\ h\^10\ M\ r\^6 + 
                  1616\ f\^6\ h\^10\ M\ r\^6 + 288\ h\^12\ M\ r\^6 - 
                  864\ f\^2\ h\^12\ M\ r\^6 + 864\ f\^4\ h\^12\ M\ r\^6 - 
                  288\ f\^6\ h\^12\ M\ r\^6 - 24\ r\^9 + 52\ f\^2\ r\^9 - 
                  48\ f\^4\ r\^9 + 16\ f\^6\ r\^9 + 60\ h\^2\ r\^9 - 
                  244\ f\^2\ h\^2\ r\^9 + 268\ f\^4\ h\^2\ r\^9 - 
                  96\ f\^6\ h\^2\ r\^9 - 184\ h\^4\ r\^9 + 
                  624\ f\^2\ h\^4\ r\^9 - 688\ f\^4\ h\^4\ r\^9 + 
                  248\ f\^6\ h\^4\ r\^9 + 300\ h\^6\ r\^9 - 
                  952\ f\^2\ h\^6\ r\^9 + 1004\ f\^4\ h\^6\ r\^9 - 
                  352\ f\^6\ h\^6\ r\^9 - 272\ h\^8\ r\^9 + 
                  832\ f\^2\ h\^8\ r\^9 - 848\ f\^4\ h\^8\ r\^9 + 
                  288\ f\^6\ h\^8\ r\^9 + 128\ h\^10\ r\^9 - 
                  384\ f\^2\ h\^10\ r\^9 + 384\ f\^4\ h\^10\ r\^9 - 
                  128\ f\^6\ h\^10\ r\^9 - 24\ h\^12\ r\^9 + 
                  72\ f\^2\ h\^12\ r\^9 - 72\ f\^4\ h\^12\ r\^9 + 
                  24\ f\^6\ h\^12\ r\^9)\))\)/\((\((\(-1\) + f)\)\^3\ \((1 + \
f)\)\^3\ \((\(-1\) + h)\)\^6\ \((1 + h)\)\^6\ r\^15)\)\),
      SequenceForm[ "(DR)^2 = ", 
        Times[ 4, 
          Power[ 
            Plus[ -1, f], -3], 
          Power[ 
            Plus[ 1, f], -3], 
          Power[ 
            Plus[ -1, h], -6], 
          Power[ 
            Plus[ 1, h], -6], 
          Power[ r, -15], 
          Plus[ 
            Times[ 2550, 
              Power[ M, 3]], 
            Times[ -7650, 
              Power[ f, 2], 
              Power[ M, 3]], 
            Times[ 7650, 
              Power[ f, 4], 
              Power[ M, 3]], 
            Times[ -2550, 
              Power[ f, 6], 
              Power[ M, 3]], 
            Times[ -15300, 
              Power[ h, 2], 
              Power[ M, 3]], 
            Times[ 45900, 
              Power[ f, 2], 
              Power[ h, 2], 
              Power[ M, 3]], 
            Times[ -45900, 
              Power[ f, 4], 
              Power[ h, 2], 
              Power[ M, 3]], 
            Times[ 15300, 
              Power[ f, 6], 
              Power[ h, 2], 
              Power[ M, 3]], 
            Times[ 38250, 
              Power[ h, 4], 
              Power[ M, 3]], 
            Times[ -114750, 
              Power[ f, 2], 
              Power[ h, 4], 
              Power[ M, 3]], 
            Times[ 114750, 
              Power[ f, 4], 
              Power[ h, 4], 
              Power[ M, 3]], 
            Times[ -38250, 
              Power[ f, 6], 
              Power[ h, 4], 
              Power[ M, 3]], 
            Times[ -51000, 
              Power[ h, 6], 
              Power[ M, 3]], 
            Times[ 153000, 
              Power[ f, 2], 
              Power[ h, 6], 
              Power[ M, 3]], 
            Times[ -153000, 
              Power[ f, 4], 
              Power[ h, 6], 
              Power[ M, 3]], 
            Times[ 51000, 
              Power[ f, 6], 
              Power[ h, 6], 
              Power[ M, 3]], 
            Times[ 38250, 
              Power[ h, 8], 
              Power[ M, 3]], 
            Times[ -114750, 
              Power[ f, 2], 
              Power[ h, 8], 
              Power[ M, 3]], 
            Times[ 114750, 
              Power[ f, 4], 
              Power[ h, 8], 
              Power[ M, 3]], 
            Times[ -38250, 
              Power[ f, 6], 
              Power[ h, 8], 
              Power[ M, 3]], 
            Times[ -15300, 
              Power[ h, 10], 
              Power[ M, 3]], 
            Times[ 45900, 
              Power[ f, 2], 
              Power[ h, 10], 
              Power[ M, 3]], 
            Times[ -45900, 
              Power[ f, 4], 
              Power[ h, 10], 
              Power[ M, 3]], 
            Times[ 15300, 
              Power[ f, 6], 
              Power[ h, 10], 
              Power[ M, 3]], 
            Times[ 2550, 
              Power[ h, 12], 
              Power[ M, 3]], 
            Times[ -7650, 
              Power[ f, 2], 
              Power[ h, 12], 
              Power[ M, 3]], 
            Times[ 7650, 
              Power[ f, 4], 
              Power[ h, 12], 
              Power[ M, 3]], 
            Times[ -2550, 
              Power[ f, 6], 
              Power[ h, 12], 
              Power[ M, 3]], 
            Times[ -1435, 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ 4465, 
              Power[ f, 2], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ -4625, 
              Power[ f, 4], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ 1595, 
              Power[ f, 6], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ 9090, 
              Power[ h, 2], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ -27910, 
              Power[ f, 2], 
              Power[ h, 2], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ 28550, 
              Power[ f, 4], 
              Power[ h, 2], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ -9730, 
              Power[ f, 6], 
              Power[ h, 2], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ -23765, 
              Power[ h, 4], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ 72255, 
              Power[ f, 2], 
              Power[ h, 4], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ -73215, 
              Power[ f, 4], 
              Power[ h, 4], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ 24725, 
              Power[ f, 6], 
              Power[ h, 4], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ 32860, 
              Power[ h, 6], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ -99220, 
              Power[ f, 2], 
              Power[ h, 6], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ 99860, 
              Power[ f, 4], 
              Power[ h, 6], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ -33500, 
              Power[ f, 6], 
              Power[ h, 6], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ -25365, 
              Power[ h, 8], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ 76255, 
              Power[ f, 2], 
              Power[ h, 8], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ -76415, 
              Power[ f, 4], 
              Power[ h, 8], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ 25525, 
              Power[ f, 6], 
              Power[ h, 8], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ 10370, 
              Power[ h, 10], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ -31110, 
              Power[ f, 2], 
              Power[ h, 10], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ 31110, 
              Power[ f, 4], 
              Power[ h, 10], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ -10370, 
              Power[ f, 6], 
              Power[ h, 10], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ -1755, 
              Power[ h, 12], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ 5265, 
              Power[ f, 2], 
              Power[ h, 12], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ -5265, 
              Power[ f, 4], 
              Power[ h, 12], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ 1755, 
              Power[ f, 6], 
              Power[ h, 12], 
              Power[ M, 2], 
              Power[ r, 3]], 
            Times[ 128, M, 
              Power[ r, 6]], 
            Times[ -432, 
              Power[ f, 2], M, 
              Power[ r, 6]], 
            Times[ 496, 
              Power[ f, 4], M, 
              Power[ r, 6]], 
            Times[ -192, 
              Power[ f, 6], M, 
              Power[ r, 6]], 
            Times[ -912, 
              Power[ h, 2], M, 
              Power[ r, 6]], 
            Times[ 3024, 
              Power[ f, 2], 
              Power[ h, 2], M, 
              Power[ r, 6]], 
            Times[ -3344, 
              Power[ f, 4], 
              Power[ h, 2], M, 
              Power[ r, 6]], 
            Times[ 1232, 
              Power[ f, 6], 
              Power[ h, 2], M, 
              Power[ r, 6]], 
            Times[ 2736, 
              Power[ h, 4], M, 
              Power[ r, 6]], 
            Times[ -8752, 
              Power[ f, 2], 
              Power[ h, 4], M, 
              Power[ r, 6]], 
            Times[ 9312, 
              Power[ f, 4], 
              Power[ h, 4], M, 
              Power[ r, 6]], 
            Times[ -3296, 
              Power[ f, 6], 
              Power[ h, 4], M, 
              Power[ r, 6]], 
            Times[ -4288, 
              Power[ h, 6], M, 
              Power[ r, 6]], 
            Times[ 13280, 
              Power[ f, 2], 
              Power[ h, 6], M, 
              Power[ r, 6]], 
            Times[ -13696, 
              Power[ f, 4], 
              Power[ h, 6], M, 
              Power[ r, 6]], 
            Times[ 4704, 
              Power[ f, 6], 
              Power[ h, 6], M, 
              Power[ r, 6]], 
            Times[ 3664, 
              Power[ h, 8], M, 
              Power[ r, 6]], 
            Times[ -11104, 
              Power[ f, 2], 
              Power[ h, 8], M, 
              Power[ r, 6]], 
            Times[ 11216, 
              Power[ f, 4], 
              Power[ h, 8], M, 
              Power[ r, 6]], 
            Times[ -3776, 
              Power[ f, 6], 
              Power[ h, 8], M, 
              Power[ r, 6]], 
            Times[ -1616, 
              Power[ h, 10], M, 
              Power[ r, 6]], 
            Times[ 4848, 
              Power[ f, 2], 
              Power[ h, 10], M, 
              Power[ r, 6]], 
            Times[ -4848, 
              Power[ f, 4], 
              Power[ h, 10], M, 
              Power[ r, 6]], 
            Times[ 1616, 
              Power[ f, 6], 
              Power[ h, 10], M, 
              Power[ r, 6]], 
            Times[ 288, 
              Power[ h, 12], M, 
              Power[ r, 6]], 
            Times[ -864, 
              Power[ f, 2], 
              Power[ h, 12], M, 
              Power[ r, 6]], 
            Times[ 864, 
              Power[ f, 4], 
              Power[ h, 12], M, 
              Power[ r, 6]], 
            Times[ -288, 
              Power[ f, 6], 
              Power[ h, 12], M, 
              Power[ r, 6]], 
            Times[ -24, 
              Power[ r, 9]], 
            Times[ 52, 
              Power[ f, 2], 
              Power[ r, 9]], 
            Times[ -48, 
              Power[ f, 4], 
              Power[ r, 9]], 
            Times[ 16, 
              Power[ f, 6], 
              Power[ r, 9]], 
            Times[ 60, 
              Power[ h, 2], 
              Power[ r, 9]], 
            Times[ -244, 
              Power[ f, 2], 
              Power[ h, 2], 
              Power[ r, 9]], 
            Times[ 268, 
              Power[ f, 4], 
              Power[ h, 2], 
              Power[ r, 9]], 
            Times[ -96, 
              Power[ f, 6], 
              Power[ h, 2], 
              Power[ r, 9]], 
            Times[ -184, 
              Power[ h, 4], 
              Power[ r, 9]], 
            Times[ 624, 
              Power[ f, 2], 
              Power[ h, 4], 
              Power[ r, 9]], 
            Times[ -688, 
              Power[ f, 4], 
              Power[ h, 4], 
              Power[ r, 9]], 
            Times[ 248, 
              Power[ f, 6], 
              Power[ h, 4], 
              Power[ r, 9]], 
            Times[ 300, 
              Power[ h, 6], 
              Power[ r, 9]], 
            Times[ -952, 
              Power[ f, 2], 
              Power[ h, 6], 
              Power[ r, 9]], 
            Times[ 1004, 
              Power[ f, 4], 
              Power[ h, 6], 
              Power[ r, 9]], 
            Times[ -352, 
              Power[ f, 6], 
              Power[ h, 6], 
              Power[ r, 9]], 
            Times[ -272, 
              Power[ h, 8], 
              Power[ r, 9]], 
            Times[ 832, 
              Power[ f, 2], 
              Power[ h, 8], 
              Power[ r, 9]], 
            Times[ -848, 
              Power[ f, 4], 
              Power[ h, 8], 
              Power[ r, 9]], 
            Times[ 288, 
              Power[ f, 6], 
              Power[ h, 8], 
              Power[ r, 9]], 
            Times[ 128, 
              Power[ h, 10], 
              Power[ r, 9]], 
            Times[ -384, 
              Power[ f, 2], 
              Power[ h, 10], 
              Power[ r, 9]], 
            Times[ 384, 
              Power[ f, 4], 
              Power[ h, 10], 
              Power[ r, 9]], 
            Times[ -128, 
              Power[ f, 6], 
              Power[ h, 10], 
              Power[ r, 9]], 
            Times[ -24, 
              Power[ h, 12], 
              Power[ r, 9]], 
            Times[ 72, 
              Power[ f, 2], 
              Power[ h, 12], 
              Power[ r, 9]], 
            Times[ -72, 
              Power[ f, 4], 
              Power[ h, 12], 
              Power[ r, 9]], 
            Times[ 24, 
              Power[ f, 6], 
              Power[ h, 12], 
              Power[ r, 9]]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]18\[InvisibleSpace]":"\
\[InvisibleSpace]36\[InvisibleSpace]":"\[InvisibleSpace]15\[InvisibleSpace]" \
on "\[InvisibleSpace]3\[InvisibleSpace]"/"\[InvisibleSpace]15\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 18, ":", 36, ":", 15, " on ", 3, "/", 15, "/", 2001, "."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(timestamp[];\)\)], "Input"],

Cell[BoxData[
    InterpretationBox[\("Time is "\[InvisibleSpace]18\[InvisibleSpace]":"\
\[InvisibleSpace]36\[InvisibleSpace]":"\[InvisibleSpace]15\[InvisibleSpace]" \
on "\[InvisibleSpace]3\[InvisibleSpace]"/"\[InvisibleSpace]15\[InvisibleSpace]\
"/"\[InvisibleSpace]2001\[InvisibleSpace]"."\),
      SequenceForm[ 
      "Time is ", 18, ":", 36, ":", 15, " on ", 3, "/", 15, "/", 2001, "."],
      Editable->False]], "Print"]
}, Open  ]],

Cell[BoxData[
    \(\(NotebookSave[EvaluationNotebook[]];\)\)], "Input"]
},
FrontEndVersion->"4.0 for X",
ScreenRectangle->{{0, 800}, {0, 600}},
CellGrouping->Manual,
WindowSize->{734, 504},
WindowMargins->{{Automatic, 2}, {Automatic, 0}},
ShowCellLabel->False
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1739, 51, 64, 1, 27, "Input"],
Cell[1806, 54, 2970, 44, 375, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[4813, 103, 2294, 42, 475, "Input"],
Cell[7110, 147, 147, 3, 23, "Print"],
Cell[7260, 152, 177, 3, 23, "Print"],
Cell[7440, 157, 213, 4, 23, "Print"],
Cell[7656, 163, 339, 9, 37, "Print"],
Cell[7998, 174, 387, 11, 42, "Print"],
Cell[8388, 187, 367, 9, 25, "Print"],
Cell[8758, 198, 320, 8, 40, "Print"],
Cell[9081, 208, 427, 7, 23, "Print"],
Cell[9511, 217, 427, 7, 23, "Print"],
Cell[9941, 226, 221, 6, 42, "Print"],
Cell[10165, 234, 427, 7, 23, "Print"],
Cell[10595, 243, 330, 10, 42, "Print"],
Cell[10928, 255, 425, 7, 23, "Print"],
Cell[11356, 264, 147, 3, 23, "Print"],
Cell[11506, 269, 177, 3, 23, "Print"],
Cell[11686, 274, 219, 4, 23, "Print"],
Cell[11908, 280, 393, 11, 42, "Print"],
Cell[12304, 293, 387, 11, 42, "Print"],
Cell[12694, 306, 248, 5, 25, "Print"],
Cell[12945, 313, 426, 11, 25, "Print"],
Cell[13374, 326, 323, 8, 40, "Print"],
Cell[13700, 336, 425, 7, 23, "Print"],
Cell[14128, 345, 425, 7, 23, "Print"],
Cell[14556, 354, 4107, 126, 96, "Print"],
Cell[18666, 482, 427, 7, 23, "Print"],
Cell[19096, 491, 12454, 375, 276, "Print"],
Cell[31553, 868, 427, 7, 23, "Print"],
Cell[31983, 877, 147, 3, 23, "Print"],
Cell[32133, 882, 177, 3, 23, "Print"],
Cell[32313, 887, 225, 4, 23, "Print"],
Cell[32541, 893, 393, 11, 42, "Print"],
Cell[32937, 906, 387, 11, 42, "Print"],
Cell[33327, 919, 438, 12, 42, "Print"],
Cell[33768, 933, 367, 9, 25, "Print"],
Cell[34138, 944, 514, 14, 25, "Print"],
Cell[34655, 960, 323, 8, 40, "Print"],
Cell[34981, 970, 427, 7, 23, "Print"],
Cell[35411, 979, 427, 7, 23, "Print"],
Cell[35841, 988, 13789, 420, 294, "Print"],
Cell[49633, 1410, 427, 7, 23, "Print"],
Cell[50063, 1419, 16940, 507, 384, "Print"],
Cell[67006, 1928, 427, 7, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[67470, 1940, 49, 1, 27, "Input"],
Cell[67522, 1943, 427, 7, 23, "Print"]
}, Open  ]],
Cell[67964, 1953, 72, 1, 27, "Input"]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

